package gov.fnal.eag.matrix;
import com.imsl.math.Complex;
/** QR Decomposition.
<P>
   For an m-by-n matrix A with m >= n, the QR decomposition is an m-by-n
   orthogonal matrix Q and an n-by-n upper triangular matrix R so that
   A = Q*R.
<P>
   The QR decompostion always exists, even if the matrix does not have
   full rank, so the constructor will never fail.  The primary use of the
   QR decomposition is in the least squares solution of nonsquare systems
   of simultaneous linear equations.  This will fail if isFullRank()
   returns false.
*/

public class QRDecomposition implements java.io.Serializable {

/* ------------------------
   Class variables
 * ------------------------ */

   /**
	 * 
	 */
	private static final long serialVersionUID = -8935674121184118052L;

/** Array for internal storage of decomposition.
   @serial internal array storage.
   */
   private Complex[][] QR;

   /** Row and column dimensions.
   @serial column dimension.
   @serial row dimension.
   */
   private int m, n;

   /** Array for internal storage of diagonal of R.
   @serial diagonal of R.
   */
   private Complex[] Rdiag;
   
   /**
    *  Some complex constants
    * 
    */
   private final static Complex czero = new Complex(0.,0.);
   private final static Complex cone = new Complex(1.0,0.0);

/* ------------------------
   Constructor
 * ------------------------ */

   /** QR Decomposition, computed by Householder reflections.
   @param A    Rectangular matrix
     Create   structure to access R and the Householder vectors and compute Q.
   */

   public QRDecomposition (Matrix A) {
      // Initialize.
      QR = A.getArrayCopy();       // fill QR
      m = A.getRowDimension();     // # rows
      n = A.getColumnDimension();  // # cols
      Rdiag = new Complex[n];		// fill diagonal with zeroes

      // Main loop.
      for (int k = 0; k < n; k++) {  // loop on cols
         // Compute 2-norm of k-th column without under/overflow.
          Complex nrm = czero;
          for (int i = k; i < m; i++) {
              nrm = Complex.sqrt(Complex.plus((Complex.times(nrm, nrm)),
                      Complex.times(QR[i][k],QR[i][k])));
          }

         if (!nrm.equals(czero) ) {
            // Form k-th Householder vector.
 //           if ((QR[k][k]).real() < 0) {
 //              nrm = Complex.negative(nrm);
 //           }
            for (int i = k; i < m; i++) {
                QR[i][k] = Complex.over(QR[i][k],nrm);
            }
                QR[k][k] =Complex.plus(cone,QR[k][k]);
            // Apply transformation to remaining columns.
            for (int j = k+1; j < n; j++) {
//               Complex s = new Complex(czero); 
                Complex s = czero;
               for (int i = k; i < m; i++) {
                   s = Complex.plus(s, Complex.times(QR[i][k],QR[i][j]));
               }
               s = Complex.negative(Complex.over(s, QR[k][k]));
               for (int i = k; i < m; i++) {
                   QR[i][j] = Complex.plus(QR[i][j],Complex.times(s,QR[i][k]));
               }
            }
         }
         Rdiag[k] =Complex.negative(nrm);
      }
   }

/* ------------------------
   Public Methods
 * ------------------------ */

   /** Is the matrix full rank?
   @return     true if R, and hence A, has full rank.
   */

   public boolean isFullRank () {
      for (int j = 0; j < n; j++) {
         if (Rdiag[j].equals(czero))
            return false;
      }
      return true;
   }

   /*
   /** Return the Householder vectors
   @return     Lower trapezoidal matrix whose columns define the reflections
   */

   /*
   public Matrix getH () {
      Matrix X = new Matrix(m,n);
      Complex[][] H = X.getArray();
      for (int i = 0; i < m; i++) {
         for (int j = 0; j < n; j++) {
            if (i >= j) {
               H[i][j] = QR[i][j];
            } else {
               H[i][j] = czero;
            }
         }
      }
      return X;
   }
   */
   
   /** Return the upper triangular factor
   @return     R
   */

   /*
   public Matrix getR () {
      Matrix X = new Matrix(n,n);
      Complex[][] R = X.getArray();
      for (int i = 0; i < n; i++) {
         for (int j = 0; j < n; j++) {
            if (i < j) {
               R[i][j] = QR[i][j];
            } else if (i == j) {
               R[i][j] = Rdiag[i];
            } else {
               R[i][j] = czero;
            }
         }
      }
      return X;
   }
	*/
   /** Generate and return the (economy-sized) orthogonal factor
   @return     Q
   */

   /*
   public Matrix getQ () {
      Matrix X = new Matrix(m,n);
      Complex[][] Q = X.getArray();
      for (int k = n-1; k >= 0; k--) {
         for (int i = 0; i < m; i++) {
            Q[i][k] = czero;
         }
         Q[k][k] = cone;
         for (int j = k; j < n; j++) {
            if (!QR[k][k].equals(czero)) {
                Complex s = czero;
                for (int i = k; i < m; i++) {
                    s = Complex.plus(s,Complex.times(QR[i][k],Q[i][j]));
                }
                s = Complex.negative(Complex.over(s,QR[k][k]));
                for (int i = k; i < m; i++) {
                    Q[i][j] = Complex.plus(Q[i][j],Complex.times(s,QR[i][k])); 
                }
            }
         }
      }
      return X;
   }
	*/
   /** Least squares solution of A*X = B
   @param B    A Matrix with as many rows as A and any number of columns.
   @return     X that minimizes the two norm of Q*R*X-B.
   @exception  IllegalArgumentException  Matrix row dimensions must agree.
   @exception  RuntimeException  Matrix is rank deficient.
   */

   /*
   public Matrix solve (Matrix B) {
      if (B.getRowDimension() != m) {
         throw new IllegalArgumentException("Matrix row dimensions must agree.");
      }
      if (!this.isFullRank()) {
         throw new RuntimeException("Matrix is rank deficient.");
      }
      
      // Copy right hand side
      int nx = B.getColumnDimension();
      Complex[][] X = B.getArrayCopy();

      // Compute Y = transpose(Q)*B
      for (int k = 0; k < n; k++) {
         for (int j = 0; j < nx; j++) { 
             Complex s = czero;
             for (int i = k; i < m; i++) {
                 s = Complex.plus(s,Complex.times(QR[i][k],X[i][j]));
             }
             s = Complex.negative(Complex.over(s,QR[k][k]));
             for (int i = k; i < m; i++) {
                 X[i][j] = Complex.plus(X[i][j],Complex.times(s,QR[i][k]));
             }
         }
      }
      // Solve R*X = Y;
      for (int k = n-1; k >= 0; k--) {
         for (int j = 0; j < nx; j++) {
             X[k][j] = Complex.over(X[k][j],Rdiag[k]);
         }
         for (int i = 0; i < k; i++) {
            for (int j = 0; j < nx; j++) {
                X[i][j] = Complex.minus(X[i][j],Complex.times(X[k][j],QR[i][k]));
            }
         }
      }
      return (new Matrix(X,n,nx).getMatrix(0,n-1,0,nx-1));
   }
   */
}
