package gov.fnal.eag.matrix;
import com.imsl.math.Complex;


   /** Singular Value Decomposition.
   <P>
   For an m-by-n matrix A with m >= n, the singular value decomposition is
   an m-by-n orthogonal matrix U, an n-by-n diagonal matrix S, and
   an n-by-n orthogonal matrix V so that A = U*S*V'.
   <P>
   The singular values, sigma[k] = S[k][k], are ordered so that
   sigma[0] >= sigma[1] >= ... >= sigma[n-1].
   <P>
   The singular value decompostion always exists, so the constructor will
   never fail.  The matrix condition number and the effective numerical
   rank can be computed from this decomposition.
   */

public class SingularValueDecomposition implements java.io.Serializable {

/* ------------------------
   Class variables
 * ------------------------ */

   /**
	 * 
	 */
	private static final long serialVersionUID = -7910123784666116879L;

/** Arrays for internal storage of U and V.
   @serial internal storage of U.
   @serial internal storage of V.
   */
   private Complex[][] U, V;

   /** Array for internal storage of singular values.
   @serial internal storage of singular values.
   */
   private Complex[] s;

   /** Row and column dimensions.
   @serial row dimension.
   @serial column dimension.
   */
   private int m, n;
   
   /**
    * 
    * Some complex constants
    */
   private final static Complex czero = new Complex(0.,0.);
   private final static Complex cone = new Complex(1.0,0.0);

/* ------------------------
   Constructor
 * ------------------------ */

   /** Construct the singular value decomposition
   @param Arg    Rectangular matrix
   */

   public SingularValueDecomposition (Matrix Arg) {

      // Derived from LINPACK code.
      // Initialize.
      Complex[][] A = Arg.getArrayCopy();
      m = Arg.getRowDimension();
      n = Arg.getColumnDimension();
      int nu = Math.min(m,n);
      s = new Complex [Math.min(m+1,n)];
//
      for (int i=0 ; i<Math.min(m+1,n); i++) s[i] = czero;
      U = new Complex [m][nu];
      for (int i=0; i<m; i++) {
      	for (int j=0; j<nu; j++) {
//
          U[i][j] = czero;
      	}
      }
      V = new Complex [n][n];
      for (int i=0; i<n; i++) {
      	for (int j=0; j<n; j++) {
//
          V[i][j] = czero;
      	}
      }
      Complex[] e = new Complex [n];
//
      for (int i=0; i<n; i++) e[i] = czero;
      Complex[] work = new Complex [m];
//
      for (int i=0; i<m; i++) work[i] = czero;
      boolean wantu = true;
      boolean wantv = true;

      // Reduce A to bidiagonal form, storing the diagonal elements
      // in s and the super-diagonal elements in e.

      int nct = Math.min(m-1,n);
      int nrt = Math.max(0,Math.min(n-2,m));
      for (int k = 0; k < Math.max(nct,nrt); k++) {
         if (k < nct) {

            // Compute the transformation for the k-th column and
            // place the k-th diagonal in s[k].
            // Compute 2-norm of k-th column without under/overflow.
//            s[k] = new Complex(0.,0.);
             s[k] = czero;
             for (int i = k; i < m; i++) {
//
                 s[k] = Complex.sqrt(Complex.plus(Complex.times(s[k],s[k]),
                         (Complex.times(A[i][k],A[i][k]))));  // hipotenuza
             }
            
//            System.out.println("s[k]="+s[k].toString());
            
            if (!s[k].equals(czero)) {
//               if (A[k][k] < 0.0) {
//                  s[k] = -s[k];
//               }
               for (int i = k; i < m; i++) {
//
                   A[i][k] = Complex.over(A[i][k],s[k]);
               }
//
               A[k][k] = Complex.plus(cone,A[k][k]);
            }
            s[k] = Complex.negative(s[k]);
         }
         for (int j = k+1; j < n; j++) {
            if ((k < nct) & (!s[k].equals(czero)))  {

            // Apply the transformation.

//
                Complex t = czero;
                for (int i = k; i < m; i++) {
//
                    t = Complex.plus(t,Complex.times(A[i][k],A[i][j]));
                }
//
                t = Complex.negative(Complex.over(t,A[k][k] ));
                for (int i = k; i < m; i++) {
//
                    A[i][j] = Complex.plus(A[i][j], Complex.times(t,A[i][k]));
                }
            }

            // Place the k-th row of A into e for the
            // subsequent calculation of the row transformation.

            e[j] = A[k][j];
         }
         if (wantu & (k < nct)) {

            // Place the transformation in U for subsequent back
            // multiplication.

            for (int i = k; i < m; i++) {
               U[i][k] = A[i][k];
            }
         }
         if (k < nrt) {

            // Compute the k-th row transformation and place the
            // k-th super-diagonal in e[k].
            // Compute 2-norm without under/overflow.
//
             e[k] = czero;
            for (int i = k+1; i < n; i++) {
//               e[k] = Maths.hypot(e[k],e[i]);
//
                e[k] = Complex.sqrt(Complex.plus(
                        Complex.times(e[k],e[k]),Complex.times(e[i],e[i])));
            }
            if (!e[k].equals(czero)) {
//               if (e[k+1] < 0.0) {
//                  e[k] = -e[k];
//               }
               for (int i = k+1; i < n; i++) {
//
                   e[i] = Complex.over(e[i],e[k]);
               }
//
               e[k+1] = Complex.plus(cone,e[k+1]);
            }
            e[k] = Complex.negative(e[k]);
            if ((k+1 < m) & (!e[k].equals(czero))) {

            // Apply the transformation.

               for (int i = k+1; i < m; i++) {

                   work[i] = czero;
               }
               for (int j = k+1; j < n; j++) {
                  for (int i = k+1; i < m; i++) {

                      work[i] = Complex.plus(work[i],Complex.times(e[j],A[i][j]));
                  }
               }
               for (int j = k+1; j < n; j++) {
                  Complex t = Complex.negative(e[j].over(e[k+1]));
                  for (int i = k+1; i < m; i++) {

                      A[i][j] = Complex.plus(A[i][j], Complex.times(t, work[i]));
                  }
               }
            }
            if (wantv) {

            // Place the transformation in V for subsequent
            // back multiplication.

               for (int i = k+1; i < n; i++) {
                  V[i][k] = e[i];
               }
            }
         }
      }

      // Set up the final bidiagonal matrix or order p.

      int p = Math.min(n,m+1);
      if (nct < n) {
         s[nct] = A[nct][nct];
      }
      if (m < p) {
         s[p-1] = czero;
      }
      if (nrt+1 < p) {
         e[nrt] = A[nrt][p-1];
      }
      e[p-1] = czero;

      // If required, generate U.

      if (wantu) {
         for (int j = nct; j < nu; j++) {
            for (int i = 0; i < m; i++) {

                U[i][j] = czero;
            }

            U[j][j] = cone;
         }
         for (int k = nct-1; k >= 0; k--) {
            if (!s[k].equals(czero)) {
               for (int j = k+1; j < nu; j++) {

                   Complex t = czero;
                  for (int i = k; i < m; i++) {

                      t = Complex.plus(t,Complex.times(U[i][k],U[i][j] ) );
                  }

                  t = Complex.negative(Complex.over(t,U[k][k]));
                  for (int i = k; i < m; i++) {

                      U[i][j] = Complex.plus(U[i][j],Complex.times(t,U[i][k]));
                  }
               }
               for (int i = k; i < m; i++ ) {
                  U[i][k] = Complex.negative(U[i][k]);
               }

               U[k][k] = Complex.plus(cone,U[k][k] );
               for (int i = 0; i < k-1; i++) {
                  U[i][k] = czero;
               }
            } else {
               for (int i = 0; i < m; i++) {
                  U[i][k] = czero;
               }
               U[k][k] = cone;
            }
         }
      }

      // If required, generate V.

      if (wantv) {
         for (int k = n-1; k >= 0; k--) {
            if ((k < nrt) & (!e[k].equals(czero))) {
               for (int j = k+1; j < nu; j++) {
                   Complex t = czero;
                   for (int i = k+1; i < n; i++) {
                       t = Complex.plus(t, Complex.times(V[i][k], V[i][j]));
                   }

                   t = Complex.negative(Complex.over(t,V[k+1][k] ));
                   for (int i = k+1; i < n; i++) {

                       V[i][j] = Complex.plus(V[i][j],Complex.times(t,V[i][k]));
                   }
               }
            }
            for (int i = 0; i < n; i++) {

                V[i][k] = czero;
            }

            V[k][k] = cone;
         }
      }

      // Main iteration loop for the singular values.

      int pp = p-1;
      int iter = 0;
      double eps = Math.pow(2.0,-52.0);
      while (p > 0) {
         int k,kase;

         // Here is where a test for too many iterations would go.

         // This section of the program inspects for
         // negligible elements in the s and e arrays.  On
         // completion the variables kase and k are set as follows.

         // kase = 1     if s(p) and e[k-1] are negligible and k<p
         // kase = 2     if s(k) is negligible and k<p
         // kase = 3     if e[k-1] is negligible, k<p, and
         //              s(k), ..., s(p) are not negligible (qr step).
         // kase = 4     if e(p-1) is negligible (convergence).

         for (k = p-2; k >= -1; k--) {
            if (k == -1) {
               break;
            }
            if (Complex.abs(e[k]) <= eps*(Complex.abs(s[k]) + Complex.abs(s[k+1]))) {
                e[k] = czero;
               break;
            }
         }
         if (k == p-2) {
            kase = 4;
         } else {
            int ks;
            for (ks = p-1; ks >= k; ks--) {
               if (ks == k) {
                  break;
               }
               double t = (ks != p ? Complex.abs(e[ks]) : 0.) + 
                          (ks != k+1 ? Complex.abs(e[ks-1]) : 0.);
               if (Complex.abs(s[ks]) <= eps*t)  {
                  s[ks] = czero;
                  break;
               }
            }
            if (ks == k) {
               kase = 3;
            } else if (ks == p-1) {
               kase = 1;
            } else {
               kase = 2;
               k = ks;
            }
         }
         k++;

         // Perform the task indicated by kase.

         switch (kase) {

            // Deflate negligible s(p).

            case 1: {
               Complex f = e[p-2];
               e[p-2] = czero;
               for (int j = p-2; j >= k; j--) {
                   Complex t = Complex.sqrt(Complex.plus(Complex.times(s[j],s[j] ),
                           Complex.times(f,f)));

                    Complex cs = Complex.over(s[j], t);

                    Complex sn = Complex.over(f, t);
                  s[j] = t;
                  if (j != k) {

                      f = Complex.negative(Complex.times(sn,e[j-1]));
                     e[j-1] = cs.times(e[j-1]);
                  }
                  if (wantv) {
                     for (int i = 0; i < n; i++) {
                         t = Complex.plus(Complex.times(cs,V[i][j]),
                                 Complex.times(sn,V[i][p-1]));
                         V[i][p-1] = Complex.minus(Complex.times(cs,V[i][p-1]),
                                 Complex.times(sn,V[i][j]));
                         V[i][j] = t;
                     }
                  }
               }
            }
            break;

            // Split at negligible s(k).

            case 2: {
               Complex f = e[k-1];
               e[k-1] = czero;
               for (int j = k; j < p; j++) {
                   Complex t = Complex.sqrt(Complex.plus(Complex.times(s[j],s[j]),
                           Complex.times(f,f)));
                   Complex cs = Complex.over(s[j],t);
                   Complex sn = Complex.over(f,t);
                   s[j] = t;
                   f = Complex.negative(Complex.times(sn,e[j]));
                   e[j] = Complex.times(cs,e[j]);
                  if (wantu) {
                     for (int i = 0; i < m; i++) {
                         t = Complex.plus(Complex.times(cs, U[i][j]),
                                 Complex.times(sn,U[i][k-1]));
                         U[i][k-1] = Complex.minus(Complex.times(cs,U[i][k-1]),
                                 Complex.times(sn,U[i][j]));
                         U[i][j] = t;
                     }
                  }
               }
            }
            break;

            // Perform one qr step.

            case 3: {

               // Calculate the shift.
   
               double scale = Math.max(Math.max(Math.max(Math.max(
                       Complex.abs(s[p-1]),Complex.abs(s[p-2])),Complex.abs(e[p-2])), 
                       Complex.abs(s[k])),Complex.abs(e[k]));
               Complex sp = s[p-1].over(scale);
               Complex spm1 = s[p-2].over(scale);
               Complex epm1 = e[p-2].over(scale);
               Complex sk = s[k].over(scale);
               Complex ek = e[k].over(scale);
               Complex b = (((spm1.plus(sp)).times(spm1.minus(sp))).plus(epm1.times(epm1))).over(2.0);
               Complex c = (sp.times(epm1)).times(sp.times(epm1));
               Complex shift = czero;
               if ((!b.equals(czero)) | (!c.equals(czero))) {
                  shift = Complex.sqrt((b.times(b)).plus(c));
//                  if (b < 0.0) {
//                     shift = -shift;
//                  }
                  shift = c.over(b.plus(shift));
               }
               Complex f = ((sk.plus(sp)).times(sk.minus(sp))).plus(shift);
               Complex g = sk.times(ek);
   
               // Chase zeros.
   
               for (int j = k; j < p-1; j++) {
                  Complex t = Complex.sqrt((f.times(f)).plus(g.times(g)));
                  Complex cs = f.over(t);
                  Complex sn = g.over(t);
                  if (j != k) {
                     e[j-1] = t;
                  }
                  f = (cs.times(s[j])).plus(sn.times(e[j]));
                  e[j] = (cs.times(e[j])).minus(sn.times(s[j]));
                  g = sn.times(s[j+1]);
                  s[j+1] = cs.times(s[j+1]);
                  if (wantv) {
                     for (int i = 0; i < n; i++) {
                        t = (cs.times(V[i][j])).plus(sn.times(V[i][j+1]));
                        V[i][j+1] = (cs.times(V[i][j+1])).minus(sn.times(V[i][j]));
                        V[i][j] = t;
                     }
                  }
                  t = Complex.sqrt((f.times(f).plus(g.times(g))));
                  cs = f.over(t);
                  sn = g.over(t);
                  s[j] = t;
                  f = (cs.times(e[j])).plus(sn.times(s[j+1]));
                  s[j+1] = (cs.times(s[j+1])).minus(sn.times(e[j]));
                  g = sn.times(e[j+1]);
                  e[j+1] = cs.times(e[j+1]);
                  if (wantu && (j < m-1)) {
                     for (int i = 0; i < m; i++) {
                        t = (cs.times(U[i][j])).plus(sn.times(U[i][j+1]));
                        U[i][j+1] = (cs.times(U[i][j+1])).minus(sn.times(U[i][j]));
                        U[i][j] = t;
                     }
                  }
               }
               e[p-2] = f;
               iter = iter + 1;
            }
            break;

            // Convergence.

            case 4: {

               // Make the singular values positive.
   
/*               if (s[k] <= 0.0) {
                  s[k] = (s[k] < 0.0 ? -s[k] : 0.0);
                  if (wantv) {
                     for (int i = 0; i <= pp; i++) {
                        V[i][k] = -V[i][k];
                     }
                  }
               }
*/   
               // Order the singular values.
   
               while (k < pp) {
                  if (Complex.abs(s[k]) >= Complex.abs(s[k+1])) {
                     break;
                  }
                  Complex t = s[k];
                  s[k] = s[k+1];
                  s[k+1] = t;
                  if (wantv && (k < n-1)) {
                     for (int i = 0; i < n; i++) {
                        t = V[i][k+1]; V[i][k+1] = V[i][k]; V[i][k] = t;
                     }
                  }
                  if (wantu && (k < m-1)) {
                     for (int i = 0; i < m; i++) {
                        t = U[i][k+1]; U[i][k+1] = U[i][k]; U[i][k] = t;
                     }
                  }
                  k++;
               }
               iter = 0;
               p--;
            }
            break;
         }
      }
   }

/* ------------------------
   Public Methods
 * ------------------------ */

   /** Return the left singular vectors
   @return     U
   */

   //public Matrix getU () {
   //   return new Matrix(U,m,Math.min(m+1,n));
  // }

   /** Return the right singular vectors
   @return     V
   */

   //public Matrix getV () {
   //   return new Matrix(V,n,n);
   //}

   /** Return the one-dimensional array of singular values
   @return     diagonal of S.
   */

   //public Complex[] getSingularValues () {
   //   return s;
  // }

   /** Return the diagonal matrix of singular values
   @return     S
   */

   //public Matrix getS () {
   //   Matrix X = new Matrix(n,n);
   //   Complex[][] S = X.getArray();
   //   for (int i = 0; i < n; i++) {
   //      for (int j = 0; j < n; j++) {
   //         S[i][j] = czero;
   //      }
   //      S[i][i] = this.s[i];
   //   }
   //   return X;
   //}

   /** Two norm
   @return     max(S)
   */

   //public Complex norm2 () {
   //   return s[0];
   //}

   /** Two norm condition number
   @return     max(S)/min(S)
   */

   //public Complex cond () {
   //   return s[0].over(s[Math.min(m,n)-1]);
   //}

   /** Effective numerical matrix rank
   @return     Number of nonnegligible singular values.
   */

   public int rank () {
      double eps = Math.pow(2.0,-52.0);
      double tol = Complex.abs(s[0].times(Math.max(m,n)*eps));
      int r = 0;
      for (int i = 0; i < s.length; i++) {
         if (Complex.abs(s[i]) > tol) {
            r++;
         }
      }
      return r;
   }
}
