package gov.fnal.eag.shapelets.functions;


import cern.jet.math.*;

/**
 * implements paper 1 equation 19 and paper 3 equations 75 - 82
 * 
 * @author stoughton, kuropat
 */
public class BasisFunction extends Phi{

	/**
	 * I made it public because it will be frequently used
	 */
	public static int[][] ind;   
	/**
	 * binomial coefficients
	 */
	public static double[][] binom;
	protected static int nMaxBasis = nMaxPhi-4; // 24
	protected static double[] XI;
	protected static double[] YI;
	private static double sqrtPI = Math.sqrt(Math.PI);
	private static double sqrt2 = Math.sqrt(2.);
/**
 *    Have to add 3 to nMaxBasis to be able to calculate 
 *   integrals in the recurrent formulas to nMaxBasis rank
 */
	static {
		XI = new double[nMaxBasis+3];
		YI = new double[nMaxBasis+3];
		ind = makeSVect(nMaxBasis+3);
		binom = BinCoef();
	}
	

	/**
	 * calculate the value of eigenfunctions at point (x,y) for the rank (ny,nx)
	 * @param ny
	 * @param nx
	 * @param y
	 * @param x
	 * @param beta
	 * @return double value of the eigenfunction at the point
	 */

	public static double evaluate(int ny, int nx, double y, double x, double beta) {
		double phin1x1 = evaluatePhi(nx, x/beta);
		double phin2x2 = evaluatePhi(ny, y/beta);
		//System.out.println(" in BasisFunction.evaluate:  nx="+nx+" ny="+ny+" x="+x+" y="+y+" beta="+beta+" phin1x1="+phin1x1+"  phin2x2="+phin2x2);
		return phin1x1 * phin2x2 /beta;
	}
	//
	/**
	 * calculates two dimensional integral of eigenfunctions at the point (y,x) 
	 * on the surface of a pixel of size (dy,dx) for functions of rank ny, nx. 
	 * @param ny 
	 * @param nx 
	 * @param y 
	 * @param x 
	 * @param beta 
	 * @param dy 
	 * @param dx 
	 * @return integrated eigenfunctions
	 * @throws Error 
	 */
	public static double integrate(int ny, int nx, double y, double x, double beta, double dy, double dx) throws Error {
		if (ny > YI.length-1) {
			throw new Error("ny="+ny+" is too large for YI.length="+YI.length);
		}
		if (nx > XI.length-1) {
			throw new Error("nx="+nx+" is too large for XI.length="+XI.length);
		}
		double res =0.;
		double fact1 = Math.sqrt(beta*sqrtPI/2.);
		double fact2 = -1.*Math.sqrt(2.*beta);
		double xp =x/beta;
		double dxp = dx/beta;
		double yp = y/beta;
		double dyp = dy/beta;
		//
		
		
		XI[0] = fact1*(Functions.erf.apply((xp + dxp/2.)/sqrt2) - Functions.erf.apply((xp - dxp/2.)/sqrt2));
		YI[0] = fact1*(Functions.erf.apply((yp + dyp/2.)/sqrt2) - Functions.erf.apply((yp - dyp/2.)/sqrt2));
		XI[1] = fact2*(evaluatePhi(0,xp + dxp/2.) - evaluatePhi(0,xp - dxp/2.));
		YI[1] = fact2*(evaluatePhi(0,yp + dyp/2.) - evaluatePhi(0,yp - dyp/2.));

		if (ny >=2) {
			for (int n=2; n<=ny; n++) {
				YI[n] = Math.sqrt((n - 1.)/n)*YI[n-2] -
				 Math.sqrt(2.*beta/n)*(evaluatePhi(n-1,yp+dyp/2.) - evaluatePhi(n-1,yp-dyp/2.));

			}
		}
		if (nx >=2) {
			for (int n=2; n<=nx; n++) {
				XI[n] = Math.sqrt((n - 1.)/n)*XI[n-2] -
				  Math.sqrt(2.*beta/n)*(evaluatePhi(n-1,xp+dxp/2.) - evaluatePhi(n-1,xp-dxp/2.));
			}
		}
		res = YI[ny]*XI[nx];
		return res;
	}
	/**
	 * prepare to integrate in pixels at position x up to and including order nMax
	 * @param nMax
	 * @param x
	 * @param beta
	 * @return double array xi
	 */
	public static double[] preIntegrate(int nMax, double x, double beta) {
		double[] xi = new double[nMax+1];
		double fact1 = Math.sqrt(beta*sqrtPI/2.);
		double fact2 = -1.*Math.sqrt(2.*beta);
		double xp =x/beta;
		double dxp = 1.0/beta;
		xi[0] = fact1*(Functions.erf.apply((xp + dxp/2.)/sqrt2) - Functions.erf.apply((xp - dxp/2.)/sqrt2));
		if (nMax == 0) return xi;
		xi[1] = fact2*(evaluatePhi(0,xp + dxp/2.) - evaluatePhi(0,xp - dxp/2.));
		if (nMax == 1) return xi;
		for (int nx=2; nx<=nMax; nx++) {
			xi[nx] = Math.sqrt((nx - 1.)/nx)*xi[nx-2] -
			  Math.sqrt(2.*beta/nx)*(evaluatePhi(nx-1,xp+dxp/2.) - evaluatePhi(nx-1,xp-dxp/2.));
		}
		return xi;
	}
	/**
	 * calculates all the integrals of eigenfunctions in specified point (y,x) up to the nMax rank.
	 * Integration is performed on the pixel size (dy,dx)
	 * @param nMax
	 * @param y
	 * @param x
	 * @param beta
	 * @param dy
	 * @param dx
	 * @return double[][] of integrated eigenfunctions at the point
	 * @throws shapeletException
	 */
	public static double[][] integrateAll(int nMax, double y, double x, double beta, double dy, double dx) {
		/* if nMax > nMaxBasis give exception             NK  */
		if (nMax > nMaxBasis) {
			throw new IllegalArgumentException(" nMax > nMaxBasis");
		}
		double[][] res = new double[nMax+1][nMax+1];
		double fact1 = Math.sqrt(beta*sqrtPI/2.);
		double fact2 = -1.*Math.sqrt(2.*beta);
		double xp =x/beta;
		double dxp = dx/beta;
		double yp = y/beta;
		double dyp = dy/beta;
		//
		
		
		XI[0] = fact1*(Functions.erf.apply((xp + dxp/2.)/sqrt2) - Functions.erf.apply((xp - dxp/2.)/sqrt2));
		YI[0] = fact1*(Functions.erf.apply((yp + dyp/2.)/sqrt2) - Functions.erf.apply((yp - dyp/2.)/sqrt2));
		XI[1] = fact2*(evaluatePhi(0,xp + dxp/2.) - evaluatePhi(0,xp - dxp/2.));
		YI[1] = fact2*(evaluatePhi(0,yp + dyp/2.) - evaluatePhi(0,yp - dyp/2.));

		if (nMax >=2) {
			for (int n=2; n<=nMax; n++) {
				YI[n] = Math.sqrt((n - 1.)/n)*YI[n-2] -
				 Math.sqrt(2.*beta/n)*(evaluatePhi(n-1,yp+dyp/2.) - evaluatePhi(n-1,yp-dyp/2.));

			}
		
			for (int n=2; n<=nMax; n++) {
				XI[n] = Math.sqrt((n - 1.)/n)*XI[n-2] -
				  Math.sqrt(2.*beta/n)*(evaluatePhi(n-1,xp+dxp/2.) - evaluatePhi(n-1,xp-dxp/2.));

			}
		}
		for (int ny=0; ny<=nMax; ny++) {
			for (int nx=0; nx<=nMax; nx++) {		
		       res[ny][nx] = YI[ny]*XI[nx];
			}
		}
		return res;
	}
	//
	/**
	 *  takes a table of eigenvalues and returns a vector
	 * @param tab 
	 * @return vector representation of a table
	 */
	public static double[] tab2Vector(double[][] tab) {

		int size = lengthFromDimension(tab.length);
		double[] vector = new double[size];

		for (int i=0; i< size; i++) {
	       	vector[i] = tab[ind[0][i]][ind[1][i]];		 			
		}
		return vector;
	}

	/**
	 * @param length - number of columns in eigenvalues table
	 * @return length of vector to store upper part of the matrix
	 */
	public static int lengthFromDimension(int length) {
		int size = length*(length+1)/2;

		return size;
	}
	
	/**
	 * takes a vector of eigenvalues and return a table
	 * @param vector 
	 * @return double[][] resulting matrix
	 */
	public static double[][] vector2Tab(double[] vector) {
		double[][] res = null;
		int size = vector.length;
		int n = dimensionFromLength(vector.length);

        res = new double[n][n]; 
        
        /*  Don't need this:  Java initializes arrays to zero
        for (int i=0; i<n; i++) {
        	for (int j=0; j<n; j++) {
        		res[i][j] = 0.;
        	}
        }
        */
        
        for (int i=0; i<size; i++ ) {
			//int ind0i = ind[0][i];
			//int ind1i = ind[1][i];
           	res[ind[0][i]][ind[1][i]] = vector[i];
		}
		return res;
	}

	/**
	 * @param vector values swapped in place to flip row,col
	 */
	public static void flipVector(double[] vector) {
		int size = vector.length;
		double[][] tab = vector2Tab(vector);
		
		for (int i=0; i<size; i++) {
	       	vector[i] = tab[ind[1][i]][ind[0][i]];  // Note that indices are swapped compared to tab2Vector() method		 			
		}
	}
	/**
	 * @param length dimension of 1-d representation
	 * @return dimension of 2-d representation
	 */
	public static int dimensionFromLength(int length) {
		int dimension = (int) Math.sqrt(2.*length);
//		System.out.println(" length="+length+" dimension="+dimension+" calculated="+Math.sqrt(2.*length));
		if (length != lengthFromDimension(dimension)) {
			throw new IllegalArgumentException("illegal length="+length+"  for dimension="+dimension+"  length is "+lengthFromDimension(dimension));
		}
		return dimension;
	}
	/**
	 *  creates 2 dimensional array of indexes
	 * to specify which number in a vector of shapelet   
	 *    coefficients corresponds to which n1 and n2, or which nr and nl. 
     *  A vector of polar shapelet coefficients, created using ShapeletsTransform.c2p
     *  can be indexed using:
     *    BasisFunction.makeSVect(nMax)
     *    n=n_r+n_l
     *    m=n_r-n_l
	 * 	 
	 * @param nMax  maximal index in eigenvalue matrix ( matrix dimension is nMax+1)
	 * @return int[][] of indexes with nMax+1 rank
	 */
	private static int[][] makeSVect(int nMax) {
 // number of elements in half matrix+diagonal
		int nCoof =(nMax+1)*(nMax+2)/2;
//		System.out.println("makeSVect nCoof="+nCoof+" nMax="+nMax);
		int i = 0;
		ind = new int[2][nCoof];
		for (int n=0; n<nMax+1;n++){
			for (int n1i=0; n1i<n+1; n1i++){
//				res[0][i] = n1i;
//				res[1][i] = n - n1i;
				ind[0][i] = n - n1i;
				ind[1][i] = n1i;
				i++;

			}
		}
		return ind;
	}
	/**
	 * 
	 * stores binomial coefficients. 
	 * 
	 *  Use coefficients as binom[n][k] where n and k are integer,
	 * result is double
	 * 
	 * @author kuropat
	 * @return double [][] of binominal coeffs
	 */
	private static double[][] BinCoef() {

		//double[][] binom;
		if (binom == null) {
			binom = new double[nMaxBasis+1 ][nMaxBasis +1 ];
			for (int i = 0; i < nMaxBasis+1; i++) {
				for (int j = 0; j <= i; j++) {
					binom[i][j] = (Factorial.factorial[i] / (Factorial.factorial[i - j] * Factorial.factorial[j]));
				}
			}
		}
		return binom;
	}
	/**
	 * convenience method to restore images.
	 * For a given point (x,y) relative to the center of image and pixel size dy X dx,
	 * calculate integrals on eigenfunctions multiplied by eigenvalues.
	 * All the work is done in a one place and should be very fast.
	 * @param Vnm  -- eigenvalues
	 * @param y    -- y coordinate relative to the image center (i - y0)
	 * @param x    -- x coordinate (j - x0) 
	 * @param beta -- scale factor
	 * @param dy   -- pixel sizes
	 * @param dx
	 * @return     -- image value in the pixel
	 */
	 
	public double restore(double[][]Vnm, double y, double x, double beta, double dy, double dx )  {
		double res = 0.;

		double fact1 = Math.sqrt(beta*sqrtPI/2.);
		double fact2 = -1.*Math.sqrt(2.*beta);
		double xp =x/beta;
		double dxp = dx/beta;
		double yp = y/beta;
		double dyp = dy/beta;
		int nVal = Math.min(nMaxBasis,Vnm.length);
		//
		
		
		XI[0] = fact1*(Functions.erf.apply((xp + dxp/2.)/sqrt2) - Functions.erf.apply((xp - dxp/2.)/sqrt2));
		YI[0] = fact1*(Functions.erf.apply((yp + dyp/2.)/sqrt2) - Functions.erf.apply((yp - dyp/2.)/sqrt2));
		XI[1] = fact2*(evaluatePhi(0,xp + dxp/2.) - evaluatePhi(0,xp - dxp/2.));
		YI[1] = fact2*(evaluatePhi(0,yp + dyp/2.) - evaluatePhi(0,yp - dyp/2.));

		if (nMaxBasis >=2) {
			for (int n=2; n<nMaxBasis+1; n++) {
				YI[n] = Math.sqrt((n - 1.)/n)*YI[n-2] -
				 Math.sqrt(2.*beta/n)*(evaluatePhi(n-1,yp+dyp/2.) - evaluatePhi(n-1,yp-dyp/2.));

			}
		
			for (int n=2; n<nMaxBasis+1; n++) {
				XI[n] = Math.sqrt((n - 1.)/n)*XI[n-2] -
				  Math.sqrt(2.*beta/n)*(evaluatePhi(n-1,xp+dxp/2.) - evaluatePhi(n-1,xp-dxp/2.));
			}
		}
		for (int ny=0; ny<nVal; ny++) {
			for (int nx=0; nx<nVal; nx++) {		
		       res += Vnm[ny][nx]* YI[ny]*XI[nx];
			}
		}

		return res;
	}
	
	/**
	 * convenience method to restore images.
	 * For a given point (x,y) relative to the center of image and pixel size dy X dx
	 * It calculate integrals on eigenfunctions multiplied by eigenvalues.
	 * So all the work is done in a one place and should be very fast.
	 * @param Vnm  -- eigenvalues in form of vector
	 * @param y    -- y coordinate relative to the image center (i - y0)
	 * @param x    -- x coordinate (j - x0) 
	 * @param beta -- scale factor
	 * @param dy   -- pixel sizes
	 * @param dx
	 * @return     -- image value in the pixel
	 */
	/*
	public double restoreV(double[]Vnm, double y, double x, double beta, double dy, double dx )  {
		double res = 0.;

		double fact1 = Math.sqrt(beta*sqrtPI/2.);
		double fact2 = -1.*Math.sqrt(2.*beta);
		double xp =x/beta;
		double dxp = dx/beta;
		double yp = y/beta;
		double dyp = dy/beta;
		int nVal = Vnm.length;
		//
		
		
		XI[0] = fact1*(Functions.erf.apply((xp + dxp/2.)/sqrt2) - Functions.erf.apply((xp - dxp/2.)/sqrt2));
		YI[0] = fact1*(Functions.erf.apply((yp + dyp/2.)/sqrt2) - Functions.erf.apply((yp - dyp/2.)/sqrt2));
		XI[1] = fact2*(evaluatePhi(0,xp + dxp/2.) - evaluatePhi(0,xp - dxp/2.));
		YI[1] = fact2*(evaluatePhi(0,yp + dyp/2.) - evaluatePhi(0,yp - dyp/2.));

		if (nMaxBasis >=2) {
			for (int n=2; n<nMaxBasis+1; n++) {
				YI[n] = Math.sqrt((n - 1.)/n)*YI[n-2] -
				Math.sqrt(2.*beta/n)*(evaluatePhi(n-1,yp+dyp/2.) - evaluatePhi(n-1,yp-dyp/2.));
			}
		
			for (int n=2; n<nMaxBasis+1; n++) {
				XI[n] = Math.sqrt((n - 1.)/n)*XI[n-2] -
				  Math.sqrt(2.*beta/n)*(evaluatePhi(n-1,xp+dxp/2.) - evaluatePhi(n-1,xp-dxp/2.));

			}
		}
		for (int k=0; k<nVal; k++){ 	
		       res += Vnm[k]* YI[ind[0][k]]*XI[ind[1][k]];
		}

		return res;
	}
	*/
	/** 
	 * @return nMax maximum index in the eigenvalue matrix.
	 * The matrix rank is nMax+1
	 * @see gov.fnal.eag.shapelets.functions.Phi#getNMax()
	 */
	public static int getNMax() {
		return nMaxBasis;
	}
	/**
	 * the method to calculate image centroids, flux and radius using
	 * analitical expressions
	 * @param nMax  maximal ranc to use
	 * @param beta  beta expected size of the image
	 * @param fnm   eigenvalues
	 * @return  double[4] moments 
	 *     moments[0] flux
	 *     moments[1] column centroid
	 *     moments[2] row centroid
	 *     moments[3] sigma or effective radius of the image
	 *     By N. Kuropatkin  07/02/2007
	 */
	public static double[] ImageAnalyser( double beta, double[][] fnm) {
		//First calculate Flux
		int nMax = Math.min(nMaxBasis, fnm.length-1);
		double[] moments = new double[4];
		double flux = 0.;
		double Xc = 0.;
		double Yc = 0.;
		double R2f = 0.;
		for (int n1=0; n1<=nMax; n1+=2) {
			for (int n2=0; n2<=nMax; n2+=2) {
				flux += Math.pow(2.,(2. -n1 -n2)/2.)*Math.sqrt(BasisFunction.binom[n1][n1/2]) * Math.sqrt(BasisFunction.binom[n2][n2/2])*fnm[n1][n2];
			}
		}
		flux *=sqrtPI*beta;
		// Now calculate centroids
		for (int n1 = 1; n1<nMax; n1+=2) {
			for (int n2 = 0; n2<nMax; n2+=2) {
				Yc += Math.sqrt(n1+1.)*Math.pow(2.,(2.-n1-n2)/2.)*Math.sqrt(BasisFunction.binom[n1+1][(n1+1)/2]) *
				      Math.sqrt(BasisFunction.binom[n2][n2/2])*fnm[n1][n2];
			}
		}
		Yc = Yc*sqrtPI*beta*beta/flux;
		for (int n2 =1; n2<nMax; n2 +=2) {
			double a = Math.sqrt(n2+1.);
			double c = Math.sqrt(BasisFunction.binom[n2+1][(n2+1)/2]);
			for (int n1=0; n1<=nMax; n1 +=2) {
				double b = Math.pow(2.,(2.-n1-n2)/2.);
				double d =  Math.sqrt(BasisFunction.binom[n1][n1/2])*fnm[n1][n2];
				Xc += a*b*c*d;
				if (Double.isNaN(Xc)) {
					throw new Error("n2="+n2+" n1="+n1+" nMax="+nMax+" a="+a+" b="+b+" c="+c+" d="+d);
				}
			}
		}
		Xc = Xc*sqrtPI*beta*beta/flux;
		// Finally calculate RMS - effective radius of the image (used to calculate beta)
		for (int n1=0; n1<=nMax; n1 +=2) {
			for (int n2=0; n2<=nMax; n2 +=2) {
				R2f +=Math.pow(2.,(4.-n1-n2)/2.)*(1.+n1+n2)*Math.sqrt(BasisFunction.binom[n1][n1/2])*
				      Math.sqrt(BasisFunction.binom[n2][n2/2])*fnm[n1][n2];
			}
		}
		R2f = Math.sqrt(R2f*sqrtPI*beta*beta*beta/flux);
		moments[0] = flux;   // integral flux
		moments[1] = Xc;     // column centroid
		moments[2] = Yc;     // row centroid
		moments[3] = R2f;    // 
		return moments;
	}
}
