package gov.fnal.eag.shapelets.functions;
import cern.colt.matrix.*;
/**
 * handles convolution of two objects defined by eigenvalues and scales, faster than Convolver.
 * @author stoughto
 */
public class Convolution {
	private double gamma;
	private double a1, a2, a3;
	private double nu;
	private double a, b, c;
	private double ct[][][];
	/**
	 * @param alpha scale of the first object
	 * @param beta scale of the second object
	 * @param nMax maximum order to use
	 */
	public Convolution(double alpha, double beta, int nMax) {
		gamma = Math.sqrt(alpha*alpha + beta*beta);
		a1 = 1.0/gamma;
		a2 = 1.0/alpha;
		a3 = 1.0/beta;
		nu = 1.0/Math.sqrt(2.0*gamma*gamma);
		a = Math.sqrt(2.0)*nu/a1;
		b = Math.sqrt(2.0)*nu/a2;
		c = Math.sqrt(2.0)*nu/a3;
		ct = getLTensor(a, b, c, nMax);
		DoubleFactory3D factory = DoubleFactory3D.dense;
		DoubleMatrix3D matrix = factory.make(nMax+1,nMax+1,nMax+1);
		double allFactor = Math.sqrt(2)*Math.pow(Math.PI, 0.25)*nu/Math.sqrt(a1*a2*a3);
		for (int l=0; l<=nMax; l++) {
			double lFactor = 1.0/Math.sqrt(Factorial.factorial[l]);
			for (int m=0; m<=nMax; m++) {
				double mFactor = 1.0/Math.sqrt(Factorial.factorial[m]);
				for (int n=0; n<=nMax; n++) {
					if ( (l+m+n)%2 == 0) {
						double nFactor = 1.0/Math.sqrt(Factorial.factorial[n]);
						ct[n][m][l] *= allFactor*Math.pow(-1, (3*n+m+l)/2)*lFactor*mFactor*nFactor/Math.sqrt(Math.pow(2, l+m+n-1));
						matrix.setQuick(n, m, l,ct[n][m][l] );
					}
				}
			}
		}
	}		
	/**
	 * convolve these two objects
	 * @param f eigenvalues for the first object
	 * @param g eigenvalues for the second object
	 * @return the eigenvalues for the convolved object
	 */
	public double[][] convolve(double[][] f, double[][] g) {
		//System.out.println("Convolution.convolve:  f.length="+f.length+" g.length="+g.length);
		int length = Math.max(f.length, g.length);
		double[][] retval = new double[length][length];
		for (int n1=0; n1<f.length; n1++) {
			for (int m1=0; m1<f.length; m1++) {
				for (int l1=0; l1<g.length; l1++) {
					if ( (n1+m1+l1)%2 == 0) {
						for (int n2=0; n2<f.length-n1; n2++) {
							for (int m2=0; m2<f.length-m1; m2++) {
								for (int l2=0; l2<g.length-l1; l2++) {
									if ( (n2+m2+l2)%2 == 0) {
										retval[n1][n2] += ct[n1][m1][l1]*ct[n2][m2][l2]*f[m1][m2]*g[l1][l2];
									}
								}
							}
						}
					}
				}
			}
		}
		return retval;
	}
	/**
	 * @return the ct matrix, for testing
	 */
	public double[][][] getCt() {return ct;}
	
	/**
	 * generate the L tensor of equation 11 paper 2
	 * @param a
	 * @param b
	 * @param c
	 * @param nMax
	 * @return the tensor
	 */
	public static double[][][] getLTensor(double a, double b, double c, int nMax) {

		double[][][] lt = new double[nMax + 1][nMax + 1][nMax + 1];
		boolean[][][] done = new boolean[nMax + 1][nMax + 1][nMax + 1];
		lt[0][0][0] = 1.0;
		done[0][0][0] = true;
		for (int i = 0; i < nMax + 1; i++) {
			for (int l = 0; l <= i; l++) {
				for (int m = 0; m <= i; m++) {
					for (int n = 0; n <= i; n++) {
						if ((l + m + n) % 2 == 0) {
							if (!done[l][m][n]) {
								double sum = 0.0;
								if ((l >= m) && (l >= n)) {
									if (((l - 2) >= 0)) {
										sum += 2 * (l - 1) * (a * a - 1)
												* lt[l - 2][m][n];
									}
									if (((l - 1) >= 0) && ((m - 1) >= 0)) {
										sum += 2. * m * a * b
												* lt[l - 1][m - 1][n];
									}
									if (((l - 1) >= 0) && ((n - 1) >= 0)) {
										sum += 2. * n * a * c
												* lt[l - 1][m][n - 1];
									}
									lt[l][m][n] = sum;
								} else if ((m >= l) && (m >= n)) {
									if (((m - 1) >= 0) && ((m - 2) >= 0)) {
										sum += 2. * (m - 1) * (b * b - 1)
												* lt[l][m - 2][n];
									}
									if (((m - 1) >= 0) && ((l - 1) >= 0)) {
										sum += 2. * l * a * b
												* lt[l - 1][m - 1][n];
									}
									if (((m - 1) >= 0) && ((n - 1) >= 0)) {
										sum += 2. * n * b * c
												* lt[l][m - 1][n - 1];
									}
									lt[l][m][n] = sum;
								} else {
									if (((n - 1) >= 0) && ((n - 2) >= 0)) {
										sum += 2. * (n - 1) * (c * c - 1)
												* lt[l][m][n - 2];
									}
									if (((n - 1) >= 0) && ((l - 1) >= 0)) {
										sum += 2. * l * a * c
												* lt[l - 1][m][n - 1];
									}
									if (((n - 1) >= 0) && ((m - 1) >= 0)) {
										sum += 2. * m * b * c
												* lt[l][m - 1][n - 1];
									}
									lt[l][m][n] = sum;
									done[l][m][n] = true;
								}
							}
						}
					}
				}
			}

		}
		return lt;
	}
	/**
	 * @return the width of the resulting convolution
	 */
	public double getGamma() {
		return gamma;
	}
}
