
package gov.fnal.eag.shapelets.functions;

import cern.colt.matrix.DoubleFactory2D;
import cern.colt.matrix.DoubleMatrix2D;
import cern.colt.matrix.linalg.LUDecompositionQuick;


/**
 * performs convolution/deconvolution of images presented by a set of eigenvalues.
 * 
 * It also calculate convolution matrix P of eq. 12 paper II and
 * inverse matrix P^-1 to be used for deconvolution
 *
 * @author kuropat
 *
 */
public class Convolver extends CCTensor {
	private int[][] ind = BasisFunction.ind;
	private int aSize = 0;        // convolved vector size
	private int bSize = 0;        // image vector size
	private int gSize = 0;        // psf vector size
	private double[][] Pconv;     // convolution matrix aSize x aSize
	private double[][] Pinv;      // inverted matrix for deconvolution


	/**
	 * fully-specified constructor
	 * @param gamma
	 * @param alpha
	 * @param beta
	 * @param psf 
	 * @param nMax
	 */
	public Convolver(double gamma, double alpha, double beta, double[][] psf, int nMax) {
		super(gamma, alpha, beta, nMax);
		//

		double[] gCoof;
		gCoof = BasisFunction.tab2Vector(psf);
		aSize = (nMax+1)*(nMax+2)/2;
		bSize = aSize;
		gSize = gCoof.length;
		Pconv = PConv(gCoof);
		Pinv = invert(Pconv);
	}

	/**
	 * fully-specified constructor
	 * @param gamma - expected convolved scale factor
	 * @param alpha - image scale factor
	 * @param beta - PSF scale factor
	 * @param psf 
	 * @param nMax
	 */
	public Convolver(double gamma, double alpha, double beta, double[] psf, int nMax) {
		super(gamma, alpha, beta, nMax);
		//


		aSize = (nMax+1)*(nMax+2)/2;
		bSize = aSize;
		gSize = psf.length;
		Pconv = PConv(psf);
		Pinv = invert(Pconv);
	}
	/**
	 * convolution of the image shapelet with the psf shapelet
	 * @param image 2d array of image shapelet eigenvalues
	 *
	 * @return 2d array of the convolved object shapelet eigenvalues 
	 */
	public double[] convolve(double[] image) {

		double[] convolved = new double[aSize];

		for (int n1 = 0; n1 < aSize; n1++) {

			convolved[n1] = 0.;
			for (int m1 = 0; m1 < bSize; m1++) {

				convolved[n1] += Pconv[n1][m1] * image[m1];
			}
		}

		return convolved;
	}
	/**
	 * convolve the image shapelet with the psf shapelet
	 * 
	 * @param image
	 *            2d array of image shapelet eigenvalues
	 *            
	 * @return 2d array of the convolved object shapelet eigenvalues
	 */
	public double[][] convolve(double[][] image) {
		//

		double[] imVect = BasisFunction.tab2Vector(image);

		double [][] convolved;

		double[] convTMP = new double[aSize];
		for (int n1 = 0; n1 < aSize; n1++){

			convTMP[n1] = 0.;
			//			for (int m1 = 0; m1 < bSize; m1++) {
			for (int m1 = 0; m1 < imVect.length; m1++) {
				convTMP[n1] += Pconv[n1][m1] * imVect[m1];
			}
		}
		convolved = BasisFunction.vector2Tab(convTMP);
		return convolved;
	}

	/**
	 * deconvolution of the psf shapelet from an image shapelet 
	 * @param image 2d array of image shapelet eigenvalues
	 *
	 * @return 2d array of the deconvolved object shapelet eigenvalues 
	 */
	public double[] deconvolve(double[] image) {

		double[] deconvolved = new double[aSize];

		for (int n1 = 0; n1 < aSize; n1++) {

			deconvolved[n1] = 0.;
			for (int m1 = 0; m1 < bSize; m1++) {

				deconvolved[n1] += Pinv[n1][m1] * image[m1];
			}
		}

		return deconvolved;
	}
	/**
	 * deconvolve the psf shapelet from the image shapelet 
	 * 
	 * @param image
	 *            2d array of image shapelet eigenvalues
	 *            
	 * @return 2d array of the deconvolved object shapelet eigenvalues
	 */
	public double[][] deconvolve(double[][] image) {
		//
		double[] imVect = BasisFunction.tab2Vector(image);
		double [][] deconvolved;

		double[] deconvTMP = new double[aSize];
		for (int n1 = 0; n1 < aSize; n1++) {

			deconvTMP[n1] = 0.;
			//			for (int m1 = 0; m1 < bSize-1; m1++) {
			for (int m1 = 0; m1 < imVect.length; m1++) {
				deconvTMP[n1] += Pinv[n1][m1] * imVect[m1];
			}
		}
		deconvolved = BasisFunction.vector2Tab(deconvTMP);
		return deconvolved;
	}

	/**
	 * 
	 * @param gCoof  psf eigenvector
	 * @return double[][] convolution matrix
	 */
	private double [][] PConv(double[] gCoof) {
		double[][] P = new double[aSize][bSize];

		//

		for (int i=0; i< aSize; i++) {
			for (int j=0; j< bSize; j++) {
				P[i][j] = 0.;

				for (int l=0; l<gSize; l++) {
					if (((ind[0][i] + ind[0][j] + ind[0][l]) % 2) == 0) {
						P[i][j]+=CC[i][j][l]*gCoof[l];
					}
				}
			}

		}
		return P;
	}
	/**
	 * @return convolution matrix
	 */
	public double[][] getPconv() { return Pconv; }

	private double[][] invert(final double[][] pconvIn) {
		double[][] Pdeconv;
		double [][] identity = new double[aSize][bSize];

		for (int i=0; i< aSize; i++) {
			for (int j =0; j<bSize; j++) {
				identity[i][j] =(i == j ?  1.0 : 0.0);
			}
		}
		DoubleMatrix2D I = DoubleFactory2D.dense.make(aSize,bSize);
		I.assign(identity);
		DoubleMatrix2D P = DoubleFactory2D.dense.make(aSize,bSize);
		P.assign(pconvIn);

		LUDecompositionQuick ld = new LUDecompositionQuick();
		ld.decompose(P);
		ld.solve(I);
		Pdeconv = I.toArray();
		return Pdeconv;
	}

}

