package gov.fnal.eag.shapelets.functions;

import java.util.Hashtable;

/**
 * implements equation 3 of paper 2
 * to decompse an image returning a set of eigenvalues.
 * 
 * @author stoughto
 * 
 */
public class Decompose {
	private double[][] f;
	private double x0, y0; // location of (x,y)=(0,0) in f
	private double scale;  // number of pixels in f that correspond to x,y changing by 1	
	private int nMaxBasis = BasisFunction.nMaxBasis;
	/**
	 *  Standard constructor
	 * @param f  - pixeled image
	 * @param y0 - row of the image center
	 * @param x0 - col of the image center
	 * @param scale -- scale (beta)
	 */
	public Decompose(double[][] f, double y0, double x0, double scale) {
		this.f = f;
		this.x0 = x0;
		this.y0 = y0;
		this.scale = scale;
	}
	
	
	/**
	 * Return a matrix of eigenvalues calculated with eigenfunctions
	 * integrated within the pixel.
	 * @param nMax
	 * @return double[][] of eigenvalues
	 * @throws shapeletException
	 */
	public double[][] getFnn2(int nMax)  {
		if (nMax > nMaxBasis ) nMax = nMaxBasis;
		double[][] sum = new double[nMax+1][nMax+1];
		double [][] V;
		double x = 0.;
		double y = 0.;
		double dx = 1.;
		double dy = 1.;
		double dxby2 = 0.5;
		double dyby2 = 0.5;
		for (int i=0; i<f.length; i++) {
			for (int j=0; j<f[0].length; j++) {
					y = i + dyby2 - y0;
					x = j + dxby2 - x0;
					V = BasisFunction.integrateAll(nMax,y,x,scale,dy,dx);
					for (int ny=0; ny<=nMax; ny++) {
						for (int nx=0; nx<=nMax; nx++) {
						sum[ny][nx] += f[i][j] * V[ny][nx];
					}
				}
			}
		}
		return sum;
	}
	
	/**
	 * performs the image decomposition with fast but rough accuracy
	 *  (accurate to half of pixel )
	 * @param nMax
	 * @return double[][] matrix of eigenvalues
	 */
	public double[][] project(int nMax) {
		if (nMax > nMaxBasis ) nMax = nMaxBasis;
		double[][] retval = new double[nMax+1][nMax+1];
		double[][] rowPrecalc = new double[nMax+1][f.length];
		double[][] colPrecalc = new double[nMax+1][f[0].length];
		for (int row=0; row<f.length; row++) {
			double[] xTemp = BasisFunction.preIntegrate(nMax, row-y0, scale);
			for (int n1=0; n1<=nMax; n1++) {
				rowPrecalc[n1][row] = xTemp[n1];
			}
		}
		for (int col=0; col<f[0].length; col++) {
			double[] yTemp = BasisFunction.preIntegrate(nMax, col-x0, scale);
			for (int n2=0; n2<=nMax; n2++) {
				colPrecalc[n2][col] = yTemp[n2];
			}
		}
		for (int row=0; row<f.length; row++) {
			for (int col=0; col<f[0].length; col++) {
				for (int n1=0; n1<=nMax; n1++) {
					for (int n2=0; n2<=nMax-n1; n2++) {
						retval[n1][n2] += rowPrecalc[n2][row]*colPrecalc[n1][col]*f[row][col];
					}
				}
			}
		}
		return retval;
	}
	/**
	 * Return a Vector of eigenvalues calculated with eigenfunctions
	 * integrated within the pixel.
	 * @param nMax
	 * @return double[] of eigenvalues
	 */
	public double[] getFnnV(int nMax)  {
		if (nMax > nMaxBasis ) nMax = nMaxBasis;
		int vectL = (nMax+1)*(nMax+2)/2;
		double[] sum = new double[vectL];
		double [][] V;
		double x = 0.;
		double y = 0.;
		double dx = 1.;
		double dy = 1.;
		double dxby2 = 0.5;
		double dyby2 = 0.5;
		int[][] ind = BasisFunction.ind;


		for (int i=0; i<f.length; i++) {
			for (int j=0; j<f[0].length; j++) {
					y = i  + dyby2 + - y0;
					
					x = j + dxby2 - x0;
					V = BasisFunction.integrateAll(nMax,y,x,scale,dy,dx);
					for (int k=0; k<vectL; k++) {
						sum[k] += f[i][j] * V[ind[0][k]][ind[1][k]];
					}	
			}
		}
//

		return sum;
	}
	
	/**
	 * @param pix
	 * @return hashtable of the centroid information
	 */
	public static Hashtable<String,Number> centroid(double[][] pix) {
		double weight;
		double sum=0;
		double iSum=0;
		double jSum=0;
		double i2Sum=0;
		double j2Sum=0;
		int nPixels=0;
		for (int i=0; i<pix.length; i++) {
			double iCent = i + 0.5;
			for (int j = 0; j < pix[0].length; j++) {
				double jCent = j + 0.5;
				weight = pix[i][j];
				sum += weight;
				iSum += iCent * weight;
				jSum += jCent * weight;
				i2Sum += iCent * iCent * weight;
				j2Sum += jCent * jCent * weight;
				nPixels++;
			}
		}
		Hashtable<String,Number> hashtable = new Hashtable<String,Number>();
		hashtable.put("nPixels", new Integer(nPixels));
		hashtable.put("iCentroid", new Double(iSum / sum));
		hashtable.put("jCentroid", new Double(jSum / sum));
		double s2 = sum * sum;
		hashtable.put("iSigma", new Double(Math.sqrt(i2Sum / sum
				- iSum * iSum / s2)));
		hashtable.put("jSigma", new Double(Math.sqrt(j2Sum / sum
				- jSum * jSum / s2)));
		return hashtable;
	}

}
