package gov.fnal.eag.sim.catalog;

import gov.fnal.eag.util.astronomy.Cosmo;
import gov.fnal.eag.wcs.SphericalCoordinate;

/**
 * Collects the properties of an astronomical source.
 * This class does little (or no) processing itself; it
 * just holds the parameters of an object, which can be
 * accessed through getters and setters.
 * 
 * @author neilsen
 * @version $Name:  $ ($Id: AstroObject.java,v 1.37 2009/02/25 17:45:57 stoughto Exp $)
 */
public class AstroObject {

	// Fields
	// 

	/**
	 * the position of the object
	 */
	protected SphericalCoordinate location;

	/**
	 * the redshift of this object z = (&lambda;<sub>obs</sub>-&lambda;<sub>rest</sub>)/(&lambda;<sub>rest</sub>)
	 */
	protected double redshift;

	protected double distancePc;

	/**
	 * the spectral energy distribution
	 */
	protected SED spectrum;
	// remember the original value of the fluxHashCode and check that it does not get changed
	private int spectrumFluxHashCode;
	private double spectrumDeltaMagnitude;
	/**
	 * the morphology
	 */
	protected Morphology shape;

	/**
	 * Fully qualified constructor for {@link gov.fnal.eag.sim.catalog.AstroObject}s
	 * 
	 * @param coordinate the astronomical coordinates of the object, in decimal degrees
	 * @param redshift the redshift (z) of the object
	 * @param distancePc non-cosmological distance to the object.
	 * @param spectrum the spectral energy distribution, in ergs/sec/cm/cm at 10 pc
	 * @param shape the shape of the object
	 * @throws IllegalArgumentException if neither redshift nor distancePc equal 0.0
	 */
	public AstroObject(SphericalCoordinate coordinate, double redshift, double distancePc, SED spectrum, Morphology shape) {
		this.setLocation(coordinate);
		this.setRedshiftAndDistancePc(redshift, distancePc);
		this.setSpectrum(spectrum);
		this.setShape(shape);
	}



	// Accessor Methods
	/**
	 * Get the location
	 * 
	 * @return the location
	 */
	public SphericalCoordinate getLocation (  ) {
		return location;
	}

	/**
	 * @param value Set the value of location
	 */
	public void setLocation ( SphericalCoordinate value  ) {
		location = value;
	}

	/**
	 * @return redshift of this object
	 */
	public double getRedshift (  ) {
		return redshift;
	}

	/**
	 * @return the non-cosmological distance to this object, in pc
	 */
	public double getDistancePc() {
		return distancePc;
	}

	/**
	 * sets the redshift and distance, after checking that exactly one of them is >= 0
	 * @param redshift 
	 * @param distancePc 
	 */
	public void setRedshiftAndDistancePc (double redshift, double distancePc) {
		if (redshift <= 0.0 && distancePc <= 0.0) {
			throw new IllegalArgumentException("either redshift or distance must be positive but redshift="+redshift+" distancePc="+distancePc);
		}
		if (redshift > 0.0 && distancePc > 0.0) {
			throw new IllegalArgumentException("only one of redshift and distance may be positive but redshift="+redshift+" distancePc="+distancePc);
		}
		this.redshift = redshift;
		this.distancePc = distancePc;
	}
	/**
	 * Get the  spectrum
	 * 
	 * @return the spectrum
	 */
	public SED getSpectrum (  ) {
		return spectrum;
	}

	/**
	 * @param value Set the value of spectrum
	 * 
	 * 
	 */
	public void setSpectrum ( SED value  ) {
		spectrum = value;
		spectrumFluxHashCode = spectrum.fluxHashCode();
		spectrumDeltaMagnitude = 0.0;
	}

	/**
	 * return the shape of the object
	 * @return shapelet coefficients for extended object; null for point source
	 */
	public Morphology getShape (  ) {
		return shape;
	}

	/**
	 * set the shape of the object
	 * @param newShape the shape object to describe the object
	 */
	public void setShape ( Morphology newShape  ) {
		shape = newShape.copy();
	}

	/**
	 * return the size of the object (in Mpc)
	 * @param cosmology the cosmology to use
	 * @return size in Mpc
	 */
	public double getSize (Cosmo cosmology) {
		return shape.getSizeInMpc(cosmology);
	}

	/**
	 * Return the diameter of the object, in arc seconds
	 * @param cosmology the cosmology to use
	 * @return the diameter of the object, in arc seconds
	 */
	public double getAngularSize(Cosmo cosmology) {
		double sizeInParsecs = 1.0e6*getSize(cosmology);
		double angularDistance = cosmology.angularDistance(getRedshift())*cosmology.getCH0();

		double angularDiameter = sizeInParsecs/angularDistance; // in radians
		angularDiameter *= (180.0/Math.PI)*3600; // now it is in arcseconds
		//System.out.println("AstroObject:  sizeInParsecs="+sizeInParsecs+" angularDistance="+angularDistance+" angularDiameter="+angularDiameter+" redshift="+getRedshift());
		return angularDiameter;
	}

	/**
	 * scales the spectrum of the object so that it has the absolute magnitude in the specified filter at the redshift
	 * @param filter the filter 
	 * @param absoluteMagnitude the desired magnitude
	 */
	public void setAbsoluteMagnitude (AstronomicalFilter filter, double absoluteMagnitude) {
		if (spectrumFluxHashCode != spectrum.fluxHashCode()) {
			throw new RuntimeException("spectrum flux changed:  spectrum="+spectrum);
		}
		double currentAbsoluteMagnitude = getAbsoluteMagnitude(filter);
		spectrumDeltaMagnitude += absoluteMagnitude - currentAbsoluteMagnitude;
	}

	/**
	 * @param filter
	 * @return the absolute magnitude in the rest-frame filter
	 */
	public double getAbsoluteMagnitude(AstronomicalFilter filter) {
		if (spectrumFluxHashCode != spectrum.fluxHashCode()) {
			throw new RuntimeException("spectrum flux changed:  spectrum="+spectrum);
		}
		double absoluteMagnitude = spectrum.getMagnitude(filter);
		absoluteMagnitude += spectrumDeltaMagnitude;
		return absoluteMagnitude;
	}
	/**
	 * scales the spectrum so the object so that it has the apparent magnitude in the specified filter
	 * @param filter the filter
	 * @param apparentMagnitude the desired apparent magnitude
	 * @param cosmo used to convert apparent to absolute magnitude, with the redshift of this object
	 */
	public void setApparentMagnitude (AstronomicalFilter filter, double apparentMagnitude, Cosmo cosmo) {
		if (spectrumFluxHashCode != spectrum.fluxHashCode()) {
			throw new RuntimeException("spectrum flux changed:  spectrum="+spectrum);
		}
		double currentApparentMagnitude = getApparentMagnitude(filter, cosmo);
		spectrumDeltaMagnitude += apparentMagnitude - currentApparentMagnitude;
	}
	
	/**
	 * @param filter
	 * @param cosmo
	 * @return the apparent magnitude in the filter in the observed filter, using the cosmology for distanceModulus
	 */
	public double getApparentMagnitude(AstronomicalFilter filter, Cosmo cosmo) {
		if (spectrumFluxHashCode != spectrum.fluxHashCode()) {
			throw new RuntimeException("spectrum flux changed:  spectrum="+spectrum);
		}
		double magnitude = spectrum.getMagnitude(filter, redshift); 
		double distanceModulus = getDistanceModulus(cosmo);
		double retval = magnitude + distanceModulus + spectrumDeltaMagnitude;		
		return retval;
	}
	
	/**
	 * @param cosmo
	 * @return the distance modulus; 5*log10(distancePc/10) or cosmos.distanceModulus(redshift)
	 */
	public double getDistanceModulus(Cosmo cosmo) {
		double retval;
		if (distancePc > 0) {
			retval = 5.0*Math.log10(distancePc/10.0);
		} else {
			retval = cosmo.distanceModulus(redshift);
		}
		return retval;
	}



	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		String retval = getClass().getCanonicalName();
		retval += " location="+location+" redshift="+getRedshift();
		return retval;

	}
	/**
	 * @param that
	 * @return true if this equals that, without chatter
	 */
	public boolean equals(AstroObject that) {
		return equals(that, false);
	}
	/**
	 * @param that
	 * @param chatty if true, talk to System.out about why they are not equal
	 * @return true if this equals that
	 */
	public boolean equals(AstroObject that, boolean chatty) {
		if (!this.location.equals(that.location)) {
			if (chatty) {
				System.out.println("AstroObject.equals:  locations not equal");
			}
			return false;
		}
		if (Math.abs(this.redshift - that.redshift) > 1e-15) {
			if (chatty) {
				System.out.println("AstroObject.equals:  redshifts not equal");
			}
			return false;
		}
		if (!this.spectrum.equals(that.spectrum)) {
			if (chatty) {
				System.out.println("AstroObject.equals:  spectra not equal");
			}
			return false;
		}
		if (!this.shape.equals(that.shape, chatty)) {
			if (chatty) {
				System.out.println("AstroObject.equals:  shapes not equal");
			}
			return false;
		}
		return true;
	}
	/**
	 * @param gamma1
	 * @param gamma2
	 */
	public void shear(double gamma1, double gamma2) {
		shape.shear(gamma1, gamma2);
	}



}


