package gov.fnal.eag.sim.catalog;

import gov.fnal.eag.sim.random.SimRandomLuminosity;
import gov.fnal.eag.sim.random.SimRandomPoisson;
import gov.fnal.eag.sim.random.SimRandomRedshift;
import gov.fnal.eag.util.stopwatch.StopwatchCollection;
import gov.fnal.eag.wcs.SphericalCoordinate;

import java.util.Iterator;


/**
 * Creates {@link gov.fnal.eag.sim.catalog.AstroObject}s. 
 * 
 * In normal use, a calling method would create (or otherwise
 * obtain) generator objects for each of an
 * {@link gov.fnal.eag.sim.catalog.AstroObject}s attributes,
 * create an AstroObjectGenerator using those attributes, and call
 * next() each time a new AstroObject is desired.
 *                
 * @author neilsen
 * @version $Name:  $ ($Id: AstroObjectGenerator.java,v 1.31 2009/02/25 17:45:57 stoughto Exp $)
 * 
 */
public class AstroObjectGenerator implements Iterable<AstroObject> {
	protected CoordinateGenerator coordinateGenerator;
	protected SimRandomRedshift redshiftGenerator;
	protected SEDGenerator sedGenerator;
	protected MorphologyGenerator morphologyGenerator;
	protected SimRandomLuminosity luminosityGenerator;
	protected SimRandomPoisson simRandomPoisson;
	protected AstronomicalFilter filter;

	protected AstroObjectGenerator() {
	}
	/**
	 * Fully qualified constructor for AstroObjectGenerators
	 * 
	 * @param coordinateGenerator generator for astronomical coordinates
	 * @param redshiftGenerator generator for redshifts
	 * @param sedGenerator generator for spectra
	 * @param morphologyGenerator generator for shapes
	 * @param luminosityGenerator generator for absolute magnitudes
	 */
	public AstroObjectGenerator(CoordinateGenerator coordinateGenerator,
			SimRandomRedshift redshiftGenerator,
			SEDGenerator sedGenerator,
			MorphologyGenerator morphologyGenerator,
			SimRandomLuminosity luminosityGenerator) {
		this.coordinateGenerator = coordinateGenerator;
		this.redshiftGenerator = redshiftGenerator;
		this.sedGenerator = sedGenerator;
		this.morphologyGenerator = morphologyGenerator;
		this.luminosityGenerator = luminosityGenerator;
		int poissonSeed = coordinateGenerator.getISeed() + redshiftGenerator.getIseedIndex()+sedGenerator.getISeed()+
		morphologyGenerator.getISeed()+luminosityGenerator.getISeed();
		simRandomPoisson = new SimRandomPoisson(poissonSeed);	  
		setFilter();
	}

	/**
	 * @return whether the redshiftGenerator defines distances as PC or redshift
	 */
	public boolean getRedshiftDefinedAsDistancePc () {
		boolean retval = redshiftGenerator.getDefinedAsDistancePc();
		return retval;
	}
	protected void setFilter() {
		filter = luminosityGenerator.getFilter();
		if (filter == null) {
			StopwatchCollection.start("    AstroObjectGenerator: getFilterName+scale");
			String filterName = luminosityGenerator.getFilterName();
			WavelengthScale scale = sedGenerator.getWavelengthScale();	  
			StopwatchCollection.stop("    AstroObjectGenerator: getFilterName+scale");
			StopwatchCollection.start("    AstroObjectGenerator: new astronomical filter");
			filter = new AstronomicalFilter(filterName, scale);
			StopwatchCollection.stop("    AstroObjectGenerator: new astronomical filter");
		} else {
			WavelengthScale sgws = sedGenerator.getWavelengthScale();
			WavelengthScale fws = filter.getLambda();
			if (!sgws.equals(fws)) {
				throw new IllegalArgumentException("wavelength scales not equal:  sgws="+sgws+" fws="+fws);
			}
		}
	}

	/** copy constructor
	 * @param astroObjectGenerator
	 */
	public AstroObjectGenerator(AstroObjectGenerator astroObjectGenerator) {
		coordinateGenerator = astroObjectGenerator.coordinateGenerator.copy();
		redshiftGenerator = astroObjectGenerator.redshiftGenerator.copy();	
		sedGenerator = astroObjectGenerator.sedGenerator.copy();
		morphologyGenerator = astroObjectGenerator.morphologyGenerator.copy();
		luminosityGenerator = astroObjectGenerator.luminosityGenerator.copy();
		simRandomPoisson = new SimRandomPoisson(astroObjectGenerator.simRandomPoisson.getIseedIndex());
		setFilter();
	}  

	/**
	 * @return the mean number of objects from this generator.
	 */
	public double getMeanNumber() {
		// how many steradians does this catalog cover?
		double steradians = this.coordinateGenerator.steradians();
		double areaFraction = steradians / (4 * Math.PI);  		
		double spaceDensity = this.luminosityGenerator.getSpaceDensity(); // in pc^-3;  
		if (!Double.isNaN(spaceDensity)) {
			double volume = this.redshiftGenerator.volume(); // in pc^3  		
			double meanNumber = volume * areaFraction * spaceDensity;
			return meanNumber;
		} 
		double surfaceDensity = luminosityGenerator.getSurfaceDensity(); // in degree^-2
		if (!Double.isNaN(surfaceDensity)) {
			double areaSquareDegrees = steradians * Math.pow(Math.toDegrees(1), 2);
			double meanNumber = areaSquareDegrees * surfaceDensity;
			return meanNumber;			
		}
		throw new Error("both space and surface density are NaN from luminosityGenerator="+luminosityGenerator);
	}

	/**
	 * @return the number of objects to generate for one instance of this generator.
	 * It selects a random number from a poisson distribution with mean of getMeanNumber()
	 */
	//public int nToGenerate() {
	//	double meanNumber = getMeanNumber();
	//	int retval = simRandomPoisson.next(meanNumber);
	//	return retval;
	//}

	/**
	 * @return a brief report of what this generator has
	 */
	public String getReport() {
		String retval = toString()+"\n";
		retval += "   "+this.coordinateGenerator.toString()+"\n";
		retval += "   "+this.redshiftGenerator.toString()+"\n";
		retval += "   "+this.sedGenerator.toString()+"\n";
		retval += "   "+this.morphologyGenerator.toString()+"\n";
		retval += "   "+this.luminosityGenerator.toString()+"\n";
		retval += "   mean number of objects = "+getMeanNumber();
		return retval;
	}
	/**
	 * @param luminosityGenerator
	 * @param redshiftGenerator
	 * @return the surfaceDensity of objects, per degree squared
	 */
	public static double calculateSurfaceDensity(SimRandomLuminosity luminosityGenerator, SimRandomRedshift redshiftGenerator) {
		double spaceDensity = luminosityGenerator.getSpaceDensity(); // objects/pc^3
		if (Double.isNaN(spaceDensity)) {
			double retval = luminosityGenerator.getSurfaceDensity();
			if (Double.isNaN(retval)) {
				throw new Error();
			}
			return retval;
		}
		double volume = redshiftGenerator.volume();                  // pc^3 in shell
		double surfaceDensity = spaceDensity*volume;                 // objects/sphere
		surfaceDensity /= 4.0*Math.PI;								 // objects/steradian
		surfaceDensity *= Math.pow(Math.toRadians(1), 2);			 // objects/degree squared
		double retval = surfaceDensity;
		if (Double.isNaN(retval)) {
			throw new Error();
		}
		return retval;
	}

	/**
	 * @see java.lang.Iterable#iterator()
	 */
	public Iterator<AstroObject> iterator() {
		return new AstroObjectIterator();
	}
	private class AstroObjectIterator implements Iterator<AstroObject> {

		Iterator<SphericalCoordinate> scIterator;
		AstroObjectIterator() {
			scIterator = coordinateGenerator.iterator();
		}
		/**
		 * reports whether there is more by calling SphericalCoordinateIterator.next
		 * @see java.util.Iterator#hasNext()
		 */
		public boolean hasNext() {
			return scIterator.hasNext();
		}

		/** 
		 * calls SphericalCoordinateIterator.next(), then fills in the redshft, sed, absoluteMagnitude, 
		 * and morphology, and returns the AstroObject constructed from all of these
		 * @see java.util.Iterator#next()
		 */
		public AstroObject next() {
			SphericalCoordinate sc = scIterator.next();
			double redshift =  Double.NaN;
			SED sed = null;
			Morphology morphology = null;
			double absoluteMagnitude = Double.NaN;
			redshift = redshiftGenerator.next(sc);
			sed = sedGenerator.next(sc);
			absoluteMagnitude = luminosityGenerator.next(sc);
			morphology = morphologyGenerator.next(sc, redshift, absoluteMagnitude);
			AstroObject source;
			if (redshiftGenerator.getDefinedAsDistancePc()) {
				source = new AstroObject(sc, 0.0, redshift, sed, morphology);
			} else {
				source = new AstroObject(sc, redshift, 0.0, sed, morphology);				
			}
			StopwatchCollection.stop("    AstroObject.next: new AstroObject");
			StopwatchCollection.start("    AstroObject.next: next luminosity");
			StopwatchCollection.stop("    AstroObject.next: next luminosity");
			StopwatchCollection.start("    AstroObject.next: setRestMagnitude");
			source.setAbsoluteMagnitude(filter, absoluteMagnitude);
			StopwatchCollection.stop("    AstroObject.next: setRestMagnitude");
			return source;
		}
		/**
		 * @throws UnsupportedOperationException
		 * @see java.util.Iterator#remove()
		 */
		public void remove() {
			throw new UnsupportedOperationException();
		}
	}

	/**
	 * @throws UnsupportedOperationException
	 */
	public void remove() {
		throw new UnsupportedOperationException();
	}
}

