package gov.fnal.eag.sim.catalog;

import gov.fnal.eag.sim.catalog.WavelengthScale.Scaling;
import gov.fnal.eag.sim.repository.image.ImageRepository;
import gov.fnal.eag.util.misc.Interpolator;
import gov.fnal.eag.util.misc.ResourceList;
import gov.fnal.eag.util.misc.StringUtil;
import gov.fnal.eag.util.plot.JFreeChartUtil;
import gov.fnal.eag.util.yanny.Table;

import java.awt.geom.Point2D;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.logging.Logger;
import java.util.regex.Pattern;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

import uk.ac.starlink.table.RowListStarTable;
import cern.colt.list.DoubleArrayList;
import cern.jet.stat.Descriptive;

/**
 * Represents the transmission curve and traditional calibration parameters for a filter.
 * 
 * @author stoughto
 * @version $Name$ ($Id$)
 */
public class AstronomicalFilter implements Iterable<ValuedBin> {

	/**
	 * Defines how the transmission is defined.
	 */
	public enum TransmissionType {
		/**
		 * shape of the transmission curve T_lambda = T_nu corresponds to energy
		 * transmitted
		 */
		ENERGY,
		/**
		 * shape of the transmission curve corresponds to number of photons
		 * transmitted
		 */
		PHOTONS,
		/**
		 * shape of the transmission curve is used for the D4000 calibration
		 */
		D4000;

		/**
		 * @param pegaseTransmissionType
		 * @return ENERGY, PHOTONS, or D4000 for 0, 1, 2, respectively
		 */
		public static TransmissionType get(int pegaseTransmissionType) {
			switch (pegaseTransmissionType) {
			case 0:
				return ENERGY;
			case 1:
				return PHOTONS;
			case 2:
				return D4000;
			default:
				throw new IllegalArgumentException("can not deal with pegaseTransmissionType="+pegaseTransmissionType);
			}
		}
	}

	private TransmissionType transmissionType;

	private CalibrationType calibrationType;

	private String name;

	private String description;

	private double[] meanTransmission;

	private WavelengthScale lambda;

	private ArrayList<Point2D.Double> points;
	private double transmissionMax;
	/**
	 * create a filter with one skinny wavelength bin centered at lambda with the specified transmission
	 * @param lambda mean wavelength of the bin in angstrom
	 * @param transmission usually 1.0
	 */
	
	public AstronomicalFilter(double lambda, double transmission) {
		this.lambda = new WavelengthScale(lambda-1e-9, lambda+1e-9, 1, Scaling.LINEAR);
		meanTransmission = new double[] {transmission};	
		this.transmissionType = TransmissionType.PHOTONS;
		this.calibrationType = CalibrationType.AB;
	}
	
	/**
	 * @param name
	 * @param points
	 * @param scale
	 */
	public AstronomicalFilter(String name, ArrayList<Point2D.Double> points, WavelengthScale scale) {
		this.name = name;
		this.points = points;
		this.lambda = scale;
		this.transmissionType = TransmissionType.PHOTONS;
		this.calibrationType = CalibrationType.AB;
		init();
	}
	/**
	 * constructs a filter from the pre-defined list of filters, which is
	 * returned by getNames()
	 * 
	 * @param name
	 *            name of the filter
	 * @param scale
	 *            wavelength scale
	 * 
	 */
	public AstronomicalFilter(String name, WavelengthScale scale) {
		this.name = name;
		loadPoints();
		lambda = scale;
		init();
	}

	/**
	 * multiplies the filters together to make an effective transmission
	 * @param filters
	 */
	public AstronomicalFilter(final AstronomicalFilter[] filters) {
		if (filters == null) {
			throw new IllegalArgumentException("filters can not be null");
		}
		if (filters.length == 0) {
			throw new IllegalArgumentException("must have at least one filter");
		}
		this.lambda = filters[0].getLambda();
		this.transmissionType = filters[0].getTransmissionType();
		this.calibrationType = filters[0].getCalibrationType();
		this.name = "";
		this.description = "product of ";
		meanTransmission = new double[lambda.size()];
		for (int i=0; i<meanTransmission.length; i++) {
			meanTransmission[i] = 1.0;
		}
		for (AstronomicalFilter filter : filters ) {
			if (!filter.getLambda().equals(lambda)) {
				throw new IllegalArgumentException("all wavelength scales need to be the same");
			}
			if (transmissionType == null) {
				if (filter.getTransmissionType() != null) {
					throw new IllegalArgumentException("all transmissionType need to be the same");					
				}
			} else if (!filter.getTransmissionType().equals(transmissionType)) {
				throw new IllegalArgumentException("all transmissionType need to be the same");
			}
			if (!filter.getCalibrationType().equals(calibrationType)) {
				throw new IllegalArgumentException("all calibrationType need to be the same");
			}
			for (int i=0; i<meanTransmission.length; i++) {
				meanTransmission[i] *= filter.meanTransmission[i];
			}			
			name += filter.name + " ";
			description += filter.name + " ";
		}

	}

	private void init() {
		meanTransmission = new double[lambda.size()];
		int i = 0;
		for (WavelengthBin wb : lambda) {
			try {
				meanTransmission[i] = wb.integrate(points)/ wb.binWidth();
			} catch (IntegralNotDefinedException e) {
				meanTransmission[i] = 0.0;
			}
			i++;
		}
		// Now that everything is filled in, we can integrate over the standard
		// flux
		calibrationType.setDenominator(this);
		transmissionMax = 0.0;
		for (Point2D.Double point : points) {
			transmissionMax = Math.max(transmissionMax, point.y);
		}

	}
	/**
	 * load the points
	 */
	private void loadPoints() {
		String[] pres = {"filter-", ""};
		for (String pre : pres) {
			String baseFileName = "data"
				+ File.separatorChar + "filters" + File.separatorChar
				+ pre + name;
			InputStream inputStream;
			inputStream = Spectrum.class.getResourceAsStream(baseFileName + ".dat");
			if (inputStream != null) {
				loadPointsText(inputStream);
				return;
			} else {
				inputStream = AstronomicalFilter.class.getResourceAsStream(baseFileName + ".fit");
				if (inputStream != null) {
					RowListStarTable rlst;
					try {
						rlst = ImageRepository.readRlst(inputStream, 0);
					} catch (IllegalArgumentException e) {
						throw new IllegalArgumentException("Can not open baseFileName="+baseFileName);
					}
					this.description = rlst.getParameterByName("FDESCRIP")
					.getValueAsString(100);
					this.transmissionType = TransmissionType.get(Integer.parseInt(rlst
							.getParameterByName("TRNSTYPE").getValueAsString(100)));
					this.calibrationType = CalibrationType.get(Integer.parseInt(rlst
							.getParameterByName("CLBRTYPE").getValueAsString(100)));
					int transmissionIndex = Table.getIndexByName(rlst, "transmission");
					int wavelengthIndex = Table.getIndexByName(rlst, "wavelength");
					points = new ArrayList<Point2D.Double>();
					for (int i = 0; i < rlst.getRowCount(); i++) {
						Object[] row = rlst.getRow(i);
						points.add(new Point2D.Double((Double) row[wavelengthIndex], (Double) row[transmissionIndex]));
					}
					return;
				}
			}
		}
		throw new IllegalArgumentException("can not loadPoints for name="+name);
	}
	private void loadPointsText(InputStream inputStream) {
		points = new ArrayList<Point2D.Double>();
		if (name==null) {
			throw new IllegalArgumentException();
		}
		BufferedReader bufferedReader = new BufferedReader(
				new InputStreamReader(inputStream));
		// Read the file one line at a time
		String line;
		try {
			while ((line = bufferedReader.readLine()) != null) {
				String[] tokens = StringUtil.tokenize(line, " ");
				if (!tokens[0].startsWith("#")) {
					if (tokens.length == 2) {
						double lambdaTemp = Double.parseDouble(tokens[0]);
						double transmission = Double.parseDouble(tokens[1]);
						points.add(new Point2D.Double(lambdaTemp, transmission));
					}
				}
			}
		} catch (NumberFormatException e) {
			throw new Error(e);
		} catch (IOException e) {
			throw new Error(e);
		}
		calibrationType = CalibrationType.STANDARD;
	}

	/**
	 * @return the points as read in by loadPoints()
	 */
	public ArrayList<Point2D.Double> getPoints() {
		return points;
	}
	/**
	 * constructs a filter from the pre-defined list of filters, which is
	 * returned by getNames()
	 * 
	 * @param name
	 *            name of the filter
	 * @param nBins
	 *            the number of bins
	 * 
	 */
	public AstronomicalFilter(String name, int nBins)  {
		this.name = name;
		loadPoints();
		double min = Double.MAX_VALUE;
		double max = -min;
		for (Point2D.Double point : points) {
			min = Math.min(min, point.x);
			max = Math.max(max, point.x);
		}
		WavelengthScale scale = new WavelengthScale(min, max, nBins,WavelengthScale.Scaling.LINEAR);
		this.lambda = scale;
		this.transmissionType = TransmissionType.PHOTONS;
		this.calibrationType = CalibrationType.AB;
		init();		
	}

	/**
	 * Creates a scale using the extreme values in a file
	 * 
	 * @param name
	 *            name of the filter
	 * @param nBins
	 *            the number of bins
	 * @return the new wavelength scale
	 * 
	 */
	public static final WavelengthScale deriveScale(String name, int nBins) {		
		InputStream inputStream = AstronomicalFilter.class.getResourceAsStream("data"
				+ File.separatorChar + "filters" + File.separatorChar
				+ "filter-" + name + ".fit");
		RowListStarTable rlst = ImageRepository.readRlst(inputStream, 0);
		int wavelengthIndex = Table.getIndexByName(rlst, "wavelength");
		Object[] row = rlst.getRow(0);
		double min = (Double) row[wavelengthIndex];
		row = rlst.getRow(rlst.getRowCount() - 1);
		double max = (Double) row[wavelengthIndex];
		WavelengthScale scale = new WavelengthScale(min, max, 512,
				WavelengthScale.Scaling.LINEAR);
		return scale;
	}

	/**
	 * A fully specified constructor.
	 * 
	 * @param name the name of the filter
	 * @param description a description of the filter
	 * @param transmissionType the transmission type for the defining curve
	 * @param calibrationType the calibration type used for this filter
	 * @param scale the wavelength scale on which to generate the transmission curve
	 * @param transmission the values for the transmission for each bin of the wavelength scale
	 */
	public AstronomicalFilter(String name, String description,
			TransmissionType transmissionType, CalibrationType calibrationType,
			WavelengthScale scale, DoubleArrayList transmission)
	{
		this.transmissionType = transmissionType;
		this.calibrationType = calibrationType;
		this.name = name;
		this.description = description;
		this.lambda = scale;
		this.meanTransmission = new double[transmission.size()];
		for (int i=0; i<meanTransmission.length; i++) {
			meanTransmission[i] = transmission.get(i);
		}
		calibrationType.setDenominator(this);
	}

	/**
	 * A fully specified constructor
	 * 
	 * @param name the name of the filter
	 * @param description a description of the filter
	 * @param transmissionType the transmission type for the defining curve
	 * @param calibrationType the calibration type used for this filter
	 * @param wavelength the wavelength scale on which to generate the transmission curve
	 * @param transmission the values for the transmission for each bin of the wavelength scale
	 */
	public AstronomicalFilter(String name, String description,
			TransmissionType transmissionType, CalibrationType calibrationType,
			WavelengthScale wavelength, double[] transmission) {
		this(name, description, transmissionType, calibrationType, wavelength,
				new DoubleArrayList(transmission));
	}

	// Accessor Methods
	/**
	 * Get the value of transmission
	 * 
	 * @return the value of transmission
	 */
	public DoubleArrayList getMeanTransmission() {
		return new DoubleArrayList(meanTransmission);
	}

	/**
	 * Set the mean transmission values for each bin in the wavelength scale. 
	 * 
	 * @param value
	 *            A list of transmissions
	 *            corresponding to bins in the wavelength scale.
	 */
	public void setMeanTransmission(DoubleArrayList value) {
		meanTransmission = value.elements();
	}

	/**
	 * Get the name of the filter.
	 * 
	 * @return name the name of this filter
	 */
	public String getName() {
		return name;
	}

	/**
	 * @param name the name to set
	 */
	public void setName(String name) {
		this.name = name;
	}
	/**
	 * Get a description of this filter.
	 * 
	 * @return the description of this filter
	 */
	public String getDescription() {
		return description;
	}

	/**
	 * Get the wavelength scale on which the transmission curve for this filter is defined.
	 * 
	 * @return lambda the wavelength scale
	 */
	public WavelengthScale getLambda() {
		return lambda;
	}

	/**
	 * get a list of Filter names available
	 * @return a Collection of names, each of which is suitable for use in a Filter constructor
	 */
	public static Collection<String> getNames() {
		ArrayList<String> retval = new ArrayList<String>();
		String[] types = {"fit", "dat"};
		for (String type : types) {
			{
				Collection<String> resourceList = ResourceList.getResources(Pattern.compile(".*gov/fnal/eag/sim/catalog/data/filters/.*."+type));
				for (String resource : resourceList) {
					String[] tokens = resource.split("/");
					retval.add(tokens[tokens.length-1].replace("filter-","").replace("."+type,""));
				}
			}
		}
		Collections.sort(retval);
		return retval;
	}

	/**
	 * Returns whether a file name follows the conventions for naming files that define filter transmission curves.
	 * 
	 * @param name file name to check
	 * @return true if it does
	 */
	public static boolean acceptFile(String name) {
		if (name.endsWith(".fit") && name.startsWith("filter"))
			return true;
		return false;
	}

	/**
	 * Get the calibration type for this filter.
	 * 
	 * @return the calibration type
	 */
	public CalibrationType getCalibrationType() {
		return calibrationType;
	}

	/**
	 * Get the transmission representation used by this filter.
	 * 
	 * @return the transmission type
	 */
	public TransmissionType getTransmissionType() {
		return transmissionType;
	}

	/**
	 * Create a plot of transmission vs wavelength for all available filters.
	 * 
	 * @param min
	 *            minimum mean wavelength of filters to include, in Angstrom
	 * @param max
	 *            maximum mean wavelength of filters to include, in Angstrom
	 * @return plot transmission vs wavelength for all of the filters
	 * @throws Exception
	 */
	public static JFreeChart plotAllFilters(double min, double max)
	throws Exception {
		Logger logger = Logger.getLogger("gov.fnal.eag.sim");
		XYSeriesCollection xySeriesCollection = new XYSeriesCollection();
		for (String name : getNames()) {
			XYSeries xySeries = new XYSeries(name);
			AstronomicalFilter filter = new AstronomicalFilter(name, 512);
			double mean = filter.getMeanWavelength();
			if (mean > min && mean < max) {
				logger.fine("name=" + name);
				for (ValuedBin bin : filter) {
					xySeries.add(bin.getWavelengthBin().meanLambda(), bin
							.getValue());
				}
				xySeriesCollection.addSeries(xySeries);
			}
		}
		JFreeChart chart = JFreeChartUtil.makeChart(xySeriesCollection, "AstronomicalFilter with "+min+"< mean wavelength < "+max, "Angstrom", "transmission", null);
		JFreeChartUtil.linesShapesVisible(chart, true, false);
		return chart;
	}

	/**
	 * Return a plot of transmission against wavelength for this filter.
	 * 
	 * @return plot of transmission vs wavelength for this Filter
	 */
	public JFreeChart getChart() {
		return getChart((Double) null, (Double) null);
	}

	/**
	 * @param filters
	 * @param chartName
	 * @return the chart
	 */


	public static JFreeChart getChart(Iterable<AstronomicalFilter> filters,  String chartName) {
		XYSeriesCollection xysc = new XYSeriesCollection();
		for (AstronomicalFilter filter : filters) {
			XYSeries xySeries = getSeries(filter, filter.getName());
			xysc.addSeries(xySeries);
		}
		JFreeChart chart = ChartFactory.createScatterPlot(chartName, "wavelength",
				"transmission", xysc, PlotOrientation.VERTICAL,
				true, true, true);
		JFreeChartUtil.linesShapesVisible(chart, true, false);
		return chart;
	}




	/**
	 * plot transmission vs. wavelength (microns)
	 * @param minIn minimum wavelength to plot (microns)
	 * @param maxIn maximum wavelength to plot (microns)
	 * @return the chart
	 */
	public JFreeChart getChart(Double minIn, Double maxIn) {
		double min,max;
		if (minIn == null) {
			min = -Double.MAX_VALUE;
		} else {
			min = minIn;
		}
		if (maxIn == null) {
			max = Double.MAX_VALUE;
		} else {
			max = maxIn;
		}
		XYSeriesCollection xySeriesCollection = new XYSeriesCollection();
		XYSeries xySeries = new XYSeries(name);
		for (ValuedBin bin : this) {
			double wavelength = bin.getWavelengthBin().meanLambda()/1e4;	
			if (wavelength > min && wavelength < max) {
				xySeries.add(wavelength, bin.getValue());
			}
		}
		xySeriesCollection.addSeries(xySeries);

		JFreeChart chart = ChartFactory.createScatterPlot(name, "wavelength",
				"transmission", xySeriesCollection, PlotOrientation.VERTICAL,
				true, true, true);
		XYLineAndShapeRenderer xyItemRenderer = (XYLineAndShapeRenderer) chart
		.getXYPlot().getRendererForDataset(xySeriesCollection);
		xyItemRenderer.setSeriesLinesVisible(0, true);
		xyItemRenderer.setSeriesShapesVisible(0, false);
		return chart;
	}

	/**
	 * Return the maximum wavelength covered by the transmission curver for this filter.
	 * 
	 * @return the minimum wavelength
	 */
	public double getMinWavelength() {
		return lambda.getLambdaMin();
	}

	/**
	 * Return the minimun wavelength covered by the transmission curve for this filter.
	 * 
	 * @return the maximum wavelength
	 */
	public double getMaxWavelength() {
		return lambda.getLambdaMax();
	}

	/**
	 * Return a DoubleArrayList of the mean wavelengths of the ranges over which the transmission is stored.
	 * 
	 * @return the mean wavelengths for the bins
	 */
	public DoubleArrayList wavelength() {
		DoubleArrayList wavelength = new DoubleArrayList();
		for (ValuedBin bin : this) {
			wavelength.add(bin.getWavelengthBin().meanLambda());
		}
		return wavelength;
	}

	/**
	 * WHAT IS THIS SUPPOSED TO DO, AND HOW DOES IT RELATE TO {@link #wavelength()}?
	 * 
	 * @return the mean wavelength
	 */
	public double getMeanWavelength() {
		return new Double(Descriptive.weightedMean(wavelength(),
				new DoubleArrayList(meanTransmission)));
	}

	/**
	 * An iterator over the wavelength bins in the filter
	 * 
	 * @return an iterator over valued bins
	 */
	public ValuedBinIterator iterator() {
		Double[] t = new Double[meanTransmission.length];
		int index = 0;
		for (double d : meanTransmission) {
			t[index++] = d;
		}
		return new ValuedBinIterator(t, lambda);
	}

	/**
	 * list the number of names filters, and then plot them all
	 * 
	 * @param args
	 *            ignored
	 * @throws Exception 
	 */

	public static void main(String[] args) throws Exception {
		Collection<String> list = AstronomicalFilter.getNames();
		System.out.println(" number of defined filters:  " + list.size());
		{
			JFreeChart chart = AstronomicalFilter.plotAllFilters(0, 5e3);		
			String fileName = StringUtil.getFileName("-1","png");
			ChartUtilities.saveChartAsPNG(new File(fileName), chart, 800, 400);		
		}
		{
			JFreeChart chart = AstronomicalFilter.plotAllFilters(5e3, 5e4);
			String fileName = StringUtil.getFileName("-2","png");
			ChartUtilities.saveChartAsPNG(new File(fileName), chart, 800, 400);		
		}
		{
			JFreeChart chart = AstronomicalFilter.plotAllFilters(5e4, 5e10);
			String fileName = StringUtil.getFileName("-3","png");
			ChartUtilities.saveChartAsPNG(new File(fileName), chart, 800, 400);		
		}

	}
	/**
	 * @return the number of wavelength bin/transmission pairs in this Filter
	 */
	public int size() {
		return meanTransmission.length;
	}
	/**
	 * interpolate linearly to get the value of the transmission at the location
	 * @param d the location
	 * @return the transmission, or 0 if d is out of range
	 */
	public double getValueAt(double d) {
		double[] lambdas = lambda.getMeanLambdas();
		return Interpolator.linear(lambdas, meanTransmission, d);		
	}
	/**
	 * get a series for plotting
	 * @param filter
	 * @param legend
	 * @return the series
	 */
	public static XYSeries getSeries(AstronomicalFilter filter, String legend) {
		XYSeries retval = new XYSeries(legend);
		Scaling scaling = filter.getLambda().getScaling();
		for (ValuedBin vb : filter) {
			double lambda = vb.getWavelengthBin().getMeanLambda(scaling);
			double value = vb.getValue();
			if (value > 0.0) {
				retval.add(lambda, value);
			}
		}
		return retval;
	}
	/**
	 * @return the minimum non-zero wavelength
	 */
	public double getMinNonzeroWavelength() {
		for (ValuedBin vb : this) {
			if (vb.getValue() != 0) {
				return vb.getWavelengthBin().getLambdaMin();
			}
		}
		return lambda.getLambdaMin();
	}

	/**
	 * @return the maximum non-zero wavelength
	 */
	public double getMaxNonzeroWavelength() {
		ValuedBin maxVb = null;
		for (final ValuedBin vb : this) {
			if (vb.getValue() != 0) {
				maxVb = vb;
			}
		}
		if (maxVb == null) {
			return lambda.getLambdaMax();
		}
		return maxVb.getWavelengthBin().getMeanLambda(lambda.getScaling());
	}	

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		String retval = "AstronomicalFilter:  name="+getName()+" transmissionType="+transmissionType;
		return retval;
	}
	/**
	 * @return the area of this filter, in Angstrom, normalized so the max input point has transmission of  1.0
	 * integral transmission(lambda) dLambda
	 */
	public double getArea() {
		double retval = 0.0;
		for (ValuedBin vb : this) {
			double value = vb.getValue();
			retval += value * vb.getWavelengthBin().binWidth();
		}
		retval /= transmissionMax;
		return retval;
	}

	/**
	 * @param sed
	 * @return integral of wavelength*transmission*flux
	 */
	public double getEffectiveWavelength(SED sed) {
		if (sed == null) {
			throw new IllegalArgumentException("sed can not be null");
		}
		if (!lambda.equals(sed.getLambda())) {
			throw new IllegalArgumentException("ws not equal:  lambda="+lambda+" sed.getLambda="+sed.getLambda());
		}
		double numerator = 0.0;
		double denominator = 0.0;
		for (ValuedBin vb : this) {
			double wavelength = vb.getWavelengthBin().getMeanLambda(lambda.getScaling());
			double dWavelength = vb.getWavelengthBin().binWidth();
			double transmission = this.getValueAt(wavelength);
			double flux = sed.getFlux(wavelength);
			numerator += transmission * flux * wavelength * dWavelength;
			denominator += transmission * flux * dWavelength;
		}
		double retval = numerator/denominator;
		return retval;
	}

	/**
	 * @param filters
	 * @param string
	 * @return the chart
	 */
	public static JFreeChart getChart(AstronomicalFilter[] filters, String string) {
		ArrayList<AstronomicalFilter> temp = new ArrayList<AstronomicalFilter>(Arrays.asList(filters));
		return getChart(temp, string);
	}
}
