package gov.fnal.eag.sim.catalog;

import gov.fnal.eag.wcs.SphericalCoordinate;

import java.lang.Math;

import uk.ac.starlink.util.PhysicalConstants;

/**
 * Generates a blackbody SED for a specified temperature, on a given wavelength scale.
 * 
 * @author neilsen
 * @version $Name:  $ ($Id: BlackBodySEDGenerator.java,v 1.11 2008/08/04 17:06:57 descvs Exp $)
 */
public class BlackBodySEDGenerator implements SEDGenerator {
	// Fields
	private double temperature = 0; // degrees kelvin
	private WavelengthScale scale = null;
 	
	/**
	 * Create a new blackbady SED generator
	 * 
	 * @param temperature the temperature (degrees K)
	 * @param scale the wavelength scale
	 */
	public BlackBodySEDGenerator(double temperature, WavelengthScale scale) {
		setTemperature(temperature);
		setScale(scale);
	}
	// Accessor Methods
	
	/**
	 * Set a new temperature
	 * 
	 * @param temperature the temperature (degrees K)
	 */
	public void setTemperature(double temperature) {
		this.temperature = temperature;
	}
	
	/**
	 * Get the temperature
	 * 
	 * @return the temperature (degrees K)
	 */
	public double getTemperature() {
		return temperature;
	}
	
	/**
	 * Set the wavelength scale
	 * 
	 * @param scale the new scale
	 */
	public void setScale(WavelengthScale scale) {
		this.scale = scale;
	}
	
	// Operations
	/**
	 * return the SED for the temperature
	 * @return the SED for this temperature
	 */
	public SED next() {
		double[] flux = new double[scale.size()];
		int fluxIndex = 0;
		double lambda = 0;
		
		// Physical constants
		double h = PhysicalConstants.PLANCK; // Planks constant, in J*sec
		double c = PhysicalConstants.SPEED_OF_LIGHT; // speed of light, m/s
		double k = PhysicalConstants.BOLTZMANN; // Boltzmans constant, J/K
		
		for(WavelengthBin bin: scale) {
			lambda = 1e-10*bin.meanLambda();
			double d1 = Math.exp(h*c/(lambda*k*temperature)) - 1.0;
			double d2 = (1e-10*bin.binWidth())*2.0*Math.PI*h*c*c/(Math.pow(lambda,5)*d1); // in Watts/m^2
			// Flux needs to be in ergs/sec/cm^2
			flux[fluxIndex++] = d2 * 1.0E3;
		}
		return new SED(flux,scale);
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.SEDGenerator#copy()
	 */
	public BlackBodySEDGenerator copy() {
		BlackBodySEDGenerator retval = new BlackBodySEDGenerator(temperature, new WavelengthScale(scale));
		return retval;
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.SEDGenerator#getISeed()
	 */
	public int getISeed() {
		return 0;
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.SEDGenerator#getWavelengthScale()
	 */
	public WavelengthScale getWavelengthScale() {
		return scale;
	}
	/**
	 * @see gov.fnal.eag.sim.catalog.SEDGenerator#next(boolean)
	 */
	public SED next(boolean returnValue) {
		SED retval = null;
		if (returnValue) {
			retval = next();
		}
		return retval;
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.SEDGenerator#next(gov.fnal.eag.wcs.SphericalCoordinate)
	 */
	public SED next(SphericalCoordinate sc) {
		return next();
	}
	
}

