package gov.fnal.eag.sim.catalog;

/**
 * defines how the flux,magnitude conversion is done
 * @author stoughto
 *
 */
public enum CalibrationType {
	/**
	 * defines the flux to magnitude conversion
	 * 
	 * @author stoughto
	 */
	/**
	 * mag = integral(4050,4250)[F_nu dLambda] / integral(3750,3950)[F_nu
	 * dLambda] Bruzual 1983
	 */
	D4000("unity", 0.0),
	/**
	 * mag =
	 * -2.5*log10(integral[F_lambda*T_lambda*dLambda]/integral[F_lambda(Vega)*T_Lambda*dLambda]) +
	 * 0.03
	 */
	STANDARD("vega", 0.03),
	/**
	 * mag = -2.5*log10(integral[F_nu*T_nu*dNu]/integral[F_nu*dNu]) - 48.60 F_nu
	 * in erg/sec/cm^2/Hz
	 */
	AB("unity", -48.6),
	/**
	 * mag =
	 * -2.5*log10(integral[F_lambda*T_lambda*dLambda]/integral[F_lambda(BD+174708)*T_Lambda*dLambda]) +
	 * 9.5
	 */
	THUAN("bd17", 9.5),
	/**
	 * mag =
	 * -2.5*log10(integral[F_lambda*T_lambda*dLambda]/integral[T_Lambda*dLambda]) -
	 * 21.10 F_lambda in erg/sec/cm^2/A
	 */
	WFPC2("unity", -21.10),
	/**
	 * mag =
	 * -2.5*log10(integral[F_lambda*T_lambda*dLambda]/integral[T_Lambda*dLambda]) -
	 * 21.175 F_lambda in erg/sec/cm^2/A
	 */
	FOCA("unity", -21.175);
	private final String standardFlux;

	private final double zeropoint;

	private double denominator = 0;

	private static SED standardFluxVega = null;

	private static SED standardFluxBd17 = null;
	
	CalibrationType(String standardFlux, double zeropoint) {
		this.standardFlux = standardFlux;
		this.zeropoint = zeropoint;
	}

	/**
	 * @return a clue about what to use as a standard flux (vega for STANDARD,
	 *         bd18 for Thuan, unity for the rest)
	 */
	public String getStandardFlux() {
		return standardFlux;
	}

	/**
	 * @return the zeropoint for this CalibrationType
	 */
	public double getZeropoint() {
		return zeropoint;
	}

	/**
	 * do the integral that is in the denominator of the magnitude conversion
	 * @param filter filter to use
	 */
	public void setDenominator(AstronomicalFilter filter) {
		SED standardFluxSed;
		WavelengthScale ws = filter.getLambda();
		if (standardFlux == "unity") {
			double[] one = new double[ws.size()];
			for (int i = 0; i < one.length; i++) {
				one[i] = 1.0;
			}
			standardFluxSed = new SED(one, ws);
		} else if (standardFlux == "vega") {
			if (standardFluxVega == null || !standardFluxVega.getLambda().equals(ws)) {
				standardFluxVega = new SED(ws, "alpha_lyr_stis_002");
			}
			standardFluxSed = standardFluxVega; // flux is in erg/sec/cm^2/A
		} else if (standardFlux == "bd17" || !standardFluxBd17.getLambda().equals(ws)) {
			if (standardFluxBd17 == null) {
				standardFluxBd17 = new SED(ws, "bd17d4708_stis_001");
			}
			standardFluxSed = standardFluxBd17;
		} else {
			standardFluxSed = new SED(ws, standardFlux);
		}
		ValuedBinIterator standardFluxIterator = standardFluxSed.iterator();
		denominator = 0;
		ValuedBin standardFluxBin = null;
		for (ValuedBin filterBin : filter) {
			standardFluxBin = standardFluxIterator.next();
			denominator += standardFluxBin.getValue() * filterBin.getValue();
		}
	}

	/**
	 * @return the appropriate denominator inside the -2.5*Math.log10() call for
	 *         calculating a magnitude
	 */
	public double getDenominator() {
		return denominator;
	}

	/**
	 * get a Calibration from the integer type used in the Pegase package
	 * @param pegaseType the type to get
	 * @return the corresponding CalibrationType
	 * @throws UnsupportedOperationException
	 */
	public static CalibrationType get(int pegaseType)
			throws UnsupportedOperationException {
		switch (pegaseType) {
		case 0:
			return D4000;
		case 1:
			return STANDARD;
		case 2:
			return AB;
		case 3:
			return THUAN;
		case 4:
			return WFPC2;
		case 5:
			return FOCA;
		default:
			throw new UnsupportedOperationException("unsupported pegaseType="
					+ pegaseType);
		}
	}
}
