package gov.fnal.eag.sim.catalog;

import java.util.Iterator;
import java.util.NoSuchElementException;

import gov.fnal.eag.sim.catalog.CoordinateGenerator;
import gov.fnal.eag.util.htm.ConvexRegion;
import gov.fnal.eag.wcs.SphericalCoordinate;


/**
 * generate coordinate on a uniform grid inside the detector region
 */
public class GridCoordinateGenerator implements CoordinateGenerator {


	SphericalCoordinate detectorMinRaDec, detectorMaxRaDec, gridCenter;
	int gridSize;
	double gridDeltaArcsec, cosDec;
	ConvexRegion convexRegion;
	int size = 0;
	double ra0, dec0;
	int raIndex, decIndex;
	int nDone;
	/**
	 * @param detectorMinRaDec defines where the detector is on the sky
	 * @param detectorMaxRaDec defines where the detector is on the sky
	 * @param gridCenter location of the center point of the grid
	 * @param gridSize number of grid points in each direction
	 * @param gridDeltaArcsec spacing between grid points, with tangential projection at gridCenter
	 */
	public GridCoordinateGenerator(SphericalCoordinate detectorMinRaDec, SphericalCoordinate detectorMaxRaDec, 
			SphericalCoordinate gridCenter, int gridSize, double gridDeltaArcsec) {
		if (gridSize <=0 ) {
			throw new IllegalArgumentException("gridSize needs to be > 0 but gridSize="+gridSize);
		}
		if (gridDeltaArcsec <=0 ) {
			throw new IllegalArgumentException("gridDeltaArcsec needs to be > 0 but gridDeltaArcsec="+gridDeltaArcsec);
		}
		this.detectorMinRaDec = detectorMinRaDec;
		this.detectorMaxRaDec = detectorMaxRaDec;
		this.gridCenter = gridCenter;
		this.gridSize = gridSize;
		this.gridDeltaArcsec = gridDeltaArcsec;
		convexRegion = new ConvexRegion(detectorMinRaDec, detectorMaxRaDec);
		cosDec = Math.cos(Math.toRadians(gridCenter.getLatitude()));
		ra0 = gridCenter.getLongitude() - (gridDeltaArcsec/3600.0/cosDec)*(gridSize-1)/2;
		dec0 = gridCenter.getLatitude() - (gridDeltaArcsec/3600.0)*(gridSize-1)/2;
		for (int i=0; i<gridSize; i++) {
			for (int j=0; j<gridSize; j++) {
				if (get(i,j) != null) {
					size++;
				}
			}
		}
		raIndex = 0;
		decIndex = 0;
	}

	private SphericalCoordinate get(int i, int j) {
		double ra = ra0 + i*gridDeltaArcsec/3600.0/cosDec;
		double dec = dec0 + j*gridDeltaArcsec/3600.0;
		SphericalCoordinate sc = new SphericalCoordinate(ra,dec);
		if (convexRegion.contains(sc)) {
			return sc;
		}
		return null;
		
	}

	private SphericalCoordinate getNext() {
		SphericalCoordinate retval = get(raIndex, decIndex);
		raIndex++;
		if (raIndex >= gridSize) {
			raIndex = 0;
			decIndex++;
			if (decIndex > gridSize) {
				throw new NoSuchElementException();
			}
		}
		return retval;
	}
	/** (non-Javadoc)
	 * @see gov.fnal.eag.sim.catalog.CoordinateGenerator#steradians()
	 */
	public double steradians() {
		return SphericalCoordinate.steradians(detectorMinRaDec, detectorMaxRaDec);
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.CoordinateGenerator#copy()
	 */
	public GridCoordinateGenerator copy() {
		GridCoordinateGenerator retval = new GridCoordinateGenerator(detectorMinRaDec, detectorMaxRaDec, gridCenter, gridSize, gridDeltaArcsec);
		return retval;
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.CoordinateGenerator#getISeed()
	 */
	public int getISeed() {
		return 0;
	}

	/**
	 * @return the number of coordinates that will be returned by this generator
	 */
	public int size() {
		return size;
	}

	/**
	 * @see java.lang.Iterable#iterator()
	 */
	public Iterator<SphericalCoordinate> iterator() {
		return new GridCoordinateIterator();
	}
	private class GridCoordinateIterator implements Iterator<SphericalCoordinate> {

		public boolean hasNext() {
			return (nDone < size);
		}

		public SphericalCoordinate next() {
			SphericalCoordinate retval;
			while ((retval = getNext()) == null) {
			}
			nDone++;
			return retval;
		}
		public void remove() {
			throw new UnsupportedOperationException();
		}
	}
}

