package gov.fnal.eag.sim.catalog;

import gov.fnal.eag.sim.pixel.DetectorObject;
import gov.fnal.eag.sim.pixel.Profiles;
import gov.fnal.eag.util.astronomy.Cosmo;

/**
 * The shape of an astronomical source.
 * 
 * @author neilsen
 * @version $Name$ ($Id$)
 */
public abstract class Morphology {
	protected boolean distanceIsRedshift = true;
	protected double size = 0.; // The object size in arcsec
	protected double distance = 0.; // The red shift of the object, or distance in pc if distanceIsRedshift is false
	protected static Profiles profiles = null;
	
	protected Morphology() {		
	}
	/**
	 * @param that
	 * @return true if this equals that, without chatter
	 */
	abstract public boolean equals(Morphology that);
	
	/**
	 * @param that
	 * @param chatty if true, send message to System.out to say why they are not equal
	 * @return true if this equals that
	 */
	public abstract boolean equals(Morphology that, boolean chatty);
	/**
	 * apply the weak lensing shear
	 * @param gamma1
	 * @param gamma2
	 */
	abstract public void shear(double gamma1, double gamma2);

	/**
	 * @param theta in degrees
	 */
	abstract public void rotate(double theta);

	/**
	 * @return a deep copy of the original
	 */
	abstract public Morphology copy();
	/**
	 * makes a DetectorObject from this Morphology, for the given pixel scale
	 * @param row location on the detector
	 * @param col location on the detector
	 * @param counts number of observed counts
	 * @param pixelSizeArcseconds
	 * @return the resulting object
	 */
	abstract public DetectorObject makeDosb(double row, double col, double counts, double pixelSizeArcseconds);
	/**
	 * @param size set the size of the object in arcsec without changing the distance
	 */
	public void setSize(double size) {
		this.size = size;  
	}
	/**
	 * @param distance the distance to set without changing the size
	 */
	void setDistance(double distance) {
		this.distance = distance;
	}
	
	/**
	 * move this object to a new distance and change the size appropriately
	 * @param newDistance the new location
	 * @param cosmo cosmology to use to figure out how to change the size
	 */
	public void move(double newDistance, Cosmo cosmo) {
		if (newDistance <= 0.0) {
			throw new IllegalArgumentException("newDistance must be > 0.0 but newDistance="+newDistance);
		}
		double ratio;
		if (distanceIsRedshift) {
			double currentAngularDistance = cosmo.angularDistance(distance);
			double newAngularDistance = cosmo.angularDistance(newDistance);
			ratio = currentAngularDistance/newAngularDistance;
		} else {
			ratio = distance/newDistance;
		}
		size *= ratio;
		distance = newDistance;
	}
	
	/**
	 * @return object size in arcsec
	 */
	public double getSize() {
		return this.size;
	}
	/**
	 * @return red shift of the object
	 */
	public double getDistance() {
		return this.distance;
	}
	
	/**
	 * calculate the comoving size of the object in mega parsecs
	 * @param cosmology the cosmology to use
	 * @return the size in Mpc
	 */
	public double getSizeInMpc(Cosmo cosmology) {
		double sizeRadians = Math.toRadians(size/3600);
		double angularDistance = cosmology.angularDistance(distance);
		double angularDistanceMpc = angularDistance*cosmology.getCH0()*1e-6;
		double sizeMpc = sizeRadians*angularDistanceMpc;
		return sizeMpc;
	}
	/**
	 * redshift or pc
	 * @return whether the distance is redshift.
	 */
	public boolean getDistanceIsRedshift() {
		return distanceIsRedshift;
	}

	/**
	 * @return the redshift, if distance is expressed as redshift
	 */
	public double getRedshift() {
		if (distanceIsRedshift) {
			return distance;
		}
		throw new IllegalArgumentException("distance is in pc for this morphology.  distance="+distance); 
	}


}

