package gov.fnal.eag.sim.catalog;

import gov.fnal.eag.shapelets.functions.BasisFunction;
import gov.fnal.eag.sim.random.SimRandomUniform;
import gov.fnal.eag.sim.repository.image.ImageRepository;
import gov.fnal.eag.util.misc.InputStreamIOException;
import gov.fnal.eag.util.misc.RowSequenceIOException;
import gov.fnal.eag.util.yanny.Table;
import gov.fnal.eag.wcs.SphericalCoordinate;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;

import uk.ac.starlink.table.RowListStarTable;
import uk.ac.starlink.table.RowSequence;

/**
 * picks a Morphology from the catalog and applies a random rotation
 * @author stoughto
 *
 */
public class MorphologyGeneratorFromCatalog implements MorphologyGenerator {

	private ArrayList<Morphology> catalog = new ArrayList<Morphology>();
	private final int seed;
	private int nMaxIndex, cc0Index, betaIndex, zIndex;
	double arcsecPerBeta = 1.0;
	private SimRandomUniform sru = new SimRandomUniform();
	/**
	 * @param fileName where to get the catalog
	 * @param seed base seed for Morphology generation
	 */
	public MorphologyGeneratorFromCatalog(String fileName, int seed) {
		this.seed = seed;
		try {
			init(fileName);
		} catch (IOException e) {
			throw new Error(e);
		}
	}

	private void init(String fileName) throws IOException {
		InputStream inputStream;
		RowListStarTable rlst =  null;
		RowSequence rs;
		if (fileName.startsWith("ioException")) {
			inputStream = new InputStreamIOException();
			rs = new RowSequenceIOException();
		} else {
			try {
				// Look for the file directly
				inputStream = new FileInputStream(fileName);
			} catch (FileNotFoundException e){
				// Look for the file by name in the data/seds directory
				inputStream = SED.class.getResourceAsStream(fileName);			
			}
			if (inputStream == null) {
				throw new IllegalArgumentException("can not open fileName="+fileName);
			}
			rlst = ImageRepository.readRlst(inputStream, 0);
			rs = rlst.getRowSequence();
			if (fileName.startsWith("data/shapelets/shapecat")) {
				nMaxIndex = Table.getIndexByName(rlst, "nMax");
				cc0Index = Table.getIndexByName(rlst, "CC0");
				betaIndex = Table.getIndexByName(rlst, "Beta");
				zIndex = -1;
				arcsecPerBeta = 0.1;
			} else if (fileName.startsWith("testCat")) {
				nMaxIndex = Table.getIndexByName(rlst, "nMax");
				cc0Index = Table.getIndexByName(rlst, "CC0");
				betaIndex = Table.getIndexByName(rlst, "Beta");
				zIndex = Table.getIndexByName(rlst, "redshift")	;
				arcsecPerBeta = 1.0;			
			} else {
				throw new IllegalArgumentException("what is the format of "+fileName);
			}
		}


		while (rs.next()) {
			Object[] row = rs.getRow();
			int nMax = ((Number) row[nMaxIndex]).intValue();
			int length = BasisFunction.lengthFromDimension(nMax);
			double[] ev = new double[length];
			for (int i=0; i<length; i++) {
				ev[i] = ((Number) row[i+cc0Index]).doubleValue();
			}
			double scale = arcsecPerBeta * ((Number) row[betaIndex]).doubleValue();
			double redshift;
			if (zIndex == -1) {
				redshift = 0.6;
			} else {
				redshift = ((Number) row[zIndex]).doubleValue();
			}
			Morphology morphology = new MorphologyShapelet(ev, scale, redshift);
			catalog.add(morphology);
		}
	}
	/**
	 * @see gov.fnal.eag.sim.catalog.MorphologyGenerator#copy()
	 * @throws UnsupportedOperationException for now
	 */
	public MorphologyGenerator copy() {
		throw new UnsupportedOperationException();
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.MorphologyGenerator#getISeed()
	 */
	public int getISeed() {
		return seed;
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.MorphologyGenerator#next(boolean)
	 * @throws UnsupportedOperationException for now
	 */
	public Morphology next(boolean calculateReturnValue) {
		throw new UnsupportedOperationException();
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.MorphologyGenerator#next(gov.fnal.eag.wcs.SphericalCoordinate, double, double)
	 */
	public Morphology next(SphericalCoordinate sc, double redshift, double absoluteMagnitude) {
		sru.setIseedIndex(sc.getSeed(seed));
		int index = (int) sru.next(0, catalog.size());
		Morphology retval = catalog.get(index).copy();
		double phi = sru.next(0.0, 360.0);
		retval.rotate(phi);
		return retval;
	}

}
