package gov.fnal.eag.sim.catalog;

import gov.fnal.eag.shapelets.functions.BasisFunction;
import gov.fnal.eag.shapelets.functions.Epanechinikov;
import gov.fnal.eag.sim.random.SimRandomUniform;
import gov.fnal.eag.sim.repository.image.ImageRepository;
import gov.fnal.eag.wcs.SphericalCoordinate;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Arrays;

import uk.ac.starlink.table.StarTable;

import com.imsl.math.Complex;

/**
 * creates a generator of shapelet
 * objects based on the provided shapelet catalog. The catalog can be in the
 * form of Cartesian shapelet coefficients or polar coefficients.   It
 * generates a lambda vector (vector of maximal variances of the coefficients)
 * using the catalog.  In case of polar catalog variance of phase part of a
 * coefficient is recommended to be 8 degrees.
 * 
 * @author kuropat
 * @version $Name$ ($Id: MorphologyGenerator.java,v 1.5 2006/04/05 16:53:59
 *          kuropat Exp $)
 */
public class MorphologyGeneratorShapeletCatalog implements MorphologyGenerator {
	
	private StarTable starTable;
	private SimRandomUniform ur;

	private Epanechinikov ep;

	/**
	 * Contains all information necessary for image reconstruction
	 */
	private double[][] catalog;

	/**
	 * rotated object
	 */
	private double[] rotObj;

	/**
	 * parameters of lambda generator fact -- linear parameter factor pfact --
	 * phase parameter factor
	 */
	private double fact = 2.;

	private double pfact = 8.;

	/**
	 * Lambda vector and deltaVector
	 */
	private double[] lambdaVector;

	private double[] deltaVector;

	/**
	 * type of catalog
	 */
	private boolean polarFlag = true;

	/**
	 * Maximum rank of shapelet coefficients stored in the catalog
	 */
	private int nMax = 15;

	/**
	 * minimal and maximal number of catalog objects to be used in the
	 * simulation use these to select subset ob the catalog
	 */
	private int minObjNum = 0;

	private int maxObjNum = 0;

	/**
	 * range of the random rotation of the generated object set to 0. to switch
	 * off rotation
	 */
	private double minRotAngle = 0.;
	private double maxRotAngle = 360.;
	
	private int seed;
	/**
	 * constructor with the default random seed 123987
	 * @param catFile
	 *            name of the fits file containing shapelet catalog
	 * @param polarFlag
	 *            flag showing which type of catalog is used true for polar
	 *            shapelet coefficients (modulus and phase ),  false for Cartesian
	 *            shapelets
	 */
	public MorphologyGeneratorShapeletCatalog(String catFile, boolean polarFlag) {
		this(123987, catFile, polarFlag);
	}
	

	/**
	 * fully-specified constructor
	 * @param seed random number seed index
	 * @param catFile
	 *            name of the fits file in this project's data/shapelets folder containing shapelet catalog
	 * @param polarFlag
	 *            flag showing which type of catalog is used true for polar
	 *            shapelet coefficients (modulus and phase ),  false for Cartesian
	 *            shapelets
	 * @throws IOException 
	 */
	public MorphologyGeneratorShapeletCatalog(int seed, String catFile, boolean polarFlag) {
		String baseFileName = "data" + File.separator + "shapelets" + File.separator + catFile;
		InputStream inputStream;
		URL inputURL = this.getClass().getResource(baseFileName);
		//System.out.println("File="+inputURL.getFile());
		try {
			inputStream = inputURL.openStream();
		} catch (IOException e) {
			throw new Error(e);
		}
		if (inputStream  != null) {
			starTable = ImageRepository.readRlst(inputStream, 0);
			init(seed, polarFlag);
		} else {
			throw new Error("can not read StarTable from "+baseFileName);
	}
	}
	/**
	 * fully-specified constructor
	 * @param seed random number seed index
	 * @param starTable
	 *            contains the coefficients
	 * @param polarFlag
	 *            flag showing which type of catalog is used true for polar
	 *            shapelet coefficients (modulus and phase ),  false for Cartesian
	 *            shapelets
	 */
	public MorphologyGeneratorShapeletCatalog(int seed, StarTable starTable, boolean polarFlag) {
		this.starTable = starTable;
		init(seed, polarFlag);
	}
	
	private void init(int seedIn, boolean polarFlag1) {
		this.polarFlag = polarFlag1;
		this.seed = seedIn;
		ur = new SimRandomUniform(seed);
		ep = new Epanechinikov();
			int nObjs = (int) starTable.getRowCount();
			int maxCoef = starTable.getColumnCount() - 5;
			if (polarFlag1) {
				nMax = (int) Math.sqrt(maxCoef) - 1;
			} else {
				nMax = (int) Math.sqrt(2. * maxCoef) - 1;
			}
			catalog = new double[maxCoef + 5][nObjs];
//			this.maxObjNum = (int) starTable.getRowCount() - 1;
			/**
			 * Warning, this is model dependent. Another catalog can have different
			 * structure and we need to make the code more flaxible
			 */
			int nsel = 0;
			for (int row = 0; row < starTable.getRowCount(); row++) {
				Object[] shapeletObject;
				try {
					shapeletObject = starTable.getRow(row);
				} catch (IOException e) {
					throw new Error(e);
				}
				// shapeletObject[2] is number of eigenvalues
				if (((Double)shapeletObject[2]).doubleValue() > 0. ) {
					catalog[0][nsel] = ((Double) shapeletObject[0]).doubleValue(); // Yc
					catalog[1][nsel] = ((Double) shapeletObject[1]).doubleValue(); // Xc
					catalog[2][nsel] = ((Double) shapeletObject[2]).doubleValue(); // Ncoef
					catalog[3][nsel] = ((Double) shapeletObject[3]).doubleValue(); // beta
					catalog[4][nsel] = ((Double) shapeletObject[4]).doubleValue(); // flux

					for (int j = 0; j < maxCoef; j++) { // coefs
						catalog[j + 5][nsel] = ((Double) shapeletObject[5 + j])
							.doubleValue();
					}
					nsel++;
				}
			}
			this.maxObjNum = nsel;
			if (polarFlag1) {
				lambdaVector = calcLambdaVect(catalog, fact, pfact);
			} else {
				lambdaVector = calcLambdaVect(catalog, fact);
			}
		}
	

	/**
	 * Calculate lambda vector. Should be called after constructor or after
	 * changing lambda factors
	 * 
	 * @param factIn -
	 *            factor of the liniar (module) variance
	 * @param pfactIn -
	 *            factor of phase variance
	 */
	public void calcVariationLimits(double factIn, double pfactIn) {
		this.fact = factIn;
		this.pfact = pfactIn;
		// Generate Lambda vector
		if (polarFlag) {
			lambdaVector = calcLambdaVect(catalog, fact, pfact);
		} else {
			lambdaVector = calcLambdaVect(catalog, fact);
		}

	}

	/**
	 * Calculate Lambda for a specified shapelet coefficient Lambda is mean
	 * value of neighbor differences (variations )
	 * 
	 * @param in -
	 *            double array of the shapelet coefficient in the catalog
	 * @param factor -
	 *            double factor to modify the lambda
	 * @return - Lambda*factor
	 */

	private double calcLambda(double[] in, double factor) {
		double Lambda = 0.;
		// count non zero elements of in
		int al = 0;
		for (int i = 0; i < in.length; i++) {
			if (in[i] != 0.)
				al++;
		}
		if (al <= 1) {
			Lambda = 0.;
			return Lambda;
		}
		double[] a = new double[al];
		int ind = 0;
		for (int i = 0; i < in.length; i++) {
			if (in[i] != 0.) {
				a[ind] = in[i];
				ind++;
			}
		}

		Arrays.sort(a);
		double nPair = 0.;

		for (int i = 1; i < a.length; i++) {
			Lambda += Math.abs(a[i] - a[i - 1]);
			nPair++;
		}
		if (nPair > 0) {
			Lambda /= nPair;
		} else {
			Lambda = 0.;
		}
		Lambda *= factor;

		return Lambda;
	}

	/**
	 * Calculate the vector of Lambda*factor for all shapelet coefficients in
	 * the catalog. This method can be used for Cartesian and polar coefficients
	 * if polar coefficients are presented in (real, imagenary) form.
	 * 
	 * @param catalogIn
	 * @param factor
	 * @return - double array of Lambda*factor
	 */
	private double[] calcLambdaVect(double[][] catalogIn, double factor) {
		double[] Lambda = new double[catalogIn.length - 3];
		for (int i = 0; i < catalogIn.length - 3; i++) {
			Lambda[i] = calcLambda(catalogIn[i + 3], factor);
		}

		return Lambda;
	}

	/**
	 * This method calculates Lambda vector for polar shapelet coefficients when
	 * polar coefficients are presented as modulus and phase form
	 * 
	 * @param catalogIn
	 * @param lmodul -
	 *            coefficient for liniar part (modulus) (recommended 4)
	 * @param lphase -
	 *            coefficient for phase (recommended 15 degrees )
	 * @return Lambda vector
	 */
	private double[] calcLambdaVect(double[][] catalogIn, double lmodul,
			double lphase) {
		double[] Lambda = new double[catalogIn.length - 3];
		double DEGRAD = Math.PI / 180.;
		double phaseFact = lphase * DEGRAD;
		Lambda[0] = calcLambda(catalogIn[3], lmodul); // beta
		Lambda[1] = calcLambda(catalogIn[4], lmodul); // flux
		int ll = 0;

		for (int i = 0; i < (catalogIn.length - 5) / 2; i++) {
			Lambda[ll + 2] = calcLambda(catalogIn[ll + 5], lmodul);
			Lambda[ll + 2 + 1] = phaseFact;

			if (ll == 0)
				Lambda[ll + 2 + 1] = 0.;
			ll += 2;
		}
		return Lambda;
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.MorphologyGenerator#next(boolean)
	 */
	public MorphologyShapelet next(boolean returnValue) {
		MorphologyShapelet retval = null;
		if (returnValue) {
			retval = next();
		} else {
			ur.next();
			ur.next();
			ur.next();
		}
		return retval;
	}
	/**
	 * Randomly select an object from the catalog and perform modifications to
	 * its shapelet coefficients.
	 * 
	 * @return - new Shape object in the catalog entry format
	 */
	public MorphologyShapelet next() {

		int minObj = this.minObjNum;
		int maxObj = this.maxObjNum;
		int range = maxObj - minObj;
		int nPars = catalog.length;
		double size = 0.;
		double redshift = 0.6;
		double beta = 0.;
		// The real eigenvalues start after the fifth element of the catalog[i][] array
		double[] newShape = new double[nPars];
		// System.out.println("MorphologyGeneratorShapeletCatalog:  ur seed = "+ur.getIseedIndex());
		int nextInd = (int) (ur.next() * range) + minObj;
		long nseed = (long) ur.next(123975., 975123.);
		

		// newShape contains the whole string from the catalog
		for (int i = 0; i < nPars; i++) {
			newShape[i] = catalog[i][nextInd];
		}
		
		ep.setSeed(nseed);
		deltaVector = ep.nextVector(lambdaVector);
		// Apply delta variation
		int vLen = 0;                   // length of the delta vector
		if (polarFlag) {
			vLen = (int) ((newShape[2] + 1.) * (newShape[2] + 2.)) + 2;
		} else {
			vLen = (int) ((newShape[2] + 1.) * (newShape[2] + 2.) / 2.) + 2;
		}
		for (int i = 3; i < vLen; i++) {
			newShape[i] += deltaVector[i - 3];
		}
		beta = newShape[3];             // beta parameter to be used for the size calculation
		double[] coefs = new double[vLen - 2];
		Complex[] pcoefs = new Complex[(vLen - 2) / 2];

		/*
		 * Now perform rotation
		 */
		double rotRange = maxRotAngle - minRotAngle;
		for (int k = 0; k < vLen - 2; k++) {
			coefs[k] = newShape[k + 5];
		}
		double phi = ur.next() * rotRange + minRotAngle;
		//System.out.println("MorphologyGeneratorShapeletCatalog.next:  nextInd="+nextInd+" nseed="+nseed+" phi="+phi);
		if (!polarFlag) {

			rotObj = gov.fnal.eag.shapelets.functions.ShapeletsTransform
					.rotate(coefs, phi);
			
		} else {
			int k = 0;
			for (int i = 0; i < coefs.length; i += 2) {
				double mod = coefs[i];
				double phi1 = coefs[i + 1];
				pcoefs[k] = new Complex(mod * Math.cos(phi1), mod
						* Math.sin(phi1));
				k++;
			}
			rotObj = gov.fnal.eag.shapelets.functions.ShapeletsTransform
					.p2c(gov.fnal.eag.shapelets.functions.ShapeletsTransform
							.rotate(pcoefs, phi));

		}
		// calculate size of the object in arcsec
		double transCoeff = 0.1;  // transforms from pixels to arcsec for HST
		size = beta * transCoeff;
		//
		// calculate red shift. should be present in the catalog
		redshift = 0.6;
		//
		MorphologyShapelet retval = new MorphologyShapelet(rotObj, size, redshift);
		return retval;
	}
	
	/**
	 * @return the number of objects in the catalog
	 */
	public int size() {
		return catalog[2].length;
	}
	/**
	 * @param i
	 * @return the ith morphology from the catalog 
	 */
	public MorphologyShapelet get(int i) {
		int nMaxTemp = (int) catalog[2][i];
		int length = BasisFunction.lengthFromDimension(nMaxTemp+1);
		double[] ev = new double[length];
		for (int ii=0; ii<length; ii++) {
			ev[ii] = catalog[ii+5][i];
		}
		double size = catalog[3][i] * 0.1; 
		double redshift = 0.6;
		MorphologyShapelet retval = new MorphologyShapelet(ev, size, redshift);
		return retval;
	}

	/**
	 * @param seed the seed to set
	 */
	public void setSeeds(int seed) {
		this.seed = seed;
		ur.setIseedIndex(seed);
	}

	/**
	 * @return the factor parameter
	 */
	public double getFactor() {
		return this.fact;
	}

	/**
	 * returm value of the pfactor parameter
	 * 
	 * @return double
	 */
	public double getPFactor() {
		return this.pfact;
	}

	/** 
	 * @return the maximum rank in this catalog
	 */
	public int getNMax() {
		return this.nMax;
	}
	
	/**
	 * get the number of objects
	 * @return the number of objects
	 */
	public int getMaxObjNum() {
		return maxObjNum;
	}

	/**
	 * sets subset of the catalog to use in the simulation
	 * 
	 * @param minObj
	 * @param maxObj
	 */
	public void setCatRange(int minObj, int maxObj) {
		this.minObjNum = minObj;
		if (maxObj < minObj) {
			System.err.println("Parameters error minObj=" + minObj + " maxObj="
					+ maxObj);
			return;
		}
		if (maxObj >= 0 && maxObj <= this.maxObjNum) {
			this.maxObjNum = maxObj;
		}

	}

	/**
	 * sets the range of the random rotation of the generated objects
	 * 
	 * @param minAngle minimum angle in degrees
	 * @param maxAngle maximum angle in degrees
	 */
	public void setRotationRange(double minAngle, double maxAngle) {
		this.maxRotAngle = maxAngle;
		this.minRotAngle = minAngle;
	}


	/**
	 * @return a deep copy of this object
	 */
	public MorphologyGeneratorShapeletCatalog copy() {
		MorphologyGeneratorShapeletCatalog retval = new MorphologyGeneratorShapeletCatalog(ur.getIseedIndex(), starTable, polarFlag);
		return retval;
	}

	/**
	 * @return the random number seed
	 */
	public int getISeed() {
		return ur.getIseedIndex();
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.MorphologyGenerator#next(gov.fnal.eag.wcs.SphericalCoordinate)
	 */
	public Morphology next(SphericalCoordinate sc, double redshift, double absoluteMagnitude) {
		// bad:  this returns the same object for all positions in the same HTM trixel
		//int iSeedIndex = HTMfunc.indexFromHtm(thisHtm+seed); 
		//
		// good:  randomizes based on position inside the trixel
		int iSeedIndex = sc.getSeed(seed);
		ur.setIseedIndex(iSeedIndex);
		return next(true);
	}
}
