package gov.fnal.eag.sim.catalog;

import gov.fnal.eag.shapelets.functions.BasisFunction;
import gov.fnal.eag.shapelets.functions.Decompose;
import gov.fnal.eag.shapelets.functions.ShapeletsTransform;
import gov.fnal.eag.sim.pixel.DetectorObjectProfile;
import gov.fnal.eag.sim.pixel.DetectorObjectSurfaceBrightness;
import gov.fnal.eag.sim.pixel.PixeledImage;
import gov.fnal.eag.sim.pixel.Profiles;

import java.util.Arrays;

/**
 * describes a Morhpology with a set of shapelet coefficients
 * @author stoughto
 *
 */
public class MorphologyShapelet extends Morphology {
	/**
	 * default construct returns a point source
	 */
	private double[] shapeletEigenValues = null;
	/**
	 * default construcor is an object with only one coefficient
	 */
	public MorphologyShapelet() {  // point source is the default
		shapeletEigenValues = new double[] {1.0};
	}
	/**
	 * @see gov.fnal.eag.sim.catalog.Morphology#equals(gov.fnal.eag.sim.catalog.Morphology)
	 */
	@Override
	public boolean equals(Morphology that) {
		return equals(that, false);
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.Morphology#equals(gov.fnal.eag.sim.catalog.Morphology, boolean)
	 */
	@Override
	public boolean equals(final Morphology that, final boolean chatty) {
		if (that instanceof MorphologyShapelet) {
			MorphologyShapelet thatMs = (MorphologyShapelet) that;
			if (this.size != that.size) {
				if (chatty) {
					System.out.println("Morphology.equals:  sizes not equal: "+this.size+" "+that.size);
				}
				return false;
			}
			if (this.distance != that.distance) {
				if (chatty) {
					System.out.println("Morphology.equals:  distances not equal");
				}
				return false;
			}
			if (!Arrays.equals(this.shapeletEigenValues, thatMs.shapeletEigenValues)) {
				if (chatty) {
					System.out.println("Morphology.equals:  shapeletEigenValues not equal");
				}
				return false;
			}
			return true;
		}
		return false;
	}
	/**
	 * Create a Shape with these eigenvalues.
	 * @param shapeletEigenValues always specified in the 1d cartesian coordinates
	 * @param size object size in arcsec at the specified distance
	 * @param distance the distance (in redshift)
	 */
	public MorphologyShapelet(double[] shapeletEigenValues, double size, double distance) {
		this(shapeletEigenValues, size, distance, true);
	}
	/**
	 * Create a Shape with these eigenvalues.
	 * @param shapeletEigenValues always specified in the 1d cartesian coordinates
	 * @param size object size in arcsec at the specified distance
	 * @param distance the distance (in redshift or pc)
	 * @param distanceIsRedshift 
	 */
	public MorphologyShapelet(double[] shapeletEigenValues, double size, double distance, boolean distanceIsRedshift) {
		this.shapeletEigenValues = shapeletEigenValues;
		setSize(size);
		setDistance(distance);
		this.distanceIsRedshift = distanceIsRedshift;
	}

	/**
	 * constructs a morphology of a gaussian object
	 * @param sizeArcsec sigma of the object
	 * @param redshift at this redshift
	 * @param nMax shapelet orders
	 */
	public MorphologyShapelet(double sizeArcsec, double redshift, int nMax) {
		this(sizeArcsec, redshift, nMax, true);
	}

	/**
	 * constructs a morphology of a gaussian object
	 * @param sizeArcsec sigma of the object
	 * @param distance at this distance
	 * @param nMax shapelet orders
	 * @param distanceIsRedshift true if it is, otherwise distance is in pc
	 */
	public MorphologyShapelet(double sizeArcsec, double distance, int nMax, boolean distanceIsRedshift) {
		if (profiles == null) {
			profiles = new Profiles();
		}
		double scale = 10;
		DetectorObjectProfile dop = new DetectorObjectProfile(0, 0, 1, "gauss", scale, 0, 0, profiles);
		int diameter = 120;
		PixeledImage pi = dop.makePixeledImage(diameter);
		Decompose decompose = new Decompose(pi.getPix(), diameter/2.0, diameter/2.0, scale);
		shapeletEigenValues = decompose.getFnnV(nMax);
		setSize(sizeArcsec);
		setDistance(distance);
	}
	/**
	 * constructs a morphology of a profile object
	 * @param shape 
	 * @param sizeArcsec sigma of the object
	 * @param ellipticity 
	 * @param positionAngle 
	 * @param distance at this distance
	 * @param nMax shapelet orders
	 * @param distanceIsRedshift 
	 */
	public MorphologyShapelet(String shape, double sizeArcsec, double ellipticity, double positionAngle, double distance, int nMax, boolean distanceIsRedshift) {
		if (profiles == null) {
			profiles = new Profiles();
		}
		double scale = 10;
		DetectorObjectProfile dop = new DetectorObjectProfile(0, 0, 1, shape, scale, ellipticity, positionAngle, profiles);
		int diameter = 120;
		PixeledImage pi = dop.makePixeledImage(diameter);
		Decompose decompose = new Decompose(pi.getPix(), diameter/2.0, diameter/2.0, scale);
		shapeletEigenValues = decompose.getFnnV(nMax);
		setSize(sizeArcsec);
		setDistance(distance);
		this.distanceIsRedshift = distanceIsRedshift;
	}

	/**
	 * copy constructor 
	 * @param original
	 */
	public MorphologyShapelet(MorphologyShapelet original) {
		size = original.size;
		distance = original.distance;
		distanceIsRedshift = original.distanceIsRedshift;
		shapeletEigenValues = original.shapeletEigenValues.clone();
	}
	/**
	 * Return the eigenvalues for this shape.
	 * @return the eigenvalues for this Shapelet, or null if it is a point source
	 */
	public double[] getEigenvalues() {
		return shapeletEigenValues;
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.Morphology#shear(double, double)
	 */
	@Override
	public void shear(double gamma1, double gamma2) {
		if (gamma1 != 0.0 && gamma2 != 0.0) {
			double[][] ev = BasisFunction.vector2Tab(shapeletEigenValues);
			shapeletEigenValues = BasisFunction.tab2Vector(ShapeletsTransform.shear(ev, gamma1, gamma2));
		}
	}
	/**
	 * @see gov.fnal.eag.sim.catalog.Morphology#rotate(double)
	 */
	@Override
	public void rotate(double theta) {
		shapeletEigenValues = ShapeletsTransform.rotate(shapeletEigenValues, theta);		
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "MorphologyShapelet:  redshift="+distance+" size="+size;		
	}
	/**
	 * @see gov.fnal.eag.sim.catalog.Morphology#copy(gov.fnal.eag.sim.catalog.Morphology)
	 */
	@Override
	public MorphologyShapelet copy() {
		MorphologyShapelet retval = new MorphologyShapelet(this);
		return retval;
	}
	/**
	 * @see gov.fnal.eag.sim.catalog.Morphology#makeDosb(double, double, double, double)
	 */
	@Override
	public DetectorObjectSurfaceBrightness makeDosb(double row, double col, double counts, double pixelSizeArcseconds) {		
		double scale = getSize()/pixelSizeArcseconds;
		double[] eigenvalues = getEigenvalues();
		DetectorObjectSurfaceBrightness retval = new DetectorObjectSurfaceBrightness(row, col, counts, eigenvalues, scale);			
		return retval;
	}
}
