package gov.fnal.eag.sim.catalog;

import gov.fnal.eag.sim.extractor.Extractor;
import gov.fnal.eag.sim.extractor.SourceObject;
import gov.fnal.eag.sim.extractor.Extractor.KernelType;
import gov.fnal.eag.sim.pixel.ColorTable;
import gov.fnal.eag.sim.pixel.DetectorObject;
import gov.fnal.eag.sim.pixel.DetectorObjectComposite;
import gov.fnal.eag.sim.pixel.DetectorObjectProfile;
import gov.fnal.eag.sim.pixel.ImageGlyph;
import gov.fnal.eag.sim.pixel.PixeledImage;
import gov.fnal.eag.sim.pixel.PixeledImagePlot;
import gov.fnal.eag.sim.pixel.ScaleTable;
import gov.fnal.eag.sim.pixel.SimplePixelMark;
import gov.fnal.eag.sim.repository.image.ImageRepository;
import gov.fnal.eag.util.misc.SDSSUtil;
import gov.fnal.eag.wcs.TransformTangentPlaneA7;

import java.awt.Point;
import java.awt.Rectangle;
import java.awt.image.BufferedImage;
import java.util.ArrayList;

/**
 * implements Morphology with a list of pixels and values
 * @author stoughto
 *
 */
public class MorphologySourceObject extends Morphology {
	private SourceObject sourceObject;
	private double theta;
	/**
	 * fully-specified constructor
	 * @param sourceObject
	 * @param size arcseconds per pixel in the sourceObject
	 * @param theta
	 * @param distance
	 * @param distanceIsRedshift
	 */
	public MorphologySourceObject(SourceObject sourceObject, double size, double theta, double distance, boolean distanceIsRedshift) {
		init(sourceObject, size, theta, distance, distanceIsRedshift);
	}
	/**
	 * constructor based on SDSS imaging data
	 * @param sdssDasRoot
	 * @param run
	 * @param rerun
	 * @param camcol
	 * @param field
	 * @param filter
	 * @param range
	 * @param redshift
	 */
	public MorphologySourceObject(String sdssDasRoot, int run, int rerun, int camcol, int field, String filter, Rectangle range, double redshift) {
		String fpcFileName = SDSSUtil.fpCFullFileName(sdssDasRoot, run, rerun, camcol, field, filter);
		PixeledImage pi = ImageRepository.getPixeledImage(fpcFileName, range.height, range.width, range.y, range.x);
		Extractor extractor = new Extractor(pi);
		extractor.setKernel(KernelType.GAUSS5x5);
		extractor.setNSigma(2.0);
		SourceObject so = extractor.getLargestSource();
		double sdssPixelSize = 0.4;
		init(so, sdssPixelSize, 0.0, redshift, true);
	}
	
	private void init(SourceObject sourceObject1, double size1, double theta1, double distance1, boolean distanceIsRedshift1) {
		double flux = sourceObject1.getFlux();
		if (flux <= 0.0) {
			throw new IllegalArgumentException("sourceObject must have positive flux but flux="+flux);
		}
		sourceObject = sourceObject1;
		size = size1;
		theta = theta1;
		distance = distance1;
		distanceIsRedshift = distanceIsRedshift1;
	}
	/**
	 * @see gov.fnal.eag.sim.catalog.Morphology#copy()
	 */
	@Override
	public Morphology copy() {
		MorphologySourceObject retval = new MorphologySourceObject(sourceObject, size, theta, distance, distanceIsRedshift);
		return retval;
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.Morphology#equals(gov.fnal.eag.sim.catalog.Morphology)
	 */
	@Override
	public boolean equals(Morphology that) {
		return equals(that, false);
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.Morphology#equals(gov.fnal.eag.sim.catalog.Morphology, boolean)
	 */
	@Override
	public boolean equals(Morphology that, boolean chatty) {
		if (!(that instanceof MorphologySourceObject)) {
			if (chatty) {
				System.out.println("not a MorphologySourceObject: that="+that);
			}
			return false;
		}
		MorphologySourceObject thatMso = (MorphologySourceObject) that;
		if (!sourceObject.equals(thatMso.sourceObject)) {
			if (chatty) {
				System.out.println("the sourceObjects are not equal");
			}
			return false;
		}
		
		if (distance != thatMso.distance) {
			if (chatty) {
				System.out.println("distances are not equal:  distance="+distance+" thatDistance="+thatMso.distance);
			}
			return false;
		}
		
		if (distanceIsRedshift != thatMso.distanceIsRedshift) {
			if (chatty) {
				System.out.println("distanceIsRedshift not equal");
			}
			return false;
		}
		
		if (theta != thatMso.theta) {
			if (chatty) {
				System.out.println("theta not equal");
			}
			return false;			
		}
		if (size != thatMso.size) {
			if (chatty) {
				System.out.println("size not equal");
			}
			return false;			
		}
		return true;
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.Morphology#rotate(double)
	 */
	@Override
	public void rotate(double thetaDegrees) {
		this.theta += thetaDegrees;
	}

	/**
	 * @return the rotation angle
	 */
	public double getTheta() {
		return theta;
	}
	/**
	 * @see gov.fnal.eag.sim.catalog.Morphology#shear(double, double)
	 */
	@Override
	public void shear(double gamma1, double gamma2) {
		throw new UnsupportedOperationException();
	}
	
	/**
	 * @param row
	 * @param col
	 * @param nPhoton
	 * @param pixelSizeArcseconds
	 * @param profileNames
	 * @param smoothRadius if > 0 convolve and smooth within this radius of the edge of the source object
	 * @param plotFileName 
	 * @return a composite detector object by re-sampling the pixel values to the new pixel size and extending in radius by fitting the named profiles
	 */
	public DetectorObjectComposite makeDoc(double row, double col,
			double nPhoton, double pixelSizeArcseconds, String[] profileNames, int smoothRadius, String plotFileName) {
		ArrayList<ImageGlyph> glyphs = null;
		if (plotFileName != null) {
			glyphs = new ArrayList<ImageGlyph>();
		}
		// find min/max pixels in
		double colSo = sourceObject.getXWin();
		double rowSo = sourceObject.getYWin();
		if (Double.isNaN(colSo)) {
			colSo = sourceObject.getX();
		}
		if (Double.isNaN(rowSo)) {
			rowSo = sourceObject.getY();
		}
		// make the PixeledImage from the SourceObject
		int xMin = sourceObject.getXMin();
		int xMax = sourceObject.getXMax();
		int yMin = sourceObject.getYMin();
		int yMax = sourceObject.getYMax();
		PixeledImage piSo = new PixeledImage(1+yMax-yMin, 1+xMax-xMin);
		piSo.setRow0Col0(yMin, xMin);
		for (Point point : sourceObject.getPoints()) {
			int r = point.y;
			int c = point.x;
			double value = sourceObject.getPixeledImage().getByRowCol(r, c);
			piSo.setByRowCol(r, c, value);
		}	
		double raRef = 180.0;
		double decRef = 0.0;
		TransformTangentPlaneA7 sourceTransform = new TransformTangentPlaneA7(colSo, rowSo, raRef, decRef,
				size/3600.0, theta, 0.0);		
		// make a blank PixeledImage for the destination
		double rowSoMin = 0;
		double rowSoMax = 0;
		double colSoMin = 0;
		double colSoMax = 0;
		double ct = Math.cos(Math.toRadians(theta));
		double st = Math.sin(Math.toRadians(theta));
		for (Point point : sourceObject.getPoints()) {
			double dr = rowSo - (point.y+0.5);
			double dc = colSo - (point.x+0.5);
			double c =  ct*dc + st*dr;
			double r = -st*dc + ct*dr;
			rowSoMin = Math.min(rowSoMin, r);
			colSoMin = Math.min(colSoMin, c);
			rowSoMax = Math.max(rowSoMax, r);
			colSoMax = Math.max(colSoMax, c);
		}
		
		// size is  arcseconds per sourceObject pixel
		// pixelSizeArcseconds is arcseconds per detector pixel
		// so to convert from sourceObject pixels to detector pixels
		// multilply by size/pixelSizeArcseconds 
		double scaling = size/pixelSizeArcseconds;
		// create the blank pixeled image		
		int rowMin = (int) (Math.floor(row+scaling*rowSoMin));
		int rowMax = (int) (Math.ceil(row+scaling*rowSoMax));
		int colMin = (int) (Math.floor(col+scaling*colSoMin));
		int colMax = (int) (Math.ceil(col+scaling*colSoMax));
		int nRow = rowMax - rowMin;
		int nCol = colMax - colMin;
		
		PixeledImage pi = new PixeledImage(nRow, nCol);
		int destinationCol0 = (int)Math.floor(colMin);
		int destinationRow0 = (int)Math.floor(rowMin);
		pi.setRow0Col0(destinationRow0, destinationCol0);
		// fill the pixeled image
		TransformTangentPlaneA7 destinationTransform = new TransformTangentPlaneA7(col, row, raRef, decRef, pixelSizeArcseconds/3600.0, 0.0, 0.0);
		

		PixeledImage.drizzle(pi, destinationTransform, piSo, sourceTransform, 1.0);
		/*
		for (int r=pi.getRowMin(); r<pi.getRowMax(); r++) {
			for (int c=pi.getColumnMin(); c<pi.getColumnMax(); c++) {
				double value = pi.getByRowCol(r, c);
			}
		}
		*/
		// Scale flux by nPhoton
		double piSum = pi.getSum();
		double fluxFactor = nPhoton / piSum;
		pi.multiplyScalar(fluxFactor);
		// Make the mask of pixels completely filled by the source object
		PixeledImage piOnes = new PixeledImage(sourceObject.getPixeledImage());
		piOnes.setScalar(0.0);
		for (Point point : sourceObject.getPoints()) {
			piOnes.setByRowCol(point.y, point.x, 1.0);
		}
		PixeledImage piMask = new PixeledImage(nRow, nCol);
		piMask.setRow0Col0(destinationRow0, destinationCol0);
		PixeledImage.drizzle(piMask, destinationTransform, piOnes, sourceTransform, 1.0);
		double sourceArcsecPerPixel = sourceTransform.arcsecPerPixel();
		double destinationArcsecPerPixel = destinationTransform.arcsecPerPixel();
		double factor = Math.pow(destinationArcsecPerPixel/sourceArcsecPerPixel, 2);
			
		// Make a SourceObject out of the pixels that have > 0.0 flux
		double threshold = 0.99*factor;
		if (profileNames.length == 0) {
			threshold = 0.0;
		}
		SourceObject dso = new SourceObject(pi, piMask, threshold);
		ArrayList<DetectorObject> detectorObjects = dso.fitProfiles(profileNames);

		// make the dosb and return it
		int smoothDiameter = 2*smoothRadius+1;
		DetectorObjectComposite doc = new DetectorObjectComposite(row, col, detectorObjects, dso, null);
		int nc = dso.getXMax() - dso.getXMin() + 2*smoothRadius + 1;
		int nr = dso.getYMax() - dso.getYMin() + 2*smoothRadius + 1;
		int c0 = dso.getXMin() - smoothRadius;
		int r0 = dso.getYMin() - smoothRadius;
		
		PixeledImage pi2 = new PixeledImage(nr, nc);
		pi2.setRow0Col0(r0, c0);
		for (int r=pi2.getRowMin(); r<pi2.getRowMax(); r++) {
			for (int c=pi2.getColumnMin(); c<pi2.getColumnMax(); c++) {
				double value = doc.getFluxInPixel(r, c);
				pi2.setByRowCol(r,c,value);
			}
		}

		if (smoothDiameter > 1) {
			DetectorObjectProfile dop = new DetectorObjectProfile(smoothDiameter/2.0, smoothDiameter/2.0, 1.0, "gauss", 1.0, 0.0, 0.0, profiles);
			PixeledImage kernel = dop.makePixeledImage(smoothDiameter, 4);
			pi2 = pi2.convolveFftw(kernel);
		}
		
		ArrayList<Point> points = new ArrayList<Point>();
		SourceObject dsoEdge = dso.getEdgePixels();
		for (int r=pi2.getRowMin(); r<pi2.getRowMax(); r++) {
			for (int c=pi2.getColumnMin(); c<pi2.getColumnMax(); c++) {
				double distanceToEdge = dsoEdge.getClosestDistance(c,r);
				if (distanceToEdge < smoothRadius) {
					// if within smoothRadius of an edge pixel, use the smoothed value
					Point point = new Point(c,r);
					points.add(point);
					if (plotFileName != null) {
						ImageGlyph glyph = new SimplePixelMark(r,c);
						glyphs.add(glyph);
					}
				} else if (dso.getClosestDistance(c, r) == 0) {
					// else, if in the original detector source object, use the dso value
					Point point = new Point(c,r);
					points.add(point);
					double value = pi.getByRowCol(r, c);
					pi2.setByRowCol(r, c, value);
				}
			}			
		}

		if (plotFileName != null) {
			ScaleTable scaleTable = new ScaleTable(pi2);
			ColorTable colorTable = new ColorTable();
			PixeledImagePlot pip = new PixeledImagePlot(pi2, scaleTable, colorTable);
			BufferedImage bufferedImage = pip.makeBufferedImage(400, 0.8, glyphs);
			ImageRepository.toPng(plotFileName, bufferedImage);
		}
		SourceObject filteredSourceObject = new SourceObject(pi2, points);
		filteredSourceObject.setFitValues(dso);
		DetectorObjectComposite retval = new DetectorObjectComposite(row, col, detectorObjects, filteredSourceObject, null);
		return retval;
	}
	/**
	 * @see gov.fnal.eag.sim.catalog.Morphology#makeDosb(double, double, double, double)
	 * fits using exp+deVauc by default
	 */
	@Override
	public DetectorObject makeDosb(double row, double col,
			double counts, double pixelSizeArcseconds) {
		String[] profileNames = {"exp", "deVauc"};
		DetectorObject retval = makeDoc(row,col,counts,pixelSizeArcseconds, profileNames, 2, null);
		return retval;
	}
	/**
	 * @return the SourceObject that defines this MorphologySourceObject
	 */
	public SourceObject getSourceObject() {
		return sourceObject;
	}

	/**
	 * @param profileNames names of profiles to fit and use for half-light radius calculation
	 * @return the radius (in arcseconds) of a circular aperture that contains half the flux
	 */
	public double getHalfLightRadius(String[] profileNames) {
		DetectorObjectComposite doc = makeDoc(0.0, 0.0, 1.0, size, profileNames, 0, null);
		double nPhoton = doc.getNPhoton();
		if (nPhoton <= 0.0) {
			throw new IllegalArgumentException("nPhoton="+nPhoton);
		}
		double sky = 0.0;
		double fluxRatioPrevious = 0.0;
		double fluxRatio = 0.0;
		double r=0.0;
		do  {
			fluxRatioPrevious = fluxRatio;
			r+= 1.0;
			fluxRatio = doc.getApertureFlux(r,sky)/nPhoton;
		} while (fluxRatio < 0.5);

		double retval = r-1 + (0.5-fluxRatioPrevious)/(fluxRatio-fluxRatioPrevious);
		retval *= size;
		return retval;
	}
}
