package gov.fnal.eag.sim.catalog;

import gov.fnal.eag.sim.catalog.WavelengthScale.Scaling;
import gov.fnal.eag.sim.repository.image.ImageRepository;
import gov.fnal.eag.util.misc.Constants;
import gov.fnal.eag.util.misc.ResourceList;
import gov.fnal.eag.util.misc.StringUtil;
import gov.fnal.eag.util.plot.JFreeChartUtil;

import java.awt.geom.Point2D;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.regex.Pattern;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.LogarithmicAxis;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

import cern.colt.list.DoubleArrayList;

import uk.ac.starlink.table.RowListStarTable;

/**
 * A Spectral Energy Distribution, that is, a spectra.
 * 
 * Objects of this class store collections of defined wavelength ranges ({@link gov.fnal.eag.sim.catalog.WavelengthScale}s)
 * and corresponding integrated energies (integrated over the range of each
 * defined bin).
 * 
 * @author neilsen
 * @version $Name:  $ ($Id: SED.java,v 1.76 2009/03/23 14:56:14 stoughto Exp $)
 * 
 */
public class SED implements Iterable<ValuedBin>, Cloneable {
	
	private ArrayList<Point2D.Double> points;
	
	private double[] flux; // The flux is stored in units of ergs/cm^2/sec

	private WavelengthScale lambda;

	private String name = "";
	private static String[] glfcFileNames = null;

	/**
	 * Create a new SED
	 * 
	 * @param flux
	 *            the fluxes in units of ergs/cm^2/sec
	 * @param scale
	 *            the scale where the bins are defined in Angstroms
	 */
	public SED(double[] flux, WavelengthScale scale) {
		setFlux(flux);
		setLambda(scale);
	}

	/*
	 * construct a spectrum that matches the given magnitudes in the filters.
	 * I hope we never have to actually use this.  Implementation is not finished.
	 * 
	public SED(double[] magnitudes, AstronomicalFilter[] filters) {
		if (magnitudes==null || filters == null) {
			throw new IllegalArgumentException("magnitudes="+magnitudes+" filters="+filters);
		}
		if (magnitudes.length != filters.length) {
			throw new IllegalArgumentException("magnitudes.length="+magnitudes.length+" filters.length="+filters.length);
		}
		if (magnitudes.length == 0) {
			throw new IllegalArgumentException("magnitudes.length="+magnitudes.length);			
		}
		lambda = filters[0].getLambda();
		for (int i=1; i<filters.length; i++) {
			if (!lambda.equals(filters[i].getLambda())) {
				throw new IllegalArgumentException("lambda="+lambda+" but filters["+i+"]="+filters[i].getLambda());
			}
		}
		double[] flux = new double[lambda.getNBin()];
		for (int i=0; i<filters.length; i++) {
			double wavelength = filters[i].getMeanWavelength();
			System.out.println("i="+i+" wavelength="+wavelength);
			
		}
	}
	*/
	/**
	 * read from the file and put the flux in a WavelengthScale with nBin,
	 * scaling, from min to max lambda in the file
	 * 
	 * @param nBin
	 *            number of bins
	 * @param scaling
	 *            scaling
	 * @param name
	 *            file to read
	 */
	public SED(int nBin, Scaling scaling, String name) {
		this.setName(name);
		points = loadPoints(name);
		double min = points.get(0).x;
		double max = points.get(points.size() - 1).x;
		WavelengthScale ws = new WavelengthScale(min, max, nBin, scaling);
		lambda = ws;
		flux = new double[ws.size()];
		int i = 0;
		for (ValuedBin vb : this) {
			try {
				flux[i] = vb.getWavelengthBin().integrate(points);
				i++;
			} catch (IntegralNotDefinedException e) {
				throw new Error(e);
			}
		}

	}

	/**
	 * @param name
	 *            spectrum name
	 * @return the points as read from the .sed or .fit file
	 * @throws Error
	 */
	public static ArrayList<Point2D.Double> loadPoints(String name)
			throws Error {
		ArrayList<Point2D.Double> points = new ArrayList<Point2D.Double>();

		String baseFileName = "data" + File.separator + "seds" + File.separator
		+ name;
InputStream inputStream;
		try {
			// Look for the file directly
			inputStream = new FileInputStream(name+".sed");
		} catch (FileNotFoundException e){
			// Look for the file by name in the data/seds directory
			String fullFileName = baseFileName;
			if (!fullFileName.endsWith(".sed")) {
				fullFileName += ".sed";
			}
			inputStream = SED.class.getResourceAsStream(fullFileName);			
		}
		if (inputStream != null)  {			
			points = loadPointsText(inputStream);
			return points;
		} 
		String fullFileName = baseFileName;
		if (!fullFileName.endsWith(".fit")) {
			fullFileName += ".fit";
		}
		inputStream = SED.class.getResourceAsStream(fullFileName);
		if (inputStream != null) {
			RowListStarTable rlst = ImageRepository.readRlst(inputStream, 0);
			for (long row = 0; row < rlst.getRowCount(); row++) {
				Object[] temp = rlst.getRow(row);
				double lambda = (Float) temp[0];
				double flux = (Float) temp[1];
				Point2D.Double lf = new Point2D.Double(lambda, flux);
				points.add(lf);
			}
			return points;
		}
		throw new IllegalArgumentException("can not load points for name="+name);
	}

	/**
	 * @param inputStream
	 * @return the points as loaded from the inputStream
	 */
	public static ArrayList<Point2D.Double> loadPointsText(InputStream inputStream) {
		BufferedReader bufferedReader = new BufferedReader(
				new InputStreamReader(inputStream));
		// Read the file one line at a time
		String line;
		ArrayList<Point2D.Double> points = new ArrayList<Point2D.Double>();

		try {
			while ((line = bufferedReader.readLine()) != null) {
				String[] tokens = StringUtil
						.removeDuplicateWhitespace(line).toString().split(
								" ");
				// Only use lines that do not start with a "#" and that have
				// two tokens
				if (!tokens[0].startsWith("#") && tokens.length == 2) {
					double lambda = Double.parseDouble(tokens[0]);
					double flux = Double.parseDouble(tokens[1]);
					Point2D.Double lf = new Point2D.Double(lambda, flux);
					points.add(lf);
				}
			}
		} catch (Exception e) {
			throw new Error(e);
		}
		return points;
	}
	/**
	 * create an SED by loading the named SED (from a file in this package)
	 * where the wavelengths are in Angstrom and Fluxes are in erg/cm^2/sec at 10 pc in
	 * the bin
	 * 
	 * @param ws
	 *            use this wavelength scale
	 * @param name
	 *            one of the names returned by getNames()
	 */
	public SED(WavelengthScale ws, String name) {
		this.setName(name);
		points = loadPoints(name);
		lambda = ws;
		init();
	}
	
	/**
	 * @param ws
	 * @param name
	 * @param points each point is wavelength (Angstroms), flux (ergs/cm^2/sec/Angstrom)
	 */
	public SED(WavelengthScale ws, String name, ArrayList<Point2D.Double> points) {
		this.setName(name);
		this.points = points;
		this.lambda = ws;
		init();
	}
	
	private void init() {
		flux = new double[lambda.size()];
		int i = 0;
		for (ValuedBin vb : this) {
			try {
				WavelengthBin wb = vb.getWavelengthBin();
				flux[i] = wb.integrate(points);
			} catch (IntegralNotDefinedException e) {
				throw new Error(e);
			}
			i++;
		}
	}
	/**
	 * create an SED by loading the named SED (from a file in this package)
	 * where the wavelengths are in Angstrom and Fluxes are in erg/cm^2/sec at 10 pc in
	 * the bin
	 * 
	 * @param ws
	 *            use this wavelength scale
	 * @param inputStream connected to a text file
	 * @param name
	 *            one of the names returned by getNames()
	 */
	public SED(WavelengthScale ws, InputStream inputStream, String name) {
		this.setName(name);
		points = loadPointsText(inputStream);

		lambda = ws;
		flux = new double[ws.size()];
		int i = 0;
		for (ValuedBin vb : this) {
			try {
				WavelengthBin wb = vb.getWavelengthBin();
				flux[i] = wb.integrate(points);
			} catch (IntegralNotDefinedException e) {
				throw new Error(e);
			}
			i++;
		}
	}

	/**
	 * copy constructor
	 * 
	 * @param sed
	 */
	public SED(SED sed) {
		flux = new double[sed.flux.length];
		for (int i=0; i<flux.length; i++) {
			flux[i] = sed.flux[i];
		}
		lambda = sed.lambda.copy();
		this.setName(new String(sed.getName()));
	}

	/**
	 * Return true if the objects are equal.
	 * 
	 * @param that
	 * @return true if the objects are equal
	 */
	public boolean equals(SED that) {
		if (!lambda.equals(that.lambda))
			return false;
		return Arrays.equals(flux, that.flux);
	}

	/**
	 * Make a deep copy.
	 * 
	 * @return a deep copy of this object
	 */
	@Override
	public SED clone() {
		return new SED(this);
	}

	/**
	 * get a list of SED names available
	 * 
	 * @return a list of names, each of which is suitable for use in SED
	 *         constructors
	 */
	public static Collection<String> getNames() {
		Collection<String> retval = new ArrayList<String>();
		String path = "/fnal/eag/sim/catalog/data/seds/";
		{
			Collection<String> resourceList = ResourceList.getResources(Pattern
					.compile(".*gov/fnal/eag/sim/catalog/data/seds/.*.sed"));
			for (String resource : resourceList) {
				int index = resource.indexOf(path);
				String name = resource.substring(index+path.length());
				retval.add(name);
			}
		}
		{
			Collection<String> resourceList = ResourceList.getResources(Pattern
					.compile(".*gov/fnal/eag/sim/catalog/data/seds/.*.fit"));
			for (String resource : resourceList) {
				int index = resource.indexOf(path);
				String name = resource.substring(index+path.length());
				retval.add(name);
			}
		}
		return retval;
	}

	/**
	 * Get the flux vector
	 * 
	 * @return the value of flux
	 */
	public double[] getFlux() {
		return flux;
	}

	/**
	 * Set the flux values to a provided array.
	 * 
	 * @param value
	 *            flux values
	 */
	public void setFlux(double[] value) {
		flux = value;
	}

	/**
	 * Get the wavelength scale on which the SED is defined.
	 * 
	 * @return the wavelength scale on which the SED is defined.
	 */
	public WavelengthScale getLambda() {
		return lambda;
	}

	/**
	 * Set the wavalangth scale.
	 * 
	 * @param value
	 *            the new scale.
	 */
	private void setLambda(WavelengthScale value) {
		assert (flux.length == value.size());
		lambda = value;
	}

	// Operations
	/**
	 * Apply a filter to the SED
	 * 
	 * @param filter
	 *            the filter to apply
	 */
	public void multiply(AstronomicalFilter filter) {
		if (!filter.getLambda().equals(getLambda())) {
			throw new IllegalArgumentException(
					"scales must be identical but filter scale="
							+ filter.getLambda() + " sed=" + getLambda());
		}
		ValuedBinIterator fluxIterator = iterator();
		int fluxIndex = 0;
		for (ValuedBin filterBin : filter) {
			ValuedBin fluxBin = fluxIterator.next();
			flux[fluxIndex++] = fluxBin.getValue() * filterBin.getValue();
		}
	}

	/**
	 * Integrate the energy a redshifted SED over a filter
	 * 
	 * @param filter
	 *            the filter to integrate over
	 * @param redshift
	 *            redshift the spectrum by this amount before integration
	 * @return the integrated flux
	 */
	public double integrateEnergy(AstronomicalFilter filter, double redshift) {
		
		SED shiftedSED;
		if (redshift == 0) {
			shiftedSED = this;
		} else {
			shiftedSED = this.clone();
			shiftedSED.redshift(redshift, filter.getLambda());
		}
		double retval = shiftedSED.integrateEnergy(filter);
		return retval;
	}

	/**
	 * Integrate the energy an SED over a filter
	 * 
	 * @param filter
	 *            the filter to integrate over
	 * @return the integrated flux
	 */
	public double integrateEnergy(AstronomicalFilter filter) {
		// Hogg:  
		//        f_nu is in energy/time/frequency
		// R is observed filter
		// Q is absolute filter
		
		ValuedBinIterator fluxIterator = iterator();
		ValuedBin fluxBin = null;
		double integratedFlux = 0.0;
		for (ValuedBin filterBin : filter) {
			fluxBin = fluxIterator.next();
			if (!(fluxBin.getWavelengthBin().equals(filterBin.getWavelengthBin()))) {
				throw new IllegalArgumentException(
						"WavelengthBins not equal:  flux="
								+ fluxBin.getWavelengthBin() + " filter="
								+ filterBin.getWavelengthBin());
			}
			double fluxValue = fluxBin.getValue();
			double filterValue = filterBin.getValue();
			integratedFlux += fluxValue * filterValue;
		}
		return integratedFlux;
	}

	/**
	 * get the number of photons in this SED observed through the Filter
	 * 
	 * @param filter
	 *            the Filter to use
	 * @return the number of photons
	 */
	public double integratePhotons(AstronomicalFilter filter) {

		if (!lambda.equals(filter.getLambda())) {
			throw new Error(
					"flux and filter wavelength scales are not identical");
		}
		ValuedBinIterator fluxIterator = photonIterator();
		ValuedBin fluxBin = null;
		double integratedFlux = 0.0;
		for (ValuedBin filterBin : filter) {
			fluxBin = fluxIterator.next();
			double temp = fluxBin.getValue() * filterBin.getValue();
			integratedFlux += temp;
		}
		return integratedFlux;
	}

	/**
	 * @param title
	 *            to put at the top of the plot
	 * @return the chart made in other methods
	 */
	public JFreeChart getChart(String title) {
		XYSeries temp = new XYSeries(getName());
		Scaling scaling = lambda.getScaling();
		for (ValuedBin vb : this) {
			double x = vb.getWavelengthBin().getMeanLambda(scaling);
			double y = vb.getValue();
			temp.add(x, y);
		}
		XYSeriesCollection xysc = new XYSeriesCollection(temp);

		JFreeChart retval = ChartFactory.createScatterPlot(title,
				"wavelength [Angstrom]", "flux [ergs/sec/cm^2] per bin", xysc,
				PlotOrientation.VERTICAL, true, true, true);
		XYLineAndShapeRenderer xyItemRenderer = (XYLineAndShapeRenderer) retval
				.getXYPlot().getRendererForDataset(xysc);
		xyItemRenderer.setSeriesLinesVisible(0, true);
		xyItemRenderer.setSeriesShapesVisible(0, false);
		return retval;
	}

	/**
	 * Rescale the flux through a provided filter to have a specified value
	 * 
	 * @param filter
	 *            the filter in which to normalize the flux
	 * @param targetFlux
	 *            the desired total flux
	 */
	public void normalizeEnergy(AstronomicalFilter filter, double targetFlux) {
		double currentFlux = integrateEnergy(filter);
		double ratio = targetFlux / currentFlux;
		scaleFlux(ratio);
	}

	/**
	 * Rescale the flux through a provided filter to have a specified value
	 * 
	 * @param filter
	 *            the filter in which to normalize the flux
	 * @param targetFlux
	 *            the desired total flux
	 */
	public void normalizePhotons(AstronomicalFilter filter, double targetFlux) {
		double currentFlux = integratePhotons(filter);
		double ratio = targetFlux / currentFlux;
		scaleFlux(ratio);
	}

	/**
	 * scale the luminosity, based on transmission type and calibration type of
	 * the filter
	 * 
	 * @param filter
	 *            used for flux integrations
	 * @param magnitude
	 *            set point for the magnitude
	 */
	public void setLuminosity(AstronomicalFilter filter, double magnitude) {
		setLuminosity(filter, magnitude, filter.getCalibrationType());
	}

	/**
	 * scale the luminosity, based on transmission type and calibration type
	 * 
	 * @param filter
	 *            used for flux integration
	 * @param magnitude
	 *            set point for the magnitude
	 * @param calibrationType
	 *            the calibration type
	 */
	public void setLuminosity(AstronomicalFilter filter, double magnitude,
			CalibrationType calibrationType) {
		double ratio = getRatio(filter, magnitude, calibrationType);
		scaleFlux(ratio);
	}

	/**
	 * @param factor
	 *            multiply flux by this factor
	 */
	
	public void scaleFlux(double factor) {
		for (int i = 0; i < flux.length; i++) {
			flux[i] *= factor;
		}
	}
	
	/**
	 * @param filter
	 * @param magnitude
	 * @return the number to multipy this SED by so it will have magnitude in
	 *         the filter for the filter's CalibrationType
	 */
	public double getRatio(AstronomicalFilter filter, double magnitude) {
		CalibrationType calibrationType = filter.getCalibrationType();
		return getRatio(filter, magnitude, calibrationType);
	}

	/**
	 * @param filter
	 * @param magnitude
	 * @param calibrationType
	 * @return the number to multiply this SED by so it will have magnitude in
	 *         the filter for calibrationType
	 */
	public double getRatio(AstronomicalFilter filter, double magnitude,
			CalibrationType calibrationType) {
		double currentMagnitude = getMagnitude(filter, calibrationType);
		double ratio = Math.pow(10.0, 0.4 * (currentMagnitude - magnitude));
		return ratio;
	}

	/**
	 * Integrate energy over the whole SED
	 * 
	 * @return the integrated energy
	 */
	public double integrateEnergy() {
		double integratedFlux = 0.0;
		for (ValuedBin bin : this) {
			integratedFlux += bin.getValue();
		}
		return integratedFlux;
	}

	/**
	 * Integrate photons over the whole SED
	 * 
	 * @return the integrated number of photons
	 */
	public double integratePhotons() {
		double integratedPhotons = 0.0;
		ValuedBinIterator vbi = this.photonIterator();
		while (vbi.hasNext()) {
			ValuedBin bin = vbi.next();
			integratedPhotons += bin.getValue();
		}
		return integratedPhotons;
	}

	/**
	 * Integrate photons over the whole SED
	 * 
	 * @return the integrated number of photons
	 */
	public double integratePhotonsSnap() {
		double integratedPhotons = 0.0;
		ValuedBinIterator vbi = this.photonIterator();
		int i = 0;
		while (vbi.hasNext()) {
			double fudge = 1.0;
			if (i == 0 || i == this.size() - 1)
				fudge = 0.5;
			ValuedBin bin = vbi.next();
			integratedPhotons += fudge * bin.getValue();
			i++;
		}
		return integratedPhotons;
	}

	/**
	 * Rescale to the targetEnergy
	 * 
	 * @param targetEnergy
	 *            the desired total energy
	 */
	public void normalizeEnergy(double targetEnergy) {
		double currentFlux = integrateEnergy();
		double ratio = targetEnergy / currentFlux;
		scaleFlux(ratio);
	}

	/**
	 * Rescale to the targetPhotons
	 * 
	 * @param targetPhotons
	 *            the desired total number of photons
	 */
	public void normalizePhotons(double targetPhotons) {
		double currentFlux = integratePhotons();
		double ratio = targetPhotons / currentFlux;
		scaleFlux(ratio);
	}

	/**
	 * rescale flux to the target magnitude at the reshift
	 * 
	 * @param filter
	 *            the filter corresponding to the magnitude
	 * @param finalMagnitude
	 *            the desired magnitude
	 * @param redshift the reshift
	 */
	public void normalizeMagnitude(AstronomicalFilter filter, double finalMagnitude, double redshift) {
		//double zeropoint = filter.getCalibrationType().getZeropoint();
		//double currentMagnitude = getMagnitude(filter, redshift) + zeropoint;
		double currentMagnitude = getMagnitude(filter, redshift);
		scaleFlux(currentMagnitude, finalMagnitude);
	}
	/**
	 * Rescale to the target magnitude at redshift=0.0
	 * 
	 * @param filter
	 *            the filter corresponding to the magnitude
	 * @param finalMagnitude
	 *            the desired magnitude
	 */
	public void normalizeMagnitude(AstronomicalFilter filter,
			double finalMagnitude) {
		double currentMagnitude = getMagnitude(filter);
		scaleFlux(currentMagnitude, finalMagnitude);
	}

	/**
	 * scale the flux so the spectrum changes from currentMagnitude to final
	 * Magnitude
	 * 
	 * @param currentMagnitude
	 * @param finalMagnitude
	 */
	public void scaleFlux(double currentMagnitude, double finalMagnitude) {
		double ratio = Math.pow(10, (currentMagnitude - finalMagnitude) / 2.5);
		scaleFlux(ratio);
	}

	/**
	 * shift the wavelengths to this redshift
	 * 
	 * @param z
	 *            the change in redshift
	 * @param newScale
	 *            the new scale for the SED
	 */
	public void redshift(double z, WavelengthScale newScale) {
		// Try it the fast way
		try {
			redshiftFast(z, newScale);
		} catch (IllegalArgumentException e) {
			// If the scales don't allow that, then use the slow method
			redshiftSlow(z, newScale);
		}
		return;
	}

	/**
	 * called if the binning does not allow the fast algorithm
	 * 
	 * @param z
	 * @param newScale
	 */
	public void redshiftSlow(double z, WavelengthScale newScale) {
		double[] newFlux = new double[newScale.size()];
		Iterator<ValuedBin> oldIterator = photonIterator();
		while (oldIterator.hasNext()) {
			ValuedBin oldBin = oldIterator.next();
			double oMin = (1 + z) * oldBin.getWavelengthBin().getLambdaMin();
			double oMax = (1 + z) * oldBin.getWavelengthBin().getLambdaMax();
			double photons = oldBin.getValue();
			Iterator<WavelengthBin> newIterator = newScale.iterator();
			int iNewBin = 0;
			while (newIterator.hasNext()) {
				WavelengthBin newBin = newIterator.next();
				double nMin = newBin.getLambdaMin();
				double nMax = newBin.getLambdaMax();
				if (nMin >= oMax) {
					// new bin is completely above old bin
					break;

				} else if (nMax > oMin) {
					if (nMin >= oMin && nMin < oMax && nMax >= oMax) {
						// new bin straddles max of old bin
						newFlux[iNewBin] += photons * (oMax - nMin)
								/ (oMax - oMin);

					} else if (nMin >= oMin && nMin < oMax && nMax >= oMin
							&& nMax < oMax) {
						// new bin contained in old bin
						newFlux[iNewBin] += photons * (nMax - nMin)
								/ (oMax - oMin);

					} else if (oMin >= nMin && oMin < nMax && oMax >= nMin
							&& oMax < nMax) {
						// old bin contained in new bin
						newFlux[iNewBin] += photons;

					} else if (nMin < oMin && nMax >= oMin && nMax < oMax) {
						// new bin straddles min of old bin
						newFlux[iNewBin] += photons * (nMax - oMin)
								/ (oMax - oMin);
					}
				}
				iNewBin++;
			}
		}

		flux = newFlux;
		lambda = newScale;

		// convert the flux back to ergs from photons
		convertPhotonsToErgs();

	}

	// this should be private
	private void convertErgsToPhotons() {
		int iBin = 0;
		for (WavelengthBin wb : lambda) {
			flux[iBin++] *= wb.getMeanLambda(lambda.getScaling()) / Constants.h / Constants.c;
		}
	}

	// this should be private
	private void convertPhotonsToErgs() {
		int iBin = 0;
		for (WavelengthBin wb : lambda) {
			flux[iBin++] /= wb.getMeanLambda(lambda.getScaling()) / Constants.h / Constants.c;
		}
	}

	/**
	 * call this to test how long the conversion take
	 */
	public void convertErgsToPhotonsAndBackAgain() {
		convertErgsToPhotons();
		convertPhotonsToErgs();
	}
	/**
	 * @param z
	 * @param newScale
	 * @throws IllegalArgumentException
	 *             if the scales are not LOGARITHMIC and have the same bin size
	 */
	public void redshiftFast(double z, WavelengthScale newScale) {
		if (lambda.getScaling() != Scaling.LOGARITHMIC) {
			throw new IllegalArgumentException(
					"redshiftFast only works for scaling=LOGARITHMIC");
		}
		if (newScale.getScaling() != Scaling.LOGARITHMIC) {
			throw new IllegalArgumentException(
					"redshiftFast only works for scaling=LOGARITHMIC");

		}
		double logBinSize = lambda.getBinSize();
		{
			double temp = 1 - logBinSize / newScale.getBinSize();
			if (Math.abs(temp) > 1e-12) {
				throw new IllegalArgumentException(
						"redshiftFast only works if binsizes are the same but temp="
								+ temp + " logBinSize=" + logBinSize
								+ " newScale.getBinSize()="
								+ newScale.getBinSize());
			}
		}
		double delta = (Math.log10((1 + z) * lambda.getLambdaMin()) - Math
				.log10(newScale.getLambdaMin()))
				/ logBinSize;
		int iDelta = (int) Math.floor(delta);
		// double fraction = delta%1.0;
		convertErgsToPhotons();
		double[] newFlux = new double[lambda.getWavelengths().length];
		for (int i = 0; i < newFlux.length; i++) {
			try {
				newFlux[i] = flux[i - iDelta];
			} catch (ArrayIndexOutOfBoundsException e) {
			}
		}
		setFlux(newFlux);
		setLambda(newScale);
		convertPhotonsToErgs();
	}

	/**
	 * redshift the spectrum to the same wavelength scale.
	 * 
	 * @param redshift
	 * @throws IllegalArgumentException
	 *             if the scaling is not LOGARITHMIC
	 */
	public void redshiftFast(double redshift) {
		WavelengthScale wsNew = lambda.redshift(redshift);
		redshiftFast(redshift, wsNew);
	}

	/**
	 * plot flux per second per wavelength unit at distance
	 * 
	 * @param title
	 *            to put on the top of the chart
	 * @param seds
	 *            the SEDs to plot
	 * @param names
	 *            legend names for the SEDs
	 * @param ergs
	 *            true to plot flux in energy; false to plot flux in photons
	 * @param logLambda
	 *            true to have the wavelengths plotted per logLambda
	 * @param logFlux true to have the flux plotted on a logarithmic scale
	 * @param photonSphereSurface
	 *            surface area at detection
	 * @return the chart
	 */
	public static JFreeChart getChart(String title, SED[] seds, String[] names,
			boolean ergs, boolean logLambda, boolean logFlux, double photonSphereSurface) {
		return getChart(title, seds, names, ergs, logLambda, false, logFlux,
				photonSphereSurface, null, null);
	}

	/**
	 * @param title
	 * @param seds
	 * @param names
	 * @param ergs
	 * @param logLambda
	 * @param plotLogLambda whether to plot the domain axis logarithmically
	 * @param logFlux 
	 * @param photonSphereSurface
	 * @param minLambda
	 *            in microns
	 * @param maxLambda
	 *            in microns
	 * @return the chart
	 */
	public static JFreeChart getChart(String title, SED[] seds, String[] names,
			boolean ergs, boolean logLambda, boolean plotLogLambda, boolean logFlux, double photonSphereSurface,
			Double minLambda, Double maxLambda) {
		XYSeriesCollection xySeriesCollection = new XYSeriesCollection();
		int j = 0;
		String rangeLabel = getRangeLabel(ergs, logLambda, photonSphereSurface);
		for (SED sed : seds) {
			XYSeries xySeries = SED.getSeries(sed, names[j++], ergs, logLambda, photonSphereSurface);
			xySeriesCollection.addSeries(xySeries);
		}

		JFreeChart tempChart = ChartFactory.createScatterPlot(title,
				"wavelength [microns]", rangeLabel, xySeriesCollection,
				PlotOrientation.VERTICAL, true, true, true);
		XYLineAndShapeRenderer xyItemRenderer = (XYLineAndShapeRenderer) tempChart
				.getXYPlot().getRendererForDataset(xySeriesCollection);
		for (int iSeries=0; iSeries<seds.length; iSeries++) {
			xyItemRenderer.setSeriesLinesVisible(iSeries,true);
			xyItemRenderer.setSeriesShapesVisible(iSeries, false);
		}
		
		LogarithmicAxis rangeAxis = new LogarithmicAxis(rangeLabel);
		rangeAxis.setLog10TickLabelsFlag(true);

		if (logFlux) {
			double minValue = getMin(logFlux, seds);		
			double maxValue = getMax(seds);
			double dLog = Math.log10(maxValue) - Math.log10(minValue);
			double fraction = 0.05;
			// double rangeBelowMin = Math.pow(10, Math.log10(minValue) - dLog*(2*fraction/(1-2*fraction)));

			double rangeMin = Math.pow(10, Math.log10(minValue) - dLog*fraction/(1-2*fraction));
			double rangeMax = Math.pow(10, Math.log10(maxValue) + dLog*fraction/(1-2*fraction));			
			rangeAxis.setLowerBound(rangeMin);
			rangeAxis.setUpperBound(rangeMax);
			((XYPlot) tempChart.getPlot()).setRangeAxis(rangeAxis);			
		}
		
		// NumberAxis range = ((NumberAxis)
		// tempChart.getXYPlot().getRangeAxis());
		// JFreeChartUtil.setTickUnit(range);
		JFreeChartUtil.setRangeTicks(tempChart.getXYPlot(), minLambda,
				maxLambda);

		if (plotLogLambda) {
			LogarithmicAxis domainAxis = new LogarithmicAxis("wavelength [microns]");
			domainAxis.setLog10TickLabelsFlag(true);
			((XYPlot) tempChart.getPlot()).setDomainAxis(domainAxis);			
		}
		NumberAxis domain = ((NumberAxis) tempChart.getXYPlot().getDomainAxis());
		if (minLambda != null) {
			domain.setLowerBound(minLambda);
		}
		if (maxLambda != null) {
			domain.setUpperBound(maxLambda);
		}
		return tempChart;
	}

	private static double getMax(SED[] seds) {
		double retval = Double.NaN;
		for (SED sed : seds) {
			double max = sed.getMaxValue();
			if (Double.isNaN(retval)) {
				retval = max;
			} if (!Double.isNaN(max)) {
				retval = Math.max(retval, max);
			}
		}
		return retval;	
		
	}

	private double getMaxValue() {
		double retval = Double.NaN;
		for (double value : flux) {
			if (!Double.isNaN(value)) {
				if (!(value < retval)) {
					retval = value;
				}
			}
		}
		return retval;		
	}

	private static double getMin(boolean logFlux, SED[] seds) {
		double retval = Double.NaN;
		for (SED sed : seds) {
			double min = sed.getMinValue(logFlux);
			if (!Double.isNaN(min)) {
				if (Double.isNaN(retval)) {
					retval = min;
				} if (!Double.isNaN(min)) {
					retval = Math.min(retval, min);
				}
			}
		}
		return retval;
	}
	private double getMinValue(boolean logarithmicRange) {
		double retval = Double.NaN;
		for (double value : flux) {
			if (!Double.isNaN(value)) {
				if (! (value > retval)) {
					if ( !logarithmicRange || value > Double.MIN_VALUE) {
						retval = value;
					}
				}
			}
		}
		return retval;		
	}


	private static String getRangeLabel(boolean ergs, boolean logLambda,
			double photonSphereSurface) {
		String retval = "flux [";
		if (ergs) {
			retval += "ergs/sec";
		} else {
			retval += "photons/sec";
		}
		if (photonSphereSurface > 0) {
			retval += "/cm^2";
		}
		if (logLambda) {
			retval += "/log lambda";
		} else {
			retval += "/Angstrom";
		}
		retval += "]";

		return retval;
	}

	/**
	 * 
	 * @param sed
	 * @param legend
	 * @param ergs
	 * @param photonSphereSurface
	 * @param logLambda
	 * @return the series
	 */
	public static XYSeries getSeries(SED sed, String legend, boolean ergs,
			boolean logLambda, double photonSphereSurface) {
		return getSeries(sed, legend, ergs, logLambda, photonSphereSurface,
				null, null);
	}

	/**
	 * 
	 * @param sed
	 * @param legend
	 * @param ergs
	 * @param photonSphereSurface surface area of the photon, in cm^2
	 * @param logLambda
	 * @param min
	 *            in microns
	 * @param max
	 *            in microns
	 * @return the series
	 */
	public static XYSeries getSeries(SED sed, String legend, boolean ergs,
			boolean logLambda, double photonSphereSurface, Double min,
			Double max) {
		XYSeries xySeries = new XYSeries(legend);
		Scaling scaling = sed.lambda.getScaling();
		Iterator<ValuedBin> iterator;
		if (ergs) {
			iterator = sed.iterator();
		} else {
			iterator = sed.photonIterator();
		}
		
		boolean notDone = true;
		while (iterator.hasNext()) {
			ValuedBin bin = iterator.next();
			double lambda = bin.getWavelengthBin().getMeanLambda(scaling) / 1e4;
			if (min == null || lambda > min) {
				if (max == null || lambda < max) {
					double value = bin.getValue();
					if (photonSphereSurface > 0.0) {
						value *= photonSphereSurface;
					} else {
						value *= Constants.tenPcCm2;
					}
					if (logLambda) {
						value /= bin.getWavelengthBin().logBinWidth();
					} else {
						value /= bin.getWavelengthBin().binWidth();
					}
					if (notDone && lambda > 0.49) {
						System.out.println("SED lambda="+lambda+" value="+value);
					}
					xySeries.add(lambda, value);
					if (notDone && lambda > 0.49) {
						notDone = false;
					}
				}
			}
		}
		return xySeries;
	}

	/**
	 * 
	 * @param sed
	 * @param redshift 
	 * @param legend
	 * @param ergs
	 * @param photonSphereSurface
	 * @param logLambda
	 * @param min
	 *            in microns
	 * @param max
	 *            in microns
	 * @return the series
	 */
	public static XYSeries getSeriesRedshifted(SED sed, double redshift,
			String legend, boolean ergs, boolean logLambda,
			double photonSphereSurface, Double min, Double max) {
		XYSeries xySeries = new XYSeries(legend);
		Scaling scaling = sed.lambda.getScaling();
		Iterator<ValuedBin> iterator;
		if (ergs) {
			iterator = sed.iterator();
		} else {
			iterator = sed.photonIterator();
		}
		while (iterator.hasNext()) {
			ValuedBin bin = iterator.next();
			double lambda = bin.getWavelengthBin().getMeanLambda(scaling) / 1e4;
			double lambdaShifted = lambda * (1 + redshift);
			if (min == null || lambdaShifted > min) {
				if (max == null || lambdaShifted < max) {
					double value = bin.getValue();
					if (photonSphereSurface > 0) {
						value *= Constants.tenPcCm2 / photonSphereSurface;
					} else {
						value *= Constants.tenPcCm2;
					}
					if (logLambda) {
						value /= bin.getWavelengthBin().logBinWidth();
					} else {
						value /= bin.getWavelengthBin().binWidth();
					}
					xySeries.add((1 + redshift) * lambda, value);
				}
			}
		}
		return xySeries;
	}

	/**
	 * 
	 * @param sed
	 * @param legend
	 * @return the series: photons/sec/logLambda vs Angstrom
	 */
	public static XYSeries getSeriesLogLambda(SED sed, String legend) {
		XYSeries xySeries = new XYSeries(legend);
		Scaling scaling = sed.lambda.getScaling();
		Iterator<ValuedBin> iterator;
		iterator = sed.photonIterator();
		while (iterator.hasNext()) {
			ValuedBin bin = iterator.next();
			double value = bin.getValue() * Constants.tenPcCm2
					/ bin.getWavelengthBin().logBinWidth();
			xySeries.add(bin.getWavelengthBin().getMeanLambda(scaling), value);
		}
		return xySeries;
	}

	/**
	 * An iterator over the wavelength bins in the SED
	 * 
	 * @return an iterator over valued bins that yields ValuedBins with values
	 *         in units of ergs/cm^2/sec
	 */
	public ValuedBinIterator iterator() {
		Double[] f = new Double[flux.length];
		int index = 0;
		for (double d : flux) {
			f[index++] = d;
		}
		return new ValuedBinIterator(f, lambda);
	}

	/**
	 * An iterator over the wavelength bins in the SED
	 * 
	 * @return an iterator over valued bins that yields ValuedBins with values
	 *         in units of photons/cm^2/sec
	 */
	public ValuedBinIterator photonIterator() {
		Double[] f = new Double[flux.length];
		int index = 0;
		WavelengthBinIterator wbi = lambda.iterator();
		for (double d : flux) {
			WavelengthBin wb = wbi.next();
			f[index++] = d * wb.meanLambda() / Constants.h / Constants.c; // in photons/sec/cm^2
		}
		return new ValuedBinIterator(f, lambda);
	}

	/**
	 * calculate the magnitude in the filter, using the CalibrationType of the
	 * filter
	 * 
	 * @param filter
	 *            the filter
	 * @return the magnitude
	 */
	public double getMagnitude(AstronomicalFilter filter) {
		return getMagnitude(filter, 0.0);
	}

	/**
	 * calculate the magnitude in the filter, using the CalibrationType of the
	 * filter
	 * 
	 * @param filter
	 *            the filter
	 * @param redshift
	 *            redshift the SED by this amount before integration
	 * @return the magnitude
	 */
	public double getMagnitude(AstronomicalFilter filter, double redshift) {
		return getMagnitude(filter, filter.getCalibrationType(), redshift);
	}

	/**
	 * calculate the magnitude in the filter
	 * 
	 * @param filter
	 *            the filter
	 * @param calibrationType
	 *            defines what kind of magnitude to calculate
	 * @return the magnitude
	 */
	public double getMagnitude(AstronomicalFilter filter,
			CalibrationType calibrationType) {
		return getMagnitude(filter, calibrationType, 0.0);
	}

	/**
	 * calculate the magnitude in the filter at a given redshift
	 * 
	 * @param filter
	 *            the filter
	 * @param calibrationType
	 *            defines what kind of magnitude to calculate
	 * @param redshift
	 *            edshift the SED by this amount before integration
	 * @return the magnitude
	 */
	public double getMagnitude(AstronomicalFilter filter,
			CalibrationType calibrationType, double redshift) {
		double numerator = integrateEnergy(filter, redshift);
		double denominator = calibrationType.getDenominator();
		double zeropoint = calibrationType.getZeropoint();
		//System.out.println("SED.getMagnitude:  num="+numerator+" den="+denominator+" zero="+zeropoint+" calibrationType="+calibrationType+" redshift="+redshift);
		double retval = -2.5 * Math.log10(numerator / denominator);
		retval += zeropoint;
		return retval;
	}

	/**
	 * @return the number of bins in the SED
	 */
	public int size() {
		return lambda.size();
	}

	/**
	 * call getNames() and print the size
	 * 
	 * @param args
	 *            ignored
	 */
	public static void main(String[] args) {
		System.out.println(" Number of SEDs defined is "
				+ SED.getNames().size());
	}

	/**
	 * @param wavelength
	 *            in angstrom
	 * @return photons/logLambda/cm^2 at 10 pc
	 */
	public double getErgsPerLogLambda(double wavelength) {
		WavelengthBin bin = new WavelengthBin(0, 1);
		int iBin = lambda.getContainingBin(wavelength, bin);
		double value = flux[iBin];
		double logLambda = bin.logBinWidth();
		double retval = value / logLambda;
		return retval;
	}
	
	/**
	 * @param wavelength in Angstrom
	 * @return ergs/sec/cm^2/Angstrom
	 */
	public double getFlux(double wavelength) {
		WavelengthBin bin = new WavelengthBin(0, 1);
		int iBin = lambda.getContainingBin(wavelength, bin);
		double value = flux[iBin];
		double binWidth = bin.binWidth();
		double retval = value / binWidth;
		return retval;
	}


	/**
	 * @param wavelength
	 *            in angstrom
	 * @return photons/logLambda
	 */
	public double getPhotonsPerLogLambda(double wavelength) {
		WavelengthBin bin = new WavelengthBin(0, 1);
		int iBin = lambda.getContainingBin(wavelength, bin);
		double value = flux[iBin];
		double meanLambda = bin.meanLambda();
		double logBinWidth = bin.logBinWidth();
		double retval = value * meanLambda / Constants.h / Constants.c / logBinWidth;
		return retval;
	}

	/**
	 * scale the flux to convert from an observed magnitude at the distance to absolute magnitude
	 * @param distance in meters
	 */
	public void setDistance(double distance) {
		double ratio = Math.PI*Math.pow(distance, 2)/Constants.tenPc;
		scaleFlux(ratio);
	}

	/**
	 * @return the original points that were loaded to produce this SED
	 */
	public ArrayList<Point2D.Double> getPoints() {
		return points;
	}

	/**
	 * @param string
	 * @param seds
	 * @param names
	 * @param ergs
	 * @param logLambda
	 * @param photonSphereSurface
	 * @return chart make with logFlux=false
	 */
	public static JFreeChart getChart(String string, SED[] seds,
			String[] names, boolean ergs, boolean logLambda, double photonSphereSurface) {
		return getChart(string, seds, names, ergs, logLambda, false, photonSphereSurface);
	}

	/**
	 * @return the name of this SED
	 */
	public String getName() {
		return name;
	}
	
	/**
	 * @return Array.hashCode of the flux array
	 */
	public int fluxHashCode() {
		int retval = Arrays.hashCode(flux);
		return retval;
	}

	/**
	 * @param absoluteFilter
	 * @param apparentMagnitude
	 * @param apparentFilter
	 * @param distanceModulus
	 * @param redshift
	 * @return the absolute magnitude of an object with this spectrum and the given apparentMagnitude, distanceModulus, and redshift
	 */
	public double getAbsoluteMagnitude(AstronomicalFilter absoluteFilter,
			double apparentMagnitude, AstronomicalFilter apparentFilter,
			double distanceModulus, double redshift) {
		double kCorrection = getKCorrection(absoluteFilter, apparentFilter, redshift);
		double retval = apparentMagnitude - distanceModulus - kCorrection;
		return retval;
	}

	
	/**
	 * apparentMagnitude = absoluteMagnitude + distanceModulus + kCorrection
	 * @param absoluteFilter
	 * @param apparentFilter
	 * @param redshift the redshift
	 * @return the value of kCorrection
	 */
	public double getKCorrection(AstronomicalFilter absoluteFilter,
			AstronomicalFilter apparentFilter, double redshift) {
		double magz = -2.5*Math.log10(redshift+1);
		magz = 0.0;
		double magAbsolute = getMagnitude(absoluteFilter);
		double magApparent = getMagnitude(apparentFilter, redshift);
		double kCorrection = magz + magAbsolute - magApparent;
		System.out.println("SED.getKCorrection magz="+magz);
		System.out.println("SED.getKCorrection magAbsolute="+magAbsolute);
		System.out.println("SED.getKCorrection magApparent="+magApparent);
		return kCorrection;
	}
	
	/**
	 * gets the GOODS luminosity function template
	 * @param wavelengthScale
	 * @param index
	 * @return the SED
	 * @throws IOException
	 */
	public static SED getGlfcTemplate(WavelengthScale wavelengthScale, int index) throws IOException {
		String fullFileName = getGlfcTemplateName(index);
		SED retval = new SED(wavelengthScale, fullFileName);
		return retval;
	}

	/**
	 * @param index
	 * @return the path to the file that contains the spectrum
	 * This corresponds to Table 1 of Jouvel et al 2009.
	 * @throws IOException 
	 */
	public static String getGlfcTemplateName(int index) throws IOException {
		String location = "lephare/sed/GAL/";
		if (index < 1 || index > 66) {
			throw new IllegalArgumentException("index must be from 1 to 66 inclusive, but index="+index);
		}
		if (glfcFileNames == null) {
			glfcFileNames = new String[66];
			String fileName = "data/seds/"+location+"CE_NEW/CE_NEW_MOD.list";
			InputStream inputStream = SED.class.getResourceAsStream(fileName);
			BufferedReader bufferedReader = new BufferedReader(new InputStreamReader(inputStream));
			for (int i=0; i<66; i++) {
				glfcFileNames[i] = StringUtil.tokenize(location+bufferedReader.readLine(), " ")[0];
			}
		}
		return glfcFileNames[index-1];
	}

	/**
	 * @param filter
	 * @return the AB magnitude for this SED in the filter
	 */
	public double getABMagnitude(AstronomicalFilter filter) {
		if (filter == null) {
			throw new IllegalArgumentException("filter can not be null");
		}
		if (!filter.getLambda().equals(lambda)) {
			throw new IllegalArgumentException("lambda="+lambda+" not equal to filter.lambda="+filter.getLambda());
		}
		double[] meanLambdas = lambda.getMeanLambdas();
		DoubleArrayList transmissions = filter.getMeanTransmission();
		double sum1 = 0.0;
		double sum2 = 0.0;
		for (int iBin=0; iBin<lambda.getNBin(); iBin++) {
			double fLambda = flux[iBin];
			double meanLambda = meanLambdas[iBin];
			double transmission = transmissions.get(iBin);
			sum1 += transmission * fLambda * meanLambda;
			sum2 += transmission * meanLambda;
		}
		double retval = -2.5*Math.log10(sum1/sum2) - 48.6;
		System.out.println("SED.getABMagnitude:  filter="+filter+" sum1="+sum1+" sum2="+sum2+" retval="+retval);
		return retval;
	}

	/**
	 * @param name new name for this SED
	 */
	public void setName(String name) {
		this.name = name;
	}
}
