package gov.fnal.eag.sim.catalog;

import java.awt.geom.Point2D;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;

import org.jfree.chart.JFreeChart;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

import uk.ac.starlink.table.RowListStarTable;
import cern.colt.list.DoubleArrayList;
import gov.fnal.eag.sim.catalog.WavelengthScale.Scaling;
import gov.fnal.eag.sim.repository.image.ImageRepository;
import gov.fnal.eag.util.misc.Constants;
import gov.fnal.eag.util.misc.StringUtil;
import gov.fnal.eag.util.plot.JFreeChartUtil;

/**
 * describes a spectral energy distribution.
 * @author stoughto
 *
 */
public class Spectrum {
	private WavelengthScale ws;
	private double[] flux;
	/**
	 * the value of denominator calculated in the getMagnitude method
	 */
	public double denominator;
	/**
	 * the value of numerator calculated in the getMagnitude method
	 */
	public double numerator;
	/**
	 * @param ws wavelength bin, in Angstrom
	 * @param flux in ergs per second
	 */
	public Spectrum(WavelengthScale ws, double[] flux) {
		init(ws, flux);
/**
	 * @param ws
	 * @param name
	 */
		}

	/**
	 * 
	 * @param ws
	 * @param name of file to read in from disk or from the package
	 */
	public Spectrum(WavelengthScale ws, String name) {
		this.ws = ws;
		ArrayList<Point2D.Double> points = loadPoints(name);
		flux = new double[ws.size()];
		int iBin = 0;
		for (WavelengthBin wb : ws) {
			try {
				flux[iBin] = wb.integrate(points);
			} catch (IntegralNotDefinedException e) {
				throw new Error(e);
			}
			iBin++;
		}
	}
	private void init(WavelengthScale wsIn, double[] fluxIn) {
		if (wsIn == null || fluxIn == null) {
			throw new IllegalArgumentException("no nulls allowed but ws="+wsIn+" flux="+fluxIn);//
		}
		if (wsIn.size() != fluxIn.length) {
			throw new IllegalArgumentException("ws.size()="+wsIn.size()+"  flux.length="+fluxIn.length);
		}
		this.flux = fluxIn.clone();
		this.ws = wsIn;
	}

	/**
	 * @param filter
	 * @param stdFlux
	 * @return the magnitude
	 * 
	 * mag = -2.5 log10[numerator/denominator] where
	 * numerator = (integral lambda * flux(lambda) * h * c * filter(lambda) dLambda)
	 * denominator = (integral lambda * stdFlux(lambda) * h * c * filter(lambda) dLambda)
	 * 
	 */
	public double getMagnitude(AstronomicalFilter filter, Spectrum stdFlux) {
		if (filter == null || stdFlux == null) {
			throw new IllegalArgumentException("no nulls allowed but filter="+filter+" stdFlux="+stdFlux);
		}
		if (!ws.equals(filter.getLambda()) || !ws.equals(stdFlux.ws)) {
			throw new IllegalArgumentException("ws must all be equal but ws="+ws+" filter.lambda="+filter.getLambda()+" stdFlux.ws="+stdFlux.ws);
		}
		numerator = 0.0;
		denominator = 0.0;
		int iBin=0; 
		Scaling scaling = ws.getScaling();
		DoubleArrayList meanTransmission = filter.getMeanTransmission();
		for (WavelengthBin bin : ws) {
			double meanLambda = bin.getMeanLambda(scaling);
			numerator += meanLambda * this.flux[iBin] * meanTransmission.get(iBin);
			denominator += meanLambda * stdFlux.flux[iBin] * meanTransmission.get(iBin);
			iBin++;
		}
		double retval = -2.5*Math.log10(numerator/denominator);
		return retval;
	}

	/**
	 * @param filter
	 * @return the integrated number of photons from this spectrum and filter
	 */
	public double getNPhoton(AstronomicalFilter filter) {
		double countsPerSecond = 0.0;
		Scaling scaling = ws.getScaling();
		DoubleArrayList meanTransmission = filter.getMeanTransmission();
		int iBin = 0;
		for (WavelengthBin bin : ws) {
			double meanLambda = bin.getMeanLambda(scaling);
			countsPerSecond += meanLambda * this.flux[iBin] * meanTransmission.get(iBin);
			iBin++;
			
		}		
		countsPerSecond /= Constants.h*Constants.c;
		return countsPerSecond;
	}
	/**
	 * @param ws
	 * @return the spectrum of the AB standard for this wavelength scale
	 * the flux g_nu = 3631E-23 erg/sec/Hz
	 */
	public static Spectrum getABStandard(WavelengthScale ws) {
		double g_nu = 3631.0E-23; //erg/cm^2/sec/Hz
		double[] abStandardFlux = new double[ws.size()];
		int iBin=0;
		for (WavelengthBin bin : ws) {
			double lambdaMin = bin.getLambdaMin();
			double lambdaMax = bin.getLambdaMax();
			double nuMin = Constants.c/lambdaMin;
			double nuMax = Constants.c/lambdaMax;
			double energyInBin = g_nu*(nuMin-nuMax);
			abStandardFlux[iBin] = energyInBin;
			iBin++;
		}
		Spectrum retval = new Spectrum(ws, abStandardFlux);
		return retval;
	}

	/**
	 * @param name
	 *            spectrum name
	 * @return the points as read from the .sed or .fit file
	 * @throws Error
	 */
	public static ArrayList<Point2D.Double> loadPoints(String name)
	throws Error {
		ArrayList<Point2D.Double> points = new ArrayList<Point2D.Double>();
		String baseFileName = "data" + File.separator + "seds" + File.separator + name;

		InputStream inputStream;		
		try {
			// Look for the file directly
			inputStream = new FileInputStream(name);
		} catch (FileNotFoundException e1) {
			// Look for the file by name in the data/seds directory
			inputStream = SED.class.getResourceAsStream(baseFileName + ".sed");			
		}

		if (inputStream != null)  {			
			points = loadPointsText(inputStream);
		} else if ((inputStream = SED.class.getResourceAsStream(baseFileName
				+ ".fit")) != null) {
			RowListStarTable rlst = ImageRepository.readRlst(inputStream, 0);
			for (long row = 0; row < rlst.getRowCount(); row++) {
				Object[] temp = rlst.getRow(row);
				double lambda = (Float) temp[0];
				double flux = (Float) temp[1];
				Point2D.Double lf = new Point2D.Double(lambda, flux);
				points.add(lf);
			}
		} else {
			throw new Error("can not read fileName=" + baseFileName);
		}
		return points;
	}

	/**
	 * @param inputStream
	 * @return the points as loaded from the inputStream
	 */
	public static ArrayList<Point2D.Double> loadPointsText(InputStream inputStream) {
		BufferedReader bufferedReader = new BufferedReader(
				new InputStreamReader(inputStream));
		// Read the file one line at a time
		String line;
		ArrayList<Point2D.Double> points = new ArrayList<Point2D.Double>();

		try {
			while ((line = bufferedReader.readLine()) != null) {
				String[] tokens = StringUtil
				.removeDuplicateWhitespace(line).toString().split(
						" ");
				// Only use lines that do not start with a "#" and that have
				// two tokens
				if (!tokens[0].startsWith("#") && tokens.length == 2) {
					double lambda = Double.parseDouble(tokens[0]);
					double flux = Double.parseDouble(tokens[1]);
					Point2D.Double lf = new Point2D.Double(lambda, flux);
					points.add(lf);
				}
			}
		} catch (Exception e) {
			throw new Error(e);
		}
		return points;
	}

	/**
	 * @param filter
	 * @return a chart of this Spectrum, and this Spectrum filtered
	 */
	public JFreeChart getChart(AstronomicalFilter filter) {
		XYSeriesCollection xysc = new XYSeriesCollection();
		XYSeries xyFlux = new XYSeries("flux");
		XYSeries xyFluxFiltered = new XYSeries("flux filtered");
		xysc.addSeries(xyFlux);
		xysc.addSeries(xyFluxFiltered);
		int iBin = 0;
		for (WavelengthBin wb : ws) {
			double lambda = wb.getMeanLambda(ws.getScaling());
			xyFlux.add(lambda, flux[iBin]);
			double transmission = filter.getValueAt(lambda);
			xyFluxFiltered.add(lambda, flux[iBin]*transmission);
			iBin++;
		}
		JFreeChart retval = JFreeChartUtil.makeChart(xysc, "Spectrum.getChart", "lambda(A", "flux", null);
		return retval;
	}
}

