package gov.fnal.eag.sim.catalog;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;

import gov.fnal.eag.sim.random.SimRandomLuminosity;
import gov.fnal.eag.sim.random.SimRandomRedshift;
import gov.fnal.eag.sim.random.SimRandomSchechter;
import gov.fnal.eag.sim.random.SimRandomUniform;
import gov.fnal.eag.sim.random.UniformCoordinateGenerator;
import gov.fnal.eag.util.astronomy.Cosmo;
import gov.fnal.eag.wcs.SphericalCoordinate;

/**
 * an AstroCatalogGenerator for Galaxies following Jouvel etl al.:  Mock Galaxy Catalogs for DE  astro-ph 0902.0625v1
 * @author stoughto
 *
 *
 */
public class StephanieGLFC extends AstroCatalogGenerator {

	// each array is LF, z, phiStar, mStar, alpha
	double[][] lf  = {
		{1, 0.50, 16.7e-4, -21.15, -0.71},
		{1, 0.75, 18.0e-4, -21.13, -0.50},
		{1, 1.00,  9.5e-4, -21.72, -0.98},
		{1, 2.00,  8.5e-4, -21.72, -0.98},
		{1, 3.00,  7.5e-4, -21.72, -0.98},
		{1, 4.00,  6.5e-4, -21.72, -0.98},
		{2, 0.50, 24.0e-4, -21.08, -1.37},
		{2, 0.75, 25.0e-4, -21.29, -1.17},
		{2, 1.00, 25.0e-4, -21.20, -1.10},
		{2, 1.50, 24.4e-4, -21.15, -0.05},
		{2, 2.00, 24.4e-4, -21.15, -0.99},
		{2, 3.00, 24.4e-4, -21.15, -0.99},
		{2, 6.00, 24.4e-4, -21.15, -0.85},
		{3, 0.50, 48.6e-4, -18.84, -1.10},
		{3, 0.75, 33.1e-4, -19.67, -1.18},
		{3, 1.00, 33.1e-4, -20.55, -1.52},
		{3, 1.50, 30.0e-4, -20.62, -1.62},
		{3, 2.00, 25.0e-4, -20.72, -1.70},
		{3, 3.00, 25.0e-4, -20.72, -1.80},
		{3, 6.00, 25.0e-4, -20.72, -1.80},
	};
	int[][] typeMinMax = {
			{1, 1, 17},
			{2, 18, 55},
			{3, 56, 66}
	};
	
	String filterName;
	double [] alpha   = new double[] { -0.16,  -0.54,  -0.90,  -1.66};
	double [] mStar   = new double[] {-19.03, -19.18, -19.31, -19.36};  // will add 5*log(h)
	double [] phiStar = new double[] {  3.35,   4.83,   7.32,   4.09};  // will multiply by h^3*1e-3
	String[] sedNames = {"CWW_E_ext", "CWW_Sbc_ext", "CWW_Scd_ext", "CWW_Im_ext"};
	SimRandomRedshift[] redshiftGenerators = new SimRandomRedshift[lf.length];
	SimRandomSchechter[] luminosityGenerators = new SimRandomSchechter[lf.length];
	/**
	 * @param bandpass which bandpass to use		double hCubed = Math.pow(h,3);
		switch (bandpass) {
			case V:
				filterName = "pegase-V_BK78";
				break;
			default:
				throw new Error("trouble initializing Zucca");
		}

	 * @param min ra,dec point defines the minimum of the box
	 * @param max ra,dec point defines the maximum of the box.  If min.ra > max.ra include ra==0
	 * @param seed random number seed index
	 * @param cosmo cosmology to use
	 * @param zMin minimum redshift
	 * @param zMax maximum redshift
	 * @param ws 
	 * @throws IOException 
	 */
	public StephanieGLFC(SphericalCoordinate min, SphericalCoordinate max, int seed, Cosmo cosmo, double zMin, double zMax, WavelengthScale ws) throws IOException {
		double h = cosmo.getLittleH();
		double hCubed = Math.pow(h,3);
		Hashtable<Integer, SEDGenerator> sedGenerators = new Hashtable<Integer, SEDGenerator>();
		
		SimRandomUniform sru = new SimRandomUniform(seed+123);
		for (int iTypeMinMax=0; iTypeMinMax<typeMinMax.length; iTypeMinMax++) {
			int ilf = typeMinMax[iTypeMinMax][0];
			ArrayList<SED> seds = new ArrayList<SED>();
			for (int index=typeMinMax[iTypeMinMax][1]; index<=typeMinMax[iTypeMinMax][2]; index++) {
				SED sed = SED.getGlfcTemplate(ws, index);
				seds.add(sed);
			}
			SEDGenerator sedGenerator = new SEDGeneratorUniform(seds, sru);
			sedGenerators.put(ilf, sedGenerator);
		}
		
		for (int i=0; i<lf.length; i++) {
			int nBin = 100;
			SimRandomRedshift redshiftGenerator = new SimRandomRedshift(cosmo, seed+10+i, zMin, zMax, nBin);
			redshiftGenerators[i] = redshiftGenerator;
			int ilf = (int) lf[i][0];
			SEDGenerator sedGenerator = sedGenerators.get(ilf);
			double normalizedMStar = mStar[i]+5*Math.log10(h);
			double minMagnitude = -22+5*Math.log10(h);
			double maxMagnitude = -14+5*Math.log10(h);
			int nBinSchechter = 100;
			
			// Table 2 give phiStar in units of (10^-3)*(h^3)*(Mpc^-3)
			double normalizedPhiStar = phiStar[i]*1e-3*hCubed*1e-18; // in pc^-3;
			SimRandomLuminosity luminosityGenerator = new SimRandomSchechter(seed+30+i, normalizedPhiStar, alpha[i], normalizedMStar, minMagnitude, maxMagnitude, nBinSchechter);
			luminosityGenerator.setFilterName(filterName);
			luminosityGenerators[i] = (SimRandomSchechter) luminosityGenerator;			
			double surfaceDensity = AstroObjectGenerator.calculateSurfaceDensity(luminosityGenerators[i], redshiftGenerator);
			CoordinateGenerator coordinateGenerator = new UniformCoordinateGenerator(min, max, seed+i, surfaceDensity);
			MorphologyGenerator morphologyGenerator = new MorphologyGeneratorSDSSImageCosmosSize();
			
			AstroObjectGenerator astroObjectGenerator = new AstroObjectGenerator(coordinateGenerator, redshiftGenerator, sedGenerator, morphologyGenerator, luminosityGenerator);

			add(astroObjectGenerator);
		}
	}
	/**
	 * call the toPNG method on each luminosity generator
	 */
	public void makePlots() {
		for (int i=0; i<alpha.length; i++) {
			try {
				luminosityGenerators[i].toPNG("zucca-"+i, 800, 600);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}		
	}
}
