/**
 * 
 */
package gov.fnal.eag.sim.catalog;

import gov.fnal.eag.sim.catalog.WavelengthScale.Scaling;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;

/**
 * Specifies range of wavelengths
 * 
 * @author neilsen
 * @version $Name:  $ ($Id: WavelengthBin.java,v 1.23 2008/11/03 17:20:11 stoughto Exp $)
 */
public class WavelengthBin {
	//Fields
	private WBComparator comparator = new WBComparator();
	/**
	 *  wavelength of the left side of the bin
	 */
	Point2D.Double lambdaMinPoint;
	/**
	 * wavelength of the right side of the bin
	 */
	Point2D.Double lambdaMaxPoint;
	
	// Methods
	// Constructors
	/**
	 * Creates a specification for a wavelength bin
	 * 
	 * @param lambdaMin wavelength at left side (in Angstroms)
	 * @param lambdaMax wavelength at right side (in Angstroms)
	 */
	public WavelengthBin(double lambdaMin, double lambdaMax) {
		setMinMax(lambdaMin, lambdaMax);
	}

	/**
	 * set the min and max value of this bin
	 * @param lambdaMin
	 * @param lambdaMax
	 */
	public void setMinMax(double lambdaMin, double lambdaMax) {
		if (lambdaMax < lambdaMin) {
			throw new IllegalArgumentException("Maximum wavelength must be greater than the minimum wavelength");
		}
		lambdaMinPoint = new Point2D.Double(lambdaMin,0);
		lambdaMaxPoint = new Point2D.Double(lambdaMax,0);
	}
		
	/**
	 * returns the wavelength (in Angstroms) of the left boundry of the bin
	 * 
	 * @return the wavelength at left side
	 */
	public double getLambdaMin() {
		return lambdaMinPoint.x;
	}
	
	/**
	 * returns the wavelength (in Angstroms) of the right boundry of the bin
	 * 
	 * @return the wavelength at right side
	 */
	public double getLambdaMax() {
		return lambdaMaxPoint.x;
	}
	
	/**
	 * returns the wavelength (in Angstroms) of the center of the bin
	 * 
	 * @return the wavelength at the center
	 */
	public double meanLambda() {
		return (lambdaMinPoint.x + lambdaMaxPoint.x)/2;
	}
	
	/**
	 * @return the mean lambda in log space
	 */
	public double meanLambdaLog() {
		final double lMin = Math.log(lambdaMinPoint.x);
		final double lMax = Math.log(lambdaMaxPoint.x);
		return Math.exp((lMin+lMax)/2);
	}
	/**
	 * returns the width of a bin, in Angstroms
	 * 
	 * @return the bin width (in Angstroms)
	 */
	public double binWidth() {
		return lambdaMaxPoint.x-lambdaMinPoint.x;
	}
	
	/**
	 * @return width of the bin in log lambda
	 */
	public double logBinWidth() {
		return Math.log(lambdaMaxPoint.x)-Math.log(lambdaMinPoint.x);
	}
	
	/**
	 * Return true if the bins are equal
	 * 
	 * @param w
	 * @return true iff the bins have the same min and max
	 */
	public boolean equals(WavelengthBin w) {
		return (lambdaMinPoint.x == w.getLambdaMin() && lambdaMaxPoint.x == w.getLambdaMax());
	}
	
	/**
	 * integrate the x,y points with linear interpolation inside the bin
	 * @param points x,y values
	 * @return the integral
	 * @throws IntegralNotDefinedException if there are fewer than 2 points
	 */
	public double integrate(ArrayList<Point2D.Double> points) throws IntegralNotDefinedException {
		if (points.size() < 2) throw new IntegralNotDefinedException(" need at least two points but size="+points.size());
		Point2D.Double[] pointArray = new Point2D.Double[points.size()];
		{
			int i=0;
			for (Point2D.Double point : points) {
				pointArray[i++] = point;
			}
		}
		int iBinaryMin = Arrays.binarySearch(pointArray, lambdaMinPoint, comparator);
		int iMin = iBinaryMin;
		if (iBinaryMin<0) {
			iMin = -iMin-1;
		} else {
			// on the left, if you match a bin boundary, increment iMin
			iMin++;
		}

		int iBinaryMax = Arrays.binarySearch(pointArray, lambdaMaxPoint, comparator);
		int iMax = iBinaryMax;
		if (iBinaryMax<0) {
			iMax = -iMax-1;
		}
		// note that if you match a bin boundary on the right, you do not need to change iMax
	
		//System.out.println(" x.size="+x.size()+" iMin="+iMin+" iMax="+iMax+" x.min="+x.get(0)+" x.max="+x.get(x.size()-1)+" lambdaMin="+lambdaMin+" lambdaMax="+lambdaMax);
		
		// These asserts do not seem to be doing anything....
		// assert(iMin >= 1);
		// assert(iMax < x.size());
		
		double retval = 0.0;

		// take care of trouble cases
		if (iMin == 0) {
			//throw new IntegralNotDefinedException("low side of the bin is below the lowest x value:  lambdaMin="+lambdaMin+" x.get(0)="+x.get(0)); 
			return 0.0;
		}
		if (iMax == points.size()) {
			//throw new IntegralNotDefinedException("high side of bin is above the highest x value:  lambdaMax="+lambdaMax+" x.get(last)="+x.get(x.size()-1));
			return 0.0;
		}
		for (int i=iMin-1; i<iMax; i++) {
			double y0, y1;
			double x0, x1;
			if (i == iMin-1) {
				x0 = lambdaMinPoint.x;
				y0 = points.get(i).y + (x0-points.get(i).x)*(points.get(iMin).y-points.get(i).y)/(points.get(iMin).x-points.get(i).x);
			} else {
				x0 = points.get(i).x;
				y0 = points.get(i).y;
			}
			if (i == iMax-1) {
				x1 = lambdaMaxPoint.x;
				y1 = points.get(i).y + (x1-points.get(i).x)*(points.get(iMax).y-points.get(i).y)/(points.get(iMax).x-points.get(i).x);
			} else {
				x1 = points.get(i+1).x;
				y1 = points.get(i+1).y;
			}
			double deltaX = x1-x0;
			double increment = deltaX*(y0+y1)/2.0;
			retval += increment;
		}
		return retval;
	}

	/**
	 * @param scaling
	 * @return meanLambda() or meanLambdaLog, depending on scaling
	 */
	public double getMeanLambda(Scaling scaling) {
		double retval;
		if (scaling == Scaling.LINEAR) {
			retval = meanLambda();
		} else if (scaling == Scaling.LOGARITHMIC) {
			retval = meanLambdaLog();
		} else {
			throw new Error("can not deal with scaling="+scaling);
		}
		return retval;
	}
	
	private class WBComparator implements Comparator<Point2D.Double> {

		public int compare(Point2D.Double o1, Point2D.Double o2) {
			if (o1.x < o2.x) {
				return -1;
			} else if (o1.x > o2.x) {
				return 1;
			} else {
				return 0;
			}
		}

		
	}
	
	/**
	 * Return the string representation of the bin
	 * 
	 * @return the string representation of the bin
	 */
	@Override
	public String toString() {
		String retval = "lambdaMin="+lambdaMinPoint.x+" lambdaMax="+lambdaMaxPoint.x;
		return retval;
	}

	/**
	 * @param lambda
	 * @return true if lambdaMin <= lambda < lambdaMax
	 */
	public boolean contains(double lambda) {
		if (lambda < getLambdaMin()) return false;
		if (lambda >= getLambdaMax()) return false;
		return true;
	}

}
