/**
 * 
 */
package gov.fnal.eag.sim.catalog;

import java.lang.Math;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.lang.UnsupportedOperationException;

/**
 * Iterator over wavelength bins specified by a wavelength scale
 * 
 * @author neilsen
 * @version $Name:  $ ($Id: WavelengthBinIterator.java,v 1.5 2006/12/16 18:54:34 stoughto Exp $)
 * 
 */
public class WavelengthBinIterator implements Iterator<WavelengthBin> {

	//Fields
	
	//the specification of how to generate wavelength bins
	private WavelengthScale wavelengthScale;
	
	/**
	 * the index of the next bin (0 is the first)
	 */
	int currentBin = 0;
	
	//Methods
	/**
	 * Create a iterator over a given set of wavelength bins.
	 * 
	 * @param wavelengthScale the wavelength bin set to iterate over.
	 */
	public WavelengthBinIterator(WavelengthScale wavelengthScale) {
		this.wavelengthScale = wavelengthScale;
		currentBin = 0;
	}
	
	/**
	 * Returns <code>true</code> if there are more bins.
	 *  
	 *  @return <code>true</code> if there is another bin
	 */
	public boolean hasNext() {
		return currentBin < wavelengthScale.size() && currentBin >= -1;
	}

	/**
	 * Returns the next bin.
	 * 
	 * @return the next bin
	 * @throws NoSuchElementException
	 */
	public WavelengthBin next() {
		if (!hasNext()) {
			throw new NoSuchElementException("No more wavelength bins.");
		}
		
		double lambdaMin = 0;
		double lambdaMax = 0;
		double binWidth = 0;
		
		switch (wavelengthScale.getScaling()) {
		case LINEAR: 
			binWidth = (wavelengthScale.getLambdaMax() - wavelengthScale.getLambdaMin()) /  wavelengthScale.size();
			lambdaMin = wavelengthScale.getLambdaMin() + binWidth * currentBin;
			// rounding errors can make lambdaMax of the last bin a little too big.
			if (currentBin == wavelengthScale.size()-1) {
				lambdaMax = wavelengthScale.getLambdaMax();
			} else {
				lambdaMax = lambdaMin + binWidth;
			}
			break;
		case LOGARITHMIC:
			/*
			binWidth = Math.log10(wavelengthScale.getLambdaMax() - wavelengthScale.getLambdaMin()) / wavelengthScale.size();
			double min = binWidth * currentBin; // the distance from the start of the scale to the start of the bin 
			double max = min + binWidth; // the distance from the start of the scale to the end of the bin 
			lambdaMin = wavelengthScale.getLambdaMin() + Math.pow(10,min);
			if (currentBin == wavelengthScale.size()-1) {
				lambdaMax = wavelengthScale.getLambdaMax();
			} else {
				lambdaMax = wavelengthScale.getLambdaMin() + Math.pow(10,max);
			}
			*/
			double logLambdaMin = Math.log10(wavelengthScale.getLambdaMin());
			double logLambdaMax = Math.log10(wavelengthScale.getLambdaMax());
			double logBinWidth = (logLambdaMax-logLambdaMin)/wavelengthScale.size();
			lambdaMin = Math.pow(10, logLambdaMin + currentBin*logBinWidth);
			lambdaMax = Math.pow(10, logLambdaMin + (currentBin+1)*logBinWidth);
			break;
		default: assert false: "wavelength scaling " + wavelengthScale.getScaling() + " not implemented.";
		}
		
		currentBin++;
		assert lambdaMax > lambdaMin : "Invalid wavelength bin!";
		return new WavelengthBin(lambdaMin,lambdaMax);
	}

	/**
	 * Not implemented
	 * @throws java.lang.UnsupportedOperationException
	 */
	public void remove() {
		throw new UnsupportedOperationException("Bins cannot be removed from a wavelength scale");
	}

}
