package gov.fnal.eag.sim.catalog;

import cern.colt.list.DoubleArrayList;

/**
 * The specification for how a range of wavelengths is divided into bins.
 * 
 * @author neilsen
 * @version $Name:  $ ($Id: WavelengthScale.java,v 1.25 2009/03/23 14:49:54 stoughto Exp $)
 */
public class WavelengthScale implements Iterable<WavelengthBin>, Cloneable {
	/**
	 * describes the wavelength scale
	 */
	public enum Scaling {
		/**
		 * bins are all the same size
		 */
		LINEAR, 
		/**
		 * log of wavelength is same for each bin
		 */
		LOGARITHMIC
	}

	// Fields
	private static double defaultLambdaMin = 91.0;
	private static double defaultLambdaMax = 97400.0;
	private static Scaling defaultScaling = Scaling.LOGARITHMIC;
	private static int defaultNBin = 10000;

	private double lambdaMin; // wavelength of left side of first bin (Angstroms)
	private double lambdaMax; // wavelength of right side of last bin (Angstroms)
	private int nBin; // number of bins
	private Scaling scaling; // how the bin sizes vary with starting value
	private double[] meanLambdas = null;
	// Methods
	// Constructors
	/**
	 * Constructs a wavelength scale
	 * 
	 * @param lambdaMin wavelength at left side of zeroth bin
	 * @param lambdaMax wavelength at right side of last bin
	 * @param nBin number of bins
	 * @param scaling either LINEAR or LOGARITHMIC
	 */
	public WavelengthScale ( double lambdaMin, double lambdaMax, int nBin, Scaling scaling) {
		if (nBin <= 0) {
			throw new IllegalArgumentException("nBin must be positive by nBin="+nBin);
		}
		if (lambdaMax <= lambdaMin) {
			throw new IllegalArgumentException(String.format("lambdaMin=+%20.15f lambdaMax=%20.15f",lambdaMin, lambdaMax));
		}
		this.lambdaMin = lambdaMin;
		this.lambdaMax = lambdaMax;
		this.nBin = nBin;
		this.scaling = scaling;
	}

	/**
	 * create a WavelengthScale based on the input values
	 * @param in
	 */
	public WavelengthScale(DoubleArrayList in)  {
		if (in.size() < 2) throw new IllegalArgumentException("need at least two wavelengths to define a WavelengthScale");
		boolean isLinear = true;
		boolean isLogarithmic = true;
		double delta = in.get(1) - in.get(0);
		if (delta <= 0) throw new IllegalArgumentException(" delta wavelength must be >= 0 ");
		double deltaLog = Math.log10(in.get(1)) - Math.log10(in.get(0));
		for (int i=1; i<in.size(); i++) {
			double thisDelta = in.get(i) - in.get(i-1);
			if (thisDelta != delta) isLinear = false;
			if (thisDelta > 0) {
				double thisDeltaLog = Math.log10(in.get(i)) - Math.log10(in.get(i-1));
				double temp = Math.abs(thisDeltaLog-deltaLog);
				if (temp > 1e-6*deltaLog) isLogarithmic = false;
			} else {
				isLogarithmic = false;
			}
		}
		if (isLinear) {
			lambdaMin = in.get(0) - delta/2.0;
			lambdaMax = in.get(in.size()-1) + delta/2.0;
			nBin = in.size();
			scaling = Scaling.LINEAR;			
		} else if (isLogarithmic) {
			lambdaMin = Math.pow(10, Math.log10(in.get(0)) - deltaLog/2.0);
			lambdaMax = Math.pow(10, Math.log10(in.get(in.size()-1)) + deltaLog/2.0);
			nBin = in.size();
			scaling = Scaling.LOGARITHMIC;
		} else {
			throw new UnsupportedOperationException(" can only deal with scales that are linear or logarithmic");
		}
	}

	/**
	 * calculate the bin size, in angstrom for linear scaling, in log10(angstrom) for logarithmic
	 * @return the bin size
	 */
	public double getBinSize() {
		double binSize;
		if (scaling == Scaling.LINEAR) {
			binSize = (lambdaMax-lambdaMin)/nBin;
		} else if (scaling == Scaling.LOGARITHMIC) {
			binSize = (Math.log10(lambdaMax)-Math.log10(lambdaMin))/nBin;
		} else {
			throw new UnsupportedOperationException(" can only deal with scales that are linear or logarithmic");
		}
		return binSize;
	}
	// Accessor Methods
	/**
	 * Get the wavelength of the left side of the first bin (Angstroms)
	 * 
	 * @return the value of the left side of the first bin
	 */
	public double getLambdaMin (  ) {
		return lambdaMin;
	}

	/**
	 * Set the wavelength of the left side of the first bin (Angstroms)
	 * 
	 * @param value the value of the left side of the last bin
	 */
	public void setLambdaMin ( double value  ) {
		lambdaMin = value;
	}

	/**
	 * return the wavelength of right side of last bin (Angstroms)
	 * 
	 * @return wavelength of right side of last bin
	 */
	public double getLambdaMax (  ) {
		return lambdaMax;
	}

	/**
	 * Set the wavelength of the right side of the last bin (Angstroms)
	 * 
	 * @param value wavelength of right side of last bin
	 */
	public void setLambdaMax ( double value  ) {
		lambdaMax = value;
	}

	/**
	 * Return the number of bins
	 * 
	 * @return number of bins
	 */
	public int size (  ) {
		return nBin;
	}

	/**
	 * Set the number of bins
	 * 
	 * @param value number of bins
	 */
	public void setNBin ( int value  ) {
		nBin = value;
	}

	/**
	 * Return how the bin sizes vary with starting value
	 * 
	 * @return the scaling
	 */
	public Scaling getScaling (  ) {
		return scaling;
	}

	/**
	 * Set how the bin sizes vary with starting value
	 * 
	 * @param value the scaling (LINEAR or LOGARITHMIC)
	 */
	public void setScaling ( Scaling value  ) {
		scaling = value;
	}

	// Operations
	/**
	 * Return an array of the starting wavelengths of each bin
	 * 
	 * @return the left wavelength of each bin
	 */
	public Double[] getWavelengths ( ) {
		Double[] binMins = new Double[nBin];
		int binIndex = 0; // the index into the binMins array

		for(WavelengthBin bin: this) {
			binMins[binIndex++] = bin.getLambdaMin();
		}
		return binMins;
	}

	/**
	 * Return an array of the mean wavelengths of each bin
	 * 
	 * @return the mean wavelength of each bin
	 */
	public double[] getMeanLambdas ( ) {
		if (meanLambdas == null) {
			meanLambdas = new double[nBin];
			int binIndex = 0; // the index into the binMins array

			for(WavelengthBin bin: this) {
				if (scaling.equals(Scaling.LINEAR)) {
					meanLambdas[binIndex++] = bin.meanLambda();
				} else if (scaling.equals(Scaling.LOGARITHMIC)) {
					meanLambdas[binIndex++] = bin.meanLambdaLog();					
				} else {
					throw new Error("can not deal with scaling="+scaling);
				}
			}
		}
		return meanLambdas;
	}

	/**
	 * Return an iterator over wavelength bins.
	 * 
	 * @return an iterator over wavelength bins
	 */
	public WavelengthBinIterator iterator() {
		return new WavelengthBinIterator(this);
	}

	/**
	 * shallow copy constructor
	 * @param ws orignal to copy
	 */
	protected WavelengthScale(WavelengthScale ws) {
		lambdaMin = ws.lambdaMin;
		lambdaMax = ws.lambdaMax;
		nBin = ws.nBin;
		scaling = ws.scaling;
	}

	/**
	 * default constructor
	 */
	public WavelengthScale() {
		this(defaultLambdaMin, defaultLambdaMax, defaultNBin, defaultScaling);
	}

	/**
	 * Returns true if the argument represents an equivalent scale.
	 * 
	 * @param that
	 * @return true if the objects are equivalent
	 */
	public boolean equals(WavelengthScale that) {

		if (lambdaMin != that.lambdaMin) {
			return false;
		}
		if (lambdaMax != that.lambdaMax) {
			return false;
		}
		if (nBin != that.nBin) {
			return false;
		}
		if (scaling != that.scaling) {
			return false;
		}
		return true;
	}
	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		String retval = "lambdaMin="+lambdaMin+" labmdaMax="+lambdaMax+" nBin="+nBin+" scaling="+scaling;
		return retval;
	}
	/**
	 * @param redshift
	 * @return redshifted version of the wavelength scale
	 */
	public WavelengthScale redshift(double redshift) {
		WavelengthScale retval = new WavelengthScale(lambdaMin*(1+redshift), lambdaMax*(1+redshift), nBin, scaling);
		return retval;
	}

	/**
	 * @param wavelength
	 * @param bin min and max set to equal the containing bin
	 * @return a bin from this WavelengthScale containing this wavelength
	 * @throws IllegalArgumentException if the wavelength is not include in any bin
	 */
	public int getContainingBin(double wavelength, WavelengthBin bin) {
		if (wavelength < lambdaMin || wavelength > lambdaMax) {
			throw new IllegalArgumentException("wavelength not contained in lambdaMin="+lambdaMin+" and lambdaMax="+lambdaMax);
		}
		int iBin;
		double l0, l1;
		switch (scaling) {
		case LINEAR:
		{
			double binWidth = (lambdaMax - lambdaMin) /  nBin;
			iBin = (int) ((wavelength-lambdaMin)/binWidth);
			// rounding errors can make lambdaMax of the last bin a little too big.
			l0 = lambdaMin + iBin*binWidth;
			l1 = l0 + binWidth;
		}
		break;
		case LOGARITHMIC:
		{
			double logLambdaMin = Math.log10(lambdaMin);
			double logLambdaMax = Math.log10(lambdaMax);
			double logBinWidth = (logLambdaMax-logLambdaMin)/nBin;
			double logWavelength = Math.log10(wavelength);
			iBin = (int) ((logWavelength-logLambdaMin)/logBinWidth);
			double logL0 = logLambdaMin + iBin*logBinWidth;
			double logL1 = logL0 + logBinWidth;
			l0 = Math.pow(10, logL0);
			l1 = Math.pow(10, logL1);
		}
		break;
		default: 
			throw new IllegalArgumentException("scaling="+scaling+" is not supported");
		}
		bin.setMinMax(l0,l1);
		return iBin;		
	}

	/**
	 * @return a deep copy of WavelengthScale
	 */
	public WavelengthScale copy() {
		WavelengthScale retval = new WavelengthScale(lambdaMin, lambdaMax, nBin, scaling);
		return retval;
	}

	/**
	 * @return the number of bins 
	 */
	public int getNBin() {
		return nBin;
	}

	/**
	 * @param wavelength
	 * @return the bin number corresponding to this wavelength
	 */
	public int getIBin(double wavelength) {
		double binSize = getBinSize();
		int retval = Integer.MIN_VALUE;
		switch (scaling) {
		case LINEAR: {
			retval = (int) Math.floor((wavelength-lambdaMin)/binSize);
			break;
		}
		case LOGARITHMIC: {
			retval = (int) Math.floor((wavelength-lambdaMin)/binSize);
			break;
		}
		}
		return retval;
		}
}

