package gov.fnal.eag.sim.catalog;

import gov.fnal.eag.sim.random.SimRandomLuminosity;
import gov.fnal.eag.sim.random.SimRandomRedshift;
import gov.fnal.eag.sim.random.SimRandomSchechter;
import gov.fnal.eag.sim.random.UniformCoordinateGenerator;
import gov.fnal.eag.util.astronomy.Cosmo;
import gov.fnal.eag.wcs.SphericalCoordinate;

/**
 * an AstroCatalogGenerator for Galaxies following Zucca et al. astro-ph/0506393 "The VIMOS VLT Deep Survey"
 * @author stoughto
 *
 *
 */
public class Zucca extends AstroCatalogGenerator {
	
	String filterName;
	double [] alpha   = new double[] { -0.16,  -0.54,  -0.90,  -1.66};
	double [] mStar   = new double[] {-19.03, -19.18, -19.31, -19.36};  // will add 5*log(h)
	double [] phiStar = new double[] {  3.35,   4.83,   7.32,   4.09};  // will multiply by h^3*1e-3
	String[] sedNames = {"CWW_E_ext", "CWW_Sbc_ext", "CWW_Scd_ext", "CWW_Im_ext"};
	SimRandomRedshift[] redshiftGenerators = new SimRandomRedshift[sedNames.length];
	SimRandomSchechter[] luminosityGenerators = new SimRandomSchechter[sedNames.length];
	/**
	 * @param bandpass which bandpass to use
	 * @param min ra,dec point defines the minimum of the box
	 * @param max ra,dec point defines the maximum of the box.  If min.ra > max.ra include ra==0
	 * @param seed random number seed index
	 * @param cosmo cosmology to use
	 * @param zMin minimum redshift
	 * @param zMax maximum redshift
	 * @param ws 
	 */
	public Zucca(ZuccaBandpass bandpass, SphericalCoordinate min, SphericalCoordinate max, int seed, Cosmo cosmo, double zMin, double zMax, WavelengthScale ws) {
		double h = cosmo.getLittleH();
		double hCubed = Math.pow(h,3);
		switch (bandpass) {
			case V:
				filterName = "pegase-V_BK78";
				break;
			default:
				throw new Error("trouble initializing Zucca");
		}
		for (int i=0; i<alpha.length; i++) {
			
			int nBin = 100;
			SimRandomRedshift redshiftGenerator = new SimRandomRedshift(cosmo, seed+10+i, zMin, zMax, nBin);
			redshiftGenerators[i] = redshiftGenerator;
			
			SED sed = new SED(ws, sedNames[i]);
			SEDGenerator sedGenerator = new SEDGeneratorConstant(sed);
			//MorphologyGenerator morphologyGenerator = new MorphologyGeneratorShapeletCatalog(seed+20+i, "polarcatS.fits", true);
			MorphologyGenerator morphologyGenerator = new MorphologyGeneratorFromCatalog("data/shapelets/shapecatS.fits", seed+20+i);
			double normalizedMStar = mStar[i]+5*Math.log10(h);
			double minMagnitude = -22+5*Math.log10(h);
			double maxMagnitude = -14+5*Math.log10(h);
			int nBinSchechter = 100;
			
			// Table 2 give phiStar in units of (10^-3)*(h^3)*(Mpc^-3)
			double normalizedPhiStar = phiStar[i]*1e-3*hCubed*1e-18; // in pc^-3;
			SimRandomLuminosity luminosityGenerator = new SimRandomSchechter(seed+30+i, normalizedPhiStar, alpha[i], normalizedMStar, minMagnitude, maxMagnitude, nBinSchechter);
			luminosityGenerator.setFilterName(filterName);
			luminosityGenerators[i] = (SimRandomSchechter) luminosityGenerator;			
			double surfaceDensity = AstroObjectGenerator.calculateSurfaceDensity(luminosityGenerators[i], redshiftGenerator);
			CoordinateGenerator coordinateGenerator = new UniformCoordinateGenerator(min, max, seed+i, surfaceDensity);
			AstroObjectGenerator astroObjectGenerator = new AstroObjectGenerator(coordinateGenerator, redshiftGenerator, sedGenerator, morphologyGenerator, luminosityGenerator);

			add(astroObjectGenerator);
		}
	}
	/**
	 * call the toPNG method on each luminosity generator
	 */
	public void makePlots() {
		for (int i=0; i<alpha.length; i++) {
			try {
				luminosityGenerators[i].toPNG("zucca-"+i, 800, 600);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}		
	}
}
