package gov.fnal.eag.sim.crt;

import gov.fnal.eag.util.misc.StringUtil;
import gov.fnal.eag.wcs.SphericalCoordinate;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Properties;

/**
 * describes a Fast Fourier Transform Cylindrical Radio Telescope
 * @author stoughto
 *
 */
public class FFTCRT extends RadioTelescope {
	private int nx, ny, nt;
	private double dx,dy,dt;
	private double nuMin, nuMax;
	private double fMin, fMax;
	String antennaProperties;
	private AntennaResponse[][] antennaResponses;
	String gainProperties;
	private Gain[][] gains;
	String noiseTemperatureProperties;
	private NoiseTemperature[][] noiseTemperatures;
	
	/**
	 * @param properties to describe the CRT, specified in the init method.
	 */
	public FFTCRT(Properties properties) {
		init(properties);
	}

	/**
	 * @param fileName containing properties to describe the CRT, specified in the init method
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public FFTCRT(String fileName) throws FileNotFoundException, IOException {
		Properties properties = new Properties();
		properties.load(new FileInputStream(fileName));
		init(properties);
	}
	/**
	 * @param properties
	 * <ul>
	 * <li> dx separation between detectors along the cylinder axis (meters)
	 * <li> dy separation between cylinders (meters)
	 * <li> dt sampling time (seconds)
	 * <li> nx number of detectors along a cylinder
	 * <li> ny number of cylinders
	 * <li> nt number of time samples
	 * <li> fMin minimum frequency (Hz)
	 * <li> antennaProperties use "default" for now
	 * <li> gainProperties use "default" for now
	 * <li> noiseTemperatureProperties use "constant T" for now where T is noise temperature in degrees Kelvin
	 * </ul>
	 */
	private void init(Properties properties) {
		double longitude, latitude;
		if (properties.containsKey("longitude")) {
			longitude = StringUtil.getPropertyAsDouble(properties, "longitude");
		} else  {
			longitude = 0;
		}
		if (properties.containsKey("latitude")) {
			latitude = StringUtil.getPropertyAsDouble(properties, "latitude");
		} else {
			latitude = 0;
		}
		location = new SphericalCoordinate(longitude, latitude);
		dx = StringUtil.getPropertyAsDouble(properties, "dx");
		dy = StringUtil.getPropertyAsDouble(properties, "dy");
		dt = StringUtil.getPropertyAsDouble(properties, "dt");
		nx = StringUtil.getPropertyAsInt(properties, "nx");
		ny = StringUtil.getPropertyAsInt(properties, "ny");
		nt = StringUtil.getPropertyAsInt(properties, "nt");
		fMin = StringUtil.getPropertyAsDouble(properties, "fMin");
		fMax = fMin + 1/(2*dt);
		//  nuMin and nuMax are acting as a bandpass filter for now....
		nuMin = fMin;
		nuMax = fMax;
		antennaResponses = new AntennaResponse[nx][ny];
		antennaProperties = StringUtil.getPropertyAsString(properties, "antennaProperties");
		if (antennaProperties.equals("default")) {
			AntennaResponse antennaResponse = new AntennaResponseDefault(getDx(), dy);
			for (int ix=0; ix<nx; ix++) {
				for (int iy=0; iy<ny; iy++) {
					antennaResponses[ix][iy] = antennaResponse;
				}
			}
		} else {
			throw new Error("Can not deal with antennaProperties="+antennaProperties);
		}
		gains = new Gain[nx][ny];
		gainProperties = StringUtil.getPropertyAsString(properties, "gainProperties");
		if (gainProperties.equals("default")) {
			Gain gain = new GainDefault();
			for (int ix=0; ix<nx; ix++) {
				for (int iy=0; iy<ny; iy++) {
					gains[ix][iy] = gain;
				}
			}
		} else {
			throw new Error("Can not deal with gainProperties="+gainProperties);
		}
		noiseTemperatures = new NoiseTemperature[nx][ny];
		noiseTemperatureProperties = StringUtil.getPropertyAsString(properties, "noiseTemperatureProperties");
		if (noiseTemperatureProperties.startsWith("constant")) {
			String[] tokens = StringUtil.tokenize(noiseTemperatureProperties, " ");
			double constant = Double.parseDouble(tokens[1]);
			NoiseTemperature noiseTemperature = new NoiseTemperatureConstant(constant);
			for (int ix=0; ix<nx; ix++) {
				for (int iy=0; iy<ny; iy++) {
					noiseTemperatures[ix][iy] = noiseTemperature;
				}
			}
		} else {
			throw new Error("Can not deal with noiseTemperatureProperties="+noiseTemperatureProperties);
		}
		
	}


	/**
	 * @return the nx
	 */
	public int getNx() {
		return nx;
	}
	/**
	 * @return the ny
	 */
	public int getNy() {
		return ny;
	}
	/**
	 * @return the nt
	 */
	public int getNt() {
		return nt;
	}
	/**
	 * @return the dt
	 */
	public double getDt() {
		return dt;
	}

	/**
	 * @return the Gains
	 */
	public Gain[][] getGains() {
		return gains;
	}
	
	/**
	 * @return (nuMax-nuMin)/nt in sec^-1
	 */
	public double getDeltaF() {
		double retval = (nuMax-nuMin)/nt;
		return retval;
	}

	/**
	 * @return the antenna responses
	 */
	public AntennaResponse[][] getAntennaResponses() {
		return antennaResponses;
	}

	/**
	 * @return the dx
	 */
	public double getDx() {
		return dx;
	}
	/**
	 * @return the dy
	 */
	public double getDy() {
		return dy;
	}
	
	/**
	 * @return the array of NoiseTemperatures for this telescope
	 */
	public NoiseTemperature[][] getNoiseTemperatures() {
		return noiseTemperatures;
	}

	/**
	 * @return the minimum frequency
	 */
	public double getNuMin() {
		return nuMin;
	}
	/**
	 * @return the maximum frequency, in Hz
	 */
	public double getNuMax() {
		return nuMax;
	}


	/**
	 * @return the minimum frequency, in Hz
	 */
	public double getFMin() {
		return fMin;
	}

	/**
	 * @param nuBins
	 * @param sinThetaBins
	 * @param phiBins
	 * @param nu
	 * @param mu
	 * @param iota
	 * @param tPointSource
	 * @return a point source in one of the bins of the FFTCRT defined by the bins
	 */
	public static SkyTemperature getSkyTemperaturePointSource(Bins nuBins, Bins sinThetaBins, Bins phiBins, int nu, int mu, int iota, double tPointSource) {
		SkyTemperaturePointSource retval = new SkyTemperaturePointSource(nuBins, sinThetaBins, phiBins, nu, mu, iota, tPointSource);
		return retval;
	}
}
