package gov.fnal.eag.sim.crt;

import flanagan.complex.Complex;
import gov.fnal.eag.util.misc.Constants;
import gov.fnal.eag.wcs.SphericalCoordinate;
/**
 * 
 * @author stoughto
 *
 */
public class FFTPointing {
	FFTCRT fftcrt;
	SkyTemperature skyTemperature;
	UniversalTime universalTime;

	int nx,ny,nt;
	int nf;
	int lMin, lMax;
	double fMin;
	double dx,dy,dt,thetaSqrt;
	Bins sinThetaBins, phiBins, nuBins;
	/**
	 * @param fftcrt the telescope
	 * @param skyTemperature foreground+signal
	 * @param universalTime when to make the observation
	 * @param sinThetaBins 
	 * @param phiBins 
	 * @param nuBins 
	 */
	public FFTPointing(FFTCRT fftcrt, SkyTemperature skyTemperature, UniversalTime universalTime, Bins sinThetaBins, Bins phiBins, Bins nuBins) {
		this.fftcrt = fftcrt;
		this.skyTemperature = skyTemperature;
		this.universalTime = universalTime;
		this.sinThetaBins = sinThetaBins;
		this.phiBins = phiBins;
		this.nuBins = nuBins;
		init();
	}
	private void init() {
		// number of receivers in each direction
		nx = fftcrt.getNx();
		ny = fftcrt.getNy();
		// number of digitization bins
		nt = fftcrt.getNt();
		if (nt%2 != 0) {
			throw new IllegalArgumentException("nt needs to be even but nt="+nt);
		}
		nf = nt/2;
		fMin = fftcrt.getFMin();
		dt = fftcrt.getDt();

		// assume that nt is an even number.
		lMin = (int) (nt*dt*fMin+1);
		lMax = lMin + nt/2;
		//System.out.println("nt="+nt+" dt="+dt+" fMin="+fMin+" lMin="+lMin+" lMax="+lMax);
		dx = fftcrt.getDx();
		dy = fftcrt.getDy();
		//thetaSqrt = Math.sqrt(dSinTheta * dPhi);		
		thetaSqrt = Math.sqrt(sinThetaBins.getDelta() * phiBins.getDelta());		
		//sqrtDeltaf = Math.sqrt(fftcrt.getDeltaF());
	}
	/**
	 * @param x index to the x bin
	 * @param y index to the y bin
	 * @param l index to the l bin
	 * @return the mean power received in this bin
	 */
	public double calculateMeanPower(int x, int y, int l) {

		//double omegaL = 2*Math.PI*(nuMin + (l+0.5)*(nuMax-nuMin)/nl);
		if (l < lMin || l > lMax) {
			throw new IllegalArgumentException("bad l value:  l="+l+" nt="+nt);
		}
		double nuL = l/(nt*dt);
		double plxy = 0.0;
		
		double timeBinningFactor = (double) nuBins.getNBin()/ (double) nf;
		int nNuBins = nuBins.getNBin();
		for (int iNu=1; iNu<nNuBins; iNu++) {	
			double nu = nuBins.minValue(iNu);
			double omegaNu = 2*Math.PI*nu;
			Complex zlxyOmegaNu = getZlxyOmegaNu(x, y, nuL, nu);

			NoiseTemperature[][] noiseTemperatures = fftcrt.getNoiseTemperatures();
			double tNoise = noiseTemperatures[x][y].getT(omegaNu);
			double etaxyomega = Math.sqrt(tNoise);
			Complex blxyOmegaNu = zlxyOmegaNu.times(etaxyomega);
			double term = Complex.squareAbs(blxyOmegaNu)/timeBinningFactor;
			plxy += term;
			if (Double.isNaN(plxy)) {
				throw new Error();
			}
			
			for (int iMu=1; iMu<sinThetaBins.getNBin(); iMu++) {
				double sinTheta = sinThetaBins.minValue(iMu);
				double theta = Math.asin(sinTheta);
				for (int iIota=1; iIota<phiBins.getNBin(); iIota++) {
					double phi = phiBins.minValue(iIota);
					Complex alxyk = getAlxyk(x, y, nu, zlxyOmegaNu, theta, phi);
					double termPlxy = Complex.squareAbs(alxyk);
					plxy += termPlxy;
					if (Double.isNaN(plxy)) {
						throw new Error();
					}
				}
			}
		}
		return plxy;
	}
	
	/**
	 * @param x
	 * @param y
	 * @param nuL frequency in Hz
	 * @param nuNu frequency in Hz
	 * @return the parameter of equation 23
	 */
	public Complex getZlxyOmegaNu(int x, int y, double nuL, double nuNu) {
		Gain gain = fftcrt.getGains()[x][y];
		Complex gxyOmegaNu = gain.gain(nuNu);
		Complex numerator = (new Complex(1,0)).minus(Complex.expPlusJayArg(nt*2*Math.PI*(nuNu-nuL)*dt));
		Complex denominator = (new Complex(1,0)).minus(Complex.expPlusJayArg(2*Math.PI*(nuNu-nuL)*dt));		
		Complex fraction;
		double delta = 2*Math.PI*Math.abs(nuNu-nuL)*dt;
		if ( delta < 1e-10) {
			fraction = new Complex(nt,0);
		} else {
			fraction = numerator.over(denominator);
		}
		Complex zlxyOmegaNu = gxyOmegaNu.times(fraction).times(1.0/nt);
		return zlxyOmegaNu;
	}
	/**
	 * @param x index in the x direction
	 * @param y index in the y direction
	 * @param nu frequency in Hz
	 * @param zlxyOmegaNu result of method getZlxyOmegaNu(x, y, omegaNu, omegaL)
	 * @param theta in radians
	 * @param phi in radians
	 * @return the result of equation 11, where X is defined in equation 26 and sigma in equation 9
	 */
	public Complex getAlxyk(int x, int y, double nu, Complex zlxyOmegaNu,
			double theta, double phi) {
		KVector k = new KVector(theta, phi, nu);
		Complex axy = fftcrt.getAntennaResponses()[x][y].response(k);
		Complex eixkxdx = Complex.expPlusJayArg(x*k.getX()*dx);
		Complex eiykydy = Complex.expPlusJayArg(y*k.getY()*dy);								
		Complex xlxyk = axy.times(eixkxdx).times(eiykydy).times(zlxyOmegaNu).times(thetaSqrt);

		SphericalCoordinate telescopeCoord = new SphericalCoordinate(Math.toDegrees(phi), Math.toDegrees(theta));
		SphericalCoordinate sc = fftcrt.getRaDec(telescopeCoord, universalTime);
		double t = skyTemperature.getT(sc, nu);
		double lambda = 1e-10*Constants.c/nu; // c in Angstroms/second, and we want lambda in meters
		double sigma = Math.sqrt(t)/lambda;
		Complex alxyk = xlxyk.times(sigma);
		return alxyk;
	}
	
	/**
	 * @param m
	 * @param n
	 * @param x
	 * @param y
	 * @return the Ymnxy of eq. 30;
	 */
	public Complex getYmnxy(int m, int n, int x, int y) {
		Complex retval = Complex.expMinusJayArg(2.0*Math.PI*( (y*n/(double)ny) + (x*m/(double)nx)));
		return retval;
	}
	
	/**
	 * @param x0
	 * @param x1
	 * @return 1 if x0==x1, 0 otherwise
	 */
	public int kronekerDelta(int x0, int x1) {
		if (x0 == x1) {
			return 1;
		}
		return 0;
	}

	
	/**
	 * @param l
	 * @param lp
	 * @param x
	 * @param y
	 * @param xp
	 * @param yp
	 * @return the value in equation 17
	 */
	
	public Complex getAllpxyxpyp(int l, int lp, int x, int y, int xp, int yp) {
		Complex retval = new Complex(0,0);
		if (l < lMin || l > lMax) {
			throw new IllegalArgumentException("bad l value:  l="+l+" lMin="+lMin+" lMax="+lMax);
		}
		double totalTime = nt*dt;
		double nuL = l/totalTime;
		if (lp < lMin || lp > lMax) {
			throw new IllegalArgumentException("bad l value:  l="+l+" lMin="+lMin+" lMax="+lMax);
		}
		double nuLp = lp/totalTime;
		for (int iNu=0; iNu<nuBins.getNBin(); iNu++) {
			double nu = nuBins.value(iNu);
			Complex zlxyOmegaNu  = getZlxyOmegaNu(x, y, nuL,  nu);
			Complex zlxyOmegaNuP = getZlxyOmegaNu(x, y, nuLp, nu);
			for (int iMu=0; iMu<sinThetaBins.getNBin(); iMu++) {
				double sinTheta = sinThetaBins.value(iMu);
				double theta = Math.asin(sinTheta);
				for (int iIota=0; iIota < phiBins.getNBin(); iIota++) {
					double phi = phiBins.value(iIota);
					Complex alxyk    = getAlxyk( x,  y, nu, zlxyOmegaNu,  theta, phi);
					Complex alpxpypk = getAlxyk(xp, yp, nu, zlxyOmegaNuP, theta, phi);
					Complex term = alxyk.times(alpxpypk.conjugate());
					/*
					if (term.abs() > 0.0) {
						System.out.println("getAllpxyxpyp:  iNu="+iNu+" iMu="+iMu+" iIota="+iIota+" term="+term);
					}
					*/
					retval = retval.plus(term);
				}
			}			
		}
		return retval;
	}
	
	
	/**
	 * @param l
	 * @param m
	 * @param n
	 * @return the parameter of equation 16, the mean pixel strength in a pixel
	 */
	
	
	public Complex getHlmn(int l, int m, int n) {
		Complex retval = new Complex(0.0, 0.0);
		Complex[][] ymn = new Complex[nx][ny];
		for (int x=0; x<nx; x++) {
			for (int y=0; y<ny; y++) {
				ymn[x][y] = getYmnxy(m,n,x,y);
				//System.out.println("m="+m+" n="+n+" x="+x+" y="+y+" ymnxy="+ymn[x][y]);
			}
		}
		for (int x=0; x<nx; x++) {
			for (int y=0; y<ny; y++) {
				for (int xp=0; xp<nx; xp++) {
					for (int yp=0; yp<ny; yp++) {
						Complex term1 = ymn[x][y].times(ymn[xp][yp].conjugate()) ;
						Complex term2 = new Complex(kronekerDelta(x,xp)*kronekerDelta(y,yp), 0);
						//double term3 = getAllpxyxpyp(l, l, x, y, xp, yp).squareAbs();
						Complex term3 = getAllpxyxpyp(l, l, x, y, xp, yp);
						Complex term = (term1.minus(term2)).times(term3);
						//if (term.getReal() < 0.0) {
							//throw new Error();
						//}
						//System.out.println("x="+x+" y="+y+" xp="+xp+" yp="+yp+" term="+term+" term3="+term3);
						retval = retval.plus(term);
					}
				}
			}
		}
		return retval;
	}
		
	/**
	 * @return the minimum l value
	 */
	public int getLMin() {
		return lMin;		
	}
	/**
	 * @return the maximum l value
	 */
	public int getLMax() {
		return lMax;		
	}
}
