package gov.fnal.eag.sim.crt;

import gov.fnal.eag.sim.repository.image.ImageRepository;
import gov.fnal.eag.util.misc.StringUtil;
import gov.fnal.eag.wcs.SphericalCoordinate;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;

/**
 * implements the Global Sky Model of Oliviera-Costa
 * @author stoughto
 *
 */
public class GlobalSkyModel implements SkyTemperature {
	String gsmRoot = System.getProperty("user.home")+File.separator+"ImageRepository.gsm";
	ImageRepository ir = new ImageRepository(gsmRoot);		

	String defaultComponentMap="408locked";
	static final int nComp = 3;
	double[] x;
	double[][] y;
	int n;
	double[][] ypp;

	double[] f = new double[nComp+1]; // computeComponents(double nu)
	double nu = -1;

	ArrayList<double[]> componentMap = null;
	/**
	 * uses the default component map
	 * @throws IOException 
	 */
	public GlobalSkyModel() throws IOException  {
		init(defaultComponentMap);
	}
	private void init(String componentMapName) throws IOException {
		loadComponents();
		loadComponentMap(componentMapName);
	}
	private void loadComponentMap(String componentMapName) throws IOException {
		System.out.println("GlobalSkyModel.loadComponentMap begin");
		componentMap = new ArrayList<double[]>();
		String fullFileName = ir.getFullFileName("component_maps_"+componentMapName+".dat");
		InputStream inputStream = ImageRepository.getInputStream(fullFileName);
		InputStreamReader inputStreamReader = new InputStreamReader(inputStream);
		BufferedReader bufferedReader = new BufferedReader(inputStreamReader);
		String line;
		while ( (line = bufferedReader.readLine()) != null) {
			double[] a = StringUtil.toDoubleArray(line, " ");
			componentMap.add(a);
		}
		System.out.println("GlobalSkyModel.loadComponentMap size="+componentMap.size());
	}
	private void loadComponents() throws IOException  {
		// load components.dat
		String fullFileName = ir.getFullFileName("components.dat");
		InputStream inputStream = ImageRepository.getInputStream(fullFileName);
		InputStreamReader inputStreamReader = new InputStreamReader(inputStream);
		BufferedReader bufferedReader = new BufferedReader(inputStreamReader);
		bufferedReader.mark(10000);
		String line;
		n = 0;
		while ( (line = bufferedReader.readLine()) != null) {
			n++;
		}
		bufferedReader.reset();
		x = new double[n];
		y = new double[n][nComp+1];
		int i = 0;
		while ( (line = bufferedReader.readLine()) != null) {
			double[] temp = StringUtil.toDoubleArray(line, " ");
			x[i] = Math.log(temp[0]);
			for (int iComp=0; iComp<nComp; iComp++) {
				y[i][iComp] = temp[iComp+2];
				//System.out.println("i="+i+" iComp="+iComp+" y="+y[i][iComp]);
			}
			y[i][nComp] = Math.log(temp[1]);
			i++;
		}
		// now do the work of SUBROUTINE myspline_r8
		ypp = new double[n][nComp+1];
		for (int iComp=0; iComp<=nComp; iComp++) {
			ypp[0][iComp] = 0.0;
			double[] u = new double[n];
			u[0] = 0.0;
			for (i=1; i<n-1; i++) {
				double sig = (x[i]-x[i-1])/(x[i+1]-x[i-1]);
				double p = sig*ypp[i-1][iComp]+2.0;
				ypp[i][iComp] = (sig-1.0)/p;
				u[i] = (6.0*((y[i+1][iComp]-y[i][iComp])/(x[i+1]-x[i])-(y[i][iComp]-y[i-1][iComp])/(x[i]-x[i-1]))/(x[i+1]-x[i-1])-sig*u[i-1])/p;
			}
			double qn = 0.0;
			double un = 0.0;
			ypp[n-1][iComp] = (un-qn*u[n-2])/(qn*ypp[n-2][iComp]+1.0);
			for (int k=n-2; k>=0; k--) {
				ypp[k][iComp] = ypp[k][iComp]*ypp[k+1][iComp]+u[k];
			}
		}
		/*
		for (int in=0; in<ypp.length; in++) {
			for (int iComp=0; iComp<ypp[in].length; iComp++) {
				System.out.println("in="+in+" iComp="+iComp+" ypp[in][iComp]="+ypp[in][iComp]);
			}
		}
		 */
	}
	private void computeComponents(double nuIn) {
		if (this.nu == nuIn) return;
		if (nuIn <= 0.0) {
			throw new IllegalArgumentException("nu must be positive by nuIn="+nuIn);
		}
		this.nu = nuIn;
		double logNu = Math.log(nu);
		for (int iComp=0; iComp<=nComp; iComp++) {
			int kLo = 0;
			int kHi = n-1;
			int nIterations = 0;
			while (kHi-kLo > 1) {
				int k = (kHi+kLo)/2;
				if (x[k] > logNu) {
					kHi = k;
				} else {
					kLo = k;
				}
				nIterations++;
			}
			double h = x[kHi] - x[kLo];
			if ((logNu-x[0])*(logNu-x[n-1]) > 0.0) {
				double aa = (y[kHi][iComp]-y[kLo][iComp])/h;
				f[iComp] = y[kLo][iComp] + aa * (logNu-x[kLo]);
			} else {
				double aa = (x[kHi]-logNu)/h;
				double bb = (logNu-x[kLo])/h;
				f[iComp] = aa*y[kLo][iComp] + bb*y[kHi][iComp] + ((Math.pow(aa,3)-aa)*ypp[kLo][iComp]+(Math.pow(bb,3)-bb)*ypp[kHi][iComp])*Math.pow(h,2)/6.0;
			}
		}
		f[nComp] =  Math.exp(f[nComp]);
	}
	private double getT(double[] comps, double nu1) {
		computeComponents(nu1);
		double retval = 0.0;
		for (int iComp=0; iComp<nComp; iComp++) {
			retval += comps[iComp]*f[iComp];
		}
		retval *= f[nComp];
		return retval;
	}
	/**
	 * @param eq ra,dec of the location
	 * @param nu1 frequency, in MHz
	 * @return T in degrees K
	 */
	public double getT(SphericalCoordinate eq, double nu1) {
		int nSide = 512;
		computeComponents(nu1);
		SphericalCoordinate gal = eq.copy();
		gal.equatorialToGalactic();
		long iPix = gal.getIpix_ring(nSide);
		double[] comps = componentMap.get((int) iPix);
		double retval = getT(comps, nu1);
		return retval;
	}
}
