/**
 * 
 */
package gov.fnal.eag.sim.extractor;

import cern.colt.list.DoubleArrayList;
import cern.jet.stat.Descriptive;
import gov.fnal.eag.sim.pixel.PixeledImage;
import gov.fnal.eag.util.densityEstimate.Histogram;

/**
 * handles background calculation for a mesh (sub-PixeledImage) of pixels
 * @author stoughto
 *
 */

public class Backmesh2 {
	PixeledImage image, weight;
	int meshRowSize, meshColumnSize, nMeshRow, nMeshColumn;
	boolean[][] isGoods;
	double[][] means;
	/**
	 * Set boolean to false to do calculations in floating point precision: otherwise calculations done in double precision
	 */
	public static boolean doDouble = true;
	// magic copied from SExtractor back.c
	static final int QUANTIF_NSIGMA = 5;
	static final int QUANTIF_AMIN = 4;
	static final int QUANTIF_NMAXLEVELS = 4096;
	static final double step = Math.sqrt(2.0/Math.PI)*QUANTIF_NSIGMA/QUANTIF_AMIN;
	static final double EPS = 1e-4;

	/**
	 * constructor with an image but no weight 
	 * @param image
	 */
	public Backmesh2(PixeledImage image) {
		this(image, null);
	}
	/**
	 * constructor with image and weight
	 * @param image
	 * @param weight
	 */
	public Backmesh2(PixeledImage image, PixeledImage weight) {
		this(image, weight, image.getRowCount(), image.getColumnCount());
	}
	/**
	 * @param image
	 * @param weight
	 * @param meshRowSize
	 * @param meshColumnSize
	 */
	public Backmesh2(PixeledImage image, PixeledImage weight, int meshRowSize ,int meshColumnSize) {
		this.image = image;
		this.weight = weight;
		this.meshRowSize = meshRowSize;
		this.meshColumnSize = meshColumnSize;
		init();
	}

	/**
	 * initialize means and isGood, based on the logic in subroutine backstat, backhisto, and backGuess
	 */
	private void init() {
		nMeshRow = image.getRowCount()/meshRowSize;
		nMeshColumn = image.getColumnCount()/meshColumnSize;
		int rowGutter    = (image.getRowCount() - nMeshRow*meshRowSize)/2;
		int columnGutter = (image.getColumnCount() - nMeshColumn*meshColumnSize)/2;
		isGoods = new boolean[nMeshRow][nMeshColumn];
		means  = new double[nMeshRow][nMeshColumn];
		for (int iMeshRow=0; iMeshRow<nMeshRow; iMeshRow++) {
			int rowMin = image.getRowMin()+rowGutter+iMeshRow*meshRowSize;
			int rowMax = rowMin+meshRowSize;
			for (int iMeshColumn=0; iMeshColumn<nMeshColumn; iMeshColumn++) {
				int colMin = image.getColumnMin()+columnGutter+iMeshColumn*meshColumnSize;
				int colMax = colMin + meshColumnSize;

				double mean = 0.0;
				double sigma = 0.0;
				double weightSum = 0.0;
				int npix = 0;
				for (int row=rowMin; row<rowMax; row++) {
					for (int col=colMin; col<colMax; col++) {
						double imageValue = image.getByRowCol(row,col);
						double weightValue = 1.0;
						if (weight != null) {
							weightValue=weight.getByRowCol(row, col);
							imageValue*=weightValue;
						}
						mean      += imageValue;
						sigma     += imageValue*imageValue;
						weightSum += weightValue; 
					}
				}
				npix = meshRowSize*meshColumnSize;
				mean /= weightSum;
				sigma = Math.sqrt(sigma/weightSum - mean*mean);
				//System.out.println("mean"+mean+"sigma"+sigma);
				double lcut = mean - 2.0*sigma;
				double hcut = mean + 2.0*sigma;
				mean = 0;
				sigma = 0;
				npix = 0;
				for (int row=image.getRowMin(); row<image.getRowMax(); row++) {
					for (int col=image.getColumnMin(); col<image.getColumnMax(); col++) {
						double value = image.getByRowCol(row,col);
						if (value >= lcut && value <= hcut) {
							mean += value;
							sigma += value*value;
							npix++;
						}
					}
				}
				mean /= npix;
				sigma = Math.sqrt(sigma/npix - mean*mean);
				int nLevels = (int) (step*npix+1);
				nLevels = Math.min(QUANTIF_NMAXLEVELS, nLevels);
				double qscale;
				if (sigma > 0.0) {
					qscale = (float) (2 * QUANTIF_NSIGMA*sigma/nLevels);
				} else {
					qscale = (float) 1.0;
				}
				double qzero = (float) (mean - QUANTIF_NSIGMA*sigma);

				//System.out.println("mean="+mean+" QUANTIF_NSIGMA="+QUANTIF_NSIGMA+" sigma="+sigma+" qzero="+String.format("%30.10f", qzero));
				double hgMin = qzero - qscale/2.0;
				double hgMax = hgMin + nLevels*qscale;
				//	System.out.println("hgMax="+hgMax+" nLevels="+nLevels);
				Histogram hg = image.getHistogram(hgMin, hgMax, nLevels);
				hcut = hg.getNBin(); 
				lcut = 0;
				double sig = 10*(hg.getNBin()-1);
				double sig1 = 1.0;
				double mea = mean;
				double med = mean;
				int iBinLowCut = 0;
				int iBinHighCut = hg.getNBin()-1;
				for (int n=99; (sig>=0.1) && Math.abs(sig/sig1-1.0)>EPS && n >= 0; n--){
					sig1 = sig;
					long sum = 0;
					sig = 0.0;
					mea = 0.0;
					long lowsum = 0;
					long highsum = 0;
					int iBinLowIndex = iBinLowCut;
					int iBinHighIndex = iBinHighCut;
					for (int iBin=iBinLowCut; iBin<=iBinHighCut; iBin++) {
						if (lowsum < highsum) {
							double contents = hg.getContents(iBinLowIndex++);
							lowsum += contents;
						} else {
							double contents = hg.getContents(iBinHighIndex--);
							highsum += contents;
						}
						double contents = hg.getContents(iBin);
						sum += contents;
						mea += contents*iBin;
						sig += contents*iBin*iBin;
					}
					double temp1;
					if (iBinHighIndex >= 0.0) {
						if (hg.getContents(iBinLowIndex) > hg.getContents(iBinHighIndex)) {
							temp1 = hg.getContents(iBinLowIndex);
						} else {
							temp1 = hg.getContents(iBinHighIndex);
						}
						// TODO Why is this?
						med = iBinHighIndex + 0.5 + (highsum-lowsum)/(2.0*temp1);
						//System.out.println("med "+med+" iBinHighIndex "+iBinHighIndex+" temp1 "+temp1+" highsum "+highsum+" lowsum "+lowsum);
					} else {
						med = 0.0;
					}
					mea /= sum;
					sig = sig/sum - mea*mea;
					if (sig > 0) {
						sig = Math.sqrt(sig);
					} else {
						sig = 0.0;
					}
					{
						double ftemp = med-3.0*sig;
						if (ftemp > 0) {
							iBinLowCut = (int) (ftemp+0.5);
						} else {
							iBinLowCut = 0;
						}
					}
					{
						double ftemp = med+3.0*sig;
						//System.out.println("med="+med+" sig="+sig);
						//System.out.println("ftemp= "+ftemp+" NBin<< subt 1 ="+hg.getNBin());
						if (ftemp < hg.getNBin()-1) {
							if (ftemp > 0) {
								iBinHighCut = (int) (ftemp+0.5);
							} else {
								iBinHighCut = (int) (ftemp-0.5);
							}
						} else {
							iBinHighCut = hg.getNBin()-1;
						}
					}
				}
				double temp1 = Math.abs(sig);
				// note that these were already defined
				//double qscale = hg.getBinSize();
				//double qzero = hg.getMin() + qscale/2.0;
				double temp3 = qzero+mea*qscale;
				double temp4 = Math.abs((mea-med)/sig);
				double temp5 = qzero+(2.5*med-1.5*mea)*qscale;
				double temp6 = qzero + med*qscale;
				//System.out.println("qzero= "+qzero+" med= "+med+" qscale= "+qscale);
				if (temp1>0.0) {
					if (temp4 < 0.3) {
						mean = temp5; 
					} else {
						mean = temp6;
					}
				}
				else {
					mean = temp3;
				}	
				means[iMeshRow][iMeshColumn] = mean;
				System.out.println("Backmesh2.init:  iMeshRow="+iMeshRow+" iMeshColumn="+iMeshColumn+" mean="+mean);
				isGoods[iMeshRow][iMeshColumn] = true;
			}
		}
	}
	/**
	 * @return the median of all good means
	 */
	public double getBackground() {
		DoubleArrayList dal = new DoubleArrayList();
		for (int iMeshRow=0; iMeshRow<nMeshRow; iMeshRow++) {
			for (int iMeshColumn=0; iMeshColumn<nMeshColumn; iMeshColumn++) {
				if (isGoods[iMeshRow][iMeshColumn]) 
					dal.add(means[iMeshRow][iMeshColumn]);
			}
		}
		dal.sort();
		double retval = Descriptive.median(dal);
		return retval;
	}
}
