package gov.fnal.eag.sim.extractor;

import java.awt.Point;

import flanagan.math.MinimisationFunction;
import gov.fnal.eag.sim.pixel.DetectorObjectProfile;
import gov.fnal.eag.sim.pixel.PixeledImage;
import gov.fnal.eag.sim.pixel.Profiles;

/**
 * used in calcuating elliptical adaptive moments
 * @author stoughto
 *
 */
public class EllipseMinimisationFunction implements MinimisationFunction{

	private static Profiles profiles = new Profiles();
	private double xPosition, yPosition;
	private DetectorObjectProfile dopWindow;
	private int nPixel;
	private double x2we, y2we, xywe;
	private SourceObject sourceObject;
	/**
	 * @param sigmaWindow sigma of the window function (a, radius in pixels)
	 * @param ellipWindow ellipticity of the window function (1 - b/a)
	 * @param thetaWindow theta in degrees
	 * @param sourceObject list of pixels included in the object
	 */
	public EllipseMinimisationFunction(double sigmaWindow, double ellipWindow, double thetaWindow, SourceObject sourceObject) {
		xPosition = sourceObject.getXWin();
		yPosition = sourceObject.getYWin();
		dopWindow = new DetectorObjectProfile(yPosition, xPosition, 1.0, "gauss", sigmaWindow, ellipWindow, thetaWindow, profiles);
		this.sourceObject = sourceObject;
		// calculate sums weighted by the window function		
		double sumxx = 0;
		double sumxy = 0;
		double sumyy = 0;
		double sumwv = 0;
		PixeledImage pi = sourceObject.getPixeledImage();
		for (Point point : sourceObject.getList()) {
			int row = point.y;
			int col = point.x;
			double w = dopWindow.getFluxByPix(row, col, 1);
			double v = pi.getByRowCol(row, col);
			double x = col + 0.5 - xPosition;
			double y = row + 0.5 - yPosition;
			sumwv += w*v;
			sumxx += w*v*x*x;
			sumxy += w*v*x*y;
			sumyy += w*v*y*y;
		}
		x2we = sumxx/sumwv;
		y2we = sumyy/sumwv;
		xywe = sumxy/sumwv;
		nPixel = sourceObject.getNPixels();
	}

	/**
	 * @param params
	 * @return the value of the function
	 */
	public double function(double[] params) {
		double sigmaTrue = params[0];
		double ellipTrue = params[1];
		double thetaTrue = params[2];
		DetectorObjectProfile dopExpected = new DetectorObjectProfile(yPosition, xPosition, 1.0, "gauss", sigmaTrue, ellipTrue, thetaTrue, profiles);
		double aRadius = Math.sqrt(nPixel/Math.PI/(1+ellipTrue));
		double bRadius = aRadius*(1+ellipTrue);
		int rowMin = (int) Math.floor(yPosition - aRadius);
		int rowMax = (int) Math.floor(yPosition + aRadius);
		int colMin = (int) Math.floor(xPosition - aRadius);
		int colMax = (int) Math.floor(xPosition + aRadius);
		double sumew = 0;
		double sumxx = 0;
		double sumxy = 0;
		double sumyy = 0;
		double a2 = aRadius*aRadius;
		double b2 = bRadius*bRadius;
		for (int row=rowMin; row<rowMax; row++) {
			for (int col=colMin; col<colMax; col++) {
				if (sourceObject.isInEllipse(col, row, xPosition, yPosition, a2, b2, thetaTrue)) {
					double w = dopWindow.getFluxByPix(row, col, 1);
					double e = dopExpected.getFluxByPix(row, col, 1);
					double x = col + 0.5 - xPosition;
					double y = row + 0.5 - yPosition;
					sumew += w*e;
					sumxx += w*e*x*x;
					sumxy += w*e*x*y;
					sumyy += w*e*y*y;
					
				}
			}
		}
		double x2wee = sumxx/sumew;
		double y2wee = sumyy/sumew;
		double xywee = sumxy/sumew;
		return Math.pow(x2wee-x2we, 2) + Math.pow(y2wee-y2we, 2) + Math.pow(xywee-xywe, 2);
	}
}
