package gov.fnal.eag.sim.extractor;

import gov.fnal.eag.sim.pixel.DetectorObjectProfile;
import gov.fnal.eag.sim.pixel.PixeledImage;
import gov.fnal.eag.sim.pixel.Profiles;
import gov.fnal.eag.sim.repository.image.ImageRepository;

import org.freehep.math.minuit.FCNBase;

/**
 * Utility class for finding elliptical values. Function is used with 
 * JMinuit to find the minimum and true values for x2, y2, and xy for
 * elliptically windowed objects. 
 * @author dcharms
 *
 */
public class EllipseMinuit implements FCNBase{

private PixeledImage win;
	
	private double x2M;
	
	private double y2M;
	
	private double xyM;
	
	private double row;
	
	private double col;
	
	/**
	 * If true, values will be printed and PixeledImages will be written 
	 * to fits files.
	 */
	public static boolean debug = false;
	
	/**
	 * @param row row value of center
	 * @param col col value of center
	 * @param eW ellipticity of windowing
	 * @param sW scale of windowing
	 * @param tW theta of windowing
	 * @param x2M the measured x2
	 * @param y2M the measured y2
	 * @param xyM the measured xy
	 */
	public EllipseMinuit(double row, double col, double eW, double sW, double tW, double x2M, double y2M, double xyM) {
		DetectorObjectProfile dop = new DetectorObjectProfile(row, col, 1, "gauss", sW, eW, tW, new Profiles());
		win = dop.makePixeledImage((int)(sW*30));
		if (debug) {
			System.out.println(" row="+row+" col="+col+" eW="+eW+" sW="+sW+" tW="+tW);
			ImageRepository.write("EllipseMinFunction.fit", win);
			System.out.println(" row="+row+" col="+col+" row0="+win.getRow0()+" col0="+win.getCol0());
			SourceObject temp = new SourceObject(win);
			System.out.println("temp x="+temp.getX()+" ellip="+temp.getEllipticity()+" a="+Math.sqrt(temp.getA())+" theta="+temp.getTheta());
		}
		this.row = row;
		this.col = col;
		this.x2M = x2M;
		this.y2M = y2M;
		this.xyM = xyM;
	}
	
	/**
	 * @see org.freehep.math.minuit.FCNBase#valueOf(double[])
	 */
	public double valueOf(double[] param) {
		double x2 = param[0];
		double y2 = param[1];
		double xy = param[2];
		double a = Math.sqrt((x2+y2)/2+Math.sqrt(Math.pow((x2-y2)/2, 2)+Math.pow(xy, 2)));
		double b = Math.sqrt((x2+y2)/2-Math.sqrt(Math.pow((x2-y2)/2, 2)+Math.pow(xy, 2)));
		double eT = 1-b/a;
		double sT = a;
		double tT =0.5*Math.atan2(2*xy, (x2-y2))*180/Math.PI;
		//System.out.println("x2="+param[0]+" y2="+param[1]+" xy="+param[2]);
		//System.out.println("e="+eT+" s="+sT+" t="+tT);
		DetectorObjectProfile dop = new DetectorObjectProfile(row, col, 1, "gauss", sT, eT, tT, new Profiles());
		PixeledImage truePi = dop.makePixeledImage(win.getRowCount());
		truePi.multiply(win);
		SourceObject so = new SourceObject(truePi);
		x2 = so.getX2();
		y2 = so.getY2();
		xy = so.getXY();
		
		double deltaX2 = x2M-x2;
		double deltaY2 = y2M-y2;
		double deltaXY = xyM-xy;
		return Math.pow(deltaX2, 2)+Math.pow(deltaY2, 2)+Math.pow(deltaXY, 2);
	}

}
