package gov.fnal.eag.sim.extractor;

import gov.fnal.eag.sim.pixel.ColorTable;
import gov.fnal.eag.sim.pixel.DetectorObjectProfile;
import gov.fnal.eag.sim.pixel.PixeledImage;
import gov.fnal.eag.sim.pixel.PixeledImagePlot;
import gov.fnal.eag.sim.pixel.PixeledMask;
import gov.fnal.eag.sim.pixel.Profiles;
import gov.fnal.eag.sim.pixel.ScaleTable;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.AffineTransform;
import java.awt.geom.Area;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Point2D;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;

import org.jfree.chart.JFreeChart;
import org.jfree.chart.annotations.XYAnnotation;
import org.jfree.chart.annotations.XYLineAnnotation;
import org.jfree.chart.annotations.XYShapeAnnotation;
import org.jfree.chart.annotations.XYTextAnnotation;
import org.jfree.chart.plot.XYPlot;

import cern.colt.list.DoubleArrayList;
import cern.jet.stat.Descriptive;


/**
 * finds object in a PixeledImage and returns a SourceObject. Algorithm similar
 * SExtractor.
 * 
 * @author dcharms
 * 
 */
public class Extractor implements Iterable<SourceObject> {
	private Profiles profiles = new Profiles();
	/**
	 * names for some pre-defined kernels
	 */
	public enum KernelType {
		/**
		 * 3x3 default 
		 */
		DEFAULT3x3,
		/**
		 * 3x3 gaussian
		 */
		GAUSS3x3, 
		/**
		 * 5x5 gaussian
		 */
		GAUSS5x5
	}

	private PixeledImage pi;

	private PixeledImage kernel;

	private PixeledImage smoothed = null;

	private PixeledImage segment = null;

	private double threshold;

	private int minPixels = 4;

	private ArrayList<ArrayList<Point>> bigList = null;

	private int largest = -1;

	private static double scale = 3;

	private double sky = 0;

	private double skySigma = 0;

	private boolean skyFound = false;

	private double nSigma = 3.0;
	/**
	 * Color of bounding box. If null, box is not drawn.
	 */
	public static Color doBoundingBox = null;

	/**
	 * Color of x's on pixels. If null, x's are not drawn.
	 */
	public static Color doPixelDetection = null;

	/**
	 * Color of intersecting lines at center of object. If null, lines not drawn.
	 */
	public static Color doABAxes = null;

	/**
	 * Color of ellipse. If null, ellipse not drawn.
	 */
	public static Color doEllipse = null;

	/**
	 * Color of pixel outline. If null, outline not drawn.
	 */
	public static Color doPixelOutline = null;

	/**
	 * Color of labels. If null, labels are not shown.
	 */
	public static Color doText = null;

	/**
	 * To calculate sky and subtract it from image automatically 
	 * or not. True by default;
	 */
	public boolean doSky = true;

	private ArrayList<Hashtable<String,Number>> skyFindings = new ArrayList<Hashtable<String,Number>>();
	/**
	 * Default constructor
	 * @param pi pixeled image to be extracted
	 */
	public Extractor (PixeledImage pi) {
		this.pi = pi;
	}

	/**
	 * Get smoothed pixeled image
	 * @return the smoothed pixeled image
	 */
	public PixeledImage getSmoothed () {
		if (smoothed == null)
			doSmoothing();
		return smoothed;
	}

	/**
	 * Set the smoothing kernel
	 * @param kernel the smoothing kernel
	 */
	public void setKernel(PixeledImage kernel) {
		this.kernel = kernel;
		smoothed = null;
	}

	/**
	 * sets the kernel to the specified type, and normalizes it.
	 * @param type
	 */
	public void setKernel(KernelType type) {
		switch (type) {
		case DEFAULT3x3:
			kernel = new PixeledImage(new double[][] { {1,2,1}, {2,4,2}, {1,2,1}}); break;
		case GAUSS3x3:
			kernel = makeGaussKernel(3,3); break;
		case GAUSS5x5:
			kernel = makeGaussKernel(5,5);
		}
		kernel.multiplyScalar(1.0/kernel.getSum());	
	}

	private PixeledImage makeGaussKernel(int nRow, int nCol) {
		DetectorObjectProfile dop = new DetectorObjectProfile(nRow/2.0, nCol/2.0, 1.0, "gauss", 1.0, 0.0, 0.0, profiles);
		PixeledImage piTemp = new PixeledImage(nRow, nCol);
		dop.renderToPixeledImage(piTemp, 10);
		return piTemp;
	}
	/**
	 * Set the minimum value that will be detected
	 * @param threshold the minimum value
	 */
	public void setAbsoluteThresh(double threshold) {
		segment = null;
		this.threshold = threshold;
		this.doSky = false;
		//System.out.println("Extractor.setAbsoluteThresh:  threshold="+threshold);
	}

	/**
	 * Set the minimum number of pixels that will be considered an object
	 * @param minPixels the minimum number of pixels
	 */
	public void setMinPixels(int minPixels) {
		this.minPixels = minPixels;
	}

	/**
	 * @return the pixeled image used to construct this Extractor
	 */
	public PixeledImage getPi() {
		return pi;
	}
	/**
	 * Create and return the segment image
	 * @return the segment image
	 */

	public PixeledImage getSegment() {
		if (doSky) {
			calculateSky();
			double thresh = sky + nSigma*skySigma;
			setAbsoluteThresh(thresh);
		}
		if (segment == null) {
			segment = new PixeledImage(pi.getRowCount(), pi.getColumnCount());
			segment.setRow0Col0(pi.getRow0(), pi.getCol0());
			bigList = new ArrayList<ArrayList<Point>>();
			getSmoothed();
			for (int i=smoothed.getRowMin(); i<smoothed.getRowMax(); i++) {
				for (int j=smoothed.getColumnMin(); j<smoothed.getColumnMax(); j++) {
					double temp = smoothed.getByRowCol(i, j); 
					if (temp - threshold > 0) {
						if (segment.getByRowCol(i-1, j) > 0 || segment.getByRowCol(i, j-1) > 0) {
							if (segment.getByRowCol(i-1, j) == segment.getByRowCol(i, j-1) ) {
								int value = (int) segment.getByRowCol(i-1, j);
								bigList.get(value-1).add(new Point(j, i));
								segment.setByRowCol(i, j, value);
							}
							else if (segment.getByRowCol(i-1, j) > 0 && segment.getByRowCol(i, j-1) == 0) {
								int value = (int) segment.getByRowCol(i-1, j);
								bigList.get(value-1).add(new Point(j, i));
								segment.setByRowCol(i, j, value);
							}
							else if (segment.getByRowCol(i-1, j) == 0 && segment.getByRowCol(i, j-1) > 0) {
								int value = (int) segment.getByRowCol(i, j-1);
								bigList.get(value-1).add(new Point(j, i));
								segment.setByRowCol(i, j, value);
							}
							else {
								int valueRow = (int) segment.getByRowCol(i-1, j);
								int valueCol = (int) segment.getByRowCol(i, j-1);
								if (valueRow > valueCol) {
									segment.setByRowCol(i, j, valueCol);
									bigList.get(valueCol-1).add(new Point(j, i));
									for (Point2D pixel : bigList.get(valueRow-1)) {
										segment.setByRowCol(pixel.getY(), pixel.getX(), valueCol);
									}
									bigList.get(valueCol-1).addAll(bigList.get(valueRow-1));
									bigList.get(valueRow-1).clear();
								}
								else {
									segment.setByRowCol(i, j, valueRow);
									bigList.get(valueRow-1).add(new Point(j, i));
									for (Point2D pixel : bigList.get(valueCol-1)) {
										segment.setByRowCol(pixel.getY(), pixel.getX(), valueRow);
									}
									bigList.get(valueRow-1).addAll(bigList.get(valueCol-1));
									bigList.get(valueCol-1).clear();
								}
							}
						}
						else {
							ArrayList<Point> list = new ArrayList<Point>();
							list.add(new Point(j, i));
							bigList.add(list);
							segment.setByRowCol(i, j, bigList.size());
						}
					}
				}
			}
			cleanUp();
		}
		return segment;
	}


	/**
	 * Single out one object from the pixeled image
	 * @param object the object you want to single out
	 * @return the object
	 */
	public SourceObject getSource(int object) {
		SourceObject retval;
		if (bigList == null)
			getSegment();
		if (bigList.size() > object) {
			retval = new SourceObject(pi, bigList.get(object));
		} else {
			retval = new SourceObject(pi, new ArrayList<Point>());
		}
		return retval;
	}

	/**
	 * Get the number of objects detected
	 * @return the number of objects
	 */
	public int getSourceCount() {
		if (bigList == null) getSegment();
		return bigList.size();
	}

	/**
	 * Get the largest object detected in pixeled image
	 * @return the largest object
	 */
	public SourceObject getLargestSource() {
		int largestObject = 0;
		if (smoothed == null) 
			getSmoothed();
		SourceObject source;
		int max = 0;
		if (segment == null)
			getSegment();
		for (int i=0; i<getSourceCount(); i++) {
			source = getSource(i);
			int nPixels = source.getNPixels();
			if (nPixels > max) {
				largestObject = i;
				max = nPixels;
			}
		}
		source = getSource(largestObject);
		largest = largestObject;
		return source;
	}

	private void doSmoothing() {
		if (kernel != null) {
			smoothed = new PixeledImage(pi);
			sky = getSky();
			smoothed.convolveIj(kernel, sky);
		} else {
			smoothed = pi;
		}
	}

	private void cleanUp() {
		int emptyCount = -1;
		ArrayList<Integer> empty = new ArrayList<Integer>();
		for (int i=0; i<bigList.size(); i++) {
			int nPix = bigList.get(i).size();
			if (nPix < minPixels) {
				for (Point pixel : bigList.get(i)) {
					segment.setByRowCol(pixel.getY(), pixel.getX(), 0);
				}
				bigList.get(i).clear();
			}
		}
		for (int i=0; i<bigList.size(); i++) {
			if (!bigList.get(i).isEmpty()) {
				if (emptyCount != -1) {
					for (Point2D pixel : bigList.get(i)) {
						segment.setByRowCol(pixel.getY(), pixel.getX(), emptyCount+1);
					}
					bigList.get(empty.get(emptyCount)).addAll(bigList.get(i));
					bigList.get(i).clear();
					empty.add(i);
					emptyCount++;
				}
			}
			else {
				if (emptyCount == -1) {
					emptyCount++;
				}
				empty.add(i);
			}
		}
		for (int i=0; i<bigList.size(); i++) {
			if (bigList.get(i).isEmpty()) {
				bigList.remove(i);
				i--;
			}
		}
	}

	/**
	 * Get the largest source object's number
	 * @return largest source number
	 */
	public int getLargestObjectNum() {
		getLargestSource();
		return largest;
	}

	/**
	 * Makes a chart of a pixeled image zoomed in to the specified size
	 * @param rMin minimum row included
	 * @param cMin minimum column included
	 * @param rMax maximum row included
	 * @param cMax maximum column included
	 * @param ct color table for chart
	 * @param st scale table for chart
	 * @return the chart of scaled pixeled image
	 */
	public JFreeChart makeChart(int rMin, int cMin, int rMax, int cMax, ColorTable ct, ScaleTable st) {
		if (segment == null)
			getSegment();
		int diffR = rMax-rMin;
		int diffC = cMax-cMin;
		int width;
		int height;
		if (diffR > diffC) {
			height = diffR;
			width = (height*pi.getColumnCount()/pi.getRowCount());
		}
		else {
			width = diffC;
			height = (width*pi.getRowCount()/pi.getColumnCount());
		}
		PixeledImage pix = new PixeledImage(height, width);
		pix.setRow0Col0(rMin, cMin);
		for (int row=rMin; row<height+pix.getRow0(); row++) {
			for (int col=cMin; col<width+pix.getCol0(); col++) {
				double value = pi.getByRowCol(row, col);
				pix.setByRowCol(row, col, value);
			}
		}
		JFreeChart chart;
		PixeledImagePlot plot = new PixeledImagePlot(pix, st, ct);
		if (doPixelDetection != null) {
			PixeledMask pm = new PixeledMask(pix.getRowCount(), pix.getColumnCount());
			for (int row=0; row<pix.getRowCount(); row++) {
				for (int col=0; col<pix.getColumnCount(); col++) {
					if (segment.getByRowCol(row+pix.getRow0(), col+pix.getCol0()) > 0) 
						pm.setByRowCol(row, col, BigInteger.ONE);	
				}
			}
			pix.setMask(pm);
			chart = plot.getChart(doPixelDetection);
		}
		else
			chart = plot.getChart();
		XYPlot plotter = chart.getXYPlot();
		ArrayList<XYShapeAnnotation> shapes = new ArrayList<XYShapeAnnotation>();
		ArrayList<XYTextAnnotation> nums = new ArrayList<XYTextAnnotation>();

		ArrayList<XYLineAnnotation> lines = new ArrayList<XYLineAnnotation>();

		for (int i=0; i<bigList.size(); i++) {
			SourceObject so = getSource(i);
			double a = Math.sqrt(so.getA());
			double b = Math.sqrt(so.getB());
			Point min = getMin(i);
			Point max = getMax(i);
			int widthX = max.x-min.x;
			int heightY = max.y-min.y;

			if (doEllipse != null) {
				Ellipse2D.Double ellipse = new Ellipse2D.Double(so.getX()-a*scale, so.getY()-b*scale, 2*a*scale, 2*b*scale);
				AffineTransform affineTransform = new AffineTransform();
				affineTransform.rotate(so.getTheta()*Math.PI/180, so.getX(), so.getY());
				Area area = new Area(ellipse);
				Area rotatedEllipse = area.createTransformedArea(affineTransform);
				shapes.add(new XYShapeAnnotation(rotatedEllipse, new BasicStroke(), doEllipse));				
			}
			if (doABAxes != null){
				double theta = so.getTheta();
				double cosTheta = Math.cos(Math.toRadians(theta));
				double sinTheta = Math.sin(Math.toRadians(theta));

				double x = so.getX();
				double y = so.getY();
				double x1 = x + a * scale * cosTheta;
				double y1 = y + a * scale * sinTheta;
				double x0 = x - a * scale * cosTheta;
				double y0 = y - a * scale * sinTheta;
				lines.add(new XYLineAnnotation(x0, y0, x1, y1, new BasicStroke(), doABAxes));

				x1 = x - b * scale * sinTheta;
				y1 = y + b * scale * cosTheta;
				x0 = x + b * scale * sinTheta;
				y0 = y - b * scale * cosTheta;
				lines.add(new XYLineAnnotation(x0, y0, x1, y1, new BasicStroke(), doABAxes));

			}
			if (doBoundingBox != null) {
				Rectangle rect = new Rectangle(min.x, min.y, widthX+1, heightY+1);
				shapes.add(new XYShapeAnnotation(rect, new BasicStroke(), doBoundingBox));
			}
			if (doText != null) {
				XYTextAnnotation text = new XYTextAnnotation(""+i, so.getX(), so.getY());
				text.setFont(new Font("Arial", Font.BOLD, 16));
				text.setPaint(doText);
				int length = (int) (Math.log10(i)+1);
				int pixelW = diffR/30*length;
				int pixelH = diffR/30;
				Rectangle letter = new Rectangle((int)(so.getX()-pixelW/2), (int)(so.getY()-pixelH/2), pixelW, pixelH);
				shapes.add(new XYShapeAnnotation(letter, new BasicStroke(), Color.white, Color.white));
				nums.add(text);
			}
		}
		if (doPixelOutline != null) {			
			for (int row=segment.getRowMin(); row<segment.getRowMax(); row++) {
				for (int col=segment.getColumnMin(); col<segment.getColumnMax(); col++) {
					if (segment.getByRowCol(row, col) > 0) {
						if (segment.getByRowCol(row, col-1) == 0) {
							XYLineAnnotation line = new XYLineAnnotation(col,   row,   col,   row+1, new BasicStroke(), doPixelOutline);
							lines.add(line);
						}
						if (segment.getByRowCol(row+1, col) == 0) {
							XYLineAnnotation line = new XYLineAnnotation(col,   row+1, col+1, row+1, new BasicStroke(), doPixelOutline);
							lines.add(line);
						}
						if (segment.getByRowCol(row-1, col) == 0) {
							XYLineAnnotation line = new XYLineAnnotation(col,   row,   col+1, row  , new BasicStroke(), doPixelOutline);
							lines.add(line);
						}
						if (segment.getByRowCol(row, col+1) == 0) {
							XYLineAnnotation line = new XYLineAnnotation(col+1, row,   col+1, row+1, new BasicStroke(), doPixelOutline);
							lines.add(line);
						}
					}
				}
			}
		}
		for (XYAnnotation shape : shapes) {
			plotter.addAnnotation(shape);
		}
		for (int i=0; i<nums.size(); i++) {
			plotter.addAnnotation(nums.get(i));
		}
		for (XYAnnotation line : lines) {
			plotter.addAnnotation(line);
		}
		return chart;
	}

	/**
	 * Get a chart for an individual SourceObject
	 * @param sourceObject the object you want the image of
	 * @param ct color table
	 * @param st scale table
	 * @return the chart of the image zoomed in
	 */
	public JFreeChart makeChart(int sourceObject, ColorTable ct, ScaleTable st) {
		if (segment == null)
			getSegment();
		Point min = getMin(sourceObject);
		Point max = getMax(sourceObject);
		return makeChart(min.y, min.x, max.y, max.x, ct, st);
	}

	/**
	 * Get minimum in column and row directions and return
	 * the two as a Point.
	 * @param source the SourceObject to find min
	 * @return Point of minimums, x corresponding to row
	 * and y corresponding to column
	 */
	public Point getMin(int source) {
		int minRow = 1000;
		int minCol = 1000;
		for (Point pix : bigList.get(source)) {
			if (minRow > pix.x)
				minRow = pix.x;
			if (minCol > pix.y)
				minCol = pix.y;
		}
		return new Point(minRow, minCol);
	}

	/**
	 * Get maximum in column and row directions and return
	 * the two as a Point.
	 * @param source the SourceObject to find max
	 * @return Point of maximums, x corresponding to row
	 * and y corresponding to column
	 */
	public Point getMax(int source) {
		int maxRow = -1000;
		int maxCol = -1000;
		for (Point pix : bigList.get(source)) {
			if (maxRow < pix.x)
				maxRow = pix.x;
			if (maxCol < pix.y)
				maxCol = pix.y;
		}
		return new Point(maxRow, maxCol);
	}


	private void calculateSky() {
		if (skyFound) return;
		DoubleArrayList pixels = pi.getDoubleArrayList();
		double mean = Descriptive.mean(pixels);
		double median = Descriptive.median(pixels);
		double sigma = Math.sqrt(Descriptive.sampleVariance(pixels, mean));
		double newMean = mean;
		double newMedian = median;
		double newSigma = sigma;
		double change = 0;
		boolean done = false;
		double oldSize = 0;
		addToSkyFindings(mean, median, sigma, pixels.size());
		do {
			DoubleArrayList temp = new DoubleArrayList();
			for (int i=0; i<pixels.size(); i++) {
				double value = pixels.get(i);
				if (newMedian-3*newSigma <= value && value <= newMedian+3*newSigma)
					temp.add(value);
			}
			newMedian = Descriptive.median(temp);
			newMean = Descriptive.mean(temp);
			newSigma = Math.sqrt(Descriptive.sampleVariance(temp, newMean));
			addToSkyFindings(newMean, newMedian, newSigma, temp.size());
			if (temp.size() == oldSize) {
				done = true;
			} else {
				oldSize = temp.size();
			}
		} while (!done);
		if (oldSize < 2) {
			throw new IllegalArgumentException("can not find sky:  oldSize="+oldSize+" newMean="+newMean+" newMedian="+newMedian);
		}
		change = Math.abs(sigma-newSigma)/sigma;
		if (change > 0.20) {
			sky = 2.5*newMedian-1.5*newMean;
		} else { 
			sky = newMean;
		}
		skySigma = newSigma;
		skyFound = true;

		// pi.addScalar(-sky);

		return;
	}

	private void addToSkyFindings(double mean, double median, double sigma, int size) {
		Hashtable<String,Number> skyFinding = new Hashtable<String,Number>();
		skyFinding.put("mean", mean);
		skyFinding.put("median", median);
		skyFinding.put("sigma", sigma);
		skyFinding.put("size", size);
		skyFindings.add(skyFinding);
	}

	/**
	 * sets the sky and skySigma values
	 * @param skyIn
	 * @param skySigmaIn
	 */
	public void setSky(double skyIn, double skySigmaIn) {
		sky = skyIn;
		skySigma = skySigmaIn;
		skyFound = true;
	}
	/**
	 * Get the calculated sky value
	 * @return sky value
	 */
	public double getSky() {
		if (!skyFound) {
			calculateSky();
		}
		return sky;
	}
	/**
	 * Get the calculated sky value
	 * @return sky value
	 */
	public double getSkySigma() {
		if (!skyFound) {
			calculateSky();
		}
		return skySigma;
	}

	/**
	 * @see java.lang.Iterable#iterator()
	 */
	public Iterator<SourceObject> iterator() {
		return new ExtractorIterator();
	}

	private class ExtractorIterator implements Iterator<SourceObject> {
		int iNext = 0;
		public boolean hasNext() {
			boolean retval;
			if (iNext < getSourceCount()) {
				retval = true;
			} else {
				retval = false;
			}
			return retval;
		}

		public SourceObject next() {
			SourceObject retval =  getSource(iNext);
			iNext++;
			return retval;
		}

		public void remove() {
			throw new UnsupportedOperationException();
		}
	}
	/**
	 * gets number of sigma above sky threshold
	 * @return nSigma
	 */
	public double getNSigma() {
		return nSigma;
	}

	/**
	 * sets number of sigma above sky threshold
	 * @param sigma
	 */
	public void setNSigma(double sigma) {
		nSigma = sigma;
	}

	/**
	 * @return record of the mean, median, sigma, and number of pixels found in calculateSky()
	 */
	public ArrayList<Hashtable<String,Number>> getSkyFindings() {
		return skyFindings;
	}

	/*
	public double calculateSkyPatrick(int nx, int ny) {	
		DoubleArrayList means = new DoubleArrayList();
		for (int ix=0; ix<nx; ix++) {
			for (int iy=0; iy<ny; iy++) {
				int nRow = pi.getRowCount();
				int nCol = pi.getColumnCount();
				PixeledImage subPi = new PixeledImage(
						pi, nRow/ny, nCol/nx, iy*nRow/ny, ix*nCol/nx);
				DoubleArrayList dal = subPi.getDoubleArrayList();
				double mean = Descriptive.mean(dal);
				System.out.println("mean="+mean);
				means.add(mean);

			}
		}
		means.sort();
		double median = Descriptive.median(means);
		return median;
	}
	public double backstat() {
		DoubleArrayList dal = pi.getDoubleArrayList();
	    int npix = 1600;
		double mean = Descriptive.mean(dal);
		double sigma = Math.sqrt(Descriptive.sampleVariance(dal, mean));
		//   not needed -- double sig = sigma/npix - mean*mean;
		//double sigma2 = sig>0.0 ? Math.sqrt(sig):0.0;
		//double sig = Math.sqrt(Descriptive.sampleVariance(dal, mean));
		//double sigma = (sig >0.0)? Math.sqrt(sig):0.0;
		// c program's line   sigma= (sig = sigma/npix - mean*mean)>0.0? Math.sqrt(sig):0.0;
		double hcut = mean + 2.0*sigma;
		double lcut = mean - 2.0*sigma;
		//return mean;
		double sum = 0;
		int npixel = 0;

	// to make only first cut, take out all below except return mean1
		for (int i=0; i<dal.size(); i++)  {
			double value = dal.get(i);
			if (value<hcut && value>lcut) {
				sum += value;
				npixel ++;
			}
		}
		double mean1 = sum/npixel;
		double sigma1 = Math.sqrt(Descriptive.sampleVariance(dal, mean1));
		double hcut1 = mean1 + 2.0*sigma1;
		double lcut1 = mean1 - 2.0*sigma1;
		double sum1 = 0;
		int npixel1 = 0;
		for (int i=0; i<dal.size(); i++)  {
			double value = dal.get(i);
			if (value<hcut1 && value>lcut1) {
				sum1 += value;
				npixel1 ++;
			}
		}
		//return mean final;
		double mean2 = sum/npixel;

		
		
		// return mean1;
		return mean2;
	}
	*/
}




