package gov.fnal.eag.sim.extractor;

import gov.fnal.eag.shapelets.functions.BasisFunction;
import gov.fnal.eag.sim.catalog.MorphologyShapelet;
import gov.fnal.eag.sim.pixel.DetectorObject;
import gov.fnal.eag.sim.pixel.DetectorObjectProfile;
import gov.fnal.eag.sim.pixel.DetectorObjectSurfaceBrightness;
import gov.fnal.eag.sim.pixel.PixeledImage;
import gov.fnal.eag.sim.pixel.Profiles;
import gov.fnal.eag.util.astronomy.ReverseFinder;

import java.awt.Point;
import java.awt.geom.Point2D;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.ArrayList;

import org.freehep.math.minuit.FCNBase;
import org.freehep.math.minuit.FunctionMinimum;
import org.freehep.math.minuit.MnMigrad;
import org.freehep.math.minuit.MnMinimize;
import org.freehep.math.minuit.MnUserParameters;

/**
 * a list of pixels associated with a PixeledImage
 * 
 * @author dcharms
 *
 */
public class SourceObject {

	private  PixeledImage pi;

	private ArrayList<Point> points;

	private double x = Double.NaN;

	private double y = Double.NaN;

	private double x2 = Double.NaN;

	private double y2 = Double.NaN;

	private double xy = Double.NaN;

	private double x3 = Double.NaN;

	private double y3 = Double.NaN;

	private double xy2 = Double.NaN;

	private double x2y = Double.NaN;

	private double x4 = Double.NaN;

	private double xy3 = Double.NaN;

	private double x3y = Double.NaN;

	private double x2y2 = Double.NaN;

	private double y4 = Double.NaN;

	private int xMin = 0;

	private int yMin = 0;

	private int xMax = 0;

	private int yMax = 0;

	private double xWin = Double.NaN;

	private double yWin = Double.NaN;

	private double x2Win = Double.NaN;

	private double y2Win = Double.NaN;

	private double xyWin = Double.NaN;

	private double xEllip = Double.NaN;

	private double yEllip = Double.NaN;

	private double x2Ellip = Double.NaN;

	private double y2Ellip = Double.NaN;

	private double xyEllip = Double.NaN;

	private boolean windowedCalculated = false;

	private boolean ellipseCalculated = false;

	private boolean thirdCalculated = false;

	private boolean fourthCalculated = false;

	private boolean errorsCalculated = false;

	private int MAX_ITERATIONS = 5;

	private double sigmaTrueX = Double.NaN;

	private double objectFlux = Double.NaN;

	private double sigmaTrueY = Double.NaN;

	private double errX2 = Double.NaN;

	private double errY2 = Double.NaN;

	private double errXY = Double.NaN;

	private double errA2 = Double.NaN;

	private double errB2 = Double.NaN;

	private double errTheta = Double.NaN;

	private double[][] eigenvalues;

	private boolean rrgCalculated = false;
	private double rrgRadius = Double.NaN;
	private double rrgX = Double.NaN;
	private double rrgY = Double.NaN;
	private double rrgX2 = Double.NaN;
	private double rrgY2 = Double.NaN;
	private double rrgXY = Double.NaN;
	private double rrgXXXX = Double.NaN;
	private double rrgXXXY = Double.NaN;
	private double rrgXXYY = Double.NaN;
	private double rrgXYYY = Double.NaN;
	private double rrgYYYY = Double.NaN;
	private double rrgXErr = Double.NaN;
	private double rrgYErr = Double.NaN;
	private double rrgX2Err = Double.NaN;
	private double rrgY2Err = Double.NaN;
	private double rrgXYErr = Double.NaN;
	private double rrgX2Y2Cov = Double.NaN;
	private double rrgX2XYCov = Double.NaN;
	private double rrgXYY2Cov = Double.NaN;

	private double rrgMult = 1.0;
	private double rrgCutOff = 2.5;

	private static final Profiles profiles = new Profiles();
	
	private double fitSky = 0.0;
	private double fitChi2 = Double.NaN;
	/**
	 * Default constructor
	 * @param pi pixeled image containing object
	 * @param points list of pixels included in object
	 */
	public SourceObject(PixeledImage pi, ArrayList<Point> points) {
		this.pi = pi;
		this.setPoints(points);
		calculateSimpleSums();
	}

	private void setPoints(ArrayList<Point> points) {
		this.points = points;
	}

	/**
	 * Make a source object for entire image
	 * @param pi pixeled image to get source from
	 */
	public SourceObject(PixeledImage pi) {
		setPoints(new ArrayList<Point>());
		for (int row=pi.getRowMin(); row<pi.getRowMax(); row++) {
			for (int col=pi.getColumnMin(); col<pi.getColumnMax(); col++) {
				getPoints().add(new Point(col, row));
			}
		}
		this.pi = pi;
		calculateSimpleSums();
	}

	/**
	 * creates a source object using the pixels in pi that are in object number iSegment of segment
	 * @param pi
	 * @param segment
	 * @param iSegment
	 */
	public SourceObject(PixeledImage pi, PixeledImage segment, int iSegment) {
		this.pi = new PixeledImage(pi);
		setPoints(new ArrayList<Point>());
		for (int row=pi.getRow0(); row<pi.getRowMax(); row++) {
			for (int col=pi.getCol0(); col<pi.getColumnMax(); col++) {
				double maskValue = segment.getByRowCol(row,col);
				if (maskValue == iSegment) {
					getPoints().add(new Point(col, row));
				}
			}
		}
	}

	/**
	 * creates a source object using the pixel values in pi if the mask value is > threshold
	 * @param pi
	 * @param mask
	 * @param threshold 
	 */
	public SourceObject(PixeledImage pi, PixeledImage mask, double threshold) {
		this.pi = new PixeledImage(pi);
		setPoints(new ArrayList<Point>());
		for (int row=pi.getRow0(); row<pi.getRowMax(); row++) {
			for (int col=pi.getCol0(); col<pi.getColumnMax(); col++) {
				double maskValue = mask.getByRowCol(row,col);
				if (maskValue > threshold) {
					getPoints().add(new Point(col, row));
				}
			}
		}
		calculateSimpleSums();
	}
	/**
	 * creates a SourceObject of the pixels that have the center within a circular aperture
	 * @param pi pixel values
	 * @param row0 row center of aperture
	 * @param col0 col center of aperture
	 * @param radius aperture radius
	 */
	public SourceObject(PixeledImage pi, double row0, double col0, double radius) {
		double r2 = Math.pow(radius, 2);
		setPoints(new ArrayList<Point>());
		for (int row=pi.getRow0(); row<pi.getRowMax(); row++) {
			double dr2 = Math.pow(row+0.5-row0, 2);			
			if (dr2 <= r2) {
				for (int col=pi.getCol0(); col<pi.getColumnMax(); col++) {
					double dc2 = Math.pow(col+0.5-col0, 2);
					if (dr2+dc2 < r2) {
						getPoints().add(new Point(col, row));
					}
				}
			}
		}
		this.pi = pi;
		calculateSimpleSums();		
	}

	/**
	 * Compute first order moment of x
	 * @return first order moment of x
	 */
	public double getX() {
		return x;
	}

	/**
	 * Compute first order moment of y
	 * @return first order moment of y
	 */
	public double getY() {
		return y;
	}

	/**
	 * Compute second order moment of x
	 * @return second order moment of x
	 */
	public double getX2() {
		return x2;
	}

	/**
	 * Compute second order moment of y
	 * @return second order moment of y
	 */
	public double getY2() {
		return y2;
	}

	/**
	 * Compute second order moment of xy
	 * @return second order moment of xy
	 */
	public double getXY() {
		return xy;
	}

	/**
	 * Get minimum x value in object
	 * @return minimum x value
	 */
	public int getXMin() {
		return xMin;
	}

	/**
	 * @return maximum x value
	 */
	public int getXMax() {
		return xMax;
	}
	/**
	 * Get minimum y value in object
	 * @return minimum y value
	 */
	public int getYMin() {
		return yMin;
	}
	/**
	 * @return maximum y value
	 */
	public int getYMax() {
		return yMax;
	}

	/**
	 * Get third order moment of x
	 * @return third moment of x
	 */
	public double getX3() {
		if (!thirdCalculated)
			calculateThirdMoment();
		return x3;
	}

	/**
	 * Get third order moment of xy2
	 * @return third order moment of xy2
	 */
	public double getXY2() {
		if (!thirdCalculated)
			calculateThirdMoment();
		return xy2;
	}

	/**
	 * Get third order moment of x2y
	 * @return third order moment of x2y
	 */
	public double getX2Y() {
		if (!thirdCalculated)
			calculateThirdMoment();
		return x2y;
	}

	/**
	 * Get third order moment of y
	 * @return third order moment of y
	 */
	public double getY3() {
		if (!thirdCalculated)
			calculateThirdMoment();
		return y3;
	}

	/**
	 * Get fourth order moment of x
	 * @return fourth order moment of x
	 */
	public double getX4() {
		if (!fourthCalculated)
			calculateFourthMoment();
		return x4;
	}

	/**
	 * Get fourth order moment of xy3
	 * @return fourth order moment of xy3
	 */
	public double getXY3() {
		if (!fourthCalculated)
			calculateFourthMoment();
		return xy3;
	}

	/**
	 * Get fourth order moment of x3y
	 * @return fourth order moment of x3y
	 */
	public double getX3Y() {
		if (!fourthCalculated)
			calculateFourthMoment();
		return x3y;
	}

	/**
	 * Get fourth order moment of x2y2
	 * @return fourth order moment of x2y2
	 */
	public double getX2Y2() {
		if (!fourthCalculated)
			calculateFourthMoment();
		return x2y2;
	}

	/**
	 * Get fourth order moment of y
	 * @return fourth order moment of y
	 */
	public double getY4() {
		if (!fourthCalculated)
			calculateFourthMoment();
		return y4;
	}



	/**
	 * Get the number of pixels included in object
	 * @return the number of pixels
	 */
	public int getNPixels() {
		return getPoints().size();
	}

	/**
	 * Get the windowed value for first moment of x
	 * @return windowed value for first moment of x
	 */
	public double getXWin() {
		if (!windowedCalculated) 
			calculateWindowedSums();
		return xWin;
	}

	/**
	 * Get the windowed value for first moment of y
	 * @return windowed value for first moment of y
	 */
	public double getYWin() {
		if (!windowedCalculated) 
			calculateWindowedSums();
		return yWin;
	}

	/**
	 * Get the windowed value for second moment of x
	 * @return the windowed value for second moment of x
	 */
	public double getX2Win() {
		if (!windowedCalculated) 
			calculateWindowedSums();
		return x2Win;
	}

	/**
	 * Get the windowed value for second moment of y
	 * @return the windowed value for second moment of y
	 */
	public double getY2Win() {
		if (!windowedCalculated) 
			calculateWindowedSums();
		return y2Win;
	}

	/**
	 * Get windowed value for first moment of xy
	 * @return the windowed value for first moment of xy
	 */
	public double getXYWin() {
		if (!windowedCalculated) 
			calculateWindowedSums();
		return xyWin;
	}

	/**
	 * Get the semi-major axis length squared
	 * @return the semi-major axis length squared
	 */
	public double getA() {
		return calculateA(x2, y2, xy);
	}

	/**
	 * Get the semi-minor axis length squared
	 * @return the semi-minor axis length squared
	 */
	public double getB() {
		return calculateB(x2, y2, xy);
	}

	/**
	 * Get the value e1
	 * @return e1 defined as (x2-y2)/(x2+y2);
	 */
	public double getE1() {
		return calculateE1(x2, y2);		
	}

	/**
	 * Get the value e2
	 * @return e2 defined as 2*xy/(x2+y2)
	 */
	public double getE2() {
		return calculateE2(x2, y2, xy);		
	}

	/**
	 * Get the position angle
	 * @return the position angle in degrees
	 */
	public double getTheta() {
		return calculateTheta(x2, y2, xy);
	}

	/**
	 * Get the semi-major axis length squared
	 * @return the semi-major axis length squared
	 */
	public double getAWin() {
		if (!windowedCalculated) 
			calculateWindowedSums();
		return calculateA(x2Win, y2Win, xyWin);
	}

	/**
	 * Get the semi-minor axis length squared
	 * @return the semi-minor axis length squared
	 */
	public double getBWin() {
		if (!windowedCalculated) 
			calculateWindowedSums();
		return calculateB(x2Win, y2Win, xyWin);
	}

	/**
	 * Get the position angle
	 * @return the position angle in degrees
	 */
	public double getThetaWin() {
		if (!windowedCalculated) 
			calculateWindowedSums();
		return calculateTheta(x2Win, y2Win, xyWin);
	}

	/**
	 * Get the windowed value of e1
	 * @return e1 windowed
	 */
	public double getE1Win() {
		if (!windowedCalculated) 
			calculateWindowedSums();
		return calculateE1(x2Win, y2Win);
	}

	/**
	 * Get the windowed value of e2
	 * @return e2 windowed
	 */
	public double getE2Win() {
		if (!windowedCalculated) 
			calculateWindowedSums();
		return calculateE2(x2Win, y2Win, xyWin);
	}


	/**
	 * Compute elliptically windowed first order moment of x
	 * @return first order moment of x
	 */
	public double getXEllip() {
		if (!ellipseCalculated)
			calculateEllipticalSums();
		return xEllip;
	}

	/**
	 * Compute elliptically windowed first order moment of y
	 * @return first order moment of y
	 */
	public double getYEllip() {
		if (!ellipseCalculated)
			calculateEllipticalSums();
		return yEllip;
	}

	/**
	 * Compute elliptically windowed second order moment of x
	 * @return second order moment of x
	 */
	public double getX2Ellip() {
		if (!ellipseCalculated)
			calculateEllipticalSums();
		return x2Ellip;
	}

	/**
	 * Compute elliptically windowed second order moment of y
	 * @return second order moment of y
	 */
	public double getY2Ellip() {
		if (!ellipseCalculated)
			calculateEllipticalSums();
		return y2Ellip;
	}

	/**
	 * Compute elliptically windowed first order moment of xy
	 * @return first order moment of xy
	 */
	public double getXYEllip() {
		if (!ellipseCalculated)
			calculateEllipticalSums();
		return xyEllip;
	}

	/**
	 * Get the elliptically windowed semi-major axis length squared
	 * @return the semi-major axis length squared
	 */
	public double getAEllip() {
		if (!ellipseCalculated)
			calculateEllipticalSums();
		return calculateA(x2Ellip, y2Ellip, xyEllip);
	}

	/**
	 * Get the elliptically windowed semi-minor axis length squared
	 * @return the semi-minor axis length squared
	 */
	public double getBEllip() {
		if (!ellipseCalculated)
			calculateEllipticalSums();
		return calculateB(x2Ellip, y2Ellip, xyEllip);
	}

	/**
	 * Get the elliptically windowed position angle
	 * @return the position angle in degrees
	 */
	public double getThetaEllip() {
		if (!ellipseCalculated)
			calculateEllipticalSums();
		return calculateTheta(x2Ellip, y2Ellip, xyEllip);
	}

	/**
	 * Get the elliptically windowed value of e1
	 * @return elliptically windowed e1
	 */
	public double getE1Ellip() {
		if (!ellipseCalculated)
			calculateEllipticalSums();
		return calculateE1(x2Ellip, y2Ellip);
	}

	/**
	 * Get the elliptically windowed value of e2
	 * @return elliptically windowed e2
	 */
	public double getE2Ellip() {
		if (!ellipseCalculated)
			calculateEllipticalSums();
		return calculateE2(x2Ellip, y2Ellip, xyEllip);
	}

	/**
	 * Get the true sigma of the source in the x direction
	 * @return the true sigma
	 */
	public double getTrueSigmaX() {
		if (!windowedCalculated)
			calculateWindowedSums();
		return sigmaTrueX;
	}

	/**
	 * Get the true sigma of the source in the y direction
	 * @return the true sigma
	 */
	public double getTrueSigmaY() {
		if (!windowedCalculated)
			calculateWindowedSums();
		return sigmaTrueY;
	}

	/**
	 * Get the error in x
	 * @return the error in x2
	 */
	public double getErrX2() {
		if (!errorsCalculated)
			calculateErrors(1);
		return errX2;
	}


	/**
	 * Get the error in y
	 * @return the error in y2
	 */
	public double getErrY2() {
		if (!errorsCalculated)
			calculateErrors(1);
		return errY2;
	}

	/**
	 * Get the error for combination of x and y
	 * @return the error in combination of x and y
	 */
	public double getErrXY() {
		if (!errorsCalculated)
			calculateErrors(1);
		return errXY;
	}

	/**
	 * Get the a^2 value for error ellipse
	 * @return a^2
	 */
	public double getErrA2() {
		if (!errorsCalculated)
			calculateErrors(1);
		return errA2;
	}

	/**
	 * Get the b^2 value for error ellipse
	 * @return b^2
	 */
	public double getErrB2() {
		if (!errorsCalculated)
			calculateErrors(1);
		return errB2;
	}

	/**
	 * Get the theta for the error ellipse
	 * @return theta
	 */
	public double getErrTheta() {
		if (!errorsCalculated)
			calculateErrors(1);
		return errTheta;
	}

	private void calculateSimpleSums() {
		xMin = pi.getColumnMax();
		yMin = pi.getRowMax();
		xMax = pi.getColumnMin();
		yMax = pi.getRowMin();
		for (Point pixel : getPoints()) {
			xMin = Math.min(xMin, pixel.x);
			xMax = Math.max(xMax, pixel.x);
			yMin = Math.min(yMin, pixel.y);
			yMax = Math.max(yMax, pixel.y);
		}
		double vSum = 0;
		double xVSum = 0;
		double yVSum = 0;
		double xxVSum = 0;
		double xyVSum = 0;
		double yyVSum = 0;
		for (Point2D pixel : getPoints()) {
			double pointX = pixel.getX();
			double pointY = pixel.getY();
			double v = pi.getByRowCol(pointY,pointX);
			double dx = pointX-xMin;
			double dy = pointY-yMin;
			vSum += v;
			xVSum += pointX*v;
			yVSum += pointY*v;
			xxVSum += dx*dx*v;
			xyVSum += dx*dy*v;
			yyVSum += dy*dy*v;		
		}

		objectFlux = vSum;

		x = 0.5+xVSum/vSum;
		y = 0.5+yVSum/vSum;
		x2 = xxVSum/vSum - Math.pow(x-xMin-0.5,2);
		y2 = yyVSum/vSum - Math.pow(y-yMin-0.5,2);
		xy = xyVSum/vSum - (x-xMin-0.5)*(y-yMin-0.5);
	}

	private void calculateThirdMoment() {
		double x3VSum = 0;
		double y3VSum = 0;
		double xy2VSum = 0;
		double x2yVSum = 0;
		double vSum = 0;
		for (Point2D pixel : getPoints()) {
			double pointX = pixel.getX();
			double pointY = pixel.getY();
			double v = pi.getByRowCol(pointY,pointX);
			double dx = pointX-x+.5;
			double dy = pointY-y+.5;
			vSum += v;
			x3VSum += dx*dx*dx*v;
			xy2VSum += dx*dy*dy*v;
			x2yVSum += dx*dx*dy*v;
			y3VSum += dy*dy*dy*v;			
		}

		x3 = x3VSum/vSum;
		y3 = y3VSum/vSum;
		xy2 = xy2VSum/vSum;
		x2y = x2yVSum/vSum;
		thirdCalculated = true;
	}

	private void calculateFourthMoment() {
		double x4VSum = 0;
		double y4VSum = 0;
		double xy3VSum = 0;
		double x3yVSum = 0;
		double x2y2VSum = 0;
		double vSum = 0;
		for (Point2D pixel : getPoints()) {
			double pointX = pixel.getX();
			double pointY = pixel.getY();
			double v = pi.getByRowCol(pointY,pointX);
			double dx = pointX-x+.5;
			double dy = pointY-y+.5;
			vSum += v;
			x4VSum += dx*dx*dx*dx*v;
			xy3VSum += dx*dy*dy*dy*v;
			x3yVSum += dx*dx*dx*dy*v;
			x2y2VSum += dx*dx*dy*dy*v;
			y4VSum += dy*dy*dy*dy*v;			
		}

		x4 = x4VSum/vSum;
		y4 = y4VSum/vSum;
		xy3 = xy3VSum/vSum;
		x3y = x3yVSum/vSum;
		x2y2 = x2y2VSum/vSum;
		fourthCalculated = true;
	}

	/**
	 * @param x2 second moment x*x
	 * @param y2 second moment y*y
	 * @param xy second moment x*y
	 * @return major radius
	 */
	public static double calculateA(double x2, double y2, double xy) {
		double retval = Math.sqrt((x2+y2)/2+Math.sqrt(Math.pow((x2-y2)/2, 2)+Math.pow(xy, 2)));
		return retval;
	}

	/**
	 * @param x2 second moment x*x
	 * @param y2 second moment y*y
	 * @param xy second moment x*y
	 * @return minor radius
	 */
	public static double calculateB(double x2, double y2, double xy) {
		double retval = Math.sqrt((x2+y2)/2-Math.sqrt(Math.pow((x2-y2)/2, 2)+Math.pow(xy, 2)));
		return retval;
	}

	/**
	 * calculate theta in degrees
	 * @param x2 second moment x*x
	 * @param y2 second moment y*y
	 * @param xy second moment x*y
	 * @return theta
	 */
	public static double calculateTheta(double x2, double y2, double xy) {
		return 0.5*Math.atan2(2*xy, (x2-y2))*180/Math.PI;
	}

	/**
	 * @param x2 second moment x*x
	 * @param y2 second moment y*y
	 * @return e1
	 */
	public static double calculateE1(double x2, double y2) {
		return (x2-y2)/(x2+y2);
	}

	/**
	 * @param x2 second moment x*x
	 * @param y2 second moment y*y
	 * @param xy second moment x*y
	 * @return e2
	 */
	public static double calculateE2(double x2, double y2, double xy) {
		return (2*xy)/(x2+y2);
	}

	/**
	 * calculate the windowed sums
	 */
	public void calculateWindowedSums() {
		calculateWindowedSums(MAX_ITERATIONS);
	}



	/**
	 * calculate the windowed sums
	 * @param nIterationsMax
	 */
	public void calculateWindowedSums(int nIterationsMax) {		
		windowedCalculated = true; // indicate that this has already been done

		// Use the results of the calculateSimpleSums() method as a first guess
		xWin = x;
		yWin = y;
		x2Win = x2;
		y2Win = y2;

		final double radius = Math.sqrt(getPoints().size()/Math.PI);
		double sigmaWindowMeasured = Math.sqrt((x2Win+y2Win)/2.0);
		SigmaExpected seFunction = new SigmaExpected(radius);
		ReverseFinder reverseFinderSe = new ReverseFinder(seFunction);
		double sigmaWindow =  reverseFinderSe.getX(sigmaWindowMeasured);
		for (int iIteration=0; iIteration<nIterationsMax; iIteration++) {
			double sigmaWindow2 = Math.pow(sigmaWindow, 2);
			double[] windowed = calculateWindowed(getPoints(), pi, xWin, yWin,sigmaWindow2);
			xWin = windowed[0];
			yWin = windowed[1];
			x2Win = windowed[2];
			xyWin = windowed[3];
			y2Win = windowed[4];
			double sigmaMeasured2 = (x2Win+y2Win)/2.0;
			double sigmaMeasured = Math.sqrt(sigmaMeasured2);

			SigmaWindowedExpected function = new SigmaWindowedExpected(radius, sigmaWindow);
			ReverseFinder reverseFinder = new ReverseFinder(function);
			sigmaWindow = reverseFinder.getX(sigmaMeasured);
		}
		if (Double.isNaN(xWin)) {
			xWin = x;
			x2Win = x2;
			xyWin = xy;
		}
		if (Double.isNaN(yWin)) {
			yWin = y;
			y2Win = y2;
			xyWin = xy;
		}
	}

	/**
	 * calculate windowed averages for one iteration
	 * @param pixels list of pixels included
	 * @param pixeledImage values
	 * @param xCenter determined center in x-direction
	 * @param yCenter determined center in y-direction
	 * @param sigmaWindow2 window sigma^2
	 * @return an array of xWin, yWin, x2Win, xyWin, and y2Win
	 */
	public static double[] calculateWindowed(ArrayList<Point> pixels, PixeledImage pixeledImage, double xCenter, double yCenter, double sigmaWindow2) {
		double sum=0;
		double sumX=0;
		double sumY=0;
		double sumXX=0;
		double sumXY=0;
		double sumYY=0;
		for (Point pixel : pixels) {
			double value = pixeledImage.getByRowCol(pixel.y, pixel.x);
			double deltaX = pixel.x+0.5-xCenter;
			double deltaY = pixel.y+0.5-yCenter;
			double r2 = Math.pow(deltaX,2) + Math.pow(deltaY,2);
			double weight = Math.exp(-r2/2/sigmaWindow2);
			sum += value*weight;
			sumX += deltaX*value*weight;
			sumY += deltaY*value*weight;
			sumXX += deltaX*deltaX*value*weight;
			sumXY += deltaX*deltaY*value*weight;
			sumYY += deltaY*deltaY*value*weight;
		}
		double[] retval = new double[5];
		retval[0] = xCenter + sumX/sum; // xWin 
		retval[1] = yCenter + sumY/sum; // yWin 
		retval[2] = sumXX/sum;          // x2Win
		retval[3] = sumXY/sum;          // xyWin
		retval[4] = sumYY/sum;          // y2Win
		return retval;
	}

	/**
	 * Calculate all of the elliptically windowed values for the source
	 * using default number of iterations.
	 */
	public void calculateEllipticalSums() {
		calculateEllipticalSums(MAX_ITERATIONS);
	}

	/**
	 * Calculate all of the elliptically windowed values for the source
	 * @param iterations number of times to perform windowing
	 */
	public void calculateEllipticalSums(int iterations) {
		ellipseCalculated = true;
		xEllip = x;
		yEllip = y;
		x2Ellip = x2;
		y2Ellip = y2;
		xyEllip = xy;
		for (int i=0; i<iterations; i++) {
			double a = Math.sqrt(calculateA(x2Ellip, y2Ellip, xyEllip));
			double b = Math.sqrt(calculateB(x2Ellip, y2Ellip, xyEllip));
			double theta = calculateTheta(x2Ellip, y2Ellip, xyEllip);
			double ellip = 1-b/a;
			DetectorObjectProfile dop = new DetectorObjectProfile(yEllip, xEllip, 1, "gauss", a, ellip, theta, new Profiles());
			PixeledImage temp = dop.makePixeledImage(pi.getRowCount());
			for (Point pixel : getPoints()) {
				double value = pi.getByRowCol(pixel.y, pixel.x)*temp.getByRowCol(pixel.y, pixel.x);
				temp.setByRowCol(pixel.y, pixel.x, value);
			}
			SourceObject so = new SourceObject(temp);
			double x2M = so.getX2();
			double y2M = so.getY2();
			double xyM = so.getXY();
			EllipseMinuit emf = new EllipseMinuit(yEllip, xEllip, ellip, a, theta, x2M, y2M, xyM);
			MnUserParameters myParameters = new MnUserParameters();
			myParameters.add("x2", x2Ellip, 0.1);
			myParameters.add("y2", y2Ellip, 0.1);
			myParameters.add("xy", xyEllip, 0.1);
			MnMinimize migrad = new MnMinimize(emf, myParameters);
			migrad.minimize();
			x2Ellip = migrad.value(0);
			y2Ellip = migrad.value(1);
			xyEllip = migrad.value(2);
		}

	}

	/**
	 * Print out all simple sums results and windowed results
	 */
	@Override
	public String toString() {
		String line = "SourceObject:  x="+x+" y="+y+" x2="+x2+" y2="+y2+" xy="+xy+" xWin="+xWin+" yWin="+yWin+" x2Win="+x2Win+" y2Win="+y2Win;
		return line;
	}

	/**
	 * Get the list of pixels in object
	 * @return the list of pixel positions in this object
	 */
	public ArrayList<Point> getList() {
		return getPoints();
	}


	/**
	 * Determine whether given coordinate is in an ellipse of given values
	 * @param xIn x-coordinate of point in question
	 * @param yIn y-coordinate of point in question
	 * @param centerX x-coordinate for center of ellipse
	 * @param centerY y-coordinate for center of ellipse
	 * @param a2 the length of the major axis squared
	 * @param b2 the length of the minor axis squared
	 * @param thetaIn position angle of ellipse in degrees
	 * @return true if point is in ellipse
	 */
	public boolean isInEllipse(double xIn, double yIn, double centerX, double centerY, double a2, double b2, double thetaIn) {
		double r = Math.sqrt(Math.pow(xIn-centerX, 2)+Math.pow(yIn-centerY, 2));
		double phi = Math.acos((xIn-centerX)/r);
		if (y<centerX)
			phi = 2*Math.PI-phi;
		double pointX = r*Math.cos(phi-thetaIn*Math.PI/180);
		double pointY = r*Math.sin(phi-thetaIn*Math.PI/180);
		double limit = (pointX*pointX)/a2+(pointY*pointY)/b2;
		if (limit <= 1 || (yIn==centerY && xIn==centerX)) {
			return true;
		}
		return false;
	}


	/**
	 * Get the image being used
	 * @return the pixeled image of this object
	 */
	public PixeledImage getPixeledImage() {
		return pi;
	}
	/**
	 * Get the total flux of object
	 * @return flux of object as calculated in calculateSimpleSums
	 */
	public double getFlux() {
		return objectFlux;
	}
	/**
	 * 
	 * @param row - y coordinate of the object
	 * @param col - x coordinate of the object
	 * @param radius - aperture radius (in pixels)
	 * @return ArrayList of points where flux > 0
	 */
	public static ArrayList<Point> getAperture(double row, double col, double radius) {
		ArrayList<Point> res = new ArrayList<Point>();
		Point center = new Point((int)col,(int)row);
		int xmin = (int)(col - radius + 0.5);
		if (xmin <0)xmin = 0;
		int xmax = (int)(col + radius + 0.5);
		int ymin = (int)(row - radius + 0.5);
		if (ymin <0) ymin = 0;
		int ymax = (int)(row + radius + 0.5);
		for (int x= xmin; x<=xmax; x++) {
			for ( int y = ymin; y <= ymax; y++) {
				Point np = new Point(x,y);
				if (center.distance(np) <= radius) {
					res.add(np);
				}

			}
		}
		return res;
	}

	/**
	 * Convert e1 and e2 into a/b and theta
	 * @param e1 the value e1
	 * @param e2 the value e2
	 * @return a/b, theta in degrees
	 * logic aped from Erin Sheldon's findabtheta idl code
	 */
	public static Point2D.Double abThetaFromE1E2(double e1, double e2) {
		double abRatio;
		double theta;
		if (Math.sqrt(e1*e1+e2*e2) > 1) {
			throw new Error ("can not convert:  e1="+e1+" e2="+e2);        	
		}
		if (e1 == 0) {
			int temp = 0;
			if (e2 > 0) { 
				temp = 1;
			} else if (e2 < 0) {
				temp = -1;
			}
			theta = temp*45.0;
			abRatio = Math.sqrt( (1-Math.abs(e2)) / (1+Math.abs(e2)) );        
		} else if (e2 == 0) {
			if (e1 < 0) {
				theta = 90;
			} else {
				theta = 0;
			}
			abRatio = Math.sqrt( (1-Math.abs(e1)) / (1+Math.abs(e1)) );
		} else {
			theta = 0.5 * Math.atan2(e2, e1);
			double temp = e1/Math.cos(2*theta);
			if (temp > 1) {
				throw new Error ("can not convert:  e1="+e1+" e2="+e2);
			}
			abRatio = Math.sqrt( (1-temp )/(1+temp) );
			theta *= 180/Math.PI;
		}
		Point2D.Double retval = new Point2D.Double(abRatio, theta);
		return retval;
	}

	/**
	 * Convert a/b and theta values into e1 and e2
	 * @param abRatio
	 * @param theta in degrees
	 * @return
	 * logic aped from Erin Sheldon's idl code
	 */
	public static Point2D.Double e1e2FromabTheta(double abRatio, double theta) {
		double abRatio2 = Math.pow(abRatio, 2);        
		double temp = (1-abRatio2)/(1+abRatio2);
		double twoTheta = theta*Math.PI/90.0;
		double e1 = temp*Math.cos(twoTheta);
		double e2 = temp*Math.sin(twoTheta);
		Point2D.Double retval = new Point2D.Double(e1, e2);
		return retval;        
	}

	/**
	 * given the ellipse parameters, return the second moments
	 * @param a major axis radius
	 * @param b minor axis radius
	 * @param theta position angle, in degrees
	 * @return x2, y2, xy
	 */
	public static double[] momentsFromABTheta(double a, double b, double theta) {
		if (Math.abs(theta) > 90.0) {
			throw new Error("theta needs to be in [-90,90] but theta="+theta);
		}
		double tan2Theta = Math.tan(2*theta*Math.PI/180.0);
		double x2, y2, xy;
		double temp1 = 1.0/tan2Theta/tan2Theta;
		xy = 0.5*(a*a-b*b)/Math.sqrt(1.0+temp1);
		if (temp1 > Double.MAX_VALUE) {
			// theta is very close to zero
			xy = 0;
			x2 = a*a;
			y2 = b*b;
		} else {
			if (theta < 0) {
				xy *= -1;
			}
			x2 = 0.5*(a*a+b*b) + xy/tan2Theta;
			y2 = 0.5*(a*a+b*b) - xy/tan2Theta;
		}
		double[] retval = {x2, y2, xy};
		return retval;
	}

	/**
	 * Calculate the errors of the position measurements
	 * @param noise the read noise
	 */
	public void calculateErrors(double noise) {
		errorsCalculated = true;
		double gain = 1;
		double vSum = 0;
		double sigXSum = 0;
		double sigYSum = 0;
		double sigXYSum = 0;
		for (Point2D pixel : getPoints()) {
			double pointX = pixel.getX();
			double pointY = pixel.getY();
			double dx = pointX-x;
			double dy = pointY-y;
			double v = pi.getByRowCol(pointY,pointX);
			double sigma2 = Math.pow(noise, 2)+v/gain;
			vSum += v;
			sigXSum += sigma2*dx*dx;
			sigYSum += sigma2*dy*dy;
			sigXYSum += sigma2*dx*dy;
		}
		errX2 = sigXSum/Math.pow(vSum, 2);
		errY2 = sigYSum/Math.pow(vSum, 2);
		errXY = sigXYSum/Math.pow(vSum, 2);

		errA2 = (errX2+errY2)/2+Math.sqrt(Math.pow((errX2-errY2)/2, 2)+Math.pow(errXY, 2));
		errB2 = (errX2+errY2)/2-Math.sqrt(Math.pow((errX2-errY2)/2, 2)+Math.pow(errXY, 2));
		errTheta = .5*Math.atan2(2*errXY, errX2-errY2);
	}

	/**
	 * Get the error in x2
	 * @return error in x2
	 */
	public double getSigX2() {
		double val = x2*Math.sqrt((2/objectFlux)*(1-(1/objectFlux)));
		return val;
	}

	/**
	 * Get error in y2
	 * @return error in y2
	 */
	public double getSigY2() {
		double val = y2*Math.sqrt((2/objectFlux)*(1-(1/objectFlux)));
		return val;
	}

	/**
	 * @return 1-sqrt(b/a)
	 */
	public double getEllipticity() {
		return 1.0-(getB()/getA());
	}
	/**
	 * @return 1-sqrt(b/a)
	 */
	public double getEllipticityWin() {
		return 1.0-(getBWin()/getAWin());
	}

	/**
	 * @return the whisker of the object page 6 from SNAP_2005_5899_Scholl.pdf
	 */
	public double getWhisker() {
		return Math.pow(Math.pow((getX2()-getY2()),2)+4*Math.pow(getXY(),2),0.25);
	}
	/**
	 * get the shapelet representation of the object
	 * @param rowCenter starting row position, copied to the result
	 * @param colCenter starting column position, copied to the result
	 * @param scale starting scale (in pixels), copied to the result
	 * @param nMax number of orders
	 * @return the fit 
	 */
	public DetectorObjectSurfaceBrightness getDosb(double rowCenter, double colCenter, double scale, int nMax) {
		double[][] retval = new double[nMax+1][nMax+1];
		double[][] rowPrecalc = new double[nMax+1][yMax-yMin+1];
		double[][] colPrecalc = new double[nMax+1][xMax-xMin+1];
		for (int row=yMin; row<=yMax; row++) {
			double[] xTemp = BasisFunction.preIntegrate(nMax, row-rowCenter, scale);
			for (int n=0; n<=nMax; n++) {
				rowPrecalc[n][row-yMin] = xTemp[n];
			}
		}
		for (int col=xMin; col<=xMax; col++) {
			double[] yTemp = BasisFunction.preIntegrate(nMax, col-colCenter, scale);
			for (int n=0; n<=nMax; n++) {
				colPrecalc[n][col-xMin] = yTemp[n];
			}
		}
		for (int row=yMin; row<=yMax; row++) {
			for (int col=xMin; col<=xMax; col++) {
				for (int n1=0; n1<=nMax; n1++) {
					for (int n2=0; n2<=nMax-n1; n2++) {
						retval[n1][n2] += rowPrecalc[n2][row-yMin]*colPrecalc[n1][col-xMin]*pi.getByRowCol(row,col);
					}
				}
			}
		}
		DetectorObjectSurfaceBrightness dosb = new DetectorObjectSurfaceBrightness(rowCenter, colCenter, getFlux(), retval, scale);
		return dosb;
	}

	/**
	 * @param nMax
	 * @param location
	 * @param scale
	 * @param flux
	 * @param redshift
	 * @return the fit morphology
	 */
	public MorphologyShapelet fitMorphology(final int nMax, final Point2D location, double scale, double flux, double redshift) {
		FCNBase myFunction = new FCNBase() {

			// parameters are x, y, scale, flux
			public double valueOf(double[] par) {
				double dy = 1.0;
				double dx = 1.0;
				double tempScale = par[2];
				eigenvalues = new double[nMax+1][nMax+1];
				for (Point point : getPoints()) {
					double col = point.x;
					double row = point.y;
					double tempX = col + 0.5 - par[0];
					double tempY = row + 0.5 - par[1];
					double[][] V = BasisFunction.integrateAll(nMax, tempY, tempX, tempScale, dy, dx);					
					for (int ny=0; ny<=nMax; ny++) {
						for (int nx=0; nx<=nMax; nx++) {
							eigenvalues[ny][nx] += pi.getByRowCol(row, col) * V[ny][nx];
						}
					}
				}

				// Get the flux in the entire object
				double projectedFlux = 0;
				for (int ny=0; ny<=nMax; ny++) {
					for (int nx=0; nx<=nMax; nx++) {
						projectedFlux += BasisFunction.integrate(ny, nx, 0, 0, tempScale, 100*tempScale, 100*tempScale);
					}
				}

				//double flux = par[3];
				double chi2 = 0;
				for (Point point : getPoints()) {
					double col = point.x;
					double row = point.y;
					double tempX = col + 0.5 - par[0];
					double tempY = row + 0.5 - par[1];
					double value = 0;
					for (int ny=0; ny<=nMax; ny++) {
						for (int nx=0; nx<=nMax; nx++) {
							value += eigenvalues[ny][nx] * BasisFunction.evaluate(ny, nx, tempY, tempX, tempScale);
						}
					}
					double dataValue = pi.getByRowCol(row, col);
					double deltaValue =  (dataValue - value)/Math.sqrt(value);
					chi2 += Math.pow(deltaValue,2);
				}
				chi2 /= getPoints().size();
				return chi2;
			}
		};

		MnUserParameters myParameters = new MnUserParameters();
		// position
		if (location == null) {			
			myParameters.add("col", getXWin(), 0.1);
			myParameters.add("row", getYWin(), 0.1);
		} else {
			myParameters.add("col", location.getX(), 0.1);
			myParameters.fix("col");
			myParameters.add("row", location.getY(), 0.1);
			myParameters.fix("row");
		}

		// scale
		if (Double.isNaN(scale)) {
			double sigma = 0.5*(Math.sqrt(getX2Win()) + Math.sqrt(getY2Win()));
			myParameters.add("scale", sigma, sigma/10.0);
		} else {
			myParameters.add("scale", scale, scale/10.0);
			myParameters.fix("scale");
		}

		// flux
		if (Double.isNaN(flux)) {
			double nPhoton = this.getFlux();
			myParameters.add("flux", nPhoton, Math.sqrt(nPhoton));
		} else {
			myParameters.add("flux", flux, Math.sqrt(flux));
			myParameters.fix("flux");
		}

		MnMigrad migrad = new MnMigrad(myFunction, myParameters);
		migrad.minimize();
		double fitScale = migrad.value("scale");
		MorphologyShapelet retval = new MorphologyShapelet(BasisFunction.tab2Vector(eigenvalues), fitScale, redshift);
		return retval;

	}	

	// mimic logic in rrg_measure_moms.pro
	/**
	 * @param p
	 * @param skySigma
	 * @return the radius measured by the rrg algorithm
	 */
	public double getRrgRadius(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgRadius;
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return the x position measured by the rrg algorithm
	 */
	public double getRrgX(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgX;
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return the y position measured by the rrg algorithm
	 */
	public double getRrgY(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgY;
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return the x2 value measured by the rrg algorithm
	 */
	public double getRrgX2(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgX2;
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return the y2 value measured by the rrg algorithm
	 */
	public double getRrgY2(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgY2;
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return the xy value measured by the rrg algorithm
	 */
	public double getRrgXY(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgXY;
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return XXXX measured by rrg algorithm
	 */
	public double getRrgXXXX(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgXXXX;
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return XXXY measured by rrg algorithm
	 */
	public double getRrgXXXY(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgXXXY;
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return XXYY measured by rrg algorithm
	 */
	public double getRrgXXYY(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgXXYY;
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return XYYY measured by rrg algorithm
	 */
	public double getRrgXYYY(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgXYYY;
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return YYYY measured by rrg algorithm
	 */
	public double getRrgYYYY(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgYYYY;
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return XErr measured by rrg algorithm
	 */
	public double getRrgXErr(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgXErr;
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return YErr measured by rrg algorithm
	 */
	public double getRrgYErr(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgYErr;
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return X2Err
	 */
	public double getRrgX2Err(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgX2Err;
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return Y2Err
	 */
	public double getRrgY2Err(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgY2Err;
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return XYErr
	 */
	public double getRrgXYErr(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgXYErr;
	}

	/**
	 * @param p
	 * @param skySigma
	 * @return X2Y2Cov measured by rrg algorithm
	 */
	public double getRrgX2Y2Cov(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgX2Y2Cov;
	}

	/**
	 * @param p
	 * @param skySigma
	 * @return X2XYCov measured by rrg algorithm
	 */
	public double getRrgX2XYCov(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgX2XYCov;
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return XYY2Cov measured by rrg algorithm
	 */
	public double getRrgXYY2Cov(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return rrgXYY2Cov;
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return e1 calculated with the rrg algorithm
	 */
	public double getE1Rrg(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return  calculateE1(rrgX2, rrgY2);
	}
	/**
	 * @param p
	 * @param skySigma
	 * @return e2 calculated with the rrg algorithm
	 */
	public double getE2Rrg(PixeledImage p, double skySigma) {
		if (!rrgCalculated) {
			calculateRrg(p, skySigma);
		}
		return  calculateE2(rrgX2, rrgY2, rrgXY);

	}

	private void calculateRrg(PixeledImage p, double skySigma) {
		rrgCalculated = true;
		double soA = SourceObject.calculateA(x2, y2, xy);
		double soB = SourceObject.calculateB(x2, y2, xy);
		rrgRadius = rrgMult*Math.sqrt(soA*soB);
		double rrgCutRadius = rrgCutOff * rrgRadius;
		double denom = 2*Math.pow(rrgRadius, 2.0);

		// calculate centroid (rrgX and rrgY)
		rrgX = x;
		rrgY = y;
		double deltax = Double.MAX_VALUE;
		double deltay = Double.MAX_VALUE;
		int count = 1;

		int rowMin = 0;
		int rowMax = 0;
		int colMin = 0;
		int colMax = 0;
		while (Math.abs(deltax) > 0.01 && Math.abs(deltay) > 0.01 && count < 500) {
			rowMin = (int) (Math.floor(rrgY-rrgCutRadius));
			rowMax = (int) (Math.ceil(rrgY+rrgCutRadius));
			colMin = (int) (Math.floor(rrgX-rrgCutRadius));
			colMax = (int) (Math.ceil(rrgX+rrgCutRadius));
			double sumInt = 0.0;
			double xInt = 0.0;
			double yInt = 0.0;
			for (int row=rowMin; row<rowMax; row++) {
				double dy = 0.5+row-rrgY;
				for (int col=colMin; col<colMax; col++) {
					double dx = 0.5+col-rrgX;
					double distance = Math.sqrt(dx*dx + dy*dy);
					if (distance < rrgCutRadius) {
						double gf = Math.exp(-distance*distance/denom);
						double val = p.getByRowCol(row,col);
						sumInt += val*gf;
						xInt += (col+0.5)*val*gf;
						yInt += (row+0.5)*val*gf;
					}
				}
			}
			double previousRrgX = rrgX;
			double previousRrgY = rrgY;
			rrgX = xInt/sumInt;
			rrgY = yInt/sumInt;
			deltax = rrgX-previousRrgX;
			deltay = rrgY-previousRrgY;
			count++;
		}
		// calculate second and fourth moments, and error on centroid
		double sumInt = 0.0;
		double xxInt = 0.0;
		double yyInt = 0.0;
		double xyInt = 0.0;
		double sumIxxxx = 0.0;
		double sumIxxxy = 0.0;
		double sumIxxyy = 0.0;
		double sumIxyyy = 0.0;
		double sumIyyyy = 0.0;
		double sumXxdi = 0.0;
		double sumXydi = 0.0;
		double sumYydi = 0.0;
		for (int row=rowMin; row<rowMax; row++) {
			double dy = 0.5+row-rrgY;
			for (int col=colMin; col<colMax; col++) {
				double dx = 0.5+col-rrgX;
				double distance = Math.sqrt(dx*dx + dy*dy);
				if (distance < rrgCutRadius) {
					double gf = Math.exp(-distance*distance/denom);
					double val = p.getByRowCol(row,col);
					xxInt += gf*dx*dx*val;
					yyInt += gf*dy*dy*val;
					xyInt += gf*dx*dy*val;
					double iErr = Math.sqrt(skySigma*skySigma + val);
					sumXxdi += Math.pow(gf*iErr*dx, 2);
					sumYydi += Math.pow(gf*iErr*dy, 2);
					sumXydi += Math.pow(gf*iErr, 2)*dx*dy;
					sumInt += gf*val;
					sumIxxxx += gf*val*dx*dx*dx*dx;
					sumIxxxy += gf*val*dx*dx*dx*dy;
					sumIxxyy += gf*val*dx*dx*dy*dy;
					sumIxyyy += gf*val*dx*dy*dy*dy;
					sumIyyyy += gf*val*dy*dy*dy*dy;
				}
			}
		}
		rrgX2 = xxInt/sumInt;
		rrgY2 = yyInt/sumInt;
		rrgXY = xyInt/sumInt;
		rrgXXXX = sumIxxxx/sumInt;
		rrgXXXY = sumIxxxy/sumInt;
		rrgXXYY = sumIxxyy/sumInt;
		rrgXYYY = sumIxyyy/sumInt;
		rrgYYYY = sumIyyyy/sumInt;
		rrgXErr = Math.sqrt(sumXxdi)/sumInt;
		rrgYErr = Math.sqrt(sumYydi)/sumInt;		

		// calculate error and covariance on second moments
		double sumXxIxx = 0.0;
		double sumYyIyy = 0.0;
		double sumXyIxy = 0.0;
		double sumCovXxYy = 0.0;
		double sumCovXxXy = 0.0;
		double sumCovXyYy = 0.0;
		for (int row=rowMin; row<rowMax; row++) {
			double dy = 0.5+row-rrgY;
			for (int col=colMin; col<colMax; col++) {
				double dx = 0.5+col-rrgX;
				double distance = Math.sqrt(dx*dx + dy*dy);
				if (distance < rrgCutRadius) {
					double gf = Math.exp(-distance*distance/denom);
					double val = p.getByRowCol(row,col);
					double iErr = Math.sqrt(skySigma*skySigma + val);
					sumXxIxx += Math.pow(gf*iErr*(dx*dx-rrgX2), 2);
					sumYyIyy += Math.pow(gf*iErr*(dy*dy-rrgY2), 2);
					sumXyIxy += Math.pow(gf*iErr*(dx*dy-rrgXY), 2);
					sumCovXxYy += Math.pow(gf*iErr,2)*(dx*dx-rrgX2)*(dy*dy-rrgY2);
					sumCovXxXy += Math.pow(gf*iErr,2)*(dx*dx-rrgX2)*(dx*dy-rrgXY);
					sumCovXyYy += Math.pow(gf*iErr,2)*(dx*dy-rrgXY)*(dy*dy-rrgY2);
				}
			}
		}
		rrgX2Err = Math.sqrt(sumXxIxx)/sumInt;
		rrgY2Err = Math.sqrt(sumYyIyy)/sumInt;
		rrgXYErr = Math.sqrt(sumXyIxy)/sumInt;
		rrgX2Y2Cov = sumCovXxYy / sumInt / sumInt;
		rrgX2XYCov = sumCovXxXy / sumInt / sumInt;
		rrgXYY2Cov = sumCovXyYy / sumInt / sumInt;

	}

	/**
	 * @param that
	 * @return whether this equals that, including the order of points in the list
	 */
	public boolean equals(SourceObject that) {
		int n = getPoints().size();
		if (n != that.getPoints().size()) {
			return false;
		}
		for (int i=0; i<n; i++) {
			Point point = getPoints().get(i);
			Point thatPoint = that.getPoints().get(i);
			if (point.x != thatPoint.x) {
				return false;
			}
			if (point.y != thatPoint.y) {
				return false;
			}
			double value = pi.getByRowCol(point.y, point.x);
			double thatValue = that.pi.getByRowCol(point.y, point.x);
			if (value != thatValue) {
				return false;
			}
		}
		return true;
	}

	/**
	 * @return the list of points
	 */
	public ArrayList<Point> getPoints() {
		return points;
	}

	/**
	 * @param i
	 * @return the value above the fitSky of pixel that the ith point references
	 */
	public double getValue(int i) {
		Point point = points.get(i);
		double retval = pi.getByRowCol(point.y, point.x);
		retval -= fitSky;
		return retval;
	}

	/**
	 * @param profileNames
	 * @return the profiles that fit the object
	 */
	public ArrayList<DetectorObject> fitProfiles(final String[] profileNames) {
		ArrayList<DetectorObject> retval = new ArrayList<DetectorObject>();
		if (profileNames.length > 0) {
			final double row = getYWin();
			final double col = getXWin();
			final double originalFlux = getFlux();
			final double nPixels = getNPixels();
			double e1 = getE1();
			double e2 = getE2();
			// for nProfiles
			// Constrain all of the profiles to have the same ellipticity and position angle.
			// ==> 2 parameters
			// allow the scale to change for each
			// ==> nProfiles parameters
			// allow fraction of flux to vary for each, constraining the sum of the fractions to be 1.0
			// ==> nProfiles-1 parameters
			// the last parameter is the sky level
			// ==> 1 parameter
			//
			// So the number of parameters =  2 + nProfiles + nProfiles - 1 + 1= 2*(nProfiles+1)

			
			int nProfiles = profileNames.length;
			MnUserParameters upar = new MnUserParameters();
			upar.add("e1", e1, 0.01);
			upar.setLimits("e1", -0.5, 0.5);
			upar.add("e2", e2, 0.01);
			upar.setLimits("e2", -0.5, 0.5);
			for (int i=0; i<nProfiles; i++) {
				double scale = getA();
				scale /= DetectorObjectProfile.sigmaFactor.get(profileNames[i].toLowerCase());
				String name = "scale_"+i;
				upar.add(name, scale, scale*0.1);
				upar.setLowerLimit(name, 0.0);
			}
			for (int i=1; i<nProfiles; i++) {
				upar.add("fluxFraction_"+i, ((double) i)/nProfiles, 1.0/nProfiles/2);
			}
			upar.add("sky", 0.0, 1.0);
			
			FCNBase chi2Function = new FCNBase() {
				int nGridBright = 4;
				public double valueOf(double[] par) {
					double fitSkyTemp = par[par.length-1];
					double nPhoton = originalFlux - nPixels*fitSkyTemp;
					int nProfiles1 = (par.length/2)-1;
					DetectorObjectProfile[] dops = new DetectorObjectProfile[nProfiles1];
					double e1Temp = par[0];
					double e2Temp = par[1];
					Point2D.Double abTheta = SourceObject.abThetaFromE1E2(e1Temp, e2Temp);
					double ellipticity11 = 1.0-abTheta.getX();
					double positionAngle11 = abTheta.getY();
					for (int i=0; i<nProfiles1; i++) {
						double scale1 = par[2+i];
						double fluxFraction;
						if (i==0) {
							fluxFraction = 1.0;
							for (int j=1; j<nProfiles1; j++) {
								int index = 2+nProfiles1+j-1;
								fluxFraction -= par[index];
							}
						} else {
							int index = 2+nProfiles1+i-1;
							fluxFraction = par[index];
						}
						dops[i] = new DetectorObjectProfile(row, col, fluxFraction*nPhoton, profileNames[i], scale1, ellipticity11, positionAngle11, profiles);
						double newNPhoton = getNewNPhoton(dops[i]);
						dops[i].setNPhoton(newNPhoton);
					}
					double chi2 = 0.0;
					int nPointsUsed = 0;
					for (Point point : points) {
						int r = point.y;
						int c = point.x;
						double actual = pi.getByRowCol(r, c);
						if (actual > 0) {
							double fit = fitSkyTemp;
							for (int i=0; i<nProfiles1; i++) {
								fit += dops[i].getFluxByPix(r, c, nGridBright);
							}
							chi2 += Math.pow((fit-actual)/Math.sqrt(actual), 2);
							nPointsUsed++;
						}
					}
					int ndf = nPointsUsed-par.length;
					chi2 /= ndf;
					return chi2;
				}
			};
			
			MnMigrad migrad = new MnMigrad(chi2Function, upar);
			FunctionMinimum min = migrad.minimize();
			MnUserParameters parameters = min.userParameters();
			fitChi2 = min.fval();
			//MnUserParameters parameters = upar;
			double fitE1 = parameters.value("e1");
			double fitE2 = parameters.value("e2");
			fitSky = parameters.value("sky");
			Point2D.Double abTheta = SourceObject.abThetaFromE1E2(fitE1, fitE2);
			double fitEllipticity = 1.0-abTheta.getX();
			double fitPositionAngle = abTheta.getY();
			for (int i=0; i<profileNames.length; i++) {
				double fluxFraction;
				if (i==0) {
					fluxFraction = 1.0;
					for (int j=1; j<nProfiles; j++) {
						fluxFraction -= parameters.value("fluxFraction_"+j);
					}				
				} else {
					fluxFraction = parameters.value("fluxFraction_"+i);				
				}
				double nPhoton = originalFlux - nPixels*fitSky;
				double fitFlux = nPhoton*fluxFraction;
				double fitScale = parameters.value("scale_"+i);
				DetectorObjectProfile dop = new DetectorObjectProfile(row, col, fitFlux, profileNames[i], fitScale, fitEllipticity, fitPositionAngle, profiles);
				double newNPhoton = getNewNPhoton(dop);
				dop.setNPhoton(newNPhoton);
				retval.add(dop);
			}
		} else {
			// if no profiles are to be fit, then calculate a fitSky and fitChi2 from the edge pixels
			SourceObject edge = getEdgePixels();
			double edgeFlux=edge.getFlux();
			double edgeNPixels=edge.getNPixels();
			fitSky = edgeFlux/edgeNPixels;
			fitChi2 = 0.0;
			int nPoint = 0;
			for (Point point : edge.getPoints()) {
				double value = pi.getByRowCol(point.y, point.x);
				if (value > 0.0) {
					double delta = value-fitSky;
					fitChi2 += delta*delta/value;
					nPoint++;
				}
			}
			fitChi2 /= nPoint;
			fitSky = 0.0;
		}
		return retval;
	}
	
	private double getNewNPhoton(DetectorObjectProfile dop) {
		int nGridBright = 4;
		double fluxInSourceObject = 0.0;
		for (Point point : points) {
			int ir = point.y;
			int ic = point.x;
			fluxInSourceObject += dop.getFluxByPix(ir, ic, nGridBright);
		}
		double oldNPhoton = dop.getNPhoton();
		double newNPhoton = oldNPhoton*oldNPhoton/fluxInSourceObject;
		return newNPhoton;
	}
	/**
	 * @return the fitSky value calculate in the fitProfiles method
	 */
	public double getFitSky() {
		return fitSky;
	}

	/**
	 * @return the chi2 calculated in the fitProfiles method
	 */
	public double getFitChi2() {
		return fitChi2;
	}
	/**
	 * @param row
	 * @param col
	 * @return the value of the PixeledImage at row,col, ignoring sky, or Double.NaN if the pixel is not int this SourceObject
	 */
	public double getPiValue(int row, int col) {
		double retval = Double.NaN;
		for (Point point : points) {
			if (point.x == col && point.y == row) {
				retval = pi.getByRowCol(row, col);
				break;
			}
		}
		return retval;
	}

	/**
	 * @param c
	 * @param r
	 * @return the distance to the pixel closest to this column,row position
	 */
	public double getClosestDistance(int c, int r) {
		double retval = Double.MAX_VALUE;
		Point cr = new Point(c,r);
		for (Point point : points) {
			double distance = cr.distance(point);
			retval = Math.min(distance, retval);
		}
		return retval;
	}

	/**
	 * @return composed of only the pixels that are on the edge of this SourceObject
	 */
	public SourceObject getEdgePixels() {
		ArrayList<Point> edgePoints = new ArrayList<Point>();
		for (Point point : points) {
			if (getClosestDistance(point.x, point.y-1) > 0.0) {
				edgePoints.add(point);
			} else if (getClosestDistance(point.x, point.y+1) > 0.0) {
				edgePoints.add(point);				
			} else if (getClosestDistance(point.x+1, point.y) > 0.0) {
				edgePoints.add(point);				
			} else if (getClosestDistance(point.x-1, point.y) > 0.0) {
				edgePoints.add(point);				
			}
		}
		SourceObject retval = new SourceObject(pi, edgePoints);
		return retval;
	}

	/**
	 * copy the values of fitSky and fitChi2 from the dso to this SourcObject
	 * @param dso
	 */
	public void setFitValues(SourceObject dso) {
		fitSky = dso.getFitSky();
		fitChi2 = dso.getFitChi2();
	}
	
	/**
	 * writes a file that can be loaded as a region in ds9, to put one green x in each pixel of this object
	 * @param regionFileName
	 * @throws FileNotFoundException
	 */
	public void makeRegionFile(String regionFileName) {
		PrintStream ps;
		try {
			ps = new PrintStream(new File(regionFileName));
		} catch (FileNotFoundException e) {
			throw new Error(e);
		}
		
		for (Point point : points) {
			double x1 = point.x+0.5-pi.getCol0();
			double y1 = point.y+0.5-pi.getRow0();
			ps.println(String.format("line %f %f %f %f", x1, y1, x1+1, y1+1));
			ps.println(String.format("line %f %f %f %f", x1, y1+1, x1+1, y1));
		}
		
		ps.close();
	}
}
