package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.shapelets.functions.Decompose;
import gov.fnal.eag.sim.catalog.AstronomicalFilter;
import gov.fnal.eag.sim.catalog.MorphologyShapelet;
import gov.fnal.eag.sim.catalog.ValuedBin;

import java.util.Hashtable;

import cern.jet.math.Bessel;

/**
 * get shapelet representations of Airy functions
 * @author stoughto
 *
 */
public class Airy {
	private static double radToArcsec = 3600*180/Math.PI;
	
	private AstronomicalFilter filter;
	private double obscuration;
	private double aperture;
	private PixeledImage pi = null;
	int pixelDiameter = 0;
	double pixelSize = 0;
	
	/**
	 * @param lambda wavelength in angstrom
	 * @param obscuration fractional radius of the central obscuration
	 * @param aperture primary aperture diameter in meters
	 * @param x distance from the optical axis in arcseconds
	 * @return the airy function, normalized to 1.0
	 */
	public static double airyFunction(double lambda, double obscuration, double aperture, double x) {
		if (obscuration < 0 || obscuration > 1) throw new Error("obscuration must be between 0 and 1 inclusive "+obscuration);
		double dRatio = aperture / (lambda * 1e-10);
		double dimensionlessDistance = x * Math.PI * dRatio / radToArcsec;
		double diff;
		if (dimensionlessDistance < 0.01) {
			diff = 1-obscuration*obscuration;
		} else {
			diff = 2*Bessel.j1(dimensionlessDistance)/dimensionlessDistance - 2*obscuration*Bessel.j1(obscuration*dimensionlessDistance)/dimensionlessDistance;
			diff *= dRatio/radToArcsec;
		}
		double retval = (Math.PI / 4 / (1-obscuration*obscuration)) * Math.pow(diff, 2);
		return retval;
	}
	
	/**
	 * @param lambda wavelength in angstroms
	 * @param obscuration fractional radius of the central obscuration
	 * @param aperture primary aperture diameter in meters
	 */
	public Airy(double lambda, double obscuration, double aperture) {
		this(new AstronomicalFilter(lambda, 1.0), obscuration, aperture);
	}
	
	/**
	 * @param filter integrate over lambda weighted by this filter
	 * @param obscuration fractional radius of the central obscuration
	 * @param aperture primary aperture diameter in meters
	 */
	public Airy(AstronomicalFilter filter, double obscuration, double aperture) {
		this.filter = filter;
		this.obscuration = obscuration;
		this.aperture = aperture;
	}
	
	/**
	 * get a pixeled-image representation of this Airy
	 * @param pixelDiameterIn number of pixels in the row and column direction
	 * @param pixelSizeIn pixel size in arcseconds
	 * @return the PixeledImage that represents this Airy function
	 */
	public PixeledImage getPixeledImage(int pixelDiameterIn, double pixelSizeIn) {
		if (pi != null && this.pixelDiameter==pixelDiameterIn && this.pixelSize==pixelSizeIn) return pi;
		this.pixelDiameter = pixelDiameterIn;
		this.pixelSize = pixelSizeIn;
		pi = new PixeledImage(pixelDiameter, pixelDiameter);
		for (int row=0; row<pixelDiameter/2; row++) {
			double dRow2 = Math.pow(0.5+row-pixelDiameter/2,2);
			for (int col=0; col<pixelDiameter/2; col++) {
				double dCol2 = Math.pow(0.5+col-pixelDiameter/2,2);
				double arcsec = pixelSize*Math.sqrt(dRow2+dCol2);
				double value = pixelSize*pixelSize*getValue(arcsec);
				pi.setByRowCol(row, col, value);
				pi.setByRowCol(row, pixelDiameter-col-1, value);
				pi.setByRowCol(pixelDiameter-row-1,col, value);
				pi.setByRowCol(pixelDiameter-row-1,pixelDiameter-col-1, value);
			}
		}
		return pi;
	}

	
	/**
	 * evaluate the Airy function, integrated over wavelength
	 * @param arcsec distance from optical axis in arcseconds
	 * @return the normalized surface brightness per arcsec^2
	 */
	public double getValue(double arcsec) {
		double valueSum = 0;
		double weightSum = 0;
		for (ValuedBin vb : filter) {
			double lambda = vb.getWavelengthBin().meanLambda();
			double airy = Airy.airyFunction(lambda, obscuration, aperture, arcsec);
			double weight = vb.getValue();
			valueSum += airy*weight;
			weightSum += weight;
		}
		double value = valueSum/weightSum;
		return value;
	}
	/**
	 * @return the morphology for this Airy function, with reasonable defaults
	 */
	public MorphologyShapelet getMorphology() {
		return getMorphology(15, 1.6);
	}
	/**
	 * get a Shapelet representation
	 * @param nMaxAiry maximum order for shapelet decomposition
	 * @param scaleFactor deconstruct with scale=scaleFactor*rowSigma
	 * @return the Shapelet that represents this Airy function, integrated over wavelength
	 * 
	 */
	public MorphologyShapelet getMorphology(int nMaxAiry, double scaleFactor) {
		// first make the correct pi for this
		int pixelDiameterIn = 200;
		double pixelSizeIn = 0.001;
		getPixeledImage(pixelDiameterIn, pixelSizeIn);
		double row0 = pixelDiameter/2;
		double col0 = pixelDiameter/2;
		Hashtable<String,Number> ht = pi.centroid(0.0);
		double rowSigma = ((Double) ht.get("rowSigma")).doubleValue();
		double scale = scaleFactor*rowSigma;
		//System.out.println(" in getMorphology:  scale="+scale+" pixels");
		//int nGrid = 3;
		Decompose decompose = new Decompose(pi.getPix(), row0, col0, scale);
		double[] eigenValues;
		eigenValues = decompose.getFnnV(nMaxAiry);
		// This fudge factor chosen to the sigma of the original and the sigma of the reconstruction are with 1e-2 of each other
		//double fudgeFactor=0.7348;
		double fudgeFactor = 1.0;
		double size = scale*pixelSize*fudgeFactor; // in arcseconds
		//System.out.println(" in getMorphology:  size="+size+" arcsec");
		double redshift = 0.0;
		MorphologyShapelet morphology = new MorphologyShapelet(eigenValues, size, redshift);
		return morphology;
	}
}
