package gov.fnal.eag.sim.pixel;

import java.io.File;
import java.io.InputStream;
import java.util.ArrayList;

import gov.fnal.eag.sim.repository.image.ImageRepository;
import gov.fnal.eag.util.densityEstimate.Polynomial1D;
import gov.fnal.eag.wcs.Coordinate;
import gov.fnal.eag.wcs.TransformLinear;

/**
 * creates background light based on pupil image defined by a file, assumed to be azimuthally symmetric
 * @author stoughto
 *
 */
public class BackgroundLightPupilImageTab implements BackgroundLight {

	private Polynomial1D rFit;
	/**
	 * set this to be non-null to make a diagnostic plot
	 */
	public static String plotFileName = null;
	private TransformLinear colRowToMm;
	private TransformLinear tabColRowToMm;	
	private double sky;
	private PixeledImage piTab;
	/**
	 * @param colRowToMm transformation from col,row to mm in focal plane
	 * @param sky sky per pixel background level
	 * @param fileName where to read in tabulated pupil image 
	 * @param tabPixelSizeMm pixel size, in mm, of the tabulated pupil image (should be in fileName but it is not)
	 * @param order fit order, 25 works OK
	 */
	public BackgroundLightPupilImageTab(TransformLinear colRowToMm, double sky, String fileName, double tabPixelSizeMm, int order) {
		this.colRowToMm = colRowToMm;
		this.sky = sky;
		
		// First, see if the file exists in the current working directory
		try {
			piTab = ImageRepository.readPixeledImage(fileName);
		} catch (IllegalArgumentException e) {
			// Second, see if the file exists in this package
			InputStream inputStream;
			String baseFileName = "data" + File.separator + "backgroundLight" + File.separator + fileName;
			inputStream = BackgroundLightPupilImageTab.class.getResourceAsStream(baseFileName);
			if (inputStream == null) {
				throw new Error("can not find fileName="+fileName);
			}
			piTab = ImageRepository.readPixeledImage(inputStream, 0);
		}
		
		tabColRowToMm = new TransformLinear(piTab.getColumnCount()/2.0, piTab.getRowCount()/2.0, 0.0, 0.0, tabPixelSizeMm, 0.0);
		
		ArrayList<double[]> xys = new ArrayList<double[]>();
		double radMax = 0.0;
		for (int tabRow=piTab.getRowMin(); tabRow < piTab.getRowMax(); tabRow++) {
			double r = 0.5 + tabRow ;
			for (int tabCol=piTab.getColumnMin(); tabCol < piTab.getColumnMax(); tabCol++) {
				double c = 0.5 + tabCol;
				Coordinate coord = new Coordinate(c,r);
				tabColRowToMm.forwardInPlace(coord);
				double x = coord.getX();
				double y = coord.getY();
				double rad = Math.sqrt(x*x + y*y);
				radMax = Math.max(rad, radMax);
				double v = piTab.getByRowCol(r,c);
				xys.add(new double[] {rad,v});
			}
		}
		int nZero = 100;
		double deltaRad = 0.1*radMax/nZero;
		for (int i=0; i<nZero; i++) {
			double rad = radMax+deltaRad*i;
			xys.add(new double[] {rad, 0.0});
		}
		rFit = new Polynomial1D(xys, order, plotFileName, plotFileName);
	}
	
	/**
	 * @see gov.fnal.eag.sim.pixel.BackgroundLight#getLightInPixel(double, double)
	 */
	public double getLightInPixel(double row, double col) {
		Coordinate coordinate = new Coordinate(col, row);
		colRowToMm.forwardInPlace(coordinate);
		double rho = Math.sqrt(Math.pow(coordinate.getX(), 2) + Math.pow(coordinate.getY(), 2));
		double pupil = rFit.valueAt(rho);
		double retval = sky*(1.0 + pupil);
		return retval;
	}
	/**
	 * @param row
	 * @param col
	 * @return the sky+pupil image in this pixel
	 */
	public double getLightInPixelDebug(double row, double col) {
		Coordinate coordinate = new Coordinate(col, row);
		colRowToMm.forwardInPlace(coordinate);
		double rho = Math.sqrt(Math.pow(coordinate.getX(), 2) + Math.pow(coordinate.getY(), 2));
		double pupil = rFit.valueAt(rho);
		double retval = sky*(1.0 + pupil);
		System.out.println("getLightInPixelDebug:  row="+row+" col="+col+" coordinate="+coordinate+" rho="+rho+" retval="+retval);
		return retval;
	}
	
	/**
	 * @param row
	 * @param col
	 * @return the sky+pupil image in this pixel by looking up the corresponding value in the tabulated pupil image
	 */
	public double getLightInPixelTab(double row, double col) {
		Coordinate coordinate = new Coordinate(col, row);
		colRowToMm.forwardInPlace(coordinate);
		tabColRowToMm.inverseInPlace(coordinate);
		double tabCol = coordinate.getX();
		double tabRow = coordinate.getY();
		double pupil = piTab.getByRowCol(tabRow, tabCol);
		double retval = sky*(1.0 + pupil);
		return retval;		
	}
}
