package gov.fnal.eag.sim.pixel;

import java.awt.geom.Point2D;

import nom.tam.fits.HeaderCardException;
import gov.fnal.eag.sim.random.SimRandomUniform;
import gov.fnal.eag.sim.repository.image.FitsDataFormat;
import gov.fnal.eag.sim.repository.image.ImageRepository;

/**
 * implements the CCD digitization strategy by using three DigitizedLookupTables
 * @author stoughto
 *
 */
public class CCDDigitizeLookupTable extends LookupTable {
	private DigitizeLookupTable[] dlts = new DigitizeLookupTable[3];		
	private double[] limit;
	static int[] indexToMask = {0, 2, 3};
	static int[] maskToIndex = {0, -9999, 1, 2};
	private int[] offset = new int[3];
	
	/**
	 * default read noise values for the three ranges
	 */
	public static final double[] readNoiseDefault = {3.56, 10.97, 21.09};
	/**
	 * default gain values for the three ranges
	 */
	public static final double[] gainDefault = {0.5, 8.0, 16.0};
	/**
	 * default bias values for the three ranges
	 */
	public static final double[] biasDefault = {4000.0, 7500.0, 7750.0};
	/**
	 * default upper limit for the three ranges
	 */
	public static final double[] limitDefault = {4000.0, 36000.0, 132000.0};

	/**
	 * default constructor
	 */
	public CCDDigitizeLookupTable() {
		this(readNoiseDefault, gainDefault, biasDefault, limitDefault);
	}
	
	/**
	 * replaces values with the digitized/undigitized values using the default lookup table
	 * @param pi
	 */
	public static void packUnpackStatic(PixeledImage pi) {
		CCDDigitizeLookupTable cdlt = new CCDDigitizeLookupTable();
		cdlt.packUnpack(pi);
		return;
	}
	
	/**
	 * replace the values in pi with the result of digitizing and unpacking the original values
	 * @param pi
	 */
	public void packUnpack(PixeledImage pi) {
		for (int row=pi.getRowMin(); row<pi.getRowMax(); row++) {
			for (int col=pi.getColumnMin(); col<pi.getColumnMax(); col++) {
				double value = pi.getByRowCol(row, col);
				int packed = ccdDigitize(value);
				value = ccdUnpack(packed);
				pi.setByRowCol(row, col, value);
			}
		}
	}
	/**
	 * replace values with the compressed/decompressed value using the default CCDDigitizeLookupTable
	 * @param pi
	 */
	public static void compressDecompressStatic(PixeledImage pi) {
		CCDDigitizeLookupTable cdlt = new CCDDigitizeLookupTable();
		cdlt.compressDecompress(pi);
		return;
	}
	
	/**
	 * replace values with the compressed/decompressed value 
	 * @param pi
	 */
	@Override
	public void compressDecompress(PixeledImage pi) {
		for (int row=pi.getRowMin(); row<pi.getRowMax(); row++) {
			for (int col=pi.getColumnMin(); col<pi.getColumnMax(); col++) {
				double value = pi.getByRowCol(row, col);
				int packed = ccdDigitize(value);
				int compressed = ccdCompress(packed);				
				value = ccdDecompress(compressed);
				pi.setByRowCol(row, col, value);
			}
		}
	}
	/**
	 * replace values with the compressed/decompressed values using the default values, smearing values with the sru
	 * @param pi
	 * @param sru
	 */
	public static void compressDecompressStatic(PixeledImage pi, SimRandomUniform sru) {
		CCDDigitizeLookupTable cdlt = new CCDDigitizeLookupTable();
		cdlt.compressDecompress(pi, sru);
		return;
	}
	
	/**
	 * replace values with the compressed/decompressed values, smearing values with the sru
	 * @param pi
	 * @param sru
	 */
	public void compressDecompress(PixeledImage pi, SimRandomUniform sru) {
		for (int row=pi.getRowMin(); row<pi.getRowMax(); row++) {
			for (int col=pi.getColumnMin(); col<pi.getColumnMax(); col++) {
				double value = pi.getByRowCol(row, col);
				int packed = ccdDigitize(value);
				int compressed = ccdCompress(packed);				
				value = ccdDecompress(compressed, sru);
				pi.setByRowCol(row, col, value);
			}
		}
	}
	/**
	 * fully-specified constructor
	 * @param readNoise the three read noise values, in electrons
	 * @param gain the three gains, in electrons/ADU
	 * @param bias the three bias values, in ADU
	 * @param limit used to calculate maximum value, in ADU
	 */
	public CCDDigitizeLookupTable(double[] readNoise, double[] gain, double[] bias, double[] limit) {		
		if (readNoise.length != 3) {
			throw new IllegalArgumentException("readNoise.length must be 3 but readNoise.length="+readNoise.length);
		}
		if (gain.length != 3) {
			throw new IllegalArgumentException("gain.length must be 3 but gain.length="+gain.length);
		}
		if (bias.length != 3) {
			throw new IllegalArgumentException("bias.length must be 3 but bias.length="+bias.length);
		}
		if (limit.length != 3) {
			throw new IllegalArgumentException("bias.length must be 3 but limit.length="+limit.length);
		}
		
		int[] nu = {1 << (DigitizeLookupTable.NSQBIT-2), 1 << (DigitizeLookupTable.NSQBIT-2), 1 << (DigitizeLookupTable.NSQBIT-1)};
		int[] is = {3, 3, 2};
		for (int i=0; i<3; i++) {
			int maxValue = (int) Math.floor((limit[i] + 10*readNoise[i])/gain[i] + bias[i]);
			dlts[i] = new DigitizeLookupTable(nu[i], readNoise[i], gain[i], bias[i], maxValue, is[i]);
			if (i>0) {
				offset[i] = offset[i-1]+nu[i-1];
			}
		}
		
		this.limit = limit;
	}
	
	/**
	 * @param value
	 * @return digitized version:  14 data bits plus 2  bits to tell you which range you are in
	 */
	public int ccdDigitize(double value) {
		int retval;
		int index = 2;
		for (int i=0; i<2; i++) {
			if (value < limit[i]) {
				index = i;
				break;
			}
		}
		retval = dlts[index].nirDigitize(value);
		retval = Math.max(0, retval);
		retval = Math.min(0x3FFF, retval);
		retval = (retval) | (indexToMask[index]<<14);
		return retval;
	}
	
	/**
	 * @param in
	 * @return unpack the digitized bits
	 */
	public double ccdUnpack(int in) {
		int index = getIndex(in);
		double retval;
		int base = in & 0x3FFF;
		retval = dlts[index].nirUnpack(base);
		return retval;
	}

	/**
	 * @param in the encoded value, with the mask in the most significant two bits
	 * @return 0, 1, or 2, corresponding to a mask value of 0, 2, or 3.
	 */
	public int getIndex(int in) {
		int mask = 3 & (in>>14);
		if (mask == 1) {
			throw new IllegalArgumentException("code=1 is not legal in="+in);
		}
		int index = maskToIndex[mask];
		return index;
	}
	
	/**
	 * @param value
	 * @return the index to the DigitizeLookupTable to use for this value
	 */
	public int getIndex(double value) {
		int index = 2;
		for (int i=0; i<2; i++) {
			if (value < limit[i]) {
				index = i;
				break;
			}
		}
		return index;
	}
	/**
	 * compress the bits
	 * @param in input value
	 * @return compressed bits
	 */
	public int ccdCompress(int in) {
		int index = getIndex(in);
		int base = in & 0x3FFF;
		int value =  dlts[index].nirCompress(base);
		int retval = value + offset[index];
		return retval;
	}
	
	/**
	 * uncompress to the value
	 * @param in compressed bits
	 * @return value
	 */
	public double ccdDecompress(int in) {
		int mask = getMaskFromIndex(in);
		int index = maskToIndex[mask];
		int inToNirDecompress = in-offset[index];
		double retval = dlts[index].nirDecompress(inToNirDecompress);
		return retval;
	}
	/**
	 * uncompress to the value
	 * @param in compressed bits
	 * @param sru to generate random values for smearing in the uncompressed bin
	 * @return value
	 */
	public double ccdDecompress(int in, SimRandomUniform sru) {
		int mask = getMaskFromIndex(in);
		int index = maskToIndex[mask];
		int inToNirDecompress = in-offset[index];
		double retval = dlts[index].nirDecompress(inToNirDecompress, sru);
		return retval;
	}

	/**
	 * @param in index:  0, 1, or 2
	 * @return the mask for the index:  00, 10, or 11
	 */
	public int getMaskFromIndex(int in) {
		int max = 1<<DigitizeLookupTable.NSQBIT;
		int mask;
		if (in < max/4) {
			mask = 0;
		} else if (in >= max/4 && in < max/2) {
			mask = 2;
		} else {
			mask = 3;
		}
		return mask;
	}

	/**
	 * @param iCompressed
	 * @return the expected tolerence for unpacking this value
	 */
	public double getCcdTolerence(int iCompressed) {
		int index = maskToIndex[getMaskFromIndex(iCompressed)];
		return dlts[index].getNirTolerence(iCompressed-offset[index]);
	}
	
	/**
	 * @param index
	 * @return the minimum value in the lookup table
	 */
	public int getIvMin(int index) {
		return dlts[index].getIvMin();
	}
	/**
	 * @param index
	 * @return the maximum value in the lookup table
	 */
	public int getIvMax(int index) {
		return dlts[index].getIvMax();
	}
	
	/**
	 * @param args inFileName outFileName
	 * @throws HeaderCardException
	 */
	public static void main (String[] args) throws HeaderCardException {
		if (args.length < 2) {
			System.out.println("Run SNAP CCD compression and decompression algorithm, and write output file as ushort FITS image");
			String name = CCDDigitizeLookupTable.class.getName();
			System.out.println("Usage:  "+name+" inputFileName outputFileName [exptime keyword]");
			return;
		}
		String inFileName = args[0];
		String outFileName = args[1];
		if (inFileName.equals(outFileName)) {
			System.out.println("Error:  inFileName and outFileName are the same");
			return;
		}
		System.out.println("CCDDigitizeLookupTable:  read inFileName="+inFileName);
		PixeledImage pi = ImageRepository.readPixeledImage(inFileName);

		double expTime = 1.0;
		if (args.length >=3 ) {
			String expKeyword = args[2];
			expTime = pi.header.getDoubleValue(expKeyword);	
			System.out.println("CCDDigitizeLookupTable.main:  " +
					"multiply by expTime="+expTime);
			pi.multiplyScalar(expTime);
		}

		double skyPerSecond = 0.0;
		if (args.length >= 4 ) {
			skyPerSecond = Double.parseDouble(args[3]);
			double skyValue = skyPerSecond*expTime;
			System.out.println("CCDDigitizeLookupTable.main: skyPerSecond="+skyPerSecond+" add sky value = "+skyValue);
			pi.addScalar(skyValue);
		}
		ImageRepository.write(inFileName+".scaled", pi);
		
		//double[] readNoise = {7.0, 5.0, 2.5};
		//double[] gain = {0.5, 8.0, 16.0};
		//double[] bias = {4000.0, 7500.0, 7750.0};
		//double[] limit = {4000.0, 36000.0, 132000.0};		
		//CCDDigitizeLookupTable cdlt = new CCDDigitizeLookupTable(readNoise, gain, bias, limit);
		CCDDigitizeLookupTable cdlt = new CCDDigitizeLookupTable();

		cdlt.compressDecompress(pi);
		
		pi.header.addValue("CCDVER", getVersion(), "CCDDigitizeLookupTable version");
		pi.header.addValue("SKY", skyPerSecond, "counts per second added before compress/uncompress and then subtracted");
		if (skyPerSecond != 0.0) {
			double skyValue = skyPerSecond*expTime;
			System.out.println("CCDDigitizeLookupTable.main:  subtract skyValue="+skyValue);
			pi.addScalar(-skyPerSecond*expTime);
		}
		if (expTime != 0.0) {
			double factor = 1.0/expTime;
			System.out.println("CCDDigitizeLookupTable.main:  " +
					"expTime="+expTime+" so multiply by "+factor);
			pi.multiplyScalar(factor);
		}
		System.out.println("CCDDigitizeLookupTable:  write outFileName="+outFileName);
		ImageRepository.write(outFileName, pi, FitsDataFormat.FLOAT, true);		
	}

	
	/**
	 * @return the cvs version string for this class
	 */
	public static String getVersion() {
		 String version = "CCDDigitizeLookupTable.java,v 1.5 2008/01/04 18:56:50 descvs Exp";
		 return version;
	}
	
	/**
	 * @return the lookup tables
	 */
	public DigitizeLookupTable[] getDigitizeLookupTables() {
		return dlts;
	}
	
	/**
	 * 
	 * @param originalValue
	 * @return min,max of the range of original values that yield the same compressed value
	 */
	public Point2D.Double getRange(double originalValue) {
		int packed = ccdDigitize(originalValue);
		int compressed = ccdCompress(packed);
		double unCompressed = ccdDecompress(compressed);
		double unCompressedPlus = ccdDecompress(compressed+1);
		double max = findEdge(unCompressed, unCompressedPlus);
		double unCompressedMinus = ccdDecompress(compressed-1);
		double min = findEdge(unCompressedMinus, unCompressed);
		Point2D.Double retval = new Point2D.Double(min, max);
		return retval;
	}

	private double findEdge(double v0, double v1) {
		int p0 = ccdCompress(ccdDigitize(v0));
		int p1 = ccdCompress(ccdDigitize(v1));
		if (p0 != p1-1) {
			throw new IllegalArgumentException("v0="+v0+" v1="+v1+" p0="+p0+" p1="+p1);
		}
		double vMin = v0;
		double vThis = v0;
		double vMax = v1;
		int pThis = p0;
		double delta = Double.MAX_VALUE;
		do {
			if (pThis == p0) {
				delta = (vMax-vThis)/2.0;
				vThis += delta;
				pThis = ccdCompress(ccdDigitize(vThis));
				if (pThis == p1) {
						vMax = vThis;
				}
			} else {
				delta = (vThis-vMin)/2.0;
				vThis -= delta;
				pThis = ccdCompress(ccdDigitize(vThis));
				if (pThis == p0) {
					vMin = vThis;
				}
			}
		} while (delta > 1e-12);
		return vThis;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#compress(double)
	 */
	@Override
	public int compress(double electrons) {
		return ccdCompress(ccdDigitize(electrons));
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#getNBits()
	 */
	@Override
	public int getNBits() {
		return DigitizeLookupTable.NSQBIT;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#restore(int)
	 */
	@Override
	public double restore(int compressed) {
		return ccdDecompress(compressed);
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#nValues(int)
	 */
	@Override
	public int nValues(int compressed) {
		throw new UnsupportedOperationException();
	}

}
