package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.sim.repository.image.ImageRepository;
import gov.fnal.eag.util.misc.StringUtil;
import gov.fnal.eag.util.stopwatch.StopwatchCollection;

import java.awt.Rectangle;
import java.io.File;
import java.util.ArrayList;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;

/**
 * convenience class for running ClockCharge in parallel
 * @author stoughto
 *
 */
public class ClockChargeRunner {
	String[] args;
	String[] clockChargeArgs;
	String action;
	ImageRepository irIn, irOut;
	String fnIn;
	int stride;
	boolean quantizeTraps, showStopwatch;
	/**
	 * @param args
	 * @throws JSAPException 
	 * <ul>
	 * <li> --action one of doSerial-i, doParallel-i, combineParallel, combineSerial
	 * <li> --stride number of rows or columns per chunk in each direction
	 * <li> --irIn image repository of input
	 * <li> --irOut image repository of output
	 * </ul>
	 */
	public static void main(String[] args) throws JSAPException {
		new ClockChargeRunner(args);
	}
	/**
	 * do the action specified in the args
	 * @param args
	 * @throws JSAPException
	 */
	public ClockChargeRunner(String[] args) throws JSAPException {
		this.args = args;
		makeClockChargeArgs();
		parseArgs(args);
		if (action.equals("combineParallel")) {
			combineParallel();
		} else if (action.equals("combineSerial")) {
			combineSerial();
		} else if (action.startsWith("do")) {
			doOneSegment();
		} 
		if (showStopwatch) {
			StopwatchCollection.report(System.out);
		}
	}

	private void makeClockChargeArgs() {
		ArrayList<String> al = new ArrayList<String>();
		for (int i=0; i<args.length; i+=2) {
			String arg = args[i];
			if (argIsGood(arg)) {
				al.add(arg);
				al.add(args[i+1]);
			}
		}
		clockChargeArgs = new String[al.size()];
		for (int i=0; i<al.size(); i++) {
			clockChargeArgs[i] = al.get(i);
			System.out.println("i="+i+" clockChargeArgs[i]="+clockChargeArgs[i]);
		}
	}
	private boolean argIsGood(String arg) {
		if (!arg.startsWith("--")) {
			return false;
		}
		if (arg.equals("--stride")) return false;
		if (arg.equals("--irIn")) return false;
		if (arg.equals("--irOut")) return false;
		if (arg.equals("--action")) return false;	
		if (arg.equals("--fnIn")) return false;	
		return true;
	}
	private void combineSerial() {
		String fnOut = fnIn+"-serial.fit";
		File fileOut = new File(fnOut);
		if (fileOut.exists()) {
			System.out.println("serial output already exists:  fnOut="+fnOut);
		} else {
			System.out.println("making fnOut="+fnOut);
			PixeledImage pi = irIn.getPixeledImage(fnIn);
			int nChunk = (int) Math.ceil(pi.getRowCount()/stride);
			for (int iChunk=0; iChunk<nChunk; iChunk++) {
				String tempName = fnIn+"-serial-"+iChunk;
				PixeledImage tempPi = irOut.getPixeledImage(tempName);
				for (int col=pi.getColumnMin(); col<pi.getColumnMax(); col++) {
					for (int iRow=0; iRow<stride; iRow++) {
						int row = pi.getRowMin()+iChunk*stride+iRow;
						double value = tempPi.getByRowCol(row, col);
						pi.setByRowCol(row, col, value);
					}
				}
			}
			irOut.put(fnOut, pi);
		}
	}
	private void combineParallel() {
		String fnOut = fnIn+"-parallel.fit";
		File fileOut = new File(fnOut);
		if (fileOut.exists()) {
			System.out.println("parallel output already exists:  fnOut="+fnOut);
		} else {
			System.out.println("making fnOut="+fnOut);
			
			PixeledImage pi = irIn.getPixeledImage(fnIn);
			int nChunk = (int) Math.ceil(pi.getColumnCount()/stride);
			for (int iChunk=0; iChunk<nChunk; iChunk++) {
				String tempName = fnIn+"-parallel-"+iChunk;
				System.out.println("iChunk="+iChunk+" nChunk="+nChunk);
				PixeledImage tempPi = irOut.getPixeledImage(tempName);
				for (int row=pi.getRowMin(); row<pi.getRowMax(); row++) {
					for (int iCol=0; iCol<stride; iCol++) {
						int col = pi.getColumnMin()+iChunk*stride+iCol;
						double value = tempPi.getByRowCol(row, col);
						pi.setByRowCol(row, col, value);
					}
				}
			}
			irOut.put(fnOut, pi);
		}		
	}
	private void doOneSegment() throws JSAPException {
		ClockCharge clockCharge = new ClockCharge(clockChargeArgs);
		int iChunk = getIchunk(action);
		Rectangle range;
		String fnOut;
		int x,y,width,height;
		PixeledImage piIn;
		if (action.startsWith("doSerial")) {
			String temp = fnIn+"-parallel.fit";
			System.out.println("ClockChargeRunner:  read "+temp);
			System.out.flush();
			piIn = irOut.getPixeledImage(temp);
			clockCharge.setParallelTrapsPerPixelZero();
			y = iChunk*stride;
			height = stride;
			x = 0;
			width = piIn.getColumnCount();
			range = new Rectangle(x, y, width, height);
			fnOut = fnIn+"-serial-"+iChunk;
		} else if (action.startsWith("doParallel")) {
			System.out.println("ClockChargeRunner:  read "+fnIn);
			System.out.flush();
			piIn = irIn.getPixeledImage(fnIn);
			clockCharge.setSerialTrapsPerPixelZero();
			x = iChunk*stride;
			width = stride;
			y = 0;
			height = piIn.getRowCount();
			range = new Rectangle(x, y, width, height);
			fnOut = fnIn+"-parallel-"+iChunk;
		} else {
			throw new IllegalArgumentException("can not deal with action="+action);
		}
		File fileOut = new File(fnOut);
		if (fileOut.exists()) {
			System.out.println("output file already exists:  fnOut="+fnOut);
		} else {
			System.out.println("making  fnOut="+fnOut);
			clockCharge.doClock2(piIn, range);
			System.out.println("after doClock2:  height="+height+" width="+width+" y="+y+" x="+x);
			PixeledImage piOut = new PixeledImage(piIn, height, width, y, x);
			irOut.put(fnOut, piOut);
		}
	}
	private int getIchunk(String action2) {
		String[] tokens = StringUtil.tokenize(action2, "-");
		int retval = Integer.parseInt(tokens[1]);
		return retval;
	}
	private void parseArgs(String[] args1) throws JSAPException {
		JSAP jsap = new JSAP();
		jsap.registerParameter( new FlaggedOption("action").setStringParser(JSAP.STRING_PARSER).setRequired(true).setLongFlag("action"));
		jsap.registerParameter( new FlaggedOption("stride").setStringParser(JSAP.INTEGER_PARSER).setRequired(true).setLongFlag("stride"));
		jsap.registerParameter( new FlaggedOption("irIn").setStringParser(JSAP.STRING_PARSER).setRequired(true).setLongFlag("irIn"));
		jsap.registerParameter( new FlaggedOption("irOut").setStringParser(JSAP.STRING_PARSER).setRequired(true).setLongFlag("irOut"));
		jsap.registerParameter( new FlaggedOption("fnIn").setStringParser(JSAP.STRING_PARSER).setRequired(true).setLongFlag("fnIn"));
		jsap.registerParameter( new FlaggedOption("quantizeTraps").setStringParser(JSAP.BOOLEAN_PARSER).setRequired(true).setLongFlag("quantizeTraps"));
		jsap.registerParameter( new FlaggedOption("showStopwatch").setStringParser(JSAP.BOOLEAN_PARSER).setRequired(true).setLongFlag("showStopwatch"));
		JSAPResult result = jsap.parse(args1);
		action = result.getString("action");
		stride = result.getInt("stride");
		irIn = new ImageRepository(result.getString("irIn"));
		irOut = new ImageRepository(result.getString("irOut"));
		fnIn = result.getString("fnIn");
		quantizeTraps = result.getBoolean("quantizeTraps");
		showStopwatch = result.getBoolean("showStopwatch");

	}
}
