package gov.fnal.eag.sim.pixel;

//import gov.fnal.eag.sim.catalog.SED;

import gov.fnal.eag.util.misc.ResourceList;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.Stroke;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.regex.Pattern;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.annotations.XYShapeAnnotation;
import org.jfree.chart.annotations.XYTextAnnotation;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.UnflaggedOption;

/**
 * transform an integer to a Color to display pictures of PixeledImages
 * @author stoughto
 *
 */
public class ColorTable {

	/**
	 * the number of colors defined for tables
	 */
	public static int ncolor = 256;
	private Color[] color = new Color[ncolor];

	/**
	 * get a list of available color look up tables
	 * @return a collection of names, each of which is suitable for use in the ColorTable constructor
	 */
	public static Collection<String> getNames() {
		Collection<String> retval = new ArrayList<String>();

		{
			Collection<String> resourceList = ResourceList.getResources(Pattern.compile(".*gov/fnal/eag/sim/pixel/data/.*.lut"));
			for (String resource : resourceList) {
				String[] tokens = resource.split("/");
				retval.add(tokens[tokens.length-1].replace(".lut",""));
			}
		}
		return retval;		
	}


	/**
	 * Default constructor with grey color map
	 */
	public ColorTable() {
		this("gray");
	}

	/**
	 * creates a color table 
	 * @param name which color mapping to use
	 */
	public ColorTable(String name) {
		String line;
		String fileName = "data" + File.separator + name + ".lut";
		URL inputURL = this.getClass().getResource(fileName);
		InputStream inputStream;
		try {
			if (inputURL != null) {
				inputStream = inputURL.openStream();
			} else {
				inputStream = new FileInputStream(name);
			}
			if (inputStream != null) {
				BufferedReader bufferedReader = new BufferedReader(
						new InputStreamReader(inputStream));
				// Read the file one line at a time
				line = bufferedReader.readLine();
				if (!line.equals("256,")) {
					throw new Error(" can only deal with look up tables of dimension 256");
				}
				for (int i=0; i<256; i++) {
					line = bufferedReader.readLine();
					String[] rgb = line.replaceAll(",","").split(" ");
					color[i] = new Color(
							Float.parseFloat(rgb[0]), 
							Float.parseFloat(rgb[1]), 
							Float.parseFloat(rgb[2]));
				}
			}
		} catch (IOException e) {
			throw new Error(e);
		}
	}

	/**
	 * interpolate these colors uniformly 
	 * @param colors
	 */
	public ColorTable(Color[] colors) {
		int nSection = colors.length-1;
		for (int i=0; i<ncolor; i++) {
			int iSection = i*nSection/ncolor;
			Color c0 = colors[iSection];
			Color c1 = colors[iSection+1];
			int delta = i-iSection*ncolor/nSection;
			int red0 = c0.getRed();
			int red1 = c1.getRed();
			float red = (float) ((red0 + delta*(red1-red0)/(ncolor/nSection))/255.0);
			int green0 = c0.getGreen();
			int green1 = c1.getGreen();
			float green = (float) ((green0 + delta*(green1-green0)/(ncolor/nSection))/255.0);
			int blue0 = c0.getBlue();
			int blue1 = c1.getBlue();
			float blue = (float) ((blue0 + delta*(blue1-blue0)/(ncolor/nSection))/255.0);
			//System.out.println(" rgb="+red+" "+green+" "+blue);
			color[i] = new Color(red, green, blue);
		}
	}


	/**
	 * look up a color for this index
	 * @param i the index
	 * @return the Color
	 */
	public Color get(int i) {
		return color[i];
	}

	/**
	 * invert the color table
	 */
	public void invert() {
		Color[] inverted = new Color[color.length];
		for (int i=0; i<color.length; i++) {
			inverted[i] = color[color.length-i-1];
		}
		color = inverted;		
	}
	/**
	 * make a PNG file demonstrating the available color tables.
	 * @param args outputFileName
	 * @throws Exception
	 */
	public static void main(String[] args) throws Exception {
		JSAP jsap = new JSAP();
		UnflaggedOption outputFileNameOpt = new UnflaggedOption("outputFileName")
		.setStringParser(JSAP.STRING_PARSER)
		.setDefault("colorTable.png")
		.setRequired(true)
		.setGreedy(false);		
		jsap.registerParameter(outputFileNameOpt);
		JSAPResult config = jsap.parse(args);    

		// check whether the command line was valid, and if it wasn't,
		// display usage information and exit.
		if (!config.success() ) {
			System.err.println();
			System.err.print("Usage: java "
					+ ColorTable.class.getName());
			System.err.println(" "
					+ jsap.getUsage());
			System.err.println();
			throw new IllegalArgumentException();
		}

		String outputFileName = config.getString("outputFileName");
		JFreeChart chart = getSampleChart();
		ChartUtilities.saveChartAsPNG(new File(outputFileName), chart, 500, 500);		

	}

	/**
	 * @return a JFreeChart with color table samples
	 * @throws Exception
	 */
	public static JFreeChart getSampleChart() throws Exception {
		Collection<String> names = ColorTable.getNames();
		int n = names.size();
		XYSeries xySeries = new XYSeries("");		
		final XYDataset data = new XYSeriesCollection(xySeries);
		final JFreeChart chart = ChartFactory.createScatterPlot(
				"ColorTable demo",
				"index", "table", 
				data, 
				PlotOrientation.VERTICAL,
				true, 
				true, 
				false
		);
		NumberAxis domainAxis = (NumberAxis) chart.getXYPlot().getDomainAxis();
		domainAxis.setRange(-50, 256);
		NumberAxis rangeAxis = (NumberAxis) chart.getXYPlot().getRangeAxis();
		rangeAxis.setRange(0,100*n);
		rangeAxis.setVisible(false);
		XYPlot xyPlot = chart.getXYPlot();
		Stroke stroke = new BasicStroke();
		int j=0;
		for (String name: names) {
			ColorTable colorTable = new ColorTable(name);
			for (int i=0; i<256; i+=8) {
				int yMin = j*100+5;
				int yMax = j*100+95;
				Shape shape = new Rectangle(i, yMin, i+1, yMax);
				Color color = colorTable.get(i);
				XYShapeAnnotation annotation = new XYShapeAnnotation(shape, stroke, color, color); 
				xyPlot.addAnnotation(annotation);
			}
			XYTextAnnotation textAnnotation = new XYTextAnnotation(name, -25, j*100+50);
			xyPlot.addAnnotation(textAnnotation);
			j++;
		}
		return chart;
	}
}