package gov.fnal.eag.sim.pixel;


/**
 * implements square root compression based on Gowen and Smith, Review of Scientific Instruments Vol 74 Num 8 (2003)
 * @author stoughto
 *
 */
public class CompressSqrt extends LookupTable {
	double a,b,c;
	int uncompressedMin, uncompressedMax;
	int compressedMin, compressedMax;
	double[] uncompressedLookupTable;
	int[] n;
	private void init(int uncompressedMinTemp, int uncompressedMaxTemp, int compressedMinTemp, int compressedMaxTemp) {
		this.uncompressedMin = uncompressedMinTemp;
		this.uncompressedMax = uncompressedMaxTemp;
		this.compressedMin = compressedMinTemp;
		this.compressedMax = compressedMaxTemp;
		a = compressedMin - 0.5;
		b = Math.pow(compressedMax-compressedMin, 2)/(uncompressedMax-uncompressedMin);
		c = b * uncompressedMin;

		//keep a sum and count for each value of i, then in next loop average
		
		//System.out.println("CompressSqrt.init:  uncompressedMin="+uncompressedMin);
		//System.out.println("CompressSqrt.init:  uncompressedMax="+uncompressedMax);
		//System.out.println("CompressSqrt.init:    compressedMin="+compressedMin);
		//System.out.println("CompressSqrt.init:    compressedMax="+compressedMax);
		//System.out.println("CompressSqrt.init:                a="+a);
		//System.out.println("CompressSqrt.init:                b="+b);
		//System.out.println("CompressSqrt.init:                c="+c);

		n = new int[compressedMax-compressedMin];
		int[] sum = new int[compressedMax-compressedMin];
		for (int i=uncompressedMin; i<uncompressedMax; i++) {
			int compressed = compress(i);
			sum[compressed-compressedMin] += i;
			n[compressed-compressedMin] ++;
			//System.out.println("CompressSqrt.init:  uncompressed="+i+" compressed="+compressed+" sum="+sum[compressed-compressedMin]+" n="+n[compressed-compressedMin]);
		}
		uncompressedLookupTable = new double[compressedMax-compressedMin];		
		for (int i=compressedMin; i<compressedMax; i++) {
			try {
				uncompressedLookupTable[i-compressedMin] = sum[i-compressedMin]/(double) n[i-compressedMin];
			} catch (ArithmeticException e) {
				uncompressedLookupTable[i-compressedMin] = Double.NaN;				
			}
			//System.out.println("CompressSqrt.init:  compressed="+i+" sum="+sum[i-compressedMin]+" n="+n[i-compressedMin]+" uncompressedLookupTable="+uncompressedLookupTable[i-compressedMin]);
		}

	}
	/**
	 * 
	 * @param uncompressedMin
	 * @param uncompressedMax
	 * @param nBits number of bits in compressed values, ranging from 0 to 2^nBits-1
	 */
	public CompressSqrt(int uncompressedMin, int uncompressedMax, int nBits) {
		int compressedMinTemp = 0;
		int compressedMaxTemp = (int) Math.pow(2, nBits);
		init(uncompressedMin, uncompressedMax, compressedMinTemp, compressedMaxTemp);
	}
	/**
	 * @param uncompressedMin
	 * @param uncompressedMax
	 * @param compressedMin
	 * @param compressedMax
	 */
	public CompressSqrt(int uncompressedMin, int uncompressedMax, int compressedMin, int compressedMax) {
		init(uncompressedMin, uncompressedMax, compressedMin, compressedMax);
	}
	
	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#compress(double)
	 */
	@Override
	public int compress(double uncompressed) {
		if (uncompressed <= uncompressedMin) {
			return compressedMin;
		}
		if (uncompressed >= uncompressedMax) {
			return compressedMax-1;
		}
		int retval = (int) Math.floor(sqrtABC(uncompressed));
		return retval;
	}
	/**
	 * @param uncompressed
	 * @return the double value calculated from a, b,c
	 */
	public double sqrtABC(double uncompressed) {
		double retval = 0.5 + a + Math.sqrt(b*uncompressed - c);	
		return retval;
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#restore(int)
	 */
	@Override
	public double restore(int compressed) {
		if (compressed <= compressedMin ) {
			return uncompressedMin;
		}
		if (compressed >= compressedMax ) {
			return uncompressedMax;
		}
		double retval = uncompressedLookupTable[compressed-compressedMin];
		//System.out.println("CompressSqrt:  compressed="+compressed+" retval="+retval);
		return retval;
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#nValues(int)
	 */
	@Override
	public int nValues(int compressed) {
		if (compressed <= compressedMin ) {
			return uncompressedMin;
		}
		if (compressed >= compressedMax ) {
			return uncompressedMax;
		}
		int retval = n[compressed-compressedMin];
		//System.out.println("CompressSqrt:  compressed="+compressed+" retval="+retval);
		return retval;
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#getNBits()
	 */
	@Override
	public int getNBits() {
		double temp = Math.log(compressedMax-compressedMin)/Math.log(2);
		int retval = (int) Math.ceil(temp);
		return retval;
	}
}
