/*
 * Created on Nov 6, 2004
 *
 */
package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.sim.random.RandomSphere;
import gov.fnal.eag.sim.random.SimRandomPoisson;
import gov.fnal.eag.sim.random.SimRandomUniform;
import gov.fnal.eag.wcs.SphericalCoordinate;

import java.util.logging.Logger;

/**
 * static classes to generate cosmic rays 
 */
public class CosmicRay   {
	Logger logger = Logger.getLogger(CosmicRay.class.toString());
	//private double epsilon = 1e-6;
	private double stepSizeInMicrons = 5.0;

	private PixeledImage pi;
	private PixelDetector pd;
	private RandomSphere randomSphere;
	private SimRandomPoisson simRandomPoisson;
	private SimRandomUniform simRandomUniform;
	
	/**
	 * @param pi PixeledImage to populate with cosmic rays
	 * @param pd detector description
	 * @param rs generates points uniformly on a sphere
	 * @param simRandomPoisson generate poisson random numbers
	 * @param simRandomUniform generate uniform random numbers
	 */
	public CosmicRay(PixeledImage pi, PixelDetector pd, RandomSphere rs, SimRandomPoisson simRandomPoisson, SimRandomUniform simRandomUniform) {
		this.pi = pi; 
		this.pd=pd;
		this.randomSphere = rs;
		this.simRandomPoisson = simRandomPoisson;
		this.simRandomUniform = simRandomUniform;
	}

	/**
	 * add cosmic rays
	 * @param nToAdd the number to add
	 */
	public void addManyCosmicRays(int nToAdd) {
		int delta = 4;
		int rowMin = -delta;
		int rowMax = pd.rowCount+delta;
		int colMin = -delta;
		int colMax = pd.columnCount + delta;
		double meanFlux = nToAdd*( (pd.rowCount+2*delta)*(pd.columnCount+2*delta)/(pd.rowCount*pd.columnCount));
		int nToDo = simRandomPoisson.next(meanFlux);
		logger.finer("in CosmicRay.addManyCosmicRays:  nToAdd="+nToAdd+" nToDo="+nToDo);
		for (int i=0; i<nToDo; i++) {
			SphericalCoordinate point = randomSphere.next();
			
			
			double azimuth = point.getLongitude(); // ranges from 0 to 360
			double altitude = point.getLatitude(); // ranges from -90 to 90
			altitude = Math.abs(altitude);
			double row = simRandomUniform.next(rowMin, rowMax);
			double col = simRandomUniform.next(colMin, colMax);
			addOneCosmicRay(row, col, altitude, azimuth);
			if (i % 100 == 0) {
				logger.finer("in addManyCosmicRays:  i="+i+"/"+nToDo);
			}
		}
	}


	/**
	 * adds one cosmic ray
	 * @param row starting row position in pixels at z=0
	 * @param col starting col position in pixels at z=0
	 * @param altitude angle of incidence in degrees from 90.0 (straight in) to 0.1 (sideways) 
	 * @param azimuth the other angle in degrees, from 0 to 360.
	 * @return true, no matter what!
	 */
	public boolean addOneCosmicRay(double row, double col, double altitude, double azimuth) {
		// Dig a few things out of the input classes to ease typing

		// If theta is not legal, return right away.
		if (Math.abs(altitude) < 0.1) {return false;}
		
		// x is the column direction; y is the row direction; z is into the ccd

		// Get the starting positions in microns
		double x = col * pd.getPixelWidth();
		double y = row * pd.getPixelWidth();
		double z = 0.0;

		// Prepare to do the trigonometry
		double sinAltitude = Math.sin(altitude*Math.PI/180.0);
		double cosAltitude = Math.cos(altitude*Math.PI/180.0);
		double sinAzimuth   = Math.sin(azimuth*Math.PI/180.0);
		double cosAzimuth   = Math.cos(azimuth*Math.PI/180.0);
		
		// Get the step sizes in microns;
		double dz = Math.abs(stepSizeInMicrons*sinAltitude);
		double dx = stepSizeInMicrons*cosAltitude*cosAzimuth;
		double dy = stepSizeInMicrons*cosAltitude*sinAzimuth;
		int irow, icol;
		double nElectronPerStep = pd.getDnedx() * stepSizeInMicrons;
		
		double[] flux, dxx, dyy, dzz;

		// If not diffusion, sample only the middle of the diffusion grid
		int nSig = 0;
		int nPerSigma = 2;
		if (pd.getDiffusion()) {
			nSig = 5; // number of sigmas to sample
			nPerSigma = 10; // number of steps to take per sigma		
		} 
			
		int nbin = 2*nSig*nPerSigma + 1;
		double delta = 1.0/nPerSigma;
		int index = 0;
		int nbin2 = nbin*nbin;
		flux = new double[nbin2];
		dxx  = new double[nbin2];
		dyy  = new double[nbin2];
		dzz  = new double[nbin2];
		double fluxSum = 0.0;
		for (int xbin=0; xbin < nbin; xbin++) {
			double xx = -nSig+xbin*delta;
			for (int ybin=0; ybin < nbin; ybin++) {
				double yy = -nSig+ybin*delta;
				double rho = Math.sqrt(xx*xx + yy*yy);
				flux[index] = Math.exp(-0.5 * rho * rho) * stepSizeInMicrons;
				fluxSum += flux[index];
				// Shift from the "declination" angle -- azimuth=90 is no change.
				double xTemp = xx*sinAltitude;
				double yTemp = yy;
				double zTemp = xx*cosAltitude;
				// The final position is due to the "right ascension" angle"
				dxx[index] =  xTemp * cosAzimuth + yTemp * sinAzimuth;
				dyy[index] = -xTemp * sinAzimuth + yTemp * cosAzimuth;
				dzz[index] = zTemp;
				index++;
			}
		} 
		for (int ibin=0; ibin<nbin2; ibin++) {
			flux[ibin] *= nElectronPerStep/fluxSum;
		}
		double zGrid;
		double pixelWidth = pd.getPixelWidth();
		double distanceTravelled = 0.0;
		double zMax = pd.getPixelThickness();
		//logger.finer("in cosmic:  zMax="+zMax+" dz="+dz+" temp="+temp);
		while (z < zMax) {
			//logger.finer("distanceTravelled="+distanceTravelled+" z="+z+" zMax="+zMax);
			double diffusionSigma = pd.getSigma()*(distanceTravelled/200.0);
			for (int ibin=0; ibin<nbin2; ibin++) {
				// If we are in a valid pixel, put in some electrons
				irow = (int) Math.floor((y+diffusionSigma*dyy[ibin])/pixelWidth);
				if (irow >= 0 && irow < pd.rowCount) {
					icol = (int) Math.floor((x+diffusionSigma*dxx[ibin])/pixelWidth);
					if (icol >=0 && icol < pd.columnCount) {
						zGrid = z+diffusionSigma*dzz[ibin];
						if (zGrid >= 0.0 && zGrid < pd.getPixelThickness() ) {
							pi.incrByRowColNoOffset(irow,icol,flux[ibin]);
						} else {
							//logger.finer("zGrid out of range zGrid="+zGrid);
						}
					} else {
						//logger.finer(" icol out of range icol="+icol);
					}
				} else {
					//logger.finer(" irow out of range irow="+irow);
				}
			}
			// Now move along to the next point
			x+=dx; y+=dy; z+=dz; distanceTravelled += stepSizeInMicrons;
		}
		return true;
	}
}
