package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.sim.random.SimRandomNormal;
import gov.fnal.eag.sim.repository.image.FitsDataFormat;
import gov.fnal.eag.sim.repository.image.ImageRepository;
import gov.fnal.eag.util.misc.StringUtil;

import java.util.Iterator;

import nom.tam.fits.HeaderCardException;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.UnflaggedOption;

/**
 * compressed data generated by the CRIC v3 board
 * @author stoughto
 *
 */
public class Cric3Compress extends LookupTable {

	private Cric3Digitize cd;
	private CompressSqrt[] compressSqrts;
	private int[] compressedLimit = new int[3];
	private int nBits;

	/**
	 * default number of bits for compression
	 */
	public static int nBitsDefault = 10;
	/**
	 * default Cric3Digitize
	 */
	public Cric3Digitize cric3DigitizeDefault = new Cric3Digitize();
	/**
	 * default constructor
	 */
	public Cric3Compress() {
		init(cric3DigitizeDefault, nBitsDefault, null);
	}
	/**
	 * @param cd
	 * @param nBits
	 */
	public Cric3Compress(Cric3Digitize cd, int nBits) {
		init(cd, nBits, null);
	}
	
	/**
	 * @param cd 
	 * @param nBits
	 * @param compressedLimit the exclusive upper limit of each of the three ranges of compressed values
	 */
	public Cric3Compress(Cric3Digitize cd, int nBits, int[] compressedLimit) {
		init(cd, nBits, compressedLimit);
	}
	private void init(Cric3Digitize cdIn, int nBitsIn, int[] compressedLimitIn) {
		this.cd = cdIn;
		this.nBits = nBitsIn;
		if (compressedLimitIn == null) {
			compressedLimit[0] = (int) Math.pow(2, nBitsIn-1);
			compressedLimit[1] = compressedLimit[0] + (int) Math.pow(2, nBitsIn-2);
			compressedLimit[2] = compressedLimit[1] + (int) Math.pow(2, nBitsIn-2);
		} else {
			compressedLimit[0] = compressedLimitIn[0];
			compressedLimit[1] = compressedLimitIn[1];
			compressedLimit[2] = compressedLimitIn[2];
		}
		compressSqrts = new CompressSqrt[3];
		/*
		for (int i=0; i<3; i++) {
			int uncompressedMin = cdIn.digitize(0);
			int compressedMin = 0;
			if (i>0) {
				uncompressedMin = cdIn.digitize(cdIn.getLimit(i-1), i) & 0x3FFF;
				compressedMin = compressedLimit[i-1];
			}
			int uncompressedMax = (int) Math.pow(2, 14);
			int compressedMax = compressedLimit[i];
			if (i<2) {
				uncompressedMax = cdIn.digitize(cdIn.getLimit(i), i) & 0x3FFF;
			}
			compressSqrts[i] = new CompressSqrt(uncompressedMin, uncompressedMax, compressedMin, compressedMax);
		}
		 */
		for (int i=0; i<3; i++) {
			double electronsMin = cdIn.getLimitMin(i);
			double electronsMax = cdIn.getLimitMax(i);
			//System.out.println("Cric3Compress.init:  i="+i+" electronsMin="+electronsMin+" electronsMax="+electronsMax);
			int uncompressedMin = cdIn.digitize(electronsMin, i) & 0x3FFF;
			int uncompressedMax = cdIn.digitize(electronsMax, i) & 0x3FFF;
			int compressedMin = 0;
			if (i>0) {
				compressedMin = compressedLimit[i-1];
			}
			int compressedMax = compressedLimit[i];
			compressSqrts[i] = new CompressSqrt(uncompressedMin, uncompressedMax, compressedMin, compressedMax);
			//System.out.println("Cric3Compress.init:  i="+i+" uncompressedMin="+uncompressedMin+" uncomresedMax="+uncompressedMax+" compressedMin="+compressedMin+" compressedMax="+compressedMax);
		}
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#compress(double)
	 */
	@Override
	public int compress(double electrons) {
		int digitized = cd.digitize(electrons);
		int index = cd.getIndex(digitized);
		int base = digitized & 0x3FFF;
		//System.out.println("Cric3Compress.compress:  electrons="+electrons+" digitized="+digitized+" index="+index);
		int retval = compressSqrts[index].compress(base);
		//System.out.println(String.format("Cric3Compress.compress: electrons=%.2f digitized=%d compressed=%d", electrons, digitized, retval));
		return retval;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#getNBits()
	 */
	@Override
	public int getNBits() {
		return nBits;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#restore(int)
	 */
	@Override
	public double restore(int compressed) {
		int index = getIndex(compressed);
		double uncompressed = compressSqrts[index].restore(compressed);
		double retval =  cd.restore(uncompressed, index);
		//System.out.println("Cric3Compress.restore: compressed="+compressed+" index="+index+" uncompressed="+uncompressed+" retval="+retval);
		return retval;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#nValues(int)
	 */
	@Override
	public int nValues(int compressed) {
		int index = getIndex(compressed);
		int retval = compressSqrts[index].nValues(compressed);
		return retval;
	}

	private int getIndex(int compressed) {
		int index;
		if (compressed < compressedLimit[0]) {
			index = 0;
		} else if (compressed < compressedLimit[1]) {
			index = 1;
		} else {
			index = 2;
		}
		return index;
	}

	/**
	 * @param args inFileName outFileName [optional flagged keywords]
	 * <p>
	 * --expTimeKeyword (String) the name of the keyword in the FITS header to get exposure time
	 * <p>
	 * --skyPerSecond (double) sky counts per second per pixel
	 * <p>
	 * --readNoises (three double values) the value of the read noise in each of the three ranges
	 * <p>
	 * --gains (three double values) the value of the gain in each of the three ranges
	 * <p>
	 * --arminBias (double) constant bias applied to all three ranges
	 * <p>
	 * --aduThreshold (double) defines the ADU level where the range changes
	 * <p>
	 * --nBits (int) number of bits in the compressed values
	 * <p>
	 * --scaleFactor (double) multiplicative scale factor
	 * <p>
	 * --showBernsteinParameter "skyLevel skySigma" prints log_2 ( (skySigma) / (bits per step) )
	 * where "bits per step" is calculated near skyLevel
	 * 
	 * <ol>
	 * <li>create the codec with the readNoises, gains, arminBias, aduThredhosd, and nBits values
	 * <li>reads in the inputFileName.  
	 * <li>Multiply by exposure time.  
	 * <li>Add skyPerSecond*expTime to each pixel.  
	 * <li>Multiply by scale factor.
	 * <li>compress then decompress the data.  
	 * <li>Divide by scale factor
	 * <li>subtract skyPerSecond*expTime
	 * <li>divide by exposure time.
	 * <li>Write result to outFileName.
	 * </ol>
	 * @throws HeaderCardException
	 * @throws JSAPException 
	 */
	public static void main (String[] args) throws HeaderCardException, JSAPException {
		JSAP jsap = new JSAP();
		jsap.registerParameter(new UnflaggedOption("inputFileName",  JSAP.STRING_PARSER, true, "input file name"));
		jsap.registerParameter(new UnflaggedOption("outputFileName", JSAP.STRING_PARSER, true, "output file name"));
		jsap.registerParameter(new FlaggedOption("expTimeKeyword",   JSAP.STRING_PARSER,  JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'e', "expTimeKeyword"));
		jsap.registerParameter(new FlaggedOption("skyPerSecond",     JSAP.DOUBLE_PARSER,  JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 's', "skyPerSecond"));
		jsap.registerParameter(new FlaggedOption("readNoises",       JSAP.STRING_PARSER,  StringUtil.toString(Cric3Digitize.readNoiseDefault), JSAP.NOT_REQUIRED, 'r', "readNoises" ) );
		jsap.registerParameter(new FlaggedOption("compressedLimits", JSAP.STRING_PARSER,  "default", JSAP.NOT_REQUIRED, 'c', "compressedLimits" ) );
		jsap.registerParameter(new FlaggedOption("gains",            JSAP.STRING_PARSER,  StringUtil.toString(Cric3Digitize.gainDefault),      JSAP.NOT_REQUIRED, 'g', "gains" ) );
		jsap.registerParameter(new FlaggedOption("arminBias",        JSAP.DOUBLE_PARSER,  Double.toString(Cric3Digitize.arminBiasDefault), JSAP.NOT_REQUIRED, 'a', "arminBias" ) );
		jsap.registerParameter(new FlaggedOption("aduThreshold",     JSAP.DOUBLE_PARSER,  Double.toString(Cric3Digitize.aduThresholdDefault), JSAP.NOT_REQUIRED, 't', "aduThreshold" ) );
		jsap.registerParameter(new FlaggedOption("nBits",            JSAP.INTEGER_PARSER, Integer.toString(Cric3Compress.nBitsDefault), JSAP.NOT_REQUIRED, 'n', "nBits" ) );
		jsap.registerParameter(new FlaggedOption("scaleFactor",      JSAP.DOUBLE_PARSER,  Double.toString(1.0), JSAP.NOT_REQUIRED, 'f', "scaleFactor" ) );
		jsap.registerParameter(new FlaggedOption("showBernsteinParameter",      JSAP.STRING_PARSER,  "null", JSAP.NOT_REQUIRED, 'b', "showBernsteinParameter" ) );

		JSAPResult result = jsap.parse(args);
		if (!result.success()) {
			Iterator<?> iterator = result.getErrorMessageIterator();
			while (iterator.hasNext()) {
				Object em = iterator.next();
				System.out.println(em);        		
			}
			System.out.print("Usage: java " + Cric3Compress.class.getName());
			System.out.println(" " + jsap.getUsage());
			return;
		}

		String inputFileName = result.getString("inputFileName");
		PixeledImage pi = ImageRepository.readPixeledImage(inputFileName);
		String outputFileName = result.getString("outputFileName");
		double expTime = 1.0;
		if (result.userSpecified("expTimeKeyword") ) {
			String expTimeKeyword = result.getString("expTimeKeyword");
			expTime = pi.header.getDoubleValue(expTimeKeyword);	
			System.out.println("Cric3Compress.main:  " +
					"multiply by expTime="+expTime);
			pi.multiplyScalar(expTime);
		}

		double skyPerSecond = 0.0;
		if (result.userSpecified("skyPerSecond")) {
			skyPerSecond = result.getDouble("skyPerSecond");
			double skyValue = skyPerSecond*expTime;
			System.out.println("Cric3Compress.main: skyPerSecond="+skyPerSecond+" add sky value = "+skyValue);
			pi.addScalar(skyValue);
		}

		double[] readNoise = StringUtil.toDoubleArray(result.getString("readNoises"));
		double[] gain = StringUtil.toDoubleArray(result.getString("gains"));
		double arminBias = result.getDouble("arminBias");
		double aduThreshold = result.getDouble("aduThreshold");		
		Cric3Digitize c3d = new Cric3Digitize(readNoise, gain, aduThreshold, arminBias);
		int nBits = result.getInt("nBits");
		String compressedLimitsString = result.getString("compressedLimits");
		int[] compressedLimits;
		if (compressedLimitsString.equalsIgnoreCase("default")) {
			compressedLimits = null;
		} else {
			compressedLimits = StringUtil.toIntArray(compressedLimitsString, ",");
		}
		Cric3Compress cdlt = new Cric3Compress(c3d, nBits, compressedLimits);
		
		String showBernsteinParameter = result.getString("showBernsteinParameter");
		if (!showBernsteinParameter.equalsIgnoreCase("null")) {
			double[] sbp = StringUtil.toDoubleArray(showBernsteinParameter);
			double sky = sbp[0];
			double skySigma = sbp[1];
			double bp = cdlt.calculateBernsteinParameter(sky, skySigma);
			System.out.println("Bernstein Parameter="+bp+" for skyLevel="+sky+" skySigma="+skySigma);
		}
		double scaleFactor = result.getDouble("scaleFactor");
		pi.multiplyScalar(scaleFactor);
		cdlt.compressDecompress(pi);
		pi.multiplyScalar(1.0/scaleFactor);
		pi.header.addValue("CCDVER", getVersion(), "Cric3Compress version");
		pi.header.addValue("SKY", skyPerSecond, "counts per second added before compress/uncompress and then subtracted");
		if (skyPerSecond != 0.0) {
			double skyValue = skyPerSecond*expTime;
			System.out.println("Cric3Compress.main:  subtract skyValue="+skyValue);
			pi.addScalar(-skyPerSecond*expTime);
		}
		if (expTime != 0.0) {
			double factor = 1.0/expTime;
			System.out.println("Cric3Compress.main:  " +
					"expTime="+expTime+" so multiply by "+factor);
			pi.multiplyScalar(factor);
		}
		System.out.println("Cric3Compress:  write outFileName="+outputFileName);
		ImageRepository.write(outputFileName, pi, FitsDataFormat.FLOAT, true);		
	}

	/**
	 * @return the cvs version string for this class
	 */
	public static String getVersion() {
		String version = "Cric3Compress.java,v 1.4 2008/04/27 19:02:37 descvs Exp";
		return version;
	}
	/**
	 * @return the Cric3Digitize associated with this
	 */
	public Cric3Digitize getCd() {
		return cd;
	}
	
	/**
	 * 
	 * calculate the b parameter described by Gary Bernstein:
	 * 
	In any case the critical parameter to examine is
	b = log_2 ( (sky sigma) / (bits per step) )
	where sky sigma is the RMS of sky pixels in the image, and (bits per step) is the number of output values per input value for the compression algorithm in the vicinity of the sky level.
	 * @param skyLevel in electrons per pixel
	 * @param skySigma rms of the sky level
	 * @return the b parameter
	*/
	public double calculateBernsteinParameter(double skyLevel, double skySigma) {
		double bitsPerStep = calculateNValuesPerStep(skyLevel, skySigma);
		double retval = Math.log(skySigma/bitsPerStep)/Math.log(2);		
		return retval;
	}
	/**
	 * @param skyLevel
	 * @param skySigma
	 * @return the average number of values for each 
	 */
	public double calculateNValuesPerStep(double skyLevel, double skySigma) {
		int n = 100;
		SimRandomNormal srn = new SimRandomNormal(123, skyLevel, skySigma, "", false);
		double sum = 0.0;
		for (int i=0; i<n; i++) {
			double value = srn.next();
			int compressed = compress(value);
			int nValues = nValues(compressed);
			sum += nValues;
		}
		double retval = sum/n;
		return retval;
	}
}
