package gov.fnal.eag.sim.pixel;

import java.util.Iterator;

import nom.tam.fits.HeaderCardException;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.UnflaggedOption;

import gov.fnal.eag.sim.random.SimRandomNormal;
import gov.fnal.eag.sim.repository.image.FitsDataFormat;
import gov.fnal.eag.sim.repository.image.ImageRepository;
import gov.fnal.eag.util.misc.StringUtil;

/**
 * digitize electronics with three ranges
 * @author stoughto
 *
 */
public class Cric3Digitize extends LookupTable{
	/*
	 From Armin Karcher April 8, 2008

"I don't think anyone really knows the numbers as no CCD has been operated with a CRIC.
I like adu=(electrons+bias)/gain +8192.5
where bias is probably negative, probably less than 1500. (you pick a number)
The CRIC gain is 1.9uV/adu, 30uV/adu, 61uV/adu  +/- 5% , The CCD gain is ~3.6 to 3.9 uV/e
giving an overall gain of about .5, 8, 16
bias should be entirely independent of gain, but this has to be verified in more detail..."

We need to translate these to the values we use.

		int retval = (int) (in/gain + bias + 0.5);

so if this bias is bias_jm, and the bias of -1000 that Armin suggests turns out to be -1000/gain + 8192

	 */
	
	/**
	 * default gain in each region, in electrons/ADU
	 */
	public static double[] gainDefault = {0.5, 8.0, 16.0};
	
	/**
	 * default ADU value that defines boundaries between ranges
	 */
	public static double aduThresholdDefault = 0.8*Math.pow(2,14);
	/**
	 * default sigma of the read noise for each range, in electrons
	 */
	public static double[] readNoiseDefault = {4.0, 9.6, 19.2};
	
	/**
	 * default value of the "Armin bias"
	 */
	public static double arminBiasDefault = -1000;
	
	private double[] readNoise, gain, limit, limitMin, limitMax, bias;

	
	
	static int[] indexToMask = {0, 2, 3};
	static int[] maskToIndex = {0, -9999, 1, 2};
	
	/**
	 * fully-specified constructor
	 * @param readNoise the read noise in each range
	 * @param gain the gain in each range
	 * @param aduThreshold threshold for switching to next range
	 * @param arminBias the bias is calculated as arminBias/gain[i] + 8192.0;
	 */
	public Cric3Digitize(double[] readNoise, double[] gain, double aduThreshold, double arminBias) {
		this.readNoise = readNoise.clone();
		this.gain = gain.clone();
		bias = new double[3];
		for (int i=0; i<3; i++) {
			bias[i] = arminBias/gain[i] + 8192.0;
		}
		limit = new double[3];
		limitMin= new double[3];
		limitMax= new double[3];
		
		// some algebra
		// ADU = electrons/gain + bias + 0.5
		// electrons = gain*(ADU - bias - 0.5)
		
		limitMin[0] =                                          - 6.0*readNoise[0];
		limitMax[0] = gain[0]*(aduThreshold   - bias[0] - 0.5) + 6.0*readNoise[0];
		limit[0] =    gain[0]*(aduThreshold   - bias[0] - 0.5);
		
		limitMin[1] = gain[0]*(aduThreshold   - bias[0] - 0.5) - 6.0*readNoise[1];
		limitMax[1] = gain[1]*(aduThreshold   - bias[1] - 0.5) + 6.0*readNoise[1];
		limit[1] =    gain[1]*(aduThreshold   - bias[1] - 0.5);

		limitMin[2] = gain[1]*(aduThreshold   - bias[1] - 0.5) - 6.0*readNoise[2];
		limitMax[2] = gain[2]*(Math.pow(2,14) - bias[2] - 0.5) + 6.0*readNoise[2];
		limit[2] =    gain[2]*(Math.pow(2,14) - bias[2] - 0.5);
		
	}
	
	/**
	 * default constructor
	 */
	public Cric3Digitize() {
		this(readNoiseDefault, gainDefault, aduThresholdDefault, arminBiasDefault);		
	}
	/**
	 * @param electrons
	 * @return the index to the DigitizeLookupTable to use for this value
	 */
	public int getIndex(double electrons) {
		int index = 2;
		for (int i=0; i<2; i++) {
			if (electrons < limit[i]) {
				index = i;
				break;
			}
		}
		return index;
	}

	/**
	 * add read noise to the electrons and digitize to a 14-bit unsigned integer
	 * @param electrons
	 * @param srn
	 * @return the digitized value
	 */
	public int digitize(double electrons, SimRandomNormal srn) {
		double mean = srn.getMean();
		double sigma = srn.getSigma();
		
		// Use the read noise of the next higher range
		int tempIndex = getTempIndex(electrons);
		double next0 = srn.next();
		double tempReadNoise = readNoise[tempIndex]*(next0-mean)/sigma;		
		int trueIndex =  getIndex(electrons+tempReadNoise);
		double next1 = srn.next();
		double trueReadNoise = readNoise[trueIndex]*(next1-mean)/sigma;
		int retval = digitize(electrons+trueReadNoise, trueIndex);
		return retval;
	}
	private int getTempIndex(double electrons) {
		int index = getIndex(electrons);
		if (index == 0) return 1;
		if (index == 2) return 2;
		if (electrons < 0.5*(limit[0]+limit[1])) return 1;
		return 2;
	}

	/**
	 * @param electrons
	 * @return the digitized value
	 */
	public int digitize(double electrons) {
		return digitize(electrons, Integer.MIN_VALUE);
	}
	/**
	 * 
	 * @param electrons the value to digitize, which already includes read noise
	 * @param index the range to digitize in.  If less than zero, calculate the range based on electrons
	 * @return the digitixed value
	 */
	public int digitize(double electrons, int index) {
		// what range are we in
		if (index < 0) {
			index = getIndex(electrons);
		}	
		int mask = indexToMask[index] << 14;

		// calculate ADU's by ADU = electrons/gain
		int retval = (int) (electrons/gain[index] + bias[index] + 0.5);
		//System.out.println("Cric3Digitize.digitize:  electrons="+electrons+" retval="+retval);
		// digitize to 14 bits
		retval = Math.min(retval, 0x3FFF);
		retval = Math.max(retval, 0);

		// add the mask bits
		retval += mask;		
		return retval;
	}
	
	/**
	 * @param electrons
	 * @return the digitized value, without casting to an integer value
	 */
	public double digitizeTruth(double electrons) {
		int index = getIndex(electrons);
		double retval = electrons/gain[index] + bias[index];
		return retval;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#restore(int)
	 */
	@Override
	public double restore(int digitized) {
		// what range are we in
		int index = getIndex(digitized);
		// get the values from the 14 bits
		int base = digitized & 0x3FFF;
		double retval = restore(base, index);
		return retval;
	}
	
	/**
	 * @param uncompressed
	 * @param index
	 * @return the restored value
	 */
	public double restore(double uncompressed, int index) {
		double retval = gain[index]*(uncompressed - bias[index]);
		return retval;
	}
	/**
	 * @param in the encoded value, with the mask in the most significant two bits
	 * @return 0, 1, or 2, corresponding to a mask value of 0, 2, or 3.
	 */
	public int getIndex(int in) {
		int mask = 3 & (in>>14);
		if (mask == 1) {
			throw new IllegalArgumentException("code=1 is not legal in="+in);
		}
		int index = maskToIndex[mask];
		return index;
	}

	/**
	 * @param index
	 * @return the limit for this range
	 */
	public double getLimit(int index) {
		return limit[index];
	}
	/**
	 * @param index
	 * @return the limitMin for this range
	 */
	public double getLimitMin(int index) {
		return limitMin[index];
	}
	/**
	 * @param index
	 * @return limitMax for this range
	 */
	public double getLimitMax(int index) {
		return limitMax[index];
	}
	/**
	 * @param index
	 * @return the read noise for this range
	 */
	public double getReadNoise(int index) {
		return readNoise[index];
	}

	/**
	 * @param index
	 * @return the gain for this range
	 */
	public double getGain(int index) {
		return gain[index];
	}
	/**
	 * @param index
	 * @return the bias for this range
	 */
	public double getBias(int index) {
		return bias[index];
	}

	/**
	 * return the 16-bit digitized value
	 * @see gov.fnal.eag.sim.pixel.LookupTable#compress(double)
	 */
	@Override
	public int compress(double electrons) {
		return digitize(electrons);
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#getNBits()
	 */
	@Override
	public int getNBits() {
		return 16;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#nValues(int)
	 */
	@Override
	public int nValues(int compressed) {
		throw new UnsupportedOperationException();
	}
	
	
	/**
	 * @param args inFileName outFileName [optional flagged keywords]
	 * <p>
	 * --expTimeKeyword (String) the name of the keyword in the FITS header to get exposure time
	 * <p>
	 * --skyPerSecond (double) sky counts per second per pixel
	 * <p>
	 * --readNoises (three double values) the value of the read noise in each of the three ranges
	 * <p>
	 * --gains (three double values) the value of the gain in each of the three ranges
	 * <p>
	 * --arminBias (double) constant bias applied to all three ranges
	 * <p>
	 * --aduThreshold (double) defines the ADU level where the range changes
	 * <p>
	 * --scaleFactor (double) multiplicative scale factor
	 * <p>
	 * 
	 * <ol>
	 * <li>create the codec with the readNoises, gains, arminBias, aduThredhosd, and nBits values
	 * <li>reads in the inputFileName.  
	 * <li>Multiply by exposure time.  
	 * <li>Add skyPerSecond*expTime to each pixel.  
	 * <li>Multiply by scale factor.
	 * <li>digitize then restore the data.  
	 * <li>Divide by scale factor
	 * <li>subtract skyPerSecond*expTime
	 * <li>divide by exposure time.
	 * <li>Write result to outFileName.
	 * </ol>
	 * @throws HeaderCardException
	 * @throws JSAPException 
	 */
	public static void main (String[] args) throws HeaderCardException, JSAPException {
		JSAP jsap = new JSAP();
		jsap.registerParameter(new UnflaggedOption("inputFileName",  JSAP.STRING_PARSER, true, "input file name"));
		jsap.registerParameter(new UnflaggedOption("outputFileName", JSAP.STRING_PARSER, true, "output file name"));
		jsap.registerParameter(new FlaggedOption("expTimeKeyword",   JSAP.STRING_PARSER,  JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'e', "expTimeKeyword"));
		jsap.registerParameter(new FlaggedOption("skyPerSecond",     JSAP.DOUBLE_PARSER,  JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 's', "skyPerSecond"));
		jsap.registerParameter(new FlaggedOption("readNoises",       JSAP.STRING_PARSER,  StringUtil.toString(Cric3Digitize.readNoiseDefault), JSAP.NOT_REQUIRED, 'r', "readNoises" ) );
		jsap.registerParameter(new FlaggedOption("compressedLimits", JSAP.STRING_PARSER,  "default", JSAP.NOT_REQUIRED, 'c', "compressedLimits" ) );
		jsap.registerParameter(new FlaggedOption("gains",            JSAP.STRING_PARSER,  StringUtil.toString(Cric3Digitize.gainDefault),      JSAP.NOT_REQUIRED, 'g', "gains" ) );
		jsap.registerParameter(new FlaggedOption("arminBias",        JSAP.DOUBLE_PARSER,  Double.toString(Cric3Digitize.arminBiasDefault), JSAP.NOT_REQUIRED, 'a', "arminBias" ) );
		jsap.registerParameter(new FlaggedOption("aduThreshold",     JSAP.DOUBLE_PARSER,  Double.toString(Cric3Digitize.aduThresholdDefault), JSAP.NOT_REQUIRED, 't', "aduThreshold" ) );
		jsap.registerParameter(new FlaggedOption("scaleFactor",      JSAP.DOUBLE_PARSER,  Double.toString(1.0), JSAP.NOT_REQUIRED, 'f', "scaleFactor" ) );

		JSAPResult result = jsap.parse(args);
		if (!result.success()) {
			Iterator<?> iterator = result.getErrorMessageIterator();
			while (iterator.hasNext()) {
				Object em = iterator.next();
				System.out.println(em);        		
			}
			System.out.print("Usage: java " + Cric3Digitize.class.getName());
			System.out.println(" " + jsap.getUsage());
			return;
		}

		String inputFileName = result.getString("inputFileName");
		PixeledImage pi = ImageRepository.readPixeledImage(inputFileName);
		String outputFileName = result.getString("outputFileName");
		double expTime = 1.0;
		if (result.userSpecified("expTimeKeyword") ) {
			String expTimeKeyword = result.getString("expTimeKeyword");
			expTime = pi.header.getDoubleValue(expTimeKeyword);	
			System.out.println("Cric3Digitize.main:  " +
					"multiply by expTime="+expTime);
			pi.multiplyScalar(expTime);
		}

		double skyPerSecond = 0.0;
		if (result.userSpecified("skyPerSecond")) {
			skyPerSecond = result.getDouble("skyPerSecond");
			double skyValue = skyPerSecond*expTime;
			System.out.println("Cric3Digitize.main: skyPerSecond="+skyPerSecond+" add sky value = "+skyValue);
			pi.addScalar(skyValue);
		}

		double[] readNoise = StringUtil.toDoubleArray(result.getString("readNoises"));
		double[] gain = StringUtil.toDoubleArray(result.getString("gains"));
		double arminBias = result.getDouble("arminBias");
		double aduThreshold = result.getDouble("aduThreshold");		
		Cric3Digitize c3d = new Cric3Digitize(readNoise, gain, aduThreshold, arminBias);
		
		double scaleFactor = result.getDouble("scaleFactor");
		pi.multiplyScalar(scaleFactor);
		c3d.compressDecompress(pi);
		pi.multiplyScalar(1.0/scaleFactor);
		pi.header.addValue("SKY", skyPerSecond, "counts per second added before compress/uncompress and then subtracted");
		if (skyPerSecond != 0.0) {
			double skyValue = skyPerSecond*expTime;
			System.out.println("Cric3Digitize.main:  subtract skyValue="+skyValue);
			pi.addScalar(-skyPerSecond*expTime);
		}
		if (expTime != 0.0) {
			double factor = 1.0/expTime;
			System.out.println("Cric3Compress.main:  " +
					"expTime="+expTime+" so multiply by "+factor);
			pi.multiplyScalar(factor);
		}
		System.out.println("Cric3Digitize:  write outFileName="+outputFileName);
		ImageRepository.write(outputFileName, pi, FitsDataFormat.FLOAT, true);		
	}

}
