package gov.fnal.eag.sim.pixel;

import java.awt.geom.Point2D;

import gov.fnal.eag.wcs.SphericalCoordinate;
import uk.ac.starlink.table.RowListStarTable;

/**
 * describes an astronomical object as it reaches a pixeled detector, in pixel coordinates 
 */
public abstract class DetectorObject {
	protected boolean isSimple;
	protected double nPhoton, row, col, ra, dec;
	protected String fromTable;
	/**
	 * describes whether the implementation is a simple object, rather than a Shapelet
	 * @return true if it is simple
	 */
	public boolean isSimple() {
		return isSimple;
	}
	/**
	 * @return the number of photons in this object
	 */
	public double getNPhoton() {
		return nPhoton;
	}
	/**
	 * get the row pixel position
	 * @return the row
	 */
	public double getRow() {
		return row;
	}
	/**
	 * get the column pixel position
	 * @return the column
	 */
	public double getCol() {
		return col;
	}
	//public void setIsSimple(boolean isSimple);
    /**
     * set ra and dec for this object, to keep track of truth
     * @param ra 
     * @param dec
     */
    public void setRaDec(double ra, double dec) {
    	this.ra = ra;
    	this.dec = dec;
    }
    
    /**
     * @param sc the location to set
     */
    public void setRaDec(SphericalCoordinate sc) {
    	setRaDec(sc.getLongitude(), sc.getLatitude());
    }
	/**
	 * set ra, dec, and a label for this object, to keep track of truth
	 * @param ra
	 * @param dec
	 * @param fromTable
	 */
	public void setRaDec(double ra, double dec, String fromTable) {
		setRaDec(ra,dec);
		this.fromTable = fromTable;
	}

    /**
     * get the ra
     * 
     * @return the right ascension in degrees
     */
    public double getRa() {
    	return ra;
    }
    /**
     * get the declination
     * @return the declination in degrees
     */
    public double getDec() {
    	return dec;
    }
    
	/**
	 * move the row position of this object
	 * @param row in pixels
	 */
	public void setRow(double row) {
		this.row = row;
	}
	/**
	 * move the column position of this object
	 * @param col in pixels
	 */
	public void setCol(double col) {
		this.col = col;
	}
	/**
	 * set the number of photons for this object
	 * @param nPhoton
	 */
	public void setNPhoton(double nPhoton) {
		this.nPhoton = nPhoton;
	}
	
	/**
	 * convolve this object with a gaussian
	 * @param d sigma of the gaussian, in pixels
	 */
	public abstract void convolveGaussian(double d);

	/**
	 * add this object to the table
	 * @param rlst the table
	 */
	public abstract void addToTable(RowListStarTable rlst);


	/**
	 * get the flux integrated in the pixel
	 * @param iRow
	 * @param iCol
	 * @return flux integrated in the pixel
	 */
	public abstract double getFluxInPixel(int iRow, int iCol);
	
	/**
	 * clear arrays to prepare for garbage collection.  
	 */
	public abstract void gc();
	
	/**
	 * @param rowTemp
	 * @param colTemp
	 * @return flux per pixel at this location
	 */
	public abstract double getFlux(double rowTemp, double colTemp);
	
	/**
	 * @param pixelDiameter
	 * @return filled in with values for this object
	 */
	public  PixeledImage makePixeledImage(int pixelDiameter) {
		int row0 = (int) (row-pixelDiameter/2);
		int col0 = (int) (col-pixelDiameter/2);
		PixeledImage pi = new PixeledImage(pixelDiameter, pixelDiameter);
		pi.setRow0Col0(row0, col0);
		for (int r=row0; r<row0+pixelDiameter; r++) {
			for (int c=col0; c<col0+pixelDiameter; c++) {
				double value = getFluxInPixel(r, c);
				pi.setByRowCol(r, c, value);
			}
		}
		return pi;
	}
	   /**
     * render this object in all pixels
     * @param pi where to render
     * @param allowNegNums whether to allow negative numbers
     */
    public void renderToPixeledImage(PixeledImage pi, boolean allowNegNums) {
		for (int r=pi.getRowMin(); r<pi.getRowMax(); r++) {
			for (int c=pi.getColumnMin(); c<pi.getColumnMax(); c++) {
				double value = getFluxInPixel(r, c);
				if (value < 0 && !allowNegNums) {
					value = 0;
				}
				pi.incrByRowCol(r, c, value);
			}
		}
	}

    /**
     * @param radius
     * @param sky
     * @return the sum of flux-sky in pixels where the center of the pixel is closer than radius to the center of the object
     */
    public double getApertureFlux(double radius, double sky) {
    	Point2D.Double center = new Point2D.Double(col, row);
    	double retval = getApertureFlux(center, radius, sky);
    	return retval;
    }
    
    /**
     * @param center
     * @param radius
     * @param sky
     * @return the sum of flux-sky in pixels where the center of the pixel is closer than radius to center
     */
    public double getApertureFlux(Point2D.Double center, double radius, double sky) {
    	double radius2 = Math.pow(radius,2);
    	int rMin = (int) Math.floor(center.y-radius);
    	int rMax = (int) Math.ceil(center.y+radius);
    	int cMin = (int) Math.floor(center.x-radius);
    	int cMax = (int) Math.ceil(center.x+radius);
    	double retval = 0.0;
    	for (int r=rMin; r<rMax; r++) {
    		double dr2 = Math.pow(r+0.5-row,2);
    		for (int c=cMin; c<cMax; c++) {
    			double dc2 = Math.pow(c+0.5-col,2);
    			double d2 = dr2+dc2;
    			if (d2 < radius2) {
    				retval += getFluxInPixel(r,c)-sky;
    			}
    		}
    	}
    	return retval;
    }
}