package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.sim.extractor.SourceObject;
import gov.fnal.eag.util.plot.JFreeChartUtil;

import java.awt.Point;
import java.util.ArrayList;

import org.jfree.chart.JFreeChart;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

import uk.ac.starlink.table.RowListStarTable;

/**
 * a composite of several DetectorObjects plus a SourceObject.
 * Return the flux in the SourceObject where it is defined OR the the sum of the flux in the DetectorObjects
 * @author stoughto
 *
 */
public class DetectorObjectComposite extends DetectorObject {
	ArrayList<DetectorObject> detectorObjects = null;
	SourceObject sourceObject;
	Point sourceOffset;
	boolean[][] included = null;
	int xMin, yMin;
	Double halfLightRadius = null;
	
	/**
	 * each DetectorObject has a row,col position, and the location of the SourceObject is moved with sourceObject
	 * so these all line up with each other.  Then row,col move the composite on the detector.
	 * @param row 
	 * @param col 
	 * @param detectorObjects
	 * @param sourceObject
	 * @param sourceOffset x,y position to add to sourceObject row,col positions
	 */
	public DetectorObjectComposite(double row, double col, ArrayList<DetectorObject> detectorObjects, SourceObject sourceObject, Point sourceOffset) {
		this.row = row;
		this.col = col;
		this.detectorObjects = detectorObjects;
		this.sourceObject = sourceObject;
		if (sourceOffset == null) {
			this.sourceOffset = new Point(0,0);
		} else {
			this.sourceOffset = sourceOffset;
		}
		nPhoton = sourceObject.getFlux();
		for (DetectorObject detectorObject : detectorObjects) {
			nPhoton += detectorObject.getNPhoton();
		}
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#addToTable(uk.ac.starlink.table.RowListStarTable)
	 */
	@Override
	public void addToTable(RowListStarTable rlst) {
		throw new UnsupportedOperationException();
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#convolveGaussian(double)
	 */
	@Override
	public void convolveGaussian(double d) {
		throw new UnsupportedOperationException();
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#gc()
	 */
	@Override
	public void gc() {
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#getFlux(double, double)
	 */
	@Override
	public double getFlux(double rowTemp, double colTemp) {
		throw new UnsupportedOperationException();
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#getFluxInPixel(int, int)
	 */
	@Override
	public double getFluxInPixel(int r, int c) {
		double retval = 0.0;
		try {
			int xPixel = (int) Math.floor(c - sourceOffset.x - col);
			int yPixel = (int) Math.floor(r - sourceOffset.y - row);
			double flux = getFluxInSourceObject(yPixel, xPixel);
			retval += flux;
		} catch (IllegalArgumentException e) { // catch this when yPixel,xPixel is not in sourceObject
			for (DetectorObject detectorObject : detectorObjects) {
				int xPixel = (int) Math.floor(c - col);
				int yPixel = (int) Math.floor(r - row);
				retval += detectorObject.getFluxInPixel(yPixel, xPixel);
			}
		}
		return retval;
	}

	private double getFluxInSourceObject(int r, int c) {
		if (included == null) {
			yMin = sourceObject.getYMin();
			xMin = sourceObject.getXMin();
			int nrIncluded = sourceObject.getYMax() - yMin + 1;
			int ncIncluded = sourceObject.getXMax() - xMin + 1;
			included = new boolean[nrIncluded][ncIncluded];
			for (Point point : sourceObject.getPoints()) {
				int x = point.x-xMin;
				int y = point.y-yMin;
				included[y][x] = true;				
			}
		}
		try {
			if (included[r-yMin][c-xMin]) {
				double retval = sourceObject.getPixeledImage().getByRowCol(r, c) -  sourceObject.getFitSky();
				return retval;
			}
		} catch (ArrayIndexOutOfBoundsException e) {
		}
		throw new IllegalArgumentException();
	}
	/**
	 * @param title
	 * @param radius in pixels, to plot values
	 * @param doLogarithmicRange whether to use a log scale for fluxes
	 * @return a plot of the profile along the major axis of the source object, using the pixel values in the center and the fit profiles outside
	 */
	public JFreeChart makeChart(String title, double radius, boolean doLogarithmicRange) {
		XYSeriesCollection xysc = new XYSeriesCollection();
		XYSeries xySourceObject = new XYSeries(String.format("flux in source object - %.2f",sourceObject.getFitSky()));
		xysc.addSeries(xySourceObject);
		XYSeries[] xyDetectorObjects =  new XYSeries[detectorObjects.size()];
		for (int i=0; i<detectorObjects.size(); i++) {
			xyDetectorObjects[i] =  new XYSeries("i="+i);
			xysc.addSeries(xyDetectorObjects[i]);
		}
		double theta = sourceObject.getTheta();
		System.out.println("doc makeChart:  theta="+theta);
		if (Double.isNaN(theta)) {
			theta = 0.0;
		}
		{
			System.out.println("WARNING set theta to 80");
			theta=80.0;
		}
		double ct = Math.cos(Math.toRadians(theta));
		double st = Math.sin(Math.toRadians(theta));
		int nBin = (int)Math.ceil(2*radius*10);
		double rMin = -radius;
		double rMax = radius;
		double delta = (rMax-rMin)/nBin;
		for (int iBin=0; iBin<=nBin; iBin++) {
		//for (double rad = -radius; rad < radius; rad += 0.1) {
			
					
			double rad = rMin + iBin*delta;
			double r = row + rad*st;
			double c = col + rad*ct;
			try {
				int ir = (int) Math.floor(r);
				int ic = (int) Math.floor(c);
				double flux = getFluxInSourceObject(ir,ic);
				/*
				if (Math.abs(Math.abs(rad)-19.2) < 0.001) {
					//System.out.println("source object rad="+rad+" r="+r+" c="+c+" ir="+ir+" ic="+ic+" flux="+flux);
					System.out.println(String.format("rad=%+.2f r=%6.2f c=%6.2f ir=%3d ic=%3d flux=%8.3f",rad, r, c, ir, ic, flux));
				}
				*/
				if (!doLogarithmicRange || flux > 0) {
					xySourceObject.add(rad,flux);
				}
			} catch (IllegalArgumentException e) {
			}
			for (int i=0; i<detectorObjects.size(); i++) {
				double flux = detectorObjects.get(i).getFlux(r, c);
				//System.out.println("detectorObjs  rad="+rad+" r="+r+" c="+c+" flux="+flux);
				if (!doLogarithmicRange || flux > 0) {
					xyDetectorObjects[i].add(rad,flux);
				}
			}
		}
		JFreeChart retval = JFreeChartUtil.makeChart(xysc, title+String.format(" chi2=%.2f",sourceObject.getFitChi2()), "radius", "flux (electrons/pixel)", null);
		JFreeChartUtil.linesShapesVisible(retval, true, false);
		if (doLogarithmicRange) {
			JFreeChartUtil.setLogarithmicRangeAxis(retval);
		}
		return retval;
	}
	/**
	 * @return the source object of this composite
	 */
	public SourceObject getDetectorSourceObject() {
		return sourceObject;
	}
	
	/**
	 * @return the DetectorObjects of this composite
	 */
	public ArrayList<DetectorObject> getDetectorObjects() {
		return detectorObjects;
	}
	
	/**
	 * @return the radius that contains half the flux, interpolted between the two closest integer values
	 */
	public double getHalfLightRadius() {
		if (halfLightRadius == null) {
			calculateHalfLightRadius();
		}
		return halfLightRadius;
	}
	private void calculateHalfLightRadius() {
		double flux = sourceObject.getFlux();
		for (DetectorObject detectorObject : detectorObjects) {
			flux += detectorObject.getNPhoton();
		}
		System.out.println("flux="+flux);
		double containedRatio = 0.0;
		double radius = 1.0;
		while (containedRatio < 0.5) {
			containedRatio = getApertureFlux(radius, 0.0);
		}
	}
	
}
