package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.shapelets.functions.BasisFunction;
import gov.fnal.eag.shapelets.functions.GaussConvolve;
import gov.fnal.eag.sim.catalog.Morphology;
import gov.fnal.eag.sim.catalog.MorphologyShapelet;
import gov.fnal.eag.wcs.Coordinate;
import gov.fnal.eag.wcs.SphericalCoordinate;
import gov.fnal.eag.wcs.Transform;

import java.util.ArrayList;

import uk.ac.starlink.table.RowListStarTable;

/**
 * keep track of shape and distortions, but don't do only do real work in the makePixeledImage method
 * @author stoughto
 *
 */
public class DetectorObjectProduct extends DetectorObject {
	Transform pixelToSky;
	Morphology morphology;
	ArrayList<PSF> psfs;
	/**
	 * @param skyCoordinate
	 * @param pixelToSky
	 * @param nPhoton
	 * @param morphology
	 * @param psfs
	 */
	public DetectorObjectProduct(SphericalCoordinate skyCoordinate, Transform pixelToSky, double nPhoton, Morphology morphology, ArrayList<PSF> psfs) {
		Coordinate coordinate = pixelToSky.inverse(skyCoordinate);
		row = coordinate.getY();
		col = coordinate.getX();
		this.pixelToSky = pixelToSky;
		this.nPhoton = nPhoton;
		this.morphology = morphology;
		this.psfs = psfs;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#addToTable(uk.ac.starlink.table.RowListStarTable)
	 */
	@Override
	public void addToTable(RowListStarTable rlst) {
		if (!rlst.getName().equals("detectorObjectSurfaceBrightness")) return;
		if (morphology instanceof MorphologyShapelet) {
			double[] ev1 = ((MorphologyShapelet) morphology).getEigenvalues();
			double scale = morphology.getSize();
			int nOrder = BasisFunction.dimensionFromLength(ev1.length);
			Object[] temp = new Object[rlst.getColumnCount()];
			temp[0] = new Double(row);
			temp[1] = new Double(col);
			temp[2] = new Double(nPhoton);
			temp[3] = new Double(scale);
			temp[4] = new Integer(nOrder);
			temp[5] = new Double(ra);
			temp[6] = new Double(dec);
			temp[7] = fromTable;
			int length = BasisFunction.lengthFromDimension(BasisFunction.getNMax()+1);
			if (ev1.length > length) {
				throw new Error("ev1 length="+ev1.length+" is greater than length="+length);
			}
			double value;
			for (int i=8; i<rlst.getColumnCount(); i++) {
				try {
					value = ev1[i-8];
				} catch (ArrayIndexOutOfBoundsException e) {
					value = 0;
				}
				temp[i] = new Double(value);
			}
			rlst.addRow(temp);        
		}
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#convolveGaussian(double)
	 */
	@Override
	public void convolveGaussian(double d) {
		throw new UnsupportedOperationException();
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#gc()
	 */
	@Override
	public void gc() {
		throw new UnsupportedOperationException();
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#getFluxInPixel(int, int)
	 */
	@Override
	public double getFluxInPixel(int iRow, int iCol) {
		throw new UnsupportedOperationException();
	}


	/**
	 * actually render this object
	 * @param countsPerPixelThreshold
	 * @param highResolution ignored for now as only the low-resolution version is implemented
	 * @return a pixeled image that is large enough to have fewer than countsPerPixelThreshold counts in each of the pixels on the edge
	 */
	public PixeledImage makePixeledImage(double countsPerPixelThreshold, boolean highResolution) {
		if (morphology instanceof MorphologyShapelet) {
			double ev[] = ((MorphologyShapelet)morphology).getEigenvalues();
			double scaleArcsec = morphology.getSize();

			// Deal with the PSFs
			PixeledImage resampledPsf = null;		
			for (PSF psf : psfs) {
				if (psf instanceof PSFPixeledImage) {
					PixeledImage temp = ((PSFPixeledImage) psf).resample(pixelToSky.arcsecPerPixel()); 
					if (resampledPsf == null) {
						resampledPsf = temp;
					} else {
						resampledPsf = resampledPsf.convolveJnt(temp);
					}
				} else if (psf instanceof PSFGaussian) {
					double gaussSigmaArcsec = ((PSFGaussian) psf).sigma;
					ev = GaussConvolve.convolve(scaleArcsec, ev, gaussSigmaArcsec);
					scaleArcsec = GaussConvolve.getGamma(scaleArcsec, gaussSigmaArcsec);
				}
			}
			double scalePixel = scaleArcsec / pixelToSky.arcsecPerPixel();	
			DetectorObjectSurfaceBrightness dosb = new DetectorObjectSurfaceBrightness(row, col, nPhoton, ev, scalePixel);
			double fluxNorm = nPhoton / (2.0 * Math.PI * Math.pow(scalePixel,2));
			PixeledImage retval = null;
			if (countsPerPixelThreshold < fluxNorm) {
				double ratio = countsPerPixelThreshold/fluxNorm;
				double radiusFactor = 3.0;
				int radius = (int) Math.ceil(radiusFactor*Math.sqrt(-2.0*Math.pow(scalePixel,2)*Math.log(ratio)));
				double maxCountsPerPixelInEdge = Double.MAX_VALUE;
				while (maxCountsPerPixelInEdge > countsPerPixelThreshold) {
					PixeledImage pi = dosb.makePixeledImage(2*radius, false);
					if (resampledPsf != null) {
						retval = pi.convolveJnt(resampledPsf);
					} else {
						retval = pi;
					}
					maxCountsPerPixelInEdge = retval.getMaxCountsPerPixelInEdge();
					radius *= 2;
				}
				if (retval != null) {
					pixelToSky.toHeader(retval.header);
				}
			}
			return retval;
		}
		throw new Error("Can not deal with morphology="+morphology);
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#getFlux(double, double)
	 */
	@Override
	public double getFlux(double rowIn, double colIn) {
		throw new UnsupportedOperationException();
	}
}
