package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.util.astronomy.Integrate;
import gov.fnal.eag.util.astronomy.SimpleFunction;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.util.Hashtable;

import uk.ac.starlink.table.ColumnInfo;
import uk.ac.starlink.table.RowListStarTable;

import cern.jet.stat.Probability;

import com.imsl.math.Sfun;

/**
 * profile implementation with ellipticity and position angle
 */
public class DetectorObjectProfile extends DetectorObject {

	/**
	 * returns the sigma for an object created with this shape and scale=1.0
	 */
	public static final Hashtable<String,Double> sigmaFactor;

	private static double devaucIntegral;
	static {
		sigmaFactor = new Hashtable<String,Double>();
		sigmaFactor.put("gauss", 1.0);
		sigmaFactor.put("exp", 1.7322);
		sigmaFactor.put("devauc", 3.1077);

		SimpleFunction func1 = new SimpleFunction() // object of anon Function subclass
		{
			@Override
			public double f(double x)
			{
				return x*Math.exp(-7.67*Math.pow(x, 0.25));
			}
		};
		double a=0.0;
		double b=100.0;
		devaucIntegral = 2.0 * Math.PI * Integrate.qRomb(func1, a, b, 1e-9);
	}
	private int nGridPoints = 5;
	protected String shape;
	/**
	 * normalization for profiles to get flux per pixel
	 */
	protected double fluxNorm;
	/**
	 * the size of this object in pixels
	 */
	protected double scale;
	/**
	 * ellipticity of this object = 1 - B/A
	 */
	protected double ellipticity = 0.0;
	/**
	 * rotation of this object.
	 */
	protected double positionAngle = 0.0;
	/**
	 * catalan number used to normalize the sech profile
	 */
	private double catalan = 0.9159655942;
	/**
	 * cos and sin of position angle
	 */
	private double cpa = 1.0;
	private double spa = 0.0;
	/**
	 * Profiles for this object;
	 */
	private Profiles profiles;

	private static Profiles staticProfiles = null;

	/**
	 * fully-specified constructor
	 * 
	 * @param row --
	 *            row position
	 * @param col --
	 *            column position
	 * @param nPhoton --
	 *            number of photons
	 * @param shape --
	 *            shape, one of deVauc, exp, or gauss, or sech
	 * @param scale --
	 *            size the object in pixels
	 * @param ellipticity --
	 *            1 - b/a
	 * @param positionAngle --
	 *            in degrees
	 * @param profiles 
	 */
	public DetectorObjectProfile(double row, double col, double nPhoton,
			String shape, double scale, double ellipticity, double positionAngle, Profiles profiles) {
		if (
				!shape.equalsIgnoreCase("point") &&
				!shape.equalsIgnoreCase("gauss") &&
				!shape.equalsIgnoreCase("exp") &&
				!shape.equalsIgnoreCase("devauc") &&
				!shape.equalsIgnoreCase("sech")
		) {
			throw new IllegalArgumentException("can not deal with shape="+shape);
		}

		setRow(row);
		setCol(col);
		this.isSimple = true;
		this.shape = shape;
		this.scale = scale;
		this.ellipticity = ellipticity;
		this.positionAngle = positionAngle;
		this.profiles = profiles;
		cpa = Math.cos(positionAngle * Math.PI / 180.0);
		spa = Math.sin(positionAngle * Math.PI / 180.0);
		setNPhoton(nPhoton);
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#setRow(double)
	 */
	@Override
	public void setRow(double row) {this.row = row;}
	/** 
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#setCol(double)
	 */
	@Override
	public void setCol(double col) {this.col = col;}
	/** 
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#setNPhoton(double)
	 */
	@Override
	public void setNPhoton(double nPhoton) {
		// This is just the normalization including ellipticity
		this.nPhoton = nPhoton;
		if (shape.equalsIgnoreCase("point")) {
			fluxNorm = nPhoton;
		} else if (shape.equalsIgnoreCase("sech")) { // the normalized sech profile
			fluxNorm = nPhoton / (2.0 * Math.PI * 2.0*catalan * scale * scale) / (1.0-ellipticity);
		} else if (shape.equalsIgnoreCase("gauss") || shape.equalsIgnoreCase("exp")) {
			fluxNorm = nPhoton / (2.0 * Math.PI * scale * scale) / (1.0-ellipticity);
		} else if (shape.equalsIgnoreCase("devauc")){
			fluxNorm = nPhoton / (devaucIntegral * scale * scale) /  (1.0-ellipticity);
		} else {
			throw new Error("shape="+shape);
		}
	}
	/**
	 * create a point source
	 * 
	 * @param row row position
	 * @param col column position
	 * @param nPhoton mean number of photons
	 */
	public DetectorObjectProfile(double row, double col, double nPhoton) {
		setRow(row);
		setCol(col);
		this.isSimple = true;
		shape = "point";
		setNPhoton(nPhoton);
	}

	/**
	 * define a bounding box for rendering this object
	 * @param photonsPerPixelThreshold the minimum number of photons per pixel you want to render
	 * @return the rectangle that bounds the interesting pixels
	 *
	 */
	public Rectangle getLimits(double photonsPerPixelThreshold) {
		double radius = -1;
		Rectangle limits = null;
		if (photonsPerPixelThreshold < fluxNorm) {
			if (shape.equalsIgnoreCase("gauss")) {
				if (photonsPerPixelThreshold > 0) {
					radius = scale * Math.sqrt(-2.0	* Math.log(photonsPerPixelThreshold	/ fluxNorm));
				} else {
					radius = scale*1e3;
				}
			} else if (shape.equalsIgnoreCase("exp")) {
				radius = -scale * Math.log(photonsPerPixelThreshold / fluxNorm);
			} else if (shape.equalsIgnoreCase("devauc")) {
				radius = scale
				* Math.pow(-Math.log(photonsPerPixelThreshold/fluxNorm) / 7.67, 4);
			} else if (shape.equalsIgnoreCase("sech")) {
				// since sech == 1/cosh, we invert photonsPerPixelThreshold and fluxNorm
				radius = scale * Sfun.acosh(fluxNorm / photonsPerPixelThreshold);
			} else if (shape.equalsIgnoreCase("point")){
				int diameter = 2;

				int ir = (int) Math.floor(row-0.5);
				int ic = (int) Math.floor(col-0.5);
				limits = new Rectangle(new Point(ic, ir), new Dimension(diameter, diameter));
			} else {
				throw new Error("Can not deal with shape="+shape);
			}
			if (limits == null) {
				int ir = (int) Math.floor(row-radius);
				int ic = (int) Math.floor(col-radius);
				int diameter = (int) Math.ceil(2*radius+1);
				//Dimension d = new Dimension(diameter, diameter);
				limits = new Rectangle(new Point(ic, ir), new Dimension(diameter, diameter));
			}
		} else {
			// Not enough light per pixel to worry about this object

			int diameter = 0;
			int ir = (int) Math.floor(row-0.5);
			int ic = (int) Math.floor(col-0.5);
			limits = new Rectangle(new Point(ic, ir), new Dimension(diameter, diameter));			
		}
		return limits;
	}
	/**
	 * gets the location of the next photon from this object
	 * @return the location
	 */
	public Point2D.Double getNextLocation() {
		if (shape.equals("point")) {
			return new Point2D.Double(col,row);
		}
		double rho = profiles.rho(shape, scale);

		// Choose an angle uniformly
		double theta  = 2*Math.PI*profiles.next();

		// Transform to the ellipse
		double xEllip = rho*Math.cos(theta)*(1.0-ellipticity);
		double yEllip = rho*Math.sin(theta);

		// Rotate this ellipse by the position angle
		Point2D.Double point = new Point2D.Double(col +  xEllip*spa + yEllip*cpa, row +  xEllip*cpa - yEllip*spa);

		return point;

	}
	/**
	 * @param r
	 * @param c
	 * @return the effective distance from the center of the object
	 */
	public double getRho(double r, double c) {
		double dr = r - row;
		double dc = c - col;
		double dcp =   dc * cpa + dr * spa;
		double drp =  -dc * spa + dr * cpa;
		double rho = Math.sqrt(Math.pow(drp/(1.0-ellipticity),2) + Math.pow(dcp, 2))/scale;
		return rho;
	}
	/**
	 * returns the flux per pixel at position (r,c) in the ccd from this object, used for bright objects
	 * 
	 * @param r row position
	 * @param c column position
	 * @return double flux in the pixel
	 */
	@Override
	public double getFlux(double r, double c) {
		double flux;
		double dr = r - row;
		double dc = c - col;
		if (shape.equalsIgnoreCase("point")) {
			if ((dr >= -1.0) && (dr < 1.0) && (dc >= -1.0) && (dc < 1.0)) {
				flux = fluxNorm * (1.0 - Math.abs(dr)) * (1.0 - Math.abs(dc));
				return flux;
			} 
			return 0.0;

		}
		double rho = getRho(r,c);
		if (shape.equalsIgnoreCase("gauss")) {
			flux = fluxNorm * Math.exp(-0.5 * rho * rho);
			return flux;
		} else if (shape.equalsIgnoreCase("exp")) {
			flux = fluxNorm * Math.exp(-rho);
			return flux;
		} else if (shape.equalsIgnoreCase("sech")) {
			flux = fluxNorm * 1.0/Math.cosh(rho);
			return flux;	
		} else  {
			// This must be devauc
			flux = fluxNorm * Math.exp(-7.67 * Math.pow(rho, 0.25));
			return flux;
		}
	}
	/**
	 * get the flux normalization
	 * @return the flux normalization
	 */
	public double getFluxNorm() {
		return fluxNorm;
	}
	/**
	 * get the integrated flux
	 * @param row0 lower corner row
	 * @param col0 lower corner column
	 * @param row1 upper corner row
	 * @param col1 upper corner column
	 * @param nGrid number of grid points per dimension for sampling
	 * @return the integrated flux
	 */

	public double getIntegratedFlux(double row0, double col0, double row1, double col1, int nGrid) {
		double integratedFlux = 0;
		double flux;
		if (shape.equals("point")) {
			for (int ir=(int) Math.floor(row0); ir < Math.floor(row1); ir++) {
				for (int ic=(int) Math.floor(col0); ic<Math.floor(col1); ic++) {
					flux = getFlux(ir+0.5, ic+0.5);
					integratedFlux += flux;
				}
			}
		} else {
			if (nGrid == 0) {
				if (ellipticity != 0.0 || !shape.equals("gauss")) {
					throw new IllegalArgumentException("for nGrid=0 can not do ellipticity="+ellipticity+" shape="+shape);
				}
				double k = 1.0/scale/Math.sqrt(2.0);
				double intRow = Probability.errorFunction(k*(row1-row)) - Probability.errorFunction(k*(row0-row));
				double intCol = Probability.errorFunction(k*(col1-col)) - Probability.errorFunction(k*(col0-col));
				integratedFlux = nPhoton * intRow * intCol / 4;
			} else {
				double dr = (row1-row0) / (nGrid+1);
				double dc = (col1-col0) / (nGrid+1);
				for (int ir=0; ir<nGrid; ir++) {
					for (int ic=0; ic<nGrid; ic++) {
						double r = row0+(ir+1)*dr;
						double c = col0+(ic+1)*dc;
						flux = getFlux(r,c);
						integratedFlux += flux;
					}
				}
				integratedFlux /= nGrid*nGrid;
			}
		}
		return integratedFlux;
	}

	/**
	 * gets the flux in a pixel
	 * @param ir row pixel location
	 * @param ic column pixel location
	 * @param nGridBright number of grid points per dimension for sampling
	 * @return the flux
	 */
	public double getFluxByPix(int ir, int ic, int nGridBright) {
		return getIntegratedFlux(ir, ic, ir+1, ic+1, nGridBright);
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#getFluxInPixel(int, int)
	 */
	@Override
	public double getFluxInPixel(int ir, int ic) {		
		double retval = getIntegratedFlux(ir, ic, ir+1, ic+1, nGridPoints);
		return retval;
	}
	/**
	 * convolves this object
	 * @param psf PSF to use for convolution
	 * @param limits bounding box on the object
	 * @param nGrid number of grid points per dimension for sampling
	 * @param rowCol0 position of the min row,col of the result, filled in by this method
	 * @return the convolved image 
	 */

	public double[][] convolve(PSF psf, Rectangle limits, int nGrid, Point rowCol0) {
		int psfColRadius = psf.ncol()/2;
		int psfRowRadius = psf.nrow()/2;
		int rowMin = (int) (Math.floor(limits.getMinY())) - psfRowRadius;
		int rowMax = (int) (Math.ceil(limits.getMaxY())) + psfRowRadius;
		int colMin = (int) (Math.floor(limits.getMinX())) - psfColRadius;
		int colMax = (int) (Math.ceil(limits.getMaxX())) + psfColRadius;
		int rowSize = rowMax-rowMin;
		int colSize = colMax-colMin;
		rowCol0.x = colMin;
		rowCol0.y = rowMin;

		PixeledImage source = new PixeledImage(rowSize, colSize);
		source.setRow0Col0(rowMin, colMin);
		for (int r=source.getRowMin(); r<source.getRowMax(); r++) {
			for (int c=source.getColumnMin(); c<source.getColumnMax(); c++) {
				double flux = getFluxByPix(r, c, nGrid);
				source.setByRowCol(r, c, flux);
			}
		}
		// TODO check the JUnint tests to see if they are complete and correct
		psf.convolve(source);
		return source.getPix();
	}
	/*
	public double[][] convolveDeleteThisMethod(PSF psf, Rectangle limits, int nGrid, Point rowCol0) {
		int rowMin = (int) (Math.floor(limits.getMinX()) - psf.nrow());
		int rowMax = (int) (Math.ceil(limits.getMaxX()) + psf.nrow());
		int colMin = (int) (Math.floor(limits.getMinY()) - psf.ncol());
		int colMax = (int) (Math.ceil(limits.getMaxY()) + psf.ncol());
		int rowSize = rowMax-rowMin;
		int colSize = colMax-colMin;
		rowCol0.x = rowMin;
		rowCol0.y = colMin;
		int nElement = rowSize*colSize;
		float[] sourceFloat = new float[nElement];
		int iElement = 0;		
		for (int col=colMin; col<colMax; col++) {
			for (int row=rowMin; row<rowMax; row++) {
				float flux = (float) getFluxByPix(row, col, nGrid);
				//if (flux > 0) {
					//System.out.println("row="+row+"  col="+col+"   flux="+flux+"  iElement="+iElement);
				//}
				sourceFloat[iElement++] = flux;
			}
		}

		DataBufferFloat dbuffer = new DataBufferFloat(sourceFloat, sourceFloat.length);
		SampleModel sampleModel = RasterFactory.createBandedSampleModel(DataBuffer.TYPE_FLOAT, rowSize, colSize, 1);
		ColorModel colorModel = PlanarImage.createColorModel(sampleModel);
		Raster raster = RasterFactory.createRaster(sampleModel, dbuffer, new java.awt.Point(0, 0));
		TiledImage tiledImage = new TiledImage(0, 0, rowSize, colSize, 0, 0, sampleModel, colorModel);
		tiledImage.setData(raster);
		PlanarImage output = JAI.create("convolve", tiledImage, psf.getKernel());
		Raster rasterOut = output.getData();
		double[] temp = new double[1];
		double[][] pix = new double[rowSize][colSize];
		for (int r = 0; r < pix.length; r++) {
			for (int c = 0; c < pix[0].length; c++) {
				temp = rasterOut.getPixel(r, c, temp);
				pix[r][c] = temp[0];
			}
		}		
		return pix;
	}
	 */

	/**
	 * make a table to keep track of this kind of object
	 * @return the table
	 */
	public static RowListStarTable createRowListStarTable() {
		ColumnInfo[] colInfoArray = new ColumnInfo[10];
		colInfoArray[0] = new ColumnInfo("row", Double.class, "row position pixels");
		colInfoArray[1] = new ColumnInfo("col", Double.class, "column position in pixels");
		colInfoArray[2] = new ColumnInfo("nPhoton", Double.class, "number of photons");
		colInfoArray[3] = new ColumnInfo("shape", String.class, "object shape");
		colInfoArray[4] = new ColumnInfo("scale", Double.class, "size in pixels");
		colInfoArray[5] = new ColumnInfo("ellipticity", Double.class, "1 - b/a");
		colInfoArray[6] = new ColumnInfo("positionAngle", Double.class, "degrees");   
		colInfoArray[7] = new ColumnInfo("ra", Double.class, "true right ascension (degrees");
		colInfoArray[8] = new ColumnInfo("dec", Double.class, "true declination (degrees)");
		colInfoArray[9] = new ColumnInfo("fromTable", String.class, "truth table from which this object was taken");

		RowListStarTable rlst = new RowListStarTable(colInfoArray);
		rlst.setName("detectorObject");
		return rlst;
	}

	/** 
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#addToTable(uk.ac.starlink.table.RowListStarTable)
	 */
	@Override
	public void addToTable(RowListStarTable rlst) {
		if (!rlst.getName().equals("detectorObject")) return;
		Object[] temp = {
				new Double(row),
				new Double(col),
				new Double(nPhoton),
				shape,
				new Double(scale),
				new Double(ellipticity),
				new Double(positionAngle),
				new Double(ra),
				new Double(dec),
				fromTable};
		rlst.addRow(temp);        
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#isSimple()
	 */
	@Override
	public boolean isSimple() {
		return isSimple;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#getNPhoton()
	 */
	@Override
	public double getNPhoton() {
		return nPhoton;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#getRow()
	 */
	@Override
	public double getRow() {
		return row;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#getCol()
	 */
	@Override
	public double getCol() {
		return col;
	}


	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#setRaDec(double, double)
	 */
	@Override
	public void setRaDec(double ra, double dec) {
		setRaDec(ra,dec,"unknown");		
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#setRaDec(double, double, java.lang.String)
	 */
	@Override
	public void setRaDec(double ra, double dec, String fromTable) {
		this.ra = ra;
		this.dec = dec;
		this.fromTable = fromTable;		
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#getRa()
	 */
	@Override
	public double getRa() {
		return ra;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#getDec()
	 */
	@Override
	public double getDec() {
		return dec;
	}

	/**
	 * if the object is a simple point, make it a gaussian
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#convolveGaussian(double)
	 */
	@Override
	public void convolveGaussian(double d) {
		if (shape.equalsIgnoreCase("point")) {
			shape = "gauss";
			scale = d;
			if (staticProfiles == null) staticProfiles = new Profiles();
			setNPhoton(nPhoton);
			return;
		}
		throw new Error(" DetectorObjectProfile:  need to implement convolveGaussian");
	}

	/**
	 * make a small PixeledImage of this object
	 * @param pixelDiameter number of rows and the number of columns
	 * @param nGridBright number of points to sample in each direction
	 * @return the pixeledImage
	 */
	public PixeledImage makePixeledImage(int pixelDiameter, int nGridBright) {
		int row0 = (int) (row-pixelDiameter/2);
		int col0 = (int) (col-pixelDiameter/2);
		PixeledImage pi = new PixeledImage(pixelDiameter, pixelDiameter);
		pi.setRow0Col0(row0, col0);
		for (int r=row0; r<row0+pixelDiameter; r++) {
			for (int c=col0; c<col0+pixelDiameter; c++) {
				double value = getFluxByPix(r, c, nGridBright);
				pi.setByRowCol(r, c, value);
			}
		}
		return pi;
	}

	/**
	 * increments the pixel values to add this object
	 * @param pi 
	 * @param nGridBright grid size to sample inside pixel
	 * 
	 * This is a convenient way to add one object to an image, but it is not an efficient way to add many objects to an image
	 */
	public void renderToPixeledImage(PixeledImage pi, int nGridBright) {
		for (int r=pi.getRowMin(); r < pi.getRowMax(); r++) {
			for (int c=pi.getColumnMin(); c<pi.getColumnMax(); c++) {
				double value = getFluxByPix(r, c, nGridBright);
				pi.incrByRowCol(r, c, value);
			}
		}
	}
	/**
	 * 
	 * @param args ignored
	 * @throws Exception 
	 */
	public static void main(String[] args) throws Exception {
		double trueRow = 100.5;
		double trueCol = 200.5;
		double trueSigma = 3.0;
		DetectorObjectProfile dos = new DetectorObjectProfile(trueRow, trueCol, 1000);
		Rectangle limits = dos.getLimits(1.0);
		PSF psf = new PSFGaussian(trueSigma);
		Point rowCol0 = new Point();
		double[][] convolved = dos.convolve(psf, limits, 1, rowCol0);
		System.out.println("trueRow="+trueRow+" trueCol="+trueCol+" row0="+rowCol0.x+"   col0="+rowCol0.y);
		//DoubleArrayList dal = new DoubleArrayList();
		PixeledImage pi = new PixeledImage(convolved);
		//pi.printAll(System.out, "convolved pi in doTestConvolvePoint");
		//ir.putPixeledImage("aaa", pi, true);
		Hashtable<String, Number> centroidSimple = pi.centroid(0.0);
		double expectedRowCentroid = trueRow - rowCol0.y;
		double actualRowCentroid = new Double(centroidSimple.get("rowCentroid").toString()).doubleValue();
		double expectedColCentroid = trueCol - rowCol0.x;
		double actualColCentroid = new Double(centroidSimple.get("colCentroid").toString()).doubleValue();
		System.out.println("expectedRowCentroid="+expectedRowCentroid+"  actualRowCentroid="+actualRowCentroid);
		System.out.println("expectedColCentroid="+expectedColCentroid+"  actualColCentroid="+actualColCentroid);

	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#makePixeledImage(int)
	 */
	@Override
	public PixeledImage makePixeledImage(int pixelDiameter) {
		int row0 = (int) (row-pixelDiameter/2);
		int col0 = (int) (col-pixelDiameter/2);
		PixeledImage pi = new PixeledImage(pixelDiameter, pixelDiameter);
		pi.setRow0Col0(row0, col0);
		for (int r=row0; r<row0+pixelDiameter; r++) {
			for (int c=col0; c<col0+pixelDiameter; c++) {
				double value = getFluxByPix(r, c, nGridPoints);
				pi.setByRowCol(r, c, value);
			}
		}
		return pi;
	}

	/**
	 * @param nGridPoints the number of grid points used in makePixeledImage when calling getGluxByPix
	 */
	public void setNGridPoints(int nGridPoints) {
		this.nGridPoints = nGridPoints;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#gc()
	 */
	@Override
	public void gc() {
	}

	/**
	 * 
	 * @return the shape that defines this profile
	 */
	public String getShape() {
		return shape;
	}
}