package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.shapelets.functions.BasisFunction;
import gov.fnal.eag.shapelets.functions.Convolution;
import gov.fnal.eag.shapelets.functions.GaussConvolve;
import gov.fnal.eag.shapelets.functions.Hermite;
import gov.fnal.eag.shapelets.functions.Phi;
import gov.fnal.eag.shapelets.functions.ShapeletsTransform;
import gov.fnal.eag.wcs.Coordinate;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;

import uk.ac.starlink.table.ColumnInfo;
import uk.ac.starlink.table.RowListStarTable;

import com.imsl.math.Complex;


/**
 * shapelet implementation translated from the IDL of Massey et al.
 */
public class DetectorObjectSurfaceBrightness extends DetectorObject {
	/**
	 * x shapelet centroid shift
	 */
	double dX = 0.;
	/**
	 * y shapelet centroid shift
	 */
	double dY = 0.;
	/**
	 *  the row location on pixel detector corrected for the shift
	 */
	double rCor = 0.;
	/**
	 *  the col location on pixel detector corrected for the shift
	 */
	double cCor = 0.;
	
	/**
	 * buffer to hold a description of this object if is defined with a PixeledImage
	 */
	PixeledImage pixeledImage = null;
	/**
     * the eigenvalues of this object
	 */
	protected double[][] eigenvalues;
	/**
	 *  The vector form of eigenvalues
	 */
	protected double[] ev1;
	/**
	 * scale of this object, in pixels
	 */
	protected double scale;

	/**
	 * integrated flux from the raw eigenvectors
	 */
	protected double normalization;
	/**
	 * dimension of the 2-d matrix for the object
	 */
	protected int nOrder;
	
	private boolean[] xFastDefined;
	private boolean[] yFastDefined;
	private double[][] xFastPhiEvaluated;
	private double[][] yFastPhiEvaluated;
	
	private boolean[] xFastIntegrateDefined;
	private boolean[] yFastIntegrateDefined;
	private double[][] xFastIntegrateEvaluated;
	private double[][] yFastIntegrateEvaluated;
	
	
	private int nFastScale = 10;
	private int xFastMin;
	private int yFastMin;
	
	/**
	 * constructor with a pixeled image
	 * @param pi defines the shape of the object
	 */
	public DetectorObjectSurfaceBrightness(PixeledImage pi) {
		double tempRow = pi.getRow0()+0.5*pi.getRowCount();
		double tempCol = pi.getCol0()+0.5*pi.getColumnCount();
		initPi(pi, tempRow, tempCol, 0);
	}

	/**
	 * constructor with pixeled image
	 * @param pi defines the shape of the object
	 * @param row pixel position
	 * @param col column position
	 * @param nPhoton number of photons
	 */
	public DetectorObjectSurfaceBrightness(PixeledImage pi, double row, double col, double nPhoton) {
		initPi(pi, row, col, nPhoton);
	}
	/**
	 * constructor with pixeled image
	 * @param pi defines the shape and flux of the object
	 * @param row pixel position
	 * @param col pixel position
	 */
	public DetectorObjectSurfaceBrightness(PixeledImage pi, double row, double col) {
		initPi(pi, row, col, 0.0);
	}
	
	private void initPi(PixeledImage pi, double rowIn, double colIn, double nPhotonPi) {
		pixeledImage = pi;
		double piSum = pi.getSum();
		if (nPhotonPi == 0.0) {
			nPhoton = piSum;
			normalization = 1.0;
		} else {
			nPhoton = nPhotonPi;
			normalization = piSum/nPhotonPi;
		}
		this.isSimple = false;		
		this.row = rowIn;
		this.col = colIn;
		this.rCor = rowIn;
		this.cCor = colIn;
		
		int row0 = (int)Math.floor(row)-pi.getRowCount()/2;
		int col0 = (int)Math.floor(col)-pi.getColumnCount()/2;		
		pi.setRow0Col0(row0,col0);
	}
	/**
	 * fully-specified constructor
	 * 
	 * @param row row position
	 * @param col column position
	 * @param nPhoton number of photons
	 * @param eigenvalues shapelet eigenvalues
	 * @param scale size of the object in pixels
	 */
	public DetectorObjectSurfaceBrightness(double row, double col, double nPhoton,
			double[][] eigenvalues, double scale) {
		init(row, col, nPhoton, eigenvalues, scale);
	}
	
	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#setRow(double)
	 */
	@Override
	public void setRow(double row) {
		super.setRow(row);
		init(row, col, nPhoton, eigenvalues, scale);
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#setCol(double)
	 */
	@Override
	public void setCol(double col) {
		super.setCol(col);
		init(row, col, nPhoton, eigenvalues, scale);
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#setNPhoton(double)
	 */
	@Override
	public void setNPhoton(double nPhoton) {
		super.setNPhoton(nPhoton);
		if (pixeledImage == null) {
			normalization = 1.0; // getIntegratedFlux uses this, so set it to 1.0 until it is calculated
			double row0 = row-100*scale;
			double col0 = col-100*scale;
			double row1 = row+100*scale;
			double col1 = col+100*scale;
			normalization = getIntegratedFlux(row0, col0, row1, col1)/nPhoton;		
			if (Math.abs(normalization) < 10*Double.MIN_VALUE) {
				// When rendering pure shapelets, the positive and negative fluxes can cancel.
				normalization = nPhoton;				
			}
		} else {
			normalization = pixeledImage.getSum()/nPhoton;
		}
		
	}
	
	/**
	 * @return the scale of this object in pixels
	 */
	public double getScale() {
		return scale;
	}
	/**
	 * @return the eigenvalues
	 */
	public double[][] getEigenvalues() {
		return eigenvalues;
	}
	/**
	 * In this init method we have calculated corrected center of the object
	 * and make compensation for the centroid shift introduced by Erin's bad
	 * decomposition. Attantion, in the correction we have switched row and col
	 * centroids as Erins decomposition was done in Fortran where assignment of raws
	 *  and cols is opposit to C++ and Java.
	 *  
	 *   By N. Kuropatkin 07/02/2007
	 *   
	 * @param rowIn
	 * @param colIn
	 * @param nPhotonIn
	 * @param eigenvaluesIn
	 * @param scaleIn
	 * 
	 * 
	 */
	private void init(double rowIn, double colIn, double nPhotonIn, double[][]eigenvaluesIn, double scaleIn)  {
		nOrder = eigenvaluesIn[0].length;
		this.row = rowIn;
		this.col = colIn;
		this.nPhoton = nPhotonIn;
		this.isSimple = false;
		this.eigenvalues = eigenvaluesIn;
		this.scale = scaleIn;
		ev1 = BasisFunction.tab2Vector(this.eigenvalues);
		double[] moments = BasisFunction.ImageAnalyser(scale,eigenvalues);
		setDxDy(new Coordinate(moments[1], moments[2]));
	}
	
	/**
	 * @return the offset required to move the centroid of the flux to the row,col position
	 */
	public Coordinate getDxDy() {
		Coordinate retval = new Coordinate(dX, dY);
		return retval;
	}
	
	/**
	 * @param coordinate set the offset
	 */
	public void setDxDy(Coordinate coordinate) {
		this.dX = coordinate.getX();
		this.dY = coordinate.getY();
		this.cCor = col - dY;
		if (Double.isNaN(cCor)) throw new Error("cCor is NaN:  col="+col+" dY="+dY);
		this.rCor = row - dX;
		if (Double.isNaN(rCor)) throw new Error("rCor is NaN:  row="+row+" dX="+dX);
		setNPhoton(nPhoton);		
		int nbin = 2*((int) Math.ceil(scale*nFastScale));
		initFast(nbin);		
	}
	
	private void initFast(int nbin) {
		if (nbin > 0) {
			xFastDefined = new boolean[nbin];
			yFastDefined = new boolean[nbin];
			xFastPhiEvaluated = new double[nOrder][nbin];
			yFastPhiEvaluated = new double[nOrder][nbin];
			xFastIntegrateDefined = new boolean[nbin];
			yFastIntegrateDefined = new boolean[nbin];
			xFastIntegrateEvaluated = new double[nOrder][nbin];
			yFastIntegrateEvaluated = new double[nOrder][nbin];
			xFastMin = (int) Math.floor(col - scale*nFastScale);
			yFastMin = (int) Math.floor(row - scale*nFastScale);
		} else {
			throw new Error("scale="+scale+" nFastScale="+nFastScale+" nbin="+nbin);
		}
	}

	/**
	 * fully-specified constructor with vector version of eigenvalues
	 * 
	 * @param row row position
	 * @param col column position
	 * @param nPhoton number of photons from this object
	 * @param ev1 shapelet eigenvalues in form of vector
	 * @param scale size the object in pixels
	 */
	public DetectorObjectSurfaceBrightness(double row, double col, double nPhoton,
			double[] ev1, double scale) {
		this(row, col, nPhoton, BasisFunction.vector2Tab(ev1), scale);
	}
	
	/**
	 * constructs an object with a gaussian profile of the specified scale
	 * @param row row position
	 * @param col column position
	 * @param nPhoton number of photons
	 * @param scale sigma 
	 */
	public DetectorObjectSurfaceBrightness(double row, double col, double nPhoton, double scale) {
		this(row, col, nPhoton, (new PSFGaussian(scale)).getEigenvalues(), scale);
	}

	/**
	 * returns the flux per pixel at position (r,c) in the ccd from this object, used for bright objects
	 * 
	 * @param rowIn row position
	 * @param colIn column position
	 * @return double flux in the pixel
	 */
	@Override
	public double getFlux(double rowIn, double colIn) {
		if (pixeledImage == null) {
			
			double flux = 0.0;
			double rowPosition = rowIn - rCor;
			double colPosition = colIn - cCor;
			for (int nx=0; nx<nOrder; nx++) {
				for (int ny=0; ny<nOrder; ny++) {
                    if (nx+ny < nOrder) {
                    	double temp = BasisFunction.evaluate(ny, nx, rowPosition, colPosition, scale);                  
                        flux += temp * eigenvalues[nx][ny];
                        //System.out.println("in getFlux:      nx="+nx+" ny="+ny+" temp="+temp);
                        //System.out.println("in getFlux:      nx="+nx+" ny="+ny+" flux="+flux);
                    }
				}
			}
			return flux*nPhoton/normalization;
		} 
		double retval = pixeledImage.getByRowCol(rowIn, colIn)/normalization; 
		return retval;
	}
	/**
	 * returns the flux per pixel at the center of pixel ir,ic in the ccd from this object
	 *       	//logger.finer(" iSquare="+iSquare+" maxFlux="+maxFlux);
	 * @param ir row position
	 * @param ic column position
	 * @return double flux at the center of the pixel
	 */
	public double getFluxAtPixFast(int ir, int ic) {
		if (pixeledImage == null) {
			double flux = 0.0;

			int iy = ir - yFastMin;
			if (iy < 0 || iy >= yFastPhiEvaluated[0].length) {
				return 0;
			}
			int ix = ic - xFastMin;
			if (ix < 0 || ix >= xFastPhiEvaluated[0].length) {
				return 0;
			}
			//System.out.println("getFluxAtPixFast:  ic="+ic+" ix="+ix+" xFastDefined.length="+xFastDefined.length);
			if (!xFastDefined[ix]) {
				xFastEvaluate(ix);
			}
			if (!yFastDefined[iy]) {
				yFastEvaluate(iy);
			}
			for (int nx=0; nx<nOrder; nx++) {
				for (int ny=0; ny<nOrder-nx; ny++) {
					double temp = xFastPhiEvaluated[nx][ix] * yFastPhiEvaluated[ny][iy]; 
					//System.out.println(" getFluxFast:  nx="+nx+" ny="+ny+" xphi="+xFastPhiEvaluated[nx][ix]+" yphi="+yFastPhiEvaluated[ny][iy]);
					flux +=   temp * eigenvalues[nx][ny];
					//System.out.println("in getFluxFast:  nx="+nx+" ny="+ny+" temp="+temp/scale);
					//System.out.println("in getFluxFast:  nx="+nx+" ny="+ny+" flux="+flux/scale);                    
				}
			}
			return flux*nPhoton/normalization/scale;
		} 
		return pixeledImage.getByRowCol(ir, ic);
	}
	/**
	 * corrected for image centroid shift
	 * @param ix
	 */
	private void xFastEvaluate(int ix) {
		xFastDefined[ix] = true;
		double x = (ix + xFastMin + 0.5 - cCor)/scale;
		//System.out.println(" in xFastEvaluate:  ix="+ix+" x="+x);
		double tempExp = Math.exp(-x*x/2.0);
		for (int n=0; n<nOrder; n++) {
			xFastPhiEvaluated[n][ix] = Hermite.getH()[n].evaluate(x) * tempExp / Phi.getCoeff()[n];
			//System.out.println(" in xFastEvaluate:  x="+x+" n="+n+" xFastPhiEvaluated[n][ix]="+xFastPhiEvaluated[n][ix]);
		}
	}
	
	private void yFastEvaluate(int iy) {
		yFastDefined[iy] = true;
		double y = (iy + yFastMin + 0.5 - rCor)/scale;
		double tempExp = Math.exp(-y*y/2.0);
		for (int n=0; n<nOrder; n++) {
			yFastPhiEvaluated[n][iy] = Hermite.getH()[n].evaluate(y) * tempExp / Phi.getCoeff()[n];
			//System.out.println(" in yFastEvaluate:  y="+y+" n="+n+" yFastPhiEvaluated[n][iy]="+yFastPhiEvaluated[n][iy]);
		}
	}


	/**
	 * get the integrated flux in the rectangles
	 * corrected for image centroid shift
	 * @param row0 min row
	 * @param col0 min col
	 * @param row1 max row
	 * @param col1 max col
	 * @return integrated flux
	 */
	public double getIntegratedFlux(double row0, double col0, double row1, double col1) {
		//double r0 = row-row0;
		//double c0 = col-col0;
		//double r1 = row-row1;
		//double c1 = col-col1;
		double rowDelta = row1-row0;
		double colDelta = col1-col0;
		double rowPosition = 0.5*(row0+row1) - rCor;
		double colPosition = 0.5*(col0+col1) - cCor;
		//double I0 = b.integrate(ny, nx, y0, x0, beta, 2.*radius, 2.*radius);
		double flux = 0.0;
		for (int nx=0; nx<nOrder; nx++) {
			for (int ny = 0; ny<nOrder; ny++) {
				if (nx+ny < nOrder) {
					double bIntegrate = BasisFunction.integrate(nx, ny, colPosition, rowPosition, scale, colDelta, rowDelta);
					//System.out.println(" getIntegratedFlux:  row0="+row0+" col0="+col0+" row="+row+" col="+col+" rowPosition="+rowPosition+" colPosition="+colPosition+" rowDelta="+rowDelta+" colDelta="+colDelta);
					//System.out.println("getIntegratedFlux:   nx="+nx+" ny="+ny+" bIntegrate="+bIntegrate+" eigenvalue="+eigenvalues[nx][ny]);
					flux += bIntegrate*eigenvalues[nx][ny];
				}
			}
		}
		//System.out.println(" nPhoton="+nPhoton+" normalization="+normalization);
		return flux*nPhoton/normalization;
	}
	/**
	 * get integrated flux in this pixel
	 * @param ir row
	 * @param ic col
	 * @return integrated flux
	 */
	public double getFluxByPixSlow(int ir, int ic) {
		if (pixeledImage == null) {
			return getIntegratedFlux(ir, ic, ir+1, ic+1);
		} 
		return pixeledImage.getByRowCol(ir,ic);
	}

	/**
	 * returns the integrated flux in pixel ir,ic from this object by calling getFluxInPixFastShapelet or looking up the value in the pixeled image
	 * 
	 * @param ir row position
	 * @param ic column position
	 * @return double integrated flux in the pixel
	 */
	public double getFluxByPix(int ir, int ic) {
		if (pixeledImage == null) {
			return getFluxInPixFastShapelet(ir, ic);
		} 
		return getFlux(ir, ic);
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#getFluxInPixel(int, int)
	 */
	@Override
	public double getFluxInPixel(int ir, int ic) {
		return getFluxByPix(ir,ic);
	}
	
	private double getFluxInPixFastShapelet(int ir, int ic) {
		double flux = 0.0;

		int iy = ir - yFastMin;
		if (iy < 0 || iy >= yFastIntegrateEvaluated[0].length) {
			return 0;
		}
		int ix = ic - xFastMin;
		if (ix < 0 || ix >= xFastIntegrateEvaluated[0].length) {
			return 0;
		}
		//System.out.println("getFluxAtPixFast:  ic="+ic+" ix="+ix+" xFastDefined.length="+xFastDefined.length);
		if (!xFastIntegrateDefined[ix]) {
			defineXFastIntegrate(ix);
		}
		if (!yFastIntegrateDefined[iy]) {
			defineYFastIntegrate(iy);
		}
		for (int nx=0; nx<nOrder; nx++) {
			for (int ny=0; ny<nOrder; ny++) {
		        if (nx+ny < nOrder) {
		        	double temp = xFastIntegrateEvaluated[nx][ix] * yFastIntegrateEvaluated[ny][iy]; 
		        	//System.out.println(" getFluxFast:  nx="+nx+" ny="+ny+" xphi="+xFastPhiEvaluated[nx][ix]+" yphi="+yFastPhiEvaluated[ny][iy]);
		            flux +=   temp * eigenvalues[nx][ny];
		            //System.out.println(" getFluxInPixFast:  nx="+nx+" ny="+ny+" temp="+temp+" eibenvalue="+eigenvalues[nx][ny]);
		            //System.out.println("in getFluxFast:  nx="+nx+" ny="+ny+" temp="+temp/scale);
		            //System.out.println("in getFluxFast:  nx="+nx+" ny="+ny+" flux="+flux/scale);
		        }
			}
		}
		return flux*nPhoton/normalization;
	}
/**
 * corrected for image centroid shift
 * @param iy
 */
	private void defineYFastIntegrate(int iy) {
		yFastIntegrateDefined[iy] = true;
		double y = (iy + yFastMin + 0.5 - rCor);
		int nMax = nOrder-1;
		double[] temp = BasisFunction.preIntegrate(nMax, y, scale);
		for (int i=0; i<nOrder; i++) {
			yFastIntegrateEvaluated[i][iy] = temp[i];
		}
	}
	/**
	 * corrected for image centroid shift
	 * @param ix
	 */
	private void defineXFastIntegrate(int ix) {
		xFastIntegrateDefined[ix] = true;
		double x = (ix + xFastMin + 0.5 - cCor);
		int nMax = nOrder-1;
		double[] temp = BasisFunction.preIntegrate(nMax, x, scale);
		for (int i=0; i<nOrder; i++) {
			xFastIntegrateEvaluated[i][ix] = temp[i];
		}
	}
	


	/**
	 * get the dimension of the 2d matrix
	 * @return nOrder
	 */
	public int getNOrder() { return nOrder;}
	/**
	 * @return the maximum order, which is the number of orders-1
	 */
	public int getNMax() {return nOrder-1;}

	/**
	 * convolve the DetectorObjectShapelet with the psf, if it is not null.  
	 * @param psf the psf for convolution
	 * @return the convolved shapelet object
	 */
	public DetectorObjectSurfaceBrightness convolve(PSF psf) {
		DetectorObjectSurfaceBrightness retval;
 		if (psf == null) {
			retval = this;
 		} else {
 			double convolvedScale = Math.sqrt(scale*scale + psf.getScale()*psf.getScale());
 			double [][] psfEigenvalues = psf.getEigenvalues();
 			//Convolver convolver = new Convolver(convolvedScale, scale, psf.getScale(), psfEigenvalues.length);
 			//double[][] convolvedEigenvalues = convolver.Convolute(eigenvalues, psfEigenvalues);
 			int maxNOrder = Math.max(nOrder, psfEigenvalues.length);
 			Convolution convolution = new Convolution(scale, psf.getScale(), maxNOrder-1);
 			double[][] convolvedEigenvalues = convolution.convolve(eigenvalues, psfEigenvalues);
 			DetectorObjectSurfaceBrightness convolvedObject = new DetectorObjectSurfaceBrightness(row, col, nPhoton, convolvedEigenvalues, convolvedScale);
 			retval = convolvedObject;
 		}
 		return retval;
	}

	/**
	 * @param psf
	 * @return the result of the convolution
	 */
	public DetectorObjectSurfaceBrightness convolve(DetectorObjectSurfaceBrightness psf) {
		double psfScale=psf.scale;
		double convolvedScale = Math.sqrt(scale*scale + psfScale*psfScale);
		int nMax = Math.max(getNMax(), psf.getNMax());

		//Convolver convolver = new Convolver(convolvedScale, scale, psfScale, nMax);
		//double[][] convolvedEigenvalues = convolver.Convolute(eigenvalues, psf.eigenvalues);
		//double[] convolvedEigenvalues = convolver.Convolute2(eigenvalues, psf.eigenvalues);

		Convolution convolution = new Convolution(scale, psf.getScale(), nMax);
		double[][] convolvedEigenvalues = convolution.convolve(eigenvalues, psf.eigenvalues);
		
		DetectorObjectSurfaceBrightness convolvedObject = new DetectorObjectSurfaceBrightness(row, col, nPhoton, convolvedEigenvalues, convolvedScale);
		return convolvedObject;		
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#convolveGaussian(double)
	 */
	@Override
	public void convolveGaussian(double gaussianScale) {
		if (pixeledImage == null) {
			double convolvedScale = GaussConvolve.getGamma(scale, gaussianScale);
			double[][] evConvolved = GaussConvolve.convolve(scale, eigenvalues, gaussianScale);
			init(row, col, nPhoton, evConvolved, convolvedScale);	
		} else {
			throw new Error(" teach me how to convolve....");
		}
	}
	
	/**
	 * @return the pixeled image of this object
	 */
	public PixeledImage getPixeledImage() {
		return pixeledImage;
	}
	
	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#makePixeledImage(int)
	 */
	@Override
	public PixeledImage makePixeledImage(int pixelDiameter) {
		return makePixeledImage(pixelDiameter, true);
	}
	
	/**
	 * make a small PixeledImage of this object
	 * @param pixelDiameter number of rows and the number of columns
	 * @param allowNegNums allow values to go below 0?
	 * @return the pixeledImage
	 */
	public PixeledImage makePixeledImage(int pixelDiameter, boolean allowNegNums) {
		if (Double.isNaN(rCor)) throw new Error("rCor is NaN");
		int row0 = (int) Math.floor((rCor-pixelDiameter/2 )) + 1;
		if (Double.isNaN(cCor)) throw new Error("cCor is NaN");
		int col0 = (int) Math.floor((cCor-pixelDiameter/2 )) + 1;
		PixeledImage pi = new PixeledImage(pixelDiameter, pixelDiameter);
		pi.setRow0Col0(row0, col0);
		renderToPixeledImage(pi, allowNegNums);
		return pi;
	}

	/**
	 * render this object in all pixels, allowing negative numbers
	 * @param pi where to render
	 */
	public void renderToPixeledImage(PixeledImage pi) {
		renderToPixeledImage(pi, true);
	}
    /**
     * render this object in all pixels
     * @param pi where to render
     * @param allowNegNums whether to allow negative numbers
     */
    @Override
	public void renderToPixeledImage(PixeledImage pi, boolean allowNegNums) {
		for (int r=pi.getRowMin(); r<pi.getRowMax(); r++) {
			for (int c=pi.getColumnMin(); c<pi.getColumnMax(); c++) {
				double value = getFluxByPix(r, c);
				if (value < 0 && !allowNegNums) {
					value = 0;
				}
				pi.incrByRowCol(r, c, value);
			}
		}
	}

	/**
     * write some debugging information to a file
     * @param fileName
     * @throws Exception
     */
    public void dumpToFile(String fileName) throws Exception {
       PrintStream ps = new PrintStream(new FileOutputStream(new File(fileName)));
       if (pixeledImage == null) {
           ps.println("This is a shapelet object");
           ps.println("scale="+scale+" normalizaton="+normalization+" nOrder="+nOrder);
           int ir=(int) row;
           int ic=(int) col;
           ps.println("row="+row+" col="+col+" ir="+ir+" ic="+ic);
           double flux = getFluxByPix(ir, ic);
           ps.println("flux in ir,ic is "+flux);
           ps.println(" eigenvalue vector");
           for (int i=0; i<ev1.length; i++) {
               ps.println("ev1["+i+"]="+ev1[i]);
           }
        double fluxSum = 0.0;
        double rowDelta = 1.0;
        double colDelta = 1.0;
        double rowPosition = 0.5;
        double colPosition = 0.5;
        for (int nx=0; nx<nOrder; nx++) {
            for (int ny = 0; ny<nOrder; ny++) {
                if (nx+ny < nOrder) {
                    double bIntegrate = BasisFunction.integrate(nx, ny, rowPosition, colPosition, scale, rowDelta, colDelta);
                    fluxSum += bIntegrate*eigenvalues[nx][ny];
                    ps.println(" nx="+nx+" ny="+ny+" eigenvalues="+eigenvalues[nx][ny]+" bIntegrate="+bIntegrate+" thisFlux="+bIntegrate*eigenvalues[nx][ny]+" fluxSum="+fluxSum);
                }
            }
        }
        ps.close();
       }
    }
    /**
     * creates the RowListStarTable to track what has been rendered
     * @return the empty table
     */
    public static RowListStarTable createRowListStarTable() {
        int length = BasisFunction.lengthFromDimension(BasisFunction.getNMax());
        ColumnInfo[] columnInfoArray = new ColumnInfo[8+length];
        columnInfoArray[0] = new ColumnInfo("row", Double.class, "row position pixels");
        columnInfoArray[1] = new ColumnInfo("col", Double.class, "column position in pixels");
        columnInfoArray[2] = new ColumnInfo("nPhoton", Double.class, "number of photons");
        columnInfoArray[3] = new ColumnInfo("scale", Double.class, "scale in pixels");
        columnInfoArray[4] = new ColumnInfo("nOrder", Integer.class, "number of shapelet orders rendered");
        columnInfoArray[5] = new ColumnInfo("ra", Double.class, "true right ascension (degrees");
        columnInfoArray[6] = new ColumnInfo("dec", Double.class, "true declination (degrees)");
        columnInfoArray[7] = new ColumnInfo("fromTable", String.class, "truth table from which this object was taken");
        for (int i=0; i<length; i++) {
            columnInfoArray[8+i] = new ColumnInfo("coeff_"+i, Double.class, "coefficient "+i);
        }
        RowListStarTable rlst = new RowListStarTable(columnInfoArray);
        rlst.setName("detectorObjectSurfaceBrightness");
        return rlst;
    }
    
    /**
     * @see gov.fnal.eag.sim.pixel.DetectorObject#addToTable(uk.ac.starlink.table.RowListStarTable)
     */
	@Override
	public void addToTable(RowListStarTable rlst) {
		if (!rlst.getName().equals("detectorObjectSurfaceBrightness")) return;
		Object[] temp = new Object[rlst.getColumnCount()];
		temp[0] = new Double(row);
		temp[1] = new Double(col);
		temp[2] = new Double(nPhoton);
		temp[3] = new Double(scale);
		temp[4] = new Integer(nOrder);
		temp[5] = new Double(ra);
		temp[6] = new Double(dec);
		temp[7] = fromTable;
		if (pixeledImage == null) {
			int length = BasisFunction.lengthFromDimension(BasisFunction.getNMax()+1);
			if (ev1.length > length) {
				throw new Error("ev1 length="+ev1.length+" is greater than length="+length);
			}
			double value;
			for (int i=8; i<rlst.getColumnCount(); i++) {
				try {
					value = ev1[i-8];
				} catch (ArrayIndexOutOfBoundsException e) {
					value = 0;
				}
				temp[i] = new Double(value);
			}
		} else {
			for (int i=8; i<rlst.getColumnCount(); i++) {
				temp[i] = new Double(0.0);
			}
		}
		rlst.addRow(temp);        
	}

    /**
     * rotate this object
     * @param phi degrees to rotate
     * @throws Exception
     */
    public void rotate(double phi) throws Exception {
    	double[] ev1Rotated = ShapeletsTransform.rotate(ev1, phi);
    	double[][] evRotated = BasisFunction.vector2Tab(ev1Rotated);
    	init(row, col, nPhoton, evRotated, scale);
    }
    /**
     * shear this object in cartesian space 
     * @param gamma1 shear factor in x
     * @param gamma2 shear factor in 45 degrees to z
    */
    public void shearCartesian(double gamma1, double gamma2) {
    	double[][] ev = BasisFunction.vector2Tab(ev1);
    	double[][] evSheared = ShapeletsTransform.shear(ev, gamma1, gamma2);
    	init(row, col, nPhoton, evSheared, scale);
    }
    /**
     * performs shear and compression operations
     * @param gamma1
     * @param gamma2
     * @param kappa
     */
    public void transform(double gamma1, double gamma2, double kappa)  {
    	Complex[] evC = ShapeletsTransform.c2p(ev1);
    	if (Math.abs(gamma1) > 0.0001 || Math.abs(gamma2) > 0.0001) {
    		Complex[] ev1Sheared = ShapeletsTransform.shear(evC, gamma1, gamma2) ;
    		for (int i=0; i<evC.length; i++) {
    			evC[i] = ev1Sheared[i];
    		}
    	}	
    	if (Math.abs(kappa) > 0.0001) {
    		Complex[] evD = ShapeletsTransform.dilate(evC, kappa);
    		for (int i=0; i<evC.length; i++ ) {
    			evC[i] = evD[i];
    		}
    	}
    	double[] evS = ShapeletsTransform.p2c(evC);
    	double[][] evTransf = BasisFunction.vector2Tab(evS);
    	init(row, col, nPhoton, evTransf, scale);
    }
	/**
	 * @return  col centroid deviation
	 *
	 */
	public double getdX() {
		return dX;
	}
	/**
	 * @return  row centroid deviation
	 * 
	 */
	public double getdY() {
		return dY;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.DetectorObject#gc()
	 */
	@Override
	public void gc() {
	}	
}
