package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.sim.random.SimRandomUniform;
import gov.fnal.eag.sim.repository.image.FitsDataFormat;
import gov.fnal.eag.sim.repository.image.ImageRepository;
import nom.tam.fits.HeaderCardException;

/**
 * used to compress and uncompress data, inspired by John Marriner's digitize.c
 * 
 * The amount of data compression is specified by the number of output bits (NSQBIT).  The table maps the
 * smallest allowed input to 0 and the largest allowed input value 2^NSQBIT-1
 * 
 * This class is used to handle NIR detectors directly.  It is called in the CCDDigitizeLookupTable class.
 * 
 * @author stoughto
 *
 */
public class DigitizeLookupTable extends LookupTable {
	static double BINPAR = 0.3;
	static double BINUND = 5.0;
	static double bigFlt = 1000000.0;
	static double epsilon=1e-10;
	int[] iv;
	/**
	 * number of bits to use for encoding
	 */
	public static final int NSQBIT = 10;
	int is;
	double gain, bias;
	
	/**
	 * default constructor with reasonable defaults
	 */
	public DigitizeLookupTable() {
		int maxtab = 1 << DigitizeLookupTable.NSQBIT;
		int nu = maxtab;
		double readNoise = 10.0;
		double gainIn = 2.0;
		double biasIn = 1000.0;
		int maxValue =  (1<<16) - 1;		
		int isIn = 1;
		init(nu, readNoise, gainIn, biasIn, maxValue, isIn);
	}
	
	/**
	 * @param nu
	 * @param readNoise noise in ADU
	 * @param gain electron/ADU
	 * @param bias in adu
	 * @param maxValue
	 * @param is integer scaling factor
	 */
	
	public DigitizeLookupTable(int nu, double readNoise, double gain, double bias, int maxValue, int is) {
		init(nu, readNoise, gain, bias, maxValue, is);
	} 
	
	private void init(int nu, double readNoise, double gainIn, double biasIn, int maxValue, int isIn) {
		// Remember the number of bits to use, gain, and bias
		this.gain = gainIn;
		this.bias = biasIn;
		this.is = isIn;
		// Make the array that holds the lookup table
		iv = new int[nu+1];
		// lbin is the number of values per bin in the linear region; must be at least one
		int lbin = (int) (BINPAR*readNoise/gain);
		lbin = Math.max(lbin, 1);
		
		// vmin is the smallest input value to be encoded in the table; 10 sigma below bias.
		// It must be positive and not too close to an integer value
		double vmin = bias - BINUND*readNoise/gain;
		vmin = Math.max(vmin, 0.0);
		double delta = vmin - Math.floor(vmin);
		if (delta < epsilon) {
			vmin = Math.floor(vmin) + epsilon; // differs from original code (line 337)
		} else if (delta > 1-epsilon) {
			vmin = Math.floor(vmin) + 1 - epsilon;
		}
		double vmax = maxValue + epsilon;

		//double urange = nu;
		double urange = nu-1;

		/*
		  
		  Solve for the value of q that will achieve the desired map. 

		  We are looking for a value of q that will use all 2^NSQBIT entries
	      in the lookup table.  Two cases are considered:
	      A) Square root prescaling can be used for the desired range of input
	         values.
	      B) Transition to a linear form is required for input values that
	         result in the argument of the square root being very small or
	         negative.

	   */
		double q = -1.0;
		double dmin = vmin - bias+readNoise*readNoise/gain; // yes, only divide by gain one time
		double dmax = vmax - bias+readNoise*readNoise/gain;
		double vc, uc, ubias;
		if (dmin > 0) { 
			q = urange/(2.0*(Math.sqrt(dmax)-Math.sqrt(dmin)));
		}
		// now make sure the derivative is smaller than slope of the linear curve
		if ( dmin > 0 && q <= Math.sqrt(dmin)/lbin) {
			vc = vmin;
			uc = 0.0;
			ubias = 2.0*q*Math.sqrt(dmin);
		} else {
			double aco = lbin;
			double bco = -2.0*Math.sqrt(dmax);
			double cco = urange+(dmin)/lbin;
			/* Argument of the square root */
			double rad = bco*bco - 4.0*aco*cco;
			/* Return error = 1 if argument is negative */
			if (rad<0.0) {
				throw new Error("rad need to be positive but rad="+rad);
			}
			/* Solution for q */
			q = (-bco - Math.sqrt(rad))/(2.0*aco);
			/* Return error =2 is q is negative */
			if (q<0.0) {
				throw new Error("q needs to be positive but q="+q);
			}
			vc = q*lbin*q*lbin - readNoise*readNoise/gain + bias;
			uc = 2.0*q*Math.sqrt(vc-bias+readNoise*readNoise/gain);
			ubias = (vmin-vc)/lbin + uc;
		}
		for (int iu=0; iu < iv.length; ++iu) {
		      /* Value of u with bias */
		      double u = iu + ubias;
		      /* See if u greater than transition point between linear & sqrt */
		      double v;
		      if (u>uc) {
		          	/* Yes.  Use sqrt formula */
		          	double ut = u/(2.*q);
		          	v = ut*ut - readNoise*readNoise/gain + bias;
		        } else {
		        	/*  No.  Use linear formula */
		        	v = lbin*(u-uc) + vc;
		        }
		      /* Store value in lookup table */
		      iv[iu] = (int) v;
		      /* iv should always be increasing. Return error if not */
		      if (iu>0 && iv[iu]<=iv[iu-1]) {
		    	  throw new Error("iv should increase but iv["+iu+"]="+iv[iu]+" and iv["+(iu-1)+"]="+iv[iu-1]);
		      }
		}
	}

	/**
	 * @param in
	 * @return the compressed value
	 */
	public int nirCompress(int in) {
		if (in < iv[0]) {
			return 0;
		}
		if (in >= iv[iv.length-1]) {
			return iv.length-1;
		}
		int i1 = 0;
		for (int i=is; i<=NSQBIT; ++i) {
			int im = (1<<(NSQBIT-i)) + i1;
			if (in >= iv[im]) {
				i1 = im;
			}
		}
		int retval = i1;
		return retval;
	}
	
	/**
	 * uncompress, but throw an ArrayIndexOutOfBoundsException if the input value is out of range
	 * @param in
	 * @return the uncompressed value
	 */
	public double nirDecompress(int in) {
		double value;
		if (in < 0) {
			value = iv[0];
		} else if (in > iv.length-2) {
			value = iv[iv.length-1];
		} else {
			value = (iv[in]+iv[in+1]-1)/2.0;
		}
		double retval = gain*(value-bias);
		return retval;
	}
	/**
	 * uncompress, but throw an ArrayIndexOutOfBoundsException if the input value is out of range
	 * @param in
	 * @param sru 
	 * @return the uncompressed value
	 */
	public double nirDecompress(int in, SimRandomUniform sru) {
		double v0 = nirDecompress(in-1);
		double v1 = nirDecompress(in);
		double v2 = nirDecompress(in+1);
		double min = 0.5*(v0+v1);
		double max = 0.5*(v1+v2);
		double retval = sru.next(min,max);
		return retval;
	}
	/**
	 * uncompress, but return the min (max) value if the input is below (above) the defined range
	 * @param in
	 * @return the uncompressed value
	 */
	public double nirDecompressNoException(int in) {
		double value;
		try {
			value = (iv[in]+iv[in+1]-1)/2.0;
		} catch (ArrayIndexOutOfBoundsException e) {
			if (in < 0) {
				value = iv[0];
			} else {
				value = iv[in];
			}
		}
		double retval = gain*(value-bias);
		return retval;
	}
	
	/**
	 * produces digitized output as produced by the sidecar ASIC
   	 * The model is an ideal 16 bit (unsigned) digitizer.
	 * @param in input value
	 * @param gain gain, in electrons per ADU
	 * @param bias in ADU
	 * @return the digitized value, in the range 0, 0xFFFF inclusive
	 */
	public static int nirDigitize(double in, double gain, double bias) {
		in = Math.min(in, bigFlt);
		in = Math.max(in, -bigFlt);
		int retval = (int) (in/gain + bias + 0.5);
		retval = Math.min(retval, 0xFFFF);
		retval = Math.max(retval, 0);
		return retval;
	}
	
	/**
	 * unpacks digitized NIR values
	 * @param in input number of electrons
	 * @param gain in electron/ADU
	 * @param bias in ADU
	 * @return the value in electrons
	 */
	public static double nirUnpack(int in, double gain, double bias) {
		double retval = gain*(in-bias);
		return retval;
	}
	/**
	 * @return the value of epsilon used to construct tables
	 */
	public static double getEpsilon() {
		return epsilon;
	}
	
	/**
	 * @param iCompressed
	 * @return the tolerence for the compress,decompress cycle
	 */
	public double getNirTolerence(int iCompressed) {
		double retval;
		try {
			retval = gain*(iv[iCompressed+1]-iv[iCompressed]+2)/2.0;
		} catch (ArrayIndexOutOfBoundsException e) {
			if (iCompressed > 0) {
				retval =  gain*(iv[iCompressed]-iv[iCompressed-1]+2)/2.0;
			} else {
				throw e;
			}
		}
		return retval;
	}
	
	/**
	 * @return the smallest value in the lookup table, iv[0]
	 */
	public int getIvMin() {
		return iv[0];
	}
	
	/**
	 * @return the largest value in the lookup table, iv[iv.length-1]
	 */
	public int getIvMax() {
		return iv[iv.length-1];
	}

	/**
	 * @param value
	 * @return the digitize value, using the gain and bias of this
	 */
	public int nirDigitize(double value) {
		return DigitizeLookupTable.nirDigitize(value, gain, bias);
	}

	/**
	 * @param in digitized value
	 * @return unpacked value, using the gain and bias of this
	 */
	public double nirUnpack(int in) {
		return DigitizeLookupTable.nirUnpack(in, gain, bias);
	}
	/**
	 * @param args inFileName outFileName
	 * @throws HeaderCardException
	 */
	public static void main (String[] args) throws HeaderCardException {
		if (args.length != 2) {
			System.out.println("Run SNAP NIR compression and decompression algorithm, and write output file as ushort FITS image");
			String name = CCDDigitizeLookupTable.class.getName();
			System.out.println("Usage:  "+name+" inputFileName outputFileName");
			return;
		}
		String inFileName = args[0];
		String outFileName = args[1];
		if (inFileName.equals(outFileName)) {
			System.out.println("Error:  inFileName and outFileName are the same");
			return;
		}
		
		PixeledImage pi = ImageRepository.readPixeledImage(inFileName);
		int maxtab = 1 << DigitizeLookupTable.NSQBIT;
		int nu = maxtab;
		double readNoise = 10.0;
		double gain = 2.0;
		double bias = 1000.0;
		int maxValue =  (1<<16) - 1;		
		int is = 1;
		DigitizeLookupTable dlt = new DigitizeLookupTable(nu, readNoise, gain, bias, maxValue, is);
		
		
		for (int row=pi.getRowMin(); row<pi.getRowMax(); row++) {
			for (int col=pi.getColumnMin(); col<pi.getColumnMax(); col++) {
				double value = pi.getByRowCol(row, col);
				
				int iv = DigitizeLookupTable.nirDigitize(value, gain, bias);
				int iCompressed = dlt.nirCompress(iv);

				value = dlt.nirDecompress(iCompressed);
				pi.setByRowCol(row, col, value);
			}
		}
		pi.header.addValue("DLTVER", getVersion(), "CCDDigitizeLookupTable version");
		ImageRepository.write(outFileName, pi, FitsDataFormat.USHORT, true);
		
	}
	
	/**
	 * @return the cvs version string for this class
	 */
	public static String getVersion() {
		 String version = "DigitizeLookupTable.java,v 1.5 2008/01/04 19:10:00 descvs Exp";
		 return version;
	}
	
	/**
	 * @return the lookup values
	 */
	public int[] getIv() {
		return iv;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#compress(double)
	 */
	@Override
	public int compress(double value) {
		return nirCompress(nirDigitize(value));
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#getNBits()
	 */
	@Override
	public int getNBits() {
		return NSQBIT;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#restore(int)
	 */
	@Override
	public double restore(int value) {
		return nirDecompress(value);
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.LookupTable#nValues(int)
	 */
	@Override
	public int nValues(int compressed) {
		throw new UnsupportedOperationException();
	}
}

