package gov.fnal.eag.sim.pixel;

import java.util.ArrayList;

import gov.fnal.eag.sim.random.SimRandom;
import gov.fnal.eag.sim.random.SimRandomExponential;

/**
 * An ElectronTrap based on my interpretation of how Chris Bebek described a simple model of how traps work
 * with notch depth and trap depth added per suggestions of Richard Massey
 * @author stoughto
 *
 */
public class ElectronTrapCB1 implements ElectronTrap {

	private double contents = 0.0;
	private SimRandomExponential sre;
	private double trapHeightFraction;
	private double trapDepth = 1.0;
	/**
	 * constructor with 1 electron per trap
	 * @param sre defined with lambda=1/lifetime of the decay, in seconds
	 * @param trapHeightFraction location of the trap, from 0 to 1
	 */
	public ElectronTrapCB1(SimRandomExponential sre, double trapHeightFraction) {
		this(sre, trapHeightFraction, 1.0);
	}
	/**
	 * @param sre defined with lambda=1/lifetime of the decay, in seconds
	 * @param trapHeightFraction location of the trap, from 0 to 1
	 * @param trapDepth the number of electrons that fit in this trap
	 */	
	public ElectronTrapCB1(SimRandomExponential sre, double trapHeightFraction, double trapDepth) {

		this.sre = sre;
		if (trapHeightFraction < 0 || trapHeightFraction > 1) {
			throw new Error("trapHeightFraction must be in [0,1] but trapHeightFraction="+trapHeightFraction);
		}
		this.trapHeightFraction = trapHeightFraction;
		this.trapDepth = trapDepth;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.ElectronTrap#decay(double, boolean, double)
	 */
	public double decay(double time, boolean quantizeCharge, double chargeHeight) {
		if (time < 0.0) {
			throw new IllegalArgumentException("time can not be negative but time="+time);
		}
		if (trapHeightFraction < chargeHeight) {
			return 0.0;
		}
		if (quantizeCharge) {
			if (contents == trapDepth) {
				double value = sre.next();
				if (value < time) {
					// it decays
					contents = 0.0;
					return trapDepth;
				}
			}
			return 0.0;		
		}
		double fraction = 1.0 - Math.exp(-time*sre.getLambda());
		double retval = contents*fraction;
		contents -= retval;
		return retval;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.ElectronTrap#getContents()
	 */
	public double getContents() {
		return contents;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.ElectronTrap#absorb(double)
	 */
	public double absorb(double chargeIn) {
		double retval = 0.0;
		if (contents < trapDepth) {
			if (chargeIn >= trapDepth) {
				retval = trapDepth-contents;
				contents = trapDepth;
			}
		}
		return retval;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.ElectronTrap#releaseTrappedCharge()
	 */
	public double releaseTrappedCharge() {
		double retval = contents;
		contents = 0.0;
		return retval;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.ElectronTrap#setDecaySeed(int)
	 */
	public void setDecaySeed(int decaySeed) {
		sre.setIseedIndex(decaySeed);
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.ElectronTrap#getSimRandoms()
	 */
	public ArrayList<SimRandom> getSimRandoms() {
		ArrayList<SimRandom> retval = new ArrayList<SimRandom>();
		retval.add(sre);
		return retval;
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.ElectronTrap#getTrapHeightFraction()
	 */
	public double getTrapHeightFraction() {
		return trapHeightFraction;
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.ElectronTrap#getTrapDepth()
	 */
	public double getTrapDepth() {
		return trapDepth;
	}
}
