

package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.sim.random.SimRandom;

import java.util.ArrayList;

/**
 * describes the mean behavior of electron traps
 * @author stoughto
 *
 */
public class ElectronTrapCB1Mean implements ElectronTrap {
	// defines the physical properties of the trap
	private double trapsPerPixel;
	private double lifetime;
	private double fullWell;
	private double power = 0.5;
	private double notchDepth = 0.0;
	//private double fullWellMinusNotchDepth = 0.0;
	// defines the state of the trap
	private int nLevels;
	/**
	 * default number of levels for mean traps
	 */
	public static final int nLevelsDefault = 1000;

	/**
	 * default power for charge height calculation
	 */
	public static final double powerDefault = 0.5;
	/**
	 * default notch depth for charge height calculation
	 */
	public static final double notchDepthDefault = 0.0;

	private double[] heights; 
	private double[] fillFraction;
	/**
	 * use the defaults of power=0.5 and notchDepth=0.0 and nLevels=10000
	 * @param trapsPerPixel
	 * @param lifetime
	 * @param fullWell
	 */
	public ElectronTrapCB1Mean(double trapsPerPixel, double lifetime, double fullWell) {
		this(trapsPerPixel, lifetime, fullWell, powerDefault, notchDepthDefault, nLevelsDefault);
	}
	/**
	 * fully-specified  constructor
	 * @param trapsPerPixel
	 * @param lifetime
	 * @param fullWell
	 * @param power
	 * @param notchDepth
	 * @param nLevels 
	 */
	public ElectronTrapCB1Mean(double trapsPerPixel, double lifetime, double fullWell, double power, double notchDepth, int nLevels) {
		this.trapsPerPixel = trapsPerPixel;
		this.lifetime = lifetime;
		this.fullWell = fullWell;
		this.power = power;
		this.notchDepth = notchDepth;
		//this.fullWellMinusNotchDepth = fullWell-notchDepth;
		this.nLevels = nLevels;
		heights = new double[nLevels];
		fillFraction = new double[nLevels];
		//double deltaCharge = (fullWell-notchDepth)/nLevels;
		for (int i=0; i<nLevels; i++) {
			heights[i] = i/(double)nLevels;
			//fillFraction[i] = 0.0;
		}
	}
	/**
	 * @param lifetime
	 * @param trapsPerPixel
	 */
	public void setLifetimeTrapsPerPixel(double lifetime, double trapsPerPixel) {
		this.lifetime = lifetime;
		this.trapsPerPixel = trapsPerPixel;
		
		for (int i=0; i<fillFraction.length; i++) {
			if (fillFraction[i] == 0) {
				break;
			}
			fillFraction[i] = 0.0;
		}
		
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.ElectronTrap#absorb(double)
	 */
	public double absorb(double chargeIn) {
		double chargeHeight = getChargeHeightFraction(chargeIn);
		double chargeAbsorbed = 0.0;
		for (int i=0; i<nLevels; i++) {
			double dh = chargeHeight - heights[i];
			if (dh > 0) {
				double temp = trapsPerPixel * (1.0-fillFraction[i]) / nLevels; 
				double hFraction = Math.min(dh*nLevels, 1.0);
				chargeAbsorbed += temp*hFraction;
				fillFraction[i] = hFraction;
			} else {
				break;
			}
		}
		return chargeAbsorbed;
	}


	/**
	 * @see gov.fnal.eag.sim.pixel.ElectronTrap#decay(double, boolean, double)
	 */
	public double decay(double time, boolean quantizeCharge, double chargeHeight) {
		if (quantizeCharge) {
			throw new IllegalArgumentException("quantizeCharge can not be true in mean electron traps");
		}

		double exp = Math.exp(-time/lifetime);
		double retval = 0.0;
		for (int i=0; i<nLevels; i++) {
			if (fillFraction[i] <= 0.0) break;
			double dh = nLevels*(heights[i] + 1.0/nLevels - chargeHeight);
			if (dh > 0) {
				double qReleased = (1.0-exp) * trapsPerPixel * fillFraction[i] / nLevels; 
				double hFraction = Math.min(dh, 1.0);
				qReleased *= hFraction;
				retval += qReleased;
				fillFraction[i] -= qReleased*nLevels/trapsPerPixel;
			}
			
		}
		return retval;
	}


	/**
	 * @see gov.fnal.eag.sim.pixel.ElectronTrap#getContents()
	 */
	public double getContents() {

		double contents = 0.0;
		//fractionExposedFull * trapsPerPixel * (1.0-chargeHeightFraction);
		for (int i=0; i<nLevels; i++) {
			double temp =  trapsPerPixel * fillFraction[i] / nLevels; 
			contents += temp;
		}
		return contents;
	}


	/**
	 * @see gov.fnal.eag.sim.pixel.ElectronTrap#getSimRandoms()
	 */
	public ArrayList<SimRandom> getSimRandoms() {
		return new ArrayList<SimRandom>();
	}


	/**
	 * @see gov.fnal.eag.sim.pixel.ElectronTrap#releaseTrappedCharge()
	 */
	public double releaseTrappedCharge() {
		double retval = getContents();
		for (int i=0; i<nLevels; i++) {
			fillFraction[i] = 0.0;
		}
		return retval;
	}


	/**
	 * @see gov.fnal.eag.sim.pixel.ElectronTrap#setDecaySeed(int)
	 * 
	 * has no effect in this class since there are not random numbers involved
	 */
	public void setDecaySeed(int decaySeed) {
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.ElectronTrap#getTrapHeightFraction()
	 * always return -1.0 for this trap
	 */
	public double getTrapHeightFraction() {
		return -1.0;
	}
	
	private double getChargeHeightFraction(double chargeIn) {
		double chargeHeightFraction;
		if (chargeIn <= notchDepth) {
			chargeHeightFraction = 0.0;
		} else {
			//chargeHeightFraction = Math.sqrt(chargeIn)/Math.sqrt(fullWell);
			chargeHeightFraction = Math.pow((chargeIn-notchDepth)/(fullWell-notchDepth), power);
		}
		return chargeHeightFraction;
	}
	
	/**
	 * @see gov.fnal.eag.sim.pixel.ElectronTrap#getTrapDepth()
	 */
	public double getTrapDepth() {
		return trapsPerPixel;
	}

}
