package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.sim.catalog.AstroCatalogGenerator;
import gov.fnal.eag.sim.catalog.AstroObject;
import gov.fnal.eag.sim.catalog.AstroObjectGenerator;
import gov.fnal.eag.sim.catalog.AstroObjectGeneratorGrid;
import gov.fnal.eag.sim.catalog.AstronomicalFilter;
import gov.fnal.eag.sim.catalog.Morphology;
import gov.fnal.eag.sim.catalog.SED;
import gov.fnal.eag.sim.catalog.WavelengthScale;
import gov.fnal.eag.sim.catalog.Zucca;
import gov.fnal.eag.sim.catalog.ZuccaBandpass;
import gov.fnal.eag.sim.random.SimRandomNormal;
import gov.fnal.eag.sim.random.SimRandomPoisson;
import gov.fnal.eag.sim.random.SimRandomUniform;
import gov.fnal.eag.sim.repository.image.FitsDataFormat;
import gov.fnal.eag.sim.repository.image.FitsWriter;
import gov.fnal.eag.sim.repository.image.ImageRepository;
import gov.fnal.eag.util.astronomy.Cosmo;
import gov.fnal.eag.util.astronomy.CosmoLambdaFlat;
import gov.fnal.eag.util.misc.StringUtil;
import gov.fnal.eag.wcs.Coordinate;
import gov.fnal.eag.wcs.SphericalCoordinate;
import gov.fnal.eag.wcs.Transform;
import gov.fnal.eag.wcs.TransformLinear;
import gov.fnal.eag.wcs.TransformTangentPlaneA7;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Properties;

/**
 * makes one pixeled image
 * @author stoughto
 *
 */
public class MakePixeledImage {

	Properties properties;

	String imageRepositoryDirectory, baseFileName, psfFileName, psfDirectory, gridLfBandpass, filterName;
	int nRow, nCol, catalogSeedIndex, noiseSeedIndex;

	double raBore, decBore;
	double rowRef, colRef, detectorXPositionMeters, detectorYPositionMeters, metersPerPixel, detectorRotationDegrees, a7;

	double expTime, countsPerSecondAt20th, pixelSizeArcseconds, skyMagPerArcsec2, readNoiseElectrons;
	boolean doPoisson;
	String[] luminosityFunctionNames, gridObjectShapes, gridSEDNames;
	double[] gridCenterRas, gridCenterDecs, gridDeltaArcsecs, gridAbsMags, gridObjectSizeArcsecs, gridObjectReferenceRedshifts;
	double[] minRedshifts, maxRedshifts;
	int[] gridSizes;
	boolean[] gridDistancesAreRedshifts;

	Cosmo cosmo = new CosmoLambdaFlat();
	AstronomicalFilter filter;
	Transform pixelToSky;
	TransformLinear pixelToFocalPlane;
	SphericalCoordinate detectorMinRadec;
	SphericalCoordinate detectorMaxRadec;
	PSFFactory psfFactory;
	BackgroundLight backgroundLight;
	Collection<DetectorObject> detectorObjects;
	double cosmicRayDensity = 0.0;
	AstroCatalogGenerator astroCatalogGenerator;
	PixeledImageFactory pixeledImageFactory;
	WavelengthScale wavelengthScale = new WavelengthScale();

	/**
	 * @param args
	 * @throws IOException 
	 * @throws FileNotFoundException 
	 */
	public static void main(String[] args) throws FileNotFoundException, IOException {
		Properties userProperties = new Properties();
		if (args.length > 0) {
			userProperties.load(new FileInputStream(args[0]));
		}
		MakePixeledImage mpi = new MakePixeledImage(userProperties);
		mpi.run();
	}

	/**
	 * @param userProperties
	 */
	public MakePixeledImage(Properties userProperties) {
		this.properties = MakePixeledImage.getDefaultProperties();
		if (userProperties != null) {
			for (Object name : userProperties.keySet()) {
				String key = (String) name;
				String value = userProperties.getProperty(key);
				properties.setProperty(key, value);				
			}
		}
		init();
	}

	private void init() {
		imageRepositoryDirectory = properties.getProperty("ImageRepositoryDirectory");
		nRow = StringUtil.getPropertyAsInt(properties, "nRow");
		nCol = StringUtil.getPropertyAsInt(properties, "nCol");
		expTime = StringUtil.getPropertyAsDouble(properties, "expTime");
		countsPerSecondAt20th = StringUtil.getPropertyAsDouble(properties, "countsPerSecondAt20th");
		pixelSizeArcseconds = StringUtil.getPropertyAsDouble(properties, "pixelSizeArcseconds");
		skyMagPerArcsec2 = StringUtil.getPropertyAsDouble(properties, "skyMagPerArcsec2");
		readNoiseElectrons = StringUtil.getPropertyAsDouble(properties, "readNoiseElectrons");
		doPoisson = StringUtil.getPropertyAsBoolean(properties, "doPoisson");
		luminosityFunctionNames = StringUtil.getPropertyAsStringArray(properties, "luminosityFunctionNames");
		gridObjectShapes = StringUtil.tokenize(StringUtil.getPropertyAsString(properties, "gridObjectShapes"),",");
		//gridObjectShapes = StringUtil.getPropertyAsStringArray(properties, "gridObjectShapes");
		gridCenterRas = StringUtil.getPropertyAsDoubleArray(properties, "gridCenterRas");
		gridCenterDecs = StringUtil.getPropertyAsDoubleArray(properties, "gridCenterDecs");
		gridDeltaArcsecs = StringUtil.getPropertyAsDoubleArray(properties, "gridDeltaArcsecs");
		gridAbsMags = StringUtil.getPropertyAsDoubleArray(properties, "gridAbsMags");
		gridObjectSizeArcsecs = StringUtil.getPropertyAsDoubleArray(properties, "gridObjectSizeArcsecs");
		gridObjectReferenceRedshifts = StringUtil.getPropertyAsDoubleArray(properties, "gridObjectReferenceRedshifts");
		gridSizes = StringUtil.getPropertyAsIntArray(properties, "gridSizes");
		gridDistancesAreRedshifts =  StringUtil.getPropertyAsBooleanArray(properties, "gridDistancesAreRedshifts");
		gridSEDNames = StringUtil.getPropertyAsStringArray(properties, "gridSEDNames");
		gridLfBandpass = StringUtil.getPropertyAsString(properties, "gridLfBandpass");
		filterName = StringUtil.getPropertyAsString(properties, "filterName");
		raBore = StringUtil.getPropertyAsDouble(properties,"raBore");
		decBore = StringUtil.getPropertyAsDouble(properties,"decBore");
		rowRef = StringUtil.getPropertyAsDouble(properties,"rowRef");
		colRef = StringUtil.getPropertyAsDouble(properties,"colRef");
		detectorXPositionMeters = StringUtil.getPropertyAsDouble(properties,"detectorXPositionMeters");
		detectorYPositionMeters = StringUtil.getPropertyAsDouble(properties,"detectorYPositionMeters");
		metersPerPixel = StringUtil.getPropertyAsDouble(properties,"metersPerPixel");
		detectorRotationDegrees = StringUtil.getPropertyAsDouble(properties,"detectorRotationDegrees");
		a7 = StringUtil.getPropertyAsDouble(properties,"a7");		
		catalogSeedIndex = StringUtil.getPropertyAsInt(properties,"catalogSeedIndex");	
		noiseSeedIndex = StringUtil.getPropertyAsInt(properties,"noiseSeedIndex");
		baseFileName = StringUtil.getPropertyAsString(properties, "baseFileName");
		psfFileName = StringUtil.getPropertyAsString(properties, "psfFileName");
		minRedshifts = StringUtil.getPropertyAsDoubleArray(properties, "minRedshifts");
		maxRedshifts = StringUtil.getPropertyAsDoubleArray(properties, "maxRedshifts");


	}
	/**
	 * @return the default properties
	 */
	public static Properties getDefaultProperties() {
		Properties retval = new Properties();
		retval.setProperty("ImageRepositoryDirectory", ".");
		retval.setProperty("nRow", "500");
		retval.setProperty("nCol", "500");
		retval.setProperty("expTime", "100");
		retval.setProperty("countsPerSecondAt20th", "1000");
		retval.setProperty("arcsecPerPixel", "0.1");
		retval.setProperty("skyMagPerArcsec2", "25.0");
		retval.setProperty("readNoiseElectrons", "0.0");
		retval.setProperty("doPoisson", "true");
		retval.setProperty("luminosityFunctionNames", "ObjectGrid");
		retval.setProperty("gridCenterRas", "20.0");
		retval.setProperty("gridCenterDecs", "30.0");
		retval.setProperty("gridSizes", "3");
		retval.setProperty("gridDeltaArcsecs", "30");
		retval.setProperty("gridAbsMags","19");
		retval.setProperty("gridObjectShapes","profile gauss 0.0 0.0 15");
		retval.setProperty("gridObjectSizeArcsecs","0.5");
		retval.setProperty("gridObjectReferenceRedshifts","10");
		retval.setProperty("gridDistancesAreRedshifts","false");		
		retval.setProperty("gridSEDNames", "CWW_E_ext");
		retval.setProperty("pixelSizeArcseconds", "0.25");

		retval.setProperty("raBore", "20.0");
		retval.setProperty("decBore", "30.0");
		retval.setProperty("rowRef", "250");
		retval.setProperty("colRef", "250");
		retval.setProperty("detectorXPositionMeters", "0.0");
		retval.setProperty("detectorYPositionMeters", "0.0");
		retval.setProperty("metersPerPixel", "25e-6");
		retval.setProperty("detectorRotationDegrees", "0.0");
		retval.setProperty("a7", "0.0");
		retval.setProperty("catalogSeedIndex", "3");
		retval.setProperty("noiseSeedIndex", "4");
		retval.setProperty("baseFileName", "defaultMakePixeledImage");
		retval.setProperty("filterName", "pegase-r");
		retval.setProperty("gridLfBandpass", "pegase-r");
		retval.setProperty("psfFileName", "null");
		retval.setProperty("minRedshifts", "10.0");
		retval.setProperty("maxRedshifts", "10.0");
		return retval;
	}

	/**
	 * create the image and write it to disk
	 */
	public void run() {
		setLocationOnSky();
		setFilter();
		setPsfFactory();
		setAstroCatalogGenerator();
		setDetectorObjects();
		setSky();
		setPixeledImageFactory();
		PixeledImage pi = pixeledImageFactory.planAndProduce();
		ImageRepository ir = new ImageRepository(imageRepositoryDirectory);
		try {
			FitsWriter fitsWriterTemp = new FitsWriter(ir.getFullFileName(baseFileName+".fit"));
			fitsWriterTemp.put(pi, FitsDataFormat.FLOAT);
			fitsWriterTemp.put(pixeledImageFactory.getDetectorObjectTables());
			fitsWriterTemp.put(pixeledImageFactory.getRenderedTables());
			fitsWriterTemp.close();
		} catch (IOException e) {
			throw new Error(e);
		}    		



	}

	private void setFilter() {
		filter = new AstronomicalFilter(filterName, wavelengthScale);
	}

	private void setDetectorObjects() {
		detectorObjects = new ArrayList<DetectorObject>();
		for (AstroObjectGenerator aog : astroCatalogGenerator.getAstroObjectGenerators()) {
			for (AstroObject ao : aog) {
				Coordinate rc = pixelToSky.inverse(ao.getLocation());
				double row = rc.getY();
				double col = rc.getX();

				double apparentMagnitude = ao.getApparentMagnitude(filter, cosmo);
				double countsPerSecond = countsPerSecondAt20th * Math.pow(10, 0.4*(20-apparentMagnitude));
				Morphology shape = ao.getShape();
				double counts = countsPerSecond * expTime;

				double newDistance;
				if (shape.getDistanceIsRedshift()) {
					newDistance = ao.getRedshift();								
				} else {
					newDistance = ao.getDistancePc();
				}
				shape.move(newDistance, cosmo);
				DetectorObject dosb = shape.makeDosb(row, col, counts, pixelSizeArcseconds);
				detectorObjects.add(dosb);
			}				
		}
	}

	private void setSky() {
		double arcsec2PerPixel = Math.pow(pixelSizeArcseconds, 2);
		double skyCountsPerSecond = countsPerSecondAt20th * Math.pow(10, 0.4*(20-skyMagPerArcsec2));
		double sky = skyCountsPerSecond*arcsec2PerPixel*expTime;
		backgroundLight = new BackgroundLightConstant(sky);
		System.out.println("MakePixeledImage.setSky:  sky="+sky);
	}
	private void setPixeledImageFactory() {
		PixelDetector pd = new PixelDetector(nRow, nCol);
		SimRandomPoisson simRandomPoisson = new SimRandomPoisson(noiseSeedIndex);
		SimRandomUniform simRandomUniform = new SimRandomUniform(catalogSeedIndex);
		SimRandomNormal simRandomNormal = new SimRandomNormal(noiseSeedIndex);
		pixeledImageFactory = new PixeledImageFactory(pd, backgroundLight, psfFactory,
				detectorObjects, cosmicRayDensity,
				simRandomPoisson, simRandomUniform, simRandomNormal);
		pixeledImageFactory.setDoPoisson(doPoisson);

	}

	private void setLocationOnSky() {
		pixelToFocalPlane = new TransformLinear(colRef, rowRef, detectorXPositionMeters, detectorYPositionMeters, metersPerPixel, detectorRotationDegrees);
		Coordinate boreSite = new Coordinate(0,0);
		pixelToFocalPlane.inverseInPlace(boreSite);
		double colRefBore = boreSite.getX();
		double rowRefBore = boreSite.getY();
		pixelToSky = new TransformTangentPlaneA7(colRefBore, rowRefBore, raBore, decBore, pixelSizeArcseconds/3600.0, detectorRotationDegrees, a7);

		double detectorMinRa = 360; 
		double detectorMaxRa = 0;
		double detectorMinDec = 90;
		double detectorMaxDec = -90;
		Coordinate[] corners = {
				new Coordinate(0.0, 0.0),
				new Coordinate(0.0, nRow),
				new Coordinate(nCol, nRow),
				new Coordinate(nCol, 0.0),
		};
		for (Coordinate corner : corners) {
			SphericalCoordinate sc = pixelToSky.forward(corner);
			sc.rotateZ(180.0-raBore);
			double ra = sc.getLongitude();
			double dec = sc.getLatitude();
			detectorMinRa = Math.min(ra, detectorMinRa);
			detectorMaxRa = Math.max(ra, detectorMaxRa);
			detectorMinDec = Math.min(dec, detectorMinDec);
			detectorMaxDec = Math.max(dec, detectorMaxDec);
		}		
		detectorMinRadec = new SphericalCoordinate(detectorMinRa, detectorMinDec);
		detectorMaxRadec = new SphericalCoordinate(detectorMaxRa, detectorMaxDec);
		detectorMinRadec.rotateZ(raBore-180.0);
		detectorMaxRadec.rotateZ(raBore-180.0);
	}		
	private void setPsfFactory() {
		if (psfFileName.equalsIgnoreCase("null")) {
			psfFactory = new PSFFactoryNull();
		} else {
			psfFactory = new PSFFactoryPixeledImage(psfDirectory, psfFileName, pixelToFocalPlane, pixelSizeArcseconds);
		}
	}

	private void setAstroCatalogGenerator() {
		//double lambdaMin = 91.0;
		//double lambdaMax = 97400.0;
		//int nBin = 100;
		//WavelengthScale ws = new WavelengthScale(lambdaMin, lambdaMax, nBin, WavelengthScale.Scaling.LOGARITHMIC);
		astroCatalogGenerator = new AstroCatalogGenerator();
		int nlf = luminosityFunctionNames.length;
		for (int ilf = 0; ilf < nlf; ilf++) {
			String lfName = luminosityFunctionNames[ilf];
			double zMin = minRedshifts[ilf];
			double zMax = maxRedshifts[ilf];
			if (lfName.equalsIgnoreCase("Zucca")) {
				Zucca zucca = new Zucca(ZuccaBandpass.V, detectorMinRadec, detectorMaxRadec, catalogSeedIndex+10, cosmo, zMin, zMax, wavelengthScale);
				astroCatalogGenerator.add(zucca);
			} else if (lfName.equalsIgnoreCase("ObjectGrid")){
				SphericalCoordinate gridCenter = new SphericalCoordinate(gridCenterRas[ilf], gridCenterDecs[ilf]);
				SED sed = new SED(wavelengthScale, gridSEDNames[ilf]);
				AstroObjectGenerator aog = new AstroObjectGeneratorGrid(detectorMinRadec, detectorMaxRadec, 
						gridCenter, gridSizes[ilf], gridDeltaArcsecs[ilf], 
						gridAbsMags[ilf], gridLfBandpass, 
						gridObjectShapes[ilf],
						gridObjectSizeArcsecs[ilf],
						gridObjectReferenceRedshifts[ilf],
						cosmo, zMin, zMax, sed, gridDistancesAreRedshifts[ilf]);
				astroCatalogGenerator.add(aog);
			} else {
				throw new IllegalArgumentException("can not make generator for luminosityFunctionName="+lfName);
			}
		}


	}

	/**
	 * saves the properties used for this image
	 * @param fileName
	 * @throws IOException
	 */
	public void saveProperties(String fileName) throws IOException {
		OutputStream writer = new FileOutputStream(fileName);
		String comments = null;
		properties.store(writer, comments);
	}
}
