package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.sim.repository.image.ImageRepository;
import gov.fnal.eag.wcs.Coordinate;
import gov.fnal.eag.wcs.SphericalCoordinate;
import gov.fnal.eag.wcs.Transform;
import gov.fnal.eag.wcs.TransformLinear;
import gov.fnal.eag.wcs.TransformTangentPlaneA7;

import java.util.ArrayList;
import java.util.LinkedList;

/**
 * deals with a set of PSFs stored in a FITS file
 * @author stoughto
 *
 */
public class PSFFactoryPixeledImage implements PSFFactory {
	ArrayList<PSFPixeledImage> psfs =  new ArrayList<PSFPixeledImage>();
	ArrayList<Coordinate> coordinates = new ArrayList<Coordinate>();
	TransformLinear pixelToFocalPlane;
	Transform pixelToSky;
	double pixelSizeArcsec;
	/**
	 * @param psfFactoryDirectory
	 * @param psfFactoryName
	 * @param pixelToFocalPlane
	 * @param pixelToSky
	 */
	public PSFFactoryPixeledImage(String psfFactoryDirectory, String psfFactoryName, TransformLinear pixelToFocalPlane, TransformTangentPlaneA7 pixelToSky) {
		this.pixelToFocalPlane = pixelToFocalPlane;
		this.pixelToSky = pixelToSky;
		this.pixelSizeArcsec = pixelToSky.arcsecPerPixel();
		loadFile(psfFactoryDirectory, psfFactoryName);
	}
	
	/**
	 * @param psfFactoryDirectory location of the psf file
	 * @param psfFactoryName name of the psf file
	 * @param pixelToFocalPlane transform from pixel to focal plane coordinates
	 * @param pixelSizeArcsec size of a pixel in arcseconds
	 */
	public PSFFactoryPixeledImage(String psfFactoryDirectory, String psfFactoryName, TransformLinear pixelToFocalPlane, double pixelSizeArcsec) {
		this.pixelToFocalPlane = pixelToFocalPlane;
		this.pixelSizeArcsec = pixelSizeArcsec;
		loadFile(psfFactoryDirectory, psfFactoryName);
	}
	
	private void loadFile(String psfFactoryDirectory, String psfFactoryName) {
		ImageRepository ir = new ImageRepository(psfFactoryDirectory);
		LinkedList<Object> list = ir.get(psfFactoryName);
		for (Object object : list) {
			if (object instanceof PixeledImage) {
				PSFPixeledImage psf = new PSFPixeledImage((PixeledImage) object);
				psfs.add(psf);
			}
		}
		if (psfs.size() == 0) {
			throw new IllegalArgumentException("there are not any PSFS in "+psfFactoryDirectory+System.getProperty("file.separator")+psfFactoryName);
		}
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.PSFFactory#convolve(gov.fnal.eag.sim.pixel.PixeledImage)
	 */
	public void convolve(PixeledImage pixeledImage) throws Exception {
		throw new UnsupportedOperationException();
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.PSFFactory#convolve(gov.fnal.eag.sim.pixel.PixeledImage, gov.fnal.eag.sim.pixel.BackgroundLight)
	 */
	public void convolve(PixeledImage bright, BackgroundLight backgroundLight) {
		throw new UnsupportedOperationException();
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.PSFFactory#getPsf(gov.fnal.eag.wcs.Coordinate)
	 */
	public PSF getPsf(Coordinate coordinate) {
		PSF retval = null;
		double separationMax = Double.MAX_VALUE;
		for (PSFPixeledImage psf : psfs) {
			double separation = psf.getXyfpMeters().linearSeparation(coordinate);
			if (separation < separationMax) {
				separationMax = separation;
				retval = psf;
			}
		}
		return retval;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.PSFFactory#getPsf(gov.fnal.eag.sim.pixel.DetectorObject)
	 */
	public PSF getPsf(DetectorObject detectorObject) {
		Coordinate coordinate = new Coordinate(detectorObject.getCol(), detectorObject.getRow());
		pixelToFocalPlane.forwardInPlace(coordinate);
		return getPsf(coordinate);
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.PSFFactory#moveTelescope(gov.fnal.eag.wcs.SphericalCoordinate)
	 */
	public void moveTelescope(SphericalCoordinate sc) {
		// Start at the center of the focal plane
        Coordinate ref = new Coordinate(0.0, 0.0);
        // Transform this back to row,col
        pixelToFocalPlane.inverseInPlace(ref);
        double boreRightAscension = sc.getLongitude();
        double boreDeclination = sc.getLatitude();
        double pixelSizeDegrees = pixelSizeArcsec/3600.0;
        double rotationDegrees = 0.0;
        double a7 = 0.0;
        pixelToSky = new TransformTangentPlaneA7(ref.getX(), ref.getY(), boreRightAscension, boreDeclination, pixelSizeDegrees, rotationDegrees, a7);
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.PSFFactory#getPsf(gov.fnal.eag.wcs.SphericalCoordinate)
	 */
	public PSF getPsf(SphericalCoordinate sc) {
		Coordinate coordinate = pixelToSky.inverse(sc);
		pixelToFocalPlane.inverseInPlace(coordinate);
		return getPsf(coordinate);
	}

}
