
package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.shapelets.functions.Decompose;
import gov.fnal.eag.sim.random.SimRandomUniform;
import gov.fnal.eag.wcs.Coordinate;

import java.awt.image.ConvolveOp;
import java.awt.geom.Point2D;

import nom.tam.fits.HeaderCardException;

//import javax.vecmath.Point2d;


/**
 * gaussian PSF
 */
public class PSFSech implements PSF {

	/** 
	 * the sigma of the sech
	 */
	double sigma;
	/**
	 * 1d profiles to help generate random numbers
	 */
	private Profiles profiles;
	/**
	 * uniform random number generator
	 */
	private SimRandomUniform uniform;
	/**
	 * the maximum radius in pixels
	 */
	private double radiusMax;
	/**
	 * radiusMax squared
	 */
	private double r2Max;
	/**
	 * sigma squared
	 */
//	private double sigma2;
	/** 
	 * normalization
	 */
	private double norm;
	/**
	 * the scale (in units of sigma) used for shapelets reconstruction
	 */
	private double sechScale = 1.5265;
	/**
	 * catalan number used to normalize the sech profile
	 */
	private double catalan = 0.9159655942;
	/** 
	 * number of rows
	 */
	private int nrow;
	/**
	 * number of columns
	 */
	private int ncol;

	private static double[][] eigenvalues;
	
	/**
	 * Use this to convolve
	 */
	ConvolveOp convolveOp = null;
	
	/**
	 * 
	 */
	PixeledImage kernel = null;

	Coordinate coordinate = new Coordinate(0,0);
	/**
	 * construct a gaussian PSF
	 * @param sigma sigma in pixels
	 */
	public PSFSech(double sigma) {
		// the sech profile decays more slowly so we need to include a larger radius, 10*sigma instead
		// of 5*sigma
		this(sigma,10.0*sigma, Profiles.DEFAULT_NPOINTS, new SimRandomUniform());
	}
	/**
	 * fully specified constructor
	 * @param sigma cutoff length of the psf
	 * @param radiusMax largest radius to consider, in pixels
	 * @param nPoints number of points to use in the profiles
	 * @param simRandomUniform a SimRandomUniform 
	 */
	public PSFSech(double sigma, double radiusMax, int nPoints, SimRandomUniform simRandomUniform) {
		profiles = new Profiles(nPoints, radiusMax);
		uniform = simRandomUniform;
		setSigma(sigma);
		this.radiusMax = radiusMax;
		r2Max = Math.pow(this.radiusMax, 2);
		nrow = (int) (1+2*Math.ceil(radiusMax));
		ncol = (int) (1+2*Math.ceil(radiusMax));		
	}
	/**
	 * set the sigma value
	 * @param sigma the value
	 */
	public void setSigma(double sigma) {
		this.sigma = sigma;
		norm = 1. / (4. * Math.PI * catalan*sigma * sigma);
		kernel = null;
		eigenvalues = null;
	}
	/**
	 * get the sigma value
	 * @return the value
	 */
	public double getSigma() {
		return sigma;
	}


	/**
	 * @param location
	 * @return the intensity
	 */
	public double intensity(Point2D.Double location) {
		double radius2 = (location.x*location.x + location.y*location.y);
		if (radius2 > r2Max) return 0;
		return norm*1.0/Math.cosh(Math.sqrt(radius2)/sigma);
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.PSF#nrow()
	 */
	public int nrow() {
		return nrow;	
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.PSF#ncol()
	 */
	public int ncol() {
		return ncol;
	}

	/**
	 * brute-force integration of flux in this pixel
	 * @param location location of the pixel
	 * @param boxSize size of the box to sample
	 * @param nSamples number of samples per pixel in each direction
	 * @return the flux
	 */
	public double integrate(Point2D.Double location, Point2D.Double boxSize, int nSamples) {
		double sum = 0;
		double dx = boxSize.x/(nSamples+1);
		double dy = boxSize.y/(nSamples+1);
		Point2D.Double loc = new Point2D.Double();
		double x,y;
		for (int ix=0; ix<nSamples; ix++){
			x = location.x-boxSize.x/2.0+(ix+1)*dx;
			loc.x = x;
			for (int iy=0; iy<nSamples; iy++) {
				y = location.y-boxSize.y/2.0+(iy+1)*dy;
				loc.y = y;
				sum += this.intensity(loc);
			}
		}
		return sum/(nSamples*nSamples);
	}

	/** (non-Javadoc)
	 * @see gov.fnal.eag.sim.pixel.PSF#nextPoint()
	 */
	public Point2D.Double nextPoint() {
		Point2D.Double next = new Point2D.Double();
		double randomNumber = uniform.next();
		int index = Profiles.binarySearch(profiles.sechProfile, randomNumber);
		double rho = sigma*profiles.radius[index];
		double thetaRad = 2.0*Math.PI*uniform.next();
		next.x = rho*Math.cos(thetaRad);
		next.y = rho*Math.sin(thetaRad);
		return next;
	}

	/** (non-Javadoc)
	 * @see gov.fnal.eag.sim.pixel.PSF#getKernel()
	 */
	public PixeledImage getKernel() {
		if (kernel == null) {
			setKernel();
		}
		return kernel;
	}

	private void setKernel() {
		kernel = new PixeledImage(this.nrow(), this.ncol());
		Point2D.Double boxSize = new Point2D.Double(1.0, 1.0);
		Point2D.Double location = new Point2D.Double();
		
		double midRow = kernel.getRowCount()/2.0;
		double midCol = kernel.getColumnCount()/2.0;
		try {
			kernel.header.addValue("midRow", midRow, "midRow");
		} catch (HeaderCardException e) {
			throw new Error(e);
		}
		try {
			kernel.header.addValue("midCol", midCol, "midCol");
		} catch (HeaderCardException e) {
			// TODO Auto-generated catch block
			throw new Error(e);
		}
		for (int r=0; r<kernel.getRowCount(); r++) {
			location.y = r+0.5-midRow;
			for (int c = 0; c<kernel.getColumnCount(); c++) {
				location.x = c+0.5-midCol;
				kernel.setByRowCol(r, c, this.integrate(location, boxSize, 4));
			}
		}
	}

	/** (non-Javadoc)
	 * @see gov.fnal.eag.sim.pixel.PSF#convolve(gov.fnal.eag.sim.pixel.PixeledImage)
	 */
	public void convolve(PixeledImage pixeledImage) {
		convolve(pixeledImage, 0.0);
		return;
	}
	
	/**
	 * convolve a pixeled image with this PSF
	 * @param pixeledImage the image to convolve
	 * @param background bacground outside the image
	 * 
	 */
	public void convolve(PixeledImage pixeledImage, double background) {
		BackgroundLight backgroundLight = new BackgroundLightConstant(background);
		convolve(pixeledImage, backgroundLight);
		return;
	}
	
	/** (non-Javadoc)
	 * @see gov.fnal.eag.sim.pixel.PSF#convolve(gov.fnal.eag.sim.pixel.PixeledImage, gov.fnal.eag.sim.pixel.BackgroundLight)
	 */
	public void convolve(PixeledImage pixeledImage, BackgroundLight backgroundLight) {
//
		//pixeledImage.printCentroid(System.out, "pixeledImage in PSFGaussian.convolve()");
		//getKernel().printCentroid(System.out, "getKernel in PSFGaussian.convolve()");
//
		pixeledImage.convolveIj(getKernel(), backgroundLight);
		return;
	}
	
	
	/** (non-Javadoc)
	 * @see gov.fnal.eag.sim.pixel.PSF#getEigenvalues()
	 */
	public double[][] getEigenvalues() {
		if (eigenvalues == null) {
			calculateEigenvalues();
		}
		return eigenvalues;
	}
	
	/**
	 * calculate the eigenvalues field
	 */
	public void calculateEigenvalues() {
		//TODO decide what nMax to use
		int nMax = 16;
		//System.out.println(" begin calculateEigenvalues with sigma="+sigma);
		eigenvalues = new double[nMax][nMax];
		//Create an array to hold analytical values of the PSF
		int size = 400;
		double nSigma = 10.;
		double internalScale = size/8./nSigma;
		//double internalScale = 2*sigma;
		//System.out.println("calculateEigenvalue: internalScale="+internalScale);
		double[][] sech = new double[size][size];
		for (int ir=0; ir<size; ir++) {
			double dr = ((size/2.0) - (ir+0.5));
			for (int ic=0; ic<size; ic++) {
				double dc = ((size/2.0) - (ic+0.5));
				double val = 1.0/Math.cosh(Math.sqrt(dr*dr + dc*dc)/internalScale);
				sech[ir][ic] = val;
			}
		}
		//BasisFunction b = new BasisFunction(nMax);
		//double scale = sigma*(size/2)/nSigma;
		//double scale = sigma;
		Decompose decompose = new Decompose(sech, size/2.0, size/2.0, sechScale*internalScale);
		eigenvalues = decompose.getFnn2(nMax);
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.PSF#getScale()
	 */
	public double getScale() {
		/*
		 *sechScale was added to make shapelets work 
		 */
		return sechScale*sigma;
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.PSF#convolve(gov.fnal.eag.sim.pixel.PSF)
	 */
	public void convolve(PSF psf) {
		throw new UnsupportedOperationException();	
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.PSF#multiplyScale(double)
	 */
	public void multiplyScale(double factor) {
		sigma *= factor;
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.PSF#getXyfpMeters()
	 */
	public Coordinate getXyfpMeters() {
		return coordinate;
	}
	/**
	 * @return the catalan value
	 */
	public double getCatalan() {
		return catalan;
	}

	
}
