
package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.sim.random.SimRandomNormal;

import java.awt.geom.Point2D;
import java.io.PrintStream;
import java.util.Iterator;
import java.util.LinkedList;

/**
 * an array of pixels with lists of PixelValueTransformations to convert photons to ADUs.
 * 
 * Note that this does not include the actual pixel values, but a representation of how pixels behave
 */
public class PixelDetector {

	/**
	 * number of rows
	 */
	int rowCount;
	/**
	 * number of columns
	 */
	int columnCount;
	/**
	 * electronic transformations from electrons to ADU such as gain, saturation, read noise, dark current, bad pixels
	 */
	private LinkedList<PixelValueTransformation> electronics = new LinkedList<PixelValueTransformation>();
	/**
	 * efficiency per pixel:  quantum efficiency, filter efficiency
	 */
	private LinkedList<PixelValueTransformation> efficiency = new LinkedList<PixelValueTransformation>();
	/**
	 * pixel width in microns
	 */
	private double pixelWidth = 10.0; 
	/**
	 * pixl thickness in microns
	 */
	private double pixelThickness = 200.0; 
	/**
	 * flag to enable diffusion
	 */
    private boolean diffusion = true; 
    /**
     * number of electrons created per micron
     */
	private double dnedx = 70.0; 
	/**
	 * diffusion width in microns after propagating 200 microns
	 */
	private double sigma = 3.5; 
	

	/**
	 * default constructor, useless by itself
	 *
	 */
	public PixelDetector() {	
		this(100,100);
	}

	/**
	 * a perfect detector with gain=1, no noise, 100% efficient
	 * @param nrow number of rows
	 * @param ncol number of columns
	 */
	public PixelDetector (int nrow, int ncol) {
		rowCount=nrow;
		columnCount=ncol;
	}
	/**
	 * a perfect detector with gain=1, no noise, 100% efficient
	 * @param nColNRow number of columns (X) and number of rows (Y)
	 */
	public PixelDetector (Point2D nColNRow) {
		this((int)nColNRow.getY(), (int)nColNRow.getX());
	}
	/**
	 * constructs a simple CCD detector
	 * @param nColNRow number of columns,rows
	 * @param quantumEfficiency constant quantum efficiency
	 * @param gain constant gain
	 * @param readnoise sigma of the constant gain
	 * @param srn normal random number generator
	 */
	public PixelDetector (Point2D nColNRow, double quantumEfficiency, double gain, double readnoise, SimRandomNormal srn) {
		this(nColNRow);
		efficiency.add(new PixelValueTransformationMultiplyConstant(quantumEfficiency, "quantum efficiency"));
		electronics.add(new PixelValueTransformationMultiplyConstant(1.0/gain, "constant gain"));
		electronics.add(new PixelValueTransformationConstantGaussianNoise(readnoise, "constant read noise", srn));		
	}
	/**
	 * get number of rows
	 * @return rowCount
	 */
	public int getRowCount() {
		return rowCount;
	}
	
	/**
	 * get number of columns
	 * @return columnCount
	 */
	public int getColumnCount() {
		return columnCount;
	}

	/**
	 * get the list of electronics transformation
	 * @return the list
	 */
	public LinkedList<PixelValueTransformation> getElectronics() {	
		return electronics;
	}
	
	/**
	 * put a transformation on the list of efficiencies
	 * @param plt PixelLevelTransformation to add
	 */
	public void putElectronics(PixelValueTransformation plt) {		
		electronics.add(plt);
	}
	/**
	 * get the list of efficiency transformation
	 * @return the list
	 */
	public LinkedList<PixelValueTransformation> getEfficiency() {	
		return efficiency;
	}
	
	/**
	 * put a transformation on the list of efficiencies
	 * @param plt PixelLevelTransformation to add
	 */
	public void putEfficiency(PixelValueTransformation plt) {		
		efficiency.add(plt);
	}
	/**
	 * set the pixel width
	 * @param pixelWidth in microns
	 */
	public void setPixelWidth(double pixelWidth) {this.pixelWidth = pixelWidth;}
	/**
	 * get the pixel width
	 * @return pixel width in microns
	 */
	public double getPixelWidth() {return pixelWidth;}
	/**
	 * set the pixel thickness
	 * @param pixelThickness in microns
	 */
	public void setPixelThickness(double pixelThickness) {this.pixelThickness = pixelThickness;}

	/**
	 * get the pixel thickness
	 * @return pixel thickness in microns
	 */	
	public double getPixelThickness() {return pixelThickness;}
	/**
	 * get the diffusion flag
	 * @return true if diffusion is included 
	 */
	public boolean getDiffusion() {return diffusion;}
	/**
	 * set the number of electrons per micron
	 * @param dnedx the value
	 */
	public void setDnedx(double dnedx) {this.dnedx = dnedx;}

	/**
	 * get the number of electrons per distance
	 * @return Dnedx in electrons per micron
	 */
	
	public double getDnedx() {return dnedx;}
	/**
	 * set the diffusion width
	 * @param sigma the width in microns after 200 microns of porpagation
	 */
	public void setSigma(double sigma) {this.sigma = sigma;}
	/**
	 * get the diffusion width, in microns, after 200 microns of propagation 
	 * @return sigma in microns
	 */
	public double getSigma() {return sigma;}

	/**
	 * set the diffusion flag
	 * @param diffusion set to true to enable diffusion
	 */
	public void setDiffusion(boolean diffusion) {
		this.diffusion = diffusion;
	}
	
	/**
	 * applies the efficiency transformations
	 * @param flux the input flux
	 * @param row row position 
	 * @param col column position
	 * @return the transformed flux
	 */
	public double efficiencyTransform(double flux, double row, double col) {
		Iterator<PixelValueTransformation> it = efficiency.iterator();
		while (it.hasNext()) {
			PixelValueTransformation pvt = it.next();
			flux = pvt.transform(flux, row, col);
		}
		return flux;
	}
	/**
	 * applies the electronics transformations
	 * @param flux the input flux
	 * @param row row position
	 * @param col column position
	 * @return the transformed flux
	 */
	public double electronicsTransform(double flux, int row, int col) {
		Iterator<PixelValueTransformation> it = electronics.iterator();
		while (it.hasNext()) {
			PixelValueTransformation pvt = it.next();
			flux = pvt.transform(flux, row, col);
		}
		return flux;
	}
	
	/**
	 * puts useful information to the PrintStream
	 * @param ps where to write
	 */
	public void report(PrintStream ps) {
		ps.println(efficiency.size()+" efficiency transforms:");
		Iterator<PixelValueTransformation> it1 = efficiency.iterator();
		while (it1.hasNext()) {
			PixelValueTransformation pvt1 = it1.next();
			pvt1.report(ps);
		}
		ps.println(electronics.size()+" electronics transforms:");
		Iterator<PixelValueTransformation> it2 = electronics.iterator();
		while (it2.hasNext()) {
			PixelValueTransformation pvt2 = it2.next();
			pvt2.report(ps);
		}
		
	}

}
