/**
 * PixelRegister.java stoughto Jan 25, 2007 11:50:10 AM
 */
package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.sim.random.SimRandom;

import java.io.PrintStream;
import java.util.ArrayList;

import org.jfree.data.xy.XYSeries;

/**
 * describes a row or column of pixels with traps for clocking charge
 * @author stoughto
 *
 */
public class PixelRegister {
	//private ArrayList<Pixel> pixels = new ArrayList<Pixel>();
	private Pixel[] pixels;
	ArrayList<SimRandom> simRandomList = new ArrayList<SimRandom>();
	
	/**
	 * @return the pixels in this register
	 */
	public Pixel[] getPixels() {
		return pixels;
	}
	/**
	 * Charge is shifted towards pixel 0
	 * @param pixels the pixels in this register
	 */
	public PixelRegister(ArrayList<Pixel> pixels) {
		this.pixels = new Pixel[pixels.size()];
		for (int i=0; i<pixels.size(); i++) {
			this.pixels[i] = pixels.get(i);
			for (SimRandom simRandom : pixels.get(i).getSimRandoms()) {
				if (!simRandomList.contains(simRandom)) {
					simRandomList.add(simRandom);
				}
			}
		}
	}
	/**
	 * @return the list of SimRandom used in this PixelRegister
	 */
	public ArrayList<SimRandom> getSimRandomList() {
		return simRandomList;
	}
	/**
	 * @return the number of pixels
	 */
	public int nPixels() {
		return pixels.length;
	}
	/**
	 * clock the charge along the register
	 * @param iPixelMax the maximum pixel to consider, to optimize normal readout modes
	 * @param time charge transfer time in seconds
	 * @param quantizeCharge true if charge is quantized
	 * @return the charge that is clocked out of pixel 0
	 */
	
	public double clock(int iPixelMax, double time, boolean quantizeCharge) {
		int iPixelMaxSafe = Math.min(iPixelMax, nPixels()-1);
		double retval = 0;
		for (int iPixel=iPixelMaxSafe; iPixel>=0; iPixel--) {
			Pixel pixel = pixels[iPixel];
			retval = pixel.transferCharge(retval, time, quantizeCharge);
		}
		return retval;
	}
	/**
	 * println the charge in each pixel
	 * @param string a message to begin the show
	 * @param out where to println
	 */
	public void show(String string, PrintStream out) {
		out.println("begin PixelRegister.show for "+string);
		int i=0;
		for (Pixel pixel : pixels) {
			out.println(i+++" charge="+pixel.getCharge()+" nTrapped="+pixel.getNTrapped());
		}		
	}
	/**
	 * @return the number of trapped electrons in this register
	 */
	public double getNTrapped() {
		double retval = 0;
		for (Pixel pixel : pixels) {
			retval += pixel.getNTrapped();
		}		
		return retval;		
	}
	
	/**
	 * @return the total untrapped charge in this register
	 */
	public double getCharge() {
		double retval = 0;
		int i=0;
		for (Pixel pixel : pixels) {
			double charge = pixel.getCharge(); 
			if (charge > 0) {
				System.out.println("PixelRegister.getCharge:  i="+i+" charge="+charge);
			}
			retval += charge;
		}		
		return retval;				
	}
	/**
	 * @return a string with the total charge and the charge in each pixel with non-zero charge
	 */
	public String getChargeString() {
		String temp = "  ";
		double totalFreeCharge = 0.0;
		double totalTrappedCharge = 0.0;
		double totalTraps = 0.0;
		int i=0;
		for (Pixel pixel : pixels) {
			totalTraps += pixel.getNTraps();
			double freeCharge = pixel.getCharge(); 
			double trappedCharge = pixel.getNTrapped();
			if (freeCharge+trappedCharge > 0 ) {
				temp += "i="+i+" "+freeCharge+":"+trappedCharge+" ";
			}
			totalFreeCharge += freeCharge;
			totalTrappedCharge += trappedCharge;
			i++;
		}		
		String retval =  "free:trapped="+totalFreeCharge+":"+totalTrappedCharge+temp+" totalTraps="+totalTraps;		
		return retval;
	}
	/**
	 * @param i
	 * @return the free charge in pixel i
	 */
	public double getCharge(int i) {
		double retval = Double.NaN;
		try {
			retval = pixels[i].getCharge();
		} catch (ArrayIndexOutOfBoundsException e) {
		}
		return retval;
	}
	/**
	 * @param i
	 * @return the trapped charge in pixel i
	 */
	public double getTrapped(int i) {
		double retval = Double.NaN;
		try {
			retval = pixels[i].getNTrapped();
		} catch (ArrayIndexOutOfBoundsException e) {
		}
		return retval;
	}
	/**
	 * @param i
	 * @return number of traps in pixel i
	 */
	public double getTraps(int i) {
		Object pixel = null;
		try {
			pixel = pixels[i];
			if (pixel instanceof PixelWithTraps) {
				return ((PixelWithTraps) pixel).getNTraps();
			}
		} catch (ArrayIndexOutOfBoundsException e){
		}
		return 0;
	}
	
	/**
	 * dumps diagnostics about this PixelRegister
	 * @param message helpful message to help you see where it got called
	 * @param ps where to dump
	 * @param iPix first pixel to list
	 * @param nPix number of pixels to list
	 */
	public void dump(String message, PrintStream ps, int iPix, int nPix) {
		ps.println(message+" ----------- iPix="+iPix+" nPix="+nPix);
		{
			String lineI       = "      i: ";
			String lineIPix    = "   iPix:";
			String lineCharge  = " charge:";
			String lineTrapped = "trapped:";
			String lineTraps   = "  traps:";
			for (int i=iPix; i<iPix+nPix; i++) {
				lineI       += String.format("%10d",(i-iPix));
				lineIPix    += String.format("%10d",i);
				lineCharge  += String.format("%10.4f",getCharge(i));
				lineTrapped += String.format("%10.4f",getTrapped(i));
				lineTraps   += String.format("%10.4f", getTraps(i));				
			}
			ps.println(lineI);
			ps.println(lineIPix);
			ps.println(lineCharge);
			ps.println(lineTrapped);
			ps.println(lineTraps);
		}
	}
	
	/**
	 * @param label
	 * @param iPix
	 * @param nPix
	 * @return a series useful for diagnostic plotting
	 */
	public XYSeries getXYSeries(String label, int iPix, int nPix) {
		XYSeries retval = new XYSeries(label);
		for (int i=iPix; i<iPix+nPix; i++) {
			double charge = getCharge(i);
			if (!Double.isNaN(charge)) {
				retval.add(i-iPix, getCharge(i));
			}
		}
		return retval;
	}
	/**
	 * @param decaySeed set the SimRandom generator to this seed
	 */
	public void setDecaySeed(int decaySeed) {
		for (SimRandom simRandom : simRandomList) {
			simRandom.setIseedIndex(decaySeed);
		}
	}
	/**
	 * @param pixelMax
	 * @return the maximum free charge from pixel 0 to pixelMax
	 */
	public double getMaxCharge(int pixelMax) {
		double retval = -Double.MAX_VALUE;
		for (int i=0; i<pixelMax; i++) {
			double charge = pixels[i].getCharge();
			retval = Math.max(retval, charge);
		}
		return retval;
	}
	/**
	 * @param string identification string
	 * @param out where to write the results of getChargeString()
	 */
	public void summarize(String string, PrintStream out) {
		out.println(string+" "+getChargeString());		
	}
	/**
	 * set the charge to zero and release trapped charge
	 */
	public void clear() {
		for (Pixel pixel : pixels) {
			pixel.setCharge(0);
			pixel.releaseTrappedCharge();
		}
	}
	/**
	 * @param pix0
	 * @param time usually 1.0
	 * @param quantizeCharge
	 * @return the charge read out from pix0 to the end
	 */
	public double[] clockCharge(int pix0, double time, boolean quantizeCharge) {
		double[] retval = new double[pixels.length-pix0];
		int iPixelMax = pixels.length;
		for (int i=0; i<pixels.length; i++) {
			double value = clock(iPixelMax, time, quantizeCharge);
			if (i >= pix0) {
				retval[i-pix0] = value;
			}
			iPixelMax--;
		}
		return retval;
	}
}
