package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.sim.repository.image.ImageRepository;
import gov.fnal.eag.util.densityEstimate.FitFunction;
import gov.fnal.eag.util.densityEstimate.FitFunctionExponentials;
import gov.fnal.eag.util.misc.StringUtil;
import gov.fnal.eag.util.stopwatch.StopwatchCollection;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;

/**
 * uses analytic formulae to estimate the effects of traps on charge transfer inefficiency
 * @author stoughto
 *
 */
public class PixelRegisterAnalytic {
	double[] rhos, taus;
	double notchDepth, power, fullWell;
	/**
	 * define the mean properties of the pixels
	 * @param rhos
	 * @param taus
	 * @param notchDepth
	 * @param power
	 * @param fullWell
	 */
	public PixelRegisterAnalytic(double[] rhos, double[] taus, double notchDepth, double power, double fullWell) {
		if (rhos==null || taus==null || taus.length != rhos.length) {
			throw new IllegalArgumentException("rhos and taus must have length of 2 but rhos="+rhos+" taus="+taus);
		}
		this.rhos = rhos;
		this.taus = taus;
		this.fullWell = fullWell;
		this.notchDepth = notchDepth;
		this.power = power;
	}
	/**
	 * uses the "version 1" formula.  A sum of exponentials where the norm is defined by the 
	 * amount of charge over sky, the lifetimes are the same as the trap lifetimes
	 * @param charge original charge per pixel
	 * @param sky value to subtract from each pixel 
	 * @return the observed clocked charge per pixel
	 */
	public double[] clockV1(double[] charge, double sky) {
		double[] denoms = new double[taus.length];
		for (int i=0; i<taus.length; i++) {
			denoms[i] = Math.exp(1.0/taus[i]) - 1.0;
		}
		double[] norms = new double[taus.length];
		double[] newCharge = new double[charge.length];
		double[] params = new double[2*taus.length];
		for (int i=0; i<charge.length; i++) {
			int pix0 = i;
			double charge0 = charge[pix0];
			if (charge0 > 0.0) {
				
				double skyNorm = Math.pow((charge0+sky-notchDepth)/(fullWell-notchDepth),power) - Math.pow((sky-notchDepth)/(fullWell-notchDepth),power);
				for (int j=0; j<taus.length; j++) {
					norms[j] = pix0*(Math.exp(1.0/taus[j])-1)*rhos[j]*skyNorm;
					params[2*j] = -1.0/taus[j];
					params[2*j+1] = norms[j];
				}
				FitFunction ffV1 = new FitFunctionExponentials(taus.length);
				ffV1.setParameters(params);
				for (int j=pix0+1; j<charge.length; j++) {
					double temp = ffV1.valueAt(j-pix0);
					newCharge[j] += temp;
				}
				double remainingCharge = charge0;
				for (int j=0; j<taus.length; j++) {
					remainingCharge -= norms[j]/denoms[j];
				}
				newCharge[pix0] = remainingCharge;
			}
		}
		return newCharge;
	}
	
	/**
	 * runs the V1 algorithm with the specified arguments (sorry, no defaults)
	 * @param args all
	 * <ul>
	 * <li> --irIn location of the input file
	 * <li> --fnIn name of the input file
	 * <li> --irOut location to write output file
	 * <li> --fnOut name of the output file
	 * <li> --parallelTrapsPerPixel comma-separated space-free list of trap density in the parallel direction
	 * <li> --parallelLifetimes comma-separated space-free list of trap lifetimes
	 * <li> --parallelNotchDepth
	 * <li> --parallelPower
	 * <li> --parallelFullWell
	 * <li> --serialTrapsPerPixel comma-separated space-free list of trap density in the serial direction
	 * <li> --serialLifetimes comma-separated space-free list of trap lifetimes
	 * <li> --serialNotchDepth
	 * <li> --serialPower
	 * <li> --serialFullWell
	 * <li> --sky
	 * <li> --showStopwatch true if you want to see the time used in each direction on stdout
	 * </ul>
	 * @throws JSAPException
	 */
	public static void main(String[] args) throws JSAPException {
		JSAP jsap = new JSAP();
		jsap.registerParameter( new FlaggedOption("irIn").setStringParser(JSAP.STRING_PARSER).setRequired(true).setLongFlag("irIn"));
		jsap.registerParameter( new FlaggedOption("fnIn").setStringParser(JSAP.STRING_PARSER).setRequired(true).setLongFlag("fnIn"));
		jsap.registerParameter( new FlaggedOption("irOut").setStringParser(JSAP.STRING_PARSER).setRequired(true).setLongFlag("irOut"));
		jsap.registerParameter( new FlaggedOption("fnOut").setStringParser(JSAP.STRING_PARSER).setRequired(true).setLongFlag("fnOut"));
		jsap.registerParameter( new FlaggedOption("parallelTrapsPerPixel").setStringParser(JSAP.STRING_PARSER).setRequired(true).setLongFlag("parallelTrapsPerPixel"));
		jsap.registerParameter( new FlaggedOption("parallelLifetimes").setStringParser(JSAP.STRING_PARSER).setRequired(true).setLongFlag("parallelLifetimes"));
		jsap.registerParameter( new FlaggedOption("parallelNotchDepth").setStringParser(JSAP.DOUBLE_PARSER).setRequired(true).setLongFlag("parallelNotchDepth"));
		jsap.registerParameter( new FlaggedOption("parallelPower").setStringParser(JSAP.DOUBLE_PARSER).setRequired(true).setLongFlag("parallelPower"));
		jsap.registerParameter( new FlaggedOption("parallelFullWell").setStringParser(JSAP.DOUBLE_PARSER).setRequired(true).setLongFlag("parallelFullWell"));
		jsap.registerParameter( new FlaggedOption("serialTrapsPerPixel").setStringParser(JSAP.STRING_PARSER).setRequired(true).setLongFlag("serialTrapsPerPixel"));
		jsap.registerParameter( new FlaggedOption("serialLifetimes").setStringParser(JSAP.STRING_PARSER).setRequired(true).setLongFlag("serialLifetimes"));
		jsap.registerParameter( new FlaggedOption("serialNotchDepth").setStringParser(JSAP.DOUBLE_PARSER).setRequired(true).setLongFlag("serialNotchDepth"));
		jsap.registerParameter( new FlaggedOption("serialPower").setStringParser(JSAP.DOUBLE_PARSER).setRequired(true).setLongFlag("serialPower"));
		jsap.registerParameter( new FlaggedOption("serialFullWell").setStringParser(JSAP.DOUBLE_PARSER).setRequired(true).setLongFlag("serialFullWell"));
		jsap.registerParameter( new FlaggedOption("sky").setStringParser(JSAP.DOUBLE_PARSER).setRequired(true).setLongFlag("sky"));
		jsap.registerParameter( new FlaggedOption("showStopwatch").setStringParser(JSAP.BOOLEAN_PARSER).setRequired(true).setLongFlag("showStopwatch"));
		JSAPResult result = jsap.parse(args);
		ImageRepository irIn = new ImageRepository(result.getString("irIn"));
		String fnIn = result.getString("fnIn");
		ImageRepository irOut = new ImageRepository(result.getString("irOut"));
		String fnOut = result.getString("fnOut");
		double[] parallelTrapsPerPixel = StringUtil.toDoubleArray(result.getString("parallelTrapsPerPixel"));
		double[] parallelLifetimes = StringUtil.toDoubleArray(result.getString("parallelLifetimes"));
		double parallelNotchDepth = result.getDouble("parallelNotchDepth");
		double parallelPower = result.getDouble("parallelPower");
		double parallelFullWell = result.getDouble("parallelFullWell");
		double[] serialTrapsPerPixel = StringUtil.toDoubleArray(result.getString("serialTrapsPerPixel"));
		double[] serialLifetimes = StringUtil.toDoubleArray(result.getString("serialLifetimes"));
		double serialNotchDepth = result.getDouble("serialNotchDepth");
		double serialPower = result.getDouble("serialPower");
		double serialFullWell = result.getDouble("serialFullWell");
		double sky = result.getDouble("sky");
		PixelRegisterAnalytic parallelPra = new PixelRegisterAnalytic(parallelTrapsPerPixel, parallelLifetimes, parallelNotchDepth, parallelPower, parallelFullWell);
		PixelRegisterAnalytic serialPra = new PixelRegisterAnalytic(serialTrapsPerPixel, serialLifetimes, serialNotchDepth, serialPower, serialFullWell);
		
		PixeledImage pi = irIn.getPixeledImage(fnIn);
		StopwatchCollection.start("PixeledImage.clockChargeAnalytic");
		pi.clockChargeAnalyticV1(parallelPra, serialPra, sky);
		StopwatchCollection.stop("PixeledImage.clockChargeAnalytic");
		irOut.put(fnOut, pi);
		if (result.getBoolean("showStopwatch")) {
			StopwatchCollection.report(System.out);
		}
	}
}
