package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.util.astronomy.SimpleFunction;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collection;

/**
 * keep track of previous exposures and add charge that persists from them
 * @author stoughto
 *
 */
public class PixelValueTransformationPersistence implements PixelValueTransformation {

	private Collection<Persist> persists;
	private double timeBegin, timeEnd;
	/**
	 * @param timeBegin time (in seconds) when this exposure begins
	 * @param timeEnd time (in seconds) when this exposure ends
	 */
	public PixelValueTransformationPersistence(double timeBegin, double timeEnd) {
		this.persists = new ArrayList<Persist>();
		this.timeBegin = timeBegin;
		this.timeEnd = timeEnd;
	}

	/**
	 * @param pixeledImage a previous image
	 * @param timeZero time (in seconds) when the persistence effect began
	 * @param a1 normalization for first term
	 * @param a2 normalization for second term
	 * @param tau1 time constant (in seconds) for first term
	 * @param tau2 time constant (in seconds) for second term
	 */
	public void addPersist(PixeledImage pixeledImage, double timeZero, double a1, double a2, double tau1, double tau2) {
		persists.add(new Persist(pixeledImage, timeZero, a1, a2, tau1, tau2));
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.PixelValueTransformation#getName()
	 */
	public String getName() {
		return String.format("persistence");
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.PixelValueTransformation#report(java.io.PrintStream)
	 */
	public void report(PrintStream ps) {
		ps.println("report for "+getName());
		for (Persist persist : persists) {
			persist.report(ps);
		}
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.PixelValueTransformation#transform(double, double, double)
	 */
	public double transform(double x, double row, double col) {
		double retval = x;
		for (Persist persist : persists) {
			retval += persist.pixeledImage.getByRowColNoOffset(row, col) * persist.evaluate(timeBegin, timeEnd);
		}
		return retval;
	}

	private class Persist {
		private PixeledImage pixeledImage;
		private double timeZero;
		private DoubleExpPersist function;
		
		private Persist(PixeledImage pixeledImage, double timeZero, double a1, double a2, double tau1, double tau2) {
			this.pixeledImage = pixeledImage;
			this.timeZero = timeZero;
			function = new DoubleExpPersist(a1, a2, tau1, tau2);
		}
		
		/**
		 * send a report on this Persist object
		 * @param ps where to send the report
		 */
		public void report(PrintStream ps) {
			String report = String.format("  persist:  timeZero=%f %s", timeZero, function.getReport());
			ps.println(report);
		}

		private double evaluate(double timeBeginIn, double timeEndIn) {
			double vBegin = function.f(timeBeginIn-timeZero);
			double vEnd = function.f(timeEndIn-timeZero);
			double retval = vEnd-vBegin;
			return retval;
		}
	}
	
	private class DoubleExpPersist extends SimpleFunction {
		private double a1, a2; // normalization
		private double tau1, tau2; // time constant, in seconds
		private DoubleExpPersist(double a1, double a2, double tau1, double tau2) {
			this.a1 = a1;
			this.a2 = a2;
			this.tau1 = tau1;
			this.tau2 = tau2;
		}
		/**
		 * @return a line summarizing this function
		 */
		public String getReport() {
			String report = String.format(" DoubleExpPersist:  a1=%f a2=%f tau1=%f tau2=%f", a1, a2, tau1, tau2);
			return report;
		}
		/**
		 * @see gov.fnal.eag.util.astronomy.SimpleFunction#f(double)
		 */
		@Override
		public double f(double t) {
			double retval = a1*(1-Math.exp(-t/tau1)) + a2*(1-Math.exp(-t/tau2));
			return retval;
		}
		
	}
}

