/**
 * PixelWithTraps.java stoughto Jan 25, 2007 4:49:46 PM
 */
package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.sim.random.SimRandom;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;

/**
 * @author stoughto
 *
 */
public class PixelWithTraps implements Pixel {
	private ArrayList<ElectronTrap> traps;
	private double charge;
	private double maxHeight;
	
	private double fullWell, power, notchDepth;
	Comparator<? super ElectronTrap> comparator = new CompareByTrapHeight();

	/**
	 * @return the traps in this pixel
	 */
	public ArrayList<ElectronTrap> getTraps() {
		return traps;
	}
	/**
	 * use default power=0.5 and notchDepth=0.0;
	 * @param traps
	 * @param charge
	 * @param fullWell
	 */
	public PixelWithTraps(ArrayList<ElectronTrap> traps, double charge, double fullWell) {
		this(traps, charge, fullWell, 0.5, 0.0);
	}
	/**
	 * @param traps the ElectronTraps that this pixel has
	 * @param charge the initial charge in this pixel; must be >= 0.
	 * @param fullWell 
	 * @param power 
	 * @param notchDepth 
	 */
	public PixelWithTraps(ArrayList<ElectronTrap> traps, double charge, double fullWell, double power, double notchDepth) {
		if (traps != null) {
			Collections.sort(traps, comparator);
		}
		this.traps = traps;
		this.charge = charge;
		this.fullWell = fullWell;
		this.power = power;
		this.notchDepth = notchDepth;
		absorbElectrons();
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.Pixel#getCharge()
	 */
	public double getCharge() {
		return charge;
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.Pixel#transferCharge(double, double, boolean)
	 */
	public double transferCharge(double newCharge, double time, boolean quantizeCharge) {
		// free up electrons to transfer out
		if (traps != null) {
			double chargeHeight = getChargeHeightFraction(charge);
			for (ElectronTrap trap : traps) {
				double decayCharge = trap.decay(time, quantizeCharge, chargeHeight);
				charge += decayCharge;
			}
		}
		// this is what will be sent out
		double retval = charge; 
		setCharge(newCharge);
		return retval;
	}

	private void absorbElectrons() {
		if (traps != null) {
			double chargeHeightFraction = getChargeHeightFraction(charge);
			maxHeight = Math.max(maxHeight, chargeHeightFraction);
			for (ElectronTrap trap : traps) {
				double trapHeight = trap.getTrapHeightFraction();
				if (trapHeight > chargeHeightFraction) {
					break;
				}
				charge -= trap.absorb(charge);
				maxHeight = Math.max(maxHeight, trapHeight);
			}
		}
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.Pixel#getNTrapped()
	 */
	public double getNTrapped() {
		double retval = 0.0;
		if (traps != null) {
			for (ElectronTrap trap : traps) {
				retval += trap.getContents();
			}
		}
		return retval;
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.Pixel#copy(double)
	 */
	public Pixel copy(double chargeIn) {
		ArrayList<ElectronTrap> newTraps = null;
		if (traps != null) {
			newTraps = new ArrayList<ElectronTrap>();
			newTraps.addAll(traps);
		}
		PixelWithTraps retval = new PixelWithTraps(newTraps, chargeIn, fullWell, power, notchDepth);
		return retval;
	}

	/**
	 * add a trap to this pixel
	 * @param trap
	 */
	public void addTrap(ElectronTrap trap) {
		if (traps == null) {
			traps = new ArrayList<ElectronTrap>();
		}
		traps.add(trap);
		Collections.sort(traps, comparator);
	}

	/**
	 * @see gov.fnal.eag.sim.pixel.Pixel#setCharge(double)
	 */
	public void setCharge(double charge) {
		// retain the new charge
		this.charge = charge;
		// see if this fills any traps
		absorbElectrons();
	}


	/**
	 * @see gov.fnal.eag.sim.pixel.Pixel#releaseTrappedCharge()
	 */
	public double releaseTrappedCharge() {
		double retval = 0.0;
		double chargeHeightFraction = getChargeHeightFraction(charge);
		
		for (ElectronTrap trap : traps) {
			double trapHeight = trap.getTrapHeightFraction();
			if (trapHeight > maxHeight) {
				break;
			}
			if (trapHeight > chargeHeightFraction) {
				retval += trap.releaseTrappedCharge();
			}
		}
		return retval;
	}
	/**
	 * @see gov.fnal.eag.sim.pixel.Pixel#setDecaySeed(int)
	 */
	public void setDecaySeed(int decaySeed) {
		if (traps != null) {
			for (ElectronTrap trap : traps) {
				trap.setDecaySeed(decaySeed);
			}
		}

	}
	/**
	 * @see gov.fnal.eag.sim.pixel.Pixel#getSimRandoms()
	 */
	public ArrayList<SimRandom> getSimRandoms() {
		ArrayList<SimRandom> retval = new ArrayList<SimRandom>();
		if (traps != null) {
			for (ElectronTrap trap : traps) {
				for (SimRandom simRandom : trap.getSimRandoms()) { 
					if (!retval.contains(simRandom)) {
						retval.add(simRandom);
					}
				}
			}
		}
		return retval;
	}
	
	/**
	 * @see gov.fnal.eag.sim.pixel.Pixel#getNTraps()
	 */
	public double getNTraps() {
		if (traps==null) return 0.0;
		return traps.size();
	}
	
	private class CompareByTrapHeight implements Comparator<ElectronTrap> {
		/**
		 * comparison based on the height of the traps
		 * @param et0 
		 * @param et1 
		 * @return -1, 1, or 0 
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 */
		public int compare(ElectronTrap et0, ElectronTrap et1) {
			double th0 = et0.getTrapHeightFraction();
			double th1 = et1.getTrapHeightFraction();
			if (th0 < th1) {
				return -1;
			} else if (th0 > th1) {
				return 1;
			} else {
				return 0;
			}
		}
		
	}
	
	/**
	 * @param chargeIn
	 * @return the height in the pixel (from 0.0 to 1.0) for this charge
	 */
	public double getChargeHeightFraction(double chargeIn) {
		double chargeHeightFraction;
		if (chargeIn <= notchDepth) {
			chargeHeightFraction = 0.0;
		} else {
			//chargeHeightFraction = Math.sqrt(chargeIn)/Math.sqrt(fullWell);
			chargeHeightFraction = Math.pow((chargeIn-notchDepth)/(fullWell-notchDepth), power);
		}
		return chargeHeightFraction;
	}

}

