
package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.shapelets.functions.Factorial;
import gov.fnal.eag.sim.random.SimRandom;
import gov.fnal.eag.sim.random.SimRandomPoisson;
import gov.fnal.eag.util.astronomy.PolySamp2;
import gov.fnal.eag.util.astronomy.SimpleFunction;
import gov.fnal.eag.util.densityEstimate.Histogram;
import gov.fnal.eag.util.densityEstimate.KernelDensityEstimate;
import gov.fnal.eag.util.densityEstimate.Polynomial2D;
import gov.fnal.eag.util.densityEstimate.UniformKernel;
import gov.fnal.eag.util.misc.Statistics;
import gov.fnal.eag.util.plot.JFreeChartUtil;
import gov.fnal.eag.util.stopwatch.StopwatchCollection;
import gov.fnal.eag.wcs.Coordinate;
import gov.fnal.eag.wcs.SphericalCoordinate;
import gov.fnal.eag.wcs.Transform;
import gov.fnal.eag.wcs.TransformTangentPlaneA7;
import ij.process.FloatProcessor;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.io.PrintStream;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Hashtable;
import java.util.List;

import jfftw.ArrayUtil;
import jfftw.FFTWComplex;
import jnt.FFT.ComplexDouble2DFFT;
import nom.tam.fits.Header;
import nom.tam.fits.HeaderCard;
import nom.tam.fits.HeaderCardException;
import nom.tam.util.Cursor;

import org.jfree.chart.JFreeChart;
import org.jfree.data.xy.XYIntervalSeries;

import cern.colt.list.DoubleArrayList;
import cern.jet.stat.Descriptive;

/**
 * @author stoughto
 *
 *
 */
/**
 * describes pixeled images
 */
public class PixeledImage {
	int rowCount, columnCount;
	/**
	 * set to true to print information while running the equals() method
	 */
	public static boolean verboseEquals = false;

	/**
	 * set to true to ignore the cards ROW0, COL0, and CHECKSUM in the equals method
	 */
	public static boolean ignoreRow0Col0Checksum = true;
	/**
	 * initial row in parent image
	 */
	private int row0 = 0;
	/**
	 * initial column in parent image
	 */
	private int col0 = 0;
	/**
	 * parent image, or null if created ex nihilo
	 */
	private PixeledImage parent = null;
	/**
	 * the pixel values
	 */
	protected double[][] pix;
	/**
	 * associated mask values, if any
	 */
	private PixeledMask mask = null;
	/**
	 * header information in addition to card required for FITS image io
	 */
	public Header header = new Header();


	private double[] fourierTransform;
	private int nRowFftw = 0;
	private int nColumnFftw = 0;

	double[] jnt;
	int nRowJnt = 0;
	int nColumnJnt = 0;

	int i0 = 0;
	int j0 = 0;
	/**
	 * constructs a PixeledImage of specific size
	 * @param nrow number of rows
	 * @param ncol number of columns
	 */
	public PixeledImage(int nrow, int ncol) {
		this(nrow, ncol, false);
	}
	/**
	 * constructs a PixeledImage of specific size
	 * @param nrow number of rows
	 * @param ncol number of columns
	 * @param withMask 
	 */
	public PixeledImage(int nrow, int ncol, boolean withMask) {
		setPix(new double[nrow][ncol]);
		if (withMask) {
			mask = new PixeledMask(nrow, ncol);
		}
	}

	/**
	 * constructs a PixeledImage from array of doubles
	 * @param pix
	 */
	public PixeledImage(double[][] pix) {
		this(pix, false);
	}
	/**
	 * constructs a PixeledImage from array of doubles
	 * @param pixIn array of pixels
	 * @param withMask true if you want to create a mask with this
	 */
	public PixeledImage(double[][] pixIn, boolean withMask) {
		setPix(new double[pixIn.length][pixIn[0].length]);
		for (int row=0; row<rowCount; row++) {
			for (int col=0; col<columnCount; col++) {
				this.pix[row][col] = pixIn[row][col];
			}
		}
		if (withMask) {
			mask = new PixeledMask(rowCount, columnCount);
		}
	}
	/**
	 * default constructor, creates 100x100 pixel image
	 *
	 */
	public PixeledImage() {
		this(100,100);
	}
	/**
	 * copy constructor
	 * @param pi the PixeledImage to copy
	 */
	public PixeledImage(PixeledImage pi) {
		row0 = pi.row0;
		col0 = pi.col0;
		parent = pi.parent;
		setPix(new double[pi.getRowCount()][pi.getColumnCount()]);
		for( int row=0; row<pi.getRowCount(); row++) {
			for (int col=0; col<pi.getColumnCount(); col++) {
				pix[row][col] = pi.pix[row][col];
			}
		}
		if (pi.mask != null) {
			mask = new PixeledMask(pi.mask);
		}
		header = headerCopy(pi.header);		
	}

	/**
	 * Copy the header card by card
	 * @param headerIn input header
	 * @return a copy 
	 */
	private static Header headerCopy(Header headerIn) {
		String[] cards = new String[headerIn.getNumberOfCards()];
		Cursor it = headerIn.iterator();
		int i=0;
		while (it.hasNext()) {
			cards[i++] = ((HeaderCard)it.next()).toString();
		}
		return new Header(cards);
	}
	/**
	 * sub-image constructor
	 * @param pi the parent PixeledImage
	 * @param nrow number of rows in the new sub-image
	 * @param ncol number of columns in the new sub-image
	 * @param row0 starting row in parent image
	 * @param col0 starting column in the parent image
	 * 
	 * Note that if row,col are out of bounds in the parent image, a zero is copied in its place
	 */
	public PixeledImage(PixeledImage pi, int nrow, int ncol, int row0, int col0) {
		this(pi, new Point2D.Float(ncol, nrow), new Point2D.Float(col0, row0), 0.0);
	}

	/**
	 * @param pi
	 * @param nColNRow
	 * @param col0Row0
	 * @param backgroundLevel
	 */

	public PixeledImage(PixeledImage pi, Point2D nColNRow, Point2D col0Row0, double backgroundLevel) {
		this(nColNRow);
		this.row0 = (int) col0Row0.getY();
		this.col0 = (int) col0Row0.getX();
		for (int row=0; row<getRowCount(); row++) {
			for (int col=0; col<getColumnCount(); col++) {
				try {
					this.pix[row][col] = pi.pix[row+row0][col+col0];
				} catch (ArrayIndexOutOfBoundsException e) {
					this.pix[row][col] = backgroundLevel;
				}
			}
		}
		if (pi.mask != null) {
			mask = new PixeledMask(pi.mask, getRowCount(), getColumnCount(), row0, col0);
		}
		parent = pi;
		///header = headerCopy(pi.header);
	}

	/**
	 * Constructor from a point that describes the col,row size
	 * @param nColNRow
	 */
	public PixeledImage(Point2D nColNRow) {
		this((int)nColNRow.getY(), (int)nColNRow.getX());
	}

	/**
	 * makes a sub image, sharing the array of pixels
	 * @param nRow number of rows in the sub image
	 * @param nCol number of columns in the sub image
	 * @param r0 starting row in this image
	 * @param c0 starting column in this image
	 * @return the sub image
	 */


	public PixeledImage subPixeledImage(int nRow, int nCol, int r0, int c0) {
		int i0Sub = r0-row0+i0;
		int j0Sub = c0-col0+j0;
		if (i0Sub+nRow > rowCount) {
			throw new IllegalArgumentException("row0="+row0+" nRow="+nRow+" rowCount="+rowCount);
		}
		if (j0Sub+nCol > columnCount) {
			throw new IllegalArgumentException("col0="+col0+" nCol="+nCol+" columnCount="+columnCount);
		}
		PixeledImage retval = new PixeledImage(1,1);
		retval.pix = pix;
		retval.i0 = r0;
		retval.j0 = c0;
		retval.rowCount = nRow;
		retval.columnCount = nCol;
		retval.setRow0Col0(row0, col0);
		if (mask != null) {
			retval.mask = mask.subPixeledMask(nRow, nCol, r0, c0);
		}
		return retval;
	}
	/**
	 * connstructs a blank PixeledImage of this size
	 * @param nRow
	 * @param nCol
	 */
	public PixeledImage(double nRow, double nCol) {
		this((int)Math.ceil(nRow), (int)Math.ceil(nCol));
	}
	/**
	 * get the pixel value at the position
	 * @param row pixel location
	 * @param col pixel location
	 * @return the value
	 */
	public double getByRowCol(int row, int col) {
		return getByRowCol(row, col, 0.0);
	}

	/**
	 * get the mask for the pixeled image
	 * @return the mask
	 */
	public PixeledMask getMask() {
		return mask;
	}

	/**
	 * get the pixel value at the position, or backgroundValue if the location is out of bounds
	 * @param row pixel location
	 * @param col pixel location
	 * @param backgroundValue value to use if out of bounds
	 * @return the value
	 */
	public double getByRowCol(int row, int col, double backgroundValue) {
		return getByRowColNoOffset(row-row0, col-col0, backgroundValue);
	}

	/**
	 * get pixel value
	 * @param row position
	 * @param col position
	 * @return the value of the pixel
	 */
	public double getByRowCol(double row, double col) {
		return getByRowCol(row, col, 0.0);
	}

	/**
	 * get pixel value
	 * @param row position
	 * @param col position
	 * @param backgroundValue default value ir row,col are out of range
	 * @return the value of the pixel or backgroundValue
	 */
	public double getByRowCol(double row, double col, double backgroundValue) {
		return getByRowColNoOffset((int)row-row0, (int)col-col0, backgroundValue);
	}
	/**
	 * get pixel value
	 * @param row position
	 * @param col position
	 * @param backgroundLight default value ir row,col are out of range
	 * @return the value of the pixel or backgroundLight
	 */
	public double getByRowCol(double row, double col, BackgroundLight backgroundLight) {
		return getByRowColNoOffset((int)row-row0, (int)col-col0, backgroundLight);
	}

	/**
	 * get pixel value
	 * @param row position, ignoring row0
	 * @param col position, ignoring col0
	 * @return pixel value
	 */
	public double getByRowColNoOffset(int row, int col) {
		return getByRowColNoOffset(row, col, 0.0);
	}

	/**
	 * get pixel value, using a default if row,col are out of range
	 * @param row position, ignoring row0
	 * @param col position, ignoring col0
	 * @param backgroundValue default to use 
	 * @return the value or the backgroundValue
	 */
	public double getByRowColNoOffset(int row, int col, double backgroundValue) {
		try {
			return pix[row+i0][col+j0];
		} catch (ArrayIndexOutOfBoundsException e) {
			return backgroundValue;
		}
	}

	/**
	 * get pixel value, using a default if row,col are out of range
	 * @param row position, ignoring row0
	 * @param col position, ignoring col0
	 * @param backgroundLight default to use 
	 * @return the value or the backgroundValue
	 */
	public double getByRowColNoOffset(int row, int col, BackgroundLight backgroundLight) {
		try {
			return pix[row+i0][col+j0];
		} catch (ArrayIndexOutOfBoundsException e) {
			return backgroundLight.getLightInPixel(row, col);
		}
	}
	/**
	 * get pixel value, or 0 if row,col are out of range
	 * @param row position, ignoring row0
	 * @param col position, ignoring col0
	 * @return the value or 0.0
	 */
	public double getByRowColNoOffset(double row, double col) {
		return getByRowColNoOffset(row, col, 0.0);
	}

	/**
	 * get pixel value, using a default if row,col are out of range
	 * @param row position, ignoring row0
	 * @param col position, ignoring col0
	 * @param backgroundValue default to use 
	 * @return the value or the backgroundValue
	 */
	public double getByRowColNoOffset(double row, double col, double backgroundValue) {
		return getByRowColNoOffset((int)row, (int)col, backgroundValue);
	}

	/**
	 * set the pixel value if row,col are in range
	 * @param row position
	 * @param col position
	 * @param value to set
	 */
	public void setByRowCol(int row, int col, double value) {
		setByRowColNoOffset(row-row0, col-col0, value);
	}
	/**
	 * set the pixel value if row,col are in range
	 * @param row position
	 * @param col position
	 * @param value to set
	 */
	public void setByRowCol(double row, double col, double value) {// do nothing if out of bounds
		setByRowColNoOffset((int)row-row0, (int)col-col0, value);
	}
	/**
	 * set the pixel value if row,col are in range
	 * @param row position ignoring row0
	 * @param col position ignoring col0
	 * @param value to set
	 */
	public void setByRowColNoOffset(int row, int col, double value) {
		try {
			pix[row+i0][col+j0] = value;
		} catch (ArrayIndexOutOfBoundsException e) {// do nothing if out of bounds
		}
	}
	/**
	 * set the pixel value if row,col are in range
	 * @param row position ignoring row0
	 * @param col position ignoring col0
	 * @param value to set
	 */
	public void setByRowColNoOffset(double row, double col, double value) {
		setByRowColNoOffset((int)row, (int) col, value);
	}
	/**
	 * increment the pixel
	 * @param row position
	 * @param col position
	 * @param value add this to the pixel
	 */
	public void incrByRowCol(int row, int col, double value) {
		incrByRowColNoOffset(row-row0, col-col0, value);
	}
	/**
	 * increment the pixel
	 * @param row position
	 * @param col position
	 * @param value add this to the pixel
	 */
	public void incrByRowCol(double row, double col, double value) {
		incrByRowColNoOffset((int)row-row0, (int)col-col0, value);
	}
	/**
	 * increment the pixel, ignoring row0,col0
	 * @param row position
	 * @param col position
	 * @param value add this to the pixel
	 */
	public void incrByRowColNoOffset(double row, double col, double value) {
		incrByRowColNoOffset((int) row, (int) col, value);
	}
	/**
	 * increment the pixel, ignoring row0,col0
	 * @param row position
	 * @param col position
	 * @param value add this to the pixel
	 */
	public void incrByRowColNoOffset(int row, int col, double value) {
		try {
			pix[row+i0][col+j0] += value;
		} catch (ArrayIndexOutOfBoundsException e) {// do nothing if out of bounds
		}
	}
	/**
	 * scale the pixel
	 * @param row position
	 * @param col position
	 * @param value multiply the pixel by this value
	 */
	public void multiplyByRowCol(int row, int col, double value) {
		multiplyByRowColNoOffset(row-row0, col-col0, value);
	}
	/**
	 * scale the pixel
	 * @param row position
	 * @param col position
	 * @param value multiply the pixel by this value
	 */
	public void multiplyByRowCol(double row, double col, double value) {
		multiplyByRowColNoOffset((int)row-row0, (int)col-col0, value);
	}
	/**
	 * scale the pixel, ignoring row0,col0
	 * @param row position
	 * @param col position
	 * @param value multiply the pixel by this value
	 */
	public void multiplyByRowColNoOffset(double row, double col, double value) {
		multiplyByRowColNoOffset((int) row, (int) col, value);
	}
	/**
	 * scale the pixel, ignoring row0,col0
	 * @param row position
	 * @param col position
	 * @param value multiply the pixel by this value
	 */
	public void multiplyByRowColNoOffset(int row, int col, double value) {
		try {
			pix[row+i0][col+j0] *= value;
		} catch (ArrayIndexOutOfBoundsException e) {// do nothing if out of bounds
		}
	}
	/**
	 * set all pixels to one value
	 * @param scalar the value to set
	 */
	public void setScalar(double scalar) {
		for (int row=0; row<pix.length; row++) {
			for (int col=0; col<pix[0].length; col++) {
				pix[row][col] = scalar;
			}
		}
	}

	/**
	 * set the mask for the pixeled image
	 * @param mask the mask to set
	 */
	public void setMask(PixeledMask mask) {
		if (mask.getRowCount() == getRowCount() && mask.getColumnCount() == getColumnCount()) {
			this.mask = mask;
		}
		else {
			throw new Error("Mask is not the same size as the pixeled image.");
		}
	}
	/**
	 * set the pixel array to the pixel aray from in
	 * @param in
	 */
	public void set(PixeledImage in) {
		if (getRowCount() != in.getRowCount()) return;
		if (getColumnCount() != in.getColumnCount()) return;
		setPix(in.pix);
	}
	/**
	 * add the same value to all pixels
	 * @param scalar the value to add
	 */
	public void addScalar(double scalar) {
		for (int row=row0; row<getRowMax(); row++) {
			for (int col=col0; col<getColumnMax(); col++) {
				incrByRowCol(row, col, scalar);
			}
		}
	}
	/**
	 * multiply all pixels by the same value
	 * @param scalar the value by which to multiply
	 */
	public void multiplyScalar(double scalar) {
		for (int row=row0; row<getRowMax(); row++) {
			for (int col=col0; col<getColumnMax(); col++) {
				multiplyByRowCol(row, col, scalar);
			}
		}
	}
	/**
	 * divide all pixels by the same value
	 * @param scalar the value by which to divide
	 */
	public void divideScalar(double scalar) {
		for (int row=row0; row<getRowMax(); row++) {
			for (int col=col0; col<getColumnMax(); col++) {
				multiplyByRowCol(row, col, 1.0/scalar);
			}
		}
	}

	/**
	 * add the pixel values of pi to this PixeledImage, using row0,col0 for both
	 * @param pi the input PixeledImage
	 */
	public void add(PixeledImage pi) {
		if (pi != null) {
			for (int row=getRowMin(); row<getRowMax(); row++) {
				for (int col=getColumnMin(); col<getColumnMax(); col++) {
					incrByRowCol(row, col, pi.getByRowCol(row,col));
				}
			}
		}
	}
	/**
	 * add the pixel values of the pi to this PixeledImage, normalizing the values to add a total of nPhotons
	 * @param pi 
	 * @param nPhotons
	 */
	public void add(PixeledImage pi, double nPhotons) {
		double normalization = pi.getSum()/nPhotons;
		for (int row=pi.row0; row<pi.getRowMax(); row++) {
			for (int col=pi.col0; col<pi.getColumnMax(); col++) {
				incrByRowCol(row, col, pi.getByRowCol(row,col)/normalization);
			}
		}
	}
	/**
	 * add the pixel values
	 * @param pi input pixel values
	 * @param pd use these efficiencies for the input pixel values
	 */
	public void add(PixeledImage pi, PixelDetector pd) {
		for (int row=pi.row0; row<pi.getRowMax(); row++) {
			for (int col=pi.col0; col<pi.getColumnMax(); col++) {
				incrByRowCol(row, col, pd.efficiencyTransform(pi.getByRowCol(row,col), row+0.5, col+0.5));
			}
		}
	}

	/**
	 * add the pixel values
	 * @param pi input pixel values
	 * @param pd use these efficiencies for the input pixel values
	 * @param nPhotons the values to add nPhotons before efficiency
	 */
	public void add(PixeledImage pi, PixelDetector pd, double nPhotons) {    	
		double normalization = pi.getSum()/nPhotons;
		for (int row=pi.row0; row<pi.getRowMax(); row++) {
			for (int col=pi.col0; col<pi.getColumnMax(); col++) {
				incrByRowCol(row, col, pd.efficiencyTransform(pi.getByRowCol(row,col), row+0.5, col+0.5)/normalization);
			}
		}
	}

	/**
	 * subtract the pixel values of pi from this PixeledImage, using this size
	 * @param pi the other PixeledImage
	 */
	public void subtract(PixeledImage pi) {
		for (int row=pi.row0; row<pi.getRowMax(); row++) {
			for (int col=pi.col0; col<pi.getColumnMax(); col++) {
				double value = pi.getByRowCol(row, col);
				incrByRowCol(row, col, -value);
			}
		}	
	}
	/**
	 * multiply the pixel values of pi by this PixeledImage, using this size
	 * @param pi the other PixeledImage
	 */
	public void multiply(PixeledImage pi) {
		for (int row=pi.row0; row<pi.getRowMax(); row++) {
			for (int col=pi.col0; col<pi.getColumnMax(); col++) {
				double value = pi.getByRowCol(row, col);
				multiplyByRowCol(row, col, value);
			}
		}				
	}
	/**
	 * divide the pixel values of pi into this PixeledImage, using pi size
	 * @param pi the other PixeledImage
	 */
	public void divide(PixeledImage pi) {
		for (int row=pi.getRowMin(); row<pi.getRowMax(); row++) {
			for (int col=pi.getColumnMin(); col<pi.getColumnMax(); col++) {
				multiplyByRowCol(row, col, 1.0/pi.getByRowCol(row,col));
			}
		}				

	}
	/**
	 * copy the pixel values of pi to this PixeledImage, using this size
	 * @param pi the other PixeledImage
	 */

	public void copy(PixeledImage pi) {
		for (int row=pi.getRowMin(); row<pi.getRowMax(); row++) {
			for (int col=pi.getColumnMin(); col<pi.getColumnMax(); col++) {
				setByRowCol(row,col,pi.getByRowCol(row,col));
			}
		}
	}

	/**
	 * copy the pixel values of pi to this PixeledImage, starting at r0,c0 of this, for all pi pixels
	 * @param pi the other PixeledImage
	 * @param r0 starting row
	 * @param c0 starting column
	 */
	public void copy(PixeledImage pi, int r0, int c0) {
		for (int row=pi.getRowMin(); row < pi.getRowMax(); row++) {
			for (int col=pi.getColumnMin(); col < pi.getColumnMax(); col++) {
				setByRowCol(row+r0, col+c0, pi.getByRowCol(row, col));
			}
		}
		/*
		for (int row=0; row<pi.getRowCount(); row++) {
			for (int col=0; col<pi.getColumnCount(); col++) {
				pix[r0+row][c0+col] = pi.pix[row][col];
			}
		}						
		 */
	}

	/**
	 * copy the pixel values while dividing by the flat values
	 * @param pi values to copy
	 * @param r0 r0 in this PixeledImage to place pi values
	 * @param c0 c0 in this PixeledImage to place pi values
	 * @param flat values to divide into pi values
	 */
	public void copyAndDivideByFlat(PixeledImage pi, int r0, int c0, PixeledImage flat) {		
		for (int row=0; row<pi.getRowCount(); row++) {
			for (int col=0; col<pi.getColumnCount(); col++) {
				pix[r0+row][c0+col] = pi.pix[row][col]/flat.pix[row][col];
			}
		}						

	}

	/**
	 * copy the pixel values of this PixeledImage back to its parent
	 */

	public void copyBack() {
		for (int row=0; row<getRowCount(); row++) {
			for (int col=0; col<getColumnCount(); col++) {
				double value = getByRowColNoOffset(row,col);
				parent.setByRowColNoOffset(row+row0, col+col0, value);
			}
		}				
		if (mask != null) {
			mask.copyBack();
		}
	}
	/**
	 * add the pixel values of this PixeledImage back to its parent
	 *
	 */
	public void addBack() {
		for (int row=0; row<getRowCount(); row++) {
			for (int col=0; col<getColumnCount(); col++) {
				double value = getByRowColNoOffset(row,col);
				parent.incrByRowColNoOffset(row+row0, col+col0, value);
			}
		}				
		if (mask != null) {
			mask.orBack();		
		}
	}
	/**
	 * apply the transformation to each pixel
	 * @param transformation
	 */
	public void apply(PixelValueTransformation transformation) {
		for (int row=0; row<getRowCount(); row++) {
			for (int col=0; col<getColumnCount(); col++) {
				pix[row][col] = transformation.transform(pix[row][col], row, col);
			}
		}
	}
	/**
	 * apply the transformation to each pixel, calling it with row,col of each pixel
	 * @param transformation
	 */
	public void applyWithRowCol(PixelValueTransformation transformation) {
		for (int row=0; row<getRowCount(); row++) {
			for (int col=0; col<getColumnCount(); col++) {
				pix[row][col] = transformation.transform(pix[row][col], row, col);
			}
		}
	}
	/**
	 * check to see if all pixel values are legal
	 * @return true if they are all legal
	 */
	public boolean legalValues() {
		for (int row=0; row<getRowCount(); row++) {
			for (int col=0; col<getColumnCount(); col++) {
				if (Double.isNaN(pix[row][col])) {return false;}
				if (Double.isInfinite(pix[row][col])) {return false;}
				if (Math.abs(pix[row][col]) == Double.MAX_VALUE) {return false;}
				if (Math.abs(pix[row][col]) == Double.MIN_VALUE) {return false;}				
			}
		}
		return true;
	}
	/**
	 * calculate the XOR of the pixel values
	 * @return the sum
	 */
	public long getXor() {
		long xor = 0;
		for (int row=0; row<getRowCount(); row++) {
			for (int col=0; col<getColumnCount(); col++) {
				xor ^= Double.doubleToLongBits(pix[row][col]);
			}
		}
		return xor;
	}
	/**
	 * fluctuate each pixel value by choosing a random number from a Poisson distribution defined with a mean of the original pixel value
	 * @param srp SimRandomPoisson to control generation
	 *
	 */
	public void poisson(SimRandomPoisson srp) {		
		poisson(srp, 1.0);
	}

	/**
	 * @param srp
	 * @param poissonFactor 1.0 for poisson statistics, ranging down to 0.0 for no poisson statistics
	 */
	public void poisson(SimRandomPoisson srp, double poissonFactor) {
		if (poissonFactor > 0.0) {
			for (int row=getRowMin(); row<getRowMax(); row++) {
				for (int col=getColumnMin(); col<getColumnMax(); col++) {
					double mean = getByRowCol(row,col);
					mean /= poissonFactor;
					double newValue = srp.next(mean);
					newValue *= poissonFactor;
					setByRowCol(row, col, newValue);
				}
			}
		}
	}

	/**
	 * add random value to pixel value
	 * @param sr random value
	 */
	public void addRandomNumber(SimRandom sr) {
		for (int row=getRowMin(); row<getRowMax(); row++) {
			for (int col=getColumnMin(); col<getColumnMax(); col++) {
				incrByRowCol(row,col,sr.next());
			}
		}
	}
	/**
	 * figure out whether everything in pi is equal to this 
	 * @param pi the other PixeledImage
	 * @return true if they are equal
	 */
	public boolean equals(PixeledImage pi) {
		return equals(pi, true, true, true);
	}
	/**
	 * figure out whether the things indicated by the boolean values are equal
	 * @param pi the other PixeledImage
	 * @param pixValues true if you want to compare pixel values
	 * @param maskValues true if you want to compare mask values
	 * @param headerValues true if you want to compare header values (keyword/value pairs, not comments, ignoring order)
	 * @return true if they are equal
	 */
	public boolean equals(PixeledImage pi, boolean pixValues, boolean maskValues, boolean headerValues) {
		return equals(pi, 0.0, pixValues, maskValues, headerValues);
	}
	/**
	 * figure out whether everything in pi is equal to this, with a tolerence on the double pixel values
	 * @param pi the other PixeledImage
	 * @param tolerence maximum allowed difference for two doubles to be considered equal
	 * @return true if they are equal
	 */
	public boolean equals(PixeledImage pi, double tolerence) {
		return equals(pi, tolerence, true, true, true);
	}

	/**
	 * figure out whether the things indicated by the boolean values are equal, with a tolerence on the double pixel values
	 * @param pi the other PixeledImage
	 * @param tolerence 
	 * @param pixValues true if you want to compare pixel values
	 * @param maskValues true if you want to compare mask values
	 * @param headerValues true if you want to compare header values (keyword/value pairs, not comments, ignoring order)
	 * @return true if they are equal
	 */
	public boolean equals(PixeledImage pi, double tolerence, boolean pixValues, boolean maskValues, boolean headerValues) {
		boolean retval = true;
		if (pixValues) {
			retval = equalsRowAndColumnCount(pi, retval);
			if (!retval) return retval;
		}
		if (pixValues | maskValues) {
			retval = equalsPixelAndMaskValues(pi, tolerence, pixValues, maskValues, retval);
			if (!retval) return retval;
		}
		if (headerValues) {
			boolean rva = headersEquivalent(header, pi.header);
			if (verboseEquals && !rva) {
				System.out.println(" PixeledImage equals:  done headersEquivalent(header, pi.header)");
			}
			boolean rvb = headersEquivalent(pi.header, header);
			if (verboseEquals && !rvb) {
				System.out.println(" PixeledImage equals:  done headersEquivalent(pi.header, header)");
			}
			retval = rva && rvb;
			if (!retval) {
				if (verboseEquals) {
					headerPrint(System.out, header, "this header");
					headerPrint(System.out, pi.header, "pi header");            		
				}
			}
			return retval;
		} 
		return true;
	}
	private void headerPrint(PrintStream ps, Header headerIn, String message) {
		ps.println(" headerPrint for message="+message);
		Cursor it = headerIn.iterator();
		while (it.hasNext()) {
			ps.println(it.next());
		}
	}

	/**
	 * copy these header cards to the header of this PixeledImage
	 * @param headerIn
	 */
	public void copyHeaderCards(Header headerIn) {
		Cursor it = headerIn.iterator();
		while (it.hasNext()) {
			HeaderCard headerCard = (HeaderCard) it.next();
			header.addLine(headerCard);
		}
	}
	private boolean equalsPixelAndMaskValues(PixeledImage pi, double tolerence, boolean pixValues, boolean maskValues, boolean retval) {
		if (pixValues) {
			retval = equalsPixel(pi, tolerence, retval);
		}
		if (retval && maskValues && mask != null && pi.mask != null) {
			retval = equalsMask(pi, retval);
		}
		return retval;
	}
	private boolean equalsMask(PixeledImage pi, boolean retval) {
		for (int row=0; row<pi.getRowCount(); row++) {
			for (int col=0; col<pi.getColumnCount(); col++) {
				BigInteger thisMaskValue = mask.getByRowColNoOffset(row, col);
				BigInteger piMaskValue   = pi.mask.getByRowColNoOffset(row, col);
				if (thisMaskValue.compareTo(piMaskValue) != 0 ) {
					if (verboseEquals) {
						System.out.println(" PixeledImage equals:  row="+row+" col="+col+" thisMaskValue="+thisMaskValue+" piMaskValue="+piMaskValue);
					}
					retval =  false;
					break;
				}
			}
		}
		return retval;
	}
	private boolean equalsPixel(PixeledImage pi, double tolerence, boolean retval) {
		for (int row=0; row<pi.getRowCount(); row++) {
			for (int col=0; col<pi.getColumnCount(); col++) {
				double d = Math.abs(this.pix[row][col]-pi.pix[row][col]);
				if (d > tolerence ) {
					if (verboseEquals) { 
						System.out.println(" PixeledImage equals:  row="+row+" col="+col+" this.pix="+this.pix[row][col]+" pi.pix="+pi.pix[row][col]+" d="+d+" tolerence="+tolerence);
					}
					retval = false;
					break;
				}
			}
		}
		return retval;
	}
	private boolean equalsRowAndColumnCount(PixeledImage pi, boolean retval) {
		if (this.getRowCount() != pi.getRowCount()) {
			if (verboseEquals) {
				System.out.println(" row dimensions not equal:  this.getRowCount()="+this.getRowCount()+" pi.getRowCount()="+pi.getRowCount());
			}
			retval = false;
		}
		if (this.getColumnCount() != pi.getColumnCount()) {
			if (verboseEquals) {
				System.out.println(" col dimensions not equal:  this.getColumnCount()="+this.getColumnCount()+" pi.getColumnCount()="+pi.getColumnCount());
			}
			retval = false;
		}
		return retval;
	}
	/**
	 * see if the two headers have equivalent values
	 * @param header2
	 * @param header3
	 * @return true if the all the keyword/value pairs in header 2 exist in header 3 and have the same value
	 */
	public static boolean headersEquivalent(Header header2, Header header3) {
		Cursor it2 = header2.iterator();
		while (it2.hasNext()) {
			HeaderCard hc2 = (HeaderCard) it2.next();			
			String key2 = hc2.getKey();
			if (!ignoreThisKey(key2)) {
				HeaderCard hc3 = header3.findCard(key2);
				if (hc3 == null) {
					if (verboseEquals) {
						System.out.println(" PixeledImage equals:  key2="+key2+" not found in the other header");                    
						System.out.println(" PixeledImage equals:  header2:");
						dumpHeader(header2, System.out);
						//header2.dumpHeader(System.out);
						System.out.println(" PixeledImage equals:  header3:");
						dumpHeader(header3, System.out);
						//header3.dumpHeader(System.out);
					}
					return false;
				}
				if (!hc2.getValue().equals(hc3.getValue())) {
					if (verboseEquals) {
						System.out.println(" PixeledImage equals:  hc2.getValue()="+hc2.getValue()+" hc3.getValue()="+hc3.getValue());
						System.out.println(" PixeledImage equals:  header2:");
						dumpHeader(header2, System.out);
						//header2.dumpHeader(System.out);
						System.out.println(" PixeledImage equals:  header3:");
						dumpHeader(header3, System.out);
						//header3.dumpHeader(System.out);
					}
					return false;
				}
			}
		}
		return true;
	}
	private static boolean ignoreThisKey(String key2) {
		if (key2.equalsIgnoreCase("SIMPLE")) return true;
		if (key2.equalsIgnoreCase("BITPIX")) return true;
		if (key2.equalsIgnoreCase("NAXIS")) return true;
		if (key2.equalsIgnoreCase("NAXIS1")) return true;
		if (key2.equalsIgnoreCase("NAXIS2")) return true;
		if (key2.equalsIgnoreCase("EXTEND")) return true;
		if (key2.equalsIgnoreCase("BZERO")) return true;
		if (key2.equalsIgnoreCase("BSCALE")) return true;
		if (key2.equalsIgnoreCase("END")) return true;
		if (ignoreRow0Col0Checksum) {
			if (key2.equalsIgnoreCase("PIROW0")) return true;
			if (key2.equalsIgnoreCase("PICOL0")) return true;
			if (key2.equalsIgnoreCase("CHECKSUM")) return true;			
		}
		return false;
	}

	/**
	 * dumps the header to out
	 * @param header
	 * @param out
	 */
	public static void dumpHeader(Header header, PrintStream out) {
		Cursor cursor = header.iterator();
		while (cursor.hasNext()) {
			HeaderCard hc = (HeaderCard) cursor.next();
			out.println("key="+hc.getKey()+" value="+hc.getValue()+" comment="+hc.getComment());
		}
	}
	/**
	 * get the number of rows
	 * @return the number of rows
	 */
	public int getRowCount() {
		return rowCount;
	}
	/**
	 * get the number of columns
	 * @return the number of columns
	 */
	public int getColumnCount() {
		return columnCount;
	}
	/**
	 * get the initial row in the parent
	 * @return the initial row
	 */
	public int getRow0() {
		return row0;
	}
	/** 
	 * get the initial column in the parent
	 * @return the initial column
	 */
	public int getCol0() {
		return col0;
	}
	/**
	 * organize the pixels into a double array list, to be fed to
	 * cern.jet.stat.Descriptive to get all kinds of descriptive statistics.
	 * 
	 * @return the double array list
	 */
	public DoubleArrayList getDoubleArrayList() {
		return getDoubleArrayList(null);
	}

	/**
	 * organize the pixels into a double array list, to be fed to
	 * cern.jet.stat.Descriptive to get all kinds of descriptive statistics.
	 * 
	 * @param maskBits if not null and if this.mask is not null, include only pixels where mask.pix & maskBits == maskBits
	 * @return dal the sorted DoubleArrayList
	 */
	public DoubleArrayList getDoubleArrayList(BigInteger maskBits) {

		DoubleArrayList dal = new DoubleArrayList();
		if (maskBits != null && maskBits.compareTo(BigInteger.ZERO)!=0 && this.mask != null) {
			for (int r = 0; r < pix.length; r++) {
				for (int c = 0; c < pix[0].length; c++) {
					if (mask.getByRowColNoOffset(r, c).and(maskBits).compareTo(maskBits) == 0) {
						dal.add(pix[r][c]);
					}
				}
			}	
		} else {
			for (int r = 0; r < pix.length; r++) {
				for (int c = 0; c < pix[0].length; c++) {
					dal.add(pix[r][c]);
				}
			}	

		}
		dal.quickSort();
		return dal;
	}
	/**
	 * get the extrema
	 * @return a Hastable, with keys of min and max, with Double values
	 */
	public Hashtable<String, Double> extrema() {
		return extrema(null);
	}

	/**
	 * find min and max pixel, where the pixel is accepted 
	 * @param maskBits mask acceptance bits
	 * @return with keywords min and max
	 */
	public Hashtable<String, Double> extrema(BigInteger maskBits) {
		Hashtable<String, Double> retval = new Hashtable<String, Double>();
		double min = Double.MAX_VALUE;
		double max = -Double.MAX_VALUE;
		for (int row=getRowMin(); row<getRowMax(); row++) {
			for (int col=getColumnMin(); col<getColumnMax(); col++) {
				if (pixelAccepted(row, col, maskBits)) {
					double temp = getByRowCol(row,col);
					if (temp<min) {
						min = temp;
					}
					if (temp>max) {
						max = temp;
					}
				}
			}
		}
		retval.put("min", new Double(min));
		retval.put("max", new Double(max));
		return retval;
	}
	/**
	 * calculate centroid and width, after subtracting a sky value from each
	 * pixel
	 * 
	 * @param sky value to subtract from each pixel value
	 * @return a Hashtable with nPixels, rowCentroid, colCentroid, rowSigma, and colSigma
	 */
	public Hashtable<String,Number> centroid(double sky) {

		PixeledImage segmentation = null;
		int iSegment = 0;
		return centroid(sky, segmentation, iSegment);

	}
	/**
	 * calculate centroid and width, after subtracting a sky value from each
	 * pixel
	 * 
	 * @param sky value to subtract from each pixel value
	 * @param maskBits if not null and if this.mask is not null, include only pixels where mask.pix & maskBits == maskBits
	 * @return Hashtable of nPixels, rowCentroid, colCentroid, rowSigma, colSigma
	 */

	public Hashtable<String, Number> centroid(double sky, BigInteger maskBits) {
		Hashtable<String, Number> hashtable = new Hashtable<String, Number>();
		double sum = 0.0;
		double weight;
		double rowSum = 0.0;
		double colSum = 0.0;
		double row2Sum = 0.0;
		double col2Sum = 0.0;
		int nPixels = 0;
		int rowMin = getRowMin();
		int rowMax = getRowMax();
		int colMin = getColumnMin();
		int colMax = getColumnMax();
		if (maskBits != null && maskBits.compareTo(BigInteger.ZERO)!=0 && this.mask != null) {
			for (int r = rowMin; r < rowMax; r++) {
				double rCent = r + 0.5;
				for (int c = colMin; c < colMax; c++) {
					double cCent = c + 0.5;

					if (maskBits.compareTo(mask.getByRowCol(r,c)) == 0) {
						weight = getByRowCol(r,c) - sky;
						sum += weight;
						rowSum += rCent * weight;
						colSum += cCent * weight;
						row2Sum += rCent * rCent * weight;
						col2Sum += cCent * cCent * weight;
						nPixels++;
					}
				}
			}
		} else {
			for (int r = rowMin; r < rowMax; r++) {
				double rCent = r + 0.5;
				for (int c = colMin; c < colMax; c++) {
					double cCent = c + 0.5;
					weight = getByRowCol(r,c) - sky;
					sum += weight;
					rowSum += rCent * weight;
					colSum += cCent * weight;
					row2Sum += rCent * rCent * weight;
					col2Sum += cCent * cCent * weight;
					nPixels++;
				}
			}
		}
		hashtable.put("nPixels", new Integer(nPixels));
		hashtable.put("rowCentroid", new Double(rowSum / sum));
		hashtable.put("colCentroid", new Double(colSum / sum));
		double s2 = sum * sum;
		hashtable.put("rowSigma", new Double(Math.sqrt(row2Sum / sum
				- rowSum * rowSum / s2)));
		hashtable.put("colSigma", new Double(Math.sqrt(col2Sum / sum
				- colSum * colSum / s2)));
		return hashtable;
	}
	/**
	 * calculate the unweighted first and second moments in each direction
	 * @param sky value to subtract from each pixel
	 * @param segmentation	segmentation.fits image from SExtractor
	 * @param iSegment object number from SExtractor
	 * @return hashtable of nPixels, rowCentroid, colCentroid, rowSigma, colSigma
	 */
	public Hashtable<String, Number> centroid(double sky, PixeledImage segmentation, int iSegment) {
		return centroid(sky, segmentation, iSegment, 0.0, 0.0, 0.0);
	}

	/**
	 * simple but slow way to calculate the first and second moments in each direction
	 * @param sky value to subtract from each pixel
	 * @param segmentation similar to segmentation.fits from SExtractor
	 * @param iSegment object number use only pixels that have this value in segmentation
	 * @param sigmaWeight width of gaussian weight function (pixels)
	 * @param rowWeight position of gaussian weight function (pixels)
	 * @param colWeight position of gaussian weight function (pixels)
	 * @return hashtable of nPixels, rowCentroid, colCentroid, rowSigma, colSigma
	 */
	public Hashtable<String, Number> centroid(double sky, PixeledImage segmentation, int iSegment, double sigmaWeight, double rowWeight, double colWeight) {
		double rowRef = 0.0;
		double colRef = 0.0;
		Hashtable<String, Number> hashtable = sums(sky, rowRef, colRef, segmentation, iSegment, sigmaWeight, rowWeight, colWeight);
		appendMoments(hashtable);
		return hashtable;
	}
	/**
	 * @param sky to subract from each pixel
	 * @param rowRef approximate row to help avoid rounding errors
	 * @param colRef approximate col to help avoid rounding errors
	 * @param segmentation use only pixels set to iSegment
	 * @param iSegment object number from SEctractor
	 * @param sigmaWeight width of gaussian weight function (pixels)
	 * @param rowWeight position of gaussian weight function (pixels)
	 * @param colWeight position of gaussian weight function (pixels)
	 * @return appended with weightSum, rowSum, colSum, row2Sum, col2Sum, rowcolSum, nPixels, rowRef, colRef
	 */
	public Hashtable<String, Number> sums(double sky, double rowRef, double colRef, PixeledImage segmentation, int iSegment, double sigmaWeight, double rowWeight, double colWeight) {
		Hashtable<String, Number> hashtable = new Hashtable<String, Number>();
		double weightSum = 0.0;
		double weight;
		double rowSum = 0.0;
		double colSum = 0.0;
		double row2Sum = 0.0;
		double col2Sum = 0.0;
		double rowcolSum = 0.0;
		/*double mx2 = 0.0;
		double mean = 0.0;
		int xMin = 30;
		int yMin = 30;
		int meanCount = 100;*/
		int nPixels = 0;
		int rowMin = getRowMin();
		int rowMax = getRowMax();
		int colMin = getColumnMin();
		int colMax = getColumnMax();
		for (int r = rowMin; r < rowMax; r++) {
			double rCent = r + 0.5;
			double deltaRow2 = Math.pow(rCent-rowWeight, 2);
			for (int c = colMin; c < colMax; c++) {
				if (segmentation==null || segmentation.getByRowCol(r,c) == iSegment) {
					double cCent = c + 0.5;
					weight = getByRowCol(r,c) - sky;
					if (sigmaWeight > 0) {
						double deltaCol2 = Math.pow(cCent-colWeight, 2);
						double gaussWeight = Math.exp(-(deltaCol2+deltaRow2)/sigmaWeight/sigmaWeight/2.0);
						weight *= gaussWeight;
					}
					double deltaRow = rCent-rowRef;
					double deltaCol = cCent-colRef;
					weightSum += weight;
					rowSum    += deltaRow * weight;
					colSum    += deltaCol * weight;
					row2Sum   += deltaRow * deltaRow * weight;
					col2Sum   += deltaCol * deltaCol * weight;
					rowcolSum += deltaRow * deltaCol * weight;
					nPixels++;	
				}
			}
		}
		hashtable.put("weightSum", new Double(weightSum));
		hashtable.put("rowSum", new Double(rowSum));
		hashtable.put("colSum", new Double(colSum));
		hashtable.put("row2Sum", new Double(row2Sum));
		hashtable.put("col2Sum", new Double(col2Sum));
		hashtable.put("rowcolSum", new Double(rowcolSum));		
		hashtable.put("nPixels", new Integer(nPixels));
		hashtable.put("rowRef", new Double(rowRef));
		hashtable.put("colRef", new Double(colRef));
		return hashtable;
	}
	/**
	 * using the results from the sums() method, append first and second moment information
	 * @param hashtable appended with rowSum, colSum, weightSum, row2Sum, col2Sum, rowcolSum, rowRef, and colRef
	 */
	public static void appendMoments(Hashtable<String, Number> hashtable) {
		double rowSum = (Double) hashtable.get("rowSum");
		double colSum = (Double) hashtable.get("colSum");
		double weightSum = (Double) hashtable.get("weightSum");
		double row2Sum = (Double) hashtable.get("row2Sum");
		double col2Sum = (Double) hashtable.get("col2Sum");
		double rowcolSum = (Double) hashtable.get("rowcolSum");
		double rowRef = (Double) hashtable.get("rowRef");
		double colRef = (Double) hashtable.get("colRef");

		double rowCentroid = rowRef+rowSum/weightSum;
		double colCentroid = colRef+colSum/weightSum;
		hashtable.put("rowCentroid", new Double(rowCentroid));
		hashtable.put("colCentroid", new Double(colCentroid));
		double s2 = weightSum * weightSum;
		double rowSigma = Math.sqrt(row2Sum/weightSum - rowSum*rowSum/s2);
		double colSigma = Math.sqrt(col2Sum/weightSum - colSum*colSum/s2);
		double rowcolSigma2 = rowcolSum/weightSum - rowSum*colSum/s2;
		hashtable.put("rowSigma", new Double(rowSigma));
		hashtable.put("colSigma", new Double(colSigma));
		hashtable.put("rowcolSigma2", new Double(rowcolSigma2));
		return;
	}

	/**
	 * append ellipticity parameters to the hastable
	 * positionAngle in degrees
	 * majorRadius in pixels
	 * minorRadius in pixels
	 * e1,e2 from Haymens et al. STEP paper equation 1
	 * epsilon1, epsilon2 from Haymens et al. STEP paper equation 2
	 * @param hashtable generated by centroid() methods
	 */
	public static void appendEllipticities(Hashtable<String,Number> hashtable) {
		// Ellipticities from SExtractor manual equations 21-25


		double col2Sum = (Double) hashtable.get("col2Sum");
		double row2Sum = (Double) hashtable.get("row2Sum");
		double rowcolSum = (Double) hashtable.get("rowcolSum");
		double weightSum = (Double) hashtable.get("weightSum");
		double colCentroid = (Double) hashtable.get("colCentroid");
		double rowCentroid = (Double) hashtable.get("rowCentroid");
		double avgx2 = col2Sum/weightSum - Math.pow(colCentroid, 2);
		double avgy2 = row2Sum/weightSum - Math.pow(rowCentroid, 2);
		double avgxy = rowcolSum/weightSum - rowCentroid*colCentroid;
		double numerator = 2*avgxy;
		double denominator = avgx2-avgy2;

		double positionAngle = Math.signum(avgxy)*0.5*Math.atan2(numerator, denominator)*180.0/Math.PI; // in degrees, thank you very much
		hashtable.put("positionAngle", new Double(positionAngle));		

		double t1 = 0.5*(avgx2+avgy2);
		double t2 = Math.sqrt(Math.pow(0.5*denominator,2) + avgxy*avgxy); 
		double majorRadius = Math.sqrt(t1+t2);
		hashtable.put("majorRadius", new Double(majorRadius));
		double minorRadius = Math.sqrt(t1-t2);
		hashtable.put("minorRadius", new Double(minorRadius));

		// Weak lensing representation of Ellipticities from Heymans et. al (STEP) 
		double beta = minorRadius/majorRadius;
		double factor = (1-beta)/(1+beta);

		// Heymans equation 1
		double e1 = factor*(Math.cos(2*positionAngle*Math.PI/180.0));
		double e2 = factor*(Math.sin(2*positionAngle*Math.PI/180.0));
		hashtable.put("e1", new Double(e1));
		hashtable.put("e2", new Double(e2));

		// Heymans equation 3
		double epsilon1 = (row2Sum-col2Sum)/(row2Sum+col2Sum);
		double epsilon2 = 2*rowcolSum/(row2Sum+col2Sum);
		hashtable.put("epsilon1", new Double(epsilon1));
		hashtable.put("epsilon2", new Double(epsilon2));

	}
	/**
	 * get the Rectangle that defines the active area of this PixeledImage
	 * @return the rectangle
	 */
	public Rectangle getRectangle() {
		Rectangle rectangle = new Rectangle(new Point(getCol0(), getRow0()), new Dimension(getColumnCount(), getRowCount()));
		return rectangle;
	}
	/**
	 * get the sum of all pixels
	 * @return the sum
	 */
	public double getSum() {
		double sum = 0.0;
		for( int row=0; row<getRowCount(); row++) {
			for (int col=0; col<getColumnCount(); col++) {
				sum += pix[row][col];
			}
		}
		return sum;
	}

	/**
	 * @param maskBits BigInteger.ZERO to reject pixels with any bit set, null to accept all pixels
	 * @param plotName if not null, histogram values and put the plot in this file name
	 * @return with keys sky and skySigma and skyError
	 */
	public Hashtable<String, Number> getSky(BigInteger maskBits, String plotName) {
		double sum = 0.0;
		double sum2 = 0.0;
		double n = 0;
		// first iteration
		for (int row=getRowMin(); row<getRowMax(); row++) {
			for (int col=getColumnMin(); col<getColumnMax(); col++) {
				if (pixelAccepted(row, col, maskBits)) {
					double value = getByRowCol(row,col);
					sum += value;
					sum2 += value*value;
					n++;
				}
			}
		}
		double mean = sum/n;
		double sigma = Math.sqrt((sum2-n*mean*mean)/(n-1));
		// second iteration
		sum = 0.0; sum2 = 0.0; n = 0;
		for (int row=getRowMin(); row<getRowMax(); row++) {
			for (int col=getColumnMin(); col<getColumnMax(); col++) {
				if (pixelAccepted(row, col, maskBits)) {
					double value = getByRowCol(row,col);
					if (Math.abs(value-mean) < 4.0*sigma) {
						sum += value;
						sum2 += value*value;
						n++;
					}
				}
			}
		}

		Hashtable<String, Number> retval = new Hashtable<String, Number>();
		mean = sum/n;
		sigma = Math.sqrt((sum2-n*mean*mean)/(n-1));
		retval.put("sky", mean);
		retval.put("skySigma", sigma);
		retval.put("skyError", sigma/Math.sqrt(n));
		if (plotName != null) {
			double min = Math.floor(mean-4*sigma);			
			double max = Math.ceil(mean+4*sigma);
			int nBin = (int) (max-min);
			Histogram hg = new Histogram(plotName, "value", "dN", nBin, min, max);
			for (int row=getRowMin(); row<getRowMax(); row++) {
				for (int col=getColumnMin(); col<getColumnMax(); col++) {
					if (pixelAccepted(row, col, maskBits)) {
						hg.fill(getByRowCol(row,col));
					}
				}
			}
			JFreeChart chart = hg.getChart(false);
			JFreeChartUtil.toPng(chart,plotName+".png");
		}
		return retval;
	}


	///**
	// * @param maskBits
	// * @return sky and skyErr
	// */
	/*
	public Hashtable<String, Number> getSkyExtractor(BigInteger maskBits, String plotName) {
		Hashtable<String, Number> simpleSky = getSky(maskBits, plotName);
		double mean = simpleSky.get("sky").doubleValue();
		double sigma = simpleSky.get("skySigma").doubleValue();
		double newMean = mean;
		double newSigma = sigma;
		boolean done = false;
		double oldSize = 0;
		return null;
	}
	 */

	/**
	 * measure the fraction of pixels with a value above threshold
	 * @param threshold the threshold
	 * @return the fraction
	 */
	public double getFractionHot(double threshold) {
		int nPix = getRowCount()*getColumnCount();			
		int n=0;
		for( int row=0; row<getRowCount(); row++) {
			for (int col=0; col<getColumnCount(); col++) {
				if (pix[row][col] > threshold) n++;
			}
		}	
		return (double) n/(double) nPix;
	}

	/**
	 * clip all pixel values greater than threshold to the threshold value
	 * @param threshold
	 */
	public void threshold(double threshold) {
		for( int row=0; row<getRowCount(); row++) {
			for (int col=0; col<getColumnCount(); col++) {
				if (pix[row][col] > threshold) pix[row][col] = threshold;
			}
		}	

	}
	/**
	 * @return the array of doubles that contains the pixel values
	 */
	public double[][] getPix() {
		double[][] retval;
		if (i0==0 && j0==0 && rowCount == pix.length && columnCount == pix[0].length) {
			retval = pix;
		} else {
			retval = new double[rowCount][columnCount];
			for (int i=0; i<rowCount; i++) {
				for (int j=0; j<columnCount; j++) {
					retval[i][j] =  pix[i+i0][j+j0];
				}
			}
		}
		return retval;
	}

	/**
	 * @return an array of floats that contain the pixel values cast to float
	 */
	public float[][] getPixAsFloat() {
		float[][] retval;		
		if (i0==0 && j0==0 && rowCount == pix.length && columnCount == pix[0].length) {
			retval = new float[getRowCount()][getColumnCount()];
			for (int row=0; row<getRowCount(); row++) {
				for (int col=0; col<getColumnCount(); col++) {
					retval[row][col] = (float) pix[row][col];
				}
			}
		} else {
			retval = new float[rowCount][columnCount];
			for (int i=0; i<rowCount; i++) {
				for (int j=0; j<columnCount; j++) {
					retval[i][j] =  (float) pix[i+i0][j+j0];
				}
			}			
		}
		return retval;
	}

	/**
	 * get the pixels cast to an unsigned integer with 16 bits of precision, with bzero subtracted
	 * @return the array of shorts
	 */
	public short[][] getPixAsUShort() {
		double bzero = Math.pow(2,15);
		short[][] retval;
		if (i0==0 && j0==0 && rowCount == pix.length && columnCount == pix[0].length) {
			retval = new short [getRowCount()][getColumnCount()];
			for (int row=0; row<getRowCount(); row++) {
				for (int col=0; col<getColumnCount(); col++) {
					retval[row][col] = (short) (pix[row][col] - bzero);
				}
			}
		} else {
			retval = new short[rowCount][columnCount];
			for (int i=0; i<rowCount; i++) {
				for (int j=0; j<columnCount; j++) {
					retval[i][j] =  (short) (pix[i+i0][j+j0] - bzero);
				}
			}			
		}
		return retval;
	}

	/**
	 * set row0,col0 to place the center of these pixels at rowPosition,colPosition
	 * @param rowPosition row position of center of rows
	 * @param columnPosition column position of center of columns
	 */
	public void setPosition(int rowPosition, int columnPosition) {
		row0 = (int) (rowPosition - 0.5*getRowCount());
		col0 = (int) (columnPosition - 0.5*getColumnCount());
	}

	/**
	 * set row0 and col0 to these values
	 * @param row0
	 * @param col0
	 */
	public void setRow0Col0(int row0, int col0) {
		this.row0 = row0;
		this.col0 = col0;
		if (mask != null) {
			mask.setRow0Col0(row0, col0);
		}
	}

	/**
	 * sets row0, col0 to 0,0 and adjusts CRPIX1 and CRPIX2 header values
	 */
	public void resetRow0Col0() {
		if (header.containsKey("CRPIX1")) {
			double crpix1 = header.getDoubleValue("CRPIX1");
			//double newCrpix1 = crpix1 - col0;
			try {
				header.addValue("CRPIX1", crpix1 - col0, "reset CRPIX1");
			} catch (HeaderCardException e) {
				throw new Error(e);
			}
		}
		if (header.containsKey("CRPIX2")) {
			double crpix2 = header.getDoubleValue("CRPIX2");
			try {
				header.addValue("CRPIX2", crpix2 - row0, "reset CRPIX2");
			} catch (HeaderCardException e) {
				throw new Error(e);
			}
		}
		setRow0Col0(0,0);
	}
	/**
	 * set the pix array to the input array
	 * @param pixIn input array
	 */
	public void setPix(double[][] pixIn) {
		this.pix = pixIn;
		rowCount = pix.length;
		columnCount = pix[0].length;
	}
	/**
	 * convolve this PixeledImage with the kernel, using the backgroundValue
	 * @param kernel
	 * @param backgroundValue
	 * @return the convolved PixeledImage
	 */
	public PixeledImage convolveExplicit(PixeledImage kernel, double backgroundValue) {
		int nrowOut = pix.length;
		int ncolOut = pix[0].length;
		int nrowKernel = kernel.getRowCount();
		int ncolKernel = kernel.getColumnCount();
		int kernelMidRow = nrowKernel/2;
		int kernelMidCol = ncolKernel/2;
		PixeledImage retval = new PixeledImage(nrowOut, ncolOut);
		for (int outRow=0; outRow<nrowOut; outRow++) {
			for (int outCol=0; outCol<ncolOut; outCol++) {
				double sum = 0;
				for (int kernelRow=0; kernelRow<nrowKernel; kernelRow++) {
					for (int kernelCol=0; kernelCol<ncolKernel; kernelCol++) {
						double kernelValue = kernel.getByRowCol(kernelRow, kernelCol);
						double thisValue = this.getByRowCol(kernelRow+outRow-kernelMidRow, kernelCol+outCol-kernelMidCol, backgroundValue);
						sum +=  kernelValue * thisValue; 
					}
				}
				retval.setByRowCol(outRow, outCol, sum);
			}
		}
		retval.setRow0Col0(row0, col0);
		return retval;
	}

	/**
	 * use the ij library to convole the PixeledImage with the kernel and overwrite the pixel value with the convolution
	 * @param kernel psf shape
	 * @param background to use outside of edges
	 */
	public void convolveIj(PixeledImage kernel, double background) {
		convolveIj(kernel, new BackgroundLightConstant(background));
	}

	/**
	 * use the ij library to convole the PixeledImage with the kernel and overwrite the pixel value with the convolution
	 * @param kernel psf shape
	 * @param backgroundLight to use outside of edges
	 */
	public void convolveIj(PixeledImage kernel, BackgroundLight backgroundLight) {
		int nrowIn = pix.length;
		int ncolIn = pix[0].length;
		int nrowGutter = (kernel.getRowCount()+1)/2;
		int ncolGutter = (kernel.getColumnCount()+1)/2;
		int nrowFull = nrowIn + 2*nrowGutter;
		int ncolFull = ncolIn + 2*ncolGutter;
		FloatProcessor fp = new FloatProcessor(nrowFull, ncolFull);
		for (int row=getRowMin()-nrowGutter; row < getRowMax()+nrowGutter; row++) {
			for (int col=getColumnMin()-ncolGutter; col < getColumnMax()+ncolGutter; col++) {
				double value = getByRowCol(row, col, backgroundLight);
				if (Double.isNaN(value)) {
					throw new IllegalArgumentException("NaN in this:  this="+this+"  row="+row+" col="+col+" backgroundLight="+backgroundLight);
				}
				fp.putPixelValue(row-row0+nrowGutter, col-col0+ncolGutter, value);
			}
		}

		int nrowKernel = kernel.getRowCount();
		int ncolKernel = kernel.getColumnCount();
		float[] kernelArray = new float[nrowKernel*ncolKernel];
		int i=0;
		for (int row=0; row<nrowKernel; row++) {
			for (int col=0; col<ncolKernel; col++) {
				double value = kernel.getByRowColNoOffset(row, col);
				if (Double.isNaN(value)) {
					throw new IllegalArgumentException("NaN in kernel:  row="+row+" col="+col);
				}   			
				kernelArray[i++] = (float) value;
			}
		}
		//NK    	
		fp.convolve(kernelArray, nrowKernel, ncolKernel);
		for (int row=getRowMin(); row<getRowMax(); row++) {
			for (int col=getColumnMin(); col<getColumnMax(); col++) {
				double value = fp.getPixelValue(row-row0+nrowGutter, col-col0+ncolGutter);
				if (Double.isNaN(value)) {
					throw new IllegalArgumentException("NaN in result:  row="+row+" col="+col);
				}   			
				setByRowCol(row, col, value);
			}
		}
	}

	/**
	 * clear the Fftw information if you want to save space
	 */
	public void clearFftw() {
		fourierTransform = null;
	}

	/**
	 * calculates the correlation of pixel values
	 * @param length maximum length of correlation
	 * @return the correlation values, with the zero-length correlation value at the pixel length,length
	 */
	public PixeledImage autoCorrelateFftw(int length) {
		double[] ft = getFtFftw(-1);
		int nRow = getRowCount();
		int nCol = getColumnCount();
		PixeledImage p = PixeledImage.modulus(ft, nRow, nCol);
		double[] ft2 = p.getFtFftw(1);
		PixeledImage c = PixeledImage.real(ft2, nRow, nCol);
		double norm = c.getByRowCol(0, 0);
		if (norm > 0) {
			c.multiplyScalar(1.0/norm);
		}
		c.shift(length,length);
		int diameter = 2*length;
		PixeledImage retval = c.subPixeledImage(diameter, diameter, 0, 0);
		return retval;
	}
	/**
	 * make a plottable series from the result of the autoCorrelateFftw method
	 * @param name
	 * @param centroidRowColumn
	 * @param radius
	 * @param nbin
	 * @return the series
	 */
	public XYIntervalSeries getCorrelationSeries(String name, Point2D.Double centroidRowColumn, double radius, int nbin) {
		ArrayList<Point2D.Double> values = new ArrayList<Point2D.Double>();
		for (int row=getRowMin(); row < getRowMax(); row++) {
			for (int col=getColumnMin(); col < getColumnMax(); col++) {
				double distance = centroidRowColumn.distance(new Point2D.Double(row,col));
				if (distance > 0.0 && distance < radius) {
					values.add(new Point2D.Double(distance, getByRowCol(row,col)));
				}
			}
		}
		Comparator<Point2D> comparator = new PointComparator();
		Collections.sort(values, comparator);
		int delta = values.size()/nbin;
		if (delta <= 0) {
			throw new IllegalArgumentException("values.size()="+values.size()+" nbin="+nbin);
		}
		DoubleArrayList dal = new DoubleArrayList();
		for (int i=0; i<values.size(); i++) {
			dal.add(values.get(i).y);
		}

		XYIntervalSeries xyIntervalSeries = new XYIntervalSeries(name);
		for (int ibin=0; ibin<nbin; ibin++) {
			DoubleArrayList subDal = (DoubleArrayList) dal.partFromTo(ibin*delta, (ibin+1)*delta-1);
			double mean = Descriptive.mean(subDal);
			double variance = Descriptive.sampleVariance(subDal, mean);
			double sigma = Descriptive.sampleStandardDeviation(subDal.size(), variance);
			double xmin = values.get(ibin*delta).x;
			double xmax = values.get((ibin+1)*delta-1).x;
			double x = 0.5*(xmin+xmax);
			xyIntervalSeries.add(x, xmin, xmax, mean, mean-sigma, mean+sigma);    		
		}
		return xyIntervalSeries;
	}
	private class PointComparator implements Comparator<Point2D> {
		/** compares the X values of the two points
		 * @param p1 
		 * @param p2 
		 * @return -1, 1, or 0
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 */
		public int compare(Point2D p1, Point2D p2) {
			if (p1.getX() < p2.getX()) {
				return -1;
			} else if (p1.getX() > p2.getX()) {
				return 1;
			} else {
				return 0;
			}
		}

	}
	/**
	 * convolve this image with the kernel, using the Fftw routines.
	 * @param kernel
	 * @return the result of the convolution
	 */
	public PixeledImage convolveFftw(PixeledImage kernel) {
		prepareFftws(kernel.getRowCount(), kernel.getColumnCount());
		kernel.prepareFftws(getRowCount(), getColumnCount());
		double[] ft = fourierTransform.clone();
		ArrayUtil.mult(ft, kernel.fourierTransform);
		FFTWComplex fftwComplex = new FFTWComplex();
		fftwComplex.twoDimensional(nColumnFftw, nRowFftw, ft, -1, true);
		PixeledImage retval = new PixeledImage(getRowCount(), getColumnCount());
		int normalization = nRowFftw * nColumnFftw;
		retval.setRow0Col0(row0, col0);
		for (int row=retval.getRowMin(); row<retval.getRowMax(); row++) {
			int rRetval = row-kernel.getRowCount()/2;
			if (rRetval < getRowMin()) {
				rRetval += getRowCount();
			}
			for (int col=retval.getColumnMin(); col<retval.getColumnMax(); col++) {
				int cRetval = col-kernel.getColumnCount()/2;
				if (cRetval < getColumnMin()) {
					cRetval += getColumnCount();
				}
				//int index = (row-retval.getRowMin()) + (col-retval.getColumnMin())*nRowFftw;
				int index = nColumnFftw*(row-retval.getRowMin()) + (col-retval.getColumnMin());
				double value = ArrayUtil.getComplex(ft, index).modulus()/normalization;
				retval.setByRowCol(rRetval,cRetval,value);
			}
		}    	
		return retval;
	}

	private void prepareFftws(int nRowOther, int nColumnOther) {
		int nRowTemp = Math.max(getRowCount(), nRowOther);
		int nColumnTemp = Math.max(getColumnCount(), nColumnOther);
		if (fourierTransform == null || nRowFftw != nRowTemp || nColumnFftw != nColumnTemp) {
			nRowFftw = nRowTemp;
			nColumnFftw = nColumnTemp;
			System.out.println("PixeledImage.prepareFftws:  nRowFftw="+nRowFftw+" nColFftw="+nColumnFftw);
			fourierTransform = new double[2*nRowFftw*nColumnFftw];
			for (int row=getRowMin(); row<getRowMax(); row++) {
				for (int col=getColumnMin(); col<getColumnMax(); col++) {
					//int index = (row-getRowMin()) + nRowFftw*(col-getColumnMin());
					int index = nColumnFftw*(row-getRowMin()) + (col-getColumnMin());
					//int i = nRowOther/2 + row + (nColumnOther/2+col)*nRowFftw;
					ArrayUtil.setComplex(fourierTransform, index, getByRowCol(row,col), 0);
				}
			}
			FFTWComplex fftwComplex = new FFTWComplex();
			fftwComplex.twoDimensional(nColumnFftw, nRowFftw, fourierTransform, 1, true);
		}
	}

	/**
	 * @param kernel
	 * @return the convolution of this PixeledImage with the kernel, using jnt fft 
	 */
	public PixeledImage convolveJnt(PixeledImage kernel) {
		prepareJnts(kernel.getRowCount(), kernel.getColumnCount());
		kernel.prepareJnts(getRowCount(), getColumnCount());
		//PixeledImage retval = new PixeledImage(getRowCount(), getColumnCount());
		PixeledImage retval = new PixeledImage(getRowCount(), getColumnCount());
		retval.setRow0Col0(getRow0(), getCol0());
		double[] ft = jnt.clone();
		//System.out.println("do not multiply");
		ArrayUtil.mult(ft, kernel.jnt);

		ComplexDouble2DFFT complexDouble2DFFT = new ComplexDouble2DFFT(nColumnJnt, nRowJnt);
		complexDouble2DFFT.backtransform(ft);
		int normalization = nRowJnt * nColumnJnt;
		for (int row=retval.getRowMin(); row<retval.getRowMax(); row++) {
			int rRetval = row - kernel.getRowCount()/2;
			if (rRetval < retval.getRowMin()) {
				rRetval += retval.getRowCount();
			}
			for (int col=retval.getColumnMin(); col<retval.getColumnMax(); col++) {
				int cRetval = col - kernel.getColumnCount()/2;
				if (cRetval < retval.getColumnMin()) {
					cRetval += retval.getColumnCount();
				}
				//int index = nColumnJnt*(row-retval.getRowMin()) + (col-retval.getColumnMin());
				int index = (row-retval.getRowMin()) + nRowJnt*(col-retval.getColumnMin());
				double value = ArrayUtil.getComplex(ft, index).modulus()/normalization;
				retval.setByRowCol(rRetval, cRetval, value);
			}
		}   	
		return retval;
	}
	private void prepareJnts(int nRowOther, int nColumnOther) {
		int nRowTemp = Math.max(getRowCount(), nRowOther);
		int nColumnTemp = Math.max(getColumnCount(), nColumnOther);
		if (jnt == null || nRowJnt != nRowTemp || nColumnJnt != nColumnTemp) {
			nRowJnt = nRowTemp;
			nColumnJnt = nColumnTemp;
			jnt = new double[2*nRowJnt*nColumnJnt];
			for (int col=getColumnMin(); col<getColumnMax(); col++) {
				for (int row=getRowMin(); row<getRowMax(); row++) {
					//int index = nColumnJnt*(row-getRowMin()) + (col-getColumnMin());
					int index = (row-getRowMin()) + nRowJnt*(col-getColumnMin());
					double value = getByRowCol(row,col); 
					ArrayUtil.setComplex(jnt, index, value, 0);
				}
			}
			ComplexDouble2DFFT complexDouble2DFFT = new ComplexDouble2DFFT(nColumnJnt, nRowJnt);
			complexDouble2DFFT.transform(jnt);
		}
	}
	/**
	 * @param kernel
	 * @return the convolution of this PixeledImage with the kernel, using jnt fft
	 */
	public PixeledImage convolveJntOld(PixeledImage kernel) {
		prepareJnts(kernel.getRowCount(), kernel.getColumnCount());
		kernel.prepareJnts(getRowCount(), getColumnCount());
		PixeledImage retval = new PixeledImage(getRowCount(), getColumnCount());
		double[] ft = jnt.clone();
		ArrayUtil.mult(ft, kernel.jnt);
		ComplexDouble2DFFT complexDouble2DFFT = new ComplexDouble2DFFT(nColumnJnt, nRowJnt);
		complexDouble2DFFT.backtransform(ft);
		int normalization = nRowJnt * nColumnJnt;
		for (int row=0; row<nRowJnt; row++) {
			int piRow = ((row+nRowJnt/2) % nRowJnt) - kernel.getRowCount()/2;
			if (piRow >= 0 && piRow < getRowCount()) {
				for (int col=0; col<nColumnJnt; col++) {
					int piColumn = (col+nColumnJnt/2) % nColumnJnt - kernel.getColumnCount()/2;
					if (piColumn >= 0 && piColumn < getColumnCount()) {
						int i = row + col*nRowJnt;
						double value = ArrayUtil.getComplex(ft, i).modulus()/normalization;
						retval.pix[piRow][piColumn] = value;
					}
				}
			}
		}
		retval.setRow0Col0(row0, col0);
		return retval;
	}



	/**
	 * print some diagnostic information
	 * @param ps where it should go
	 * @param string an informative string to lable this output
	 */
	public void printAll(PrintStream ps, String string) {
		printAll(ps, string, 0.0);
	}
	/**
	 * print pixel values above the threshold
	 * @param ps where to print
	 * @param string informative message at the beginning
	 * @param threshold print only pixel values above this threshold
	 */
	public void printAll(PrintStream ps, String string, double threshold) {
		ps.println("PixeledImage.printAll called:"+ string+"  threshold="+threshold);
		ps.println(" row0="+getRow0()+"  col0="+getCol0());

		int n = header.getNumberOfCards();
		ps.println(" number of header cards is "+n);
		Cursor cursor = header.iterator();
		while (cursor.hasNext()) {
			HeaderCard card = (HeaderCard) cursor.next();
			System.out.println(card.toString());
		}
		for (int r=getRowMin(); r<getRowMax(); r++) {
			for (int c=getColumnMin(); c<getColumnMax(); c++) {
				double value = getByRowCol(r,c);
				if (value > threshold) ps.println(" r="+r+" c="+c+" pix="+value);
			}
		}
	}
	/**
	 * print diagnostic centroid information
	 * @param ps where to print it
	 * @param message a lable for this output
	 */
	public void printCentroid(PrintStream ps, String message) {
		printCentroid(ps, message, 0.0);
	}
	/**
	 * print diagnostic centroid information, after subtracting background
	 * @param ps where to print it
	 * @param message a label for this output
	 * @param background background to subtract frome each pixel
	 */
	public void printCentroid(PrintStream ps, String message, double background) {
		Hashtable<String, Number> centroidSimple = centroid(background);
		double rowCentroid = new Double(centroidSimple.get("rowCentroid").toString()).doubleValue();
		double colCentroid = new Double(centroidSimple.get("colCentroid").toString()).doubleValue();
		ps.println(message+" rowCentroid="+rowCentroid+" colCentroid="+colCentroid);
	}
	/**
	 * @return initial row position for iteration over row, inclusive
	 */
	public int getRowMin() {return getRow0();}	
	/**
	 * @return final row position for iteration over row, exclusive
	 */
	public int getRowMax() {return getRow0()+getRowCount();}	
	/**
	 * @return intial column position for iteration over column, inclusive
	 */
	public int getColumnMin() {return getCol0();}	
	/**
	 * @return final column position for iteration over column, exclusive
	 */
	public int getColumnMax() {return getCol0()+getColumnCount();}

	/**
	 * copy flux from source to destination using the drizzle factor.
	 * Note that the headers of source and destination must contain coherent WCS information.
	 * @param destination add flux to these pixels
	 * @param source copy flux from these pixels
	 * @param drizzle the size of the drizzle patch inside the source pixels, ranges from 0 to 1, try 0.5 at first.
	 */
	public static void drizzle(PixeledImage destination, PixeledImage source, double drizzle) {
		TransformTangentPlaneA7 destinationTransform = new TransformTangentPlaneA7(destination.header);
		TransformTangentPlaneA7 sourceTransform = new TransformTangentPlaneA7(source.header);
		drizzle(destination, destinationTransform, source, sourceTransform, drizzle);
	}
	/**
	 * copy flux from source to destination using the drizzle factor.
	 * Note that this method ignores header information in the PixeledImages
	 * @param destination add flux to these pixels
	 * @param destinationTransform 
	 * @param source copy flux from these pixels
	 * @param sourceTransform 
	 * @param drizzle the size of the drizzle patch inside the source pixels, ranges from 0 to 1, try 0.5 at first.
	 */
	public static void drizzle(PixeledImage destination, Transform destinationTransform, PixeledImage source, Transform sourceTransform, double drizzle) {
		PolySamp2 polySamp2 = new PolySamp2(source.getPix(), drizzle, false);
		int nrow = destination.getRowCount()+1;
		int ncol = destination.getColumnCount()+1;
		double[][] x = new double[nrow][ncol];
		double[][] y = new double[nrow][ncol];
		for (int row=0; row<nrow; row++) {
			for (int col=0; col<ncol; col++) {
				Coordinate rowColDestination = new Coordinate(col+destination.col0,row+destination.row0);
				SphericalCoordinate raDec = destinationTransform.forward(rowColDestination);
				Coordinate rowColSource = sourceTransform.inverse(raDec);
				x[row][col] = rowColSource.getX()-source.col0;
				y[row][col] = rowColSource.getY()-source.row0;				
			}
		}
		polySamp2.polySampIncrement(x, y, destination.pix);		
		return;
	}

	/**
	 * bleed charge that exceeds the threshold up and down each column
	 * @param threshold
	 */
	public void bleed(double threshold) {
		for (int col=0; col<getColumnCount(); col++) {
			// i=0: row = [ 0, getRowCount()-1 )
			// i=1: row = [ getRowCount()-1, 0 ]
			for (int i=0; i < 2; i++) {
				double excess = 0.0;
				for (int r = 0; r < getRowCount(); r++) {
					int row = (1-2*i)*r + i*(getRowCount()-1);
					double value = pix[row][col];
					if (value > threshold) {
						double extra = (value-threshold)/(2-i); // half when i=0; all when i=1
						pix[row][col] -= extra;
						excess += extra;
					} else if (excess > 0) {
						double delta = Math.min(excess, threshold-value);
						pix[row][col] += delta;
						excess -= delta;
					}
				}
			}
		}
	}
	/**
	 * add background light to each pixel
	 * @param backgroundLight
	 */
	public void addBackgroundLight(BackgroundLight backgroundLight) {
		for (int row = getRow0(); row < getRowCount()+getRow0(); row++) {
			for (int col = 0; col < getColumnCount(); col++) {
				double backgroundValue = backgroundLight.getLightInPixel(row+0.5, col+0.5);
				incrByRowCol(row, col, backgroundValue);
			}
		}
	}

	/**
	 * add a fraction of the charge in each pixel to the pixel across from it
	 * @param fraction
	 */
	public void addMirrorCrosstalk(double fraction) {

		for (int col1=0; col1 < getColumnCount()/2; col1++) {
			int col2 = getColumnCount()-col1-1;
			for (int row=0; row<getRowCount(); row++) {
				double f1 = fraction*getByRowColNoOffset(row,col1);
				double f2 = fraction*getByRowColNoOffset(row,col2);
				incrByRowColNoOffset(row, col2, f1);
				incrByRowColNoOffset(row, col1, f2);
			}
		}
	}
	/**
	 * copy the pixel values while multiplying by the flat values
	 * @param pi values to copy
	 * @param r0 r0 in this PixeledImage to place pi values
	 * @param c0 c0 in this PixeledImage to place pi values
	 * @param flat values to multiply pi values
	 */
	public void copyAndMultiplyByFlat(PixeledImage pi, int r0, int c0, PixeledImage flat) {
		for (int row=0; row<pi.getRowCount(); row++) {
			for (int col=0; col<pi.getColumnCount(); col++) {
				pix[r0+row][c0+col] = pi.pix[row][col]*flat.pix[row][col];
			}
		}						
	}

	/**
	 * clip all values to be between min and max
	 * @param min
	 * @param max
	 */
	public void clip(double min, double max) {
		for (int row=0; row<getRowCount(); row++) {
			for (int col=0; col<getColumnCount(); col++) {
				pix[row][col] = Math.max(pix[row][col], min);
				pix[row][col] = Math.min(pix[row][col], max);
			}
		}
	}

	/**
	 * @param min
	 * @param max
	 * @param maskValue
	 */
	public void maskExtremePixels(double min, double max, BigInteger maskValue) {
		if (mask != null) {
			for (int row=row0; row<getRowMax(); row++) {
				for (int col=col0; col<getColumnMax(); col++) {
					double value = getByRowCol(row,col);
					if (value < min || value > max) {
						mask.orByRowCol(row, col, maskValue);
					}
				}
			}
		}
	}
	/**
	 * excluding special cards BITPIX, NAXIS*, BZERO and BSCALE
	 * @return a copy of the header
	 */
	public Header getHeader() {
		Header retval = new Header();
		Cursor iterator = header.iterator();
		while (iterator.hasNext()) {
			HeaderCard headerCard = (HeaderCard) iterator.next();
			if (headerCardIsNotSpecial(headerCard)) {
				retval.addLine(headerCard);
			}
		}
		return retval;
	}
	private boolean headerCardIsNotSpecial(HeaderCard headerCard) {
		String[] specialKeys = {"BITPIX", "NAXIS", "BZERO", "BSCALE"};
		String thisKey = headerCard.getKey();
		if (thisKey != null) {
			for (String key : specialKeys) {
				if (thisKey.startsWith(key)) {
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * replace the values with what is read out with a charge transfer inefficiency
	 * @param cti
	 * @param nTailMax
	 */
	public void doCtiColFast(double cti, int nTailMax) {
		double[] ctin = new double[nTailMax];
		for (int nTail=0; nTail<nTailMax; nTail++) {
			ctin[nTail] = Math.pow(cti, nTail);
		}
		double[] oneMinusCtin = new double[getRowCount()+nTailMax];
		//for (int row=getRow0()-nTailMax; row<getRowMax(); row++) {
		for (int row=0; row<getRowCount()+nTailMax; row++) {
			int exponent = row-nTailMax+1;
			oneMinusCtin[row] = Math.pow(1.0-cti, exponent);
		}
		for (int col=getCol0(); col<getColumnMax(); col++) {
			double[] tempCol = new double[getRowCount()];
			for (int row=getRow0(); row<getRowMax(); row++) {
				for (int nTail=0; nTail<nTailMax; nTail++) {
					int index = row-getRow0()+nTail;
					if (index < getRowCount()) {
						double q0 = getByRowCol(row,col);
						if (q0 > 0) {
							double fact = getFact(row+1, nTail);
							double omctin = oneMinusCtin[nTailMax+row-nTail-getRow0()];
							double ctinNTail = ctin[nTail];
							double nTailFactorial = Factorial.getFactorial()[nTail];
							double product = q0*fact*omctin*ctinNTail/nTailFactorial;
							tempCol[index] += product;	
						}
					}
				}
			}
			for (int row=getRow0(); row<getRowMax(); row++) {
				setByRowCol(row,col,tempCol[row-getRow0()]);
			}
		}
	}

	/**
	 * replace the values with what is read out with a charge transfer inefficiency
	 * @param cti
	 */
	public void doCtiCol(double cti) {
		for (int col=getCol0(); col<getColumnMax(); col++) {
			double[] tempCol = new double[getRowMax()+1];
			for (int row=getRow0(); row<getRowMax(); row++) {
				tempCol[row] = getByRowCol(row,col);
			}
			for (int row=0; row<getRowMax(); row++) {
				// copy the bottom pixel to the array
				double value = (1-cti)*tempCol[0];
				setByRowCol(row, col, value);
				tempCol[0] *= cti;
				// shift the charge down
				for (int i=0; i<getRowMax()-row; i++) {
					tempCol[i] += tempCol[i+1]*(1-cti);
					tempCol[i+1] *= cti;
				}
			}
		}
	}


	private double getFact(int np, int n) {
		double retval = 1;
		for (int i=0; i<n; i++) {
			retval*= np-i;
		}
		return retval;
	}

	/**
	 * make a hitogram of pixel values in a column
	 * @param title
	 * @param column
	 * @return the histogram
	 */
	public Histogram getColumnHistogram(String title, int column) {
		Histogram retval = new Histogram(title, "row", "value", getRowCount(), getRowMin(), getRowMax());
		for (int row=getRowMin(); row<getRowMax(); row++) {
			retval.fill(row, getByRowCol(row,column));
		}
		return retval;
	}

	/**
	 * values for determining direction of collapse
	 * @author dcharms
	 *
	 */
	public enum Direction {

		/**
		 * Collapse on x-axis, adding columns
		 */
		XDIRECTION,

		/**
		 * Collapse on y-axis, adding rows
		 */
		YDIRECTION
	}


	/**
	 * Make a histgram by collapsing image on x or y-axis
	 * @param xOry direction to collapse image
	 * @return the histogram of the collapsed image
	 */
	public Histogram getHistogram(Direction xOry) {
		Histogram retval = getHistogram(xOry, getRow0(), getCol0(), getRowMax(), getColumnMax());
		return retval;
	}

	/**
	 * Make a histogram by collapsing image on x or y-axis
	 * @param xOry direction to collapse image
	 * @param rowMin minimum row to be included in histogram
	 * @param colMin minimum column to be included in histogram
	 * @param rowMax maximum row to be included in histogram
	 * @param colMax maximum column to be included in histogram
	 * @return the histogram of the collapsed image
	 */
	public Histogram getHistogram(Direction xOry, double rowMin, double colMin, double rowMax, double colMax) {
		return getHistogram(xOry, rowMin, colMin, rowMax, colMax, BigInteger.ZERO);
	}
	/**
	 * Make a histogram by collapsing image on x or y-axis
	 * @param xOry direction to collapse image
	 * @param rowMin minimum row to be included in histogram
	 * @param colMin minimum column to be included in histogram
	 * @param rowMax maximum row to be included in histogram
	 * @param colMax maximum column to be included in histogram
	 * @param maskBits ignore pixels with mask XOR maskBits != 0
	 * @return the histogram of the collapsed image
	 */
	public Histogram getHistogram(Direction xOry, double rowMin, double colMin, double rowMax, double colMax, BigInteger maskBits) {
		//Histogram hist = new Histogram("PixeledImage", "flux", "dN", 100, 0, rowMax+colMax+row0+col0);
		Histogram hist;
		if (xOry == Direction.XDIRECTION) {
			hist = new Histogram("PixeledImage", "row", "counts averaged over columns", (int) (rowMax-rowMin), rowMin-0.5, rowMax-0.5);
			for (double row=rowMin; row<rowMax; row++) {
				DoubleArrayList dal = new DoubleArrayList();
				for (double col=colMin; col<colMax; col++) {
					if (pixelAccepted(row, col, maskBits)) {
						double value = getByRowCol(row,col);
						dal.add(value);
					}
				}
				double mean = Descriptive.mean(dal);
				hist.fill(row, mean);
			}
		}
		else if (xOry == Direction.YDIRECTION) {
			hist = new Histogram("PixeledImage", "column", "counts averaged over rows", (int) (colMax-colMin), colMin-0.5, colMax-0.5);
			for (double col=colMin; col<colMax; col++) {
				DoubleArrayList dal = new DoubleArrayList();
				for (double row=rowMin; row<rowMax; row++) {
					if (pixelAccepted(row, col, maskBits)) {
						double value = getByRowCol(row,col);
						dal.add(value);
					}
				}
				double mean = Descriptive.mean(dal);
				hist.fill(col, mean);
			}
		}
		else {
			throw new Error("Direction not recognized xOry="+xOry);
		}
		return hist;
	}

	/**
	 * if maskBits or the PixeledMask are null, return true.
	 * @param row
	 * @param col
	 * @param maskBits
	 * @return if the pixel at row,col is accepted by this maskBits
	 */
	public boolean pixelAccepted(double row, double col, BigInteger maskBits) {
		if (mask == null) return true;
		if (maskBits == null) return true;
		BigInteger xor = mask.getByRowCol(row, col).xor(maskBits);
		return (xor.equals(BigInteger.ZERO));
	}

	/**
	 * @param min
	 * @param max
	 * @param nbin 
	 * @return the histogram of values in the range
	 */
	public Histogram getHistogram(Double min, Double max, Integer nbin) {
		if (Double.isNaN(min) || Double.isNaN(max)) {
			Hashtable<String,Double> extrema = extrema();
			if (Double.isNaN(min)) {
				min = extrema.get("min");
			}
			if (Double.isNaN(max)) {
				max = extrema.get("max");
			}
		}
		if (nbin == null) {
			nbin = 100;
		}			

		Histogram hist = new Histogram("PixeledImage", "value", "dN", nbin, min, max);
		for (int row=row0; row<row0+getRowCount(); row++) {
			for (double col=col0; col<col0+getColumnCount(); col++) {
				hist.fill(getByRowCol(row, col), 1.0);
			}			
		}
		return hist;

	}
	/**
	 * @return a histogram of all the values in this image
	 */
	public Histogram getHistogram() {
		return getHistogram(Double.NaN, Double.NaN, null);
	}

	/**
	 * @return the KernelDensityEstimate of these pixel values
	 */
	public KernelDensityEstimate getKernelDensityEstimate() {
		return getKernelDensityEstimate(Double.NaN, Double.NaN);
	}
	/**
	 * @param min
	 * @param max
	 * @return a KernelDensityEstimate for a values bounded by min,max
	 */
	public KernelDensityEstimate getKernelDensityEstimate(double min, double max) {
		Hashtable<String,Double> extrema = null;
		if (Double.isNaN(min)) {
			extrema = extrema();
			min = extrema.get("min");
		}
		if (Double.isNaN(max)) {
			if (extrema == null) {
				extrema = extrema();
			}
			max = extrema.get("max");
		}

		final SimpleFunction kernel = new UniformKernel(10.0);
		int nValues = getRowCount()*getColumnCount();
		KernelDensityEstimate retval = new KernelDensityEstimate(min, max, nValues, kernel);
		for (int row=getRowMin(); row<getRowMax(); row++) {
			for (int col=getColumnMin(); col<getColumnMax(); col++) {
				retval.add(getByRowCol(row,col));
			}
		}
		return retval;
	}

	/**
	 * @param images
	 * @return the median image calculated by the QuantileBin1D.median() method
	 */
	public static PixeledImage median(List<PixeledImage> images) {
		if (images.size() == 0) {
			throw new IllegalArgumentException("need at least one PixeledImage input");
		}
		if (images.size() == 1) {
			return images.get(0);
		}
		int rowCount = -1;
		int columnCount = -1;
		for (PixeledImage pi : images) {
			if (rowCount < 0) {
				rowCount = pi.getRowCount();
				columnCount = pi.getColumnCount();
			} else {
				if (rowCount != pi.getRowCount()   || columnCount != pi.getColumnCount()) {
					throw new IllegalArgumentException("pixeled images must all have the same dimensions");
				}
			}
		}
		PixeledImage retval = new PixeledImage(rowCount, columnCount);
		for (int row=0; row<rowCount; row++) {
			for (int col=0; col<columnCount; col++) {
				ArrayList<Double> values = new ArrayList<Double>();
				for (int i=0; i<images.size(); i++) {
					values.add(images.get(i).getByRowColNoOffset(row, col));
				}
				double median = Statistics.median(values);
				retval.setByRowColNoOffset(row, col, median);
			}
		}
		return retval;
	}

	/**
	 * @see java.lang.Object#toString()
	 * reports row,column counts and initial pixel
	 */
	@Override
	public String toString() {
		return super.toString()+" nrow="+getRowCount()+" ncol="+getColumnCount()+" row0="+getRow0()+" col0="+getCol0();
	}
	/**
	 * clear the stored fourier transform used by the jnt algorithm, to allow gc to save space
	 */
	public void clearJnt() {
		jnt = null;
	}
	/**
	 * @return the maximum value of all pixels in the outside edge of this image
	 */
	public double getMaxCountsPerPixelInEdge() {
		double retval = Double.MIN_VALUE;
		for (int row=0; row<getRowCount(); row++) {
			retval = Math.max(retval, pix[row][0]);
			retval = Math.max(retval, pix[row][getColumnCount()-1]);			
		}
		for (int col=1; col<getColumnCount()-1; col++) {
			retval = Math.max(retval, pix[0][col]);
			retval = Math.max(retval, pix[getRowCount()-1][col]);			
		}
		return retval;
	}

	/**
	 * clock the charge
	 * @param parallel constant for all pixels for parallel shifts
	 * @param serial constant for all pixels in the serial register
	 * @param quantizeCharge whether to quantize charge
	 */
	public void clockCharge(Pixel parallel, Pixel serial, boolean quantizeCharge) {
		for (int col=getColumnMin(); col<getColumnMax(); col++) {
			ArrayList<Pixel> pixels = new ArrayList<Pixel>();
			for (int row=getRowMin(); row<getRowMax(); row++) {
				double charge = getByRowCol(row,col);
				Pixel pixel = parallel.copy(charge);
				pixels.add(pixel);
			}
			PixelRegister pixelRegister = new PixelRegister(pixels);
			int iPixelMax = getRowMax()-getRowMin();
			for (int row=getRowMin(); row<getRowMax(); row++) {
				double charge = pixelRegister.clock(iPixelMax, Double.NaN, quantizeCharge);
				setByRowCol(row,col,charge);
				iPixelMax--;
			}
		}

		for (int row=getRowMin(); row<getRowMax(); row++) {
			ArrayList<Pixel> pixels = new ArrayList<Pixel>();
			for (int col=getColumnMin(); col<getColumnMax(); col++) {
				double charge = getByRowCol(row,col);
				Pixel pixel = serial.copy(charge);
				pixels.add(pixel);
			}
			PixelRegister pixelRegister = new PixelRegister(pixels);
			int iPixelMax = getRowMax()-getRowMin();
			for (int col=getColumnMin(); col<getColumnMax(); col++) {
				double charge = pixelRegister.clock(iPixelMax, Double.NaN, quantizeCharge);
				setByRowCol(row,col,charge);
				iPixelMax--;
			}
		}
	}
	/**
	 * clock the charge with these traps and times
	 * @param parallelTraps
	 * @param parallelTime
	 * @param parallelFullWell 
	 * @param parallelPower 
	 * @param parallelNotchDepth 
	 * @param serialTraps
	 * @param serialTime
	 * @param serialFullWell 
	 * @param serialPower 
	 * @param serialNotchDepth 
	 * @param quantizeCharge whether to quantize charge
	 */
	public void clockCharge(
			Hashtable<Point, ArrayList<ElectronTrap>> parallelTraps, double parallelTime, double parallelFullWell, double parallelPower, double parallelNotchDepth,
			Hashtable<Integer, ArrayList<ElectronTrap>> serialTraps, double serialTime, double serialFullWell, double serialPower, double serialNotchDepth,
			boolean quantizeCharge) {
		for (int col=getColumnMin(); col<getColumnMax(); col++) {
			ArrayList<Pixel> pixels = new ArrayList<Pixel>();
			for (int row=getRowMin(); row<getRowMax(); row++) {
				double charge = getByRowCol(row,col);
				Point point = new Point(col,row);
				ArrayList<ElectronTrap> traps = parallelTraps.get(point);
				Pixel pixel = new PixelWithTraps(traps, charge, parallelFullWell, parallelPower, parallelNotchDepth);
				pixels.add(pixel);
			}
			PixelRegister pixelRegister = new PixelRegister(pixels);
			int iPixelMax = getRowMax()-getRowMin();
			for (int row=getRowMin(); row<getRowMax(); row++) {
				double charge = pixelRegister.clock(iPixelMax, parallelTime, quantizeCharge);
				setByRowCol(row,col,charge);
				iPixelMax--;
			}
		}
		for (int row=getRowMin(); row<getRowMax(); row++) {
			ArrayList<Pixel> pixels = new ArrayList<Pixel>();
			for (int col=getColumnMin(); col<getColumnMax(); col++) {
				double charge = getByRowCol(row,col);
				ArrayList<ElectronTrap> traps = serialTraps.get(col);
				Pixel pixel = new PixelWithTraps(traps, charge, serialFullWell, serialPower, serialNotchDepth);
				pixels.add(pixel);
			}
			PixelRegister pixelRegister = new PixelRegister(pixels);
			int iPixelMax = getColumnMax()-getColumnMin();
			for (int col=getColumnMin(); col<getColumnMax(); col++) {
				double charge = pixelRegister.clock(iPixelMax, serialTime, quantizeCharge);
				setByRowCol(row,col,charge);
				iPixelMax--;
			}
		}		
	}
	/**
	 * clock the charge with these traps and times
	 * @param parallelTraps
	 * @param parallelTime 
	 * @param parallelFullWell 
	 * @param parallelPower 
	 * @param parallelNotchDepth 
	 * @param serialTraps
	 * @param serialTime
	 * @param serialFullWell 
	 * @param serialPower 
	 * @param serialNotchDepth 
	 * @param verbosity 0 is quiet, 1 reports a little; 2 reports each column and row
	 * @param range 
	 * @param quantizeCharge whether to quantize charge
	 * @param decaySeed set the SimRandom seed that controls trap decay to this value
	 */
	public void clockCharge(
			ArrayList<ElectronTrap>[][] parallelTraps, double parallelTime, double parallelFullWell, double parallelPower, double parallelNotchDepth,
			ArrayList<ElectronTrap>[] serialTraps, double serialTime, double serialFullWell, double serialPower, double serialNotchDepth, int verbosity, 
			Rectangle range, boolean quantizeCharge,
			int decaySeed) {
		if (range == null) return;
		int colMin = Math.max(getColumnMin(), range.x);
		int colMax = Math.min(getColumnMax(), range.x+range.width);
		int rowMin = Math.max(getRowMin(), range.y);
		int rowMax = Math.min(getRowMax(), range.y+range.height);
		// Parallel
		if (parallelTraps != null) {
			if (verbosity > 0) {
				System.out.println("PixeledImage.clockCharge:  parallel begin col from "+colMin+" to "+colMax+" with rowMin="+rowMin+" rowMax="+rowMax);
			}
			for (int col=colMin; col<colMax; col++) {
				ArrayList<Pixel> pixels = new ArrayList<Pixel>();
				int nParallelTraps = 0;
			
				for (int row=getRowMin(); row<rowMax; row++) {
					double charge = getByRowCol(row,col);
					ArrayList<ElectronTrap> traps = parallelTraps[row][col];
					int nTrapsThisPixel = 0;
					if (traps != null) {
						nTrapsThisPixel += traps.size();
					}
					nParallelTraps += nTrapsThisPixel;
					Pixel pixel = new PixelWithTraps(traps, charge, parallelFullWell, parallelPower, parallelNotchDepth);
					pixels.add(pixel);
				}
				if (nParallelTraps > 0) {
					PixelRegister pixelRegister = new PixelRegister(pixels);
					int iPixelMax = rowMax-getRowMin();
					int tempSeed = -1;
					for (int row=getRowMin(); row<rowMax; row++) {
						if (quantizeCharge) {
							tempSeed = decaySeed + row + col*10000;
							pixelRegister.setDecaySeed(tempSeed);
						}
						double charge = pixelRegister.clock(iPixelMax, parallelTime, quantizeCharge);
						setByRowCol(row,col,charge);
						iPixelMax--;
					}
				}
			}
		}
		// Serial
		if (serialTraps != null) {
			if (verbosity > 0) {
				System.out.println("PixeledImage.clockCharge:  begin clocking serial rows from "+rowMin+" to "+rowMax);
			}
			ArrayList<Pixel> pixels = new ArrayList<Pixel>();
			int nSerialTraps = 0;
			for (int col=getColumnMin(); col<colMax; col++) {
				ArrayList<ElectronTrap> traps = serialTraps[col];
				if (traps != null) {
					nSerialTraps += traps.size();
				}
				Pixel pixel = new PixelWithTraps(traps, 0, serialFullWell, serialPower, serialNotchDepth);
				pixels.add(pixel);
			}
			if (nSerialTraps > 0) {
				PixelRegister pixelRegister = new PixelRegister(pixels);
				for (int row=rowMin; row<rowMax; row++) {
					if (verbosity > 1) {
						System.out.println("PixeledImage.clockCharge:  begin row="+row+" in rowMin="+rowMin+" rowMax="+rowMax+" columnMin="+getColumnMin()+" colMax="+colMax);
					}
					for (int col=getColumnMin(); col<colMax; col++) {
						double charge = getByRowCol(row,col);
						Pixel pixel = pixels.get(col-getColumnMin());
						pixel.setCharge(charge);
					}
					int iPixelMax = colMax-getColumnMin();
					for (int col=getColumnMin(); col<colMax; col++) {
						if (quantizeCharge) {
							int tempSeed = decaySeed + col + row;
							pixelRegister.setDecaySeed(tempSeed);
						}
						double charge = pixelRegister.clock(iPixelMax, serialTime, quantizeCharge);
						setByRowCol(row,col,charge);
						iPixelMax--;
					}
				}
			}		
		}
	}

	/**
	 * applies the analytic "version 1" clock charging
	 * @param praParallel
	 * @param praSerial
	 * @param sky
	 */
	public void clockChargeAnalyticV1(PixelRegisterAnalytic praParallel, PixelRegisterAnalytic praSerial, double sky) {
		StopwatchCollection.start("PixeledImage.clockChargeAnalytic parallel");
		for (int c=getColumnMin(); c < getColumnMax(); c++) {
			double[] chargeIn = new double[getRowCount()];
			{
				int i=0;
				for (int r=getRowMin(); r < getRowMax(); r++) {
					chargeIn[i++] = getByRowCol(r,c);
				}
			}
			double[] chargeOut = praParallel.clockV1(chargeIn, sky);
			{
				int i=0;
				for (int r=getRowMin(); r < getRowMax(); r++) {
					setByRowCol(r,c,chargeOut[i++]);
				}
			}
		}
		StopwatchCollection.stop("PixeledImage.clockChargeAnalytic parallel");
		StopwatchCollection.start("PixeledImage.clockChargeAnalytic serial");
		for (int r=getRowMin(); r < getRowMax(); r++) {
			double[] chargeIn = new double[getColumnCount()];
			{
				int i=0;
				for (int c=getColumnMin(); c < getColumnMax(); c++) {
					chargeIn[i++] = getByRowCol(r,c);
				}
			}
			double[] chargeOut = praSerial.clockV1(chargeIn, sky);
			{
				int i=0;
				for (int c=getColumnMin(); c < getColumnMax(); c++) {
					setByRowCol(r,c,chargeOut[i++]);
				}
			}
		}
		StopwatchCollection.stop("PixeledImage.clockChargeAnalytic serial");
	}

	/**
	 * rebin by binFactor in each direction
	 * @param binFactor
	 * @return the binned image
	 * WARNING -- row0, col0 are set to zero; no header information is copied; the mask is lost
	 */
	public PixeledImage rebin(int binFactor) {
		if (binFactor <= 1) {
			return this;
		}
		int newRowCount = getRowCount()/binFactor;
		int newColumnCount = getColumnCount()/binFactor;
		PixeledImage retval = new PixeledImage(newRowCount, newColumnCount);
		for (int r=0; r<getRowCount(); r++) {
			int newRow = r/binFactor;
			for (int c=0; c<getColumnCount(); c++) {
				int newCol = c/binFactor;
				retval.incrByRowColNoOffset(newRow, newCol, getByRowColNoOffset(r,c));
			}
		}
		return retval;
	}

	/**
	 * 
	 * @param row
	 * @param col
	 * @param radius include pixel values inside this radius
	 * @param order
	 * @return the interpolated value
	 */
	public double interpolate(double row, double col, double radius, int order) {
		return interpolate(row, col, radius, order, 0);
	}

	/**
	 * @param row
	 * @param col
	 * @param radius
	 * @param order
	 * @param width sigma of gaussian weight function
	 * @return the interpolated value
	 */
	public double interpolate(double row, double col, double radius, int order, double width) {
		ArrayList<double[]> points = new ArrayList<double[]>();
		for (int r=(int) Math.floor(row-radius); r<= (int) Math.floor(row+radius); r++) {			
			if (r >= getRowMin() && r < getRowMax()) {
				for (int c=(int) Math.floor(col-radius); c<= (int) Math.floor(col+radius); c++) {
					if (c >= getColumnMin() && c < getColumnMax()) {
						double x = c+0.5;
						double y = r+0.5;
						double v = getByRowCol(r,c);
						double gaussRadius = Math.sqrt(Math.pow(col-x,2)+Math.pow(row-y,2));
						if (gaussRadius < radius) {
							if (width <= 0) {
								double[] point = new double[] {x,y,v};
								points.add(point);
							} else {
								double gaussSigma = Math.exp(gaussRadius/width/width/2);
								double[] point = new double[] {x,y,v, gaussSigma};
								points.add(point);							
							}
						}
					}
				}
			}
		}


		Polynomial2D fitP2d = new Polynomial2D(points, order);
		//double[] ps = fitP2d.getParams();
		double retval = fitP2d.value(col, row);
		return retval;
	}
	/**
	 * @param format for each falue
	 * @param ps where to dump the numbers
	 */
	public void print(String format, PrintStream ps) {
		for (int r=getRowMin(); r<getRowMax(); r++) {
			for (int c=getColumnMin(); c<getColumnMax(); c++) {
				ps.print(String.format(format+" ", getByRowCol(r,c)));
			}
			ps.println("");
		}
	}

	/**
	 * calculate the amount of charge to remove from q1 and add to q2 and q
	 * @param q1 amount of charge on the output capacitor of pixel 1
	 * @param q2 amount of charge on the output capacitor of pixel 2
	 * @param q amount of charge on the capacitor that couples pixels 1 and 2
	 * @param ratio coupling/output capacitance ratio
	 * @return the amount of charge to move
	 */
	public static double ipcCalculateDeltaQ(double q1, double q2, double q, double ratio) {
		double numerator = q1 - q2 - q/ratio;
		double denominator = 2 + 1.0/ratio;
		double retval = numerator/denominator;
		return retval;
	}

	/**
	 * apply inter-pixel capacitance
	 * @param ratio coupling/output capacitance ratio
	 * @param maxDeltaQThreshold interate until the maximum amount of charge moved is less than this threshold
	 */
	public void ipcApply(double ratio, double maxDeltaQThreshold) {
		PixeledImage columnCoupling = new PixeledImage(getRowCount(), getColumnCount()-1);
		PixeledImage rowCoupling = new PixeledImage(getRowCount()-1, getColumnCount());		
		PixeledImage columnCouplingTemp = new PixeledImage(getRowCount(), getColumnCount()-1);
		PixeledImage rowCouplingTemp = new PixeledImage(getRowCount()-1, getColumnCount());		
		double maxDeltaQ;
		do {
			PixeledImage piTemp = new PixeledImage(getRowCount(), getColumnCount());
			maxDeltaQ = 0.0;
			for (int row=getRowMin(); row<getRowMax(); row++) {
				for (int col=getColumnMin()+1; col<getColumnMax(); col++) {
					double q1 = getByRowCol(row, col-1);
					double q2 = getByRowCol(row, col);
					double q = columnCoupling.getByRowCol(row,col);
					double deltaQ = PixeledImage.ipcCalculateDeltaQ(q1, q2, q, ratio);
					maxDeltaQ = Math.max(maxDeltaQ, Math.abs(deltaQ));
					piTemp.incrByRowCol(row, col-1,   -deltaQ);
					piTemp.incrByRowCol(row, col,  deltaQ);
					columnCouplingTemp.setByRowCol(row, col,    deltaQ);
				}
			}
			for (int col=getColumnMin(); col<getColumnMax(); col++) {
				for (int row=getRowMin()+1; row<getRowMax(); row++) {
					double q1 = getByRowCol(row-1,   col);
					double q2 = getByRowCol(row  , col);
					double q = rowCoupling.getByRowCol(row, col);
					double deltaQ = PixeledImage.ipcCalculateDeltaQ(q1, q2, q, ratio);
					maxDeltaQ = Math.max(maxDeltaQ, Math.abs(deltaQ));
					piTemp.incrByRowCol(row-1, col, -deltaQ);
					piTemp.incrByRowCol(row,   col,  deltaQ);
					rowCouplingTemp.setByRowCol(row, col,    deltaQ);
				}
			}
			add(piTemp);
			columnCoupling.add(columnCouplingTemp);
			rowCoupling.add(rowCouplingTemp);
		} while (maxDeltaQ > maxDeltaQThreshold);
	}
	/**
	 * @param direction +1 for forward; -1 for inverse
	 * @return the ft
	 */
	public double[] getFtFftw(int direction) {
		if (Math.abs(direction) != 1) {
			throw new IllegalArgumentException("direction must be 1 or -1 but direction="+direction);
		}
		double[] retval = new double[2*getRowCount()*getColumnCount()];
		for (int row=getRowMin(); row<getRowMax(); row++) {
			for (int col=getColumnMin(); col<getColumnMax(); col++) {
				int index = getColumnCount()*(row-getRowMin()) + (col-getColumnMin());
				ArrayUtil.setComplex(retval, index, getByRowCol(row,col), 0);
			}
		}
		FFTWComplex fftwComplex = new FFTWComplex();
		fftwComplex.twoDimensional(getColumnCount(), getRowCount(), retval, direction, true);
		if (direction == -1) {
			ArrayUtil.mult(retval, Math.pow(getRowCount()*getColumnCount(), direction));
		}
		return retval;
	}
	/**
	 * @param ft
	 * @param nRow
	 * @param nCol
	 * @return the modulus
	 */
	public static PixeledImage modulus(double[] ft, int nRow, int nCol) {
		PixeledImage pi = new PixeledImage(nRow, nCol);
		int i=0;
		for (int row=0; row<nRow; row++) {
			for (int col=0; col<nCol; col++) {
				double value = ArrayUtil.getComplex(ft, i).modulus();
				pi.setByRowCol(row, col, value);
				i++;
			}
		}
		return pi;
	}
	/**
	 * @param ft
	 * @param nRow
	 * @param nCol
	 * @return real part of the ft
	 */
	public static PixeledImage real(double[] ft, int nRow, int nCol) {
		PixeledImage pi = new PixeledImage(nRow, nCol);
		int i=0;
		for (int row=0; row<nRow; row++) {
			for (int col=0; col<nCol; col++) {
				double value = ArrayUtil.getComplex(ft, i).r;
				pi.setByRowCol(row, col, value);
				i++;
			}
		}
		return pi;
	}

	/**
	 * shift and wrap image by nRowShift,nColShift
	 * @param nRowShift
	 * @param nColShift
	 */
	public void shift(int nRowShift, int nColShift) {
		nColShift = getColumnCount() + nColShift%getColumnCount();
		{
			double[] thisRow = new double[getColumnCount()];
			for (int row=0; row<getRowCount(); row++) {
				for (int col=0; col<getColumnCount(); col++) {
					int colPrime = (col+nColShift)%getColumnCount();
					thisRow[colPrime] = pix[row][col];
				}
				for (int col=0; col<getColumnCount(); col++) {
					pix[row][col] = thisRow[col];
				}
			}
		}
		nRowShift = getRowCount() + nRowShift%getRowCount();
		{
			double[] thisColumn = new double[getRowCount()];
			for (int col=0; col<getColumnCount(); col++) {
				for (int row=0; row<getRowCount(); row++) {
					int rowPrime = (row+nRowShift)%getRowCount();
					thisColumn[rowPrime] = pix[row][col];
				}
				for (int row=0; row<getRowCount(); row++) {
					pix[row][col] = thisColumn[row];
				}
			}
		}
	}
	/**
	 * histogram values on a line from rc0 to rc1
	 * @param cr0 col,row position to begin
	 * @param cr1 col,row position to end
	 * @param nBin number of bins
	 * @param plotName 
	 * @param fraction 
	 * @return the resulting histogram
	 */
	public Histogram getHistogram(Coordinate cr0, Coordinate cr1, int nBin, String plotName, double fraction) {
		double x0 = cr0.getX();
		double y0 = cr0.getY();

		double x1 = cr1.getX();
		double y1 = cr1.getY();

		double distance = Math.sqrt(Math.pow(x1-x0,2) + Math.pow(y1-y0,2));

		double crFx = x0 + (x1 - x0) * fraction;
		double crFy = y0 + (y1 - y0) * fraction;

		double min = -Math.sqrt(Math.pow(crFx-x0,2) + Math.pow(crFy-y0,2));
		double max = Math.sqrt(Math.pow(crFx-x1,2) + Math.pow(crFy-y1,2));
		Histogram histogram = new Histogram(plotName, "value", "dN", nBin, min, max);

		double deltax = x1-x0;
		double deltay = y1-y0;

		for (int iBin=0; iBin < nBin; iBin++) {
			double d = histogram.getBinPos(iBin);
			// find x value and y value of a point defined by x0,y0 and x1,y1 that is at a distance of d from x0,y0
			//			double x = x0 + deltax*d/distance;
			//			double y = y0 + deltay*d/distance;
			// find the flux at that location
			double g = d - min;
			double x = x0 + g * deltax / distance;
			double y = y0 + g * deltay / distance;
			double value = getByRowCol(y,x);
			histogram.fill(d, value);	
		}
		return histogram;
	}
	/**
	 * applies the transform to all the pixel values
	 * @param pvt
	 */
	public void transform(PixelValueTransformation pvt) {
		for (int row=getRowMin(); row<getRowMax(); row++) {
			for (int col=getColumnMin(); col<getColumnMax(); col++) {
				double value = getByRowCol(row,col);
				double transformedValue = pvt.transform(value, row, col);
				setByRowCol(row, col, transformedValue);
			}
		}
	}
}