package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.sim.random.RandomSphere;
import gov.fnal.eag.sim.random.SimRandomNormal;
import gov.fnal.eag.sim.random.SimRandomPoisson;
import gov.fnal.eag.sim.random.SimRandomUniform;
import gov.fnal.eag.util.misc.LoggerUtil;
import gov.fnal.eag.util.misc.Memory;
import gov.fnal.eag.util.stopwatch.StopwatchCollection;
import gov.fnal.eag.util.yanny.Table;

import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.logging.Logger;

import uk.ac.starlink.table.ColumnInfo;
import uk.ac.starlink.table.RowListStarTable;

/**
 * orchestrate the complete simulation of a pixeled image.
 * 
 */
public class PixeledImageFactory {
	/**
	 * the logger for this class
	 */
	public Logger logger = LoggerUtil.getLogger("PixeledImageFactory");
	
    /**
     * status if the last object rendered is bright
     */
    public static int STATUS_BRIGHT = -1;

    /**
     * status if the last object rendered is faint
     */
    public static int STATUS_FAINT = -2;

    /**
     * status if we don't know about the last object rendered
     */
    public static int STATUS_UNKNOWN = -3;

    /**
     * store the status of the last object rendered
     */
    public int status = STATUS_UNKNOWN;

    /**
     * status if plan and produce used the faint strategey
     */
    public static int PLAN_PRODUCE_FAINT = 1;

    /**
     * status if plan and produce used the bright strategey
     */
    public static int PLAN_PRODUCE_BRIGHT = 2;

    /**
     * status if plan and produce used the shapelet strategey
     */
    public static int PLAN_PRODUCE_SHAPELET = 4;

    /**
     * status if plan and produce used convolving each object strategey
     */
    public static int PLAN_PRODUCE_CONVOLVE_EACH = 8;

    /**
     * status if plan and produce used the convolution of the entire image strategy
     */
    public static int PLAN_PRODUCE_CONVOLVE_ALL = 16;

    /**
     * store which strategy plan and produce used
     */
    public int planProduceStatus = 0;

    /**
     * the PixelDetector for this factory
     */
    public PixelDetector pd;

    private BackgroundLight backgroundLight;

    private PSFFactory psfFactory;

    /**
     * the detector objects for this factory
     */
    Iterable<DetectorObject> detectorObjects;

    private double cosmicRayDensity;

    /**
     * microsecond per evaluation
     */
    private double timePerSingleEvaluation = 1.2;

    /**
     * microseconds per photon
     */
    private double timePerPhoton = 6.1;

    private double timePerPixelConvolve = 0.22;

    //private double timePerPhotonConvolve = 1.0;

    private double countsPerPixelThreshold = 1.0;

    private boolean doPoisson = true;

    private double poissonFactor = 1.0;
    
    private boolean brightConvolveEachObject = true;

    private int nGridBright = 1;

    private SimRandomPoisson simRandomPoisson;

    private SimRandomUniform simRandomUniform;
    
    private SimRandomNormal simRandomNormal;

    private RandomSphere randomSphere;
    
    private PixeledImage bright;

    private PixeledImage faint;

    private PixeledImage shapelet;
    
    private int minSquaresToRender = 15;
    private boolean preventNegativeFluxValues = true;
    
    //public LinkedList rowColShapeletFluxThrown = new LinkedList();
    
    private RowListStarTable renderedSimple;
    private RowListStarTable renderedShapelet;
    private RowListStarTable renderedProduct;
    private int overscanColumns = 0;
    
    private double mirrorCrosstalkFraction = 0.0;

    private ClockCharge clockCharge = null;  
    private int clockChargeVerbosity = 0;
    private Rectangle clockChargeRange = null;
    boolean quantizeCharge = true;
    /**
     * default constructor with 100x100 pixels, default background light and
     * constant PSF, no objects, and no read noise
     * 
     * 
     */
    public PixeledImageFactory() {
    	this(100,100);
    }
    /**
     * default constructor with nrow x ncol pixels, default background light and
     * constant, no objects, and no read noise
     * @param nrow
     * @param ncol
     */
    public PixeledImageFactory(int nrow, int ncol) {
        this(new PixelDetector(nrow, ncol), new BackgroundLightConstant(),
                new PSFFactoryConstantGaussian(), new LinkedList<DetectorObject>(), 0.0,
                new SimRandomPoisson(), new SimRandomUniform(), new SimRandomNormal());
    }

    /**
     * fully-specified constructor
     * 
     * @param pd
     *            the PixelDetector to use
     * @param backgroundLight
     *            sky, zodiacal, and scattered light
     * @param psfFactory
     *            returns a PSF at a given position OR for each member of
     *            detectorObjects
     * @param detectorObjects
     *            objects to render
     * @param cosmicRayDensity
     *            number of high energy cosmic rays per square meter
     * @param simRandomPoisson
     *            poisson random number generator
     * @param simRandomUniform 
     * 			  uniform random number generator
     * @param simRandomNormal 
     */
    public PixeledImageFactory(PixelDetector pd,
            BackgroundLight backgroundLight, PSFFactory psfFactory,
            Collection<DetectorObject> detectorObjects, double cosmicRayDensity,
            SimRandomPoisson simRandomPoisson, SimRandomUniform simRandomUniform, SimRandomNormal simRandomNormal) {
        init(pd, backgroundLight, psfFactory, detectorObjects, cosmicRayDensity,
                simRandomPoisson, simRandomUniform, simRandomNormal);
    }
    
    private void init(PixelDetector pdIn, BackgroundLight backgroundLightIn,
            PSFFactory psfFactoryIn, Collection<DetectorObject> detectorObjectsIn,
            double cosmicRayDensityIn, SimRandomPoisson simRandomPoissonIn,
            SimRandomUniform simRandomUniformIn, SimRandomNormal simRandomNormalIn) {
        this.pd = pdIn;
        this.backgroundLight = backgroundLightIn;
        this.psfFactory = psfFactoryIn;
        this.detectorObjects = detectorObjectsIn;
        this.cosmicRayDensity = cosmicRayDensityIn;
        this.simRandomPoisson = simRandomPoissonIn;
        this.simRandomUniform = simRandomUniformIn;
        this.simRandomNormal = simRandomNormalIn;
        this.randomSphere = new RandomSphere(simRandomPoisson.getIseedIndex()+simRandomUniform.getIseedIndex());
        
        ColumnInfo[] simpleColumnInfo = {new ColumnInfo("NTHROWN", Double.class, "number of photons thrown"), new ColumnInfo("BORP", String.class, "generated with renderBright or renderProfile")};
        renderedSimple = new RowListStarTable(simpleColumnInfo);
        
        ColumnInfo[] shapeletColumnInfo = {new ColumnInfo("NTHROWN", Double.class, "number of photons thrown"), new ColumnInfo("PIORS", String.class, "generated by adding a PixeledImage or by shaplelets out the nSquares")};
        renderedShapelet = new RowListStarTable(shapeletColumnInfo);

        renderedProduct = Table.makeBlankTable("double NTHROWN");
        
    }

    /**
     * set the threshold for deciding whether to use the bright or faint
     * algorithm
     * 
     * @param timePerSingleEvaluation
     *            the setting
     */
    public void setTimePerSingleEvaluation(double timePerSingleEvaluation) {
        this.timePerSingleEvaluation = timePerSingleEvaluation;
    }

    /**
     * get the threshold for deciding whether to use the bright or faint
     * algorithm
     * 
     * @return the threshold
     */
    public double getTimePerSingleEvaluation() {
        return timePerSingleEvaluation;
    }

    /**
     * set the threshold for deciding whether to use the bright or faint
     * algorithm
     * 
     * @param timePerPhoton
     *            the setting
     */
    public void setTimePerPhotonNoConvolve(double timePerPhoton) {
        this.timePerPhoton = timePerPhoton;
    }

    /**
     * get the threshold for deciding whether to use the bright or faint
     * algorithm
     * 
     * @return the threshold
     */
    public double getTimePerPhoton() {
        return timePerPhoton;
    }

    /**
     * set the threshold
     * 
     * @param countsPerPixelThreshold
     *            the value
     */
    public void setCountsPerPixelThreshold(double countsPerPixelThreshold) {
        this.countsPerPixelThreshold = countsPerPixelThreshold;
    }

    /**
     * set the doPoisson flag
     * 
     * @param doPoisson
     *            true to render objects with poisson noise
     */
    public void setDoPoisson(boolean doPoisson) {
        this.doPoisson = doPoisson;
    }
    
    /**
     * sets the poisson factor
     * @param poissonFactor
     */
    public void setPoissonFactor(double poissonFactor) {
    	this.poissonFactor = poissonFactor;
    }

    /**
     * get the doPoisson flag
     * 
     * @return true to render objects with poisson noise
     */
    public boolean getDoPoisson() {
        return doPoisson;
    }

    /**
     * add this DetectorObject to the list detectorObjects
     * @param detectorObject
     */
    //public void addDetectorObject(DetectorObject detectorObject) {
    //    detectorObjects.add(detectorObject);
   // }

    /**
     * set detectorObjects
     * @param detectorObjects
     */
    public void setDetectorObjects(Iterable<DetectorObject> detectorObjects) {
        this.detectorObjects = detectorObjects;
    }

    /**
     * @return detectorObjects
     */
    public Iterable<DetectorObject> getDetectorObjects() {
        return detectorObjects;
    }

    /**
     * set brightConvolveEachObject to convolve each bright object
     * @param brightConvolveEachObject
     */
    public void setBrightConvolveEachObject(boolean brightConvolveEachObject) {
        this.brightConvolveEachObject = brightConvolveEachObject;
    }

    /**
     * @return the value of brightConvolveEachObject
     */
    public boolean getBrightConvolveEachObject() {
        return brightConvolveEachObject;
    }

    /**
     * set which PSFFactory to use
     * @param psfFactory
     */
    public void setPsfFactory(PSFFactory psfFactory) {
        this.psfFactory = psfFactory;
    }

    /**
     * get the PSFFactory used
     * @return the factory
     */
    public PSFFactory getPsfFactory() {
        return psfFactory;
    }

    /**
     * set which BackgroundLight to use
     * @param backgroundLight
     */
    public void setBackgroundLight(BackgroundLight backgroundLight) {
        this.backgroundLight = backgroundLight;
    }

    /**
     * get the BackgroundLight used
     * @return the BackgroundLight
     */
    public BackgroundLight getBackgroundLight() {
        return backgroundLight;
    }

    /**
     * set the number of overscan columns
     * @param overscanColumns
     */
    public void setOverscanColumns(int overscanColumns) {
    	this.overscanColumns = overscanColumns;
    }
    
    /**
     * set bias values on the left and right of the detector
     * @param bias0 in electrons
     * @param bias1 in electrons
     */
    public void setBias(double bias0, double bias1) {
    	int[] splitColumns = {pd.columnCount/2};
    	PixelValueTransformation[] darks = {
    			new PixelValueTransformationAddConstant(bias0,"bias value on left"),
    			new PixelValueTransformationAddConstant(bias1,"bias value on right")
    	};
    	PixelValueTransformation pvtDark = new PixelValueTransformationSplit(splitColumns, darks);
    	pd.putElectronics(pvtDark);

    }
    /**
     * do the actual work, after looking at the list of objects and planning the
     * best way to proceed
     * @return the PixeledImage created
     * 
     * 
     */

    public PixeledImage planAndProduce() {
    	StopwatchCollection.start("PixeledImageFactory.planAndProduce plan");
    	faint=null;
    	shapelet=null;
    	logger.fine("begin planAndProduce");
        // Count what we have
        int nSimple = 0;
        int nBright = 0;
        int nFaint = 0;
        double nPixelConvolveEach = 0;
        double nPixelPsf = 0;
        LinkedList<Object[]> produceList = new LinkedList<Object[]>();
        int nShapelet = 0;
                
        int nDetectorObject = 0;
        for (DetectorObject detectorObject : detectorObjects) {
            if (detectorObject != null) {
            	PSF psf = null;
            	if (psfFactory != null) {
            		psf = psfFactory.getPsf(detectorObject);
            	} 
            	if (detectorObject.isSimple()) {
            		nSimple++;
            		DetectorObjectProfile detectorObjectSimple = (DetectorObjectProfile) detectorObject;
            		Rectangle limits = detectorObjectSimple
            			.getLimits(countsPerPixelThreshold);
            		boolean brightOrFaint = brightOrFaint(detectorObjectSimple,
            				limits, psf);
            		if (brightOrFaint) {
            			if (nBright == 0) {
            				bright = new PixeledImage(pd.rowCount, pd.columnCount);
            				planProduceStatus += PLAN_PRODUCE_BRIGHT;
            			}
            			nBright++;
            			Rectangle brightRectangle = bright.getRectangle();
            			limits = limits.intersection(brightRectangle);
            			int nPixelPsfThisObject;
            			if (psf == null) {
            				nPixelPsfThisObject = 0;
            			} else {
            				try {
            					nPixelPsfThisObject = psf.nrow() * psf.ncol();
            				} catch (UnsupportedOperationException e) {
            					nPixelPsfThisObject = 0;
            				}
            			}
            			nPixelPsf += nPixelPsfThisObject;
            			nPixelConvolveEach += nPixelPsfThisObject * limits.height * limits.width;
            		} else {
            			nFaint++;
            		}
            		produceList.add(new Object[] { "Simple", detectorObject,
                        limits, psf, new Boolean(brightOrFaint) });
            	} else {
            		nShapelet++;
            		produceList.add(new Object[] { "Shapelet", detectorObject,
            				null, psf });
            	}
            	nDetectorObject++;
            	if (nDetectorObject % 100 == 0) {
            		logger.finer(" in planAndProduce: nDetectorObject="+nDetectorObject+" nSimple="+nSimple+" nShapelet="+nShapelet);
            		logger.finer(Memory.report());
            	}
            }
        }

        // Decide whether to convolve bright object one at a time or all at the
        // end
        brightConvolveEachObject = true;
        if (nBright > 0) {
            double nPixelConvolveAll = (nPixelPsf / nBright)
                    * bright.getRowCount() * bright.getColumnCount();
            if (nPixelConvolveAll < nPixelConvolveEach) {
                // faster to convolve at the end
                brightConvolveEachObject = false;
                planProduceStatus += PLAN_PRODUCE_CONVOLVE_ALL;
            } else {
                // faster to convolve each object
                brightConvolveEachObject = true;
                planProduceStatus += PLAN_PRODUCE_CONVOLVE_EACH;
            }
        }

        // Create the rest of the PixeledImages that we will need. (bright was
        // made above already)
        if (nFaint > 0) {
            faint = new PixeledImage(pd.rowCount, pd.columnCount);
            planProduceStatus += PLAN_PRODUCE_FAINT;
        }
        if (nShapelet > 0) {
        	shapelet = new PixeledImage(pd.rowCount, pd.columnCount);
            planProduceStatus += PLAN_PRODUCE_SHAPELET;
        }
    	StopwatchCollection.stop("PixeledImageFactory.planAndProduce plan");
    	StopwatchCollection.start("PixeledImageFactory.planAndProduce produce");
        // Render the objects
        int nToDo = produceList.size();
        int iDone = 0;
        Iterator<Object[]> it = produceList.iterator();
        while (it.hasNext()) {
            if (iDone < 10 || iDone%100 == 0) logger.finer(" in planAndProduce:  iDone="+iDone+" / "+nToDo);
            iDone++;
            Object[] oa = it.next();
            PSF psf = (PSF) oa[3];
            if (((String) oa[0]).equals("Simple")) {
                DetectorObjectProfile detectorObjectSimple = (DetectorObjectProfile) oa[1];
                Rectangle limits = (Rectangle) oa[2];
                boolean brightOrFaint = ((Boolean) oa[4]).booleanValue();
                if (!doPoisson || brightOrFaint) {
                	StopwatchCollection.start("PixeledImageFactory.planAndProduce renderBright");
                    renderBright(detectorObjectSimple, limits, psf);
                	StopwatchCollection.stop("PixeledImageFactory.planAndProduce renderBright");
                    status = STATUS_BRIGHT;
                } else {
                	StopwatchCollection.start("PixeledImageFactory.planAndProduce renderProfile");
                    renderProfile(detectorObjectSimple, psf);
                	StopwatchCollection.stop("PixeledImageFactory.planAndProduce renderProfile");
                    status = STATUS_FAINT;
                }
            } else if (oa[1] instanceof DetectorObjectSurfaceBrightness ){
                DetectorObjectSurfaceBrightness detectorObjectShapelet = (DetectorObjectSurfaceBrightness) oa[1];
            	StopwatchCollection.start("PixeledImageFactory.planAndProduce renderSurfaceBrightness");
                renderSurfaceBrightness(detectorObjectShapelet, psf);
             	StopwatchCollection.stop("PixeledImageFactory.planAndProduce renderSurfaceBrightness");
            } else if (oa[1] instanceof DetectorObjectProduct) {
            	StopwatchCollection.start("PixeledImageFactory.planAndProduce renderDetectorObjectProduce");
            	renderDetectorObjectProduct((DetectorObjectProduct) oa[1]);
            	StopwatchCollection.stop("PixeledImageFactory.planAndProduce renderDetectorObjectProduce");
            }
        }
    	StopwatchCollection.stop("PixeledImageFactory.planAndProduce produce");
    	StopwatchCollection.start("PixeledImageFactory.planAndProduce finish");
        PixeledImage retval = finish();
    	StopwatchCollection.stop("PixeledImageFactory.planAndProduce finish");
    	return retval;
    }

    private void renderDetectorObjectProduct(DetectorObjectProduct detectorObjectProduct) {
    	boolean highResolution = false;
    	PixeledImage piDop = detectorObjectProduct.makePixeledImage(countsPerPixelThreshold, highResolution);
    	double nPhotonsThrown = 0.0;
    	if (piDop != null) {
    		nPhotonsThrown = piDop.getSum();    	
    	}
    	renderedProduct.addRow(new Object[] {nPhotonsThrown});
     	shapelet.add(piDop);
	}
	/**
     * do the actual work, in a straightforward way
     * @return the produced image
     * 
     */
    public PixeledImage produce() {
    	
        // Create the blanks
    	StopwatchCollection.start("   PixeledImage.produce() init");
        bright = new PixeledImage(pd.rowCount, pd.columnCount);
        faint = new PixeledImage(pd.rowCount, pd.columnCount);
        shapelet = new PixeledImage(pd.rowCount, pd.columnCount);
    	StopwatchCollection.stop("   PixeledImage.produce() init");
        // Render objects one by one
    	StopwatchCollection.start("   PixeledImage.produce() render");
        int iDone = 0;
        for (DetectorObject detectorObject : detectorObjects) {
            render(detectorObject);
            if (iDone%100 == 0) {
                logger.finer("in PixeledImageFactory.produce():  iDone="+iDone+" "+Memory.report());
        	}
            iDone++;
        }
    	StopwatchCollection.stop("   PixeledImage.produce() render");
    	StopwatchCollection.start("   PixeledImage.produce() finish");
        PixeledImage retval = finish();
    	StopwatchCollection.stop("   PixeledImage.produce() finish");

    	return retval;
    }
    
    /**
     * finish rendering this image by convolution, adding cosmics, background light, and electronic effects
     * @param pi the image to finish
     * @return the finished image
     */
    public PixeledImage finish(PixeledImage pi) {
    	bright = pi;
    	return finish();
    }
    
    private PixeledImage finish() {
    	PixeledImage retval;
    	logger.finer(" begin PixeledImageFactory.finish()");
        // Convolve the bright image, if necessary
        if (!brightConvolveEachObject) {
             psfFactory.convolve(bright, backgroundLight);
        }

        // Add the images, but don't let nulls get you down
        if (bright == null) {
            if (shapelet == null) {
                if (faint == null) {
                    bright = new PixeledImage(pd.rowCount, pd.columnCount);
                } else {
                    bright = faint;
                    faint = null;
                }
            } else {
                bright = shapelet;
                shapelet = null;
            }
        }
        if (doPoisson) {
        	bright.poisson(simRandomPoisson, poissonFactor);
        }
        if (shapelet != null) {
        	logger.finer(" in PixeledImageFactory.produce():  bright.add(shapelet)");
            bright.add(shapelet);
        }
        if (faint != null) {
        	logger.finer(" in PixeledImageFactory.produce():  bright.add(faint)");
            bright.add(faint);
        }
        logger.finer(" in PixeledImageFactory.produce():  before new CosmicRay");
        // Cosmic Rays
        CosmicRay cosmicRay = new CosmicRay(bright, pd, randomSphere, simRandomPoisson, simRandomUniform);
        double pixelSizeInMicrons = pd.getPixelWidth();
        double areaInSquareMeters = bright.getRowCount()*bright.getColumnCount()*Math.pow(pixelSizeInMicrons,2)/1e12;
        double meanNToAdd = cosmicRayDensity * areaInSquareMeters;
        int nToAdd = simRandomPoisson.next(meanNToAdd);
        logger.finer(" in PixeledImageFactory.finish:  meanNToAdd="+meanNToAdd+" nToAdd="+nToAdd);
        cosmicRay.addManyCosmicRays(nToAdd);
        
        logger.finer(" in finish:  backgroundLight = "+backgroundLight);
        logger.finer(" in finish:  pd="+pd);
        // Background light and electronics
        if (mirrorCrosstalkFraction > 0) {
        	logger.finer("PixeledImageFactory.finish():  call bright.addMirrorCrosstalk");
        	bright.addMirrorCrosstalk(mirrorCrosstalkFraction);        	
        }
        logger.finer(" PixeledImageFactory.finish(): mirrorCrosstalkFraction="+mirrorCrosstalkFraction);
        logger.finer(" PixeledImageFactory.finish(): electronics size="+pd.getElectronics().size());
        logger.finer(" PixeledImageFactory.finish(): efficiency size="+pd.getEfficiency().size());
        logger.finer(" PixeledImageFactory.finish(): backgroundLight="+backgroundLight);
        logger.finer(" PixeledImageFactory.finish(): doPoisson="+doPoisson);
        logger.finer(" PixeledImageFactory.finish(): overscanColumns="+overscanColumns);

        
        boolean somethingToDo = false;        
        if (pd.getElectronics().size() > 0) somethingToDo = true;
        if (pd.getEfficiency().size() > 0) somethingToDo = true;
        if (backgroundLight != null) somethingToDo = true;
        if (doPoisson) somethingToDo = true;
        if (overscanColumns != 0) somethingToDo = true;
        logger.finer(" finish:  background and electronics: somethingToDo="+somethingToDo);
        if (somethingToDo) {
            logger.finest(" finish:  make faint with overscanColumns="+overscanColumns);
            faint = new PixeledImage(bright.getRowCount(), bright.getColumnCount()+2*overscanColumns);
            PixelValueTransformationCrosstalk.cross = bright;  
        	for (int row = faint.getRowMin(); row < faint.getRowMax(); row++) {
                logger.finest(" finish:  row="+row+" rowMax="+bright.getRowMax());
                int ocMax = faint.getColumnMax()-overscanColumns;
        		for (int col = faint.getColumnMin(); col < faint.getColumnMax(); col++) {
         			double sky = 0.0;
        			if (backgroundLight!= null && col >= overscanColumns && col < ocMax) {
        				double backgroundLightLevel = backgroundLight.getLightInPixel(row+0.5, col+0.5);
        				sky = pd.efficiencyTransform(backgroundLightLevel, row, col);
        			}
        			if (doPoisson) {
        				if (poissonFactor > 0.0) {
        					sky /= poissonFactor;
        					sky = simRandomPoisson.next(sky);
        					sky *= poissonFactor;
        				}
        			}
        			// increment by sky
        			double brightValue = bright.getByRowColNoOffset(row, col-overscanColumns);
         			faint.incrByRowColNoOffset(row, col, brightValue+sky);
        		}
        	}        
           	logger.finer(" return faint from PixeledImageFactory.finish() ");
           	retval = faint;
        } else {
            logger.finer(" return bright from PixeledImageFactory.finish() ");
        	retval = bright;
        }
        // apply CTE 
        if (clockCharge != null) {
        	clockCharge.setVerbosity(clockChargeVerbosity);
        	clockCharge.doClock2(retval, clockChargeRange);
        }
        // add electronics noise
        if (pd.getElectronics().size() > 0) {
        	for (int row=retval.getRowMin(); row<retval.getRowMax(); row++) {
        		for (int col=retval.getColumnMin(); col<retval.getColumnMax(); col++) {
        			double temp = pd.electronicsTransform(retval.getByRowCol(row,col), row, col-overscanColumns);
        			retval.setByRowCol(row, col, temp);
        		}
        	}
        }
        return retval;
     }

    /**
     * put one detector object in the appropriate PixeledImage
     * 
     * @param detectorObject
     */
    private double render(DetectorObject detectorObject) {
    	PSF psf = null;
    	if (psfFactory != null) {
    		psf = psfFactory.getPsf(detectorObject);
    	}
        if (detectorObject.isSimple()) {
            DetectorObjectProfile detectorObjectSimple = (DetectorObjectProfile) detectorObject;
            Rectangle limits = detectorObjectSimple
                    .getLimits(countsPerPixelThreshold);
            // logger.finer(" before intersection:
            // row0="+limits.getMinX()+" col0="+limits.getMinY()+"
            // row1="+limits.getMaxX()+" col1="+limits.getMaxY());
            limits = limits.intersection(bright.getRectangle());
            // logger.finer(" after intersection:
            // row0="+limits.getMinX()+" col0="+limits.getMinY()+"
            // row1="+limits.getMaxX()+" col1="+limits.getMaxY());
            // logger.finer(" in render for simple:
            // doPoisson="+doPoisson);
            if (!doPoisson || brightOrFaint(detectorObjectSimple, limits, psf)) {
                status = STATUS_BRIGHT;
                return renderBright(detectorObjectSimple, limits, psf);
            } 
            status = STATUS_FAINT;
            return renderProfile(detectorObjectSimple, psf);
        }
        return renderSurfaceBrightness((DetectorObjectSurfaceBrightness) detectorObject, psf);
    }

    /**
     * @param detectorObjectSimple
     * @param limits
     * @param psf 
     * @return true if this object should be rendered using the bright algorithm
     */
    private boolean brightOrFaint(DetectorObjectProfile detectorObjectSimple,
            Rectangle limits, PSF psf) {
    	if (!doPoisson) {
    		return true;
    	}
        double nPixel = limits.width * limits.height;
        double timeBright = nPixel * nGridBright * nGridBright
                * timePerSingleEvaluation;
        if (brightConvolveEachObject) {
        	if (psf == null) {
        		timeBright = 0.0;
        	} else {
        		try {
        			timeBright += nPixel * psf.ncol() * psf.nrow() * timePerPixelConvolve;
        		} catch (UnsupportedOperationException e) {
        			timeBright = 0.0;
        		}
        	}
        }
        double timeFaint = detectorObjectSimple.getNPhoton() * timePerPhoton;
        return (timeBright < timeFaint);
    }

    /*
    private double renderDetectorObjectObserved(DetectorObjectProduct dop) {
    	double retval = 0.0;
    	boolean highResolution = false;
    	PixeledImage pi = dop.makePixeledImage(countsPerPixelThreshold, highResolution);
    	if (pi != null) {
    		shapelet.add(pi);
    		retval = pi.getSum();
    	}
    	return retval;
    }
    */
    
    private double renderSurfaceBrightness(DetectorObjectSurfaceBrightness detectorObjectShapelet, PSF psf) {
        double shapeletFluxThrown = 0.0;

        PixeledImage pi;
    	double nPhotons = detectorObjectShapelet.getNPhoton();
        if ((pi = detectorObjectShapelet.getPixeledImage()) != null) {
        	
        	shapelet.add(pi, pd, nPhotons);
//        	
//        	Hashtable centroid = pi.centroid(0.);
//        	double rowSigma = ((Double) centroid.get("rowSigma")).doubleValue();
//			double rowCentroid = ((Double) centroid.get("rowCentroid"))
//					.doubleValue();
//			double colCentroid = ((Double) centroid.get("colCentroid"))
//					.doubleValue();
//			System.out.println("sigma=" + rowSigma + " row centroid="
//					+ rowCentroid + " colCentroid=" + colCentroid);

        	status = pi.getRow0()+pi.getCol0();
        	Object[] row = {new Double(nPhotons), "pixeledImage"};
        	renderedShapelet.addRow(row);  
         	return nPhotons;
        } 
        new Squares();
        //

        int iRow = (int) Math.floor(detectorObjectShapelet.row);
        int iCol = (int) Math.floor(detectorObjectShapelet.col);

        DetectorObjectSurfaceBrightness convolvedObject;
        if (psf == null) {
        	convolvedObject = detectorObjectShapelet;
        } else {       	       	
        	convolvedObject = detectorObjectShapelet.convolve(psf);
        }
        double maxFlux = convolvedObject.nPhoton;
        int[] nPixelsPerSquare = Squares.getNPixelsPerSquare();

        if (convolvedObject.pixeledImage != null) {
        	shapelet.add(convolvedObject.pixeledImage, pd);
        } else {
        	//ArrayList<Double> fluxToAdd = new ArrayList<Double>();
        	for (int iSquare = 0; iSquare < Squares.SQUARES_MAX_RADIUS; iSquare++) {
        		//logger.finer(" iSquare="+iSquare+" maxFlux="+maxFlux);
        		if (maxFlux < countsPerPixelThreshold && iSquare > minSquaresToRender) {
        			status = iSquare;
        			//Double[] temp = {new Double(detectorObjectShapelet.getRow()), new Double(detectorObjectShapelet.getCol()), new Double(shapeletFluxThrown)};
        			//rowColShapeletFluxThrown.add(temp);
        			Object[] row = {new Double(shapeletFluxThrown), "nSquares="+status};
        			renderedShapelet.addRow(row);
        	       	return shapeletFluxThrown;
        		}
        		maxFlux = 0;

        		for (int iPixel = 0; iPixel < nPixelsPerSquare[iSquare]; iPixel++) {       		
        			int ir = iRow + Squares.squares[0][iSquare][iPixel];
        			int ic = iCol + Squares.squares[1][iSquare][iPixel];
        			double deltaFlux = convolvedObject.getFluxByPix(ir, ic);
//      			double deltaFlux = convolvedObject.getFluxByPixSlow(ir, ic);

        			if (preventNegativeFluxValues) {
        				deltaFlux = Math.max(deltaFlux, 0.0);
        			}	
        			// TODO integrate the efficiency inside the pixel correctly
        			double flux = pd.efficiencyTransform(deltaFlux, ir + 0.5, ic + 0.5);  
        			try {
        				shapeletFluxThrown += flux;
        				shapelet.incrByRowColNoOffset(ir, ic, flux);

        			} catch (ArrayIndexOutOfBoundsException e) { // do nothing if out of bounds
        			}
        			if (flux > maxFlux) {
        				maxFlux = flux;
        			}	
        		}
        	}
        }
        status = Squares.SQUARES_MAX_RADIUS;

        //Double[] temp = {new Double(detectorObjectShapelet.getRow()), new Double(detectorObjectShapelet.getCol()), new Double(shapeletFluxThrown)};
        //rowColShapeletFluxThrown.add(temp);
        Object[] row = {new Double(shapeletFluxThrown), "allSquares="+status};
        renderedShapelet.addRow(row);
        return shapeletFluxThrown;
    }


    private double renderProfile(DetectorObjectProfile detectorObject, PSF psf) {
        int nToDo = simRandomPoisson.next(detectorObject.getNPhoton());
        int nThrown = 0;
        for (int i = 0; i < nToDo; i++) {
            Point2D.Double detectorPoint = detectorObject.getNextLocation();
            Point2D.Double psfPoint = psf.nextPoint();
            
            int r = (int) Math.floor(detectorPoint.y+psfPoint.y);
            int c = (int) Math.floor(detectorPoint.x+psfPoint.x);

            double eff = pd.efficiencyTransform(1.0, r, c);
            if (simRandomUniform.next() < eff) {
                try {
                    faint.incrByRowColNoOffset(r, c, 1.0);
                    nThrown++;
                } catch (ArrayIndexOutOfBoundsException e) { // do nothing if out of bounds

                }
            }
        }
        double fluxThrown = nThrown;
        Object[] row = {new Double(fluxThrown), "profile"};
        renderedSimple.addRow(row);
        return fluxThrown;
    }

    /**
     * render this object using the bright strategy
     * @param detectorObject object to render 
     * @param limits only render inside this rectangle
     * @param psf convole with this psf
     * @return the thrown flux, used for debuggin
     */
    public double renderBright(DetectorObjectProfile detectorObject, Rectangle limits, PSF psf) {
        // logger.finer("in renderBright:
        // brightConvolveEachObject="+brightConvolveEachObject);
        if (brightConvolveEachObject) {
            try {
            	StopwatchCollection.start("PixeledImageFactory.planAndProduce rbWithConv");
                double retval =  renderBrightWithConvolution(detectorObject,limits, psf);
            	StopwatchCollection.stop("PixeledImageFactory.planAndProduce rbWithConv");
                return retval;
            } catch (NullPointerException e) {
                // if psf is null, fall through to renderBrightNoConvolution
            }
        }
    	StopwatchCollection.start("PixeledImageFactory.planAndProduce rbNoConv");
        double retval = renderBrightNoConvolution(detectorObject, limits);
    	StopwatchCollection.stop("PixeledImageFactory.planAndProduce rbNoConv");
        return retval;
    }


    private double renderBrightNoConvolution(DetectorObjectProfile detectorObject,
            Rectangle limits) {    	
        int rowMin = (int) Math.max(0, limits.getMinY());
        int rowMax = (int) Math.min(bright.getRowCount(), limits.getMaxY());
        int colMin = (int) Math.max(0, limits.getMinX());
        int colMax = (int) Math.min(bright.getColumnCount(), limits.getMaxX());
        double fluxThrown = 0;
        // logger.finer(" in renderBrightNoConvolution: rowMin="+rowMin+"
        // colMin="+colMin+" rowMax="+rowMax+" colMax="+colMax);
        for (int row = rowMin; row < rowMax; row++) {
            for (int col = colMin; col < colMax; col++) {
                double flux = detectorObject.getIntegratedFlux(row,
                        col, row + 1,  col + 1,
                        nGridBright);
                // logger.finer("in renderBrightNoConvolution: row="+row+"
                // col="+col+" before flux="+flux);
                flux = pd.efficiencyTransform(flux, row, col);
                if (doPoisson) {
                	if (poissonFactor > 0.0) {
                		flux /= poissonFactor;
                		flux = simRandomPoisson.next(flux);
                		flux *= poissonFactor;
                	}
                }
                bright.incrByRowColNoOffset(row, col, flux);
                fluxThrown += flux;
            }
        }
        Object[] row = {new Double(fluxThrown), "brightNoConvolution"};
        renderedSimple.addRow(row);
        return fluxThrown;
    }


    private double renderBrightWithConvolution(DetectorObjectProfile detectorObject, Rectangle limits, PSF psf) {
    	Point row0Col0 = new Point();
        double[][] convolvedArray = detectorObject.convolve(psf, limits, nGridBright, row0Col0);
        int row0 = (int) Math.floor(row0Col0.y);
        int col0 = (int) Math.floor(row0Col0.x);

        // logger.finer(" in renderBrightWithConvolution: row0="+row0+"
        // col0="+col0);
        double fluxThrown = 0.0;
        double flux;
        for (int row = 0; row < convolvedArray.length; row++) {
            int bRow = row + row0;
            if (bRow >= 0 && bRow < bright.getRowCount()) {
                for (int col = 0; col < convolvedArray[0].length; col++) {
                    int bCol = col + col0;
                    if (bCol >= 0 && bCol < bright.getColumnCount()) {
                    	flux = convolvedArray[row][col];
                        bright.incrByRowColNoOffset(bRow, bCol, flux);
                        fluxThrown++;
                    }
                }
            }
        }
        Object[] row = {new Double(fluxThrown), "brightWithConvolution"};
        renderedSimple.addRow(row);
        return fluxThrown;
    }

    /**
     * set the number of grid points in each direction to sample pixels for bright images
     * @param nGridBright the number
     */
    public void setNGridBright(int nGridBright) {
        this.nGridBright = nGridBright;
    }

    /**
     * @return nGridBright
     */
    public int getNGridBright() {
        return nGridBright;
    }

    /**
     * automatically set the parameters used to select the most efficient algorithms to use for rendering
     * @param nToDo number of objects to render in this test
     * @param nGridBrightTemp temportary number of grid points to sample pixels, in each direction, for bright objects
     * @param ps destination for informative commentary
     * @throws Exception
     */
    public void setTimings(int nToDo, int nGridBrightTemp, PrintStream ps)
            throws Exception {
        timePerPixelConvolve = measureTimePerPixelConvolve(nToDo);
        ps.println("timePerPixelConvolve=" + timePerPixelConvolve);
        timePerSingleEvaluation = measureTimePerSingleEvaluation(nToDo,
                nGridBrightTemp);
        ps.println("timePerSingleEvaluation=" + timePerSingleEvaluation);
        timePerPhoton = measureTimePerPhoton(nToDo);
        ps.println("timePerPhoton=" + timePerPhoton);
    }

    /**
     * measure the time for a single evaluation; called from setTimings
     * @param nToDo
     * @param nGridBrightTemp
     * @return the time for a single evaluation, in seconds
     * @throws Exception
     */
    public double measureTimePerSingleEvaluation(int nToDo, int nGridBrightTemp)
            throws Exception {
        boolean brightConvolveEachObjectSave = brightConvolveEachObject;
        setBrightConvolveEachObject(false);
        PixeledImage brightSave = bright;
        int nGridBrightSave = nGridBright;
        nGridBright = nGridBrightTemp;
        bright = new PixeledImage(1000, 1000);
        double time;
        double row = 51.1;
        double col = 62.9;
        double nPhoton = 1000000;
        String shape = "gauss";
        double scale = 15.0;
        double ellipticity = 0.0;
        double positionAngle = 0.0;
        Profiles profiles = new Profiles();
        DetectorObjectProfile dos = new DetectorObjectProfile(row, col, nPhoton,
                shape, scale, ellipticity, positionAngle, profiles);

        double countsPerPixelThresholdCalculated = dos.getFluxNorm() / 1000.0;
        Rectangle limits = dos.getLimits(countsPerPixelThresholdCalculated);
        int nPixel = limits.width * limits.height;

        long start = System.currentTimeMillis();
        for (int i = 0; i < nToDo; i++) {
            this.renderBright(dos, limits, null);
        }
        long stop = System.currentTimeMillis();
        //long millis = stop - start;

        time = 1e3 * (stop - start) / nToDo / nPixel / nGridBright
                / nGridBright;
        bright = brightSave;
        nGridBright = nGridBrightSave;
        setBrightConvolveEachObject(brightConvolveEachObjectSave);
        return time;

    }

    /**
     * measure the time per photon to render faint objects, called from setTimings()
     * @param nToDo
     * @return the time per photon, in seconds
     * @throws Exception
     */
    public double measureTimePerPhoton(int nToDo) throws Exception {
        boolean brightConvolveEachObjectSave = brightConvolveEachObject;
        setBrightConvolveEachObject(false);
        PixeledImage faintSave = faint;
        int nGridBrightSave = nGridBright;
        faint = new PixeledImage(1000, 1000);
        double time;
        double row = 51.1;
        double col = 62.9;
        double nPhoton = 100000;
        String shape = "gauss";
        double scale = 15.0;
        double ellipticity = 0.0;
        double positionAngle = 0.0;
        Profiles profiles = new Profiles();
        DetectorObjectProfile dos = new DetectorObjectProfile(row, col, nPhoton,
                shape, scale, ellipticity, positionAngle, profiles);
        PSF psf = psfFactory.getPsf(dos);
        long start = System.currentTimeMillis();
        for (int i = 0; i < nToDo; i++) {
            this.renderProfile(dos, psf);
        }
        long stop = System.currentTimeMillis();
        long millis = stop - start;

        time = 1e3 * (millis) / nToDo / nPhoton;
        faint = faintSave;
        setBrightConvolveEachObject(brightConvolveEachObjectSave);
        nGridBright = nGridBrightSave;
        return time;

    }

    /**
     * measure the time per pixel for convolving, called from setTimings()
     * @param nToDo
     * @return the time per pixel, in seconds
     * @throws Exception
     */
    public double measureTimePerPixelConvolve(int nToDo) throws Exception {
        double time;
        double trueRow = 34.56;
        double trueCol = 76.54;
        double trueSigma1 = 5.5;
        double trueSigma2 = 4.4;
        Profiles profiles = new Profiles();
        DetectorObjectProfile dos = new DetectorObjectProfile(trueRow, trueCol,
                1000, "gauss", trueSigma1, 0.0, 0.0, profiles);
        Rectangle limits = dos.getLimits(1.0);
        double nPixObject = limits.width * limits.height;
        PSF psf = new PSFGaussian(trueSigma2);
        double nPixPsf = psf.nrow() * psf.ncol();
        Point rowCol0 = new Point();
        long start = System.currentTimeMillis();
        for (int i = 0; i < nToDo; i++) {
            dos.convolve(psf, limits, 1, rowCol0);
        }
        long stop = System.currentTimeMillis();
        time = 1e3 * (stop - start) / nToDo / nPixObject / nPixPsf;
        return time;
    }
    
    /**
     * get the objects (profile and surface brightness) from this factory in a linked list, 
     * @return a linked list containing the two StarTables
     */
    public LinkedList<RowListStarTable> getDetectorObjectTables() {
        RowListStarTable profile = DetectorObjectProfile.createRowListStarTable();
        RowListStarTable surfaceBrightness = DetectorObjectSurfaceBrightness.createRowListStarTable();
        int nDetectorObjects = 0;
        for (DetectorObject temp : detectorObjects) {
        	nDetectorObjects++;
        	if (temp != null) {
        		if (temp.getClass() == DetectorObjectProfile.class) {
        			temp.addToTable(profile);
        		} else if (temp.getClass() == DetectorObjectSurfaceBrightness.class) {
        			((DetectorObjectSurfaceBrightness)temp).addToTable(surfaceBrightness);
        		} else if (temp.getClass() == DetectorObjectProduct.class ){
        			((DetectorObjectProduct)temp).addToTable(surfaceBrightness);
        		} else {
        			throw new Error(" what kind of object is temp with temp.getClass()="+temp.getClass());
        		}
        	}
         }
        LinkedList<RowListStarTable> retVal = new LinkedList<RowListStarTable>();
        retVal.add(profile);
        retVal.add(surfaceBrightness);
        return retVal;
    }
    
    /**
     * @return a list with the two RowListStarTable tables in it that track the number of photons thrown, renderedSimple and renderedShapelet
     */
    public LinkedList<RowListStarTable> getRenderedTables() {
    	LinkedList<RowListStarTable> retval = new LinkedList<RowListStarTable>();
    	retval.add(renderedSimple);
    	retval.add(renderedShapelet);
    	retval.add(renderedProduct);
    	return retval;
    }
    
    /**
     * write a file to be used in ds9 to define regions
     * @param baseFileName the base of the file name to write
     */
    public void writeDs9RegionFiles(String baseFileName)  {
        PrintStream rowcolPs, radecPs;
		try {
			rowcolPs = new PrintStream(new FileOutputStream(new File(baseFileName+".ds9.rowcol.reg")));
	        radecPs = new PrintStream(new FileOutputStream(new File(baseFileName+".ds9.radec.reg")));
		} catch (FileNotFoundException e) {
			throw new Error(e);
		}
        rowcolPs.println("IMAGE");
        radecPs.println("FK5");
        for (DetectorObject obj : detectorObjects) {
        	if (obj != null) {
        		double circleRadiusInPix = 10;
        		double circleRadiusInDeg = circleRadiusInPix*0.2/3600;
        		rowcolPs.println("circle "+(obj.getCol()+overscanColumns)+" "+obj.getRow()+" "+circleRadiusInPix);
        		radecPs.println("circle "+obj.getRa()+" "+obj.getDec()+" "+circleRadiusInDeg);
        	}
       }
       rowcolPs.close();
       radecPs.close();
    }

    /**
     * set read noise, gain, bias for a 2-amp detector
     * @param readNoise0 sigma (electrons)
     * @param readNoise1 sigma (electrons)
     * @param gain0 electrons/ADU
     * @param gain1 electrons/ADU
     * @param bias0 electrons
     * @param bias1 electrons
     */
    public void setReadnoiseGainBias(double readNoise0, double readNoise1, double gain0, double gain1, double bias0 ,double bias1) {
    	int[] splitColumns = {pd.columnCount/2};

    	// readnoise
    	PixelValueTransformation[] bias = {
    			new PixelValueTransformationAddConstant(bias0,"bias"),
    			new PixelValueTransformationAddConstant(bias1,"bias")
    	};
    	PixelValueTransformation pvtBias = new PixelValueTransformationSplit(splitColumns, bias);
    	pd.putElectronics(pvtBias);
    	
    	// readnoise
    	PixelValueTransformation[] readNoise = {
    			new PixelValueTransformationConstantGaussianNoise(readNoise0,"read noise", simRandomNormal),
    			new PixelValueTransformationConstantGaussianNoise(readNoise1,"read noise", simRandomNormal)
    	};
    	PixelValueTransformation pvtReadNoise = new PixelValueTransformationSplit(splitColumns, readNoise);
    	pd.putElectronics(pvtReadNoise);

    	// gain
    	PixelValueTransformation[] gain = {
    			new PixelValueTransformationMultiplyConstant(1.0/gain0,"gain"),
    			new PixelValueTransformationMultiplyConstant(1.0/gain1,"gain")
    	};
    	PixelValueTransformation pvtGain = new PixelValueTransformationSplit(splitColumns, gain);
    	pd.putElectronics(pvtGain);    	
    }
    
    /**
     * set the fraction used to call PixeledImage.addMirrorCrosstalk()
     * @param mirrorCrosstalkFraction
     */
    public void setMirrorCrosstalkFraction(double mirrorCrosstalkFraction) {
    	this.mirrorCrosstalkFraction = mirrorCrosstalkFraction;
    }
    
    /**
     * sets up clocking charge for the factory
     * @param clockCharge
     * @param clockChargeRange
     * @param clockChargeVerbosity
     */
    public void setClockCharge(ClockCharge clockCharge, Rectangle clockChargeRange, int clockChargeVerbosity) {
    	this.clockCharge = clockCharge;
    	this.clockChargeRange = clockChargeRange;
    	this.clockChargeVerbosity = clockChargeVerbosity;
    }
}
