package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.sim.repository.image.ImageRepository;
import gov.fnal.eag.wcs.TransformLinear;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Hashtable;

import javax.imageio.ImageIO;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.annotations.XYLineAnnotation;
import org.jfree.chart.annotations.XYShapeAnnotation;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeriesCollection;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.UnflaggedOption;

/**
 * 
 * make a plot (or PNG) of pixel values for display
 * @author stoughto
 *
 */
public class PixeledImagePlot {

	private PixeledImage pi;
	private ScaleTable st;
	private ColorTable ct;

	/**
	 * @param pi the image to plot
	 * @param st scale from pixel value to an 8-bit value
	 * @param ct transform the 8-bit value to a color
	 */
	public PixeledImagePlot(PixeledImage pi, ScaleTable st, ColorTable ct) {
		this.pi = pi;
		this.st = st;
		this.ct = ct;
	}
	
	/**
	 * write the plot to a file and return the JFreeChart
	 * @param fileName name of the file to create
	 * @param pngWidth width of the plot in display pixels (the height is calculate so the shape matches the PixeledImage)
	 * @return the chart
	 */
	public JFreeChart toPng(String fileName, int pngWidth) {
		JFreeChart chart = getChart();
		int pngHeight = (int) (pngWidth * (double) pi.getRowCount() / pi.getColumnCount());		
		try {
			ChartUtilities.saveChartAsPNG(new File(fileName), chart, pngWidth, pngHeight);
		} catch (IOException e) {
			throw new Error(e);
		}		
		return chart;
	}

	/**
	 * @param width
	 * @param zoom
	 * @param glyphs
	 * @return the buffered image
	 */
	public BufferedImage makeBufferedImage(int width, double zoom, ArrayList<ImageGlyph> glyphs) {
		double scaleFactor = (double) width / (double) pi.getColumnCount();
		int height = (int) (pi.getRowCount() * scaleFactor);
		BufferedImage bufferedImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
		
		//take row0,col0 into consideration correctly
		
		AffineTransform affineTransform = new AffineTransform();
		affineTransform.scale(scaleFactor, -scaleFactor);
		affineTransform.translate(0, -pi.getRowCount());
		affineTransform.translate(pi.getColumnCount()/2.0, pi.getRowCount()/2.0);
		affineTransform.scale(zoom, zoom);
		affineTransform.translate(-pi.getColumnCount()/2.0, -pi.getRowCount()/2.0);
		
		affineTransform.translate(-pi.getCol0(), -pi.getRow0());
		
        Graphics2D g2d = bufferedImage.createGraphics();
        double strokeWidth = 1.0/scaleFactor/zoom;
        g2d.setStroke(new BasicStroke((float) strokeWidth));

        g2d.setColor(Color.blue);
        g2d.fillRect(0, 0, width, height);
        
        g2d.setTransform(affineTransform);
        Color defaultColor = g2d.getColor();
        for (int r=pi.getRowMin(); r<pi.getRowMax(); r++) {
        	for (int c=pi.getColumnMin(); c<pi.getColumnMax(); c++) {
				double value = pi.getByRowCol(r,c);
				double scaledValue = st.lookup(value);
				Color color = ct.get((int) scaledValue);
				if (color == null) throw new Error(" scaledValue="+scaledValue+" return null color for ct="+ct);
				g2d.setColor(color);
				g2d.fillRect(c, r, 1, 1);
        	}
        }
        g2d.setColor(defaultColor);
        if (glyphs != null) {
        	for (ImageGlyph glyph : glyphs) {
        		glyph.render(g2d);
        	}
        }
        /*
        if (!Double.isNaN(fraction)) {
        	g2d.setColor(Color.black);
        	g2d.fillRect(0, 0, width, 10);
        	g2d.setColor(Color.white);
        	g2d.fillRect(0, 0, (int) (width*fraction), 10);
        }
        */
        // Graphics context no longer needed so dispose it
        g2d.dispose();
        return bufferedImage;
	}
	
	
	/**
	 * plots the PixeledImage in the BufferedImage
	 * @param bufferedImage
	 * @param biColRowToMm
	 * @param colRowToMm
	 * @param glyphs
	 */
	public void plotInBufferedImage(BufferedImage bufferedImage, TransformLinear biColRowToMm, TransformLinear colRowToMm, ArrayList<ImageGlyph> glyphs) {
		AffineTransform affineTransform = new AffineTransform();
		affineTransform.scale(1.0, -1.0);
		affineTransform.translate(0, -bufferedImage.getHeight());
		affineTransform.concatenate(biColRowToMm.getInverse());
		affineTransform.concatenate(colRowToMm.getForward());
        Graphics2D g2d = bufferedImage.createGraphics();
        g2d.setTransform(affineTransform);
        Color defaultColor = g2d.getColor();
        int stepSize = Math.max(1, (int) (biColRowToMm.arcsecPerPixel()/colRowToMm.arcsecPerPixel()));
        for (int r=pi.getRowMin(); r<pi.getRowMax(); r+=stepSize) {
        	for (int c=pi.getColumnMin(); c<pi.getColumnMax(); c+=stepSize) {
				double value = pi.getByRowCol(r+stepSize/2.0,c+stepSize/2);
				double scaledValue = st.lookup(value);
				Color color = ct.get((int) scaledValue);
				if (color == null) throw new Error(" scaledValue="+scaledValue+" return null color for ct="+ct);
				g2d.setColor(color);
				g2d.fillRect(c, r, stepSize, stepSize);
        	}
        }
        g2d.setColor(defaultColor);
        if (glyphs != null) {
        	for (ImageGlyph glyph : glyphs) {
        		glyph.render(g2d);
        	}
        }
	}

	
	/**
	 * make the actual PNG file
	 * @param fileName where to write the PNG
	 * @param binFactor number of pixels to combine
	 * @param expand if true, expand one input pixel to binFactor plotted pixels; if false, combine binFactor input pixels to one plot pixel
	 * @param fraction fraction of the plot area to fill with the image
	 */
	public void makePng(String fileName, int binFactor, boolean expand, double fraction) {
		int width = pi.getColumnCount();
		int height = pi.getRowCount();
		if (expand) {
			width *= binFactor;
			height *= binFactor;
		} else {
			width /= binFactor;
			height /= binFactor;
		}
		BufferedImage bufferedImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
        Graphics2D g2d = bufferedImage.createGraphics();
        for (int iw=0; iw<width; iw++) {
        	for (int ih=0; ih<height; ih++) {
        		if (expand) {
    				double value = pi.getByRowColNoOffset(ih,iw);
    				double scaledValue = st.lookup(value);
    				Color color = ct.get((int) scaledValue);
    				if (color == null) throw new Error(" scaledValue="+scaledValue+" return null color for ct="+ct);
    				g2d.setColor(color);
    				g2d.fillRect(iw*binFactor, ih*binFactor, binFactor, binFactor);        			
        		} else {
        			double value = 0;
        			for (int row=iw*binFactor; row<(iw+1)*binFactor; row++) {
            			for (int col=ih*binFactor; col<(ih+1)*binFactor; col++) {
            				value += pi.getByRowColNoOffset(ih,iw);
            			}
        			}
        			value /= (binFactor*binFactor);
    				double scaledValue = st.lookup(value);
    				Color color = ct.get((int) scaledValue);
    				if (color == null) throw new Error(" scaledValue="+scaledValue+" return null color for ct="+ct);
    				g2d.setColor(color);
    				g2d.fillRect(iw, ih, 1, 1);        			
        		}
        	}
        }

        if (!Double.isNaN(fraction)) {
        	g2d.setColor(Color.black);
        	g2d.fillRect(0, 0, width, 10);
        	g2d.setColor(Color.white);
        	g2d.fillRect(0, 0, (int) (width*fraction), 10);
        }
        g2d.dispose();
        writeBufferedImage(fileName, bufferedImage); 
	}

	/**
	 * @param fileName where to write the image as a png file
	 * @param bufferedImage the image to write
	 * @throws Error if there is IO trouble
	 */
	public static void writeBufferedImage(String fileName, BufferedImage bufferedImage) throws Error {
		File file = new File(fileName);
        try {
			ImageIO.write(bufferedImage, "png", file);
		} catch (IOException e) {
			throw new Error(e);
		}
	}
	

	/**
	 * Default getChart() automatically sets mask markers to green
	 * @return chart of the pixeled image
	 */
	public JFreeChart getChart() {
		return getChart(Color.GREEN);
	}
	/**
	 * create a JFreeChart ready for plotting with pixel markers for mask
	 * @param pixelMarkerColor the color to make masked pixel markers
	 * @return chart of the pixeled image
	 */
	public JFreeChart getChart(Color pixelMarkerColor) {
		XYDataset xySeriesCollection =  new XYSeriesCollection();
		JFreeChart chart = ChartFactory.createScatterPlot("","col","row", xySeriesCollection, PlotOrientation.VERTICAL, true, true, true);
		XYPlot xyPlot = chart.getXYPlot();
		
        NumberAxis domainAxis = (NumberAxis) chart.getXYPlot().getDomainAxis();
        domainAxis.setRange(pi.getColumnMin(), pi.getColumnMax());
        NumberAxis rangeAxis = (NumberAxis) chart.getXYPlot().getRangeAxis();
        rangeAxis.setRange(pi.getRowMin(), pi.getRowMax());
		
		Stroke stroke = new BasicStroke();

		BigInteger zero = BigInteger.ZERO;
		for (int col=pi.getColumnMin(); col<pi.getColumnMax(); col++) {
			for (int row=pi.getRowMin(); row<pi.getRowMax(); row++) {
				double value = pi.getByRowCol(row,col);
				double scaledValue = st.lookup(value);
				Color color = ct.get((int) scaledValue);
				if (color == null) throw new Error(" scaledValue="+scaledValue+" return null color for ct="+ct);
				Shape shape = new Rectangle(col, row, 1, 1);
				XYShapeAnnotation annotation = new XYShapeAnnotation(shape, stroke, color, color); 
				xyPlot.addAnnotation(annotation);
				BigInteger maskValue = getMaskValue(pi, row,col);
				if (maskValue != zero) {
					xyPlot.addAnnotation(new XYLineAnnotation(col, row, col+1, row+1, new BasicStroke(), pixelMarkerColor));
					xyPlot.addAnnotation(new XYLineAnnotation(col+1, row, col, row+1, new BasicStroke(), pixelMarkerColor));
				}
			}
		}
		return chart;
	}

	private BigInteger getMaskValue(PixeledImage pi2, int row, int col) {
		if (pi2.getMask() == null) {
			return BigInteger.ZERO;
		}
		return pi2.getMask().getByRowCol(row-pi2.getRow0(),col-pi2.getCol0());
	}

	/**
	 * generate a PNG file from a FITS image
	 * @param args  -scale is one of linear, histeq, log, sqrt; -color is rainbow by default; -upper and -lower optionally set upper and lower pixel values inputFitsFile outputPNGFile
	 * @throws Exception if there is trouble
	 */
	public static void main(String[] args) throws Exception {
        String imageRepositoryDefault = "imageRepositoryDefault";
		
		JSAP jsap = new JSAP();
        
		FlaggedOption bufferedImageOpt = new FlaggedOption("bufferedImage")
        .setStringParser(JSAP.BOOLEAN_PARSER)
        .setDefault("true")
        .setRequired(false)
        .setLongFlag("bufferedImage");
        jsap.registerParameter(bufferedImageOpt);
        		
		FlaggedOption factorOpt = new FlaggedOption("factor")
        .setStringParser(JSAP.DOUBLE_PARSER)
        .setDefault("2.0")
        .setRequired(false)
        .setShortFlag('f')
        .setLongFlag("factor");
        jsap.registerParameter(factorOpt);
        		
		FlaggedOption scaleOpt = new FlaggedOption("scale")
        .setStringParser(JSAP.STRING_PARSER)
        .setDefault("histeq")
        .setRequired(false)
        .setShortFlag('s')
        .setLongFlag("scale");
        jsap.registerParameter(scaleOpt);

		FlaggedOption colorOpt = new FlaggedOption("color")
        .setStringParser(JSAP.STRING_PARSER)
        .setDefault("rainbow")
        .setRequired(false)
        .setShortFlag('c')
        .setLongFlag("color");
        jsap.registerParameter(colorOpt);

        FlaggedOption imageRepositoryOpt = new FlaggedOption("imageRepository")
        .setStringParser(JSAP.STRING_PARSER)
        .setDefault(imageRepositoryDefault)
        .setRequired(false)
        .setShortFlag('r')
        .setLongFlag("repository");
        jsap.registerParameter(imageRepositoryOpt);
        

        FlaggedOption widthOpt = new FlaggedOption("width")
        .setStringParser(JSAP.INTEGER_PARSER)
        .setDefault("500")
        .setRequired(false)
        .setShortFlag('w')
        .setLongFlag("width");
        jsap.registerParameter(widthOpt);

        FlaggedOption lowerOpt = new FlaggedOption("lowerValue")
        .setStringParser(JSAP.DOUBLE_PARSER)
        .setRequired(false)
        .setShortFlag('l')
        .setLongFlag("lowerValue");
        jsap.registerParameter(lowerOpt);

        FlaggedOption upperOpt = new FlaggedOption("upperValue")
        .setStringParser(JSAP.DOUBLE_PARSER)
        .setRequired(false)
        .setShortFlag('u')
        .setLongFlag("upperValue");
        jsap.registerParameter(upperOpt);
        
        UnflaggedOption inputFileNameOpt = new UnflaggedOption("inputFileName")
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(true);
        jsap.registerParameter(inputFileNameOpt);

        UnflaggedOption outputFileNameOpt = new UnflaggedOption("outputFileName")
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(true);
        jsap.registerParameter(outputFileNameOpt);

        JSAPResult config = jsap.parse(args);    

        // check whether the command line was valid, and if it wasn't,
        // display usage information and exit.
        if (!config.success() ) {
            System.err.println();
            System.err.print("Usage: java "
                                + PixeledImagePlot.class.getName());
            System.err.println(" "
                                + jsap.getUsage());
            System.err.println();
            for (int i=0; i<args.length; i++) {
            	System.err.println("args["+i+"] = "+args[i]);
            }
            System.exit(1);
        }        
        
        String imageRepository = config.getString("imageRepository");
        ImageRepository ir;
        if (imageRepository.equals(imageRepositoryDefault)) {
        	ir = new ImageRepository();
        } else {
        	ir = new ImageRepository(imageRepository);
        }
        
        PixeledImage pi = ir.getPixeledImage(config.getString("inputFileName"));
        
        Hashtable<String, Double> extrema = pi.extrema();
        double minValue = extrema.get("min");
        if (config.userSpecified("lowerValue")) minValue = config.getDouble("lowerValue");
        double maxValue = extrema.get("max");
        if (config.userSpecified("upperValue")) maxValue = config.getDouble("upperValue");
        
        String scaleType = config.getString("scale");
        ScaleTable st;
        if (scaleType.equals("linear")) {
        	st = new ScaleTable(minValue, maxValue);
        } else if (scaleType.equals("histeq")) {
        	st = new ScaleTable(pi, minValue, maxValue);
        } else {
        	double factor = config.getDouble("factor");
         	st = new ScaleTable(scaleType, minValue, maxValue, factor);
        }
        
        String colorType = config.getString("color");
        ColorTable ct = new ColorTable(colorType);
        PixeledImagePlot pip = new PixeledImagePlot(pi, st, ct);
        String outputFileName = config.getString("outputFileName");
        int width = config.getInt("width");
        if (config.getBoolean("bufferedImage")) {
        	BufferedImage bufferedImage = pip.makeBufferedImage(width, 1.0, null);
        	ImageRepository.toPng(outputFileName, bufferedImage);
        } else {
        	pip.toPng(outputFileName, width);
        	System.out.println("done writing outputFileName="+outputFileName+" with width="+width);
        }
	}
	
	/**
	 * Add shapes to a pixeled plot
	 * @param list the list of shapes to add to the plot
	 * @param plot the plot to add shapes to
	 */
	public void addXYShapeAnnotations(ArrayList<XYShapeAnnotation> list, XYPlot plot) {
		for (XYShapeAnnotation shape : list) {
			plot.addAnnotation(shape);
		}
	}
}