
package gov.fnal.eag.sim.pixel;

import java.math.BigInteger;

import nom.tam.fits.Header;

/**
 * array of mask values to keep score of pixels in a PixeledImage
 */
public class PixeledMask {
	/**
	 * if the equals methods is going to return false, and if verbose is true, it prints an informative message to System.out before returning
	 */
	public static boolean verbose = false;
	/**
	 * starting row of parent mask
	 */
	private int row0 = 0;
	/**
	 * starting column of parent mask
	 */
	private int col0 = 0;
	/**
	 * the mask values
	 */
	private BigInteger[][] pix;
	/**
	 * the parent PixeledMask, null if this was created ex nihilo
	 */
	public PixeledMask parent = null;
	/**
	 * FITS header cards not associated with writing this as a FITS image
	 */
	public Header header = new Header();
	
	int i0 = 0;
	int j0 = 0;
	int rowCount;
	int columnCount;
	/**
	 * construct a PixeledMask of specified size
	 * @param nrow number of rows
	 * @param ncol number of columns
	 */
	public PixeledMask(int nrow, int ncol) {
		BigInteger[][] temp = new BigInteger[nrow][ncol];
		setPix(temp);
		for (int row=0; row<rowCount; row++) {
			for (int col=0; col<columnCount; col++) {
				temp[row][col] = BigInteger.ZERO;
			}
		}
	}
	
	/**
	 * create a PixeledMask using the values in the 16 bits of pi values, shifted by the keword header value PIXMASK
	 * @param pi
	 */
	public PixeledMask(PixeledImage pi) {
		this(pi.getRowCount(), pi.getColumnCount());
		int fifteenBits = (int) (Math.pow(2, 16) - 1);
		int bitsToShift = pi.header.getIntValue("PIXMASK");
		setRow0Col0(pi.getRow0(),pi.getCol0());
		for (int row=pi.getRowMin(); row<pi.getRowMax(); row++) {
			for (int col=pi.getColumnMin(); col<pi.getColumnMax(); col++) {
				int value = (int) pi.getByRowCol(row, col);
				value = Math.max(0, value);
				value = Math.min(fifteenBits, value);
				BigInteger biValue = (new BigInteger(String.format("%d", value))).shiftLeft(bitsToShift);
				setByRowCol(row, col, biValue);
			}
		}
	}
	/**
	 * set the values of this PixeledMask
	 * @param pix array of values to set
	 */
	public void setPix(BigInteger[][] pix) {
    	this.pix = pix;
    	rowCount = pix.length;
    	columnCount = pix[0].length;
	}
	/**
	 * create a mask from the input array
	 * @param pixIn array of mask values
	 */
	public PixeledMask(BigInteger[][] pixIn) {
		BigInteger[][] tempPix = new BigInteger[pixIn.length][pixIn[0].length];
		setPix(tempPix);
		for (int row=0; row<rowCount; row++) {
			for (int col=0; col<columnCount; col++) {
				BigInteger temp = new BigInteger(pixIn[row][col].toByteArray());
				pix[row][col] = temp;
			}
		}
	}
	/**
	 * default constructor of 100x100 pixels
	 *
	 */
	public PixeledMask() {
		this(100,100);
	}
	/**
	 * copy constructor 
	 * @param pm the PixeledMask to copy
	 */
	public PixeledMask(PixeledMask pm) {
		this(pm.pix);
		row0 = pm.row0;
		col0 = pm.col0;
		parent = pm.parent;
		i0 = pm.i0;
		j0 = pm.j0;
	}
	/**
	 * sub-mask constructor
	 * @param pm parent pixel mask
	 * @param nrow number of rows in sub-mask
	 * @param ncol number of columns in sub-mask
	 * @param row0 initial row in parent
	 * @param col0 initial column in parent
	 */
	public PixeledMask(PixeledMask pm, int nrow, int ncol, int row0, int col0) {
		this(nrow,ncol);
		this.row0 = row0;
		this.col0 = col0;
		this.i0 = pm.i0;
		this.j0 = pm.j0;
		parent = pm;
		for (int row=0; row<nrow; row++) {
			boolean goodRow = row<pm.getRowCount()-row0;
			for (int col=0; col<ncol; col++) {
				boolean goodCol = col<pm.getColumnCount()-col0;
				if (goodRow && goodCol) {
					pix[row][col] = new BigInteger(pm.pix[row+row0][col+col0].toByteArray()); 
					} else {
					pix[row][col] = BigInteger.ZERO;					
				}
			}
		}
		
	}
	
	/**
	 * get the mask value
	 * @param row row position
	 * @param col column position
	 * @return the mask value
	 */
	public BigInteger getByRowCol(int row, int col) {
		return getByRowColNoOffset(row-row0,col-col0);
	}
	/**
	 * get the mask value
	 * @param row row position
	 * @param col column position
	 * @return the mask value
	 */
	public BigInteger getByRowCol(double row, double col) {
		return getByRowColNoOffset((int)Math.floor(row-row0), (int)Math.floor(col-col0));
	}
	/**
	 * get the mask value by position, ignoring row0, col0
	 * @param row row position
	 * @param col column position
	 * @return the mask value
	 */
	public BigInteger getByRowColNoOffset(int row, int col) {
		try {
			return pix[row+i0][col+j0];
		} catch (ArrayIndexOutOfBoundsException e) {
			return BigInteger.ZERO;
		}
	}
	/**
	 * get the mask value by position, ignoring row0, col0
	 * @param row row position
	 * @param col column position
	 * @return the mask value
	 */
	public BigInteger getByRowColNoOffset(double row, double col) {
		return getByRowColNoOffset((int)Math.floor(row), (int)Math.floor(col));
	}
	/**
	 * set the mask value by position
	 * @param row row position
	 * @param col column position
	 * @param value to set
	 */
	public void setByRowCol(int row, int col, BigInteger value) {
		setByRowColNoOffset(row-row0, col-col0, value);
	}
	/**
	 * and the value at row,col by the value
	 * @param row row position
	 * @param col column position
	 * @param value to set
	 */
	public void andByRowCol(int row, int col, BigInteger value) {
		andByRowColNoOffset(row-row0, col-col0, value);
	}
	/**
	 * or the value at row,col by the value
	 * @param row row position
	 * @param col column position
	 * @param value to set
	 */
	public void orByRowCol(int row, int col, BigInteger value) {
		orByRowColNoOffset(row-row0, col-col0, value);
	}
	/**
	 * set the mask value by position
	 * @param row row position
	 * @param col column position
	 * @param value to set
	 */
	public void setByRowCol(double row, double col, BigInteger value) {
		setByRowCol((int)Math.floor(row), (int)Math.floor(col), value);
	}
	/**
	 * set the mask value by position, ignoring row0, col0
	 * @param row row position
	 * @param col column position
	 * @param value to set
	 */
	public void setByRowColNoOffset(int row, int col, BigInteger value) {
		try {
			pix[row+i0][col+j0] = value;
		} catch (ArrayIndexOutOfBoundsException e) {  // Don't do anything if row,col is out of bounds
		}
	}
	/**
	 * and the mask value by position, ignoring row0, col0
	 * @param row row position
	 * @param col column position
	 * @param value to set
	 */
	public void andByRowColNoOffset(int row, int col, BigInteger value) {
		try {
			pix[row+i0][col+j0] = pix[row+i0][col+j0].and(value);
		} catch (ArrayIndexOutOfBoundsException e) {  // Don't do anything if row,col is out of bounds
		}
	}
	/**
	 * or the mask value by position, ignoring row0, col0
	 * @param row row position
	 * @param col column position
	 * @param value to set
	 */
	public void orByRowColNoOffset(int row, int col, BigInteger value) {
		try {
			pix[row+i0][col+j0] = pix[row+i0][col+j0].or(value);
		} catch (ArrayIndexOutOfBoundsException e) {  // Don't do anything if row,col is out of bounds
		}
	}
	/**
	 * set the mask value by position, ignoring row0, col0
	 * @param row row position
	 * @param col column position
	 * @param value to set
	 */
	public void setByRowColNoOffset(double row, double col, BigInteger value) {
		setByRowColNoOffset((int)row, (int)col, value);
	}
	/**
	 * set all pixels to value
	 * @param scalar the value
	 */
	public void setScalar(BigInteger scalar) {
		for (int row=row0; row<getRowMax(); row++) {
			for (int col=col0; col<getColumnMax(); col++) {
				setByRowCol(row,col,scalar);
			}
		}
	}
	/**
	 * AND all pixel values with value
	 * @param scalar the value
	 */
	public void andScalar(BigInteger scalar) {
		for (int row=row0; row<getRowMax(); row++) {
			for (int col=col0; col<getColumnMax(); col++) {
				andByRowCol(row,col,scalar);
			}
		}
	}
	/**
	 * OR all pixel values with value
	 * @param scalar the value
	 */
	public void orScalar(BigInteger scalar) {
		for (int row=row0; row<getRowMax(); row++) {
			for (int col=col0; col<getColumnMax(); col++) {
				orByRowCol(row,col,scalar);
			}
		}
	}
	/**
	 * AND pi values with this, using size of this
	 * @param pm the PixeledMask to use in AND
	 */
	public void and(PixeledMask pm) {
		for (int row=row0; row<getRowMax(); row++) {
			for (int col=col0; col<getColumnMax(); col++) {
				BigInteger newValue = getByRowCol(row,col).and(pm.getByRowCol(row,col));
				setByRowCol(row, col, newValue);
			}
		}	
	}
	/**
	 * OR pi values with this, using size of this
	 * @param pm the PixeledMask to use in OR
	 */
	public void or(PixeledMask pm) {
		for (int row=row0; row<getRowMax(); row++) {
			for (int col=col0; col<getColumnMax(); col++) {
				BigInteger newValue = getByRowCol(row,col).or(pm.getByRowCol(row,col));
				setByRowCol(row, col, newValue);
			}
		}	
	}
	/**
	 * copy pi values to this, using size of this
	 * @param pm the PixeledMask to copy
	 */
	public void copy(PixeledMask pm) {
		for (int row=row0; row<getRowMax(); row++) {
			for (int col=col0; col<getColumnMax(); col++) {
				BigInteger newValue = pm.getByRowCol(row,col);
				setByRowCol(row, col, newValue);
			}
		}
	}
	/**
	 * copy the pixel values back to parent
	 */
	public void copyBack() {
		for (int row=row0; row<getRowMax(); row++) {
			for (int col=col0; col<getColumnMax(); col++) {
				BigInteger newValue = getByRowCol(row,col);
				parent.setByRowCol(row, col, newValue);
			}
		}				
	}
	/**
	 * set the pixel values in the parent by AND with this pixel values 
	 */
	public void andBack() {
		for (int row=row0; row<getRowMax(); row++) {
			for (int col=col0; col<getColumnMax(); col++) {
				BigInteger newValue = parent.getByRowCol(row, col).and(getByRowCol(row,col));
				parent.setByRowCol(row, col, newValue);
			}
		}				
		
	}
	/**
	 * set the pixel values in the parent by OR with this pixel values 
	 */
	public void orBack() {
		for (int row=row0; row<getRowMax(); row++) {
			for (int col=col0; col<getColumnMax(); col++) {
				BigInteger newValue = parent.getByRowCol(row, col).or(getByRowCol(row,col));
				parent.setByRowCol(row, col, newValue);
			}
		}		
	}
	/**
	 * calculate the check sum
	 * @return the sum
	 */
	public BigInteger checkSum() {
		BigInteger retval = BigInteger.ZERO;
		for (int row=row0; row<getRowMax(); row++) {
			for (int col=col0; col<getColumnMax(); col++) {
				retval = retval.xor(getByRowCol(row,col));
			}
		}
		return retval;
	}
	/**
	 * figure out whether all pixel values are equal
	 * @param pm the other PixeledMask
	 * @return true if they are
	 */
	public boolean equals(PixeledMask pm) {
		for (int row=row0; row<getRowMax(); row++) {
			for (int col=col0; col<getColumnMax(); col++) {
				BigInteger thatValue = pm.getByRowCol(row,col);
				BigInteger thisValue = getByRowCol(row,col);
				if (thisValue.compareTo(thatValue) != 0) {
					if (verbose) {
						System.out.println("row0="+row0+" col0="+col0+" pm.row0="+pm.row0+" pm.col0="+pm.col0);
						System.out.println("PixeledMask.equal:  row="+row+" col="+col+" thisValue="+thisValue+" thatValue="+thatValue);
					}
					return false;
				}
			}
		}
		return true;
	}
	/**
	 * get the number of rows
	 * @return rowCount
	 */
	public int getRowCount() {
		return rowCount;
	}
	/**
	 * get the number of columns
	 * @return the columnCount
	 */
	public int getColumnCount() {
		return columnCount;
	}
	/**
	 * get the starting row position in the parent
	 * @return row0
	 */
	public int getRow0() {
		return row0;
	}
	/**
	 * @return the maximum row (not inclusive)
	 */
	public int getRowMax() {
		return row0+getRowCount();
	}
	/**
	 * get the starting column position in the parent
	 * @return col0
	 */
	public int getCol0() {
		return col0;
	}
	
	/**
	 * get the starting column position in the parent
	 * @return col0
	 */
	public int getColumnMin() {
		return getCol0();
	}
	/**
	 * @return the maximum column (not inclusive)
	 */
	public int getColumnMax() {
		return col0+getColumnCount();
	}
	/**
	 * get the smallest number of bytes required such that each pixel value can be expressed
	 * @return the number of bytes
	 */
	public int getMinByteLength() {
		int minByteLength = 0;
		for (int row=row0; row<getRowMax(); row++) {
			for (int col=col0; col<getColumnMax(); col++) {
				BigInteger newValue = getByRowCol(row,col);
				minByteLength = Math.max(minByteLength, newValue.toByteArray().length);
			}
		}
		return minByteLength;
	}
	/**
	 * @return a clone of the array of mask values
	 */
	public Object getPix() {
		BigInteger[][] retval = new BigInteger[getRowCount()][getColumnCount()];
		for (int row=row0; row<getRowMax(); row++) {
			for (int col=col0; col<getColumnMax(); col++) {
				BigInteger newValue = getByRowCol(row,col);
				retval[row-row0][col-col0] = newValue;
			}
		}
		return retval;
	}
	/**
	 * makes a PixeledMask sharing the values of this PixeledMask
	 * @param nRow number of rows in sub PixeledMask
	 * @param nCol number of columns in sub PixeledMask
	 * @param r0 starting row
	 * @param c0 ending row
	 * @return the resulting sub mask
	 * @throws IllegalArgumentException if the sub-array does not fit in the existing array
	 */
	public PixeledMask subPixeledMask(int nRow, int nCol, int r0, int c0) {
		int i0Sub = r0-row0+i0;
		int j0Sub = c0-col0+j0;
		if (i0Sub+nRow > rowCount) {
			throw new IllegalArgumentException("row0="+row0+" nRow="+nRow+" rowCount="+rowCount);
		}
		if (j0Sub+nCol > columnCount) {
			throw new IllegalArgumentException("col0="+col0+" nCol="+nCol+" columnCount="+columnCount);
		}
		PixeledMask retval = new PixeledMask(1,1);
		retval.pix = pix;
		retval.i0 = r0;
		retval.j0 = c0;
		retval.rowCount = nRow;
		retval.columnCount = nCol;
		retval.setRow0Col0(row0, col0);
		return retval;		
	}

    /**
     * set row0 and col0 to these values
     * @param row0
     * @param col0
     */
    public void setRow0Col0(int row0, int col0) {
    	this.row0 = row0;
    	this.col0 = col0;
    }
 }

