package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.sim.random.SimRandomUniform;
import gov.fnal.eag.util.astronomy.SimpleFunction;
import gov.fnal.eag.util.densityEstimate.KernelDensityEstimate;
import gov.fnal.eag.util.densityEstimate.KernelDensityEstimatePlotter;
import gov.fnal.eag.util.densityEstimate.UniformKernel;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;

/**
 * defines the profiles to be used for rendering objects in a PixeledImage.
 * This should be used in generating random numbers of specified 
 * distribution. So, these are functions of probability distribution.
 * 
 */

public class Profiles extends SimRandomUniform {

	/**
	 * the radius (in pixels) values for the profiles
	 */
	protected double radius[] = null;

	/**
	 * to generate gaussian shapes
	 */
	protected double gaussProfile[] = null;
	/**
	 * characteristic radius for the gaussian shape
	 */
	protected double reGauss = 1.0;
	/**
	 * to generate deVauc shapes
	 */
	protected double deVaucProfile[] = null;

	/**
	 * default maximum radius (in pixels) to use
	 */
	public static final double DEFAULT_RADIUS_MAX = 30.0;

	/**
	 * default number of points to use
	 */
	public static final int DEFAULT_NPOINTS = 3000;
	/**
	 * characteristic radius for the deVauc shape
	 */
	protected double reDevauc = 1.0;	
	/**
	 * to generate exponential shapes
	 */
	protected double expProfile[] =  null;
	/**
	 * characteristic radius for the exponential profile;
	 */
	protected double reExp = 1.0;
	/**
	 * to generate hyperbolic secant shapes
	 */
	protected double sechProfile[] =  null;
	/**
	 * characteristic radius for the hyperbolic secant profile;
	 * Better defined as the diffusion length
	 */
	protected double reSech = 1.0;
	
	private SimpleFunction kernel = new UniformKernel(1.0);
	private KernelDensityEstimate hgGauss, hgExp, hgDevauc, hgSech;

	
	
	/**
	 * creates a Profile with 3000 points from 0 to 30 pixels (default)
	 */
	public Profiles()  {
		this(DEFAULT_NPOINTS, DEFAULT_RADIUS_MAX);
	}
	/**
	 * creates a Profile with npoints from 0 to radiusMax pixels
	 * @param npoints number of points to sample
	 * @param radiusMax maximum radius, in pixels
	 */
	public Profiles(int npoints, double radiusMax) {
		
		this.setIseedIndex(123);
		
		radius = new double[npoints];
		double dr=radiusMax/npoints;

		gaussProfile = new double[npoints];		
		double gaussSum = 0;

		deVaucProfile = new double[npoints];
		double deVaucSum = 0;
			
		expProfile = new double[npoints];
		double expSum = 0;

		sechProfile = new double[npoints];
		double sechSum = 0;
	
		/* 
		 * Calculates the means of the (not normalized) probability distributions from radius = 0 
		 * to radiusMax using the formula: sum_over_i_of{x[i]*f(x[i])} and puts the results into
		 * the [shape]Sum variables. For example, if the mean of the gaussian distribution is 100,
		 * that is, gaussSum == 100, the electrons from a single source are on average 100 pixels
		 * away from the center (I think).
		 * The individual [shape]Profile arrays hold the mean radius of the probability density 
		 * distribution as a function of radius
		 */
		for (int i=0; i<npoints; i++) {
			radius[i] = (i)*dr;

			gaussSum += radius[i]*Math.exp(-0.5*radius[i]*radius[i]/(reGauss*reGauss));
			gaussProfile[i] = gaussSum;

			deVaucSum += radius[i]*Math.exp(-7.67*Math.pow(radius[i]/reDevauc,0.25));
			deVaucProfile[i] = deVaucSum;
				
			expSum += radius[i]*Math.exp(-radius[i]/reExp);
			expProfile[i] = expSum;
			
			sechSum += radius[i]/Math.cosh(radius[i]/reSech);
			sechProfile[i] = sechSum;
		}
		
		/* 
		 * Divides the mean radius of a given bin by the mean radius of the entire (not normalized)
		 * probability distribution. Continuing the example, where gaussSum used to equal 100, it is now
		 * equal to 1.
		 */
		for (int i=0; i<npoints; i++) {
			gaussProfile[i] /= gaussSum;
			deVaucProfile[i] /= deVaucSum;
			expProfile[i] /= expSum;
			sechProfile[i] /= sechSum;
		}

		try {
			// Create "blank" kernels
			hgGauss  = new KernelDensityEstimate(0,10*reGauss,  200,kernel);
			hgExp    = new KernelDensityEstimate(0,10*reExp,    200,kernel);
			hgDevauc = new KernelDensityEstimate(0,10*reDevauc, 200,kernel);
			hgSech   = new KernelDensityEstimate(0,10*reSech, 200,kernel);

		} catch (Exception e) {
			throw new Error(e);
		} 
	}	
	/**
	 * pick the next value of rho
	 * @param shape for this shape
	 * @param scale for this scale (in pixels)
	 * @return the rho value
	 */
	/*
	 * Randomly populates the kernel estimate using our profiles.
	 * Guesses random numbers and finds their indecies 
	 */ 
	public double rho(String shape, double scale) {
		double rho = 0.0;
		double r0, r1, p0, p1;
		// Choose a radius point according to the profile
		double uniformRandomNumber = next();
		if (shape.equalsIgnoreCase("gauss")) {
			int index = binarySearch(gaussProfile, uniformRandomNumber);
			r0 = radius[index-1];
			r1 = radius[index];
			p0 = gaussProfile[index-1];
			p1 = gaussProfile[index];
			rho = (r0 + (r1-r0)*(uniformRandomNumber-p0)/(p1-p0))/reGauss;
			//rho = reGauss*radius[index];
			hgGauss.add(rho); // 
			rho *= scale;
		} else if (shape.equalsIgnoreCase("devauc")) {
			int index = binarySearch(deVaucProfile, uniformRandomNumber);
			r0 = radius[index-1];
			r1 = radius[index];
			p0 = deVaucProfile[index-1];
			p1 = deVaucProfile[index];
			rho = (r0 + (r1-r0)*(uniformRandomNumber-p0)/(p1-p0))/reDevauc;
			hgDevauc.add(rho);
			rho *= scale;			
		} else if (shape.equalsIgnoreCase("exp")) {
			int index = binarySearch(expProfile, uniformRandomNumber);
			r0 = radius[index-1];
			r1 = radius[index];
			p0 = expProfile[index-1];
			p1 = expProfile[index];
			rho = (r0 + (r1-r0)*(uniformRandomNumber-p0)/(p1-p0))/reExp;
			hgExp.add(rho);
			rho *= scale;
		} else if (shape.equalsIgnoreCase("sech")) {
			int index = binarySearch(sechProfile, uniformRandomNumber);
			r0 = radius[index-1];
			r1 = radius[index];
			p0 = sechProfile[index-1];
			p1 = sechProfile[index];
			rho = (r0 + (r1-r0)*(uniformRandomNumber-p0)/(p1-p0))/reSech;
			hgSech.add(rho);
			rho *= scale;
		}
		
		return rho;		

	}

	/**
	 * get the index from the array of doubles
	 * @param profile the profile to which the index refers
	 * @param number to search for in the profile
	 * @return the index
	 */
	public static int binarySearch(double[] profile, double number) {
		int index = Arrays.binarySearch(profile, number);
		if (index < 0) index = -index-1;
		return index;
	}
	/**
	 * 
	 * plot the internal histograms
	 * @param baseFileName 
	 */
	public void kdePlot(String baseFileName) {
		KernelDensityEstimatePlotter.toPng(hgGauss, baseFileName+"Gauss.png", "gauss");
		KernelDensityEstimatePlotter.toPng(hgExp, baseFileName+"Exp.png", "exp");
		KernelDensityEstimatePlotter.toPng(hgDevauc, baseFileName+"Devauc.png", "deVauc");
		KernelDensityEstimatePlotter.toPng(hgSech, baseFileName+"Sech.png", "sech");
		ArrayList<KernelDensityEstimate> kdeList = new ArrayList<KernelDensityEstimate>();
		kdeList.add(hgGauss);
		kdeList.add(hgExp);
		kdeList.add(hgDevauc);
		kdeList.add(hgSech);
		ArrayList<String> legends = new ArrayList<String>();
		legends.add("Gaussian");
		legends.add("Exponential");
		legends.add("DeVaucoleurs");
		legends.add("Hyperbolic Secant");
		KernelDensityEstimatePlotter kdep = new KernelDensityEstimatePlotter(kdeList, null, null, legends);
		kdep.setXSize(500);
		kdep.setYSize(500);
		kdep.setXLabel("radius (scale length)");
		kdep.setYLabel("dNdr");
		kdep.toPng(baseFileName+"All.png");
	}
	/**
	 * get some informative statistics
	 * @return Hashtable of gaussMean, gaussRms, expMean, expRms, devaucMean, devaucRms, 
	 * 					sechMean, sechRMS, uniformMean, uniformRms
	 */
	public Hashtable<String,Double> getHgStats() {
		Hashtable<String,Double> retval = new Hashtable<String,Double>();
		retval.put("gaussMean", new Double(hgGauss.getMean()));
		retval.put("gaussRms", new Double(hgGauss.getSigma()));
		retval.put("expMean", new Double(hgExp.getMean()));
		retval.put("expRms", new Double(hgExp.getSigma()));
		retval.put("devaucMean", new Double(hgDevauc.getMean()));
		retval.put("devaucRms", new Double(hgDevauc.getSigma()));
		retval.put("sechMean", new Double(hgSech.getMean()));
		retval.put("sechRms", new Double(hgSech.getSigma()));
//		retval.put("uniformMean", new Double(iHistogram1D.mean()));
//		retval.put("uniformRms", new Double(iHistogram1D.rms()));
		return retval;
	}
}
