package gov.fnal.eag.sim.pixel;

import java.util.Collections;
import java.util.Hashtable;
import java.util.LinkedList;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

import cern.colt.list.DoubleArrayList;

/**
 * lookup an input value and convert to an index into a collection of colors
 * @author stoughto
 *
 */
public class ScaleTable {

	private LinkedList<Double> xList = new LinkedList<Double>();
	private String type;
	private int nLookup;
	private int minLookup;
	private double factor;
	
	/**
	 * default number of indeces for a color table for printing
	 */
	public static int N_LOOKUP_DEFAULT = 256;
	
	/**
	 * default lookup for the minimum value
	 */
	public static int MIN_LOOKUP_DEFAULT = 0;
	/**
	 * constructs a linear lookup table from xMin to xMax returning nLookup ints starting from minLookup
	 * @param xMin minimum input value to be mapped
	 * @param xMax maximum input value to be mapped
	 * @param nLookup number of indices to return
	 * @param minLookup index to return for input of xMin
	 */
	public ScaleTable(double xMin, double xMax, int nLookup, int minLookup) {
		set("linear", nLookup, minLookup);
		for (int i=minLookup; i< minLookup+nLookup; i++) {
			double x = xMin + ((i-minLookup)/(double) (nLookup-1))*(xMax-xMin);
			xList.add(x);
		}
	}
	
	/**
	 * construct a linear lookup table with defaults for nLookup and minLookup
	 * @param xMin
	 * @param xMax
	 */
	public ScaleTable(double xMin, double xMax) {
		this(xMin, xMax, N_LOOKUP_DEFAULT, MIN_LOOKUP_DEFAULT);
	}
	
	/**
	 * @param type one of sqrt, log, linear, histeq
	 * @param xMin minimum pixel value to map
	 * @param xMax maximum pixel value to map
	 * @param factor power/base in sqrt/log
	 * @param nLookup number values in the map, usually 256
	 * @param minLookup minimum number of the lookup indices, usually 0
	 */
	public ScaleTable(String type, double xMin, double xMax, double factor, int nLookup, int minLookup) {
		if (type.toUpperCase().startsWith("LOG") && factor==1) {
			throw new Error("can not use factor=1 for log scale");
		}
		set(type, nLookup, minLookup);
		this.factor = factor;
		for (int i=minLookup; i< minLookup+nLookup; i++) {
			double scale;
			double temp = (i-minLookup)/(double)nLookup;
			if (type.toUpperCase().startsWith("SQRT")) {
				scale = Math.pow(temp, factor);			
			} else if (type.toUpperCase().startsWith("LOG")) {
				scale = (Math.pow(factor, temp)-1)/(factor-1);				
			} else {
				throw new Error("unknown type="+type);
			}
			double x = xMin + (xMax-xMin)*scale;
			xList.add(x);
		}
	}
	
	/**
	 * create a ScaleTable with the default lookup index range
	 * @param type "sqrt" or "log"
	 * @param xMin minimum value to be indexed
	 * @param xMax maximum value to be indexed
	 * @param factor multiplicative factor for scaling shape of the sqrt or log
	 */
	public ScaleTable(String type, double xMin, double xMax, double factor)  {
		this(type, xMin, xMax, factor, N_LOOKUP_DEFAULT, MIN_LOOKUP_DEFAULT);
	}

	/**
	 * create a ScaleTable by histogramming values of the PixeledImage
	 * @param xMin minimum pixel value to scale
	 * @param xMax maximum pixel value to scale
	 * @param pi contains the the values to be histogrammed
	 * @param nLookup number of lookup values, usually 256
	 * @param minLookup minimum lookup value, usually 0
	 */
	public ScaleTable(double xMin, double xMax, PixeledImage pi, int nLookup, int minLookup) {
		histInit(xMin, xMax, pi, nLookup, minLookup);
	}

	private void set(String type, int nLookup, int minLookup) {
		this.type = type;
		this.nLookup = nLookup;
		this.minLookup = minLookup;		
	}

	/**
	 * creates a ScaleTable by histogramming all values of the input PixeledImage
	 * @param pi contains the values to be histogrammed
	 */
	public ScaleTable(PixeledImage pi) {
		Hashtable<String,Double> extrema = pi.extrema();
		double xMin = extrema.get("min");
		double xMax = extrema.get("max");
		histInit(xMin, xMax, pi, N_LOOKUP_DEFAULT, MIN_LOOKUP_DEFAULT);
		
	}
	
	/**
	 * creates a histeq ScaleTable  
	 * @param pi contains the values to be histogrammed
	 * @param xMin minimum pixel value
	 * @param xMax maximum pixel value
	 */
	public ScaleTable(PixeledImage pi, double xMin, double xMax) {
		histInit(xMin, xMax, pi, N_LOOKUP_DEFAULT, MIN_LOOKUP_DEFAULT);
	}
	
	private void histInit(double xMin, double xMax, PixeledImage pi, int nLookupIn, int minLookupIn) {
		set("histeq", nLookupIn, minLookupIn);
		DoubleArrayList dal = pi.getDoubleArrayList();
		dal.sort();
		int iMin = binarySearch(dal, xMin);
		int iMax = binarySearch(dal, xMax);
		//int nBin = iMax-iMin+1;
		set("histeq", nLookup, minLookup);
		//System.out.println("iMin="+iMin+" iMax="+iMax);
		for (int i=minLookup; i< minLookup+nLookup; i++) {
			//double x = xMin + ((i-minLookup)/(double) nLookup)*(xMax-xMin);
			//int j = i*nBin/nLookup;
			int j = (int) (iMin + ((double) i/(double) nLookup)*(iMax-iMin));
			double x = dal.get(j);
			//System.out.println(" i="+i+" j="+j+"  x="+x);
			xList.add(x);
		}	
	}
	
	/**
	 * lookup a value
	 * @param dal sorted list of all values
	 * @param x the value to look u
	 * @return the index
	 */
	public static int binarySearch(DoubleArrayList dal, double x) {
		int i = dal.binarySearch(x);
		if (i <= 0) i = -i-2;
		if (i < 0) i = 0;
		if (i >= dal.size()) i = dal.size()-1;
		return i;
	}
	
	/**
	 * lookup the index for the input value
	 * @param x the input value
	 * @return the index
	 */
	public int lookup(double x) {
		int lookup;
		lookup = Collections.binarySearch(xList, x);
		// account for the index convention of the binary search method
		if (lookup <= 0) lookup = -lookup-2;		
		// always return legal values
		if (lookup < 0) lookup = 0;
		if (lookup >= nLookup) lookup = nLookup-1;
		return lookup+minLookup;
	}
	
	/**
	 * get the type of this table
	 * @return the type
	 */
	public String getType() {
		return type;
	}

	/**
	 * get the xValues as an array
	 * @return the array
	 */
	public double[]  getXValues() {
		double[] xValues = new double[xList.size()];
		int i=0;
		for (double x : xList) {
			xValues[i++] = x;
		}
		return xValues;
	}
	/**
	 * get the starting index
	 * @return the index
	 */
	public int getMinLookup() {
		return minLookup;
	}
	
	/**
	 * make a chart of the list of scale tables
	 * @param scaleTables the list of ScaleTables to plot
	 * @return the chart
	 */
	static public JFreeChart makeChart(LinkedList<ScaleTable> scaleTables){
		XYSeriesCollection xySeriesCollection =  new XYSeriesCollection();
		
		for (ScaleTable scaleTable: scaleTables) {
			XYSeries xySeries = new XYSeries(scaleTable.getDescription());
			double[] xValues = scaleTable.getXValues();
			for (int i=0; i<xValues.length; i++) {
				xySeries.add( xValues[i], (i+scaleTable.getMinLookup()));
			}			
			xySeriesCollection.addSeries(xySeries);
		}
		JFreeChart chart = ChartFactory.createScatterPlot("ScaleTable","value","iColor", xySeriesCollection, PlotOrientation.VERTICAL, true, true, true);
		return chart;
	}

	/**
	 * get a simple description, suitable for plot legends
	 * @return a string containg the type and, if relevant, the factor
	 */
	public String getDescription() {
		if (type.equalsIgnoreCase("linear")) return "linear";
		if (type.equalsIgnoreCase("histeq")) return "histeq";
		return type+" "+factor;
	}

	/**
	 * @return minimum x value mapped
	 */
	public double getXMin() {
		return xList.get(0);
	}
	/**
	 * @return maximum x value mapped
	 */
	public double getXMax() {
		return xList.get(xList.size()-1);
	}
	
	/**
	 * @return the number of entries in the lookup table
	 */
	public int getNLookup() {
		return nLookup;
	}
}
