package gov.fnal.eag.sim.pixel;

import gov.fnal.eag.sim.random.SimRandomNormal;
import gov.fnal.eag.sim.random.SimRandomPoisson;
import gov.fnal.eag.sim.random.SimRandomUniform;
import gov.fnal.eag.sim.repository.image.FitsDataFormat;
import gov.fnal.eag.sim.repository.image.ImageRepository;
import gov.fnal.eag.util.misc.StringUtil;

import java.util.LinkedList;

import nom.tam.fits.HeaderCardException;

import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;

/**
 * simulates a single exposure of a CCD to the K-alpha peak of an Fe55 source
 * @author stoughto
 *
 */
public class SimulateXRays {

	ImageRepository irOut;
	String fnOut;
	int nRow, nCol, eventSeedIndex, noiseSeedIndex, nGridBright;
	double meanEventsPerPixel, meanElectronsPerEvent, sigmaElectronsPerEvent, sigmaReadNoise, sigmaChargeDiffusion, sky, gain, poissonFactor, countsPerPixelThreshold;
	boolean doDigitization;
	Profiles profiles = new Profiles();

	PixeledImageFactory pifXRay;
	
	/**
	 * constructor defined as command-line arguments
	 * @param args
	 * @throws JSAPException
	 */
	public SimulateXRays(String[] args) throws JSAPException {
		parse(args);
	}

	/**
	 * do the actual work here
	 * @throws HeaderCardException
	 */
	public void run() throws HeaderCardException {
		PixeledImage xRay = makeXRayImage();
		PixeledImage pi = makeBackgroundImage();
		pi.add(xRay);
		
		pi.header.addValue("gain", gain, "electrons/adu");
		pi.header.addValue("mepp", meanEventsPerPixel, "meanEventsPerPixel");
		pi.header.addValue("mepe", meanElectronsPerEvent, "meanElectronsPerEvent");
		pi.header.addValue("sepe", sigmaElectronsPerEvent, "sigmaElectronsPerEvent");
		pi.header.addValue("srn", sigmaReadNoise, "sigmaReadNoise (electrons)");
		pi.header.addValue("scd", sigmaChargeDiffusion, "sigmaChargeDiffusion (pixel)");
		pi.header.addValue("sky", sky, "sky (electrons)");
		pi.header.addValue("esi", eventSeedIndex, "eventSeedIndex");
		pi.header.addValue("nsi", noiseSeedIndex , "noiseSeedIndex");
		pi.header.addValue("ngb", nGridBright , "nGridBright srt(number of samples/pixel)");
		pi.header.addValue("pf", poissonFactor , "0.0 (none) to 1.0 (complete) poisson noise");
		pi.header.addValue("cppt", countsPerPixelThreshold , "fill pixels with counts/pixel > cppt ");
		
		LinkedList<Object> results =  new LinkedList<Object>();
		results.add(pi);
		results.addAll(pifXRay.getDetectorObjectTables());
		FitsDataFormat outputFormat;
		if (doDigitization) {
			PixelValueTransformation digitize = new PixelValueTransformationDigitize(16, "digitize");
			pi.transform(digitize);
			outputFormat = FitsDataFormat.USHORT;
		} else {
			outputFormat = FitsDataFormat.DOUBLE;
		}
		irOut.put(fnOut, results, outputFormat, true);
	}

	private PixeledImage makeXRayImage() {
		SimRandomPoisson simRandomPoisson = new SimRandomPoisson(noiseSeedIndex+1);
		SimRandomUniform simRandomUniform = new SimRandomUniform(noiseSeedIndex+2);
		SimRandomNormal simRandomNormal = new SimRandomNormal(noiseSeedIndex+3);
		PixelDetector pd = new PixelDetector(nRow, nCol);
		PixelValueTransformation inverseGain = new PixelValueTransformationMultiplyConstant(1.0/gain, "gain");
		pd.putElectronics(inverseGain);
		PSFFactory psfFactory = null;
		LinkedList<DetectorObject> detectorObjects = new LinkedList<DetectorObject>();
		{
			SimRandomPoisson srp = new SimRandomPoisson(eventSeedIndex+1);
			int nEvents = srp.next(meanEventsPerPixel*nRow*nCol);
			SimRandomUniform sru = new SimRandomUniform(eventSeedIndex+2);
			SimRandomNormal srn = new SimRandomNormal(eventSeedIndex+3, meanElectronsPerEvent, sigmaElectronsPerEvent, "", false);
			for (int i=0; i<nEvents; i++) {
				double row = (int) (sru.next(0,nRow)) + 0.5;
				double col = (int) (sru.next(0,nCol)) + 0.5;
				//double row = sru.next(0,nRow);
				//double col = sru.next(0,nCol);
				double nPhoton = srn.next();
				//nPhoton = srn.getMean();
				//System.out.println("i="+i+" nPhoton="+nPhoton);
				DetectorObjectProfile dop = new DetectorObjectProfile(row, col, nPhoton, "gauss", sigmaChargeDiffusion, 0.0, 0.0, profiles);
				detectorObjects.add(dop);
			}
		}				
		double cosmicRayDensity = 0.0;
		BackgroundLight backgroundLight = new BackgroundLightConstant(0.0);
		pifXRay = new PixeledImageFactory(pd, backgroundLight, psfFactory, detectorObjects, cosmicRayDensity, simRandomPoisson, simRandomUniform, simRandomNormal);
		//System.out.println("nGridBright="+pif.getNGridBright());
		pifXRay.setNGridBright(nGridBright);
		pifXRay.setPoissonFactor(0.0);
		pifXRay.setCountsPerPixelThreshold(countsPerPixelThreshold);		
		PixeledImage pi = pifXRay.planAndProduce();
		return pi;
	}
	
	private PixeledImage makeBackgroundImage() {
		SimRandomPoisson simRandomPoisson = new SimRandomPoisson(noiseSeedIndex+1);
		SimRandomUniform simRandomUniform = new SimRandomUniform(noiseSeedIndex+2);
		SimRandomNormal simRandomNormal = new SimRandomNormal(noiseSeedIndex+3);
		PixelDetector pd = new PixelDetector(nRow, nCol);
		PixelValueTransformation readNoise = new PixelValueTransformationConstantGaussianNoise(sigmaReadNoise, "read noise", simRandomNormal);
		pd.putElectronics(readNoise);
		PixelValueTransformation inverseGain = new PixelValueTransformationMultiplyConstant(1.0/gain, "gain");
		pd.putElectronics(inverseGain);
		BackgroundLight backgroundLight = new BackgroundLightConstant(sky);
		PSFFactory psfFactory = null;
		LinkedList<DetectorObject> detectorObjects = new LinkedList<DetectorObject>();
		double cosmicRayDensity = 0.0;
		PixeledImageFactory pif = new PixeledImageFactory(pd, backgroundLight, psfFactory, detectorObjects, cosmicRayDensity, simRandomPoisson, simRandomUniform, simRandomNormal);
		//System.out.println("nGridBright="+pif.getNGridBright());
		pif.setPoissonFactor(poissonFactor);
		pif.setCountsPerPixelThreshold(countsPerPixelThreshold);	
		PixeledImage pi = pif.planAndProduce();
		return pi;
	}
	private void parse(String[] args) throws JSAPException {
		JSAP jsap = new JSAP();
		StringUtil.register(jsap, "irOut", JSAP.STRING_PARSER);
		StringUtil.register(jsap, "fnOut", JSAP.STRING_PARSER);
		StringUtil.register(jsap, "nRow", JSAP.INTEGER_PARSER);
		StringUtil.register(jsap, "nCol", JSAP.INTEGER_PARSER);
		StringUtil.register(jsap, "gain", JSAP.DOUBLE_PARSER);
		StringUtil.register(jsap, "meanEventsPerPixel", JSAP.DOUBLE_PARSER);
		StringUtil.register(jsap, "meanElectronsPerEvent", JSAP.DOUBLE_PARSER);
		StringUtil.register(jsap, "sigmaElectronsPerEvent", JSAP.DOUBLE_PARSER);
		StringUtil.register(jsap, "sigmaReadNoise", JSAP.DOUBLE_PARSER);
		StringUtil.register(jsap, "sigmaChargeDiffusion", JSAP.DOUBLE_PARSER);
		StringUtil.register(jsap, "sky", JSAP.DOUBLE_PARSER);
		StringUtil.register(jsap, "eventSeedIndex", JSAP.INTEGER_PARSER);
		StringUtil.register(jsap, "noiseSeedIndex", JSAP.INTEGER_PARSER);
		StringUtil.register(jsap, "nGridBright", JSAP.INTEGER_PARSER);
		StringUtil.register(jsap, "poissonFactor", JSAP.DOUBLE_PARSER);
		StringUtil.register(jsap, "countsPerPixelThreshold", JSAP.DOUBLE_PARSER);
		
		JSAPResult result = jsap.parse(args);
		irOut = new ImageRepository(result.getString("irOut"));
		fnOut = result.getString("fnOut");
		nRow = result.getInt("nRow");
		nCol = result.getInt("nCol");
		gain = result.getDouble("gain");
		meanEventsPerPixel = result.getDouble("meanEventsPerPixel");
		meanElectronsPerEvent = result.getDouble("meanElectronsPerEvent");
		sigmaElectronsPerEvent = result.getDouble("sigmaElectronsPerEvent");
		sigmaReadNoise = result.getDouble("sigmaReadNoise");
		sigmaChargeDiffusion = result.getDouble("sigmaChargeDiffusion");
		sky = result.getDouble("sky");
		nGridBright = result.getInt("nGridBright");
		poissonFactor = result.getDouble("poissonFactor");
		countsPerPixelThreshold = result.getDouble("countsPerPixelThreshold");
	}

	/**
	 * @param args
	 * @throws JSAPException 
	 * @throws HeaderCardException 
	 */
	public static void main(String[] args) throws JSAPException, HeaderCardException {
		SimulateXRays sxr = new SimulateXRays(args);
		sxr.run();
	}

}
