package gov.fnal.eag.sim.random;

import gov.fnal.eag.wcs.SphericalCoordinate;

/**
 * generates points randomly and uniformly distributed on a sphere with a cone
 * 
 * iSeedIndex specifies 
 * @author stoughto
 *
 */
public class RandomCone {
	private SphericalCoordinate center;
	private double openingAngleDegrees;
	private SimRandomUniform uniform1, uniform2;
	//private LinkedList<SphericalCoordinate> points;
	//private boolean keepPoints;
	
	/**
	 * @param iSeedIndex
	 * @param center
	 * @param openingAngleDegrees
	 */
	public RandomCone(int iSeedIndex, SphericalCoordinate center, double openingAngleDegrees) {
		if (openingAngleDegrees <= 0.0) {
			throw new IllegalArgumentException("opening angle must be > 0 but openingAngleDegrees="+openingAngleDegrees);
		}
		uniform1 = new SimRandomUniform();
		uniform2 = new SimRandomUniform();
		initSeeds(iSeedIndex);
		this.center = center;
		this.openingAngleDegrees = openingAngleDegrees;
	}
	
	private void initSeeds(int iSeedIndex) {
		uniform1.setIseedIndex(iSeedIndex);
		uniform2.setIseedIndex(iSeedIndex+5);
	}
	/**
	 * default constructor with iSeedIndex=1 and opening angle of one degree at ra,dec=0,0
	 */
	public RandomCone() {
		this(1, new SphericalCoordinate(0,0), 1);
	}
	
	
	/**
	 * Get the next (azimuth,altitude) point (in degrees) that is uniformly distributed on a sphere.  
	 * See http://mathworld.wolfram.com/SpherePointPicking.html
	 * @return the point
	 */
	public SphericalCoordinate next() {
		double separation;
		SphericalCoordinate point;
		boolean isGood;
		while (true) {
			double u = uniform1.next();
			double v = uniform2.next();
			double azimuth = u * 360.0;
			double phi = Math.acos(2*v-1)*180.0/Math.PI; // Phi ranges from 0 to 180 degrees
			double altitude = 90-phi; // Altitude ranges from -90 to 90 degrees
			point = new SphericalCoordinate(azimuth, altitude);
			if (center == null) {
				separation = 0.0;
			} else {
				separation = center.separation(point);
			}
			isGood = (separation < openingAngleDegrees);
			//System.out.println("ra="+azimuth+" dec="+altitude+" center="+center);
			//System.out.println("isGood="+isGood+" separation="+separation+" openingAngleDegrees="+openingAngleDegrees);
			if (isGood) {
				break;
			}
		}	
		return point;
	}
	
	/**
	 * set the random number seed index
	 * @param seedIndex
	 */
	public void setIseedIndex(int seedIndex) {
		initSeeds(seedIndex);		
	}
	
	/**
	 * @return the seed that was used to create this generator
	 */
	public int getSeed() {
		return uniform1.getIseedIndex();
	}

}
