package gov.fnal.eag.sim.random;

import gov.fnal.eag.util.plot.XYErrorBarRenderer;
import gov.fnal.eag.util.plot.XYErrorBarSeries;
import gov.fnal.eag.util.plot.XYErrorBarSeriesCollection;
import gov.fnal.eag.wcs.SphericalCoordinate;

import java.io.File;
import java.util.Iterator;
import java.util.LinkedList;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.data.statistics.HistogramDataset;
import org.jfree.data.statistics.HistogramType;
import org.jfree.data.xy.XYDataset;

/**
 * generates points randomly and uniformly distributed on a sphere
 * @author stoughto
 *
 */
public class RandomSphere {
	private double minAlt, maxAlt, minAz, maxAz;
	private SimRandomUniform uniform1, uniform2;
	private LinkedList<SphericalCoordinate> points;
	private boolean keepPoints;
	
	/**
	 * constructor, with the seed index specified
	 * @param iSeedIndex random number seed index
	 */
	public RandomSphere(int iSeedIndex) {
		init(iSeedIndex);
	}
	
	/**
	 * constructor, with the seed index and coordinate range specified
	 * @param iSeedIndex random number seed index
	 * @param minAlt minimum altitude
	 * @param maxAlt maximum altitude
	 * @param minAz  minimum azimuth
	 * @param maxAz maximum azimuth
	 */
	public RandomSphere(int iSeedIndex, double minAlt, double maxAlt,
			double minAz, double maxAz) {
		init(iSeedIndex, minAlt, maxAlt, minAz, maxAz);
	}

	private void init(int iSeedIndex) {
		init(iSeedIndex,-90.0,90.0,0.0,360.0);
	}
	
	private void init(int iSeedIndex, double minAltIn, double maxAltIn,
			double minAzIn, double maxAzIn) {
		minAlt = minAltIn;
		maxAlt = maxAltIn;
		minAz  = minAzIn;
		maxAz  = maxAzIn;
		double minAzFraction = minAz / 360.0;
		double maxAzFraction = maxAz / 360.0;
		double minAltFraction = 0.5*(Math.cos(Math.PI*(0.5 - minAlt/180.0))+1.0);
		double maxAltFraction = 0.5*(Math.cos(Math.PI*(0.5 - maxAlt/180.0))+1.0);
		if (uniform1 == null) {
			uniform1 = new SimRandomUniform(iSeedIndex, minAzFraction,
					maxAzFraction);
		} else {
			uniform1.setIseedIndex(iSeedIndex);
		}
		if (uniform2 == null) {
			uniform2 = new SimRandomUniform(iSeedIndex + 5, minAltFraction, maxAltFraction);
		} else {
			uniform2.setIseedIndex(iSeedIndex + 5);
		}
	}
	
	
	/**
	 * default constructor with iSeedIndex=1
	 */
	public RandomSphere() {
		this(1);
	}
	
	/**
	 * @param keepPointsIn set true to keep a LinkedList of all points generated
	 */
	public void keepPoints(boolean keepPointsIn) {
		keepPoints = keepPointsIn;
		if (keepPoints) {
			points = new LinkedList<SphericalCoordinate>();			
		} else {
			points = null;
		}
	}
	
	/**
	 * Get the next (azimuth,altitude) point (in degrees) that is uniformly distributed on a sphere.  
	 * See http://mathworld.wolfram.com/SpherePointPicking.html
	 * @return the point
	 */
	public SphericalCoordinate next() {
		double u = uniform1.next();
		double v = uniform2.next();
		double azimuth = u * 360.0;
		double phi = Math.acos(2*v-1)*180.0/Math.PI; // Phi ranges from 0 to 180 degrees
		double altitude = 90-phi; // Altitude ranges from -90 to 90 degrees
		SphericalCoordinate point = new SphericalCoordinate(azimuth, altitude);
		if (keepPoints) {
			points.add(point);
		}
		return point;
	}
	
	/**
	 * histogram the saved points and write plots to baseName_*.png
	 * @param baseName base name for the plot files
	 * @throws Exception 
	 */
	public void plotHistograms(String baseName) throws Exception  {
		if (keepPoints) {
			double[] altitude = new double[points.size()];
			double[] azimuth = new double[points.size()];
			double[] altitudeByCos2 = new double[points.size()];
			int i=0;
			for (Iterator<SphericalCoordinate> it = points.iterator(); it.hasNext(); ) {
				SphericalCoordinate point = it.next();
				azimuth[i] = point.getLongitude();
				altitude[i] = point.getLatitude();
				double cos2 = Math.pow(Math.cos(altitude[i]*Math.PI/180.0), 2);			
				altitudeByCos2[i] = altitude[i]/cos2;
				i++;
			}
			writeHistogram(azimuth,baseName+"_Azimuth", 90, 0, 360);
			writeHistogram(altitude,baseName+"_Altitude", 90, -90, 90);
			writeWeightedHistogram(altitude, baseName+"_WeightedAltitude", 90, -90, 90);
		}
	}
	
	private void writeHistogram(double[] values, String name, int n, double min, double max) throws Exception {
		final HistogramDataset dataset = new HistogramDataset();
        dataset.setType(HistogramType.FREQUENCY);
        dataset.addSeries("H1", values, n, min, max);
        JFreeChart chart = ChartFactory.createHistogram(
                name, 
                null,
                null, 
                dataset, 
                PlotOrientation.VERTICAL, 
                false, 
                false, 
                false
            );
        ChartUtilities.saveChartAsPNG(new File(name+".png"), chart, 500, 300);
	}
	
	private void writeWeightedHistogram(double[] values, String name, int n, double min, double max) throws Exception {
		final HistogramDataset dataset = new HistogramDataset();
        dataset.setType(HistogramType.FREQUENCY);
        dataset.addSeries("values", values, n, min, max);
        
		XYErrorBarSeriesCollection xyebsc = new XYErrorBarSeriesCollection();
		XYErrorBarSeries xyebs = new XYErrorBarSeries("Series 1");

        for (int i=0; i<n; i++) {
        	double theta = dataset.getX(0, i).doubleValue();
        	//double cos = Math.cos(theta*Math.PI/180.0);
        	double theta0 = dataset.getStartX(0, i).doubleValue();
        	double theta1 = dataset.getEndX(0, i).doubleValue();
        	double area = 2*Math.PI*(Math.sin(theta1*Math.PI/180)-Math.sin(theta0*Math.PI/180));
        	double value = dataset.getY(0, i).doubleValue();
        	double y = value/area;
        	double sigy = Math.sqrt(value)/area;
        	xyebs.add(theta, y, y-sigy, y+sigy);
        }
        xyebsc.addSeries(xyebs);
		JFreeChart chart = createChart(xyebsc);
		ChartUtilities.saveChartAsPNG(new File(name+".png"), chart, 500, 300);				

	}
	private JFreeChart createChart(XYDataset dataset) {
		JFreeChart chart = ChartFactory.createScatterPlot(
				"Density of Points",
		// title
				"theta (degrees)",
		// x-axis label
				"density",
		// y-axis label
				dataset,
		// data
				PlotOrientation.VERTICAL,
		// create legend?
				false,
		// generate tooltips?
				false, false
		// generate URLs?
				);
		XYErrorBarRenderer r = new XYErrorBarRenderer();
		chart.getXYPlot().setRenderer(r);
		r.addChangeListener(chart.getXYPlot());// see bug 1177884

		return chart;
	}
	
	/**
	 * set the random number seed index
	 * @param seedIndex
	 */
	public void setIseedIndex(int seedIndex) {
		init(seedIndex);		
	}
	
	/**
	 * @return the point at the minimum of the rectangular area covered
	 */
	public SphericalCoordinate getMin() {
		return new SphericalCoordinate(minAz, minAlt);		
	}
	/**
	 * @return the point at the maximum of the rectangular area covered
	 */
	public SphericalCoordinate getMax() {
		return new SphericalCoordinate(maxAz, maxAlt);		
	}

	/**
	 * @return the seed that was used to create this generator
	 */
	public int getSeed() {
		return uniform1.getIseedIndex();
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		String retval = "RandomSphere:";
		retval += " minAz="+minAz;
		retval += " maxAz="+maxAz;
		retval += " minAlt="+minAlt;
		retval += " maxAlt="+maxAlt;
		return retval;
	}
}
