package gov.fnal.eag.sim.random;

import java.io.File;

import gov.fnal.eag.util.densityEstimate.Histogram;
import gov.fnal.eag.util.densityEstimate.KernelDensityEstimate;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

import cern.jet.random.engine.RandomSeedGenerator;

/**
 * manages random number generation.
 * 
 * The random number generators that implement this interface are deterministic.  
 * 
 * iSeedIndex points into an array of a "good" random number seeds.   If you use the setIseedIndex method to set the seed, it repeats the same series of random numbers for
 * a given iSeedIndex.  
 * 
 * A histogram of generated values is stored and can be accessed with the getIHistogram1d() method, which returns
 * a IHistogram1D object. 
 *
 */
public abstract class SimRandom {
	protected Histogram histogram = null;

	/**
	 * provides an array of random seeds
	 */
	protected transient RandomSeedGenerator rsg;
	/**
	 * index into the array of random seeds
	 */
	protected transient int iSeedIndex;
	/**
	 * counts the number of random numbers generated
	 */
	protected transient int iGenerated;
	/**
	 * keeps a histogram of generated random numbers
	 */
	//protected transient IHistogram1D iHistogram1D = null;

	protected transient KernelDensityEstimate kernelDensityEstimate = null;
	
	/**
	 * true to keep hisograms of generated points
	 */
	protected transient boolean doHistogram = false;
	
	/**
	 * get the index for seed generation
	 * @return the seed index
	 */
	public int getIseedIndex() {
		return iSeedIndex;
	}

	/**
	 * get the count of how many numbers have been generated from this seed index
	 * @return -- the count
	 * 
	 */
	public int getIgenerated() {
		return iGenerated;
	}

	/**
	 * write a plot to the fileName
	 * @param fileName file name to save .PNG to
	 * @param sizeX width of graph
	 * @param sizeY height of graph
	 * @throws Exception 
	 */
	public void toPNG(String fileName, int sizeX, int sizeY) throws Exception {
		JFreeChart chart = getChart();
		ChartUtilities.saveChartAsPNG(new File(fileName+".png"), chart, sizeX, sizeY);
	}

	/**
	 * save a chart of true and thrown as a .PNG
	 * @return a chart with the true and thrown distributions
	 * @throws Exception 
	 */
	public JFreeChart getChart() throws Exception {	
		XYSeriesCollection c = new XYSeriesCollection();
		c.addSeries(getTrue());
		XYSeries thrown = getThrown();
		if (thrown != null) {
			c.addSeries(thrown);
		}
		JFreeChart chart = ChartFactory.createScatterPlot(getClass().toString(), "x", "dn", c, PlotOrientation.VERTICAL, true, true, true);
		for (int i=0; i<chart.getXYPlot().getDatasetCount(); i++) {
			XYLineAndShapeRenderer xyItemRenderer = (XYLineAndShapeRenderer) chart.getXYPlot().getRenderer(i);
			xyItemRenderer.setBaseLinesVisible(true);
			xyItemRenderer.setBaseShapesVisible(false);
		}
        return chart;
	}

	
	/**
	 * create generic chart of true and thrown
	 * @return the true points
	 */
	public abstract XYSeries getTrue();	
	
	/**
	 * get points suitable for plotting
	 * @return the next random number from this generator
	 */
	abstract public double next();

	/**
	 * 
	 * @param doHistogram whether to histogram values
	 */
	public void setDoHistogram(boolean doHistogram) {
		this.doHistogram = doHistogram;
		makeHistogram();
	}
	
	protected abstract void makeHistogram();

	/**
	 * @return whether values are being histogrammed
	 */
	public boolean getDoHistogram() {
		return doHistogram;
	}
	
	protected XYSeries getThrown() {
		XYSeries retval = null;
		if (histogram != null) {
			retval = histogram.getXySeries(false, 0.0);
		}
		return retval;
	}
	
	/**
	 * set the seed index
	 * @param iSeedIndex
	 */
	public abstract void setIseedIndex(int iSeedIndex);
}
