package gov.fnal.eag.sim.random;

import gov.fnal.eag.util.astronomy.Integrate;
import gov.fnal.eag.util.astronomy.SimpleFunction;
import gov.fnal.eag.util.densityEstimate.Histogram;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.annotations.XYAnnotation;
import org.jfree.chart.annotations.XYLineAnnotation;
import org.jfree.data.xy.XYSeries;

import cern.colt.list.DoubleArrayList;

/**
 * generates random numbers according to a distribution.
 * @author stoughto
 * 
 * This class selects random numbers with the distribution defined by the SimpleFunction.  The code 
 * in SimRandomDistributionTest.testSimRandomDistributionSquared() is a trivial example of generating
 * random numbers with the distribution y = x*x, for x between 5 and 10.  
 * 
 * In the constructor, we step through the range of x values from minValue to maxValue, with step size deltaValue.
 * We create a lookup table of the x values and the of the integral of the y value up to that x value.  The integral
 * between minValue and maxValue (stored in the private variable sum) is the last integral in the lookup table.  
 * To generate a random number, we select a number from a random distribution between 0 and sum.  We do a binary search
 * on the lookup table based on the sum, and return the corresponding x value.
 * 
 * For normalization, the method integral() returns the integral of the distribution between minValue and maxValue.
 * 
 *
 */
public class SimRandomDistribution extends SimRandom {

	/**
	 * the distribution
	 */
	protected SimpleFunction distribution;
	/**
	 * minimum value generated
	 */
	protected double minValue;
	/**
	 * maximum value generated
	 */
	protected double maxValue;
	/**
	 * bin for smoothing generating values
	 */
	protected double deltaValue;
	
	/**
	 * number of bins in lookup table
	 */
	protected int nBin;
	protected SimRandomUniform simRandomUniform;
	private double sum;
	private DoubleArrayList dal;
	private double minValueThrown, maxValueThrown;
	
	private double uniform = -1;
	
	/**
	 * used by subclasses
	 */
	protected SimRandomDistribution() { // used by subclasses
		simRandomUniform = new SimRandomUniform();
	}
	/**
	 * prepare to generate random numbers
	 * @param distribution with this distribution
	 * @param minValue min value
	 * @param maxValue max value
	 * @param nBin number of bins in the lookup table
	 * @param iSeedIndex a seed
	 */
	public SimRandomDistribution(SimpleFunction distribution, double minValue, double maxValue, int nBin, int iSeedIndex) {
		this();
		this.minValue = minValue;
		this.maxValue = maxValue;
		this.deltaValue = (maxValue-minValue)/(nBin);
		this.nBin = nBin;
		setDistribution(distribution);
		setIseedIndex(iSeedIndex);
	}

	/**
	 * reset set the random seed index and clear the history
	 * @param iSeedIndex
	 */
	@Override
	public void setIseedIndex(int iSeedIndex) {
		this.iSeedIndex = iSeedIndex;
		simRandomUniform.setIseedIndex(iSeedIndex);
	}

	protected void setDistribution (SimpleFunction distribution) {
		this.distribution = distribution;
		dal = new DoubleArrayList();
		sum = 0;
		for (int iBin=0; iBin<nBin; iBin++ ) {
			double value = minValue + iBin*deltaValue;
			sum+=distribution.f(value);
			dal.add(sum);
		}
		minValueThrown = maxValue;
		maxValueThrown = minValue;		
		
	}
	/**
	 * @see gov.fnal.eag.sim.random.SimRandom#getIseedIndex()
	 */
	@Override
	public int getIseedIndex() {
		return iSeedIndex;
	}
	
	/**
	 * @see gov.fnal.eag.sim.random.SimRandom#next()
	 */
	@Override
	public double next() {
		return next(true);
	}
	/**
	 * @param returnValue if false, only call the random number next methods
	 * @return the next random number, or Double.NaN if returnValue is false
	 */
	public double next(boolean returnValue) {
		double retval;
		if (returnValue) {
			double val0 = 0;
			double val1 = dal.get(dal.size()-1);
			uniform = simRandomUniform.next(val0, val1);
			double uniformBin = simRandomUniform.next(0,1);
			int i = getI(uniform);
			retval = getRetval(uniformBin, i);
		} else {
			simRandomUniform.next();
			simRandomUniform.next();
			retval = Double.NaN;
		}
		return retval;
	}
	/**
	 * @param uniformIn uniformly-distributed value from min to max
	 * @return index into the dal of values
	 */
	public int getI(double uniformIn) {
		int iBinary = dal.binarySearch(uniformIn);
		int i = iBinary;
		if (iBinary<0) {
			i = -i-1;
		}
		return i;
	}
	
	/**
	 * @param uniformBin position within the bin, from 0 to 1
	 * @param i index into the data array list of values
	 * @return the random value
	 * @throws IllegalArgumentException if something went wrong and the generated value is > max value or minimum value < min value
	 */
	public double getRetval(double uniformBin, int i) throws IllegalArgumentException {
		double retval = minValue+(i+uniformBin)*deltaValue;
		minValueThrown = Math.min(minValueThrown, retval);
		maxValueThrown = Math.max(maxValueThrown, retval);
		if (retval > maxValue) {
			throw new IllegalArgumentException("  too big:  retval="+retval+" maxValue="+maxValue+" uniformBin="+uniformBin+" i="+i+" dal.size()="+dal.size());
		}
		if (retval < minValue) {
			throw new IllegalArgumentException("  too small:  retval="+retval+" maxValue="+maxValue+" uniformBin="+uniformBin+" i="+i+" dal.size()="+dal.size());
		}
		if (doHistogram) {
			histogram.fill(retval);
		}
		iGenerated++;
		return retval;
	}

	/**
	 * @return the integral of the distribution between min and max
	 */
	public double integral() {
		return Integrate.qRomb(distribution, minValue, maxValue, 1e-9);
	}
	/**
	 * @return the minimum of all values thrown or maxValue if next() has not been called
	 */
	public double getMinValueThrown() {
		return minValueThrown;
	}
	/**
	 * @return the maximum of all values thrown or minValue if next() has not been called
	 */
	public double getMaxValueThrown() {
		return maxValueThrown;
	}
	/**
	 * @return the distribution that defines this generator
	 */
	public SimpleFunction getDistribution() {
		return distribution;
	}

	/**
	 * write a plot to the fileName
	 * @param fileName file name to save .PNG to
	 * @param sizeX width of graph
	 * @param sizeY height of graph
	 * @throws Exception 
	 */
	@Override
	public void toPNG(String fileName, int sizeX, int sizeY) throws Exception {
		JFreeChart chart = getChart();
		ChartUtilities.saveChartAsPNG(new File(fileName+".png"), chart, sizeX, sizeY);
	}


	/**
	 * @return the true points for plotting
	 */ 
	@Override
	public XYSeries getTrue() {
		XYSeries xyDistribution = new XYSeries("true");
		double binSize = deltaValue;
		if (histogram != null) {
			binSize = histogram.getBinSize();
		}
		int nx = (int) ((maxValue-minValue)/binSize);
		double density = Integrate.qRomb(distribution, minValue, maxValue, 1e-3);
		for (int ix=0; ix<nx+1; ix++) {
			double x = minValue+ix*binSize;
			double value = iGenerated*distribution.f(x)*binSize/density;
			xyDistribution.add(x, value);
		}
		return xyDistribution;
	}
	/**
	 * @return a list of XYLineAnnotations of the true distribution
	 */
	public List<XYAnnotation> getTrueAnnotations() {
		ArrayList<XYAnnotation> retval = new ArrayList<XYAnnotation>();
		for (int i=0; i<dal.size()-1; i++) {
			double val0 = dal.get(i);
			double val1 = dal.get(i+1);
			double retval0 = minValue+(i+0.5)*deltaValue;
			double retval1 = minValue+(i+1.5)*deltaValue;
			retval.add(new XYLineAnnotation(val0, retval0, val1, retval1));
		}
		return retval;
	}

	@Override
	protected void makeHistogram() {
		if (histogram == null) {
			histogram = new Histogram("SimRandomDistribution", "value", "dN/dValue", 100, minValue, maxValue);
		}
	}

	/**
	 * @return the uniform value from the previous invocation of next()
	 */
	public double getUniform() {
		return uniform;
	}
	/**
	 * @return points suitable for plotting.
	 */
	/*
	public XYErrorBarSeries getThrown()  {
		int n = 100000;
		double value[] = new double[n];
		
		for (int i=0; i<n; i++) {
			value[i] = next();
		}
		
		int samples = nBin;
		SimpleFunction kernel = new UniformKernel(deltaValue);			
		Estimate density = new KernelDensityEstimate(minValue, maxValue, samples, kernel);
		
		for (int i=0; i<n; i++) {
			density.add(value[i]);
		}
		
		double[] densY = density.getEstimates();
		XYErrorBarSeries xyThrown = new XYErrorBarSeries("thrown");
		double integral = integral();
		double chi2Sum = 0;
		double ndf = 0;
		double lowEdge = minValue;
		for (int iBin=0; iBin<nBin; iBin++) {
			double nEntries = densY[iBin]*n*(maxValue - minValue)/nBin;
			if (nEntries > 0) {
				double x = lowEdge;
				double binSize = (maxValue-minValue)/nBin;
				double y = integral*nEntries/binSize/n;
				double yPlus = integral*(nEntries+Math.sqrt(nEntries))/binSize/n;
				double yMinus;
				if (nEntries == 1) {
					yMinus = y;
				} else {
					yMinus = integral*(nEntries-Math.sqrt(nEntries))/binSize/n;
					ndf++;
					lowEdge += binSize;
					double xMiddle = 0.5*(x+lowEdge);
					double delta = ((integral*nEntries/binSize/n) - distribution.f(xMiddle))/(integral*Math.sqrt(nEntries)/binSize/n);
					chi2Sum += delta*delta;
				}
				
				y = Math.log10(y);
				yPlus = Math.log10(yPlus);
				yMinus = Math.log10(yMinus);
					
				xyThrown.add(x,y, yPlus, yMinus);
				
			}
		}
		
		
		return xyThrown;
	}
	*/
}
