package gov.fnal.eag.sim.random;

import gov.fnal.eag.util.densityEstimate.Histogram;

import java.io.File;

import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.data.xy.XYSeries;

import cern.jet.random.Exponential;
import cern.jet.random.engine.MersenneTwister;
import cern.jet.random.engine.RandomEngine;
import cern.jet.random.engine.RandomSeedGenerator;


/**
 * implements uniform random number generation
 */
public class SimRandomExponential  extends SimRandom {
	/**
	 * generates random numbers with a Normal distribution
	 */
	private transient Exponential exponential;
	/**
	 * the lambda of the distribution
	 */
	protected transient double lambda;
	
	/**
	 * construct a fully specified exponential generator
	 * @param iseedIndex chooses a seed to use
	 * @param lambdaIn the lambda of the distribution exp(-x*lambda)
	 * @param name a name you can choose
	 * @param doHistogram true to keep track of generated random numbers in a histogram
	 */
	public SimRandomExponential(int iseedIndex, double lambdaIn, String name, boolean doHistogram) {
		lambda = lambdaIn; 
		setIseedIndex(iseedIndex);
		setDoHistogram(doHistogram);
	}
	/**
	 * default constructor with the 0th seed, and lambda of 1
	 */
	public SimRandomExponential() {
		this(0, 1.0,"1d exponential",true);
	}
	
	/**
	 * constructor with iseedIndex and lambda of 1
	 * @param iseedIndex the seed index
	 */
	public SimRandomExponential(int iseedIndex) {
		this(iseedIndex, 1.0,"1d normal",true);
	}

	/**
	 * resets the random number sequence with the index
	 * @param iseedIndexIn
	 */
	@Override
	public void setIseedIndex(int iseedIndexIn) {
		iSeedIndex = iseedIndexIn;
		rsg = new RandomSeedGenerator(iSeedIndex, 0);
		RandomEngine randomGenerator = new MersenneTwister(rsg.nextSeed());		
		exponential = new Exponential(lambda, randomGenerator);
		iGenerated = 0;
		return;
	}

	/**
	 * get the next random number
	 * @return double next random number
	 */
	@Override
	public double next() {
		iGenerated++;
		double next = exponential.nextDouble();
		if (doHistogram) {
			histogram.fill(next);
		}
		return next;
	}

	/**
	 * write a plot to the fileName
	 * @param fileName file name to save .PNG to
	 * @param sizeX width of graph
	 * @param sizeY height of graph
	 * @throws Exception 
	 */
	@Override
	public void toPNG(String fileName, int sizeX, int sizeY) throws Exception {
		JFreeChart chart = getChart();
		ChartUtilities.saveChartAsPNG(new File(fileName+".png"), chart, sizeX, sizeY);
	}

	/**
	 * @return the true points, with zero-sized error bars, suitable for plotting on the histogram
	 */
	@Override
	public XYSeries getTrue() {
		XYSeries xyDistribution = new XYSeries("true");
		if (histogram != null) {			
			double contents = histogram.getNFill();
			if (Math.abs(contents) < 10*Double.MIN_VALUE) {
				contents = 1;
			}
			double vAtMax = Math.exp(-lambda*histogram.getMax());
			double vAtMin = Math.exp(-lambda*histogram.getMin());
			double vDelta = vAtMin-vAtMax;
			double binSize = histogram.getBinSize();
			double a0 = binSize*contents*lambda/vDelta;
			for (double x=histogram.getMin(); x <histogram.getMax(); x+=histogram.getBinSize()/10.0) {
				double y = a0*Math.exp(-lambda*x);
				xyDistribution.add(x,y);
			}			
		}
		return xyDistribution;
	}

	@Override
	protected void makeHistogram() {
		if (histogram == null) {
			histogram = new Histogram("SimRandomNormal", "value", "dN/dValue", 100, 0.0, 5/lambda);
		}
	}
	/**
	 * @return the value of lambda (1/lifetime)
	 */
	public double getLambda() {
		return lambda;
	}
}
