package gov.fnal.eag.sim.random;

import gov.fnal.eag.util.astronomy.NormalDistribution;
import gov.fnal.eag.util.astronomy.SimpleFunction;
import gov.fnal.eag.util.densityEstimate.Histogram;

import java.io.File;

import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.data.xy.XYSeries;

import cern.jet.random.Normal;
import cern.jet.random.engine.MersenneTwister;
import cern.jet.random.engine.RandomEngine;
import cern.jet.random.engine.RandomSeedGenerator;


/**
 * implements uniform random number generation
 */
public class SimRandomNormal  extends SimRandom {
	/**
	 * generates random numbers with a Normal distribution
	 */
	private transient Normal normal;
	/**
	 * the mean of the normal distribution
	 */
	protected transient double mean;
	/**
	 * the sigma of the normal distribution
	 */
	protected transient double sigma;
	
	/**
	 * construct a fully specified normal random number generator
	 * @param iseedIndex chooses a seed to use
	 * @param meanIn the mean of the normal distribution
	 * @param sigmaIn the sigma of the normal distribution
	 * @param name a name you can choose
	 * @param doHistogram true to keep track of generated random numbers in a histogram
	 */
	public SimRandomNormal(int iseedIndex, double meanIn, double sigmaIn, String name, boolean doHistogram) {
		mean = meanIn; sigma = Math.abs(sigmaIn);
		if (sigma == 0) sigma = 1.0;
		setIseedIndex(iseedIndex);
		setDoHistogram(doHistogram);
	}
	/**
	 * default constructor with the 0th seed, mean of 0, and sigma of 1
	 */
	public SimRandomNormal() {
		this(0, 0.0, 1.0,"1d normal",true);
	}
	
	/**
	 * constructor with iseedIndex, with mean of 0 and sigma of 1
	 * @param iseedIndex the seed index
	 */
	public SimRandomNormal(int iseedIndex) {
		this(iseedIndex, 0.0, 1.0,"1d normal",true);
	}


	/**
	 * get the next random number
	 * @return double next random number
	 */
	@Override
	public double next() {
		iGenerated++;
		double next = normal.nextDouble();
		if (doHistogram) {
			histogram.fill(next);
		}
		return next;
	}

	/**
	 * write a plot to the fileName
	 * @param fileName file name to save .PNG to
	 * @param sizeX width of graph
	 * @param sizeY height of graph
	 * @throws Exception 
	 */
	@Override
	public void toPNG(String fileName, int sizeX, int sizeY) throws Exception {
		JFreeChart chart = getChart();
		ChartUtilities.saveChartAsPNG(new File(fileName+".png"), chart, sizeX, sizeY);
	}

	/**
	 * @return the true points, with zero-sized error bars
	 */
	@Override
	public XYSeries getTrue() {
		double minValue = mean-5*sigma;
		SimpleFunction normalTrue = new NormalDistribution(mean, sigma);
		XYSeries xyDistribution = new XYSeries("true");
		int nx = 200;
		if (histogram != null) {
			nx = histogram.getNBin();
		}
		double deltaValue = 10*sigma/nx;
		for (int ix=0; ix<nx+1; ix++) {
			double x = minValue+deltaValue*(ix);
			double value = iGenerated*normalTrue.f(x)*deltaValue;
			xyDistribution.add(x, value);
		}
		return xyDistribution;
	}

	/**
	 * get the maximum value
	 * @return the max
	 */
	public double getMax() {
		double max = 0;
		double minValue = mean-5*sigma;
		double maxValue = mean+5*sigma;
		double deltaValue = .1;
		SimpleFunction normalMax = new NormalDistribution(mean, sigma);
		int nx = (int) ((maxValue-minValue)/deltaValue);
		for (int ix=0; ix<nx; ix++) {
			double x = minValue+deltaValue*(ix+0.5);
			double y = normalMax.f(x);
			if (y > max)
				max = y;
		}
		return max;
	}
	@Override
	protected void makeHistogram() {
		if (histogram == null) {
			histogram = new Histogram("SimRandomNormal", "value", "dN/dValue", 100, mean-5*sigma, mean+5*sigma);
		}
	}

	/**
	 * @return the mean of the generated distribution
	 */
	public double getMean() {
		return mean;
	}
	/**
	 * @return the sigma of the generated distribution
	 */
	public double getSigma() {
		return sigma;
	}
	/**
	 * resets the random number sequence with the index
	 * @param iseedIndexIn
	 */
	@Override
	public void setIseedIndex(int iseedIndexIn) {
		iSeedIndex = iseedIndexIn;
		rsg = new RandomSeedGenerator(iSeedIndex, 0);
		RandomEngine randomGenerator = new MersenneTwister(rsg.nextSeed());		
		normal = new Normal(mean, sigma, randomGenerator);
		iGenerated = 0;
		return;
	}
}
