package gov.fnal.eag.sim.random;

import gov.fnal.eag.util.astronomy.NormalDistribution;
import gov.fnal.eag.util.astronomy.SimpleFunction;
import gov.fnal.eag.util.densityEstimate.Histogram;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;

import org.jfree.chart.JFreeChart;
import org.jfree.data.xy.XYSeries;


/**
 * implements normal random number generation by reading in a table of pre-generated numbers
 */
public class SimRandomNormalFast  extends SimRandom {
	private int n;
	private float mean, sigma;
	private float[] x;
	private boolean doneReading = false;
	private DataInputStream inputStream;
	/**
	 * initialize reading the list of numbers
	 * @param iSeedIndexIn offset into the list
	 * @throws IOException 
	 */
	public SimRandomNormalFast(int iSeedIndexIn) throws IOException {
		this.iSeedIndex = Math.abs(iSeedIndexIn);
		String fileName = "data"+File.separator+"simRandomNormalFast.dat";
		URL inputURL = this.getClass().getResource(fileName);
			InputStream tempInputStream = inputURL.openStream();
			inputStream = new DataInputStream(tempInputStream);
			n = inputStream.readInt();
			iSeedIndex %= n;
			mean = inputStream.readFloat();
			sigma = inputStream.readFloat();
			x = new float[n];
			if (iSeedIndex > 0) {
				for (int i=0; i<iSeedIndex; i++) {
					x[i] = (inputStream.readFloat()/sigma)-mean;
				}
			}
	}
	
	/**
	 * create the file simRandomNormalFast.dat to be installed in the data directory of this package
	 * @param args makeNewData to create the data file; otherwise, print the first 10 values
	 * @throws IOException 
	 */
	public static void main(String[] args) throws IOException {
		if (args.length == 1 && args[0].equals("makeNewData")) {
			
			DataOutputStream dos = new DataOutputStream(new FileOutputStream("simRandomNormalFast.dat"));
			int n = 100000;
			dos.writeInt(n);
			
			float mean = 0.0f;
			dos.writeFloat(mean);
			
			float sigma = 1.0f;
			dos.writeFloat(sigma);
			SimRandomNormal srn = new SimRandomNormal(1, mean, sigma, "", false);
			for (int i=0; i<n; i++) {				
				float x = (float) srn.next();
				if (i % 100000 == 0) {
					System.out.println("i="+i+"/"+n+" x="+x);
				}
				dos.writeFloat(x);				
			}
			dos.close();
		
		} else {
			SimRandomNormalFast srnf = new SimRandomNormalFast(1);
			for (int i=0; i<10; i++) {
				System.out.println("i="+i+" random="+srnf.next());
			}
		}
	}
	/**
	 * @see gov.fnal.eag.sim.random.SimRandom#getChart()
	 * returns null
	 */
	@Override
	public JFreeChart getChart() throws Exception {
		return null;
	}


	/**
	 * @see gov.fnal.eag.sim.random.SimRandom#next()
	 */
	@Override
	public double next() {
		if (iSeedIndex >= x.length) {
			iSeedIndex=0;
			doneReading = true;
			try {
				inputStream.close();
			} catch (IOException e) {
				throw new Error(e);
			}
		}
		if (!doneReading) {
			try {
				x[iSeedIndex] = (inputStream.readFloat()/sigma)-mean;
			} catch (IOException e) {
				throw new Error(e);
			}
		}
		double retval = x[iSeedIndex];
		iSeedIndex++;
		return retval;
	}

	/**
	 * @see gov.fnal.eag.sim.random.SimRandom#toPNG(java.lang.String, int, int)
	 */
	@Override
	public void toPNG(String fileName, int sizeX, int sizeY) throws Exception {
		throw new UnsupportedOperationException();
	}

	/**
	 * @return the number of pre-computed random numbers
	 */
	public int getN() {
		return n;		
	}

	@Override
	protected void makeHistogram() {
		if (histogram == null) {
			histogram = new Histogram("SimRandomNormal", "value", "dN/dValue", 100, mean-5*sigma, mean+5*sigma);
		}
	}
	
	/**
	 * @return the true points, with zero-sized error bars
	 */
	@Override
	public XYSeries getTrue() {
		double minValue = mean-5*sigma;
		SimpleFunction normalTrue = new NormalDistribution(mean, sigma);
		XYSeries xyDistribution = new XYSeries("true");
		int nx = 200;
		if (histogram != null) {
			nx = histogram.getNBin();
		}
		double deltaValue = 10*sigma/nx;
		for (int ix=0; ix<nx+1; ix++) {
			double tempX = minValue+deltaValue*(ix);
			double value = iGenerated*normalTrue.f(tempX)*deltaValue;
			xyDistribution.add(tempX, value);
		}
		return xyDistribution;
	}

	/**
	 * @see gov.fnal.eag.sim.random.SimRandom#setIseedIndex(int)
	 */
	@Override
	public void setIseedIndex(int seedIndex) {
		iSeedIndex = seedIndex;
	}
}
