package gov.fnal.eag.sim.random;


import gov.fnal.eag.util.astronomy.SimpleFunction;
import gov.fnal.eag.util.densityEstimate.Histogram;
import gov.fnal.eag.util.densityEstimate.PoissonDistribution;

import org.jfree.data.xy.XYSeries;

import cern.jet.random.Poisson;
import cern.jet.random.engine.MersenneTwister;
import cern.jet.random.engine.RandomEngine;
import cern.jet.random.engine.RandomSeedGenerator;


/**
 * implements Poisson random number generation
 */
public class SimRandomPoisson  extends SimRandom {
	/**
	 * the mean of the distribution
	 */
	protected double mean;
	/**
	 * the generator from Colt
	 */
	protected Poisson poisson;

	/**
	 * generates random numbers in a Poisson distribution
	 * @param iseedIndex seed index
	 * @param meanIn mean of the distribution
	 * @param name name of this generator
	 * @param doHistogram whether to histogram thrown values
	 */
	public SimRandomPoisson(int iseedIndex, double meanIn, String name, boolean doHistogram) {
		mean = Math.abs(meanIn);
		setIseedIndex(iseedIndex);
		setDoHistogram(doHistogram);
	}
	/**
	 * default constructor 
	 */
	public SimRandomPoisson()  {
		this(0, 1.0,"1d poisson", true);
	}
	/**
	 * constructor, with random number seed specified
	 * @param iseedIndex
	 */
	public SimRandomPoisson(int iseedIndex)  {
		this(iseedIndex, 1.0,"1d poisson", true);
	}

	/**
	 * set the random number seed
	 * @param iseedIndexIn
	 */
	@Override
	public void setIseedIndex(int iseedIndexIn) {
		iSeedIndex = iseedIndexIn;
		rsg = new RandomSeedGenerator(iSeedIndex, 0);
		RandomEngine randomGenerator = new MersenneTwister(rsg.nextSeed());		
		poisson = new Poisson(mean, randomGenerator);
		//System.out.println(" SimRandomPoisson.setIseedIndex():  mean="+mean);
		iGenerated = 0;
		return;
	}

	/**
	 * get the next randomly generated number and fill one entry of the histogram if requested
	 * @return a random number drawn from the Poisson distribution with the mean
	 */
	@Override
	public double next() {
		return next(mean);
	}

	/**
	 * get the next randomly generated number and fill one entry of the histogram if requested
	 * @param meanIn specify the mean
	 * @return a random number drawn from the Poisson distribution with the specified mean
	 */
	public int next(double meanIn) {
		iGenerated++;
		int next = poisson.nextInt(meanIn);
		if (doHistogram) {
			histogram.fill(next);
		}
		return next;
	}
	
	/**
	 * @return the true points for plotting
	 */ 
	@Override
	public XYSeries getTrue() {
		XYSeries xyDistribution = new XYSeries("true");
		double iMin = Math.floor(Math.max(0, mean-5.0*Math.sqrt(mean)));
		double iMax = Math.ceil(mean+5.0*Math.sqrt(mean));
		if (histogram != null) {
			iMin = histogram.getMin();
			iMax = histogram.getMax();			
		}
		SimpleFunction distribution = new PoissonDistribution(mean);
		for (int i=(int) iMin; i<iMax+1; i++) {
			double weight = distribution.f(i);
			//System.out.println(" getTrue:  mean="+mean+" i="+i+" weight="+weight+" iGenerated="+iGenerated);
			xyDistribution.add(i, weight*iGenerated);
		}
		return xyDistribution;
	}

	@Override
	protected void makeHistogram() {
		if (histogram == null) {
			double iMin = Math.floor(Math.max(0, mean-5.0*Math.sqrt(mean)));
			double iMax = Math.ceil(mean+5.0*Math.sqrt(mean));
			int nBin = (int) (iMax-iMin+1);
			//System.out.println(" in SimRandomPoisson.makeHistogram:  iMin="+iMin+" iMax="+iMax+" nBin="+nBin);
			histogram = new Histogram(String.format("SimRandomPoisson mean=%.2f",mean), 
					"value/mean", "dN/dValue", nBin, iMin-0.5, iMax+0.5);
		}
	}
	
}
