package gov.fnal.eag.sim.random;

import gov.fnal.eag.sim.catalog.AstronomicalFilter;
import gov.fnal.eag.sim.catalog.CoordinateGenerator;
import gov.fnal.eag.util.astronomy.Integrate;
import gov.fnal.eag.util.astronomy.SimpleFunction;
import gov.fnal.eag.util.htm.HTMfunc;
import gov.fnal.eag.wcs.SphericalCoordinate;

import java.io.File;

import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;

/**
 * Class Schecter generates numbers with the Schecter galaxy luminosity function
 */
public class SimRandomSchechter extends SimRandomDistribution implements SimRandomLuminosity {
	private String filterName;
	AstronomicalFilter filter = null;
	double phiStar, alpha, mStar;
	int seed;
	double spaceDensity;
	/**
	 * fully-specified constructor
	 * @param seed random number seed index
	 * 
	 * @param phiStar
	 *            space density in pc^-3;
	 * @param alpha
	 *            slope
	 * @param mStar
	 *            characteristic magnitude
	 * @param mMin
	 *            bright limit
	 * @param mMax
	 *            faint limit
	 * @param nBin 
	 * 			  number of bins in lookup table
	 */
	public SimRandomSchechter(int seed, double phiStar, double alpha, double mStar,
			double mMin, double mMax, int nBin) {
		set(seed, phiStar, alpha, mStar, mMin, mMax, nBin);
	}
	private void set(final int seed, final double phiStar, final double alpha, final double mStar, double minMagnitude,
			double maxMagnitude, int nBin) {		
		this.phiStar = phiStar;
		this.alpha = alpha;
		this.mStar = mStar;
		this.minValue = minMagnitude;
		this.maxValue = maxMagnitude;
		this.deltaValue = (maxMagnitude-minMagnitude)/(nBin+1);
		this.nBin = nBin;
		this.seed = seed;
	    SimpleFunction temp = new SimpleFunction() {
			@Override
			public double f(double magnitude) {
	    		double norm = 0.4*Math.log(10.0)*phiStar;
	    		double powerTerm = Math.pow(10.0, 0.4*(mStar-magnitude)*(alpha+1));
	    		double inside = -Math.pow(10.0, 0.4*(mStar-magnitude));
	    		double expTerm = Math.exp(inside);
	    		return norm*powerTerm*expTerm;
	    	}
	    };
	    setDistribution(temp);
		simRandomUniform = new SimRandomUniform();
		setIseedIndex(seed);
		spaceDensity = Integrate.qRomb(distribution, minValue, maxValue, 1e-3);
		if (spaceDensity <= 0.0) {
			throw new IllegalArgumentException("not positive space density: "+spaceDensity+" minValue="+minValue+" maxValue="+maxValue);
		}
	}
	
	/**
	 * @see gov.fnal.eag.sim.random.SimRandomLuminosity#setFilterName(java.lang.String)
	 */
	public void setFilterName(String filterName) {
		this.filterName = filterName;
	}
	
	/**
	 * sets the filter and its name
	 * @param filter
	 */
	public void setFilter(AstronomicalFilter filter) {
		this.filter = filter;
		filterName = filter.getName();
	}
	/**
	 * @see gov.fnal.eag.sim.random.SimRandomLuminosity#getFilterName()
	 */
	public String getFilterName() {
		return filterName;
	}
	/**
	 * @see gov.fnal.eag.sim.random.SimRandomLuminosity#getSpaceDensity()
	 */
	public double getSpaceDensity() {
		return spaceDensity;
	}
	/**
	 * @see gov.fnal.eag.sim.random.SimRandomLuminosity#getSurfaceDensity()
	 */
	public double getSurfaceDensity() {		
		return Double.NaN;
	}
	
	/**
	 * write a plot to the fileName
	 * @param fileName file name to save .PNG to
	 * @param sizeX width of graph
	 * @param sizeY height of graph
	 * @throws Exception 
	 */
	@Override
	public void toPNG(String fileName, int sizeX, int sizeY) throws Exception{
		JFreeChart chart = getChart();
		ChartUtilities.saveChartAsPNG(new File(fileName+".png"), chart, sizeX, sizeY);
	}
	/**
	 * @see gov.fnal.eag.sim.random.SimRandomLuminosity#getISeed()
	 */
	public int getISeed() {
		return getIseedIndex();
	}
	
	/**
	 * @see gov.fnal.eag.sim.random.SimRandomLuminosity#copy()
	 */
	public SimRandomLuminosity copy() {
		SimRandomSchechter retval = new SimRandomSchechter(getISeed(), phiStar, alpha, mStar,
				minValue, maxValue, nBin);
		return retval;
		
	}
	/**
	 * @see gov.fnal.eag.sim.random.SimRandomLuminosity#getFilter()
	 */
	public AstronomicalFilter getFilter() {
		return filter;
	}
	/**
	 * @see gov.fnal.eag.sim.random.SimRandomLuminosity#next(gov.fnal.eag.wcs.SphericalCoordinate)
	 */
	public double next(SphericalCoordinate sc) {
		long thisHtm = HTMfunc.lookupId(sc.getLongitude(), sc.getLatitude(), CoordinateGenerator.depth);
		int iSeedIndexTemp = HTMfunc.indexFromHtm(thisHtm+seed);
		setIseedIndex(iSeedIndexTemp);
		return next(true);
	}
}
