package gov.fnal.eag.sim.random;

import gov.fnal.eag.util.astronomy.SimpleFunction;
import gov.fnal.eag.util.densityEstimate.Histogram;
import gov.fnal.eag.util.densityEstimate.KernelDensityEstimate;
import gov.fnal.eag.util.densityEstimate.UniformKernel;

import org.jfree.data.xy.XYSeries;

import cern.jet.random.Uniform;
import cern.jet.random.engine.MersenneTwister;
import cern.jet.random.engine.RandomEngine;
import cern.jet.random.engine.RandomSeedGenerator;

/**
 * implements uniform random number generation
 */
public class SimRandomUniform  extends SimRandom {
	private Uniform uniform;
	private double min,max;
	
	/**
	 * fully-specified constructor
	 * @param iseedIndex random number seed
	 * @param minIn minimum value to generate
	 * @param maxIn maximum value to generate
	 * @param name a descriptive hame
	 * @param doHistogram true if you want to keep histograms
	 */
	public SimRandomUniform(int iseedIndex, double minIn, double maxIn, String name, boolean doHistogram) {
		this(iseedIndex, minIn, maxIn);
		setDoHistogram(doHistogram);
		SimpleFunction kernel = new UniformKernel(.1);
		kernelDensityEstimate = new KernelDensityEstimate(min, max, 100, kernel);
	}
	
	
	/**
	 * constructor with no histogram creation
	 * @param iseedIndex random number seed
	 * @param minIn minimum value to generate
	 * @param maxIn maximum value to generate
	 */
	public SimRandomUniform(int iseedIndex, double minIn, double maxIn) {
		setIseedIndex(iseedIndex);
		min = minIn;
		max = maxIn;		
		setDoHistogram(false);
	}
		
	
	/**
	 * default constructor
	 */
	public SimRandomUniform() {
		this(0);
	}
	
	/**
	 * Constructor with write to histogram 
	 * @param name a descriptive name
	 */
	public SimRandomUniform(String name) {
		this(0,name);
	}

	/**
	 * constructor with only the random seed specified
	 * @param iseedIndex the random number seed index
	 */
	public SimRandomUniform(int iseedIndex) {
		this(iseedIndex,0.0,1.0);
	}

	/**
	 * constructor with only the random seed and name of histogram specified
	 * @param iseedIndex the random number seed index
	 * @param name the name of the histogram
	 */
	public SimRandomUniform(int iseedIndex, String name) {
		this(iseedIndex, 0.0, 1.0, name, true);
	}

	/**
	 * set the random number seed index
	 * @param iseedIndexIn
	 */
	@Override
	public void setIseedIndex(int iseedIndexIn) {
		if (iseedIndexIn < 0) {
			throw new IllegalArgumentException("iseedIndexIn can not be negative but iseedIndexIn="+iseedIndexIn);
		}
		iSeedIndex = iseedIndexIn;
		rsg = new RandomSeedGenerator(iSeedIndex, 0);
		RandomEngine randomGenerator = new MersenneTwister(rsg.nextSeed());
		uniform = new Uniform(randomGenerator);
		iGenerated = 0;
		return;
	}
	/**
	 * get the next value
	 * @return a number drawn uniformly from min to max and update the histogram if requested
	 */
	@Override
	public double next() {
		iGenerated++;
		double next = uniform.nextDoubleFromTo(min,max);
		if (doHistogram) {
			histogram.fill(next);
		}
		return next;
	}
	/**
	 * get the next value
	 * @param minIn minimum value
	 * @param maxIn maximum value
	 * @return a number drawn from minIn to maxIn but with no histogram update
	 */
	public double next(double minIn, double maxIn) {
		double retval = uniform.nextDoubleFromTo(minIn, maxIn);
		return retval;
		//return minIn + (maxIn-minIn)*next();		
	}



	/**
	 * @return the true points, with zero-sized error bars
	 */
	@Override
	public XYSeries getTrue() {
		double minValue = min;
		double maxValue = max;
		XYSeries xyDistribution = new XYSeries("true");
		int nx = 200;
		if (histogram != null) {
			nx = histogram.getNBin();
		}
		double deltaX = (maxValue-minValue)/nx;
		for (int ix=0; ix<nx+1; ix++) {
			double x = minValue+deltaX*ix;
			double value = iGenerated/nx;
			xyDistribution.add(x, value);
		}
		return xyDistribution;
	}


	@Override
	protected void makeHistogram() {
		if (histogram == null) {
			histogram = new Histogram("SimRandomUniform", "value", "dN/dValue", 100, min, max);
		}
	}
}
