package gov.fnal.eag.sim.random;

import gov.fnal.eag.sim.catalog.CoordinateGenerator;
import gov.fnal.eag.util.htm.ConvexRegion;
import gov.fnal.eag.util.htm.HTMfunc;
import gov.fnal.eag.util.htm.HTMrange;
import gov.fnal.eag.util.htm.Vector3d;
import gov.fnal.eag.wcs.SphericalCoordinate;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * Class UniformCoordinateGenerator
 * 
 */
public class UniformCoordinateGenerator implements CoordinateGenerator {
	/**
	 * set to true to keep track of debugging information
	 */
	public boolean debug = false;
	/**
	 * coordinates accepted by this generator
	 */
	public ArrayList<SphericalCoordinate> accepted = new ArrayList<SphericalCoordinate>();
	/**
	 * coordinates considered, but rejected, by this generator
	 */
	public ArrayList<SphericalCoordinate> rejected = new ArrayList<SphericalCoordinate>();
	/**
	 * the htmids used by this generator
	 */
	public ArrayList<Long> htmids = new ArrayList<Long>();
	/**
	 * the number of objects in each htmid
	 */
	public ArrayList<Integer> nPerHtmid = new ArrayList<Integer>();
	int nPrepareNext = 0;

	ConvexRegion convexRegion;
	SphericalCoordinate min,max;
	int seed;
	double meanObjectsPerSteradian;

	/**
	 * default constructor
	 * @param min ra,dec min value
	 * @param max ra,dec max value
	 * @param seed random seed index
	 * @param meanObjectsPerDegreeSquared surface density of objects
	 */
	public UniformCoordinateGenerator(SphericalCoordinate min, SphericalCoordinate max, int seed, double meanObjectsPerDegreeSquared) {
		this.min = min;
		this.max = max;
		this.seed = seed;
		this.meanObjectsPerSteradian = meanObjectsPerDegreeSquared;
		meanObjectsPerSteradian *= Math.pow(Math.toDegrees(1),2);
		convexRegion = new ConvexRegion(min,max);
	}


	/** 
	 * @see gov.fnal.eag.sim.catalog.CoordinateGenerator#steradians()
	 */
	public double steradians() {
		return SphericalCoordinate.steradians(min, max);		
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.CoordinateGenerator#copy()
	 */
	public UniformCoordinateGenerator copy() {
		UniformCoordinateGenerator retval = new UniformCoordinateGenerator(min, max, seed, meanObjectsPerSteradian/Math.pow(Math.toDegrees(1), 2));
		return retval;
	}

	/**
	 * @see gov.fnal.eag.sim.catalog.CoordinateGenerator#getISeed()
	 */
	public int getISeed() {
		return seed;
	}


	/**
	 * @see java.lang.Iterable#iterator()
	 */
	public Iterator<SphericalCoordinate> iterator() {
		return new UniformCoordinateIterator();
	}

	private class UniformCoordinateIterator implements Iterator<SphericalCoordinate> {

		
		private HTMrange htmRange = (new HTMrange(convexRegion, depth, true)).setDepth(depth);

		Iterator<Long> htmIterator = htmRange.iterator();
		int nObjectsThisHtm = 0;
		Long thisHtm = null;
		private SimRandomPoisson srp = new SimRandomPoisson();
		boolean hasNext = false;
		boolean htmValid = false;
		SphericalCoordinate next;
		RandomSphere randomSphere;
		double azimuthOffset;

		UniformCoordinateIterator() {
			nextHtm();
			prepareNext();
		}
		/**
		 * @see java.util.Iterator#hasNext()
		 */
		public boolean hasNext() {
			return hasNext;
		}

		/**
		 * @see java.util.Iterator#next()
		 */
		public SphericalCoordinate next() {
			SphericalCoordinate retval = next;
			prepareNext();
			return retval;
		}
		/**
		 * @see java.util.Iterator#remove()
		 * @throws UnsupportedOperationException
		 */
		public void remove() {
			throw new UnsupportedOperationException();
		}
		private void prepareNext() {
			if (!htmValid) {
				hasNext = false;
				next = null;
				return;
			}
			while (true) {
				if (nObjectsThisHtm == 0) {
					nextHtm();
					if (!htmValid) {
						hasNext = false;
						next = null;
						return;
					}
				}
				next = nextInThisHtm();
				nObjectsThisHtm--;
				boolean contained = convexRegion.contains(next);
				if (contained) {
					if (debug) {
						accepted.add(next);
					}
					hasNext = true;
					return;
				} 					
				if (debug) {
					rejected.add(next);
				}
			}
		}

		private SphericalCoordinate nextInThisHtm() {
			SphericalCoordinate tempSc;
			Vector3d tempV3d;
			int thisDepth = HTMfunc.idLevel(thisHtm);
			while (true) {
				tempSc = randomSphere.next();
				tempV3d = new Vector3d(tempSc.getLongitude(), tempSc.getLatitude());
				tempV3d.rotZ(-azimuthOffset);
				Long tempHtm = HTMfunc.lookupid(tempV3d, thisDepth);
				if (thisHtm.equals(tempHtm)) {
					SphericalCoordinate retval = new SphericalCoordinate(tempV3d.ra(), tempV3d.dec());
					return retval;
				}
			}
		}

		private void nextHtm() {
			while (true) {
				if (htmIterator.hasNext()) {
					thisHtm = htmIterator.next();
					if (debug) {
						htmids.add(thisHtm);
					}
					double htmArea = HTMfunc.steradians(thisHtm);
					double mean = htmArea*meanObjectsPerSteradian;
					int srpSeed = HTMfunc.indexFromHtm(thisHtm + seed);
					srp.setIseedIndex(srpSeed);
					nObjectsThisHtm =  srp.next(mean);
					if (debug) {
						nPerHtmid.add(nObjectsThisHtm);
					}
					if (nObjectsThisHtm > 0) {
						htmValid=true;
						makeRandomSphere();
						return;
					}
				} else {
					htmValid = false;
					return;					
				}
			}
		}
		private void makeRandomSphere() {
			Vector3d[] corners = HTMfunc.idToTriangle(thisHtm);
			azimuthOffset = 180.0-corners[0].ra();
			double azMin = 360;
			double azMax = 0;
			double altMin = 90;
			double altMax = -90;
			for (int i=0; i<3; i++) {
				corners[i].rotZ(azimuthOffset);
				azMin=Math.min(azMin, corners[i].ra());
				azMax=Math.max(azMax, corners[i].ra());
				altMin=Math.min(altMin, corners[i].dec());
				altMax=Math.max(altMax, corners[i].dec());
			}
			int iSeedIndex = HTMfunc.indexFromHtm(thisHtm+seed);
			randomSphere = new RandomSphere(iSeedIndex, altMin, altMax, azMin, azMax);
		}
	}
	/**
	 * @return some diagnostic information
	 */
	public String checkSums() {
		String retval = "debug not enabled";
		if (debug) {
			int nAccepted = accepted.size();
			int nRejected = rejected.size();
			int sum = 0;
			for (Integer nper : nPerHtmid) {
				sum += nper;
			}
			int nHtmid = nPerHtmid.size();
			retval = String.format("nAccepted=%5d  nRejected=%5d  nAR=%5d sum=%5d nHtmid=%3d nPrepareNext=%5d ", nAccepted, nRejected, (nAccepted+nRejected), sum, nHtmid, nPrepareNext);
			for (int ihtmid=0; ihtmid<nPerHtmid.size(); ihtmid++) {
				retval += "  ithmid="+ihtmid+":"+nPerHtmid.get(ihtmid);
			}
		}
		return retval;
	}
}

