package gov.fnal.eag.sim.repository.image;

import gov.fnal.eag.sim.pixel.PixeledImage;
import gov.fnal.eag.sim.pixel.PixeledMask;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.zip.GZIPOutputStream;

import nom.tam.fits.BasicHDU;
import nom.tam.fits.Data;
import nom.tam.fits.Fits;
import nom.tam.fits.FitsException;
import nom.tam.fits.Header;
import nom.tam.fits.HeaderCard;
import nom.tam.fits.HeaderCardException;
import nom.tam.fits.ImageHDU;
import nom.tam.util.BufferedFile;
import uk.ac.starlink.table.RowListStarTable;
import uk.ac.starlink.table.RowSequence;
import uk.ac.starlink.table.StarTable;
import uk.ac.starlink.table.jdbc.JDBCStarTable;

/**
 * manages writing PixeledImage and RowListStarTable objects to a file one at a time
 * @author stoughto
 *
 */
public class FitsWriter {
	private BufferedFile bf;
	private boolean isFirst = true;
	private String baseFileName;
	private String suffix = "";
	//private String version = "";
	/**
	 * @param fileName file to write
	 * @throws IOException if there is trouble opening the file
	 */
	public FitsWriter(String fileName) throws IOException {
		this(fileName, true);
	}
	
	/**
	 * @param fileName file to write
	 * @param overwrite throw error if overwrite is true and file exists
	 * @throws IOException if there is trouble opening the BufferedFile
	 */
	public FitsWriter(String fileName, boolean overwrite) throws IOException {
		//version = Fits.version();
		if (!overwrite) {
			File file = new File(fileName);
			if (file.exists()) {
				throw new Error(fileName+" already exists");
			}
		}
		init(fileName);
	}
	private void init(String fileName) throws IOException {
		if (fileName.endsWith(".gz")) {
			suffix = ".gz";
			baseFileName = fileName.substring(0, fileName.length()-3);
		} else {
			baseFileName = fileName;
		}
		bf = new BufferedFile(baseFileName, "rw");		
	}
	/**
	 * @param pixeledImage to put to the stream, defaulting to FLOAT format for FITS
	 */
	public void put(PixeledImage pixeledImage) {
		put(pixeledImage, FitsDataFormat.FLOAT);
	}
	
	/**
	 * write the PixeledMask as ushort, shifting if necessary
	 * @param pixeledMask
	 */
	public void put(PixeledMask pixeledMask) {
		BigInteger twoSixteenth = BigInteger.ONE.shiftLeft(16);
		int minByteLength = pixeledMask.getMinByteLength();
		int nHdu = 1 + minByteLength/2;
		for (int iHdu=0; iHdu<nHdu; iHdu++) {
			PixeledImage pi = new PixeledImage(pixeledMask.getRowCount(), pixeledMask.getColumnCount());
			pi.setRow0Col0(pixeledMask.getRow0(), pixeledMask.getCol0());
			for (int row=pi.getRowMin(); row<pi.getRowMax(); row++) {
				for (int col=pi.getColumnMin(); col<pi.getColumnMax(); col++) {
					BigInteger value = pixeledMask.getByRowCol(row, col).shiftRight(16*iHdu).remainder(twoSixteenth);
					pi.setByRowCol(row, col, value.doubleValue());
				}
			}
			try {
				pi.header.addValue("PIXMASK", 16*iHdu, "these are mask bits, shift left this many bits");
			} catch (HeaderCardException e) {
				throw new Error(e);
			}
			put(pi,FitsDataFormat.USHORT);
		}
	
	}
	/**
	 * @param pixeledImage to put to the stream
	 * @param fitsDataFormat gives bitpix, bzero, and bscale for FITS.
	 */
	public void put(PixeledImage pixeledImage, FitsDataFormat fitsDataFormat) {
		try {
			int bitpix = fitsDataFormat.getBitpix();
			Data data;
			//double min = fitsDataFormat.getMin();
			//double max = fitsDataFormat.getMax();
			if (bitpix == -64) {
				data = Fits.makeHDU(pixeledImage.getPix()).getData();
			} else if (bitpix == -32) {
				data = Fits.makeHDU(pixeledImage.getPixAsFloat()).getData();
			} else if (bitpix == 16) {
				data = Fits.makeHDU(pixeledImage.getPixAsUShort()).getData();
			} else {
				throw new Error("need to implement getPixAs method for bitpix="+bitpix);
			}
			Header header = getHeader(pixeledImage, fitsDataFormat);
			ImageHDU imageHDU = new ImageHDU(header, data);
			
			if (isFirst) {
				imageHDU.setPrimaryHDU(true);
				isFirst = false;
			} else {
				imageHDU.setPrimaryHDU(false);
			}
			
			try {
				Fits.setChecksum(imageHDU);
			} catch (IOException e) {
				e.printStackTrace();
			}
			
			imageHDU.write(bf);
		} catch (FitsException e) {
			throw new Error(e);
		}		
	}
	
	/**
	 * @param rlst RowListStarTable to put to the stream
	 */
	public void put(StarTable rlst) {
			if (isFirst) {
				BasicHDU dummyHDU = BasicHDU.getDummyHDU();
				try {
					dummyHDU.setPrimaryHDU(isFirst);
				} catch (FitsException e) {
					throw new Error(e);
				}
				try {
					dummyHDU.write(bf);
				} catch (FitsException e) {
					throw new Error(e);
				}
				isFirst = false;
			}
			FitsTableSerializerEag fts;
			try {
				fts = new FitsTableSerializerEag(rlst);
			} catch (IOException e) {
				throw new Error(e);
			}
			try {
				fts.writeHeader(bf);
			} catch (IOException e) {
				throw new Error(e);
			}
			try {
				fts.writeData(bf);
			} catch (IOException e) {
				throw new Error(e);
			}
	}
	/**
	 * simple method to  write checksum
	 * this method should be called after close() method
	 * will not work on .gz files
	 * will not work on very big files (>1GB)
	 * @throws IOException
	 * @throws FitsException 
	 */
	public void makeCheckSum() throws IOException, FitsException {
//		System.out.println("in checsum");
		File ff = new File(baseFileName);
		FileInputStream fis = new FileInputStream(ff);
		Fits ftf = new Fits(fis);
		
		//int nhdu = ftf.getNumberOfHDUs();
		int i = 0;
		boolean inp = true;
		while (inp) {
			System.out.println(" read HDIU "+i);
			i++;
			BasicHDU hdui = ftf.readHDU();
			if (hdui == null) {
				inp = false;
				break;
			}
		
		Fits.setChecksum(hdui);
		hdui.write(bf);
		System.gc();
		}
		System.out.println("number of HDUs="+i);
		fis.close();
		ff.delete();
		System.out.println(" replacing original file "+baseFileName);
		bf.flush();
		bf.setLength(bf.getFilePointer());
		bf.close();
		File resF = new File(baseFileName+".tmp");
		File ff2 = new File(baseFileName);
		resF.renameTo(ff2);
/*		
		BasicHDU[] data = ftf.read();

		ftf.setChecksum();
		fis.close();
		BufferedFile tempBf = new BufferedFile(baseFileName, "rw");	

		for (BasicHDU bu:data) {

			bu.write(tempBf);
		}
		tempBf.flush();
		tempBf.setLength(tempBf.getFilePointer());
		tempBf.close();
*/
	}
	
	/**
	 * close the stream
	 * @throws IOException 
	 */
	public void close() throws IOException {
		bf.flush();		
		bf.setLength(bf.getFilePointer());
		bf.close();
		if (suffix.equals(".gz")) {
			gzip(baseFileName);
		}

	}
	
	/**
	 * gzip the file
	 * @param fileName
	 */
	public static void gzip(String fileName) {
		FileInputStream in;
		GZIPOutputStream out;
		File inFile;
		try {
			in = new FileInputStream(fileName);
			inFile = new File(fileName);
		} catch (FileNotFoundException e) {
			throw new Error(e);
		}
		try {
			out = new GZIPOutputStream(new FileOutputStream(fileName+".gz"));
		} catch (FileNotFoundException e) {
			throw new Error(e);
		} catch (IOException e) {
			throw new Error(e);
		}
		int bufferSize = 1024;
		byte[] buffer = new byte[bufferSize];
		int len;
		try {
			while ( (len = in.read(buffer)) >= 0 ) {
				out.write(buffer, 0, len);
			}
		} catch (IOException e) {
			throw new Error(e);
		}
		try {
			in.close();
			inFile.delete();
		} catch (IOException e) {
			throw new Error(e);
		}		
		try {
			out.finish();
			out.close();
		} catch (IOException e) {
			throw new Error(e);
		}
	}
	/**
	 * copy the file
	 * @param args inputFileName outputFileName
	 * @throws Exception
	 */
	public static void main(String[] args) throws Exception {
		if (args.length != 2) {
			System.out.println("Usage:  gov.fnal.eag.sim.repository.image.FitsWrite inputFileName outputFileName");
		}
		ImageRepository ir = new ImageRepository(".");
		String inputFileName = args[0];
		String outputFileName = args[1];
		
		FitsWriter fitsWriter = new FitsWriter(outputFileName);
		Header phdr = new Header();
		
		LinkedList<Object> objInRep = ir.get(inputFileName, 0);
		int nToDo = objInRep.size();
		System.out.println("n to do ="+nToDo);
		try {
			phdr.addValue("SIMPLE",true," confirm to FITS standard");
		
		phdr.addValue("BITPIX", 8, "Bits per Pixel");
		phdr.addValue("EXTEND",true," Fits dataset may contain extensions");
		
		phdr.addValue("NEXTEND",nToDo," Number of extentions");
		} catch (HeaderCardException e) {
			
			e.printStackTrace();
		}
		fitsWriter.put(phdr);
		int hduToKeep = 0;
		LinkedList<Object> objects;
		do {
			objects = ir.get(inputFileName, hduToKeep);
			if (objects.size() == 1) {
				fitsWriter.put(objects.get(0));
			}
			hduToKeep++;
		} while (objects.size() > 0);
		

		//fitsWriter.makeCheckSum();
		fitsWriter.close();
		
	}


	/**
	 * put a PixeledImage or RowListStarTable in the file.
	 * @param object
	 */
	public void put(Object object) {
		put(object, FitsDataFormat.FLOAT);
	}
	/**
	 * put a PixeledImage or RowListStarTable in the file.
	 * @param object
	 * @param fitsDataFormat used when the object is a PixeledImage
	 */
	public void put(Object object, FitsDataFormat fitsDataFormat) {
		if (object instanceof Header) {
			put((Header) object);
		} else if (object instanceof PixeledMask) {
			put((PixeledMask) object);			
		} else if (object instanceof PixeledImage) {
			put((PixeledImage) object, fitsDataFormat);
		} else if (object instanceof StarTable) {
			put((StarTable) object);
		} else if (object instanceof JDBCStarTable) {
			JDBCStarTable jdbcStarTable = (JDBCStarTable) object;
			RowListStarTable rlst = new RowListStarTable((JDBCStarTable) object);
			try {
				RowSequence rowSequence = jdbcStarTable.getRowSequence();
				while (rowSequence.next()) {
					rlst.addRow(rowSequence.getRow());
				}
			} catch (IOException e) {
				throw new Error(e);
			}
			put(rlst);
		} else if (object instanceof Iterable) {
			for (Object thisOne : (Iterable<?>) object) {
				put(thisOne, fitsDataFormat);
			}
		} else if (object instanceof Object[]) {
			Object[] things = (Object[]) object;
			for (Object thisOne : things) {
				put(thisOne);
			}
		} else {
			throw new Error("can only put PixeledImage or RowListStarTable or Iterable but this is "+object.getClass());
		}
	}
	
	// Return a Header suitable for nom.tam.fits.Fits()
	/**
	 * generates a header suitable for nom.tam.fits.Fits
	 * @param pi image that will be written
	 * @param fitsDataFormat number of bits to use
	 * @return the filled-in header
	 */
	public static Header getHeader(PixeledImage pi, FitsDataFormat fitsDataFormat) {
		Header retval = new Header();
		try {
			retval.addValue("BITPIX", fitsDataFormat.getBitpix(), "");
			retval.addValue("NAXIS", 2, "Dimensionality");
			retval.addValue("NAXIS1", pi.getColumnCount(), "number of columns");
			retval.addValue("NAXIS2", pi.getRowCount(), "number of rows");
			retval.addValue("BZERO", fitsDataFormat.getBzero(), "");
			retval.addValue("BSCALE", fitsDataFormat.getBscale(), "");
		} catch (HeaderCardException e) {
			throw new Error(e);
		}
		Iterator<?> iterator = pi.getHeader().iterator();
		while (iterator.hasNext()) {
			HeaderCard headerCard = (HeaderCard) iterator.next();
			if (headerCard != null) {
				String key = headerCard.getKey();
				if (key != null) {
					if (key.equalsIgnoreCase("CRPIX1")) {
						double crpix1 = Double.parseDouble(headerCard.getValue());
						crpix1 += pi.getCol0();
						headerCard.setValue(String.format("%f",crpix1));
					}
					if (headerCard.getKey().equalsIgnoreCase("CRPIX2")) {
						double crpix2 = Double.parseDouble(headerCard.getValue());
						crpix2 += pi.getRow0();
						headerCard.setValue(String.format("%f",crpix2));
					}
					retval.addLine(headerCard);
				}
			}
		}
		
		
		try {
			retval.addValue("PIROW0", pi.getRow0(), "row0 for gov.fnal.eag.sim.image.PixeledImage");
			retval.addValue("PICOL0", pi.getCol0(), "col0 for gov.fnal.eag.sim.image.PixeledImage");
		} catch (HeaderCardException e) {
			throw new Error(e);
		}
		
		return retval;
	}
    /**
     * write the header to the stream
     * @param header
     */
    public void put(Header header) {
        if (isFirst) {
            BasicHDU dummyHDU = BasicHDU.getDummyHDU();
            try {
                dummyHDU.setPrimaryHDU(isFirst);
            } catch (FitsException e) {
                throw new Error(e);
            }
            Iterator<?> it = header.iterator();
            Header phdr = dummyHDU.getHeader();
            while (it.hasNext()) {
            	
 
                	
                	HeaderCard line = (HeaderCard) it.next();
                	System.out.println("card: "+line.toString());
                    phdr.addLine(line);
                
            }
            try {
            	
              //  try {
              //      Fits.setChecksum(dummyHDU);
              //      phdr.addValue("CHECKVER",Fits.version(), "Checksum software version");
              //  } catch (IOException e) {
              //      throw new Error(e);
              //  }
                dummyHDU.write(bf);
            } catch (FitsException e) {
                throw new Error(e);
            }
            isFirst = false;
        } 
    }
}
