package gov.fnal.eag.sim.repository.image;

import gov.fnal.eag.sim.pixel.LookupTable;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.PriorityQueue;

import org.freehep.util.io.BitInputStream;

/**
 * tree for encoding and decoding using Huffman, where the values range from 0 to (2^nBits)-1 inclusive.
 * @author stoughto
 *
 */
public class HuffmanTree {
	HuffmanNode root;
	private static int[] bitMasks;
	private static final int nBitsPutMax = 16;
	static {
		bitMasks = new int[nBitsPutMax+1];
		for (int iBit=0; iBit<nBitsPutMax+1; iBit++) {
			bitMasks[iBit] = 1 << iBit;
		}
	}

	/**
	 * @param frequencies the number of times value i occurs in the input stream
	 */
	public HuffmanTree(int[] frequencies) {
		init(frequencies);
	}
	private void init(int[] frequencies) {
		if (frequencies.length > Math.pow(2, nBitsPutMax)) {
			throw new IllegalArgumentException("frequencies is too long:  frequencies.length="+frequencies.length+" must be <= "+Math.pow(2,16));
		}
		PriorityQueue<HuffmanNode> nodes = new PriorityQueue<HuffmanNode>(frequencies.length);
		for (int value=0; value<frequencies.length; value++) {
			if (frequencies[value] > 0) {
				nodes.add(new HuffmanNode(frequencies[value], value));
			}
		}
		while (nodes.size() > 1) {
			HuffmanNode left = nodes.poll();
			HuffmanNode right = nodes.poll();
			HuffmanNode node = new HuffmanNode(left, right);
			nodes.add(node);
		}
		root = nodes.poll();
	}

	/**
	 * creates a HuffmanTree equivalent to the one that made the encoding with the getEncoding method
	 * @param encoding
	 */
	public HuffmanTree(ArrayList<Boolean> encoding) {
		int nBitsEncoding = getBits(encoding, 0, 6);
		ArrayList<HuffmanNode> nodes = new ArrayList<HuffmanNode>();
		//System.out.println("HuffmanTree:  nBitsEncoding="+nBitsEncoding);
		{
			int index = 6;
			while (index < encoding.size()) {
				boolean hasValue = encoding.get(index++);
				HuffmanNode node;
				if (hasValue) {
					int value = getBits(encoding, index, nBitsEncoding);
					index+= nBitsEncoding;
					//System.out.println("index="+index+"   value="+value);
					node = new HuffmanNode(0, value);
				} else {
					//System.out.println("index="+index+"   node");
					node = null;
				}
				nodes.add(node);
			}
		}
		//System.out.println("nodes.size="+nodes.size());
		root = null;
		while (nodes.size() > 1) {
			//System.out.println("begin while with nodes.size="+nodes.size());
			int index = 0;
			//System.out.println("begin with nodes.size()="+nodes.size());
			while (index < nodes.size()-2) {
				//System.out.println(" hello with index="+index+" and nodes.size()="+nodes.size());
				HuffmanNode indexNode = nodes.get(index);
				if ((indexNode == null) && (nodes.get(index+1) != null && nodes.get(index+2) != null)) {
					HuffmanNode node1 = nodes.remove(index+1);
					HuffmanNode node2 = nodes.remove(index+1);
					//System.out.println("combine:  node1="+node1+" node2="+node2);
					HuffmanNode node12 = new HuffmanNode(node1, node2);
					nodes.remove(index);
					nodes.add(index, node12);
					index--;
				} else {
					index++;
				}

			}
		}
		//System.out.println("done with nodes.size()="+nodes.size());
		root = nodes.get(0);
	}


	/**
	 * creates a HuffmanTree based on the frequencies of the values in pix compressed by lookupTable 
	 * 
	 * @param lookupTable
	 * @param pix
	 * @throws Error if any of the compressed values are < 0
	 */
	public HuffmanTree(LookupTable lookupTable, double[][] pix) {
		int cMin = Integer.MAX_VALUE;
		int cMax = Integer.MIN_VALUE;
		for (int row=0; row<pix.length; row++) {
			for (int col=0; col<pix[0].length; col++) {
				double value = pix[row][col];
				int compressed = lookupTable.compress(value);
				cMin = Math.min(cMin, compressed);
				cMax = Math.max(cMax, compressed);
			}
		}
		if (cMin < 0) {
			throw new Error("cMin="+cMin);
		}
		int[] frequencies = new int[cMax+1];
		for (int row=0; row<pix.length; row++) {
			for (int col=0; col<pix[0].length; col++) {
				double value = pix[row][col];
				int compressed = lookupTable.compress(value);
				frequencies[compressed]++;
			}
		}		
		init(frequencies);
	}

	private int getBits(ArrayList<Boolean> encoding, int index, int nBitsEncoding) {
		int retval = 0;
		for (int i=index; i<index+nBitsEncoding; i++) {
			retval <<= 1;
			if (encoding.get(i)) {
				retval++;
			}
		}
		return retval;
	}
	/**
	 * reads in the next set of bits from the stream and look up the value
	 * @param stream
	 * @return the value
	 */
	public int getValue(BitInputStream stream) {
		HuffmanNode current = root;
		while (current.left != null || current.right != null) {
			boolean bit;
			try {
				bit = stream.readBitFlag();
			} catch (IOException e) {
				throw new Error(e);
			}
			if (bit) {
				current = current.right;
			} else {
				current = current.left;
			}
		}
		return current.value;

	}

	/**
	 * @return a map that returns a String representation of the bits that encode a value
	 */
	public HashMap<Integer, String> getMap() {
		HashMap<Integer, String> map = new HashMap<Integer, String>();
		return getMap(map, root, "");
	}
	private HashMap<Integer, String> getMap(HashMap<Integer, String> map, HuffmanNode head, String code) {
		if (head.left == null && head.right == null) {
			map.put(head.value, code);
		} else {
			getMap(map, head.left, code+"0");
			getMap(map, head.right, code+"1");
		}
		return map;
	}

	/**
	 * @return the maximum number of bits required to encode the lookup table
	 */
	public int getNBitsEncoding() {
		HashMap<Integer, String> map = getMap();
		int nBitsEncoding = 0;
		for (Integer value : map.keySet()) {
			nBitsEncoding = Math.max(nBitsEncoding, Integer.toBinaryString(value).length()+1);
		}
		return nBitsEncoding;
	}
	/**
	 * @return the bits that encode this HuffmanTree
	 */
	public ArrayList<Boolean> getEncoding() {	
		final int nBitsEncoding = getNBitsEncoding();
		final ArrayList<Boolean> encoding = new ArrayList<Boolean>();
		putBits(encoding, nBitsEncoding, 6);
		getEncoding(encoding, root, nBitsEncoding);
		return encoding;
	}

	private void putBits(ArrayList<Boolean> encoding, int value, int nBitsEncoding) {
		for (int iBit=nBitsEncoding-1; iBit>=0; iBit--) {
			boolean bitOn = (value&bitMasks[iBit]) != 0;
			encoding.add(bitOn);
		}		
	}

	private void getEncoding(ArrayList<Boolean> encoding, HuffmanNode node, int nBitsEncoding) {
		if (node.left == null && node.right == null) {
			// leaf
			encoding.add(true);
			putBits(encoding, node.value, nBitsEncoding);
		} else {
			// node
			encoding.add(false);
			getEncoding(encoding, node.left, nBitsEncoding);
			getEncoding(encoding, node.right, nBitsEncoding);
		}
	}

}
