package gov.fnal.eag.sim.repository.image;

import gov.fnal.eag.sim.jzlib.JZlib;
import gov.fnal.eag.sim.jzlib.ZInputStream;
import gov.fnal.eag.sim.jzlib.ZOutputStream;
import gov.fnal.eag.sim.jzlib.ZStreamException;
import gov.fnal.eag.sim.pixel.LookupTable;
import gov.fnal.eag.sim.pixel.PixeledImage;
import gov.fnal.eag.util.misc.CommandLine;

import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StreamCorruptedException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;
import java.util.zip.ZipInputStream;

import javax.imageio.ImageIO;

import nom.tam.fits.BasicHDU;
import nom.tam.fits.Fits;
import nom.tam.fits.FitsException;
import nom.tam.fits.Header;
import nom.tam.fits.HeaderCard;
import nom.tam.fits.ImageData;
import nom.tam.fits.ImageHDU;
import nom.tam.fits.TableHDU;
import nom.tam.image.ImageTiler;
import nom.tam.util.BufferedDataOutputStream;

import org.freehep.util.io.BitInputStream;
import org.freehep.util.io.BitOutputStream;

import uk.ac.starlink.table.DefaultValueInfo;
import uk.ac.starlink.table.DescribedValue;
import uk.ac.starlink.table.RowListStarTable;
import uk.ac.starlink.table.ValueInfo;



/**
 * a place to save files, a directory on local disk
 */


public class ImageRepository {
	/**
	 * the types of lossless compression 
	 * @author stoughto
	 *
	 */
	public enum LosslessCompressionType {
		/**
		 * no compression
		 */
		NONE(null),
		/**
		 * use GZIPOutputStream and GZIPInputStream
		 */
		GZIP(null), 
		/**
		 * uses GZIPOutputStream
		 */
		ZLIB(null),
		/**
		 * uses CBZip2OutputStream
		 */
		BZIP2(null),

		/**
		 * 
		 */
		HUFFMAN(null);

		private final static String huffmanMagicString="CSHU";
		private static int huffmanMagicNumber;

		static {
			for (int i=0; i<4; i++) {
				huffmanMagicNumber <<= 8;
				char c = huffmanMagicString.charAt(i);
				huffmanMagicNumber += c;
			}
		}
		private HuffmanTree huffmanTree = null;
		LosslessCompressionType(HuffmanTree huffmanTree) {
			this.huffmanTree = huffmanTree;
		}

		/**
		 * @return the tree used for Huffman encoding; null otherwise
		 */
		public HuffmanTree getHuffmanTree() {
			return huffmanTree;
		}

		/**
		 * sets parameters for Huffman encoding
		 * @param huffmanTree
		 */
		public void setHuffman(HuffmanTree huffmanTree) {
			if (this.compareTo(LosslessCompressionType.HUFFMAN) != 0) {
				throw new IllegalArgumentException("can only call setHuffman on HUFFMAN but this="+this);
			}
			this.huffmanTree = huffmanTree;
		}
		/**
		 * @return the magic number written to the beginning of files
		 */
		public static int getHuffmanMagicNumber() {
			return huffmanMagicNumber;
		}
	}
	//private static Logger logger = Logger.getLogger("gov.fnal.eag.sim");
	/**
	 * String that points to the directory on local disk
	 */
	protected String repositoryName;
	/**
	 * Remember the name if this repository was destroyed
	 */
	private String destroyedRepositoryName; 

	/**
	 * defaults to construction an image repository called imageRepository in the user's home directory
	 */
	public ImageRepository() {
		//		this((new File(System.getProperty("user.home"), "ImageRepository")).getAbsolutePath());
		this("default");
		//this((new File("/tmp/", "ImageRepository"+"."+System.getProperty("user.name"))).getAbsolutePath());
	}
	/**
	 * Create a new image repository if the directory exists or can be created
	 * @param repositoryName String that is a directory name on local disk
	 */

	public ImageRepository(String repositoryName) {

		if (repositoryName == null) {
			throw new Error("repository name is null");
		}
		File td;
		if (repositoryName.equalsIgnoreCase("default")) {
			td = new File("/tmp/", "ImageRepository"+"."+System.getProperty("user.name"));
			this.repositoryName = td.getAbsolutePath();
		} else {
			td = new File(repositoryName);
			this.repositoryName = repositoryName;
		}
		if (!td.isDirectory()) {
			if (!td.mkdirs()) {
				throw new Error ("ImageRepository:  can not create directory "+repositoryName);
			}
		}
		if (!td.canRead()) {
			throw new Error ("ImageRepository:  can not read directory "+repositoryName);
		}
		if (!td.canWrite()) {
			throw new Error ("ImageRepository:  can not write in directory "+repositoryName);
		}
	}


	/**
	 * delete all files from this image repository (not recursive)
	 */
	public void destroy() {
		if (repositoryName == null) throw new Error("this repository has been destroyed:  "+destroyedRepositoryName);
		File root = new File(repositoryName);
		destroyedRepositoryName = repositoryName;
		repositoryName = null;
		File[] files = root.listFiles();	
		for (int i=0; i<files.length; i++) {
			if (!files[i].delete()) {
				throw new Error("can not delete "+files[i].getAbsolutePath());
			}
		}
		if (!root.delete()) {
			throw new Error("can not delete "+root.getAbsolutePath());
		}
	}
	/**
	 * count the number of files in this repository
	 * @return HashMap with the keys nFiles and nBytes
	 */
	public HashMap<String,Integer> inventory()  {
		if (repositoryName == null) throw new Error("this repository has been destroyed:  "+destroyedRepositoryName);
		HashMap<String,Integer> retval = new HashMap<String,Integer>();
		int nFiles = 0;
		int nBytes = 0;
		File[] files = (new File(repositoryName)).listFiles();
		for (int i=0; i<files.length; i++) {
			nFiles++;
			nBytes += files[i].length();
			//if (i == 0) {
			//	logger.log(Level.INFO, repositoryName+" file inventory");
			//}
			//String line = String.format("    %15d %s", files[i].length(), files[i].getName());			
			//logger.log(Level.INFO, line);
		}
		//logger.log(Level.INFO, (String.format("%30s   nFiles=%4d   nBytes=%d", repositoryName,nFiles,nBytes)));
		retval.put("nFiles", new Integer(nFiles));		
		retval.put("nBytes", new Integer(nBytes));
		return retval;
	}
	/**
	 * @return the file names in this image repository
	 */
	public ArrayList<String> getFileNames() {
		ArrayList<String> retval = new ArrayList<String>();
		File[] files = (new File(repositoryName)).listFiles();
		for (File file : files) {
			retval.add(file.getName());
		}
		Collections.sort(retval);
		return retval;
	}
	/**
	 * tests to see if the file exists
	 * @param fileName
	 * @return true if it exists
	 */
	public boolean exists(String fileName) {
		File file = new File(repositoryName, fileName);
		//String fullFileName = file.getAbsolutePath();
		return file.exists();
	}


	/**
	 * flush and close
	 * @param stream
	 * @throws Error
	 */
	public static void flushAndClose(BufferedDataOutputStream stream) throws Error 
	{
		try {
			stream.flush();
			stream.close();
		} catch (IOException e) {
			throw new Error(e);
		}
	}
	/**
	 * open the stream 
	 * @param fileName
	 * @param overwrite
	 * @return the stream
	 * @throws Error
	 */
	public BufferedDataOutputStream getBufferedDataOutputStream(String fileName, boolean overwrite) throws Error {
		if (repositoryName == null)
			throw new Error("this repository has been destroyed:  "
					+ destroyedRepositoryName);
		File file = new File(repositoryName, fileName);
		String fullFileName = file.getAbsolutePath();
		if (file.exists() && !overwrite) {
			throw new Error(fullFileName + " already exists");
		}
		if (file.exists() && !file.canWrite()) {
			throw new Error(fullFileName + " can not write");
		}
		BufferedDataOutputStream stream;
		try {
			stream = new BufferedDataOutputStream(new FileOutputStream(file));
		} catch (FileNotFoundException e) {
			throw new Error(e);
		}
		return stream;
	}
	/**
	 * read everything from the file using the nom.tam.fits library
	 * 
	 * @param databaseUrl
	 *            file to read
	 * @return a LinkedList of RowListStarTable and/or PixeledImage objects
	 */
	public LinkedList<Object> get(String databaseUrl) {
		return get(databaseUrl, -1);
	}

	/**
	 * read a FITS file
	 * @param fileName name of the file in this ImageRepository
	 * @param hduToKeep the hdu to read
	 * @return a linked list with one Object on it, filled from the hduToKeep
	 */
	public LinkedList<Object> get(String fileName, int hduToKeep) {
		try {
			if (repositoryName == null) throw new Error("this repository has been destroyed:  "+destroyedRepositoryName);
			String fullFileName = getFullFileName(fileName);
			InputStream inputStream = getInputStream(fullFileName);
			LinkedList<Object> retval = read(inputStream, hduToKeep);
			return retval;
		} catch (Exception e) {
			throw new Error(e);
		}
	}
	/**
	 * @param fileName
	 * @return the full absolute name of this file
	 */
	public String getFullFileName(String fileName) {
		File file = new File(repositoryName, fileName);
		String fullFileName = file.getAbsolutePath();
		return fullFileName;
	}
	/**
	 * @param inputStream
	 * @return the objects read from the inputStream
	 */
	static public LinkedList<Object> read(InputStream inputStream)  {
		return read(inputStream, -1);
	}

	/**
	 * read one FITS object from the stream
	 * @param inputStream where to read
	 * @param hduToRead which hdu to keep
	 * @return a linked list with the one object on it
	 */
	static public LinkedList<Object> read(InputStream inputStream, int hduToRead) {
		Header primaryHeader = null;
		LinkedList<Object> retVal = new LinkedList<Object>();
		Fits f;
		try {
			f = new Fits(inputStream);
		} catch (FitsException e1) {
			throw new IllegalArgumentException("trouble with inputStream="+inputStream, e1);
		}
		BasicHDU bh = null;
		int iHdu = 0;
		while (true) {
			try {
				//  Sorry, this is the only way to see if we are done!
				bh = f.readHDU();
			} catch (java.io.IOException e) {
				break;
			} catch (FitsException e) {
				throw new Error(" In ImageRepository:  FitsException caught: ",e);
			}
			if (bh == null) break;
			if (primaryHeader == null) {
				primaryHeader = bh.getHeader();
			}
			if (bh instanceof nom.tam.fits.ImageHDU) {
				Header header = bh.getHeader();
				if (header.containsKey("NAXIS1") && header.containsKey("NAXIS2")) {
					if (hduToRead < 0 || hduToRead == iHdu) {
						int ndim1 = header.getIntValue("NAXIS1");
						int ndim2 = header.getIntValue("NAXIS2");
						double[][] image = new double [ndim2][ndim1];
						ImageData imD  =  new ImageData(bh.getKernel());
						Object object = imD.getData();
						if (object instanceof double[][]) {
							image = (double[][]) object;            				
						} else if (object instanceof float[][]) {
							float[][] temp = (float[][]) object;
							image = new double[temp.length][temp[0].length];
							for (int i=0; i<temp.length; i++) {
								for (int j=0; j<temp[0].length; j++) {
									image[i][j] = temp[i][j];
								}
							}
						} else if (object instanceof short[][]) {
							short[][] temp = (short[][]) object;
							image = new double[temp.length][temp[0].length];
							for (int i=0; i<temp.length; i++) {
								for (int j=0; j<temp[0].length; j++) {
									//short s = temp[i][j];
									//System.out.println(String.format("s=%+6d %x", s, s));
									//int x = y & 0xFFFF; 
									image[i][j] = temp[i][j]+32768;
								}
							}
						} else if (object instanceof int[][]) {
							int[][] temp = (int[][]) object;
							image = new double[temp.length][temp[0].length];
							for (int i=0; i<temp.length; i++) {
								for (int j=0; j<temp[0].length; j++) {
									image[i][j] = temp[i][j];
								}
							}

						} else {
							throw new Error ("implement conversion for class="+object.getClass());
						}
						PixeledImage pi = new PixeledImage(image);
						{
							//int numberOfCards = header.getNumberOfCards();
							//System.out.println("ImageRepository:  numberOfCards="+numberOfCards);
							Iterator<?> it = header.iterator();
							int row0 = 0;
							int col0 = 0;
							while (it.hasNext()) {
								HeaderCard hc = (HeaderCard) it.next();
								if (!hc.toString().startsWith("END                                              ")) {
									Header h = pi.header;
									h.addLine(hc);
									if (hc.isKeyValuePair()) {
										if (hc.getKey().equalsIgnoreCase("PIROW0")) {
											try {
												row0 = Integer.parseInt(hc.getValue());
											} catch (NumberFormatException e) {
											}
										}
										if (hc.getKey().equalsIgnoreCase("PICOL0")) {
											try {
												col0 = Integer.parseInt(hc.getValue());
											} catch (NumberFormatException e) {
											}           							
										}
									}
								}
							}
							pi.setRow0Col0(row0, col0);
						}
						// Now delete the FITS formatting header cards from the PI  header
						{
							String[] keys = {"SIMPLE", "BITPIX", "NAXIS", "NAXIS1", "NAXIS2", "EXTEND", "XTENSION", "PCOUNT", "GCOUNT", "END"};
							for (String key : keys) {
								pi.header.deleteKey(key);
							}
							// We found an IMAGE so add it to the list of things read
							retVal.add(pi);
							if (hduToRead >= 0) return retVal;
						}
					}
				} else {
					// this is not really an image HDU, just the primary header, so nothing will be added to retval.
					iHdu--;
				}
			} else
				if (bh instanceof nom.tam.fits.BinaryTableHDU) {
					if (hduToRead < 0 || hduToRead == iHdu) {
						FitsStarTableEag fst = new FitsStarTableEag((TableHDU) bh);
						RowListStarTable rlst = new RowListStarTable(fst);
						for (long row=0; row<fst.getRowCount(); row++) {
							try {
								rlst.addRow(fst.getRow(row));
							} catch (IOException e) {
								throw new Error(e);
							}
						}
						Iterator<?> it = primaryHeader.iterator();
						while (it.hasNext()) {
							HeaderCard headerCard = (HeaderCard) it.next();
							String key = headerCard.getKey();
							if (doNotIgnore(key)) {
								String valueString = headerCard.getValue();
								ValueInfo valueInfo;
								Object value;
								if (headerCard.isStringValue()) {
									valueInfo = new DefaultValueInfo(key, String.class);
									value = valueString;
								} else if (valueString.equals("T") || valueString.equals("F")) {
									int booleanValue = "FT".indexOf(valueString);
									valueInfo = new DefaultValueInfo(key, Integer.class);
									value = new Integer(booleanValue);
								} else {
									valueInfo = new DefaultValueInfo(key, Double.class);
									value = new Double(valueString);
								}
								DescribedValue describedValue = new DescribedValue(valueInfo, value);			        	
								rlst.setParameter(describedValue);
							}
						}            
						// We found a BINARY TABLE so add it to the list of things read
						retVal.add(rlst);
						if (hduToRead >= 0) return retVal;
					}
				} else {
					throw new Error("can not convert from "+bh.getClass());
				}
			iHdu++;
		}
		return retVal;
	}

	private static boolean doNotIgnore(String key) {
		String[] ignoreds = {"SIMPLE", "BITPIX", "EXTEND", "COMMENT", "END", "NAXIS", "HISTORY"};
		for (String ignored : ignoreds) { 	
			if (key.toUpperCase().startsWith(ignored)) return false;
		}
		return true;
	}

	/**
	 * read from the inputStream and make a new table based on the contents of the hdu
	 * @param inputStream where to read
	 * @param hdu where 0 is the first one
	 * @return the table
	 */
	public static RowListStarTable readRlst(InputStream inputStream, int hdu)  {
		LinkedList<Object> list = read(inputStream, hdu);
		if (list.size() == 0) {
			throw new Error("list size is zero");
		}
		RowListStarTable rlst = (RowListStarTable) list.get(0);
		return rlst;
	}
	/**
	 * read from the inputStream and make a new table based on the contents of the hdu
	 * @param inputFileName where to read
	 * @param hdu where 0 is the first one
	 * @return the table
	 */
	public static RowListStarTable readRlst(String inputFileName, int hdu)  {
		InputStream inputStream;
		try {
			inputStream = new FileInputStream(inputFileName);
		} catch (FileNotFoundException e) {
			throw new Error(e);
		}
		LinkedList<Object> list = read(inputStream, hdu);
		RowListStarTable rlst = (RowListStarTable) list.get(0);
		try {
			inputStream.close();
		} catch (IOException e) {
			throw new Error(e);
		}
		return rlst;
	}

	/**
	 * @return the directory where this repository is located
	 */
	public String getRepositoryName() {
		return repositoryName;
	}
	/**
	 * read a PixeledImage
	 * @param inputStream data source
	 * @return what was read at the 0th hdu
	 */
	public static PixeledImage readPixeledImage(InputStream inputStream) {
		return readPixeledImage(inputStream, 0);
	}

	/**
	 * read a pixeled image
	 * @param inputStream
	 * @param hdu
	 * @return what was read from the hdu
	 */
	public static PixeledImage readPixeledImage(InputStream inputStream, int hdu) {
		LinkedList<Object> list = read(inputStream, hdu);
		if (list.size() == 1) {
			PixeledImage pi = (PixeledImage) list.get(0);
			return pi;
		}
		return null;
	}

	/**
	 * @param fullFileName
	 * @return read from fullFileName
	 */
	public static PixeledImage readPixeledImage(String fullFileName) {
		return readPixeledImage(getInputStream(fullFileName));
	}

	/**
	 * @param fullFileName
	 * @param hdu which hdu to read
	 * @return read from fullFileName
	 */
	public static PixeledImage readPixeledImage(String fullFileName, int hdu) {
		return readPixeledImage(getInputStream(fullFileName), hdu);
	}

	/**
	 * read a pixeled image
	 * @param fileName data source
	 * @return what was read from hdu 0
	 */
	public PixeledImage getPixeledImage(String fileName) {
		return getPixeledImage(fileName, 0);
	}
	/**
	 * read a PixeledImage
	 * @param fileName data source
	 * @param i
	 * @return what was read from hdu i
	 */
	public PixeledImage getPixeledImage(String fileName, int i) {
		InputStream inputStream = getInputStream(getFullFileName(fileName));
		return readPixeledImage(inputStream, i);
	}

	/**
	 * get an InputStream, taking into account .gz or .ZIP suffix
	 * @param fullFileName
	 * @return a FileInputStrem, GZIPInputStream, or ZipInputStream
	 */
	public static InputStream getInputStream(String fullFileName) {
		InputStream fis;
		try {
			fis = new FileInputStream(new File(fullFileName));
			if (fullFileName.endsWith(".gz")) {
				try {
					fis = new GZIPInputStream(fis);
				} catch (IOException e) {
					throw new Error(e);
				}
			} else if (fullFileName.endsWith(".ZIP")) {
				fis = new ZipInputStream(fis);
			}
		} catch (FileNotFoundException e) {
			throw new IllegalArgumentException(" can not get FileInputStream for fullFileName="+fullFileName, e);
		}
		return fis;
	}
	/**
	 * @param fileName where to read
	 * @return the reader
	 */
	public BufferedReader getBufferedReader(String fileName) {
		File file = new File(repositoryName, fileName);
		try {
			BufferedReader retval = new BufferedReader(new FileReader(file));
			return retval;
		} catch (FileNotFoundException e) {
			throw new Error(e);
		}

	}
	/**
	 * write a pixeled image
	 * @param fileName where to write
	 * @param pi what to write
	 */
	public static void write(String fileName, PixeledImage pi) {
		write(fileName, pi, FitsDataFormat.DOUBLE, true);
	}
	/**
	 * write a PixeledImage to the file
	 * @param fileName
	 * @param pi the PixeledImage to write
	 * @param fitsDataFormat 
	 * @param overwrite
	 */
	public static void write(String fileName, PixeledImage pi, FitsDataFormat fitsDataFormat, boolean overwrite) {
		try {
			FitsWriter fitsWriter = new FitsWriter(fileName, overwrite);
			fitsWriter.put(pi, fitsDataFormat);
			fitsWriter.close();
		} catch (IOException e) {
			throw new Error(e);
		}

	}
	/**
	 * write an object to the file
	 * @param fileName
	 * @param object the PixeledImage or RowListStarTable to write
	 * @param overwrite
	 */
	public static void write(String fileName, Object object, boolean overwrite) {
		try {
			FitsWriter fitsWriter = new FitsWriter(fileName, overwrite);
			fitsWriter.put(object);
			fitsWriter.close();
		} catch (IOException e) {
			throw new Error(e);
		}

	}

	/**
	 * write the things to the file
	 * @param fileName
	 * @param things
	 * @param fitsDataFormat
	 * @param overwrite
	 */
	public static void write(String fileName, Iterable<Object> things, FitsDataFormat fitsDataFormat, boolean overwrite) {
		try {
			FitsWriter fitsWriter = new FitsWriter(fileName, overwrite);
			for (Object thing : things) {
				fitsWriter.put(thing, fitsDataFormat);
			}
			fitsWriter.close();
		} catch (IOException e) {
			throw new Error(e);
		}
	}
	/**
	 * write a PixeledImage to the file in this ImageRepository
	 * @param fileName
	 * @param pi the PixeledImage to write
	 * @param fitsDataFormat data format
	 * @param overwrite
	 */

	public void put(String fileName, PixeledImage pi, FitsDataFormat fitsDataFormat, boolean overwrite) {
		write( (new File(repositoryName, fileName)).getAbsolutePath(), pi, fitsDataFormat, overwrite);
	}
	/**
	 * write a PixeledImage to the file in this ImageRepository
	 * @param fileName
	 * @param object the PixeledImage or RowListStarTable to write
	 * @param overwrite
	 */

	public void put(String fileName, Object object, boolean overwrite) {
		write( (new File(repositoryName, fileName)).getAbsolutePath(), object, overwrite);
	}
	/**
	 * write a PixeledImage to the file in this ImageRepository, overwrite existing file
	 * @param fileName
	 * @param object the PixeledImage or RowListStarTable to write
	 */

	public void put(String fileName, Object object) {
		write( (new File(repositoryName, fileName)).getAbsolutePath(), object, true);
	}

	/**
	 * put the thing in the file
	 * @param fileName the file to write
	 * @param things contains PixeledImage and RowListStarTable objects
	 * @param fitsDataFormat used for PixeledImage
	 * @param overwrite
	 */
	public void put(String fileName, Iterable<Object> things, FitsDataFormat fitsDataFormat, boolean overwrite) {
		write((new File(repositoryName, fileName)).getAbsolutePath(), things, fitsDataFormat, overwrite);		
	}

	/**
	 * @param fileName
	 * @param overwrite
	 * @return the writer
	 */
	public FitsWriter getFitsWriter(String fileName, boolean overwrite) {
		FitsWriter fitsWriter;
		try {
			fitsWriter = new FitsWriter((new File(repositoryName, fileName)).getAbsolutePath(), overwrite);
		} catch (IOException e) {
			throw new Error(e);
		}
		return fitsWriter;
	}

	/**
	 * run fitsverify on the file in this ImageRepository
	 * @param fileName
	 * @return true if fitsverify is happy
	 */
	public boolean fitsverify(String fileName) {
		String fullFileName = (new File(repositoryName, fileName)).getAbsolutePath();
		int valid = CommandLine.run("fitsverify "+fullFileName);
		return (valid==0);
	}
	/**
	 * get a RowListStarTable
	 * @param fileName file in the ImageRepository
	 * @param hdu
	 * @return the table
	 */
	public RowListStarTable getRlst(String fileName, int hdu) {
		String fullFileName = getFullFileName(fileName);
		InputStream inputStream = getInputStream(fullFileName);
		return readRlst(inputStream, hdu);
	}
	/**
	 * @param fileName where to write
	 * @param bufferedImage what to write as a png
	 */
	public static void toPng(String fileName, BufferedImage bufferedImage) {
		File file = new File(fileName);
		try {
			ImageIO.write(bufferedImage, "png", file);
		} catch (IOException e) {
			throw new Error(e);
		}

	}

	/**
	 * read sub-image from the file
	 * @param fileName
	 * @param nRow number of rows to read
	 * @param nCol number of columns to read
	 * @param row0 initial row 
	 * @param col0 initial column
	 * @return the image
	 */
	public static PixeledImage getPixeledImage(String fileName, int nRow, int nCol, int row0, int col0) {

		Fits f;
		try {
			f = new Fits(fileName);
		} catch (FitsException e) {
			throw new Error("trouble opening fileName="+fileName, e);
		}
		ImageHDU h;
		Object img;
		try {
			h = (ImageHDU) f.readHDU();
			ImageTiler tiler = h.getTiler();
			int nAxis1 = h.getHeader().getIntValue("NAXIS1");
			int nAxis2 = h.getHeader().getIntValue("NAXIS2");
			nCol = Math.min(nCol, nAxis1);
			nRow = Math.min(nRow, nAxis2);
			img= tiler.getTile(new int[]{row0, col0}, new int[] {nRow, nCol});
		} catch (FitsException e) {
			throw new Error(e);
		} catch (IOException e) {
			throw new Error(e);
		}
		double[][] pix = new double[nRow][nCol];
		if (img instanceof double[]) {
			double[] array = (double[]) img;
			for (int r=0; r<nRow; r++) {
				for (int c=0; c<nCol; c++) {
					double value = array[r*nCol + c];
					pix[r][c] =  value;
				}
			}
		} else if (img instanceof float[]) {
			float[] array = (float[]) img;
			for (int r=0; r<nRow; r++) {
				for (int c=0; c<nCol; c++) {
					double value = array[r*nCol + c];
					pix[r][c] =  value;
				}
			}
		} else if (img instanceof short[]) {
			double bzero = h.getBZero();
			double bscale = h.getBScale();
			short[] array = (short[]) img;
			for (int r=0; r<nRow; r++) {
				for (int c=0; c<nCol; c++) {
					double value = array[r*nCol + c];
					pix[r][c] =  bscale*(value + bzero);
				}
			}
		} else {
			throw new Error("can not deal with img="+img);
		}
		PixeledImage pi = new PixeledImage(pix);
		pi.setRow0Col0(row0, col0);
		return pi;		
	}
	/**
	 * @param fileName where to read the sub image
	 * @param nRow size of sub image
	 * @param nCol size of sub image
	 * @param row0 starting row in full image
	 * @param col0 starting column in full image
	 * @return the sub image
	 */

	public  PixeledImage readPixeledImage(String fileName, int nRow, int nCol, int row0, int col0) {

		return getPixeledImage(getFullFileName(fileName), nRow, nCol, row0, col0);
	}

	/**
	 * compress and write
	 * @param fileName
	 * @param lookupTable
	 * @param losslessCompressionType
	 * @param pixels
	 * @throws IOException
	 */
	public void writeCompressed(String fileName, LookupTable lookupTable, LosslessCompressionType losslessCompressionType, double[][] pixels) throws IOException {
		FileOutputStream fileOutputStream = new FileOutputStream(getFullFileName(fileName));
		
		// Initialize these variables assuming NONE compression
		OutputStream outputStream = fileOutputStream;
		BitOutputStream bitOutputStream = new BitOutputStream(outputStream);
		switch (losslessCompressionType) {

		case GZIP: 
			outputStream = new GZIPOutputStream(fileOutputStream);
			bitOutputStream = new BitOutputStream(outputStream);
			break;

		case ZLIB:
			outputStream = new ZOutputStream(fileOutputStream, JZlib.Z_BEST_COMPRESSION);
			bitOutputStream = new BitOutputStream(outputStream);
			break;

		case NONE:
			//outputStream = fileOutputStream;
			//bitOutputStream = new BitOutputStream(outputStream);
			break;

		case BZIP2:
			outputStream = new CBZip2OutputStream(fileOutputStream);
			bitOutputStream = new BitOutputStream(outputStream);
			break;

		case HUFFMAN:
			outputStream = new HuffmanOutputStream(fileOutputStream, losslessCompressionType);
			break;
		}
		for (int n : new int[] {pixels.length, pixels[0].length} ) {
			switch (losslessCompressionType) {
			case HUFFMAN:
				((HuffmanOutputStream) outputStream).writeInteger(n);				
				break;
			default:
				for (int iBit=16; iBit>0; iBit--) {
					boolean bit =  (n & (1<<iBit-1)) != 0;
					bitOutputStream.writeBitFlag(bit);
				}
			}
		}
		for (int i=0; i<pixels.length; i++) {
			for (int j=0; j<pixels[0].length; j++) {
				long value = lookupTable.compress(pixels[i][j]);
				switch (losslessCompressionType) {
				case HUFFMAN:
					((HuffmanOutputStream) outputStream).encode(value);
					break;
				default:
					bitOutputStream.writeUBits(value, lookupTable.getNBits());
				}
			}
		}
		
		switch (losslessCompressionType) {
		case HUFFMAN:
			outputStream.close();
			break;
		default:
			bitOutputStream.close();
		}
	}
	/**
	 * read and decompress
	 * @param fileName
	 * @param lookupTable
	 * @return the decompressed array
	 * @throws IOException
	 */
	public double[][] readCompressed(String fileName, LookupTable lookupTable) throws IOException {
		String fullFileName = getFullFileName(fileName);
		FileInputStream fileInputStream = new FileInputStream(getFullFileName(fileName));
		
		// initialize these assuming NONE compression
		InputStream inputStream = fileInputStream;
		BitInputStream bitInputStream = new BitInputStream(inputStream);

		LosslessCompressionType losslessCompressionType = detectLosslessCompressionType(fullFileName);
		switch (losslessCompressionType) {

		case NONE:
			//inputStream = fileInputStream;
			//bitInputStream = new BitInputStream(inputStream);
			break;

		case GZIP: 
			inputStream = new GZIPInputStream(fileInputStream);
			bitInputStream = new BitInputStream(inputStream);
			break;

		case ZLIB:
			inputStream = new ZInputStream(fileInputStream);
			bitInputStream = new BitInputStream(inputStream);
			break;

		case BZIP2:
			inputStream = new CBZip2InputStream(fileInputStream);
			bitInputStream = new BitInputStream(inputStream);
			break;

		case HUFFMAN:
			inputStream = new HuffmanInputStream(fileInputStream);
			break;

		}

		int[] temp = new int[2];
		for (int irc=0; irc<2; irc++) {
			switch (losslessCompressionType) {
			case HUFFMAN:
				temp[irc] = ((HuffmanInputStream) inputStream).readInteger();
				break;
			default:
				temp[irc] = 0;
			for (int iBit=16; iBit>0; iBit--) {
				temp[irc] <<= 1;
				boolean bit = bitInputStream.readBitFlag();
				if (bit) {
					temp[irc]++;
				}
			}
			}
		}


		int nrow = temp[0];
		int ncol = temp[1];
		if (nrow > 10000 || ncol > 10000) {
			throw new IllegalArgumentException("nrow="+nrow+" ncol="+ncol);
		}
		double[][] pixels = new double[nrow][ncol];
		for (int i=0; i<pixels.length; i++) {
			for (int j=0; j<pixels[0].length; j++) {
				long compressed;
				switch (losslessCompressionType) {
				case HUFFMAN:
					compressed = ((HuffmanInputStream) inputStream).read();
					break;
				default:
					compressed = bitInputStream.readUBits(lookupTable.getNBits());		
				}				
				double value = lookupTable.restore((int) compressed);
				pixels[i][j] = value;
			}
		}
		return pixels;
	}
	private LosslessCompressionType detectLosslessCompressionType(String fullFileName) throws FileNotFoundException {
		// Is this a GZIP file?
		try {
			FileInputStream fileInputStream =new FileInputStream(fullFileName);
			new GZIPInputStream(fileInputStream);
			return LosslessCompressionType.GZIP;			
		} catch (IOException e) {
		}
		// Is this a ZLIB file?
		try {
			FileInputStream fileInputStream =new FileInputStream(fullFileName);
			InputStream inputStream = new ZInputStream(fileInputStream);
			BitInputStream bitInputStream = new BitInputStream(inputStream);
			// see if you can get one number from it
			bitInputStream.readUBits(16);
			return LosslessCompressionType.ZLIB;			
		} catch (ZStreamException e) {
		} catch (IOException e) {
		}

		// Is this a BZIP2 file?
		try {
			FileInputStream fileInputStream =new FileInputStream(fullFileName);
			new CBZip2InputStream(fileInputStream);
			return LosslessCompressionType.BZIP2;
		} catch (NullPointerException e) {			
		}
		
		// Is this a HUFFMAN file?
		try {
			FileInputStream fileInputStream =new FileInputStream(fullFileName);
			new HuffmanInputStream(fileInputStream);
			return LosslessCompressionType.HUFFMAN;
		} catch (StreamCorruptedException e) {
		} catch (IOException e) {
		}
		return LosslessCompressionType.NONE;
	}


}
