package gov.fnal.eag.sim.utils;

import gov.fnal.eag.util.densityEstimate.Polynomial1D;
import gov.fnal.eag.util.densityEstimate.Polynomial2D;
import gov.fnal.eag.wcs.Coordinate;
import gov.fnal.eag.wcs.TransformPlanar;

import java.util.ArrayList;

/**
 * fits a linear transform for a pair of matched catalogs
 * @author stoughto
 *
 */
public class CatalogFitter {
	ArrayList<Coordinate> measuredCoordinates;
	ArrayList<Coordinate> expectedCoordinates;
	TransformPoly measuredToExpectedPoly = null;
	TransformPolyRadial measuredToExpectedPolyRadial = null;
	int order;
	/**
	 * @param measuredCoordinates
	 * @param expectedCoordinates
	 */
	public CatalogFitter(ArrayList<Coordinate> measuredCoordinates, ArrayList<Coordinate> expectedCoordinates) {
		this(measuredCoordinates, expectedCoordinates, 1);
	}
	/**
	 * @param measuredCoordinates
	 * @param expectedCoordinates
	 * @param order 1 for linear, etc.
	 */
	public CatalogFitter(ArrayList<Coordinate> measuredCoordinates, ArrayList<Coordinate> expectedCoordinates, int order) {
		if (measuredCoordinates == null || expectedCoordinates == null) {
			throw new IllegalArgumentException("coordinates can not be null but measuredCoordinates="+measuredCoordinates+" expectedCoordinates="+expectedCoordinates);
		}
		if (measuredCoordinates.size() != expectedCoordinates.size()) {
			throw new IllegalArgumentException("coordinates must be same size but measuredCoordinates="+measuredCoordinates.size()+" expectedCoordinates="+expectedCoordinates.size());
		}
		if (order < 1) {
			throw new IllegalArgumentException("order must be 1 or larger but order="+order);
		}
		this.measuredCoordinates = measuredCoordinates;
		this.expectedCoordinates = expectedCoordinates;
		this.order = order;
	}
	/**
	 * @return the LinearTransform that transformed measured coordinates to expected coordinates using a polynomial in each direction
	 */
	public TransformPlanar getMeasuredToExpectedPoly() {
		if (measuredToExpectedPoly == null) {
			calculateMeasuredToExpectedPoly();
		}
		return measuredToExpectedPoly;
	}
	
	private void calculateMeasuredToExpectedPoly() {
		ArrayList<double[]> xPoints = new ArrayList<double[]>();
		ArrayList<double[]> yPoints = new ArrayList<double[]>();
		for (int i=0; i<measuredCoordinates.size(); i++) {
			double mx = measuredCoordinates.get(i).getX();
			double my = measuredCoordinates.get(i).getY();
			double ex = expectedCoordinates.get(i).getX();
			double ey = expectedCoordinates.get(i).getY();
			xPoints.add(new double[] {mx, my, ex});
			yPoints.add(new double[] {mx, my, ey});
		}
		Polynomial2D xPoly = new Polynomial2D(xPoints, order, false);
		Polynomial2D yPoly = new Polynomial2D(yPoints, order, false);
		measuredToExpectedPoly = new TransformPoly(xPoly, yPoly);
	}
	/**
	 * @return the LinearTransform that transformed measured coordinates to expected coordinates using a polynomial in each direction
	 */
	public TransformPlanar getMeasuredToExpectedPolyRadial() {
		if (measuredToExpectedPolyRadial == null) {
			calculateMeasuredToExpectedPolyRadial(null);
		}
		return measuredToExpectedPolyRadial;
	}
	/**
	 * @param tpdPlotFileName
	 * @return the transform
	 */
	public TransformPlanar getMeasuredToExpectedPolyRadial(String tpdPlotFileName) {
		calculateMeasuredToExpectedPolyRadial(tpdPlotFileName);
		return measuredToExpectedPolyRadial;
	}
	
	private void calculateMeasuredToExpectedPolyRadial(String plotFileName) {
		ArrayList<double[]> points = new ArrayList<double[]>();
		for (int i=0; i<measuredCoordinates.size(); i++) {
			double mx = measuredCoordinates.get(i).getX();
			double my = measuredCoordinates.get(i).getY();
			double mr = Math.sqrt(mx*mx + my*my);
			double ex = expectedCoordinates.get(i).getX();
			double ey = expectedCoordinates.get(i).getY();
			double er = Math.sqrt(ex*ex + ey*ey);
			points.add(new double[] {mr, er});
		}
		double[] paramsIn = new double[order+1];
		// leave paramsIn[0] set to 0.0 to constrain the fit.  Make the rest Double.NaN so they will be fit.
		for (int i=1; i<paramsIn.length; i++) {
			paramsIn[i] = Double.NaN;
		}
		Polynomial1D poly = new Polynomial1D(points, order, plotFileName, plotFileName, paramsIn);
		measuredToExpectedPolyRadial = new TransformPolyRadial(poly);
	}
}
