package gov.fnal.eag.sim.utils;

import gov.fnal.eag.sim.pixel.PixeledImage;
import gov.fnal.eag.sim.repository.image.ImageRepository;
import gov.fnal.eag.util.misc.CommandLineStderrStdout;
import gov.fnal.eag.util.misc.StringUtil;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.StringTokenizer;

import uk.ac.starlink.table.ColumnInfo;
import uk.ac.starlink.table.RowListStarTable;

/**
 * manages the execution of SExtractor
 */
public class Sexec {
	private static HashMap<String,Class<?>> stringClass = null;
	/**
	 * holds the contents of the .sex file, which tells SExtractor how to run
	 */
	private HashMap<String,String> sex;
	/**
	 * holds the contents of the .param file, which tells SExtractor what
	 * parameters to measure
	 */
	private ArrayList<String> param;
	/**
	 * working directory
	 */
	private String workingDir;
	/**
	 * is the convolution function that SExtractor will use
	 */
	private PixeledImage conv;
	/**
	 * by default, clean up when done, but set this to false to leave
	 * intermediate files on disk for debugging.
	 */
	private boolean cleanupWhenDone = true;
	
	private String segRegion = null;	
	/**
	 * default constructor
	 *  
	 */
	public Sexec()  {
		// Verify the the executable sex is available; if not throw exception
		String command = "which sex";
		String line = null;
		Process p;
		try {
			p = Runtime.getRuntime().exec(command);
		} catch (IOException e) {
			throw new Error(e);
		}
		BufferedReader is = new BufferedReader(new InputStreamReader(p.getInputStream()));

		try {
			if ( (line = is.readLine()) != null) {
			  if (line.indexOf("sex") == -1) {
				System.out.println("sex command is not found -- exiting");
				System.exit(-1);
				}
				
			} else {
				System.out.println("command '"+command+"' return null");
				System.exit(-1);
			}
		} catch (IOException e) {
			throw new Error(e);
		}
		// Create tmp directory where to put files for the run
		ImageRepository ir = new ImageRepository();
		File tmpF = new File(ir.getRepositoryName(), "sexTest");
		workingDir = tmpF.getAbsolutePath();
		if (!tmpF.exists()) {
			if (!tmpF.mkdirs()) {
				throw new Error("can not create workingDir="+workingDir);
			}
		}
		//tmpF.deleteOnExit();

		

		
		// Create default configuration file default.sex
		InputStream defaultSexStream = this.getClass().getResourceAsStream("sextractor/config/default.sex");
		createDefSex(defaultSexStream);
		
		// Create default parameter file default.param
		InputStream defaultParamStream = this.getClass().getResourceAsStream("sextractor/config/default.param");
		createDefParam(defaultParamStream);
		// call setConv with "default"
		setDefConv("default.conv");
		// copy default NNW file
		setNNW("default.nnw");
	}
	/**
	 * Creates default configuration file and fills sex HashMap
	 * 
	 * @param inputStream made from from sextractor/config/default.sex
	 */
		private void createDefSex(InputStream inputStream) {
			String line = null;
			String parString = null;
			String key = null;
			sex = new HashMap<String,String>();
			// First read input file and create a dictionary of configs
			BufferedReader is = new BufferedReader(new InputStreamReader(inputStream));
			try {
				while ((line = is.readLine()) != null) {
					if ( ! line.startsWith("#")) { // skip comments
						StringTokenizer st = new StringTokenizer(line," ");
						if (st.countTokens() >= 2 ) { // if not empty line
							key = st.nextToken();
							parString = "";
							while (st.hasMoreTokens()) {
								parString =parString + "  " + st.nextToken();
							}
//						if (key.equalsIgnoreCase("CATALOG_NAME")) parString = workingDir+"/test.cat";
							if (key.equalsIgnoreCase("PARAMETERS_NAME")) parString = workingDir+"/default.param";
							if (key.equalsIgnoreCase("FILTER_NAME")) parString = workingDir+"/default.conv";
							if (key.equalsIgnoreCase("FLAG_IMAGE")) parString = workingDir+"/flag.fits";
							if (key.equalsIgnoreCase("STARNNW_NAME")) parString = workingDir+"/default.nnw";
							if (key.equalsIgnoreCase("CHECKIMAGE_NAME")) parString = workingDir+"/check.fits";
							sex.put(key,parString);
						}
					}
				}
			} catch (IOException e) {
				throw new Error(e);
			}
			writeConfFile("default.sex");
		}
		/**
		 * Writes default.sex from sex HashMap
		 * @param fn
		 */
		private void writeConfFile(String fn) {
			// Now open the output file and write config
			PrintWriter os;
			try {
				os = new PrintWriter( new FileOutputStream(workingDir+"/default.sex"));
			} catch (FileNotFoundException e) {
				throw new Error(e);
			}
			Collection<String> keys = sex.keySet();
			Iterator<String> it = keys.iterator();
			while (it.hasNext()) {
				String nextKey = it.next();
				String outline = nextKey +" " + sex.get(nextKey);
				os.println(outline);
			}
			os.close();	
		}
		/**
		 * Fill param HashMap and writes default.param file
		 * @param inputStream created from sextractor/config/default.param
		 */
		private void createDefParam(InputStream inputStream) {
			String line = null;
			param = new ArrayList<String>();
			// First read input file and create a dictionary of parameters
			BufferedReader is = new BufferedReader(new InputStreamReader(inputStream));
			try {
				while ((line = is.readLine()) != null) {
					if ( ! line.startsWith("#")) { // skip comments
						StringTokenizer st = new StringTokenizer(line," ");
						if (st.countTokens() >= 1 ) { // if not empty line
							param.add(line);
						}
					}
				}
			} catch (IOException e) {
				throw new Error(e);
			}
			writeParFile("default.param");
		}
		/**
		 * create default paramete file
		 * @param fn
		 */
		private void writeParFile(String fn) {
			// Now open the output file and write config
			PrintWriter os;
			try {
				os = new PrintWriter( new FileOutputStream(workingDir+"/default.param"));
			} catch (FileNotFoundException e) {
				throw new Error(e);
			}
			Iterator<String> it = param.iterator();
			while (it.hasNext()) {
				String nextPar = it.next();
				os.println(nextPar);
			}
			os.close();		
		}
	/**
	 * sets the debugging field cleanupWhenDone
	 * 
	 * @param cleanupWhenDone
	 */
	public void setCleanupWhenDone(boolean cleanupWhenDone) {
		this.cleanupWhenDone = cleanupWhenDone;
	}
	/**
	 * for each line of the stream that is not a comment, set the configuration
	 * parameter to the value, or create a new configuration parameter with the
	 * value
	 * 
	 * @param inputstream - InpusStream
	 * @throws IOException
	 */
	public void setSex(InputStream inputstream) throws IOException {
		String line = null;
		String parString = null;
		String key = null;
		sex = new HashMap<String,String>();
		String fn = workingDir+"default.sex";
		// First read input file and create a dictionary of configs
		BufferedReader is = new BufferedReader(new InputStreamReader(inputstream));
		while ((line = is.readLine()) != null) {
			if ( ! line.startsWith("#")) { // skip comments
				StringTokenizer st = new StringTokenizer(line," ");
				if (st.countTokens() >= 2 ) { // if not empty line
					key = st.nextToken();
					parString = "";
					while (st.hasMoreTokens()) {
						parString =parString + "  " + st.nextToken();
					}
//					if (key.equalsIgnoreCase("CATALOG_NAME")) parString = workingDir+"/test.cat";
					if (key.equalsIgnoreCase("PARAMETERS_NAME")) parString = workingDir+"/default.param";
					if (key.equalsIgnoreCase("FILTER_NAME")) parString = workingDir+"/default.conv";
					if (key.equalsIgnoreCase("FLAG_IMAGE")) parString = workingDir+"/flag.fits";
					if (key.equalsIgnoreCase("STARNNW_NAME")) parString = workingDir+"/default.nnw";
					if (key.equalsIgnoreCase("CHECKIMAGE_NAME")) parString = workingDir+"/check.fits";
					sex.put(key,parString);
				}
			}
		}
		writeConfFile( fn);
	}
	/**
	 * set HashList sex from the file
	 * 
	 * @param fn
	 */
	public void setSex(String fn) {
		String line = null;
		String parString = null;
		String key = null;
		sex = new HashMap<String,String>();
		String outF = workingDir + "default.sex";
		// First read input file and create a dictionary of configs
		BufferedReader is;
		try {
			is = new BufferedReader(new InputStreamReader(new FileInputStream(fn)));
		} catch (FileNotFoundException e) {
			throw new Error(e);
		}
		try {
			while ((line = is.readLine()) != null) {
				if ( ! line.startsWith("#")) { // skip comments
					StringTokenizer st = new StringTokenizer(line," ");
					if (st.countTokens() >= 2 ) { // if not empty line
						key = st.nextToken();
						parString = "";
						while (st.hasMoreTokens()) {
							parString =parString + "  " + st.nextToken();
						}
//					if (key.equalsIgnoreCase("CATALOG_NAME")) parString = workingDir+"/test.cat";
						if (key.equalsIgnoreCase("PARAMETERS_NAME")) parString = workingDir+"/default.param";
						if (key.equalsIgnoreCase("FILTER_NAME")) parString = workingDir+"/default.conv";
						if (key.equalsIgnoreCase("FLAG_IMAGE")) parString = workingDir+"/flag.fits";
						if (key.equalsIgnoreCase("STARNNW_NAME")) parString = workingDir+"/default.nnw";
						if (key.equalsIgnoreCase("CHECKIMAGE_NAME")) parString = workingDir+"/check.fits";
						sex.put(key,parString);
					}
				}
			}
		} catch (IOException e) {
			throw new Error(e);
		}
		writeConfFile(outF);

	}
	/**
	 * Set the name to value, or add it if it is not on the HashList sex
	 * 
	 * @param name
	 * @param value
	 */
	public void setSex(String name, String value) {
		String outF = workingDir + "default.sex";
		if (sex.containsKey(name)) {
			sex.remove(name);
		}
		sex.put(name,value);
		writeConfFile(outF);
	}
	/**
	 * returns the value of the parameter
	 * @param name
	 * @return String containing result of sextractor query
	 */
	public String getSex(String name) {
		String res = sex.get(name);
		return res;
	}
	/**
	 * for each line of the stream that is not a comment, set the parameter to
	 * the boolean value, or create a new parameter set to the boolean value
	 * 
	 * @param stream
	 * @throws IOException
	 */
	public void setParam(InputStream stream) throws IOException {
		String line = null;
		param = new ArrayList<String>();
		String outF = workingDir+"default.param";
		// First read input file and create a dictionary of parameters
		BufferedReader is = new BufferedReader(new InputStreamReader(stream));
		while ((line = is.readLine()) != null) {
			if ( ! line.startsWith("#")) { // skip comments
				StringTokenizer st = new StringTokenizer(line," ");
				if (st.countTokens() >= 1 ) { // if not empty line
					param.add(line);
				}
			}
		}
		writeParFile(outF);

		
	}
	/**
	 * set HashList param from the file
	 * 
	 * @param fn
	 */
	public void setParam(String fn) {
		String line = null;
		param = new ArrayList<String>();
		String outF = workingDir+"default.param";
		// First read input file and create a dictionary of parameters
		BufferedReader is;
		try {
			is = new BufferedReader(new InputStreamReader(new FileInputStream(fn)));
		} catch (FileNotFoundException e) {
			throw new Error(e);
		}
		try {
			while ((line = is.readLine()) != null) {
				if ( ! line.startsWith("#")) { // skip comments
					StringTokenizer st = new StringTokenizer(line," ");
					if (st.countTokens() >= 1 ) { // if not empty line
						param.add(line);
					}
				}
			}
		} catch (IOException e) {
			throw new Error(e);
		}
		writeParFile(outF);

	}
	/**
	 * set the name to the value, or create a new entry with that value, in the
	 * HashList param
	 * 
	 * @param name
	 * @param value
	 */
	public void setParam(String name, boolean value) {
		if (param.contains(name)) {
			if (! value) {
				param.remove(name);
			}
		} else {
			if (value) param.add(name);
		}
		String outF = workingDir+"default.param";
		writeParFile(outF);
	}
	/**
	 * check if the value of the parameter name exists on HashList param,
	 * return true if name exists or false if it
	 * is not on the HashList
	 * 
	 * @param name
	 * @return boolean flag
	 */
	public boolean getParam(String name) {
		return param.contains(name);
	}
	/**
	 *  Create default weight file for neural network
	 * @param defaultNNW
	 */
	public void setNNW(String defaultNNW)  {
		
		String fn = "sextractor"+File.separator+"config"+File.separator+"default.nnw";
		InputStream is = this.getClass().getResourceAsStream(fn);

		OutputStream os;
		try {
			os = new FileOutputStream(workingDir+"/default.nnw");
		} catch (FileNotFoundException e) {
			throw new Error(e);
		}
		copyFile(is,os,true);
	}
	/**
	 * creates check.fits file
	 * @param checkFile - should include an absolute path 
	 * @throws IOException
	 */
	public void setCheckImage( String checkFile) throws IOException {
		InputStream is = new FileInputStream(checkFile);
		OutputStream os = new FileOutputStream(workingDir+"/check.fits");
		copyFile(is,os,true);
	}
	/**
	 * sets the PixeledImage conv with one of the standards provided by
	 * SExtractor
	 * 
	 * @param defaultName
	 */
	public void setDefConv(String defaultName) {
		String line = null;
		double[][] pixels;
		ArrayList<String> convLine = new ArrayList<String>();
				
		String baseFileName = "sextractor" + File.separator + "config" + File.separator + defaultName;
		InputStream is = Sexec.class.getResourceAsStream(baseFileName);
		if (is == null) {
			throw new Error("InputStream is null for baseFileName="+baseFileName);
		}
		OutputStream os;
		try {
			os = new FileOutputStream(workingDir+"/default.conv");
		} catch (FileNotFoundException e) {
			throw new Error(e);
		}
		copyFile(is,os,true);
		BufferedReader inR;
		InputStream is2 = Sexec.class.getResourceAsStream(baseFileName);
		inR = new BufferedReader(new InputStreamReader(is2));
		try {
			while ((line = inR.readLine()) != null ) {
				convLine.add(line);
			}
		} catch (IOException e) {
			throw new Error(e);
		}
		int nrows = convLine.size();
		StringTokenizer st = new StringTokenizer(convLine.get(2)," ");
		int ncols = st.countTokens();
		pixels = new double[nrows][ncols];
		Iterator<String> it = convLine.iterator();
		int i = 0;
		it.next();   //Skip two comment lines
		it.next();
		while (it.hasNext()) {
			st = new StringTokenizer(it.next()," ");
			int j= 0;
			while (st.hasMoreTokens()) {
				pixels[i][j] = Double.parseDouble(st.nextToken());
				j++;
			}
			i++;
		}
		conv = new PixeledImage(pixels);
	}
	/**
	 *  Simple file copy procedure
	 * @param is
	 * @param os
	 * @param close
	 */
	private static void copyFile(InputStream is, OutputStream os,boolean close) {
		int b;   // byte from input file
		try {
			while ((b = is.read()) != -1 ) {
				os.write(b);
			}
			is.close();
			if (close) os.close();
		} catch (Exception e) {
			throw new Error(e);
		}
	}
	/**
	 * sets the PixeldImage conv with the input PixeledImage
	 * 
	 * @param conv
	 */
	public void setConv(PixeledImage conv)  {
		int nrow = conv.getRowCount();
		int ncol = conv.getColumnCount() ;
		this.conv = conv;
		String line = null;
		PrintWriter os;
		try {
			os = new PrintWriter( new FileOutputStream(workingDir+"/default.conv"));
		} catch (FileNotFoundException e) {
			throw new Error(e);
		}
		// create small header
		line = "CONV NORM";
		os.println(line);
		line = "#  "+nrow + "x" + ncol;
		os.println(line);
		for (int i = 0; i< nrow; i++ ) {
			line = " ";
			for (int j=0; j<ncol; j++) {
				line = line + Double.toString(conv.getPix()[i][j]) + " ";
			}
			os.println(line);
		}
		os.close();
	}
	/**
	 * gets the PixeledImage that will be used for convolution in SExtractor
	 * @return PixeledImage
	 */
	public PixeledImage getConv() {
		return conv;
	}	
	/**
	 * gets the name of the segmentation region file
	 * @return the name of the region file
	 */
	public String getSegRegion() {
		return segRegion;
	}	
	/**
	 * sets the name of the segmentation region file
	 * @param segRegion name of region file
	 */
	public void setSegRegion(String segRegion) {
		this.segRegion = segRegion;
	}
	
	/**
	 * get the name of the .fits file for SEGMENTATION
	 * @param index index of the .fits file for SEGMENTATION
	 * @return the name of the .fits file
	 */
	public String getSegmentation(int index) {
		String line = getSex("CHECKIMAGE_NAME");
		String[] tokens = line.split(" ");
		return tokens[index];
	}
	
	/**
	 * does the actual work of writing files, calling sex, and returning the
	 * result
	 * 
	 * @param pi
	 * @return ArrayList of HashMaps. Each HashMap contains parameters
	 *  of a  reconstructed object
	 */
	public String exec(PixeledImage pi) {
		// Write the image to a file
		String tempFitsImage = workingDir+"/defaultImg.fits";
		ImageRepository ir = new ImageRepository(workingDir);
		ir.put("defaultImg.fits", pi, true);		
		
		CommandLineStderrStdout.run(new String[] {"which", "sex"});
		String[] command;
		command = new String[]{"sex",tempFitsImage,"-c",workingDir+"/default.sex"};
		System.out.println("command");
		CommandLineStderrStdout.run(command);
		
		if (segRegion!=null) {
			String line = getSex("CHECKIMAGE_TYPE");
			String[] tokens = line.split(" ");
			String file = null;
			for (int i=0; i<tokens.length; i++) {
				if (tokens[i].compareTo("SEGMENTATION")==0) {
					file = getSegmentation(i);
				}
			}
			if (file!=null) {
				ImageRepository irSeg = new ImageRepository(".");
				PixeledImage piSeg = irSeg.getPixeledImage(file);
				double[][] pixels = piSeg.getPix();
				File fileOut = new File(segRegion);
				FileWriter fwOut;
				try {
					fwOut = new FileWriter(fileOut);
				} catch (IOException e) {
					throw new Error(e);
				}
				for (int i=0; i<piSeg.getRowCount(); i++) {
					for (int j=0; j<piSeg.getColumnCount(); j++) {
						if (pixels[i][j]!=0) {
							PrintWriter pwOut = new PrintWriter(fwOut);											
							pwOut.println("cross point "+(j+1)+" "+(i+1));
						}
					}
				}
				try {
					fwOut.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}		
		if (cleanupWhenDone) cleanup();
		return StringUtil.removeDuplicateWhitespace(sex.get("CATALOG_NAME")).toString().split(" ")[0];	
	}
	/**
	 *  cleanup the working directory
	 *
	 */
	public void cleanup() {
		String command = "rm -rf "+workingDir;
		try {
		    Process p = Runtime.getRuntime().exec(command);
		    p.waitFor();
		} catch (Exception e) {
			System.out.println("Error cleaning workDir");
			e.printStackTrace();
		}
	}
	/**
	 * 
	 * @return Array list of parameter keywords
	 */
	public ArrayList<String> getParameters() {
		return param;
	}
	/**
	 * @return cleanupWhenDone flag boolean
	 */
	public boolean getCleanupWhenDone() {		
		return cleanupWhenDone;
	}
	/**
	 * convert the sextractor catalog file to a RLST
	 * @param catalogName file name to read
	 * @return the table
	 */
	public RowListStarTable getRlst(String catalogName) {
		RowListStarTable rlst = null;
		BufferedReader bufferedReader;
		try {
			bufferedReader = new BufferedReader(new FileReader(catalogName));
		} catch (FileNotFoundException e) {
			throw new Error(e);
		}
		ColumnInfo[] columnInfo = new ColumnInfo[param.size()];
		Class<?>[] dataType = new Class<?>[param.size()];
		int column = 0;
		String line;
		try {
			while ( (line = bufferedReader.readLine()) != null) {
				if (line.toUpperCase().startsWith("SIMPLE")) {
					bufferedReader.close();
					rlst = ImageRepository.readRlst(catalogName, 1);
					return rlst;
				}
				String[] lineTokens = StringUtil.removeDuplicateWhitespace(line).toString().split(" ");
				if (lineTokens[0].equals("#")) {
					String name = lineTokens[2];
					Class<?> columnClass = getColumnClass(name);
					String description = "";
					for (int i=3; i<lineTokens.length; i++) {
						description += lineTokens[i]+" ";
					}
					dataType[column] = columnClass;
					columnInfo[column++] = new ColumnInfo(name, columnClass, description);
					if (column == param.size()) {
						rlst = new RowListStarTable(columnInfo);
					}
				} else {
					Object[] row = new Object[param.size()];
					for (int i=0; i<row.length; i++) {
						row[i] = getNewNumber(lineTokens[i], dataType[i]);
					}
					if (rlst != null) {
						rlst.addRow(row);
					} else {
						throw new Error("poorly formatted catalogName="+catalogName);
					}
				}
			}
		} catch (IOException e) {
			throw new Error(e);
		}
		return rlst;
	}
	private Object getNewNumber(String string, Class<?> type) {
		if (type.equals(Double.class)) {
			return new Double(string);
		} else if (type.equals(Integer.class)) {
			return new Integer(string);
		} else {
			throw new Error("teach me how to getNewNumber for string="+string+" type="+type);
		}
	}
	private Class<?> getColumnClass(String name) {
		// Add new column names and classes here.  Change getNewNumber if you use new classes
		if (stringClass == null) {
			stringClass = new HashMap<String,Class<?>>();
			stringClass.put("NUMBER", Integer.class);
			stringClass.put("FLAGS", Integer.class);
		}
		Class<?> retval = stringClass.get(name);
		if (retval == null) {
			retval =  Double.class;
		}
		return retval;
	}
	/**
	 * @return the working directory
	 */
	public String getWorkingDir() {
		return workingDir;
	}
}