package gov.fnal.eag.sim.utils;

import gov.fnal.eag.util.yanny.Table;
import gov.fnal.eag.wcs.SphericalCoordinate;

import java.awt.geom.Point2D;
import java.io.IOException;
import java.util.Iterator;

import cern.colt.list.DoubleArrayList;
import cern.jet.stat.Descriptive;

import uk.ac.starlink.table.ColumnInfo;
import uk.ac.starlink.table.RowListStarTable;
import uk.ac.starlink.table.RowSequence;
import uk.ac.starlink.table.StarTable;
import uk.ac.starlink.table.join.HEALPixMatchEngine;
import uk.ac.starlink.table.join.LinkSet;
import uk.ac.starlink.table.join.MatchEngine;
import uk.ac.starlink.table.join.RowLink;
import uk.ac.starlink.table.join.RowMatcher;

/**
 * matches two star tables
 * @author stoughto
 *
 */
public class StarTableMatcher {
	StarTable table0, table1, t0;
	private String ra0Name, dec0Name, ra1Name, dec1Name;
	double xVariance=-1, yVariance=-1;
	private DoubleArrayList deltaRaDal, deltaDecDal;
	
	private static ColumnInfo[] ci = {
		new ColumnInfo("ra", Double.class, "ra"),
		new ColumnInfo("dec", Double.class, "dec"),
	};
	
	/**
	 * @param table0 first table 
	 * @param ra0 name of the longitude coordinate
	 * @param dec0 name of the latitude coordinate
	 * @param table1 to match to the first stable
	 * @param ra1 name of the longitude coordinate
	 * @param dec1 name of the latitude coordinate
	 */
	public StarTableMatcher(StarTable table0, String ra0, String dec0, StarTable table1, String ra1, String dec1) {
		this.ra0Name = ra0;
		this.dec0Name = dec0;
		this.ra1Name = ra1;
		this.dec1Name = dec1;
		this.table0 = table0;
		this.table1 = table1;
		t0 = makeStarTable(table0, ra0Name, dec0Name, 0.0, 0.0);
	}
	
	private StarTable makeStarTable(StarTable table, String raName, String decName, double deltaRa, double deltaDec) {
		int raIndex = Table.getIndexByName(table, raName);
		int decIndex = Table.getIndexByName(table, decName);
		RowListStarTable retval = new RowListStarTable(ci);
		try {
			RowSequence rs = table.getRowSequence();
			while (rs.next()) {
				Object row[] = rs.getRow();
				Double dec = (Double) row[decIndex] + deltaDec/3600.0;
				double cosDec = Math.cos(Math.toRadians(dec));
				Double ra = ((Double) row[raIndex]) + (deltaRa/3600/cosDec);
				SphericalCoordinate sc = new SphericalCoordinate(ra,dec);
				retval.addRow(new Object[] {Math.toRadians(sc.getLongitude()), Math.toRadians(sc.getLatitude())});
			}
		} catch (IOException e) {
			throw new Error(e);
		}
		return retval;
	}
	
	/**
	 * do the match
	 * @param separation arcseconds of cone search size
	 * @param bestOnly report only the best matches if true
	 * @return the result of the match
	 */
	public LinkSet match(double separation, boolean bestOnly) {
		return match(separation, new Point2D.Double(), bestOnly);
	}
	
	/**
	 * do the match
	 * @param separation arcseconds of cone search size
	 * @param offset offest in arcseconds applied to the second table before match
	 * @param bestOnly report only the best matches if true
	 * @return the result of the match
	 */
	public LinkSet match(double separation, Point2D.Double offset, boolean bestOnly) {
		StarTable t1 = makeStarTable(table1, ra1Name, dec1Name, offset.x, offset.y);
		MatchEngine matchEngine = new HEALPixMatchEngine(Math.toRadians(separation/3600), false);
		RowMatcher matcher = new RowMatcher(matchEngine, new StarTable[] {t0, t1});
		LinkSet retval;
		try {
			retval = matcher.findPairMatches(bestOnly);
		} catch (Exception e) {
			throw new Error(e);
		} 
		return retval;
	}
	/**
	 * calculates the offset to be added to the second table to make the tables match
	 * @param match
	 * @return the offset
	 */
	public Point2D.Double calculateOffset(LinkSet match) {
		StarTable t1 = makeStarTable(table1, ra1Name, dec1Name, 0.0, 0.0);
		
		deltaRaDal = new DoubleArrayList();
		deltaDecDal = new DoubleArrayList();
		
		Iterator<?> it = match.iterator();		
		while (it.hasNext()) {
			RowLink rowLink  =  (RowLink) it.next();
			long i0 = rowLink.getRef(0).getRowIndex();
			long i1 = rowLink.getRef(1).getRowIndex();
			try {
				Object[] row0 = t0.getRow(i0);
				double ra0 = ((Number) row0[0]).doubleValue();
				double dec0 = ((Number) row0[1]).doubleValue();
				Object[] row1 = t1.getRow(i1);
				double ra1 = ((Number) row1[0]).doubleValue();
				double dec1 = ((Number) row1[1]).doubleValue();
				double cosDec0 = Math.cos(dec0);
				double deltaRa  = 3600 * cosDec0 * Math.toDegrees(ra0 - ra1);
				deltaRaDal.add(deltaRa);
				double deltaDec = 3600 * Math.toDegrees(dec0 - dec1);
				deltaDecDal.add(deltaDec);
			} catch (IOException e) {
				throw new Error(e);
			}
		}
		deltaRaDal.sort();
		Point2D.Double retval = new Point2D.Double();
		retval.x = Descriptive.median(deltaRaDal);
		xVariance = Descriptive.sampleVariance(deltaRaDal, retval.x);
		
		deltaDecDal.sort();
		retval.y = Descriptive.median(deltaDecDal);
		yVariance = Descriptive.sampleVariance(deltaDecDal, retval.y);
		
		return retval;
	}
	
	/**
	 * @return a separation based on the previous call to calculateOffset
	 */
	public double suggestSeparation() {
		if (xVariance<0 || yVariance<0) {
			throw new IllegalArgumentException("must call calculateOffset before asking for a suggestion");
		}
		double retval = 0.5*(Math.sqrt(xVariance) + Math.sqrt(yVariance));
		return retval;
	}

	/**
	 * @param match
	 * @return the average position of the match set
	 */
	public SphericalCoordinate getCenter(LinkSet match) {
		double raSum = 0;
		double decSum = 0;
		int n = 0;
		Iterator<?> it = match.iterator();		
		while (it.hasNext()) {
			RowLink rowLink  =  (RowLink) it.next();
			long i0 = rowLink.getRef(0).getRowIndex();
			try {
				Object[] row0 = t0.getRow(i0);
				raSum += ((Number) row0[0]).doubleValue();
				decSum += ((Number) row0[1]).doubleValue();
				n++;
			} catch (IOException e) {
				throw new Error(e);
			}
		}
		SphericalCoordinate retval = new SphericalCoordinate(raSum/n, decSum/n);
		return retval;		
	}

	/**
	 * @param linkSet
	 * @return the average of sqrt of the x and y variance
	 */
	public Point2D.Double getWidth(LinkSet linkSet) {
		Point2D.Double retval = new Point2D.Double(Math.sqrt(xVariance), Math.sqrt(yVariance));
		return retval;
	}
}
