package gov.fnal.eag.sim.wavelet;

import gov.fnal.eag.sim.pixie.PixieImage;
import gov.fnal.eag.sim.pixie.PixieImageException;

import java.io.File;

/**
 * implements the A Trous (with holes) wavelet technique
 * @author mcginnis
 *
 */
public class Wavelet 
{
	private	int			scaleMax;
	private	int[]		radiusAvg = new int[10];
	private	boolean		hardEdge;
	private	double[]	scaleSigmaCut = new double[10];
	private	int			KERNAL_SIZE = 257;
	
	/**
	 * Class uses Wavelet a trous procedure to form a wavelet transform and filter
	 */
	public Wavelet()
	{
		int ii;
		for (ii = 0; ii < 10; ++ii)
		{
			scaleSigmaCut[ii] = 0.0;
			radiusAvg[ii] = 0;
		}
		scaleSigmaCut[0] = 3.0;
		scaleSigmaCut[1] = 1.0;
		scaleSigmaCut[2] = 1.0;
		scaleSigmaCut[3] = 1.0;
		scaleSigmaCut[4] = 1.0;
		radiusAvg[0] = 1;
		radiusAvg[1] = 2;
		radiusAvg[2] = 4;
		radiusAvg[3] = 0;
		radiusAvg[4] = 0;
		scaleMax = 2;
		hardEdge = true;
	}
	/**
	 * copy constructor
	 * @param wavecopy the original to copy
	 */
	public Wavelet(Wavelet wavecopy)
	{
		int ii;
		for (ii = 0; ii < 10; ++ii)
		{
			scaleSigmaCut[ii] = wavecopy.getScaleSigmaCut(ii);
			radiusAvg[ii] = wavecopy.getRadiusAvg(ii);
		}
		hardEdge = wavecopy.getHardEdge();
		scaleMax = wavecopy.getScaleMax();
	}
	/**
	 * @param inputImage
	 * @return filtered image
	 */
	public PixieImage waveFilter(PixieImage inputImage)
	{
		int			ii;
		int			ij;
		double		davg;
		int			nrows;
		int			ncols;
		int			scale;
		PixieImage	outputImage = null;
		PixieImage[]	waveletImages = null;
		double[][]	outputPixelArray = null;
		double[][]	waveletPixelArray = null;
		
		
		double[] inputNoiseMeanSigma = inputImage.findNoiseMeanSigma();
		double inputImageNoiseMean = inputNoiseMeanSigma[0];
		double inputImageNoiseSigma = inputNoiseMeanSigma[1];
		inputImage.rescalePixels(1.0, -inputImageNoiseMean);
		inputImage.rescalePixels(1.0 / inputImageNoiseSigma, 0.0);

		nrows = inputImage.getRowCount();
		ncols = inputImage.getColCount();
		outputImage = new PixieImage(inputImage);

		outputPixelArray = outputImage.getPix();

		for (ii = 0; ii < nrows; ii++)
		{  
			for (ij = 0; ij < ncols; ij++) 
			{
				outputPixelArray[ii][ij] = 0.0;
			}
		}
		waveletImages = multipleWaveletTransform(inputImage);

		for (scale = 0; scale <= scaleMax; ++scale)
		{
			waveletPixelArray = waveletImages[scale].getPix();
			double[] noiseMeanSigma = waveletImages[scale].findNoiseMeanSigma();
			double imageNoiseSigma = noiseMeanSigma[1];

			for (ii = 0; ii < nrows; ii++)
			{  
				for (ij = 0; ij < ncols; ij++) 
				{
					davg = nearNeighborAvg(waveletPixelArray, nrows, ncols, 
							ii, ij, 
							radiusAvg[scale], true );

					if (davg >= (scaleSigmaCut[scale] * imageNoiseSigma))
					{
						if (hardEdge)
						{
							outputPixelArray[ii][ij] 
							  = outputPixelArray[ii][ij] 
							  + waveletPixelArray[ii][ij];
						}
						else
						{
							if (waveletPixelArray[ii][ij] > 0.0)
							{
								outputPixelArray[ii][ij] 
								    = outputPixelArray[ii][ij] 
									+ waveletPixelArray[ii][ij] 
									- (scaleSigmaCut[scale] * imageNoiseSigma);
							}
							else
							{
								outputPixelArray[ii][ij] 
								    = outputPixelArray[ii][ij] 
									+ waveletPixelArray[ii][ij] 
									+ (scaleSigmaCut[scale] * imageNoiseSigma);
							}
						}
					}
				}
			}
		}
		// Now handel the residual wavelet image
		waveletPixelArray = waveletImages[scaleMax + 1].getPix();
		double[] noiseMeanSigma = waveletImages[scaleMax + 1].findNoiseMeanSigma();
		double imageNoiseMean = noiseMeanSigma[0];
		double imageNoiseSigma = noiseMeanSigma[1];
		
		for (ii = 0; ii < nrows; ii++)
		{  
			for (ij = 0; ij < ncols; ij++) 
			{
				davg = nearNeighborAvg(waveletPixelArray, nrows, ncols, 
						ii, ij, 
						radiusAvg[scaleMax + 1], false );
				if ((davg - imageNoiseMean) 
					>= (scaleSigmaCut[scaleMax + 1] * imageNoiseSigma))
				{
					if (hardEdge)
					{
						outputPixelArray[ii][ij] = outputPixelArray[ii][ij] 
						   + waveletPixelArray[ii][ij];
					}
					else
					{
						if (waveletPixelArray[ii][ij] > 0.0)
						{
							outputPixelArray[ii][ij] = outputPixelArray[ii][ij] 
								+ waveletPixelArray[ii][ij] 
								- (scaleSigmaCut[scaleMax + 1] * imageNoiseSigma)
								+ imageNoiseMean;
						}
						else
						{
							outputPixelArray[ii][ij] = outputPixelArray[ii][ij] 
								+ waveletPixelArray[ii][ij] 
								+ (scaleSigmaCut[scaleMax + 1] * imageNoiseSigma)
								- imageNoiseMean;
						}
					}
				}
				else
				{
					outputPixelArray[ii][ij] = imageNoiseMean;
				}
				if (outputPixelArray[ii][ij] < imageNoiseMean ) 
					outputPixelArray[ii][ij] = imageNoiseMean;
			}
		}
		waveletImages = null;
		System.gc();

		inputImage.rescalePixels(inputImageNoiseSigma, 0.0);
		inputImage.rescalePixels(1.0, inputImageNoiseMean);
		outputImage.rescalePixels(inputImageNoiseSigma, 0.0);
		outputImage.rescalePixels(1.0, inputImageNoiseMean);
		return outputImage;

	}
	/**
	 * @param inputFile
	 * @return PixieImage
	 * @throws PixieImageException
	 */
	public PixieImage waveFilter(File inputFile) throws PixieImageException
	{
		PixieImage inputImage = new PixieImage(inputFile);
   		return waveFilter(inputImage);
	}
	/**
	 * @param inputFile
	 * @param outputFile
	 * @param shortFormat
	 * @throws PixieImageException
	 */
	public void waveFilterToFile(File inputFile, File outputFile, boolean shortFormat) throws PixieImageException
	{
		PixieImage inputImage = new PixieImage(inputFile);
   		PixieImage outputImage = waveFilter(inputImage);
		outputImage.writeToFitsFile(outputFile, shortFormat);
	}
	/**
	 * @param inputImage
	 * @return PixieImage[]
	 */
	public PixieImage[] multipleWaveletTransform(PixieImage inputImage)
	{
		int		scale;
		int	nkernLen;
		double[][]	kernal = null;
		PixieImage[] waveletImage = null;
		
		
		waveletImage = new PixieImage[scaleMax + 2];
		kernal = new double[KERNAL_SIZE][KERNAL_SIZE];
		for (scale = 0; scale <= scaleMax; scale++)
		{
			nkernLen = waveletKernal(scale, kernal);
			waveletImage[scale] = convolution(inputImage, kernal, 
					nkernLen);

		}
		nkernLen = residualWaveletKernal(scaleMax + 1,kernal);
		waveletImage[scaleMax + 1] = convolution(inputImage, kernal, 
				nkernLen);
		return waveletImage;
	}
	/**
	 * @param inputFile
	 * @return PixieImage[]
	 * @throws PixieImageException
	 */
	public PixieImage[] multipleWaveletTransform(File inputFile) throws PixieImageException
	{
		PixieImage inputImage = new PixieImage(inputFile);
   		return multipleWaveletTransform(inputImage);
	}
	/**
	 * @param inputFile
	 * @param outputFileDirectory
	 * @param shortFormat
	 * @throws PixieImageException
	 */
	public void multipleWaveletTransformToFile(File inputFile, File outputFileDirectory, boolean shortFormat) throws PixieImageException
	{
		int		ii;
		String delim = System.getProperty("file.separator");
		String outputfileName;

		PixieImage inputImage = new PixieImage(inputFile);
    	
		PixieImage[] outputImage =  multipleWaveletTransform(inputImage);
    	for (ii = 0; ii <= getScaleMax(); ++ii)
    	{
    		outputfileName 
				= outputFileDirectory.getParent() 
				+ delim 
				+ "wave" + Integer.toString(ii) + ".fit"; 
    		outputImage[ii].writeToFitsFile(outputfileName, shortFormat);
    	}
		outputfileName 
			= outputFileDirectory.getParent() 
			+ delim 
			+ "wave_resd.fit"; 
 		outputfileName = "wave_resd.fit"; 
		outputImage[getScaleMax() + 1].writeToFitsFile(outputfileName, shortFormat);
	}
	private int waveletKernal(int scale, double[][] kernal)
	{
		int		nkernLenSmall;
		int		nkernLen;
		int		ii;
		int		ij;
		int		offset;
		double[][]	kernalSmall = null;

		kernalSmall = new double[KERNAL_SIZE][KERNAL_SIZE];
		nkernLenSmall = residualWaveletKernal(scale,  kernalSmall);
		nkernLen = residualWaveletKernal(scale + 1, kernal);
		offset = (nkernLen - nkernLenSmall) / 2;
		for (ii = 0; ii < nkernLen; ++ii)
		{
			for (ij = 0; ij < nkernLen; ++ij)
			{
				kernal[ii][ij] = -kernal[ii][ij];
			}
		}
		for (ii = 0; ii < nkernLenSmall; ++ii)
		{
			for (ij = 0; ij < nkernLenSmall; ++ij)
			{
				kernal[ii + offset][ij + offset] 
					= kernal[ii + offset][ij + offset]
					+ kernalSmall[ii][ij];
			}
		}
		return nkernLen;
	}
	private int residualWaveletKernal(int scale, double[][] dkernal)
	{
		int		ii;
		int		ij;
		int		nkernLen;
		int		radius;
		double	amplitude;
		double[]	kernal1d = null;

		kernal1d = new double[KERNAL_SIZE];
		if (scale > 8)
		{
			nkernLen = 3;
			for (ii = 0; ii < nkernLen; ++ii) 
			{
				for (ij = 0; ij < nkernLen; ++ij) 
				{
					dkernal[ii][ij] = 0.0;
				}
			}
			return nkernLen;
		}
		if ( scale == 0 )
		{
			radius = 0;
			amplitude = 1.0;
		}
		else
		{
			radius = 1;
			if (scale > 1)
			{
				for (ii = 2; ii <= scale; ++ii)
				{
					radius = radius * 2;
				}
			}
			amplitude = 2.0 * radius;
			amplitude = 1.0 / amplitude;
		}
		nkernLen = 2 * radius + 1;
		if ( scale == 0 )
		{
			dkernal[0][0] = amplitude;
			return nkernLen;
		}
		for (ii = -radius; ii <= radius; ++ii)
		{
			kernal1d[ii + radius] = amplitude * 
				waveletMother(amplitude  * ii);
		}
		for (ii = -radius; ii <= radius; ++ii)
		{
			for (ij = -radius; ij <= radius; ++ij)
			{
				dkernal[ii + radius][ij + radius] 
					= kernal1d[ii + radius] * kernal1d[ij + radius];
			}
		}
		return nkernLen;
	}
	private double waveletMother(double x)
	{
		double	mother;

		if ( x <= -1.0)
		{
			mother = 0.0;
		}
		else
		{
			if ( x <= 0.0)
			{
				mother = 1.0 + x;
			}
			else
			{
				if ( x <= 1.0 )
				{
					mother = 1.0 - x;
				}
				else
				{
					mother = 0.0;
				}
			}
		}
		return mother;
	}
	private PixieImage convolution(PixieImage inputImage, double[][] kernal, 
							int nkernLen)
	{
		int			ii;
		int			ij;
		int			ik;
		int			il;
		int			iispot;
		int			ijspot;
		int			nkernLen2;
		int			nrows;
		int			ncols;
		PixieImage	outputImage = null;
		double[][]	inputPixelArray = null;
		double[][]	outputPixelArray = null;
		
		nrows = inputImage.getRowCount();
		ncols = inputImage.getColCount();
		outputImage = new PixieImage(inputImage);
		inputPixelArray = inputImage.getPix();
		outputPixelArray = outputImage.getPix();
		nkernLen2 = nkernLen / 2;
		for (ii = 0; ii < nrows; ++ii)
		{
			for (ij = 0; ij < ncols; ++ij)
			{
				outputPixelArray[ii][ij] = 0.0;
				for (ik = 0; ik < nkernLen; ++ik)
				{
					iispot = ii - nkernLen2 + ik;
					while ( iispot >= nrows ) 
					{	
						iispot = iispot - nrows;
					}
					while ( iispot <  0 ) 
					{
						iispot = iispot + nrows;
					}
					for (il = 0; il < nkernLen; ++il)
					{
						ijspot = ij - nkernLen2 + il;
						while ( ijspot >= ncols ) 
						{	
							ijspot = ijspot - ncols;
						}
						while ( ijspot <  0 ) 
						{
							ijspot = ijspot + ncols;
						}
						outputPixelArray[ii][ij] = outputPixelArray[ii][ij] 
							+ kernal[ik][il] 
							* inputPixelArray[iispot][ijspot];
					}
				}
			}
		}
		return outputImage;
	}
	private double nearNeighborAvg(double[][] dpix, int nrows, int ncols, 
										int irow, int icol, 
										int iradius, boolean absFlag )
	{
		int		ik;
		int		im;
		int		il;
		int		in;
		double	dsum;
		double	dcount;
		int		iradius2;

		dsum = 0.0;
		dcount = 0.0;
		iradius2 = iradius;
		if (iradius2 < 0 ) iradius2 = 0;
		for (ik = -iradius2; ik <= iradius2; ik++) 
		{
			im = irow + ik;
			if ( (0 <= im) && (im < nrows) )
			{
				for (il = -iradius2; il <= iradius2; il++) 
				{
					in = icol + il;
					if ( (0 <= in) && (in < ncols) )
					{
						if (absFlag)
						{
							dsum = dsum + Math.abs(dpix[im][in]);
						}
						else
						{
							dsum = dsum + dpix[im][in];
						}
						dcount = dcount + 1;
					}
				}
			}
		}
		dsum = dsum  / dcount;
		return dsum;
	}
	/**
	 * @param newScaleMax
	 */
	public void setScaleMax(int newScaleMax)
	{
		scaleMax = newScaleMax;
		if (scaleMax > 9) scaleMax = 9;
		if (scaleMax < 1) scaleMax = 1;
	}
	/**
	 * @return int
	 */
	public int getScaleMax()
	{
		return scaleMax;
	}
	/**
	 * @param newHardEdge
	 */
	public void setHardEdge(boolean newHardEdge)
	{
		hardEdge = newHardEdge;
	}
	/**
	 * @return boolean
	 */
	public boolean getHardEdge()
	{
		return hardEdge;
	}
	/**
	 * @param index
	 * @param newScaleSigmaCut
	 */
	public void setScaleSigmaCut(int index, double newScaleSigmaCut)
	{
		if ((0 <= index) && (index < 10) && (newScaleSigmaCut >= 0.0))
		{
			scaleSigmaCut[index] = newScaleSigmaCut;
		}
	}
	/**
	 * @param index
	 * @return double
	 */
	public double getScaleSigmaCut(int index)
	{
		if ((0 <= index) && (index < 10))
		{
			return  scaleSigmaCut[index];
		}
		
		return -1.0;
		
	}
	/**
	 * @param index
	 * @param newRadiusAvg
	 */
	public void setRadiusAvg(int index, int newRadiusAvg)
	{
		if ((0 <= index) && (index < 10) && (newRadiusAvg >= 0))
		{
			radiusAvg[index] = newRadiusAvg;
		}
	}
	/**
	 * @param index
	 * @return int
	 */
	public int getRadiusAvg(int index)
	{
		if ((0 <= index) && (index < 10))
		{
			return  radiusAvg[index];
		}
		return -1;
	}
}
