
package gov.fnal.eag.util.astronomy;

import gov.fnal.eag.util.misc.Constants;
import uk.ac.starlink.util.PhysicalConstants;


/**
 * 
 * @author stoughto
 *
 */
public abstract class Cosmo {
	
	
	/**
	 * standard value of H, in m/sec/pc
	 */
	public static final double standardHubbleConstant = 0.10; //  in m/sec/pc
	
	private double hubbleConstant = 0.071; // in m/sec/pc
	
	/**
	 * @return a copy of this cosmology
	 */
	public abstract Cosmo copy();

	/**
	 * @return angular diameter distance in units of c/H0
	 * @param z
	 */
	public abstract double angularDistance(double z);
    
	/**
	 * calculate differential comoving volume
	 * @param z redshift
	 * @return dVdz in units of (c/H)^-3
	 */
	public abstract double dVdz(double z);
	
	/**
	 * @param that
	 * @return true if that equals this Cosmo
	 */
	public abstract boolean equals(Cosmo that);
	
	/**
	 * @return the function that calculates dVdz as a function of z
	 */
	public abstract SimpleFunction getdVdzFunc();
	
	/**
	 * integrate dVdz to get the total volume in the shell
	 * @param zMin minimum redshift
	 * @param zMax maximum redshift
	 * @return the volume, in units of (c/H)^3
	 */
	
	public double volume(double zMin, double zMax) {
		return Integrate.qRomb(getdVdzFunc(), zMin, zMax, 1e-10);
	}

    /**
      * @param z
     * @return luminosity distance, in units of c/H0
     */
    public double luminosityDistance(double z) {
    	double luminosityDistance = angularDistance(z) * Math.pow(1+z,2);
    	return luminosityDistance;
    }
    
    /**
     * The distance modulus of an object an a given redshift
     * @param z the redshift
     * @return the distance modulus
     */
    public double distanceModulus(double z) {
    	// double distanceModulus = 5.0 * Math.log10(luminosityDistance(z)/10.0);
    	// double distanceModulus = 5.0 * Math.log10(luminosityDistance(z)/10.0);
    	double luminosityDistance = luminosityDistance(z);
    	double luminosityDistanceParsec = luminosityDistance *getCH0();
    	double distanceModulus = 5.0 * Math.log10(luminosityDistanceParsec/10.0);
    	return distanceModulus;
    }	
    /**
     * Change the value of the Hubble constant, where 0.07 is the default
     * @param value of the Hubble constant to set, in m/sec/pc.
     */
    public void setHubbleConstant(double value) {
    	hubbleConstant = value;
    }
    /**
     * @return the value of c/H in pc
     */
    public double getCH0() {
    	double c = PhysicalConstants.SPEED_OF_LIGHT; // in meters/sec
    	return c/hubbleConstant;
    }

	/**
	 * @return hubbleConstant/standardHubbleConstant, where standardHubbleConstant=0.1 m/sec/pc
	 */
	public double getLittleH() {
		return hubbleConstant/standardHubbleConstant;
	}
	
	/**
	 * @param redshift
	 * @return the surface area of a sphere at the redshift, in cm*cm
	 */
	public double getPhotonSphereSurface(double redshift) {
		double distanceCH0 = angularDistance(redshift);
		double ch0 = getCH0();
		double distanceMeters = distanceCH0 * ch0 * Constants.METERS_PER_PARSEC;
		double retval = 4 * Math.PI * Math.pow(1+redshift, 3) * Math.pow(100*distanceMeters, 2);
		return retval;		
	}
	
	/*
	/**
	 * @param apparentMagnitude
	 * @param redshift
	 * @return the absolute magnitude
	 */
	/*
	public double absoluteMagnitude(double apparentMagnitude, double redshift) {
		double distanceModulus = distanceModulus(redshift);
		double retval = apparentMagnitude - distanceModulus;
		return retval;
	}
	*/
	/**
	 * @param absoluteMagnitude
	 * @param redshift
	 * @return the apparent magnitude
	 */
	/*
	public double apparentMagnitude(double absoluteMagnitude, double redshift) {
		double distanceModulus = distanceModulus(redshift);
		double retval = absoluteMagnitude + distanceModulus;
		return retval;
	}
	*/

}
