package gov.fnal.eag.util.astronomy;


/**
 * simple implementation of a flat cosmology with lambda
 * @author stoughto
 *
 */
public class CosmoLambdaFlat extends Cosmo {
    private double lambda;
    
     /**
     * default constructor with a popular cosmology lambda=0.7
     */
    public CosmoLambdaFlat() {
    	this(0.7);
    }
    /**
     * specify lambda for this cosmology, for example, 0.7
     * @param lambda
     */
    public CosmoLambdaFlat(double lambda) {
        this.lambda = lambda;        
    }
    
  
    /**
     * @param z
     * @return angular distance, in units of c/H0
     */
    @Override
	public double angularDistance(double z) {
        SimpleFunction func1 = new SimpleFunction() // object of anon Function subclass
        {
           @Override
		public double f(double x)
           {
        	   return 1.0/Math.sqrt((1-lambda)*Math.pow(1+x, 3) + lambda);
           }
        };
        double integral = Integrate.qRomb(func1, 0, z, 1e-9);
        double retval = integral/(1+z);
        return retval;
    }


	/**
	 * @see gov.fnal.eag.util.astronomy.Cosmo#dVdz(double)
	 */
	@Override
	public double dVdz(double z) {
		return getdVdzFunc().f(z);
	}
 
	/**
	 * @see gov.fnal.eag.util.astronomy.Cosmo#getdVdzFunc()
	 */
	@Override
	public SimpleFunction getdVdzFunc() {
		SimpleFunction func = new SimpleFunction() {
			@Override
			public double f(double z) {
				double retval =  4 * Math.PI * Math.pow(dC(z), 2) / expansion(z);
				return retval;				
			}
		};
		return func;
	}
	private double expansion(double z) {
        return Math.sqrt(((1 - lambda) * Math.pow((1 + z), 3)) + lambda);
	}
	
	private double dC(double z) {
		return morph(z,0);
	}
	
	private double morph(double zmax, final double g) {
		SimpleFunction func = new SimpleFunction() {
	           @Override
			public double f(double z) {
	                return Math.pow(1+z,g)/expansion(z);
	            }
		};
		return Integrate.qRomb(func,0,zmax,1e-10);	       
	}
	
	/**
	 * @return a copy of this cosmology
	 */
	@Override
	public Cosmo copy() {
		CosmoLambdaFlat retval = new CosmoLambdaFlat(lambda);
		return retval;
	}
	/**
	 * @see gov.fnal.eag.util.astronomy.Cosmo#equals(gov.fnal.eag.util.astronomy.Cosmo)
	 */
	@Override
	public boolean equals(Cosmo that) {
		if (that instanceof CosmoLambdaFlat) {
			if (this.lambda == ((CosmoLambdaFlat) that).lambda) {
				return true;
			}
		}
		return false;
	}
 }