package gov.fnal.eag.util.astronomy;

import java.util.ArrayList;

/**
 * numerical integration
 * @author stoughto
 */
public class Integrate {
	
	 /**
	  * maximum number of interval halvings
	  */
	public static int STEP_MAX = 20;
	
	private static SimpleFunction xGaussian = new SimpleFunction() {
		@Override
		public double f(double x) {
			return x * Math.exp(-x * x / 2.0);
		}
	};


	private static SimpleFunction x3Gaussian = new SimpleFunction() {
		@Override
		public double f(double x) {
			return x * x * x * Math.exp(-x * x / 2.0);
		}
	};

	/**
	 * approximate the definite integral of func from a to b by Romberg's method with maximum number of calls=20
	 * @param func
	 * @param a
	 * @param b
	 * @param eps
	 * @return the approximation
	 */
	public static double qRomb(SimpleFunction func, double a, double b, double eps) {
		return qRomb(func, a, b, eps, 20);
	}
	/**
	 * approximate the definite integral of func from a to b by Romberg's method
	 * @param func
	 * @param a
	 * @param b
	 * @param eps the desired accuracy
	 * @param qRombNMax maximum number of calls
	 * @return the approximation
	 *
	 * shamelessly copied algorithm from Wikipedia http://en.wikipedia.org/wiki/Romberg's_method

def print_row(lst):
    print ' '.join('%11.8f' % x for x in lst)
 
def romberg(f, a, b, eps = 1E-8):
    """Approximate the definite integral of f from a to b by Romberg's method.
    eps is the desired accuracy."""
    R = [[0.5 * (b - a) * (f(a) + f(b))]]  # R[0][0]
    print_row(R[0])
    n = 1
    while True:
        h = float(b-a)/2**n
        R.append((n+1)*[None])  # Add an empty row.
        R[n][0] = 0.5*R[n-1][0] + h*sum(f(a+(2*k-1)*h) for k in range(1, 2**(n-1)+1)) # for proper limits
        for m in range(1, n+1):
            R[n][m] = R[n][m-1] + (R[n][m-1] - R[n-1][m-1]) / (4**m - 1)
        print_row(R[n])
        if abs(R[n][n-1] - R[n][n]) < eps:
            return R[n][n]
        n += 1
 
from math import sqrt,exp
 
# In this example, the error function erf(1) is evaluated.
print romberg(lambda t: 2/sqrt(pi)*exp(-t*t), 0, 1)

	 * 
	 */
	public static double qRomb(SimpleFunction func, double a, double b, double eps, int qRombNMax) {
		ArrayList<double[]> r = new ArrayList<double[]>();
		double r00 = 0.5 * (b-a) * (func.f(a) + func.f(b));
		//System.out.println(String.format("%.8f", r00));
		r.add(new double[] {r00});
		int n=1;
		while (true) {
			double h = (b-a)/(1<<n);
			double[] row = new double[n+1];
			double sum = 0;
			int kMin = 1;
			int kMax = (1 << (n-1)) + 1;
			for (int k=kMin; k<kMax; k++) {
				double x = a + (2*k-1) * h;
				sum += func.f(x);
			}
			double[] rowNm1 = r.get(n-1);
			row[0] = 0.5 * rowNm1[0] + h*sum;
			for (int m = 1; m < n+1; m++) {
				row[m] = row[m-1] + (row[m-1] - rowNm1[m-1]) / (Math.pow(4, m) - 1);
				//System.out.print(String.format("%.8f ", row[m-1]));
			}
			//System.out.println(String.format("%.8f", row[n]));
			double delta = Math.abs(row[n-1] - row[n]);
			if (n > 3) {
				if (delta < eps/100.0) {
					return row[n];
				}
			}
			r.add(row);
			n++;
			if (n >= qRombNMax) {
				throw new Error("too many iterations:  n="+n+" qRombNMax="+qRombNMax);
			}
		}
	}

	/**
	 * integrate x*exp(-x^2) 
	 * @param a begin integral
	 * @param b end integral
	 * @param tol tolerance
	 * @return the integral
	 */
	public static double qRombXGaussian(double a, double b, double tol) {
		return qRomb(xGaussian, a, b, tol);
	}
	

	/**
	 * integrate (x^3)*exp(-x^2) 
	 * @param a begin integral
	 * @param b end integral
	 * @param tol tolerence
	 * @return the integral
	 */
	public static double qRombX3Gaussian(double a, double b, double tol) {
		return qRomb(x3Gaussian, a, b, tol);
	}
	

	/**
	 * @param function defined with inputs theta,phi in radians
	 * @param sinThetaMin -1.0 for entire sphere
	 * @param sinThetaMax 1.0 for entire sphere
	 * @param phiMin 0 for entire sphere
	 * @param phiMax 2*pi for entire sphere
	 * @param dSinTheta
	 * @param dPhi
	 * @return integral of function over the unit sphere in the specified range
	 * calculated by sampling the function at the center of each sinTheta,phi bin
	 */
	public static double integrateSphere(Simple2dFunction function, 
			double sinThetaMin, double sinThetaMax, 
			double phiMin, double phiMax, 
			double dSinTheta, double dPhi) {
		double retval = 0.0;
		int nSinTheta = (int) ((sinThetaMax-sinThetaMin)/dSinTheta);
		int nPhi = (int) ((phiMax-phiMin)/dPhi);
		for (int iSinTheta=0; iSinTheta<nSinTheta; iSinTheta++) {
			double sinTheta = sinThetaMin + (iSinTheta+0.5)*dSinTheta;
			double theta = Math.asin(sinTheta);
			for (int iPhi=0; iPhi<nPhi; iPhi++) {
				double phi = phiMin + (iPhi+0.5)*dPhi;
				double phiPrime = Math.atan(Math.tan(phi));
				if (phiPrime < 0) {
					phiPrime += 2*Math.PI;
				}
				retval += function.f(theta,phiPrime);
			}
		}
		double norm = dSinTheta*dPhi;
		retval *= norm;
		return retval;
	}
}

