package gov.fnal.eag.util.densityEstimate;

import flanagan.analysis.ErrorProp;
import gov.fnal.eag.util.misc.StringUtil;
import gov.fnal.eag.util.plot.JFreeChartUtil;
import gov.fnal.eag.util.yanny.Table;

import java.awt.geom.Point2D;
import java.util.ArrayList;

import org.freehep.math.minuit.MnMigrad;
import org.freehep.math.minuit.MnUserParameters;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.LogarithmicAxis;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

import uk.ac.starlink.table.DefaultValueInfo;
import uk.ac.starlink.table.DescribedValue;
import uk.ac.starlink.table.RowListStarTable;
import cern.colt.list.DoubleArrayList;

/**
 * is a Histogram with minimal functionality
 * @author stoughto
 *
 */
public class Histogram {
	String title, xAxisLabel, yAxisLabel;
	int nBin, nFill;
	double min, max, undefined, underflow, overflow, binSize;
	double[] contents;
	boolean includeZero = true;
	double fitChi2 = Double.NaN;
	boolean plotNormalizeByBinWidth = false;
	protected Histogram() {

	}
	/**
	 * @param includeZero if true, include zero in y axis when plotting
	 */
	public void setIncludeZero(boolean includeZero) {
		this.includeZero = includeZero;
	}
	/**
	 * @param titleIn
	 * @param xAxisLabelIn
	 * @param yAxisLabelIn
	 * @param nBinIn
	 * @param minIn
	 * @param maxIn
	 */
	public Histogram(String titleIn, String xAxisLabelIn, String yAxisLabelIn, 
			int nBinIn, double minIn, double maxIn) {
		init(titleIn, xAxisLabelIn, yAxisLabelIn, nBinIn, minIn, maxIn);
	}
	private void init(String titleIn, String xAxisLabelIn, String yAxisLabelIn, 
			int nBinIn, double minIn, double maxIn) {
		nBin = nBinIn;
		if (nBin <=0 ) {
			throw new Error("nBin must be > 0 but nBin="+nBin);
		}
		min = minIn;
		max = maxIn;
		if (min >= max) {
			max = minIn;
			min = maxIn;
		}
		setTitle(titleIn);
		xAxisLabel = xAxisLabelIn;
		yAxisLabel = yAxisLabelIn;
		binSize = (max-min)/nBin;
		underflow = 0;
		overflow = 0;
		contents =  new double[nBin];
	}
	/**
	 * constructs a histogram from an rlst generated by the getRlst method
	 * @param rlst
	 */
	public Histogram(RowListStarTable rlst) {
		int iMinIndex = Table.getIndexByName(rlst, "min");
		int iMaxIndex = Table.getIndexByName(rlst, "max");
		int iContentsIndex = Table.getIndexByName(rlst, "contents");
		int nBinIn = (int) rlst.getRowCount();
		double min0 = (Double) (rlst.getRow(0)[iMinIndex]);
		double max0 = (Double) (rlst.getRow(0)[iMaxIndex]);
		double minIn = min0;
		double maxIn = min0 + nBinIn*(max0-min0);
		init("","","", nBinIn, minIn, maxIn);
		for (int i=0; i<nBin; i++) {
			Object[] row = rlst.getRow(i);
			double binMin = (Double) (row[iMinIndex]);
			double binMax = (Double) (row[iMaxIndex]);
			double temp = binMax-binMin;
			if (Math.abs(temp - binSize) > 1e-12) {
				throw new IllegalArgumentException("variable bin size:  i="+i+" binSize="+binSize+" thisBinSize="+temp);
			}
			contents[i] = (Double) (row[iContentsIndex]);
		}
		// This causes trouble when dealing with rlst read from a file
		//title = (String) rlst.getParameterByName("title").getValue();
		//xAxisLabel = (String) rlst.getParameterByName("xAxisLabel").getValue();
		//yAxisLabel = (String) rlst.getParameterByName("yAxisLabel").getValue();
	}
	/**
	 * deep copy constructor
	 * @param source
	 */
	public Histogram(Histogram source) {
		nBin = source.nBin;
		min = source.min;
		max = source.max;
		title = source.title;
		xAxisLabel = source.xAxisLabel;
		yAxisLabel = source.yAxisLabel;
		binSize = source.binSize;
		underflow = source.underflow;
		overflow = source.overflow;
		contents = source.contents.clone();		
		nFill = source.nFill;
	}
	/**
	 * clear this histogram;
	 */
	public void clear() {
		underflow = 0.0;
		overflow = 0.0;
		nFill = 0;
		for (int i=0; i<contents.length; i++) {
			contents[i] = 0.0;
		}
	}
	/**
	 * increment the bin that contains x by weight
	 * @param x the position
	 * @param weight 
	 */
	public void fill(double x, double weight) {
		if (Double.isNaN(x) || Double.isInfinite(x)) {
			undefined+=weight;
		} else if (x < min) {
			underflow+=weight;
		} else if (x >= max) {
			overflow+=weight;
		} else {
			int iBin = getBin(x);
			contents[iBin]+=weight;
		}
		nFill++;
	}

	/**
	 * @param x
	 * @return the bin that contains x, or Integer.MIN_VALUE,MAX_VALUE if it underflow, overflow.
	 */
	public int getBin(double x) {
		if (x < min) {
			return Integer.MIN_VALUE;
		} else if (x >= max) {
			return Integer.MAX_VALUE;
		} 
		return (int) ((x-min)/binSize);
	}


	/**
	 * increment the bin that contains x by one
	 * @param x the position
	 */
	public void fill(double x) {
		fill(x, 1.0);
	}

	/**
	 * fill with the values in dal
	 * @param dal
	 */
	public void fill(DoubleArrayList dal) {
		for (int i=0; i<dal.size(); i++) {
			fill(dal.get(i));
		}
	}

	/**
	 * @param logarithmicRange true to make the y values save for logarithmic ranges
	 * @return a chart with default min and max domain
	 */
	public JFreeChart getChart(boolean logarithmicRange) {
		return getChart(min-binSize/2, max+binSize/2, logarithmicRange);
	}

	/**
	 * @param ff
	 * @return a table contains iBin, min, max, contents and, if ff is not null, the fit value for each bin
	 */
	public RowListStarTable getRlst(FitFunction ff) {
		String rlstDef = "int iBin; double min; double max; double contents";
		if (ff != null) {
			rlstDef += "; double fit";
		}
		RowListStarTable retval = Table.makeBlankTable(rlstDef);
		for (int iBin=0; iBin<getNBin(); iBin++) {			
			if (ff == null) {
				Object[] row = {iBin, getBinMin(iBin), getBinMax(iBin), getContents(iBin)};
				retval.addRow(row);
			} else {
				Object[] row = {iBin, getBinMin(iBin), getBinMax(iBin), getContents(iBin), ff.valueAt(getBinPos(iBin))};				
				retval.addRow(row);
			}
		}
		retval.setParameter(new DescribedValue(new DefaultValueInfo("title", String.class, "the title"), title));
		retval.setParameter(new DescribedValue(new DefaultValueInfo("xAxisLabel", String.class, "the xAxisLabel"), xAxisLabel));
		retval.setParameter(new DescribedValue(new DefaultValueInfo("yAxisLabel", String.class, "the yAxisLabel"), yAxisLabel));
		return retval;
	}
	/**
	 * 
	 * @param xPlotMin x minimum of the plot
	 * @param xPlotMax x maximum of the plot
	 * @param logarithmicRange true to make the range axis logarithmic
	 * @return a chart
	 */
	public JFreeChart getChart(double xPlotMin, double xPlotMax, boolean logarithmicRange) {
		Histogram[] list = {this};
		return Histogram.getChart(list, title, xAxisLabel, yAxisLabel, xPlotMin, xPlotMax, logarithmicRange);
	}

	/**
	 * @param list the Histograms to plot
	 * @param title title for the histogram
	 * @param logarithmicRange true if you want the y axis to be logarithmic
	 * @return the charg
	 */
	public static JFreeChart getChart(Histogram[] list, String title, boolean logarithmicRange) {
		String xAxisLabelTemp="";
		String yAxisLabelTemp="";
		if (list.length > 0) {
			xAxisLabelTemp=list[0].xAxisLabel;
			yAxisLabelTemp=list[0].yAxisLabel;
		}
		return getChart(list, title, xAxisLabelTemp, yAxisLabelTemp, Histogram.getXMin(list), Histogram.getXMax(list), logarithmicRange);
	}

	/**
	 * @param list the histograms
	 * @return the minumum of the x values for the histograms
	 */
	public static double getXMin(Histogram[] list) {
		double retval = Double.MAX_VALUE;
		for (Histogram hg : list) {
			retval = Math.min(retval, hg.min);
		}
		return retval;
	}
	/**
	 * @param list the histograms
	 * @return the maximum of the x values for the histograms
	 */
	public static double getXMax(Histogram[] list) {
		double retval = -Double.MAX_VALUE;
		for (Histogram hg : list) {
			retval = Math.max(retval, hg.max);
		}
		return retval;
	}
	/**
	 * @param list
	 * @param title
	 * @param xAxisLabel
	 * @param yAxisLabel
	 * @param xPlotMin
	 * @param xPlotMax
	 * @param logarithmicRange true to make the range axis logarithmic
	 * @return a JFreeChart suitable for plotting
	 */
	public static JFreeChart getChart(Histogram[] list, String title, String xAxisLabel, String yAxisLabel, 
			double xPlotMin, double xPlotMax, boolean logarithmicRange) {
		return getChart(list, null, title, xAxisLabel, yAxisLabel, xPlotMin, xPlotMax, logarithmicRange);
	}
	/**
	 * @param hg
	 * @param ff
	 * @param title
	 * @param xAxisLabel
	 * @param yAxisLabel
	 * @param xPlotMin
	 * @param xPlotMax
	 * @param logarithmicRange
	 * @return the chart
	 */
	public static JFreeChart getChart(Histogram hg, FitFunction ff, 
			String title, String xAxisLabel, String yAxisLabel, 
			double xPlotMin, double xPlotMax, boolean logarithmicRange) {
		return getChart(new Histogram[] {hg}, new FitFunction[] {ff}, title, xAxisLabel, yAxisLabel, xPlotMin, xPlotMax, logarithmicRange);
	}
	/**
	 * @param hg
	 * @param ff
	 * @param logarithmicRange
	 * @return the chart
	 */
	public static JFreeChart getChart(Histogram hg, FitFunction ff, boolean logarithmicRange) {
		return getChart(hg, ff, hg.title, hg.xAxisLabel, hg.yAxisLabel, Double.NaN, Double.NaN, logarithmicRange);
	}
	/**
	 * @param hgList Histograms to plot
	 * @param ffList FitFunctions to plot
	 * @param title
	 * @param xAxisLabel
	 * @param yAxisLabel
	 * @param xPlotMin set to NaN for auto range
	 * @param xPlotMax set to NaN for auto range
	 * @param logarithmicRange true to make the range axis logarithmic
	 * @return a JFreeChart suitable for plotting
	 */
	public static JFreeChart getChart(Histogram[] hgList, FitFunction[] ffList, 
			String title, String xAxisLabel, String yAxisLabel, double xPlotMin, double xPlotMax, boolean logarithmicRange) {
		boolean includeZero = false;

		XYSeriesCollection xySeriesCollection = new XYSeriesCollection();
		JFreeChart retval = ChartFactory.createScatterPlot(title, xAxisLabel, yAxisLabel, xySeriesCollection, PlotOrientation.VERTICAL, true, true, true);
		XYLineAndShapeRenderer renderer = (XYLineAndShapeRenderer) retval.getXYPlot().getRendererForDataset(xySeriesCollection);
		int i = 0;
		double minValue = getMin(logarithmicRange, hgList, ffList);		
		double maxValue =  getMax(hgList, ffList);
		double dLog = Math.log10(maxValue) - Math.log10(minValue);
		double fraction = 0.05;
		double rangeBelowMin = Math.pow(10, Math.log10(minValue) - dLog*(2*fraction/(1-2*fraction)));

		for (Histogram histogram : hgList) {
			includeZero = includeZero | histogram.includeZero;
			xySeriesCollection.addSeries(histogram.getXySeries(logarithmicRange, rangeBelowMin));
			renderer.setLinesVisible(true);
			renderer.setSeriesShapesVisible(i++, false);
		}
		if (ffList != null) {
			for (FitFunction ff : ffList) {
				if (ff != null) {
					if (hgList.length > 0) {
						XYSeries temp = hgList[0].getXySeries(ff, xPlotMin, xPlotMax, logarithmicRange, rangeBelowMin);
						xySeriesCollection.addSeries(temp);
						renderer.setLinesVisible(true);
						renderer.setSeriesShapesVisible(i++, false);				
					}
				}
			}
		}
		if (!Double.isNaN(xPlotMin) && !Double.isNaN(xPlotMax) && xPlotMin < xPlotMax) {
			retval.getXYPlot().getDomainAxis().setRange(xPlotMin, xPlotMax);
		}
		if (logarithmicRange) {
			LogarithmicAxis rangeAxis = new LogarithmicAxis(yAxisLabel);
			rangeAxis.setLog10TickLabelsFlag(true);
			double rangeMin = Math.pow(10, Math.log10(minValue) - dLog*fraction/(1-2*fraction));
			double rangeMax = Math.pow(10, Math.log10(maxValue) + dLog*fraction/(1-2*fraction));			
			rangeAxis.setLowerBound(rangeMin);
			rangeAxis.setUpperBound(rangeMax);
			((XYPlot) retval.getPlot()).setRangeAxis(rangeAxis);			
		} else if (!includeZero) {
			// if none of the histograms want you to include zero, reset the range axis
			XYPlot plot = retval.getXYPlot();
			ValueAxis rangeAxis = plot.getRangeAxis();
			double delta = maxValue-minValue;
			if (delta > 0.0) {
				rangeAxis.setLowerBound(minValue-delta*0.05);
				rangeAxis.setUpperBound(maxValue+delta*0.05);
			}
		}
		return retval;
	}

	private static double getMax(Histogram[] list) {
		double retval = Double.NaN;
		for (Histogram histogram : list) {
			double max = histogram.getMaxValue();
			if (Double.isNaN(retval)) {
				retval = max;
			} if (!Double.isNaN(max)) {
				retval = Math.max(retval, max);
			}
		}
		return retval;	
	}

	private static double getMax(Histogram[] hgList, FitFunction[] ffList) {
		double retval = getMax(hgList);
		Point2D.Double domainRange = getDomainRange(hgList);
		if (ffList != null) {
			for (FitFunction ff : ffList) {
				double max = getMaxValue(domainRange, ff);
				if (!Double.isNaN(max)) {
					retval = Math.max(retval, max);
				}
			}
		}
		return retval;
	}
	private static double getMin(boolean logarithmicRange, Histogram[] hgList, FitFunction[] ffList) {
		double retval = Double.NaN;
		for (Histogram histogram : hgList) {
			double min = histogram.getMinValue(logarithmicRange);
			if (!Double.isNaN(min)) {
				if (Double.isNaN(retval)) {
					retval = min;
				} if (!Double.isNaN(min)) {
					retval = Math.min(retval, min);
				}
			}
		}
		Point2D.Double domainRange = getDomainRange(hgList);
		if (ffList != null) {
			for (FitFunction ff : ffList) {
				double min = getMinValue(domainRange, ff, logarithmicRange);
				if (Double.isNaN(retval)) {
					retval = min;
				}
				if (!Double.isNaN(min)) {
					retval = Math.min(retval, min);
				}
			}
		}
		return retval;
	}

	private static double getMinValue(Point2D.Double domainRange, FitFunction ff, boolean logarithmicRange) {		
		if (ff == null) {
			return Double.NaN;
		}
		double retval = Double.MAX_VALUE;
		int nToDo = 100;
		double delta = (domainRange.y - domainRange.x)/nToDo;
		for (double i=0; i<nToDo; i++) {
			double x = domainRange.x + i*delta;
			double value = ff.valueAt(x);
			if (!Double.isNaN(value)) {
				if (!logarithmicRange || value > Double.MIN_VALUE) {
					retval = Math.min(retval, value);
				}
			}
		}
		return retval;
	}

	private static double getMaxValue(Point2D.Double domainRange, FitFunction ff) {	
		if (ff == null) {
			return Double.NaN;
		}
		double retval = -Double.MAX_VALUE;
		double delta = domainRange.y - domainRange.x;
		for (double i=0; i<100; i++) {
			double x = domainRange.x + i*delta;
			double value = ff.valueAt(x);
			retval = Math.max(retval, value);
		}
		return retval;
	}

	private static Point2D.Double getDomainRange(Histogram[] hgList) {
		double min = Double.MAX_VALUE;
		double max = -min;
		for (Histogram hg : hgList) {
			min = Math.min(min, hg.min);
			max = Math.max(max, hg.max);
		}
		Point2D.Double retval = new Point2D.Double(min, max);
		return retval;
	}
	/**
	 * @param logarithmicRange
	 * @return the minimum value, excluding zero if logarithmicRange is true
	 */
	public double getMinValue(boolean logarithmicRange) {
		double retval = Double.NaN;
		for (int iBin=0; iBin<nBin; iBin++) {
			double value = contents[iBin];
			if (plotNormalizeByBinWidth) {
				value /= getBinWidth(iBin);
			}
			if (!Double.isNaN(value)) {
				if (!(value > retval)) {
					if ( !logarithmicRange || value > Double.MIN_VALUE) {
						retval = value;
					}
				}
			}
		}
		return retval;		
	}
	/**
	 * @param iBin
	 * @return the width of this bin
	 */
	public double getBinWidth(int iBin) {
		return binSize;
	}
	private double getMaxValue() {
		double retval = Double.NaN;
		for (int iBin=0; iBin<nBin; iBin++) {
			double value = contents[iBin];
			if (plotNormalizeByBinWidth) {
				value /= getBinWidth(iBin);
			}
			if (!Double.isNaN(value)) {
				if (!(value < retval)) {
					retval = value;
				}
			}
		}
		return retval;		
	}
	/**
	 * @param logarithmicRange true to make the range axis logarithmic
	 * @param minValue for a logarithmic scale, the minimum value to use
	 * @return the series of points to plot the bins and values for this histogram
	 */
	public XYSeries getXySeries(boolean logarithmicRange, double minValue) {
		XYSeries xySeries = new XYSeries(title);
		double y = safeForLog(logarithmicRange, 0.0, minValue);
		xySeries.add(min, y);
		for (int iBin=0; iBin<nBin; iBin++) {		
			double xMin = getBinMin(iBin);
			double xMax = getBinMax(iBin);
			double value = contents[iBin];
			if (plotNormalizeByBinWidth) {
				value /= getBinWidth(iBin);
			}
			y = safeForLog(logarithmicRange, value, minValue);
			xySeries.add(xMin, y);
			xySeries.add(xMax, y);
		}
		y = safeForLog(logarithmicRange, 0.0, minValue);
		xySeries.add(max, y);
		return xySeries;
	}

	private XYSeries getXySeries(FitFunction ff, double xPlotMin, double xPlotMax, boolean logarithmicRange, double minValue) {
		XYSeries xySeries = new XYSeries(ff.getName());
		//xySeries.add(min, safeForLog(logarithmicRange, 0.0, minValue));
		if (Double.isNaN(xPlotMin)) {
			xPlotMin = min;
		}
		if (Double.isNaN(xPlotMax)) {
			xPlotMax = max;
		}
		double delta = (xPlotMax-xPlotMin)/nBin/4;
		for (int i=0; i<nBin*4; i++) {
			double x=xPlotMin+i*delta;
			double y = ff.valueAt(x);			
			if (!plotNormalizeByBinWidth) {
				y *= binSize;
			}
			double ySafe = safeForLog(logarithmicRange, y, minValue);
			xySeries.add(x, ySafe);
		}
		return xySeries;
	}


	private double safeForLog(boolean logarithmicRange, double value, double minValue) {
		if (!logarithmicRange) {
			if (Double.isNaN(value)) {
				return -Double.MAX_VALUE;
			} 
			return value;
		}
		if (Double.isNaN(value) || value < minValue) {
			return minValue;
		}
		return value;
	}
	/**
	 * @param iBin
	 * @return the location of the center of the bin
	 */
	public double getBinPos(int iBin) {
		return min+(iBin+0.5)*binSize;
	}
	/**
	 * @param iBin
	 * @return the location of the minimum of the bin
	 */
	public double getBinMin(int iBin) {
		return min+iBin*binSize;
	}
	/**
	 * @param iBin
	 * @return the location of the maximum of the bin
	 */
	public double getBinMax(int iBin) {
		return min+(iBin+1)*binSize;
	}

	/**
	 * @return the mean of the histogram
	 */
	public double getMean() {
		double sumContents = 0;
		double sumXContents = 0;
		for (int iBin=0; iBin<nBin; iBin++) {
			if (!Double.isNaN(contents[iBin])) {
				sumContents+=contents[iBin];
				sumXContents += contents[iBin]*getBinPos(iBin);
			}
		}
		return sumXContents/sumContents;
	}

	/**
	 * @param mean
	 * @return sigma of the histogram about the mean
	 */
	public double getSigma(double mean) {
		double sumContents = 0;
		double sumXXContents = 0;
		for (int iBin=0; iBin<nBin; iBin++) {
			sumContents+=contents[iBin];
			sumXXContents += contents[iBin]*Math.pow(getBinPos(iBin)-mean,2);
		}
		return Math.sqrt(sumXXContents/sumContents);
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		double mean = getMean();
		double sigma = getSigma(mean);
		return String.format("%s mean=%f sigma=%f", title, mean, sigma);
	}

	/**
	 * @return the underflow value
	 */
	public double getUnderflow() {
		return underflow;	
	}
	/**
	 * @return the overvlow value
	 */
	public double getOverflow() {
		return overflow;	
	}
	/**
	 * @return the undefined value
	 */
	public double getUndefined() {
		return undefined;	
	}
	/**
	 * @return the number of times a fill() method was called
	 */
	public double getNFill() {
		return nFill;
	}

	/**
	 * @return the number of bins
	 */
	public int getNBin() {
		return nBin;
	}

	/**
	 * @return bin size
	 */
	public double getBinSize() {
		return (max-min)/nBin;
	}

	/**
	 * @return minimum x value of bins 
	 */
	public double getMin() {
		return min;
	}

	/**
	 * @return maximum x value of bins 
	 */
	public double getMax() {
		return max;
	}

	/**
	 * @return the error in the mean
	 */
	public double getMeanError() {
		double s = getSigma(getMean());
		double n = Math.sqrt(getNFill());
		return s/n;
	}

	/**
	 * @return the error of sigma^2
	 */
	public double getSigma2Error() {
		double retVal = Math.pow(getSigma(getMean()), 2)*Math.sqrt((2/getNFill())*(1-1/getNFill()));
		return retVal;
	}
	/**
	 * @return the error in sigma 
	 */
	public double getSigmaError() {
		double sigma = getSigma(getMean());
		double sigma2Error = getSigma2Error();
		ErrorProp sigErrorProp = ErrorProp.sqrt(new ErrorProp(sigma*sigma, sigma2Error));
		double retval = sigErrorProp.getError();
		return retval;
	}

	/**
	 * @return the sum of the contents of all bins
	 */
	public double getContents() {
		double retval = 0;
		for (int iBin=0; iBin<nBin; iBin++) {
			retval += contents[iBin];
		}		
		return retval;
	}

	/**	 * 
	 * @param iBin
	 * @return the contents of this bin
	 */
	public double getContents(int iBin) {
		return contents[iBin];
	}

	/**
	 * add the contents of source to this histogram, if they are similar
	 * @param source
	 */
	public void add(Histogram source) {
		if (similar(source)) {
			for (int i=0; i<nBin; i++) {
				contents[i] += source.contents[i];
			}
			underflow += source.underflow;
			overflow += source.overflow;
			nFill += source.nFill;
		} else {
			throw new IllegalArgumentException("histograms must be similar");
		}
	}
	/**
	 * @param source
	 * @return true if nBin, min, and max are equal for source and this histograms
	 */
	public boolean similar(Histogram source) {
		if (nBin != source.nBin) return false;
		if (min != source.min) return false;
		if (max != source.max) return false;
		return true;
	}


	/**
	 * @return a gaussian fit to the histogram using chi2 normalization, poisson errors, and all parameters fit
	 */
	public FitFunctionGaussian fitGaussian() {
		return fitGaussian(false, false, false);
	}
	/**
	 * @param nGaussians
	 * @return the resulting fit
	 */
	public FitFunctionGaussians fitGaussians(int nGaussians) {
		return fitGaussians(nGaussians, false, false, false);
	}

	/**
	 * @param nGaussians
	 * @param fixMeans
	 * @param fixSigmas
	 * @param fixNormalizations
	 * @return the resulting fit
	 */
	public FitFunctionGaussians fitGaussians(int nGaussians, boolean fixMeans, boolean fixSigmas, boolean fixNormalizations) {
		if (nGaussians < 1) {
			throw new IllegalArgumentException("nGaussians="+nGaussians+" needs to be 1 or larger");
		}
		Histogram hgCopy = new Histogram(this);
		
		double[] parameters = new double[3*nGaussians];
		double[] errors = new double[3*nGaussians];
		for (int i=0; i<nGaussians; i++) {
			{
				JFreeChart chart = Histogram.getChart(hgCopy, null, false);
				JFreeChartUtil.toPng(chart, StringUtil.getFileName("-"+i, "png"));
			}
			FitFunctionGaussian ffg = hgCopy.fitGaussian(fixMeans, fixSigmas, fixNormalizations);
			{
				JFreeChart chart = Histogram.getChart(hgCopy, ffg, false);
				JFreeChartUtil.toPng(chart, StringUtil.getFileName("-fit-"+i, "png"));
			}
			for (int j=0; j<3; j++) {
				parameters[3*i+j] = ffg.getParameters()[j];
				errors[3*i+j] = ffg.getErrors()[j];
			}
			for (int iBin=0; iBin<hgCopy.getNBin(); iBin++) {
				double integral = ffg.integrate(hgCopy.getBinMin(iBin), hgCopy.getBinMax(iBin));				
				double binWidth = hgCopy.getBinWidth(iBin);
				double integratedContents = integral*binWidth;
				double value = Math.min(integratedContents, hgCopy.getContents(iBin));
				hgCopy.fill(hgCopy.getBinPos(iBin), -value);
			}
		}
		FitFunctionGaussians retval = new FitFunctionGaussians(nGaussians);
		Chi2 chi2 = new Chi2(this, retval);
		MnUserParameters upar = new MnUserParameters();
		for (int i=0; i<nGaussians; i++) {
			double mean = parameters[3*i];
			double meanError = parameters[3*i];
			upar.add("mean_"+i, mean, meanError);
			if (fixMeans) {
				upar.fix(3*i);
			}
			double sigma = parameters[3*i+1];
			double sigmaError = errors[3*i+1];
			upar.add("sigma_"+i, sigma, sigmaError);
			if (fixSigmas) {
				upar.fix(3*i+1);
			}
			double norm = parameters[3*i+2];
			double normError = errors[3*i+2];
			upar.add("normalization+"+i, norm, normError);
			if (fixNormalizations) {
				upar.fix(3*i+2);
			}
		}
		MnMigrad migrad = new MnMigrad(chi2, upar);
		migrad.setUseAnalyticalDerivatives(false);
		retval.functionMinimum = migrad.minimize();
		MnUserParameters up = retval.functionMinimum.userParameters();
		for (int i=0; i<nGaussians; i++) {
			for (int j=0; j<3; j++) {
				parameters[3*i+j] = up.value(3*i+j);
				errors[3*i+j] = up.error(3*i+j);
			}
		}
		retval.setParameters(parameters);
		retval.setErrors(errors);
		fitChi2 = chi2.valueOf(parameters);
		return retval;
	}
	/**
	 * @param fixMean 
	 * @param fixSigma 
	 * @param fixNormalization 
	 * @return a gaussian fit to the histogram using chi2 normalization and poisson errors
	 */
	public FitFunctionGaussian fitGaussian(boolean fixMean, boolean fixSigma, boolean fixNormalization) {
		FitFunctionGaussian ffg = new FitFunctionGaussian();
		Chi2 chi2 = new Chi2(this, ffg);

		MnUserParameters upar = new MnUserParameters();
		double mean = getMean();
		double meanError = getMeanError();
		upar.add("mean", mean, meanError);
		if (fixMean) {
			upar.fix(0);
		}

		double sigma = getSigma(mean);
		double sigmaError = getSigmaError();
		upar.add("sigma", sigma, sigmaError);
		if (fixSigma) {
			upar.fix(1);
		}

		double normalization = 0;
		for (int i=0; i<contents.length; i++) {
			normalization += contents[i];
		}
		double normalizationError = Math.sqrt(normalization);
		upar.add("normalization", normalization, normalizationError);
		if (fixNormalization) {
			upar.fix(2);
		}

		MnMigrad migrad = new MnMigrad(chi2, upar);
		migrad.setUseAnalyticalDerivatives(false);
		ffg.functionMinimum = migrad.minimize();
		MnUserParameters up = ffg.functionMinimum.userParameters();
		double[] answer = new double[3];
		double[] errors = new double[3];
		for (int i=0; i<3; i++) {
			answer[i] = up.value(i);
			errors[i] = up.error(i);
		}
		ffg.setParameters(answer);		
		ffg.setErrors(errors);		
		fitChi2 = chi2.valueOf(answer);
		return ffg;
	}
	/**
	 * fits a polynomial to the histogram
	 * @param order of the polynomial to fit
	 * @param poissonErrors whether to use sqrt(contents) as error in each bin, or to do an unweighted fit
	 * @param binsToUse each element is true to use the corresponding bin, or null to use all bins
	 * @return the fit polynomial
	 */
	public Polynomial1D fitPolynomial(int order, boolean poissonErrors, boolean[] binsToUse) {
		ArrayList<double[]> xys = new ArrayList<double[]>();
		for (int iBin=0; iBin<nBin; iBin++) {
			
			if (binsToUse == null || binsToUse[iBin]) {
			double x = getBinPos(iBin);
			double y = contents[iBin]/getBinWidth(iBin);
			if (poissonErrors) {
				if (y > 0) {
					double s = Math.sqrt(y);
					xys.add(new double[]{x,y,s});
				}
			} else {
				xys.add(new double[]{x,y});
			}
			}
		}
		Polynomial1D retval = new Polynomial1D(xys, order);
		return retval;
	}
	/**
	 * fits a polynomial to the histogram, using chi2
	 * @param order
	 * @return the result of the fit
	 */
	public FitFunctionPolynomial fitPolynomial2(int order) {
		FitFunctionPolynomial ffp = new FitFunctionPolynomial(order);
		Chi2 chi2 = new Chi2(this, ffp);
		MnUserParameters upar = new MnUserParameters();
		double sum = 0;
		int n = 0;
		for (int i=0; i<contents.length; i++) {
			if (!Double.isNaN(contents[i])) {
				sum += contents[i];
				n++;
			}			
		}
		double mean = sum/n;
		double meanError = Math.abs(0.1*mean);
		upar.add("a_0", mean, meanError);
		for (int i=1; i<=order; i++) {
			upar.add("a_"+i, 0.0, Math.pow(0.1,i));
		}
		MnMigrad migrad = new MnMigrad(chi2, upar);
		migrad.setUseAnalyticalDerivatives(false);
		/*FunctionMinimum min = */
		migrad.minimize();
		double[] answer = migrad.params();
		ffp.setParameters(answer);		
		ffp.setErrors(migrad.errors());
		fitChi2 = chi2.valueOf(answer);
		return ffp;		
	}
	/**
	 * fits an exponential to the histogram, using chi2
	 * @return the result of the fit
	 */
	public FitFunctionExponentials fitExponential() {
		FitFunctionExponentials ffp = new FitFunctionExponentials(1);

		int n=0;
		double sumx = 0.0;
		double sumy = 0.0;
		double sumx2 = 0.0;
		double sumxy = 0.0;

		for (int iBin=0; iBin<nBin; iBin++) {
			if (contents[iBin] > 0.0) {
				n++;
				double x = getBinPos(iBin);
				double y = Math.log(contents[iBin]);
				sumx += x;
				sumy += y;
				sumx2 += x*x;
				sumxy += x*y;
			}
		}
		if (n < 2) {
			throw new IllegalArgumentException("need 2 or more non-zero bins but n="+n);
		}
		double denom = n*sumx2 - sumx*sumx;
		if (Math.abs(denom) < 1e-15) {
			throw new IllegalArgumentException("n="+n+" sumx2="+sumx2+" sumx="+sumx+" denom="+denom);			
		}
		double intercept = (sumy*sumx2 - sumx*sumxy)/denom;
		double norm = Math.exp(intercept);

		double slope = (n*sumxy - sumx*sumy)/denom;
		double lambda = slope;

		Chi2 chi2 = new Chi2(this, ffp);
		MnUserParameters upar = new MnUserParameters();
		upar.add("lambda", lambda, 0.1*Math.abs(lambda));
		upar.add("norm", norm, 0.1*Math.abs(norm));
		MnMigrad migrad = new MnMigrad(chi2, upar);
		migrad.setUseAnalyticalDerivatives(false);
		/*FunctionMinimum min = */
		migrad.minimize();
		double[] answer = migrad.params();
		ffp.setParameters(answer);		
		fitChi2 = chi2.valueOf(answer);
		return ffp;		
	}
	/**
	 * set the title
	 * @param titleIn
	 */
	public void setTitle(String titleIn) {
		if (titleIn != null) {
			title = titleIn;
		} else {
			title = "";
		}
	}
	/**
	 * set the xAxisLabel
	 * @param xAxisLabelIn
	 */
	public void setXAxisLabel(String xAxisLabelIn) {
		if (xAxisLabelIn != null) {
			xAxisLabel = xAxisLabelIn;
		} else {
			xAxisLabel = "";
		}
	}
	/**
	 * set the yAxisLabel
	 * @param yAxisLabelIn
	 */
	public void setYAxisLabel(String yAxisLabelIn) {
		if (yAxisLabelIn != null) {
			yAxisLabel = yAxisLabelIn;
		} else {
			yAxisLabel = "";
		}
	}
	/**
	 * @return the title
	 */
	public String getTitle() {
		return title;
	}
	/**
	 * @return the contents of the bin with the minimum contents
	 */
	public double getMinContents() {
		double retVal = Double.MAX_VALUE;
		for (double content : contents) {
			retVal = Math.min(content, retVal);
		}
		return retVal;
	}
	/**
	 * @return the contents of the bin with the maximum contents
	 */
	public double getMaxContents() {
		double retVal = -Double.MAX_VALUE;
		for (double content : contents) {
			retVal = Math.max(content, retVal);
		}
		return retVal;
	}
	/**
	 * @param value multiply all contents by this value
	 */
	public void multiplyScalar(double value) {
		for (int i=0; i<contents.length; i++) {
			contents[i] *= value;
		}		
	}

	/**
	 * @param that
	 * @return true if the two histograms have the same title, xAxisLable, yAxisLable, nBin, min, max, and contents
	 */
	public boolean equals(Histogram that) {
		return equals(that, true);
	}

	/**
	 * @param that
	 * @param checkNames if true, insist that the tile and axisLabels are also equal
	 * @return whether the histograms are equivalent
	 */
	public boolean equals(Histogram that, boolean checkNames) {

		if (checkNames) {
			if (!title.equals(that.title)) {
				return false;
			}
			if (!xAxisLabel.equals(that.xAxisLabel)) {
				return false;
			}
			if (!yAxisLabel.equals(that.yAxisLabel)) {
				return false;
			}
		}
		if (nBin != that.nBin) {
			return false;
		}
		if (min != that.min) {
			return false;
		}
		if (max != that.max) {
			return false;
		}
		for (int i=0; i<nBin; i++) {
			if (contents[i] != that.contents[i]) {
				return false;
			}
		}
		return true;
	}
	/**
	 * @param name set the name of the return value 
	 * @param fixMean whether to fix the mean to the first approximation
	 * @param fixSigma whether to fix the sigma to the first approximation
	 * @param fixNormalization whether to fix the normalization to the first approximation
	 * @return the FitFunctionGaussian fit to this histogram
	 */
	public FitFunctionGaussian fitGaussian(String name, boolean fixMean, boolean fixSigma, boolean fixNormalization) {
		FitFunctionGaussian retval = fitGaussian(fixMean, fixSigma, fixNormalization);
		retval.setName(name);
		return retval;
	}
	/**
	 * @param name set the name of the returned function
	 * @return a gaussian fit to the data in the histogram
	 */
	public FitFunctionGaussian fitGaussian(String name) {
		return fitGaussian(name, false, false, false);
	}
	/**
	 * @param ff
	 * @param constantError constant error for all points, or Double.NaN to use fit value squared for the error in each bin
	 * @return chi2 of the function fitting the histogram, assuming gaussian statistics
	 */
	public double getChi2(FitFunction ff, double constantError) {
		double retval = 0;
		int nPoints = 0;
		for (int iBin=0; iBin<nBin; iBin++) {
			double y = contents[iBin];
			if (y > 0) {
				double x = getBinPos(iBin);
				double fit = ff.valueAt(x);
				double error2 = Math.pow(constantError, 2);
				if (Double.isNaN(error2)) {
					error2 = fit;
				}
				retval += Math.pow(y-fit, 2)/error2;
				nPoints++;
			}
		}
		//System.out.println("Histogram.getChi2:  retval="+retval+" nPoints="+nPoints);
		return retval;
	}

	/**
	 * add this constant to the contents of each bin
	 * @param c
	 */
	public void addConstant(double c) {
		for (int iBin=0; iBin<nBin; iBin++) {
			contents[iBin] += c;
		}
	}
	/**
	 * @return the label for the x axis
	 */
	public String getXAxisLabel() {
		return xAxisLabel;
	}
	/**
	 * @return the label for the x axis
	 */
	public String getYAxisLabel() {
		return yAxisLabel;
	}

	/**
	 * @return the chi2 value of the most recent fit, or Double.NaN if none has been done
	 */
	public double getFitChi2() {
		return fitChi2;
	}
}


