package gov.fnal.eag.util.densityEstimate;

import gov.fnal.eag.util.yanny.Table;
import uk.ac.starlink.table.RowListStarTable;
import cern.colt.list.DoubleArrayList;

/**
 * handles variable-sized bins for the Histogram
 * @author stoughto
 *
 */
public class HistogramVariableBinWidth extends Histogram {
	DoubleArrayList edges = new DoubleArrayList();
	/**
	 * @param title
	 * @param xAxisLabel
	 * @param values
	 * @param nBin
	 */
	public HistogramVariableBinWidth(String title, String xAxisLabel, DoubleArrayList values, int nBin) {
		int deltaBin = values.size()/nBin;
		if (deltaBin < 1) {
			throw new IllegalArgumentException("nBin="+nBin+" is greater than values.size="+values.size());
		}
		for (int i=0; i<values.size()-1; i++) {
			if (!(values.get(i) <= values.get(i+1))) {
				throw new IllegalArgumentException("values is not sorted:  i="+i+" values(i)="+values.get(i)+" values(i+1)="+values.get(i+1));
			}
		}
		DoubleArrayList edges = new DoubleArrayList();
		for (int iBin=0; iBin<nBin; iBin++) {
			double edge = values.get(iBin*deltaBin);
			edges.add(edge);
		}
		edges.add(values.get(values.size()-1));
		init(title, xAxisLabel, "dN/dx", edges);
		for (int i=0; i<values.size(); i++) {
			fill(values.get(i));
		}
	}
	/**
	 * @param rlst contains the information to construct a HistogramVariableBinWidth
	 */
	public HistogramVariableBinWidth(RowListStarTable rlst) {
		int iMinIndex = Table.getIndexByName(rlst, "min");
		int iMaxIndex = Table.getIndexByName(rlst, "max");
		int iContentsIndex = Table.getIndexByName(rlst, "contents");
		nBin = (int) rlst.getRowCount();
		contents = new double[nBin];
		for (int i=0; i<rlst.getRowCount(); i++) {
			Object[] row = rlst.getRow(i);
			double tempMin = (Double) (row[iMinIndex]);
			edges.add(tempMin);
			contents[i] = (Double) (row[iContentsIndex]);
			max = (Double) (row[iMaxIndex]);
		}
		
		edges.add(max);
		min = edges.get(0);
		undefined = 0;
		underflow = 0;
		binSize = Double.NaN;
		plotNormalizeByBinWidth = true;
	}
	/**
	 * @param title
	 * @param xAxisLabel
	 * @param yAxisLabel
	 * @param edges the location of the edges of the bins, starting with the min of the first bin and the max of the last bin
	 */
	public HistogramVariableBinWidth(String title, String xAxisLabel, String yAxisLabel, DoubleArrayList edges) {
		init(title, xAxisLabel, yAxisLabel, edges);
	}
	private void init (String title, String xAxisLabel, String yAxisLabel, DoubleArrayList edges) {
		this.title = title;
		this.xAxisLabel = xAxisLabel;
		this.yAxisLabel = yAxisLabel;

		for (int i=0; i<edges.size(); i++) {
			double value = edges.get(i);
			if (!this.edges.contains(value)) {
				this.edges.add(value);
			}
		}
		if (this.edges.size() < 2) {
			throw new IllegalArgumentException("edges does not contain two unique values");
		}
		this.edges.sort();
		nBin = this.edges.size()-1;
		contents = new double[nBin];
		this.min = this.edges.get(0);
		this.max = this.edges.get(contents.length);
		undefined = 0;
		underflow = 0;
		binSize = Double.NaN;
		plotNormalizeByBinWidth = true;
	}
	/**
	 * @param source
	 * @return true if nBin, min, and max are equal for source and this histograms
	 */
	public boolean similar(Histogram source) {
		if (source instanceof HistogramVariableBinWidth) {

			if (this.edges.size() != ((HistogramVariableBinWidth)source).edges.size()) return false;
			for (int i=0; i<this.edges.size(); i++) {
				double thisValue = this.edges.get(i);
				double sourceValue = ((HistogramVariableBinWidth)source).edges.get(i);
				if (thisValue != sourceValue) {
					return false;
				}
			}
		} else {
			return false;
		}
		return true;
	}
	/**
	 * @param x
	 * @return the bin that contains x, or Integer.MIN_VALUE,MAX_VALUE if it underflow, overflow.
	 */
	public int getBin(double x) {
		if (x < min) {
			return Integer.MIN_VALUE;
		} else if (x >= max) {
			return Integer.MAX_VALUE;
		} 
		int retval = edges.binarySearch(x);
		if (retval < 0) {
			retval = -retval-2;
		}
		return retval;
	}

	/**
	 * @see gov.fnal.eag.util.densityEstimate.Histogram#getBinMin(int)
	 */
	public double getBinMin(int iBin) {
		if (iBin < 0 || iBin >= getNBin()) {
			throw new IllegalArgumentException("iBin="+iBin+" nBin="+getNBin());
		}
		double retval = edges.get(iBin);
		return retval;
	}
	/**
	 * @see gov.fnal.eag.util.densityEstimate.Histogram#getBinMax(int)
	 */
	public double getBinMax(int iBin) {
		if (iBin < 0 || iBin >= getNBin()) {
			throw new IllegalArgumentException("iBin="+iBin+" nBin="+getNBin());
		}
		double retval = edges.get(iBin+1);
		return retval;
	}
	/**
	 * @see gov.fnal.eag.util.densityEstimate.Histogram#getBinPos(int)
	 */
	public double getBinPos(int iBin) {
		double retval = 0.5*(getBinMin(iBin)+getBinMax(iBin));
		return retval;		
	}
	/**
	 * @see gov.fnal.eag.util.densityEstimate.Histogram#getBinWidth(int)
	 */
	public double getBinWidth(int iBin) {
		double retval = getBinMax(iBin)-getBinMin(iBin);
		return retval;				
	}
	/**
	 * @return the edges of the bins of this 
	 */
	public DoubleArrayList getEdges() {
		return edges;
	}

}
