/**
 * 
 */
package gov.fnal.eag.util.densityEstimate;

import cern.colt.list.DoubleArrayList;
import cern.jet.stat.Descriptive;
import gov.fnal.eag.util.astronomy.SimpleFunction;

/**
 * Estimate the probability density function using kernel density estimation (Parzen Window)
 * See <cite>Smoothing Methods in Statistics</cite> (1996) 
 * by J. C. Simonoff or <a href="http://www.maths.uwa.edu.au/~duongt/seminars/intro2kde/">
 * Tarn Duong's tutorial</a> on kernel density estimation for a good introduction.
 * 
 * @author neilsen
 * @version $Name:  $ ($Id: KernelDensityEstimate.java,v 1.8 2006/09/26 20:02:45 stoughto Exp $)
 */
public class KernelDensityEstimate {

	private double min;
	private double max;
	private int nValues;
	private int nPoints;
	private SimpleFunction kernel;
	private double[] estimates;
	private double[] variances;
	
	/**
	 * Create a new density estimator
	 * 
	 * @param min the minimum value to estimate
	 * @param max the maximum value to estimate
	 * @param nValues the number of points to estimate
	 * @param kernel the kernel with which to estimate
	 */
	public KernelDensityEstimate(double min, double max, int nValues, SimpleFunction kernel) {
		this.min = min;
		this.max = max;
		this.nValues = nValues;
		this.kernel = kernel;
		estimates = new double[nValues];
		variances = new double[nValues];
		for (int i=0; i<nValues; i++) {
			estimates[i] = 0.0;
			variances[i] = 0.0;
		}
		nPoints = 0;
	}
	
	/**
	 * Create a new density estimator using a DoubleArrayList
	 * 
	 * @param list the array of all points to be added to the density
	 * @param nValues the number of points to estimate
	 */
	public KernelDensityEstimate(DoubleArrayList list, int nValues) {
		if (list.size() != 0) {
			this.min = Descriptive.min(list);
			this.max = Descriptive.max(list);
			this.nValues = nValues;
			this.kernel = new UniformKernel((max-min)/nValues);
			estimates = new double[nValues];
			variances = new double[nValues];
			for (int i=0; i<nValues; i++) {
				estimates[i] = 0.0;
				variances[i] = 0.0;
			}
			nPoints = 0;
		}
	}
	
	/**
	 * Return the point estimated by a given point index
	 * 
	 * @param index the index of the point
	 * @return the value estimated at that point
	 */
	public double x(int index) {
		double valueSpacing = (this.max-this.min)/this.nValues;
		double x = this.min + index*valueSpacing;
		return x;
	}

	/**
	 * Add a data point with a weight
	 * 
	 * @param value the value of the new data point
	 * @param weight the weight of the new data point
	 */
	public void add(double value, double weight) {
		nPoints += weight;
		double increment;
		for(int index = 0; index < nValues; index++) {
			increment = weight * kernel.f(value - x(index));
			estimates[index] += increment;
			variances[index] += increment * increment;
		}
	}
	
	/**
	 * Add a new data point with a weight of one
	 * 
	 * @param value the value of the new data point
	 */
	public void add(double value) {
		add(value,1.0);
	}
	
	/**
	 * Add a collection of new points
	 *  
	 * @param values the collection of new data points
	 */
	public void add(Iterable<Double> values) {
		for(double value: values) {
			add(value);
		}
	}
	
	/**
	 * Return an array of data values whose densities are estimated
	 * 
	 * @return data values whose densities are estimated
	 */
	public double[] getX() {
		double[] x = new double[nValues];
		for(int i=0; i<nValues; i++) {
			x[i] = x(i);
		}
		return x;
	}
	
	/**
	 * Return the density estimates
	 * 
	 * @return an array of density estimates
	 */
	public double[] getEstimates() {
		double[] densities = new double[nValues];
		for(int i=0; i<nValues; i++) {
			densities[i] = estimates[i] / nPoints;
		}
		return densities;
	}
	
	/**
	 * Return the density estimates
	 * 
	 * @return an array of density estimates
	 */
	public double[] getStandardDeviations() {
		double[] sigma = new double[nValues];
		for(int i=0; i<nValues; i++) {
			sigma[i] = Math.sqrt(variances[i]) / nPoints;
		}
		return sigma;
	}
	
	/**
	 * @return the mean
	 */
	public double getMean() {
		double valueSum = 0;
		double weightSum = 0;
		for (int i=0; i<nValues; i++) {
			double value = x(i);
			double weight = estimates[i];
			valueSum += value*weight;
			weightSum += weight;
		}		
		return valueSum/weightSum;
	}
	
	/**
	 * @return the sigma of this distribution
	 */
	public double getSigma() {
		double mean = getMean();
		double valueSum = 0;
		double weightSum = 0;
		for (int i=0; i<nValues; i++) {
			double value = Math.pow(x(i),2);
			double weight = estimates[i];
			valueSum += value*weight;
			weightSum += weight;
		}		
		double sigma = Math.sqrt(valueSum/weightSum - mean*mean);
		return sigma;
	}

	/**
	 * @return the minimum x value for this estimate
	 */
	public double getMin() {
		return min;
	}
	
	/**
	 * @return the maximum x value for this estimate
	 */
	public double getMax() {
		return max;
	}
	
	/**
	 * @return the sum of the weights
	 */
	public double getNPoints() {
		return nPoints;
	}
}
