package gov.fnal.eag.util.densityEstimate;

import java.awt.BasicStroke;
import java.awt.Color;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.annotations.XYLineAnnotation;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

import cern.colt.list.DoubleArrayList;

/**
 * plots a KernelDensityEstimate
 * @author stoughto
 *
 */
public class KernelDensityEstimatePlotter {	
	
	private String plotTitle = "";
	
	private String xLabel = "value"; // the label on the x axis
	
	private String yLabel = "estimate"; // the label on the y axis
	
	private int xSize = 500;
	
	private int ySize = 300;
	
	private Color meanColor;
	
	private Color sigmaColor;
	
	private ArrayList<KernelDensityEstimate> list = new ArrayList<KernelDensityEstimate>();
	
	private ArrayList<String> legends = new ArrayList<String>();
	
	/**
	 * construct a plotter for the kde
	 * @param kde
	 */
	public KernelDensityEstimatePlotter(KernelDensityEstimate kde) {
		this(kde, null, null,"");
	}
	
	/**
	 * construct a plotter for kde
	 * @param kde the KernelDensityEstimate to plot
	 * @param meanColor the color of the mean line
	 * @param sigmaColor the color of the sigma line
	 * @param legend for this kde
	 */
	public KernelDensityEstimatePlotter(KernelDensityEstimate kde, Color meanColor, Color sigmaColor, String legend) {
		list.add(kde);
		this.meanColor = meanColor;
		this.sigmaColor = sigmaColor;
		legends.add(legend);
	}
	
	/**
	 * @param list
	 * @param meanColor 
	 * @param sigmaColor 
	 * @param legendsIn 
	 */
	public KernelDensityEstimatePlotter(ArrayList<KernelDensityEstimate> list, Color meanColor, Color sigmaColor, ArrayList<String> legendsIn) {
		this.list = list;
		this.meanColor = meanColor;
		this.sigmaColor = sigmaColor;
		this.legends = legendsIn;
	}
	/**
	 * make a plot
	 * @param kde the KernelDenistyEstimate to plot
	 * @param fileName where to plot it
	 * @param legend 
	 */
	public static void toPng(KernelDensityEstimate kde, String fileName, String legend) {
		(new KernelDensityEstimatePlotter(kde, null, null, legend)).toPng(fileName);
	}

	/**
	 * write a plot
	 * @param fileName where to write it
	 */
	public void toPng(String fileName) {
		JFreeChart chart = getPlot();
		try {
			ChartUtilities.saveChartAsPNG(new File(fileName), chart, xSize, ySize);
		} catch (IOException e) {
			throw new Error(e);
		}		

	}
	/**
	 * Get a chart of this KernelDensityEstimate
	 * @return a chart of this KernelDensityEstimate
	 */
	public JFreeChart getPlot() {
		XYSeriesCollection xySeriesCollection = new XYSeriesCollection();
		DoubleArrayList max = new DoubleArrayList();
		for (int j=0; j<list.size(); j++) {
			// Create a data set to plot
			XYSeries xySeries = new XYSeries(legends.get(j));
			double[] y = list.get(j).getEstimates(); // get the x values to plot
			double[] x = list.get(j).getX(); // get the y value to plot
			max.add(0);
			double xMin = list.get(j).getMin();
			double xMax = list.get(j).getMax();
			for(int i=0; i<x.length; i++) {
				if (x[i] >= xMin && x[i]<=xMax)
					xySeries.add(x[i], y[i]);
				if (max.get(j) < y[i])
					max.set(j, y[i]);
			}
			// add each point to the data set
			for (int i=0; i < x.length; i++) {
				xySeries.add(x[i],y[i]);
			}
	
			// Create a set of data sets to add to the plot. We just have one
			// data set, so it is short
			xySeriesCollection.addSeries(xySeries);
		}
		// Create a chart object
		PlotOrientation orientation = PlotOrientation.VERTICAL; 
		boolean showLegend = true; // show the legend on the plot?
		boolean toolTips = false; // generate tooltips?
		boolean urls = false; // generate url's?
		
		JFreeChart chart = ChartFactory.createXYLineChart(
				plotTitle, xLabel, yLabel, 
				xySeriesCollection, 
				orientation, showLegend, toolTips, urls);
		for (int i=0; i<list.size(); i++) {
			if (meanColor != null) {
				XYPlot plot = chart.getXYPlot();
				plot.addAnnotation(new XYLineAnnotation(list.get(i).getMean(), 0, list.get(i).getMean(), max.get(i), new BasicStroke(), meanColor));
			}
			
			if (sigmaColor != null) {
				XYPlot plot = chart.getXYPlot();
				plot.addAnnotation(new XYLineAnnotation(list.get(i).getMean()-list.get(i).getSigma()*2.35/2, max.get(i)/2, list.get(i).getMean()+list.get(i).getSigma()*2.35/2, max.get(i)/2, new BasicStroke(), sigmaColor));
			}
		}
		
		return chart;
	}
	
	/**
	 * Set the title for plot
	 * @param plotTitle the title for plot
	 */
	public void setPlotTitle(String plotTitle) {
		this.plotTitle = plotTitle;
	}

	/**
	 * Get the label for the x-axis
	 * @return the label for the x-axis
	 */
	public String getXLabel() {
		return xLabel;
	}

	/**
	 * Get the size of the x-axis
	 * @return the size of the x-axis
	 */
	public int getXSize() {
		return xSize;
	}

	/**
	 * Get the label for the y-axis
	 * @return the label for the y-axis
	 */
	public String getYLabel() {
		return yLabel;
	}

	/**
	 * Get the size of the y-axis
	 * @return the size of the y-axis
	 */
	public int getYSize() {
		return ySize;
	}

	/**
	 * Set the label for the x-axis
	 * @param xLabel the label for the x-axis
	 */
	public void setXLabel(String xLabel) {
		this.xLabel = xLabel;
	}

	/**
	 * Set the size of the x-axis
	 * @param xSize  the size of the x-axis
	 */
	public void setXSize(int xSize) {
		this.xSize = xSize;
	}

	/**
	 * Set the label for y-axis
	 * @param yLabel the label for y-axis
	 */
	public void setYLabel(String yLabel) {
		this.yLabel = yLabel;
	}

	/**
	 * Set the size of the y-axis
	 * @param ySize the size of the y-axis
	 */
	public void setYSize(int ySize) {
		this.ySize = ySize;
	}
	
	
	
}
