package gov.fnal.eag.util.densityEstimate;

import gov.fnal.eag.util.plot.JFreeChartUtil;

import java.util.ArrayList;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.xy.XYDataItem;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.netlib.lapack.DGELS;
import org.netlib.lapack.DGGGLM;
import org.netlib.util.intW;



/**
 * describes a d scalar polynomial where
 * the value is given by the parameters and x, where
 * value = p[0] for order=0
 * value = p[0] + p[1]*x for order=1, and so on
 * 
 * @author stoughto
 *
 */
public class Polynomial1D extends FitFunction {
	int order;

	/**
	 * construct the the specified parameters
	 * @param params
	 */
	public Polynomial1D(double[] params) {
		if (params == null || params.length==0) {
			throw new IllegalArgumentException("params needs to have length > 0 but params="+params);
		}
		order = params.length-1;
		this.parameters = params.clone();
	}

	
	/**
	 * @param xy
	 * @param order
	 * @param plotFileName
	 * @param plotTitle
	 */
	public Polynomial1D(XYSeries xy, int order, String plotFileName, String plotTitle) {
		this(seriesToArrayList(xy), order, plotFileName, plotTitle);
	}
	private static ArrayList<double[]> seriesToArrayList(XYSeries xy) {
		ArrayList<double[]> retval = new ArrayList<double[]>();
		for (int i=0; i<xy.getItemCount(); i++) {
			XYDataItem xydi = xy.getDataItem(i);
			double[] temp = {xydi.getX().doubleValue(), xydi.getY().doubleValue()};
			retval.add(temp);
		}
		return retval;
	}

	/**
	 * constructs a fit to the data points
	 * @param xys x, y, and, optionally, sigmaValue for a weighted fit 
	 * @param order the order to fit
	 * @param plotFileName where to make a diagnostic plot; null to not make a plot
	 * @param plotTitle title to put in the diagnostic plot
	 */
	public Polynomial1D(ArrayList<double[]> xys, int order, String plotFileName, String plotTitle) {
		double[] paramIn = new double[order+1];
		for (int i=0; i<paramIn.length; i++) {
			paramIn[i] = Double.NaN;
		}
		init(xys, order, plotFileName, plotTitle, paramIn);
	}

	/**
	 * fully-specified constructor
	 * @param xys
	 * @param order
	 * @param plotFileName null for no plot
	 * @param plotTitle
	 * @param paramIn elements set to Double.NaN will be fit
	 */
	public Polynomial1D(ArrayList<double[]> xys, int order, String plotFileName, String plotTitle, double[] paramIn) {
		init(xys, order, plotFileName, plotTitle, paramIn);
	}

	private void init(ArrayList<double[]> xys, int order1, String plotFileName, String plotTitle, double[] paramIn) {
		if (order1 < 0) {
			throw new IllegalArgumentException("order must be >= 0 but order="+order);
		}
		
		if (xys.size() == 0) {
			throw new IllegalArgumentException("need at least one data point");
		}
		this.order = order1;
		int nCoef = (order+1);
		if (paramIn.length != nCoef) {
			throw new IllegalArgumentException("paramIn.length="+paramIn.length+" needs to equal order+1="+nCoef);
		}
		for (double param : paramIn) {
			if (!Double.isNaN(param)) {
				nCoef--;
			}
		}
		int nData = xys.size();
		if (nData < nCoef) {
			throw new IllegalArgumentException("nData must be >= nCoef but nData="+nData+" nCoef="+nCoef);
		}
		double[] first = xys.get(0);
		if (first.length == 2) {
			int nrhs = 1;
			double[][] a = new double[nCoef][nData];
			double[][] b = new double[nData][1];
			for (int iData=0; iData<nData; iData++) {
				double[] xyw = xys.get(iData);
				double x = xyw[0];
				double[] xVector = getXVector(x,paramIn);
				for (int iCoef=0; iCoef<nCoef; iCoef++) {
					a[iCoef][iData] = xVector[iCoef];
				}
				double y = xyw[1];
				for (int i=0; i<paramIn.length; i++) {
					if (!Double.isNaN(paramIn[i])) {
						y -= paramIn[i] * Math.pow(xyw[0], i);
					}
				}
				b[iData][0] = y;
			}
			int mn = Math.min(nData, nCoef);
			int lwork = Math.max(1, mn + Math.max(mn, nrhs));
			double[] work = new double[100];
			intW info = new intW(0);
			DGELS.DGELS("T", nCoef, nData, nrhs, a, b, work, lwork, info);

			parameters = new double[paramIn.length];
			int iCoef = 0;
			for (int i=0; i<parameters.length; i++) {
				if (Double.isNaN(paramIn[i])) {
					parameters[i] = b[iCoef][0];
					iCoef++;
				} else {
					parameters[i] = paramIn[i];
				}
			}
			if (plotFileName != null) {
				makePlot(xys, plotFileName, plotTitle, nData);
			}

		} else if (first.length == 3) {        
			// DGGGLM(int n, int m, int p, double[][] a, double[][] b, double[] d, double[] x, double[] y, double[] work, int lwork, intW info)
			double[][] a = new double[nData][nCoef];
			double[][] b = new double[nData][nData];
			double[] d = new double[nData];
			double[] yTemp = new double[nData];
			double[] work = new double[1];
			intW info = new intW(0);
			DGGGLM.DGGGLM(nData, nCoef, nData, a, b, d, parameters, yTemp, work, -1, info);
			int lwork = (int) work[0];
			work = new double[lwork];
			for (int iData=0; iData<nData; iData++) {
				double[] xyw = xys.get(iData);
				double x = xyw[0];
				double[] xVector = getXVector(x, paramIn);
				for (int iCoef=0; iCoef<nCoef; iCoef++) {
					a[iData][iCoef] = xVector[iCoef];
				}
				double y = xyw[1];
				for (int i=0; i<paramIn.length; i++) {
					if (!Double.isNaN(paramIn[i])) {
						y -= paramIn[i] * Math.pow(xyw[0], i);
					}
				}
				d[iData] = y;
				double s = xyw[2];
				b[iData][iData] = s*s;
			}			
			double[] someParams = new double[nCoef];
			DGGGLM.DGGGLM(nData, nCoef, nData, a, b, d, someParams, yTemp, work, lwork, info);
			parameters = new double[paramIn.length];
			int iSomeParam=0;
			for (int iParam=0; iParam<parameters.length; iParam++) {
				if (Double.isNaN(paramIn[iParam])) {
					parameters[iParam] = someParams[iSomeParam];
					iSomeParam++;
				} else {
					parameters[iParam] = paramIn[iParam];
				}
			}
			if (plotFileName != null) {
				makePlot(xys, plotFileName, plotTitle, nData);
			}
		} else {
			throw new IllegalArgumentException("elements of xyzws need to be dimension 2 for unweighted or 3 for weighted but length="+first.length);
		}
		name = String.format("Polynomial1D nOrder=%d",order);
		parameterNames = new String[paramIn.length];
		for (int i=0; i<parameterNames.length; i++) {
			parameterNames[i] = "a_"+i;
		}
		//System.out.println("after DGELS.DGELS:  info="+info.val);
	}

	private void makePlot(ArrayList<double[]> xys, String plotFileName,
			String plotTitle, int nData) {
		double xMin = Double.MAX_VALUE;
		double xMax = -xMin;
		XYSeries data = new XYSeries("data");
		for (int iData=0; iData<nData; iData++) {
			double[] xyw = xys.get(iData);
			double x = xyw[0];
			xMin = Math.min(x, xMin);
			xMax = Math.max(x, xMax);
			double y = xyw[1]-valueAt(x);
			data.add(x,y);
		}
		XYSeries fit = new XYSeries("fit");
		int nToDo = 100;
		double dx = (xMax-xMin)/nToDo;
		for (int i=0; i<=nToDo; i++) {
			double x = xMin + i*dx;
			double y = 0;
			fit.add(x,y);
		}
		XYSeriesCollection xysc = new XYSeriesCollection();
		xysc.addSeries(fit);
		xysc.addSeries(data);
		JFreeChart chart = ChartFactory.createScatterPlot(plotTitle,"x","y", xysc, PlotOrientation.VERTICAL, true, true, true);
		XYLineAndShapeRenderer xylsr = (XYLineAndShapeRenderer) chart.getXYPlot().getRenderer();
		xylsr.setSeriesLinesVisible(0, true);
		xylsr.setSeriesShapesVisible(0, false);
		xylsr.setSeriesLinesVisible(1, false);
		xylsr.setSeriesShapesVisible(1, true);
		JFreeChartUtil.toPng(chart, plotFileName);
	}


	/**
	 * @param xys list of x, y, and (optionally) sigma values
	 * @param order
	 */
	public Polynomial1D(ArrayList<double[]> xys, int order) {
		this(xys, order, null, null);
	}

	/**
	 * @return the order of this polynomial
	 */
	public int getOrder() {
		return order;
	}

	/**
	 * @param x
	 * @return the value at x
	 */
	public double valueAt(double x) {
		double[] xVector = getXVector(x, null);
		double retval = 0;
		for (int i=0; i<xVector.length; i++) {
			retval += parameters[i]*xVector[i];
		}
		return retval;
	}

	private double[] getXVector(double x, double[] paramIn) {
		int nCoef = order+1;
		if (paramIn != null) {
			for (double param : paramIn) {
				if (!Double.isNaN(param)) {
					nCoef--;
				}
			}
		}
		double[] retval = new double[nCoef];
		int iCoef = 0;
		for (int i=0; i<=order; i++) {
			if (paramIn == null || Double.isNaN(paramIn[i])) {
				retval[iCoef] = Math.pow(x, i);
				iCoef++;
			}
		}
		return retval;
	}
	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		String retVal = "Polynomial1D: ";
		for (int i=0; i<parameters.length; i++) {
			retVal += i+"="+parameters[i]+" ";
			retVal += i+"="+parameters[i]+" ";
		}
		return retVal;
	}
}
