package gov.fnal.eag.util.densityEstimate;


import java.util.ArrayList;

import org.freehep.math.minuit.FunctionMinimum;
import org.freehep.math.minuit.MnApplication;
import org.freehep.math.minuit.MnMinimize;
import org.freehep.math.minuit.MnUserParameters;
import org.netlib.lapack.DGELS;
import org.netlib.lapack.DGGGLM;
import org.netlib.util.intW;

/**
 * describes a 2d scalar field as a polynomial where
 * the value is given by the parameters, x, and y, where
 * value = p[0] for order=0
 * value = p[0] + p[1]*x + p[2]*y for order=1, and so on
 * 
 * @author stoughto
 *
 */
public class Polynomial2D {
	double[] params;
	double[] errs = null;
	int order;

	/**
	 * construct the the specified parameters
	 * @param params
	 */
	public Polynomial2D(double[] params) {
		double temp = (Math.sqrt(1+8*params.length)-3)/2;
		if (temp%1 != 0 || temp < 0) {
			throw new IllegalArgumentException("length of coefficient must be 1, 3, 6, 10, ... but length="+params.length);
		}
		order = (int) temp;
		this.params = params;
	}

	/**
	 * constructs a fit to the data points
	 * @param xyzs x, y, value, and, optionally, sigmaValue for a weighted fit 
	 * @param order the order to fit
	 */
	public Polynomial2D(ArrayList<double[]> xyzs, int order) {
		this(xyzs, order, false);
	}
	/**
	 * constructs a fit to the data points
	 * @param xyzs x, y, value, and, optionally, sigmaValue for a weighted fit 
	 * @param order the order to fit
	 * @param doChi2 
	 */
	public Polynomial2D(ArrayList<double[]> xyzs, int order, boolean doChi2) {
		if (order < 0) {
			throw new IllegalArgumentException("order must be >= 0 but order="+order);
		}
		if (xyzs.size() == 0) {
			throw new IllegalArgumentException("need at least one data point");
		}
		this.order = order;
		int nCoef = (order+1)*(order+2)/2;
		int nData = xyzs.size();
		if (nData < nCoef) {
			throw new IllegalArgumentException("nData must be >= nCoef but nData="+nData+" nCoef="+nCoef);
		}
		params = new double[nCoef];
		double[] first = xyzs.get(0);
		double[][] aCopy = null;
		double[][] bCopy = null;
		if (first.length == 3) {
			int nrhs = 1;
			double[][] a = new double[nCoef][nData];
			double[][] b = new double[nData][1];
			for (int iData=0; iData<nData; iData++) {
				double[] xyzw = xyzs.get(iData);
				double x = xyzw[0];
				double y = xyzw[1];
				double[] xVector = getXVector(x,y);
				for (int iCoef=0; iCoef<nCoef; iCoef++) {
					a[iCoef][iData] = xVector[iCoef];
				}
				double z = xyzw[2];
				b[iData][0] = z;
			}
			if (doChi2) {
				aCopy = new double[nCoef][nData];
				bCopy = new double[nData][1];
				for (int iData=0; iData<nData; iData++) {
					bCopy[iData][0] =  b[iData][0];
					for (int iCoef=0; iCoef<nCoef; iCoef++) {
						aCopy[iCoef][iData] = a[iCoef][iData];
					}
				}
			}

			int mn = Math.min(nData, nCoef);
			int lwork = Math.max(1, mn + Math.max(mn, nrhs));
			double[] work = new double[100];
			intW info = new intW(0);
			DGELS.DGELS("T", nCoef, nData, nrhs, a, b, work, lwork, info);
			for (int iCoef=0; iCoef<nCoef; iCoef++) {
				params[iCoef] = b[iCoef][0];
				//System.out.println("iCoef="+iCoef+" p="+params[iCoef]);
			}
		} else if (first.length == 4) {        
			// DGGGLM(int n, int m, int p, double[][] a, double[][] b, double[] d, double[] x, double[] y, double[] work, int lwork, intW info)
			double[][] a = new double[nData][nCoef];
			double[][] b = new double[nData][nData];
			double[] d = new double[nData];
			double[] yTemp = new double[nData];
			double[] work = new double[1];
			intW info = new intW(0);
			System.out.println("Polynomial2D:  nData="+nData+" nCoef="+nCoef);
			DGGGLM.DGGGLM(nData, nCoef, nData, a, b, d, params, yTemp, work, -1, info);
			int lwork = (int) work[0];
			work = new double[lwork];
			for (int iData=0; iData<nData; iData++) {
				double[] xyzw = xyzs.get(iData);
				double x = xyzw[0];
				double y = xyzw[1];
				double[] xVector = getXVector(x,y);
				for (int iCoef=0; iCoef<nCoef; iCoef++) {
					a[iData][iCoef] = xVector[iCoef];
				}
				double z = xyzw[2];
				d[iData] = z;
				double s = xyzw[3];
				b[iData][iData] = s*s;
			}			
			if (doChi2) {
				throw new Error("IMPLEMENT THIS");
			}
			DGGGLM.DGGGLM(nData, nCoef, nData, a, b, d, params, yTemp, work, lwork, info);
		} else {
			throw new IllegalArgumentException("elements of xyzws need to be dimension 3 for unweighted or 4 for weighted but length="+first.length);
		}
		if (doChi2) {
			LSChi2 lsChi2 = new LSChi2(aCopy, bCopy);
			MnUserParameters upar = new MnUserParameters();
			for (int i=0; i<params.length; i++) {
				upar.add(String.format("parameter_%d",i), params[i], 1.0);
			}
			MnApplication application = new MnMinimize(lsChi2, upar);
			
			FunctionMinimum fm = application.minimize();
			MnUserParameters up = fm.userParameters();
			
			errs = new double[params.length];
			for (int i=0; i<params.length; i++) {
				params[i] = up.value(i);
				errs[i] = up.error(i);
			}
		}
	}


	/**
	 * @return the order of this polynomial
	 */
	public int getOrder() {
		return order;
	}

	/**
	 * @return the parameters
	 */
	public double[] getParams() {
		return params;
	}
	
	/**
	 * @return the error array
	 */
	public double[] getErrs() {
		return errs;
	}
	/**
	 * @param x
	 * @param y
	 * @return the value at x,y
	 */
	public double value(double x, double y) {
		double[] xVector = getXVector(x, y);
		double retval = 0;
		for (int i=0; i<xVector.length; i++) {
			retval += params[i]*xVector[i];
		}
		return retval;
	}

	private double[] getXVector(double x, double y) {
		double[] xpow = new double[order+1];
		double[] ypow = new double[order+1];
		xpow[0] = 1;
		ypow[0] = 1;
		for (int i=1; i<=order; i++) {
			xpow[i] = xpow[i-1]*x;
			ypow[i] = ypow[i-1]*y;
		}
		int index = 0;
		int coefLength = (order+1)*(order+2)/2;
		double[] retval = new double[coefLength];
		for (int iOrder=0; iOrder<=order; iOrder++) {
			for (int jIndex=0; jIndex<=iOrder; jIndex++) {
				int iIndex = iOrder-jIndex;
				retval[index] = xpow[iIndex]*ypow[jIndex];
				index++;
			}
		}
		return retval;
	}
}
