package gov.fnal.eag.util.densityEstimate;

import java.util.ArrayList;

import org.netlib.lapack.DGELS;
import org.netlib.util.intW;

/**
 * describes a 2d scalar field as a polynomial
 * 
 * the value is given by the parameters, x, and y, where
 * r = sqrt(x^2+y^2)
 * value = p[0] for order=0
 * value = p[0] + p[1]*r  for order=1, and so on
 * value = p[0] + p[1]*r + p[2]*r^2 for order=2, and so on
 * 
 * @author stoughto
 *
 */
public class Polynomial2DRadial {
	double[] params;
	int order;
	/**
	 * construct the the specified parameters
	 * @param params
	 */
	public Polynomial2DRadial(double[] params) {
		if (params == null || params.length == 0) {
			throw new IllegalArgumentException("params must have length > 0 but params="+params);
		}
		order = params.length - 1;
		this.params = params;
	}

	/**
	 * constructs a fit to the data points
	 * @param xyzws x, y, and value 
	 * @param order the order to fit
	 */
	public Polynomial2DRadial(ArrayList<double[]> xyzws, int order) {
		if (order < 0) {
			throw new IllegalArgumentException("order must be >= 0 but order="+order);
		}
		this.order = order;
		int nCoef = order+1;
		int nData = xyzws.size();
		if (nData < nCoef) {
			throw new IllegalArgumentException("nData must be >= nCoef but nData="+nData+" nCoef="+nCoef);
		}
        int nrhs = 1;
        double[][] a = new double[nCoef][nData];
        double[][] b = new double[nData][1];
        for (int iData=0; iData<nData; iData++) {
        	double[] xyzw = xyzws.get(iData);
        	double x = xyzw[0];
        	double y = xyzw[1];
        	double[] xVector = getXVector(x,y);
        	for (int iCoef=0; iCoef<nCoef; iCoef++) {
        		a[iCoef][iData] = xVector[iCoef];
        	}
        	double z = xyzw[2];
        	b[iData][0] = z;
        }
        int mn = Math.min(nData, nCoef);
		int lwork = Math.max(1, mn + Math.max(mn, nrhs));
		double[] work = new double[100];
		intW info = new intW(0);
        DGELS.DGELS("T", nCoef, nData, nrhs, a, b, work, lwork, info);
        //System.out.println("after DGELS.DGELS:  info="+info.val);
        params = new double[nCoef];
        for (int iCoef=0; iCoef<nCoef; iCoef++) {
        	params[iCoef] = b[iCoef][0];
        	//System.out.println("iCoef="+iCoef+" p="+params[iCoef]);
        }
	}
	

	/**
	 * @return the order of this polynomial
	 */
	public int getOrder() {
		return order;
	}
	
	/**
	 * @return the parameters
	 */
	public double[] getParams() {
		return params;
	}
	/**
	 * @param x
	 * @param y
	 * @return the value at x,y
	 */
	public double value(double x, double y) {
		double[] xVector = getXVector(x, y);
		double retval = 0;
		for (int i=0; i<xVector.length; i++) {
			retval += params[i]*xVector[i];
		}
		return retval;
	}
	
	private double[] getXVector(double x, double y) {
		double[] rpow = new double[order+1];
		double r = Math.sqrt(x*x+y*y);
		rpow[0] = 1;
		for (int i=1; i<=order; i++) {
			rpow[i] = rpow[i-1]*r;
		}
		return rpow;
	}
}
