package gov.fnal.eag.util.htm;

/**
 * an circular area on the unit sphere, defined by a direction vector and distance of a 
 * a plane perpendicular to the direction that cuts the sphere to define the circle.  
 * 
 * A distance of zero yields a halph sphere, and negative distance is a hole where more than
 * half the sphere is included.
 * 
 * The default constructors do not include the points on the circle in the constraint.
 * 
 * @author stoughto
 *
 */

public class  Constraint {

	private Vector3d direction;			// normal vector
	private double        distance;			// distance from origin
	private double        angle;			// cone angle in radians
	private boolean includeEdge;


	/** Initialization constructor
    @param direction Vector3d specifying the direction of the cap
    @param distance The distance of the cutting plane from the origin
	 */
	public Constraint(Vector3d direction, double distance) {
		this(direction, distance, false);
	}
	/** Initialization constructor
	   @param direction Vector3d specifying the direction of the cap
	   @param distance The distance of the cutting plane from the origin
	 * @param includeEdge whether a point on the constraint is included
	 */
	public Constraint(Vector3d direction, double distance, boolean includeEdge) {
		this.direction=direction;
		setDistance(distance);
		this.direction.normalize();
		angle = Math.acos(distance);
		this.includeEdge = includeEdge;
	}

	/** Initialization constructor with includeEdge=false
     * @param x x-direction of vector  specifying the direction of the cap
     * @param y y-direction of vector  specifying the direction of the cap
     * @param z z-direction of vector  specifying the direction of the cap
     * @param distance The distance of the cutting plane from the origin
	 */
	public Constraint(double x, double y, double z, double distance) {
		this(x, y, z, distance, false);
	}
	
	/** Initialization constructor
     * @param x x-direction of vector  specifying the direction of the cap
     * @param y y-direction of vector  specifying the direction of the cap
     * @param z z-direction of vector  specifying the direction of the cap
     * @param distance The distance of the cutting plane from the origin
	 * @param includeEdge 
	 */
	public Constraint(double x, double y, double z, double distance, boolean includeEdge) {
		this(new Vector3d(x,y,z), distance, includeEdge);
	}

	/** Copy constructor
	 * @param original
	 */
	public Constraint( Constraint original) {
		direction=original.direction;
		setDistance(original.distance);
		angle=original.angle;
		includeEdge=original.includeEdge;
	}


	/** Invert a constraint */
	public void invert() {
		direction = direction.mul(-1);
		setDistance(-distance);
		includeEdge = !includeEdge;
	}

	/** check whether a vector is inside this constraint
      @param vector Vector3d to be checked
      @return True if inside constraint, False if outside
	 */
	public boolean contains(Vector3d vector) {
		double dot = vector.mul(direction);
		double theta = Math.acos(dot);
		if (includeEdge) {
			if ( theta <= angle ) return true;
		} else {
			if ( theta < angle ) return true;			
		}
		return false;
	}

	/** convert data to string
      @return string containing vector and distance (4 numbers)
	 */
	@Override
	public String toString() {
		return "direction="+direction.toString()+" distance="+distance;
	}
	/**
	 * set the distance from the origin
	 * @param d the distance
	 */
	public void setDistance(double d) {
		distance = d;
		angle = Math.acos(distance);		
	}
	/**
	 * set the direction vector
	 * @param direction
	 */
	public void setDirection(Vector3d direction) {
		this.direction = direction;
	}
	/**
	 * set the opening angle
	 * @param angle opening angle in radians
	 */
	public void setAngle(double angle) {
		this.angle = angle;
		distance = Math.cos(angle);
	}

	/**
	 * @return the normal vector
	 */
	public Vector3d getDirection() {
		return direction;
	}
	/**
	 * @return cone angle in radians
	 */
	public double getAngle() {
		return angle;
	}

	/**
	 * @return distance from origin
	 */
	public double getDistance() {
		return distance;
	}

	/**
	 * @return whether the edge is included
	 */
	public boolean isIncludeEdge() {
		return includeEdge;
	}
	/**
	 * @param includeEdge true to include edges.
	 */
	public void setIncludeEdge(boolean includeEdge) {
		this.includeEdge = includeEdge;
	}
	/**
	 * is this equal to that?
	 * @param that
	 * @return true if the direction and angle are equal
	 */
	public boolean equals(Constraint that) {
		if (!direction.equals(that.direction)) return false;
		if (angle != that.angle) return false;
		if (includeEdge != that.includeEdge) return false;
		return true;
	}
}

