package gov.fnal.eag.util.htm;

import gov.fnal.eag.wcs.SphericalCoordinate;

import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

/**
 * describes a region on the sphere that is convex.  It can be either one small circle or a set of points connected by arcs of great circles.
 * @author stoughto
 *
 */
public class  ConvexRegion {
	List<Constraint>  constraints_ = new ArrayList<Constraint>(); // The vector of constraints
	ArrayList<Vector3d>  corners_	 = new ArrayList<Vector3d>(); // The corners of a zERO convex
	Constraint boundingCircle_;	  // For zERO convexes, the bc.
	//protected int olevel = 20; // output level for the ranges

	/**
	 * a region with no constraints
	 */
	public ConvexRegion() {
		
	}
	
	private static ArrayList<Vector3d> toArrayList(Vector3d[] vertices) {
		 ArrayList<Vector3d> retval = new  ArrayList<Vector3d>();
		 for (Vector3d v : vertices) {
			 retval.add(v);
		 }
		 return retval;
	}
	
	/**
	 * constructs the ConvexRegion that contains all of the vertices
	 * @param vertices 
	 */
	public ConvexRegion(Vector3d[] vertices) {
		init(ConvexRegion.toArrayList(vertices));
	}
	/**
	 * constructs the ConvexRegion that contains all of the vertices
	 * @param vertices 
	 */
	public ConvexRegion(ArrayList<Vector3d> vertices) {
		init(vertices);
	}
	
	/**
	 * constructs the ConvexRegion of one small circle
	 * @param direction
	 * @param openingAngle degrees
	 */
	public ConvexRegion(Vector3d direction, double openingAngle) {
		double distance = Math.cos(Math.toRadians(openingAngle));
		Constraint constraint = new Constraint(direction, distance);
		addConstraint(constraint);
		boundingCircle_ = constraint;
		
	}
	private void addConstraint(Constraint c) {
		constraints_.add(c);
		boundingCircle_ = constraints_.get(constraints_.size()-1);		
	}
	private void init(ArrayList<Vector3d> allVertices) { 
		corners_ = ConvexRegion.orderVerticies(allVertices);
		if (corners_.size() > 1) {
			for (int i=0; i<corners_.size(); i++) {
				Vector3d v0 = corners_.get(i);
				try {
					Vector3d v1 = corners_.get(i+1);
					addConstraint(new Constraint(v1.cross(v0), 0));				
				} catch (IndexOutOfBoundsException e) {
					if (corners_.size() > 2) { // do not close the polygon if there are only two vertices
						Vector3d v1 = corners_.get(0);
						addConstraint(new Constraint(v1.cross(v0), 0));				
					}
				}
			}
			boundingCircle_ = constraints_.get(0);
		}
		setIncludeEdges(true);
	}
	
	/**
	 * calls setIncludeEdgs(includeEdges) on all the constraints
	 * @param includeEdges
	 */
	public void setIncludeEdges(boolean includeEdges) {
		for (Constraint constraint : constraints_) {
			constraint.setIncludeEdge(includeEdges);
		}
	}
	/** Constructor from a triangle
	Initialize domain from a triangle. The corners of these vectors
	form a triangle, so we just add three ZERO convexes to the domain
	where the direction is given by the cross product of the corners.
	Of course, the sign has to be determined (on which side of the triangle
	are we?) If the three points lie on one line, no convexes are added.
	 * @param v1 normalized vector of the first corner
	 * @param v2 normalized vector of the second corner
	 * @param v3 normalized vector of the third corfer
	 */
	public ConvexRegion( Vector3d  v1, Vector3d  v2, Vector3d  v3) {
		this(new Vector3d[] {v1, v2, v3});
	}


	/** Constructor for a rectangle
	Initialize convex from a rectangle. The vectors that form a rectangle
	may be in any order, the code finds the edges by itself.
	If one of the vectors lies within the triangle formed by the
	other three vectors, the previous constructor is used.
	 * @param v1 
	 * @param v2 
	 * @param v3 
	 * @param v4 
	 */
	public ConvexRegion( Vector3d  v1,
			Vector3d  v2,
			Vector3d  v3,
			Vector3d  v4) {
		this(new Vector3d[] {v1, v2, v3, v4});
	}
	
	/**
	 * @param min min ra,dec point
	 * @param max max ra,dec point
	 */
	public ConvexRegion(SphericalCoordinate min, SphericalCoordinate max) {
		ArrayList<Vector3d> al = new ArrayList<Vector3d>();
		al.add(new Vector3d(min.getLongitude(), min.getLatitude()));
		al.add(new Vector3d(max.getLongitude(), min.getLatitude()));
		al.add(new Vector3d(max.getLongitude(), max.getLatitude()));
		al.add(new Vector3d(min.getLongitude(), max.getLatitude()));
		init(al);
	}
	
	/**
	 * makes a new list of verticies in the correct order for making a ConvexRegion
	 * @param verticiesOriginal
	 * @return the list.
	 */
	public static ArrayList<Vector3d> orderVerticies(ArrayList<Vector3d> verticiesOriginal) {
		ArrayList<Vector3d> verticies = new ArrayList<Vector3d>();
		for (Vector3d originalVertex : verticiesOriginal) {
			boolean isUnique = true;
			for (Vector3d savedVertex : verticies) {
				if (originalVertex.mul(savedVertex) > 1 - HTMfunc.Epsilon) {
					isUnique = false;
				}
			}
			if (isUnique) {
				verticies.add(originalVertex);
			}
		}
		while (verticies.size() > 2) {
			// Find a pair of verticies such that all the other verticies are on the right side of them.
			// Along the way, remove the middle vertex of three that are in line
			boolean vertexRemoved = false;
			for (int i=0; /* i<verticies.size()-1 */; i++) { // the clause is never false
				Vector3d vi = verticies.get(i);
				for (int j=i+1; j<verticies.size(); j++) {
					Vector3d vj =  verticies.get(j);
					boolean allOnTheRight = true;
					for (int k=0; k<verticies.size(); k++) {
						if ((k==i) || k==j) {
							continue;
						}
						Vector3d vk = verticies.get(k);
						double theta = Vector3d.getDirection(vi, vj, vk);
						if (Math.abs(Math.sin(Math.toRadians(theta))) < HTMfunc.Epsilon) { // three in a row
							if (Math.abs(theta) < 10*HTMfunc.Epsilon) {
								vertexRemoved = true;
								verticies.remove(j);
								break;
							} else if (Math.abs(theta+180) < 10*HTMfunc.Epsilon) {
								vertexRemoved = true;
								verticies.remove(k);
								break;
							} else {
								vertexRemoved = true;
								verticies.remove(i);
								break;
							}
						}
						if (theta > 180) {
							allOnTheRight = false;
							break;
						}						
					}
					if (vertexRemoved) {
						break;
					}
					if (allOnTheRight) {
						ArrayList<Vector3d> retval = new ArrayList<Vector3d>();
						retval.add(vi);
						retval.add(vj);
						verticies.remove(j);
						boolean moreToDo = true;
						do {
							double minTheta = 180;
							Vector3d vMin = vi;
							for (int k=0; k<verticies.size(); k++) {
								Vector3d vk = verticies.get(k);
								if (vk != vi) {
									double theta = Vector3d.getDirection(vi, vj, vk);
									if (theta < minTheta) {
										minTheta = theta;
										vMin = vk;
									}
								}
							}
							if (vMin == retval.get(0)) {
								moreToDo = false; // when you find the original vertex you are done
							} else {
								retval.add(vMin);
								verticies.remove(vMin);
								vi = retval.get(retval.size()-2);
								vj = retval.get(retval.size()-1);						
							}
						} while (moreToDo);
						return retval;
					}
				}
				if (vertexRemoved) {
					break;
				}
			}
		}
		// If there are two points, and they are at antipodes, remove them
		if (verticies.size() == 2) {
			double dot = verticies.get(0).mul(verticies.get(1));
			if (dot < -1+HTMfunc.Epsilon) {
				verticies.clear();
			}
		}
		return verticies;
	}

	/**
	 * describes the four ways a ConvexRegion and a triangle specified by an HTMid overlap
	 * @author stoughto
	 *
	 */
	public enum Overlap {
		/**
		 * the triangle is completely inside the ConvexRegion
		 */
		ConvexRegionContainsHTMId,
		/**
		 * there is no overlap between the two
		 */
		NoOverlap,
		/**
		 * the convex region is completely inside the triangle
		 */
		HTMIdContainsConvexRegion,
		
		/**
		 * there is partial overlap between the two
		 */
		PartialOverlap
	}
	
	/**
	 * describes how this ConvexRegion and a triangle overlap
	 * @param HTMId defines the triangle
	 * @return the Overlap
	 */
	public Overlap getOverlap(long HTMId) {
		// if there are no constraints there is nothing to overlap
		if (constraints_.size() == 0) {
			return Overlap.NoOverlap;
		}

		// if all three corners of the triangle are included in the ConvexRegion, then ConvexRegionContainsHTMId
		Vector3d[] triangleCorners = HTMfunc.idToTriangle(HTMId);
		int nCornersInConvexRegion = 0;
		for (Vector3d corner : triangleCorners) {
			if (contains(corner)) {
				nCornersInConvexRegion++;
			}
		}
		if (nCornersInConvexRegion == 3) {
			return Overlap.ConvexRegionContainsHTMId;
		}

		// at least one corner of the triangle is not in ConvexRegion
		int depth = HTMfunc.idLevel(HTMId);
		if (constraints_.size() == 1) {
			// There is only one constraint.  At least one corner of the triangle is not in it.
			Constraint constraint = constraints_.get(0);
			double constraintAngle = Math.toDegrees(constraint.getAngle());
			Vector3d direction = constraint.getDirection();

			Vector3d[] corners = HTMfunc.idToTriangle(HTMId);
			for (int i=0; i<3; i++) {
				Vector3d gc0 = new Vector3d(corners[i]);
				Vector3d gc1;
				try {
					gc1 = new Vector3d(corners[i+1]);
				} catch (ArrayIndexOutOfBoundsException e) {
					gc1 = new Vector3d(corners[0]);
				}
				Vector3d point = new Vector3d(direction);
				{
					// put gc0 at ra=0
					double t = -gc0.ra();
					gc0.rotZ(t); gc1.rotZ(t); point.rotZ(t);
				}
				{
					// put gc0 at dec=90
					double t = gc0.dec()-90;
					gc0.rotY(t); gc1.rotY(t); point.rotY(t);
				}
				{
					// put gc1 at ra=0
					double t = -gc1.ra();
					gc0.rotZ(t); gc1.rotZ(t); point.rotZ(t);
				}
				{
					// put gc0 and gc1 on the equator
					double t = -90;
					gc0.rotX(t); gc1.rotX(t); point.rotX(t);
				}
				double pointRa = point.ra();
				double distance;
				if ((pointRa > gc0.ra()) && (pointRa < 225)) {
					// outside gc0,gc1 closer to gc0
					distance = gc0.separation(point);
				} else if ((pointRa > gc1.ra()) && (pointRa < gc0.ra())) {
					// between gc0 and gc1
					distance = Math.abs(point.dec());
				} else {
					// outside of gc0,gc1 closer to gc1
					distance = gc1.separation(point);
				}
				if (distance < constraintAngle) {
					return Overlap.PartialOverlap;
				}				
			}
			// either the constraint is completely inside the triangle or they do not overlap at all.
			// Test the central point of the constraint to see which it is.
			long tempHTMid = HTMfunc.lookupid(direction, depth);
			if (tempHTMid == HTMId) {
				return Overlap.HTMIdContainsConvexRegion;
			}
			return Overlap.NoOverlap;
		}
		// At least one corner of the triangle is not included in the ConvexRegion. If any of the 
		// corners are in the htm triangle we have a partial overlap
		int nCornersInTriangle = 0;
		for (Vector3d corner : corners_) {
			long cornerHTMId = HTMfunc.lookupid(corner, depth);
			if (cornerHTMId == HTMId) {
				// this corner of the convex is in the triangle, so PartialOverlap;
				nCornersInTriangle++;
			}
		}
		// either the constraint is completely inside the triangle or they do not overlap at all.
		// Test the central point of the constraint to see which it is.
		if (nCornersInTriangle == corners_.size()) {
			return Overlap.HTMIdContainsConvexRegion;
		}
		if (nCornersInTriangle == 0) {
			if (nCornersInConvexRegion == 0) {
				// All corners of the ConvexRegion are outside the triangle.  
				// Check all combination of triangle and ConvexRegion edges.  
				// If one of these pairs intersects, then there is a partial overlap.
				// Otherwise, there is no overlap.
				for (int iTriangleCorner=0; iTriangleCorner<3; iTriangleCorner++) {
					Vector3d tca = triangleCorners[iTriangleCorner];
					Vector3d tcb;
					try {
						tcb = triangleCorners[iTriangleCorner+1];
					} catch (IndexOutOfBoundsException e) {
						tcb = triangleCorners[0];
					}
					Vector3d tcaxb = tca.cross(tcb);
					double tcSeparation = tca.separation(tcb);
					for (int iConvexRegionCorner=0; iConvexRegionCorner<corners_.size(); iConvexRegionCorner++) {
						Vector3d cra = corners_.get(iConvexRegionCorner);
						Vector3d crb;
						try {
							crb = corners_.get(iConvexRegionCorner+1);
						} catch (IndexOutOfBoundsException e) {
							crb = corners_.get(0);
						}
						Vector3d craxb = cra.cross(crb);
						double crSeparation = cra.separation(crb);
						Vector3d intersectionPoint = tcaxb.cross(craxb);
						for (int i=0; i<2; i++) {
							double tcaIntersectionSeparation = intersectionPoint.separation(tca);
							if (tcaIntersectionSeparation < tcSeparation) {
								double tcbIntersectionSeparation = intersectionPoint.separation(tcb);
								if (tcbIntersectionSeparation < tcSeparation) {
									double craIntersectionSeparation = intersectionPoint.separation(cra);
									if (craIntersectionSeparation < crSeparation) {
										double crbIntersectionSeparation = intersectionPoint.separation(crb);
										if (crbIntersectionSeparation < crSeparation) {										
											return Overlap.PartialOverlap;
										}
									}
								}
							}
							intersectionPoint.mul(-1);
						}
					}
					
				}
				return Overlap.NoOverlap;
			}
		}
		return Overlap.PartialOverlap;
	}
	
	/**
	 * see if the constraint extends beyond the great circle defined by c0 and c1
	 * @param c0 first corner
	 * @param c1 second corner
	 * @param p the center of the constraint
	 * @param angle opening angle (degrees)
	 * @return true if it intersects
	 */
	public static boolean edgeConstraintIntersect(Vector3d c0, Vector3d c1, Vector3d p, double angle) {
		Vector3d corner0 = new Vector3d(c0);
		Vector3d corner1 = new Vector3d(c1);
		Vector3d point = new Vector3d(p);
		double degrees;
		
		// get corner 0 to ra=0;
		degrees = -corner0.ra();
		corner0.rotZ(degrees); corner1.rotZ(degrees); point.rotZ(degrees);
		
		// now put it at the north pole
		degrees = -90;
		corner0.rotY(degrees); corner1.rotY(degrees); point.rotY(degrees);

		// if the point is below the declination of corner 1, return false
		if (point.dec() < corner1.dec()) {
			return false;
		}
		
		// now rotate corner 1 to ra=0
		degrees = -corner1.ra();
		corner0.rotZ(degrees); corner1.rotZ(degrees); point.rotZ(degrees);
		double pointRa = point.ra();
		if (pointRa > 180) {
			pointRa -= 360;
		}
		double cosDec = Math.cos(Math.toRadians(point.dec()));
		double distance = Math.abs(pointRa*cosDec);
		boolean retval = distance < angle;
		return retval;
	}
	
	/**
	 * add HTMIds to the HTMrange that intersect this ConvexRegion
	 * @param htmRange
	 * @param maxLevel maximum level to consider
	 * @param includeIntersecting true to include triangles that intersect an edge of the ConvexRegion
	 */
	public void addToRange(HTMrange htmRange, int maxLevel, boolean includeIntersecting) {
		Stack<Long> todo = new Stack<Long>();
		// Put the eight highest-level htmIds on the stack
		for (int quad=0; quad<4; quad++) {
			for (int hemi=0; hemi<2; hemi++) {
				double ra = 45+quad*90;
				double dec=45-hemi*90;
				long htmid = HTMfunc.lookupId(ra, dec, 0);
				todo.push(htmid);
			}
		}
		while (todo.size() > 0) {
			long htmid = todo.pop();
			//System.out.println("htmid="+HTMfunc.idToName(htmid)+" begin with todo.size()="+todo.size());
			//if (HTMfunc.idToName(htmid).equals("N313312")) {
				//System.out.println(" here we are");
			//}
			Overlap overlap = getOverlap(htmid);
			//System.out.println("htmid="+HTMfunc.idToName(htmid)+" overlap="+overlap);
			switch(overlap) {
			case ConvexRegionContainsHTMId: {
				htmRange.add(htmid);		
				break;
			}
			case HTMIdContainsConvexRegion:
			case PartialOverlap: {
				int level = HTMfunc.idLevel(htmid);
				if (level == maxLevel) {
					if (includeIntersecting) {
						htmRange.add(htmid);
					}
				} else {
					long[] children = HTMfunc.getChildren(htmid);
					for (long child : children) {
						todo.push(child);
					}
				}
				break;
			}
			case NoOverlap: {
				break;
			}
			}
		}
		htmRange.merge();
	}









	/**
	 * @param i
	 * @return the ith constraint
	 */
	public Constraint getConstraint (int i) {
		return constraints_.get(i);
	}

	/**
	 * @return the number of constraints
	 */
	public int getNumberOfConstraints () {
		return constraints_.size();
	}

	/**
	 * @return the number of corners
	 */
	public int getNumberOfCorners() {
		return corners_.size();
	}
	/**
	 * @param i
	 * @return the ith corner
	 */
	public Vector3d getCorner (int i) {
		return corners_.get(i);
	}


	/** (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		String out = "#CONVEX\n"+constraints_.size()+"\n";
		int i;
		for( i = 0; i < constraints_.size(); i++)
			out += getConstraint(i).toString()+'\n';
		return out;
	}
	/**
	 * @param p
	 * @return true if p is contained in this
	 */
	public boolean contains(Vector3d p) {
		for (int c=0; c < constraints_.size(); c++) {
			if ( !getConstraint(c).contains(p) ) {
				return false;
			}
		}
		return true;
	}
	
	/**
	 * @param sc
	 * @return true if sc is contained in this
	 */
	public boolean contains(SphericalCoordinate sc) {
		return contains(new Vector3d(sc));
	}
	
	/**
	 * @param that
	 * @return the convex region that is the intersection of these two
	 */
	public ConvexRegion intersect(ConvexRegion that) {
		ArrayList<Vector3d> newCorners = new ArrayList<Vector3d>();
		for (Vector3d thisCorner : corners_) {
			boolean contains = that.contains(thisCorner); 
			//System.out.println("ConvexRegion.intersect:  contains="+contains+" for thisCorner="+thisCorner);
			if (contains) {
				newCorners.add(thisCorner);
			}
		}
		for (Vector3d thatCorner : that.corners_) {
			boolean contains = this.contains(thatCorner); 
			//System.out.println("ConvexRegion.intersect:  contains="+contains+" for thatCorner="+thatCorner);
			if (contains) {
				newCorners.add(thatCorner);
			}
		}

		for (int iThis=0; iThis<corners_.size(); iThis++) {
			Vector3d this0 = corners_.get(iThis);
			Vector3d this1 = corners_.get((iThis+1)%corners_.size());			
			boolean thisIncludeEnds = constraints_.get(iThis).isIncludeEdge();
			//System.out.println("ConvexRegion.intersect:  this0="+this0+" this1="+this1+" ============================");
			for (int iThat=0; iThat<that.corners_.size(); iThat++) {
				Vector3d that0 = that.corners_.get(iThat);
				Vector3d that1 = that.corners_.get((iThat+1) % that.corners_.size());
				boolean thatIncludeEnds = constraints_.get(iThat).isIncludeEdge();
				ArrayList<Vector3d> intersections = Vector3d.getIntersection(this0, this1, thisIncludeEnds, that0, that1, thatIncludeEnds);
				//System.out.println("ConvexRegion.intersect:  that0="+that0+" that1="+that1);
				//System.out.println("ConvexRegion.intersect:  intersections.size="+intersections.size());
				//for (Vector3d intersection : intersections) {
					//System.out.println("ConvexRegion.intersect:  intersection="+intersection);
				//}
				newCorners.addAll(intersections);
			}
			
		}
		ConvexRegion retval = new ConvexRegion(newCorners);
		return retval;
	}
	
	/**
	 * @return the area of the ConvexRegion, in degrees squared
	 */
	public double area() {
		Vector3d[] temp = new Vector3d[corners_.size()];
		for (int i=0; i<temp.length; i++) {
			temp[i] = corners_.get(i);
		}
		double retval = Vector3d.getArea(temp);
		return retval;
	}
}

