package gov.fnal.eag.util.htm;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * keep track of long integers.  This implementation is useful for HTM id numbers, since interesting lists of them are clumped together
 * @author stoughto
 *
 */
public class  HTMrange implements Iterable<Long> {
	ArrayList<long[]> ranges = new ArrayList<long[]>();

	//public static final int BELOW = 0;
	//public static final int EXTEND_LOW = 1;
	//public static final int EXTEND_HIGH = 2;
	//public static final int INCLUDED = 4;
	//public static final int ABOVE = 8;
	
	/**
	 * default constructor makes an empty list
	 */
	public HTMrange() {
	}
	/**
	 * @param convexRegion
	 * @param maxLevel
	 * @param includeIntersecting
	 */
	public HTMrange(ConvexRegion convexRegion, int maxLevel, boolean includeIntersecting) {
		convexRegion.addToRange(this, maxLevel, includeIntersecting);
	}

	/**
	 * add one entry as its own range
	 * @param a
	 */
	public void add(long a) {
		ranges.add(new long[] {a,a});
	}
	
	/**
	 * add a range of values
	 * @param a min value of the range, inclusive
	 * @param b max value of the range, inclusive
	 */
	public void addRange(long a, long b) {
		long min = Math.min(a, b);
		long max = Math.max(a, b);
		ranges.add(new long[] {min,max});
	}
	
	
	/**
	 * clean up the list by collecting adjacent entries together
	 */
	public void merge() {
		if (ranges.size() > 1) {
		boolean merged;
		do {
			merged = ranges.size() > 1;
			for (int i=0; i<ranges.size()-1; i++) {
				long[] one = ranges.get(i);
				for (int j=i+1; j<ranges.size(); j++) {
					long[] two = ranges.get(j);
					merged = attemptMerge(one, two);
					if (merged ){
						ranges.remove(i);
						break;
					}
				}
				if (merged) {
					break;
				}
			}
		} while (merged);
	}
	}
	
	boolean attemptMerge(long[] a, long[] b) {
		boolean retval;
		//System.out.println(" begin attemptMerge:  a="+a[0]+" "+a[1]+" b="+b[0]+" "+b[1]);
		if (a[0] > b[1]) {
			retval = false;
		} else if (b[0] > a[1]) {
			retval = false;
		} else {
			b[0] = Math.min(a[0], b[0]);
			b[1] = Math.max(a[1], b[1]);
			retval = true;
		}
		return retval;
	}
	
	/** (non-Javadoc)
	 * @see java.lang.Iterable#iterator()
	 */
	public Iterator<Long> iterator() {
		return new HTMrangeIterator(ranges);
	}

	/**
	 * @param i
	 * @return true if i is in this set of HTMRanges
	 */
	public boolean isIn(long i) {
		for (long[] range :ranges) {
			if (i > range[1]) return false;
			if (i >= range[0]) return true;
		}
		return false;
	}

	/**
	 * @return the number of ranges in the list
	 */
	public int nRanges() {
		return ranges.size();
	}

	/**
	 * @return the number of elements on the list
	 */
	public long size() {
		long retval = 0;
		for (long[] range : ranges) {
			retval += range[1]-range[0]+1;
		}
		return retval;
	}

	/**
	 * write a brief report
	 * @param out destination
	 */
	public void report(PrintStream out) {
		out.println("nRanges="+nRanges()+" size="+size());
		long i=0;
		for (long[] range : ranges) {
			out.println("i="+i+++" "+range[0]+" "+range[1]);
		}
	}
	
	/**
	 * get the minimum and maximum depths in this set of htm id's
	 * @return minimum depth in the 0 element and maximum depth in the 1 element
	 */
	public int[] getDepth() {
		int minDepth = Integer.MAX_VALUE;
		int maxDepth = Integer.MIN_VALUE;
		for (long[] range : ranges) {
			minDepth = Math.min(minDepth, HTMfunc.idLevel(range[0]));
			minDepth = Math.min(minDepth, HTMfunc.idLevel(range[1]));
			maxDepth = Math.max(maxDepth, HTMfunc.idLevel(range[0]));
			maxDepth = Math.max(maxDepth, HTMfunc.idLevel(range[1]));
		}
		int[] retval = {minDepth, maxDepth};
		return retval;
	}
	
	/**
	 * set the depth of the triangle to newDepth
	 * @param newDepth
	 * @return this HTMrange, with all depths set to newDepth
	 */
	public HTMrange setDepth(int newDepth) {
		if (newDepth >= HTMfunc.HTMNAMEMAX-2 ) {
			throw new IllegalArgumentException("new depth can not be >= "+(HTMfunc.HTMNAMEMAX-2)+" but newDepth="+newDepth);
		}
		for (long[] range : ranges) {
			int depth0 = HTMfunc.idLevel(range[0]);
			int depth1 = HTMfunc.idLevel(range[1]);
			if (depth0 != depth1) {
				throw new IllegalArgumentException("depth0="+depth0+" depth1="+depth1);
			}
			if (depth0 > newDepth) {
				throw new IllegalArgumentException("current depth can not be > new depth but depth0="+depth0+" newDepth="+newDepth);
			}
			if (newDepth > depth0) {			
				int bitsToShift=2*(newDepth-depth0);
				long min = range[0] << bitsToShift;
				long max = range[1] << bitsToShift;
				for (int i=0; i<bitsToShift; i++) {
					max |= 1l << i;
				}
				range[0] = Math.min(min, max);
				range[1] = Math.max(min, max);
			}
		}		
		return this;
	}
	/**
	 * @param name name of the htmid in the SQL table to be queried
	 * @return a where clause to select all htmids in this range
	 */
	public String getWhereClause(String name) {
		String retval = "";
		String firstWord = "where ";
		for (long[] range : ranges) {
			retval += firstWord+name+" between "+range[0]+" and "+range[1];
			firstWord = " or ";
		}
		return retval;
	}
}