package gov.fnal.eag.util.htm;

import java.util.ArrayList;

import gov.fnal.eag.wcs.SphericalCoordinate;


/**********************************************************************

	 3-d vector class with conversion to ra/dec
<p>
The Vector3d is a standard 3D vector class with the addition
that each coordinate (x,y,z) is also kept in ra,dec since we expect
the vector to live on the surface of the unit sphere, i.e.
<pre>
  x<sup>2</sup> + y<sup>2</sup> + z<sup>2</sup>  = 1
</pre>

This is not enforced, so you can specify a vector that does not have unit
length.  If you request the ra/dec of such a vector, it will be
automatically normalized to length 1 and you get the ra/dec of that
vector (the intersection of the vector's direction with the unit
sphere.

Vector multiplication addition etc are inplemented on this class
  */

public class  Vector3d {

	double x_;
	double y_;
	double z_;
	double ra_;
	double dec_;
	boolean okRaDec_;


	/** Default constructor constructs (1,0,0), ra=0, dec=0.*/
	public Vector3d() {
		ra_=0; dec_=0; 
		okRaDec_=true;
		updateXYZ();
	}


	/**
	 * construct a vector from the three coordinates
	 * @param x
	 * @param y
	 * @param z
	 */
	public Vector3d(double x, double y, double z ) {
		x_=x; y_=y; z_=z;
	}


	/**
	 * construct a vector from ra,dec 
	 * @param ra in degrees
	 * @param dec in degrees
	 */
	public Vector3d(double ra , double dec ) {
		ra_=ra; dec_=dec;
		okRaDec_=true;
		updateXYZ();
	}



	/**
	 * copy constructor
	 * @param copy
	 */
	public Vector3d( Vector3d  copy) {
		x_=copy.x_;
		y_=copy.y_;
		z_=copy.z_;
		ra_=copy.ra_;
		dec_=copy.dec_;
		okRaDec_=copy.okRaDec_;
	}



	/**
	 * construct from the x,y,z values in the array
	 * @param d input array
	 */
	public Vector3d(double[] d) {
		this(d[0], d[1], d[2]);
	}


	/**
	 * converts a SphericalCoordinate
	 * @param sc location to copy
	 */
	public Vector3d(SphericalCoordinate sc) {
		this(sc.getX(), sc.getY(), sc.getZ());
	}


	/**
	 * set the values from xyz
	 * @param x
	 * @param y
	 * @param z
	 */
	public void set( double x, double y, double z) {
		x_=x; y_=y; z_=z;
		normalize();
		updateRaDec();
	}


	/**
	 * set the values from ra,dec
	 * @param ra in degrees
	 * @param dec in degrees
	 */
	public void set( double ra, double dec) {
		ra_=ra; dec_=dec;
		okRaDec_=true;
		updateXYZ();
	}


	/**
	 * @return x,y,z values
	  */
	public double[] get() {
		double ret[] = new double[3];
		ret[0]=x_; ret[1]=y_; ret[2]=z_;
		return ret;
	}

	/**
	 * @return length of the vector
	 */
	public double length() {
		double sum = x_*x_ + y_*y_ + z_*z_;
		return Math.sqrt(sum);
	}


	/**
	 * @return x value
	 */
	public double x() {
		return x_;
	}


	/**
	 * @return y value
	 */
	public double y() {
		return y_;
	}


	/**
	 * @return z value
	 */
	public double z() {
		return z_;
	}


	/**
	 * normalize the vector
	 */
	public void normalize() {
		double sum = norm();
		x_ /= sum;
		y_ /= sum;
		z_ /= sum;
	}
	/**
	 * rotate around the X axis
	 * @param theta degrees
	 */
	public void rotX(double theta) {
		double ct = Math.cos(Math.toRadians(theta));
		double st = Math.sin(Math.toRadians(theta));
		double temp = y_*ct - z_*st;
		z_ = y_*st + z_*ct;
		y_ = temp;	
		okRaDec_ = false;
	}
	/**
	 * rotate around the Y axis
	 * @param theta degrees
	 */
	public void rotY(double theta) {
		double ct = Math.cos(Math.toRadians(theta));
		double st = Math.sin(Math.toRadians(theta));
		double temp = z_*ct - x_*st;
		x_ = z_*st + x_*ct;
		z_ = temp;	
		okRaDec_ = false;
	}
	/**
	 * rotate around the Z axis
	 * @param theta degrees
	 */
	public void rotZ(double theta) {
		double ct = Math.cos(Math.toRadians(theta));
		double st = Math.sin(Math.toRadians(theta));
		double temp = x_*ct - y_*st;
		y_ = x_*st + y_*ct;
		x_ = temp;	
		okRaDec_ = false;
	}
	/**
	 * convert to string as "x y z"
	 */
	@Override
	public String toString() {
		//return "" +x_ + " " + y_ + " " + z_;
		String retval ="ra="+ra()+" dec="+dec();
		return retval;
	}
	/**
	 * calculate the cross product of this vector with
	 * @param v the other vector
	 * @return a new Vector3d
	 */
	public Vector3d cross(Vector3d v) {
		return  new Vector3d(y_ * v.z_ - v.y_ * z_,
				z_ * v.x_ - v.z_ * x_,
				x_ * v.y_ - v.x_ * y_);
	}



	/**
	 * @param n
	 * @return a new Vector3d scaled by n
	 */
	public Vector3d mul(double n) {
		return  new Vector3d(x_*n, y_*n, z_*n);
	}

	/**
	 * calculates the dot product of this vector with
	 * @param v the other vector
	 * @return the dot product
	 */
	public double mul(Vector3d v) {
		return (x_*v.x_)+(y_*v.y_)+(z_*v.z_);
	}

	/**
	 * add this vector to
	 * @param v the other vector
	 * @return a new Vector3d
	 */
	public Vector3d add(Vector3d v) {
		return  new Vector3d(x_+v.x_, y_+v.y_, z_+v.z_);
	}

	/**
	 * subtract the other vector from this
	 * @param v the other vector
	 * @return a new Vector3d
	 */
	public Vector3d sub(Vector3d v) {
		return  new Vector3d(x_-v.x_, y_-v.y_, z_-v.z_);
	}

	
	/**
	 * @return the dec, in degrees
	 */
	public double dec() {
		if(!okRaDec_) {
			normalize();
			updateRaDec();
		}
		return dec_;
	}

	/**
	 * @return the ra, in degrees
	 */
	public double ra() {
		if(!okRaDec_) {
			normalize();
			updateRaDec();
		}
		return ra_;
	}

	protected void updateXYZ() {
		double cd = Math.cos(Math.toRadians(dec_));
		x_ = Math.cos(Math.toRadians(ra_)) * cd;
		y_ = Math.sin(Math.toRadians(ra_)) * cd;
		z_ = Math.sin(Math.toRadians(dec_));
	}

	protected void updateRaDec() {
		dec_ = Math.toDegrees(Math.asin(z_));
		double temp = Math.atan2(y_, x_);
		if (temp < 0) {
			temp += 2*Math.PI;
		}
		ra_=Math.toDegrees(temp);
		okRaDec_ = true;
	}

	/**
	 * @param that
	 * @return true if the x,y,z values are equal
	 */
	public boolean equals(Vector3d that) {
		double thisRa = ra();
		double thatRa = that.ra();
		if (Math.abs(thisRa-thatRa) > 1e-12) return false;
		double thisDec = dec();
		double thatDec = that.dec();
		if (Math.abs(thisDec-thatDec) > 1e-12) return false;
		return true;
	}
	/**
	 * An ant walks from v1 to v2 to v3.  How many degrees to the right does it need to turn at v2 to get to v3?
	 * @param v1
	 * @param v2
	 * @param v3
	 * @return degrees to turn, from 0 to 360, or -180 if v3 is in between v1 and v2.
	 */
	public static double getDirection(Vector3d v1, Vector3d v2, Vector3d v3) {
		Vector3d a23 = v2.cross(v3); // set directions of half-spheres
		a23.normalize();

		
		Vector3d a12 = v1.cross(v2);
		a12.normalize();
		
		double dot123 = a12.mul(a23);
		if (1-Math.abs(dot123) < HTMfunc.Epsilon) { // special case:  the three points are on the same great circle
			if (dot123 > 0) return 0.0; // v2 is in the middle
			Vector3d a31 = v3.cross(v1);
			a31.normalize();
			double dot231 =  a23.mul(a31);
			if (dot231 > 0) return -180; // v3 is in the middle
			return +180; // v1 is in the middle
		}
		
		Vector3d cross123 = a12.cross(a23);
		//double dot1231 = cross123.mul(v1);
		double cosTheta = a12.mul(a23);
		double sinTheta = cross123.length();
		double theta = Math.toDegrees(Math.atan2(sinTheta, -cosTheta));
		double degrees;
		Constraint c12 = new Constraint(a12, 0.0);
		if (c12.contains(v3)) {
			degrees = theta-180;
		} else {
			 degrees = 180-theta;
		}
		if (degrees < 0) {
			degrees += 360;
		}
		return degrees;
	}


	/**
	 * apply three rotations in order
	 * @param tx degrees to rotate around X
	 * @param ty degrees to rotate around Y
	 * @param tz degrees to rotate around Z
	 */
	public void rotXYZ(double tx, double ty, double tz) {
		rotX(tx);
		rotY(ty);
		rotZ(tz);
	}
	
	/**
	 * @param that
	 * @return the angular separation in radians
	 */
	public double separationRad(Vector3d that) {
		double dot = mul(that)/length()/that.length();
		dot = Math.max(dot, -1.0);
		dot = Math.min(dot,  1.0);
		double retval = Math.acos(dot);
		return retval;
	}

	/**
	 * @param that
	 * @return the angular separation in degrees
	 */
	public double separation(Vector3d that) {
		return Math.toDegrees(separationRad(that));
	}

	/**
	 * @param corners of a convex polygon
	 * @return area in steradians
	 */
	public static double getAreaSr(Vector3d[] corners) {
		double retval = 0.0;
		for (int iTriangle=0; iTriangle<corners.length-2; iTriangle++) {
			double a = corners[0].separationRad(corners[1+iTriangle]);
			double b = corners[1+iTriangle].separationRad(corners[2+iTriangle]);
			double c = corners[2+iTriangle].separationRad(corners[0]);
			double s = (a + b + c)/2.0;
			double A = 2.0*Math.acos(Math.sqrt(Math.sin(s)*Math.sin(s-c)/(Math.sin(b)*Math.sin(a))));
			double B = 2.0*Math.acos(Math.sqrt(Math.sin(s)*Math.sin(s-a)/(Math.sin(c)*Math.sin(b))));
			double C = 2.0*Math.acos(Math.sqrt(Math.sin(s)*Math.sin(s-b)/(Math.sin(a)*Math.sin(c))));
			double E = A + B + C - Math.PI;
			retval += E;
		}		
		return retval;
	}
	
	/**
	 * @param corners
	 * @return area of the polygon, in degrees squared
	 */
	public static double getArea(Vector3d[] corners) {
		double retval = getAreaSr(corners);
		retval *= Math.pow(Math.toDegrees(1),2);
		return retval;
	}


	/** 
	 * get the intersection of two great circle segments
	 * @param a0 start of first segment
	 * @param a1 end of first segment
	 * @param b0 start of second segment
	 * @param b1 end of second segment
	 * @return the point where they intersect, or null if they do not intersect
	 */
	public static ArrayList<Vector3d> getIntersection(Vector3d a0, Vector3d a1, Vector3d b0, Vector3d b1) {
		return getIntersection(a0, a1, true, b0, b1, true);
	}
	
	/** 
	 * get the intersection of two great circle segments
	 * @param a0 start of first segment
	 * @param a1 end of first segment
	 * @param aIncludeEnds whether to include the end points of the first segment
	 * @param b0 start of second segment
	 * @param b1 end of second segment
	 * @param bIncludeEnds  whether to include the end points of the second segment
	 * @return the point where they intersect, or null if they do not intersect
	 */
	public static ArrayList<Vector3d> getIntersection(Vector3d a0, Vector3d a1, boolean aIncludeEnds, Vector3d b0, Vector3d b1, boolean bIncludeEnds) {
		ArrayList<Vector3d> retval = getIntersectionOrdered(a0, a1, aIncludeEnds, b0, b1, bIncludeEnds);
		//System.out.println("Vector3d:  retval 1 = "+retval);
		if (retval.size() > 0) {
			return retval;
		}
		retval = getIntersectionOrdered(a1, a0, aIncludeEnds, b0, b1, bIncludeEnds);
		//System.out.println("Vector3d:  retval 2 = "+retval);
		return retval;
	}
	private static ArrayList<Vector3d> getIntersectionOrdered(Vector3d a0, Vector3d a1, boolean aIncludeEnds, Vector3d b0, Vector3d b1, boolean bIncludeEnds) {
		ArrayList<Vector3d> retval = new ArrayList<Vector3d>();
		Vector3d ac = a0.cross(a1);
		/*
		System.out.println("Vector3d.getIntersection() a0="+a0);
		System.out.println("Vector3d.getIntersection() a1="+a1);
		System.out.println("Vector3d.getIntersection() ac="+ac);
	*/
		Vector3d bc = b0.cross(b1);
		/*
		System.out.println("Vector3d.getIntersection() b0="+b0);
		System.out.println("Vector3d.getIntersection() b1="+b1);
		System.out.println("Vector3d.getIntersection() bc="+bc);
	*/
		Vector3d intersection = bc.cross(ac);
		intersection.normalize();
		
		//System.out.println("Vector3d.getIntersectionOrdered() retval="+retval);
		if (Double.isNaN(intersection.ra()) || Double.isNaN(intersection.dec())) {
			return retval;
		}
		if (notBetween(a0, intersection, a1)) {
			return retval;
		}
		if (notBetween(b0, intersection, b1)) {
			return retval;
		}
		boolean endConstraint = true;
		if (!aIncludeEnds) {
			if (a0.equals(intersection)) {
				endConstraint = false;
			}
			if (a1.equals(intersection)) {
				endConstraint = false;
			}
		}
		if (!bIncludeEnds) {
			if (b0.equals(intersection)) {
				endConstraint = false;
			}
			if (b1.equals(intersection)) {
				endConstraint = false;
			}
		}
		if (endConstraint) {
			retval.add(intersection);
		}
		return retval;
	}


	private static boolean notBetween(Vector3d va, Vector3d vb, Vector3d vc) {
		Vector3d axb = va.cross(vb);
		Vector3d bxc = vb.cross(vc);
		Vector3d axc = va.cross(vc);
		/*
		System.out.println("Vector3d.notBetween: va="+va);
		System.out.println("Vector3d.notBetween: vb="+vb);
		System.out.println("Vector3d.notBetween: vc="+vc);
		System.out.println("Vector3d.notBetween: axb.separation(bxc) "+axb.separation(bxc));
		System.out.println("Vector3d.notBetween: axb.separation(axc) "+axb.separation(axc));
		System.out.println("Vector3d.notBetween: bxc.separation(axc) "+bxc.separation(axc));
		*/
		if (axb.separation(bxc) >  1e-5) {
			return true;
		}
		if (axb.separation(axc) > 1e-5) {
			return true;
		}
		if (bxc.separation(axc) > 1e-5) {
			return true;
		}
		return false;
	}


	/**
	 * @return the length
	 */
	public double norm() {
		double x2 = x_*x_;
		double y2 = y_*y_;
		double z2 = z_*z_;
		double retval = Math.sqrt(x2 + y2 + z2);
		return retval;
	}


	/**
	 * @param that
	 * @return the dot product of this and that
	 */
	public double dot(Vector3d that) {
		double x2 = x_*that.x_;
		double y2 = y_*that.y_;
		double z2 = z_*that.z_;
		double retval = x2 + y2 + z2;
		return retval;
	}


	

	/*
	 * shamelessly stolen from http://en.wikipedia.org/wiki/Great-circle_distance
	 * @param that
	 * @return the separation (in degrees) between this and that
	 */
	/*
	public double separationWikipedia(Vector3d that) {
		double cd1 = Math.cos(Math.toRadians(dec()));
		double cd2 = Math.cos(Math.toRadians(that.dec()));
		double sd1 = Math.sin(Math.toRadians(dec()));
		double sd2 = Math.sin(Math.toRadians(that.dec()));
		double sra = Math.sin(Math.toRadians(-Math.abs(ra()-that.ra())));
		double cra = Math.cos(Math.toRadians(ra()-that.ra()));
		System.out.println("sra="+sra+" cra="+cra+" ra="+ra()+" that.ra="+that.ra());
		double y = Math.sqrt(Math.pow(cd2*sra,2) + Math.pow(cd1*sd1-sd1*cd2*cra, 2));
		double x = sd1*sd2 + cd1*cd2*cra;
		double retval = Math.toDegrees(Math.atan2(y, x));
		System.out.println("y="+y+" x="+x+" retval="+retval);
		return retval;
	}
	*/
}


