package gov.fnal.eag.util.jdbc;

import java.io.IOException;
import java.sql.Connection;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import uk.ac.starlink.table.StarTable;
import uk.ac.starlink.table.jdbc.Connector;
import uk.ac.starlink.table.jdbc.JDBCAuthenticator;
import uk.ac.starlink.table.jdbc.JDBCFormatter;
import uk.ac.starlink.table.jdbc.JDBCStarTable;
import uk.ac.starlink.table.jdbc.TerminalAuthenticator;

/**
 * mimics uk.ac.starlink.table.jdbc.JDBCHandler.java but has the method getUserPasswordFromUrl to aid Oracle connections
 * @author stoughto
 *
 */
public class JDBCHandler {

    private JDBCAuthenticator auth;
    private String user;
    private String passwd;

    /**
     * default constructor
     */
    public JDBCHandler() {
        this( new TerminalAuthenticator() );
    }

    /** 
     * construtors with authenticator specified
     * @param auth
     */
    public JDBCHandler( JDBCAuthenticator auth ) {
        this.auth = auth;
    }

    /**
     * Clone constructor.  This constructor creates a new JDBCHandler
     * with the all the same fields as the given one <tt>jh</tt>.
     *
     * @param   jh  the handler whose fields to copy
     */
    public JDBCHandler( JDBCHandler jh ) {
        this( jh.auth );
    }

    /**
     * @return the authenticator
     */
    public JDBCAuthenticator getAuthenticator() {
        return auth;
    }

    /**
     * set the authenticor
     * @param auth
     */
    public void setAuthenticator( JDBCAuthenticator auth ) {
        auth.getClass();  // check for null
        this.auth = auth;
    }

    /**
     * do the actual work of creating a StarTable from a SQL query
     * @param spec JDBC query string
     * @param wantRandom true if you want to try creating a random access StarTable
     * @return the StarTable
     * @throws IOException if there is trouble
     */
    public StarTable makeStarTable( String spec, boolean wantRandom ) throws IOException {
    	
        /* Reject if it doesn't look like a JDBC URL. */
        if ( ! spec.startsWith( "jdbc:" ) ) {
            throw new IllegalArgumentException( "Not a JDBC-protocol URL" );
        }

        /* Reject if no SQL query is present. */
        spec = unEscape( spec );
        int hashPos = spec.indexOf( "#" );
        if ( hashPos < 0 ) {
            throw new IOException( 
                "Bad JDBC specification, should be jdbc:...#SQL-query" );
        }
        String frag = spec.substring( hashPos + 1 );
        final String url = spec.substring( 0, hashPos );

        /* Try to get a StarTable using the query. */
        try {
            Connector connector = new Connector() {
                public Connection getConnection() throws SQLException {
                    return JDBCHandler.this.getConnection( url );
                }
            };
            try {
                return new JDBCStarTable( connector, frag, wantRandom );
            }

            /* The open may fail if we've asked for a random one due to 
             * server restrictions - in this case try getting a sequential
             * one instead. */
            catch ( SQLException e ) {
                if ( wantRandom ) {
                    return new JDBCStarTable( connector, frag, false );
                }
				throw e;
            }
        }
        catch ( SQLException e ) {
            StringBuffer sbuf = new StringBuffer()
                .append( "Error making connection " )
                .append( url );
            Enumeration<Driver> den = DriverManager.getDrivers();
            if ( den.hasMoreElements() ) {
                sbuf.append( " - known JDBC drivers:\n" );
                while ( den.hasMoreElements() ) {
                    sbuf.append( "   " )
                        .append( den.nextElement().getClass().getName() )
                        .append( '\n' );
                }
            }
            else {
                sbuf.append( " - no known JDBC drivers" );
            }
            throw (IOException) new IOException( sbuf.toString() )
                               .initCause( e );
        }
    }

    
    /**
     * writes a StarTable to the JDBC connection
     * @param startab table to write
     * @param spec SQL specification
     * @throws IOException if spec does not contain a #
     * @throws SQLException if there is SQL trouble
     */
    public void createJDBCTable( StarTable startab, String spec ) 
            throws IOException, SQLException {

        /* Reject if it doesn't look like a JDBC URL. */
        if ( ! spec.startsWith( "jdbc:" ) ) { 
            throw new IllegalArgumentException( "Not a JDBC-protocol URL" );
        }

        /* Reject if no table name is present. */
        int hashPos = spec.indexOf( "#" );
        if ( hashPos < 0 ) {
            throw new IOException(
                "Bad JDBC specification, should be \"jdbc:...#table-name\"" );
        }
        String frag = spec.substring( hashPos + 1 );
        String url = spec.substring( 0, hashPos );

        /* Try to get a connection using the URL. */
        Connection conn = getConnection( url );
 
        /* And write the data to the table. */
        new JDBCFormatter( conn ).createJDBCTable( startab, frag );
    }

    private Connection getConnection( String url )
            throws SQLException {
       /* First attempt a connection without any authentication. */
    	
        try {
        	Connection retval = DriverManager.getConnection( url );
        	return retval;
        }

        /* If that fails for any reason, request authentication details
         * so we can try again. */
        catch ( SQLException e ) {
            url = getUserPasswordFromUrl(url);
       }
        return DriverManager.getConnection( url, user, passwd );
    }

    private String getUserPasswordFromUrl(String url) {
		// This is what the connection string needs to be   jdbc:postgresql://host:port/database?username=name&password=pw
    	StringTokenizer st = new StringTokenizer(url, "?=&");
    	String[] tokens = new String[st.countTokens()];
    	int i=0;
    	while (st.hasMoreTokens()) {
    		tokens[i] = st.nextToken();
     		i++;
    	}
    	user=tokens[2];
    	if (tokens.length == 5) {
    		passwd = tokens[4];
    	} else {
    		passwd = "";
    	}
    	String retval = url;
    	return retval;
	}


	/**
     * Indicates whether any drivers are installed.
     *
     * @return  <tt>true</tt> iff at least one JDBC driver is available
     *          for URL resolution
     */
    public static boolean hasDrivers() {
        return DriverManager.getDrivers().hasMoreElements();
    }

    /**
     * Unescapes URL-type escaped characters (%xx).
     *
     * @param   url   string which may contain escaped characters
     * @return  string with escaped characters replaced by literals
     */
    private static String unEscape( String url ) {
        StringBuffer buf = new StringBuffer();
        Matcher m = Pattern.compile( "%[0-9a-fA-F][0-9a-fA-F]" )
                           .matcher( url );
        while ( m.find() ) {
            char c = (char) Integer.parseInt( m.group().substring( 1, 3 ), 16 );
            m.appendReplacement( buf, new String( new char[] { c } ) );
        }
        m.appendTail( buf );
        return buf.toString();
    }

}
