package gov.fnal.eag.util.jdbc;

import gov.fnal.eag.util.misc.StringUtil;
import gov.fnal.eag.util.yanny.Table;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Properties;

import uk.ac.starlink.table.RowSequence;
import uk.ac.starlink.table.StarTable;
import uk.ac.starlink.table.TableFormatException;

/**
 * manages the creation of a Connection to an Oracle server
 * @author stoughto
 * 
 * properties read from the .postgres file in the user's home area
 * host localhost
 * port 5432
 * database databaseName
 * username userName
 * password passWord
 *
 */
public class PostgresConnection {

	/**
	 * 
	 * @return the connection
	 */
	public static Connection getConnection() {
		String postgresClass = "org.postgresql.Driver";
		try {
			Class.forName(postgresClass);
		} catch (ClassNotFoundException e) {
			throw new Error(e);
		}
		Properties properties = getProperties();
		String connectionString = getConnectionString(properties);
		//System.out.println(" connectionString="+connectionString);
		String username = properties.getProperty("username","");
		String password = properties.getProperty("password","");

		Connection conn = null;
		try {
			conn = DriverManager.getConnection(connectionString, username, password);
		} catch (SQLException e) {
			throw new Error(e);
		}
		return conn;
	}

	private static String getConnectionString(Properties properties) {
		// This is what the connection string needs to be   jdbc:postgresql://host:port/database?username=name&password=pw
		String host = properties.getProperty("host","localhost");
		String port = properties.getProperty("port","5432");
		String database = properties.getProperty("database","");
		String username = properties.getProperty("username","");
		String password = properties.getProperty("password","");
		if (database.equals("")) {
			database = System.getProperty("user.name");
		}
		String connectionString = "jdbc:postgresql://"+host+":"+port+"/"+database+"?username="+username+"&password="+password;
		return connectionString;
	}

	private static Properties getProperties() {
		File propertyFile = new File(System.getProperty("user.home"), ".postgres");
		Properties properties = new Properties();
		try {
			properties.load(new FileInputStream(propertyFile));
		} catch (FileNotFoundException e) {
			throw new Error("Can not find file="+propertyFile.getAbsolutePath());
		} catch (IOException e) {
			throw new Error("IOException while reading file="+propertyFile.getAbsolutePath());
		}
		

		return properties;
	}
	
	/**
	 * run a sample query on the idproc table and print the results
	 * @param args ignored
	 */
	public static void main(String[] args) {
		String query = "select * from idproc where run = 2074";
		StarTable st;
		try {
			st = PostgresConnection.getStarTable(query);
		} catch (Exception e) {
			throw new Error(e);
		}
		int runIndex = Table.getIndexByName(st, "run");
		int mjdIndex = Table.getIndexByName(st, "mjd");
		int rerunIndex = Table.getIndexByName(st, "rerun");
		int columnIndex = Table.getIndexByName(st, "column");
		int pipelineIndex = Table.getIndexByName(st, "pipeline");
		int versionIndex = Table.getIndexByName(st, "version");
		int startMJDIndex = Table.getIndexByName(st, "startMJD");
		int endMJDIndex = Table.getIndexByName(st, "endMJD");
		int pdQualIndex = Table.getIndexByName(st, "pdQual");
		int rdQualIndex = Table.getIndexByName(st, "rdQual");

		RowSequence rs;
		try {
			rs = st.getRowSequence();
		} catch (IOException e) {
			throw new Error(e);
		}
		int i=1;
		try {
			while (rs.next()) {
				Object[] row = rs.getRow();
				if (i <= 5) {
				int run = ((Number)row[runIndex]).intValue();
				int rerun = ((Number)row[rerunIndex]).intValue();
				int mjd = ((Number)row[mjdIndex]).intValue();
				int column = ((Number)row[columnIndex]).intValue();
				String pipeline = (String) StringUtil.removeDuplicateWhitespace((String)row[pipelineIndex]);
				String version=(String) StringUtil.removeDuplicateWhitespace((String)row[versionIndex]);
				int startMJD = ((Number)row[startMJDIndex]).intValue();
				int endMJD = ((Number)row[endMJDIndex]).intValue();
				String pdQual=(String) StringUtil.removeDuplicateWhitespace((String)row[pdQualIndex]);
				String rdQual=(String) StringUtil.removeDuplicateWhitespace((String)row[rdQualIndex]);
				
				System.out.println(String.format("i=%4d  run=%4d rerun=%3d mjd=%6d column=%1d pipeline=%10s version=%10s startMJD=%6d endMJD=%6d pdQual=%8s rdQual=%8s",
						i, run, rerun, mjd, column, pipeline, version, startMJD, endMJD, pdQual, rdQual));
				}
				i++;
			}
		} catch (IOException e) {
			throw new Error(e);
		}
		System.out.println("number of rows is "+i);
	}
	
	/**
	 * @param query
	 * @return the resulting table
	 * @throws TableFormatException
	 * @throws IOException
	 */
	public static StarTable getStarTable(String query) throws TableFormatException, IOException {
		return getStarTable(getProperties(), query);
	}
	/**
	 * @param properties 
	 * @param query
	 * @return the resulting table
	 * @throws TableFormatException
	 * @throws IOException
	 */
	public static StarTable getStarTable(Properties properties, String query) throws TableFormatException, IOException {
		getConnection();
		String connectionString = getConnectionString(properties);
	   	String databaseUrl = connectionString+"#"+query;
		StarTableFactory starTableFactory = new StarTableFactory(false);
		StarTable retval = starTableFactory.makeStarTable(databaseUrl);
		return retval;
	}
}
