package gov.fnal.eag.util.misc;

import gov.fnal.eag.util.stopwatch.StopwatchCollection;
import gov.fnal.eag.wcs.SphericalCoordinate;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Random;

import uk.ac.starlink.table.join.HEALPixMatchEngine;
import uk.ac.starlink.table.join.HTMMatchEngine;
import uk.ac.starlink.table.join.MatchEngine;
/**
 * prepare a catalog for matching individual objects to it.
 * @author stoughto
 *
 * @param <Element> the object that will be returned by the match
 */
public class MatchTable<Element> {
	ArrayList<Element> elements = new ArrayList<Element>();
	ArrayList<double[]> coordinates = new ArrayList<double[]>();
	MatchEngine matchEngine;
	double bestSeparation;
	//List<IndexList> sortedList = new LinkedList<IndexList>();
	Hashtable<Object,ArrayList<Integer>> hashtable = new Hashtable<Object,ArrayList<Integer>>();
	IndexList tempIndexList = new IndexList(0);
	/**
	 * @return the elements from this MatchTable
	 */
	public ArrayList<Element> getElements() {
		return elements;
	}
	/**
	 * @return the coordinates for this MatchTable
	 */
	public ArrayList<double[]> getCoordinates() {
		return coordinates;
	}
	/**
	 * fully-specified constructor.  Note that the MatchEngine has a separation
	 * @param matchEngine tested for HTMMatchEngine and HEALPixMatchEngine
	 */
	public MatchTable(MatchEngine matchEngine) {
		this.matchEngine = matchEngine;
	}

	/**
	 * add a new entry to this table.
	 * @param sc
	 * @param element
	 */
	public void add(SphericalCoordinate sc, Element element) {
		Object[] bins = getBins(sc);		
		for (Object bin : bins) {
			ArrayList<Integer> list = hashtable.get(bin);
			if (list == null) {
				list = new ArrayList<Integer>();
				hashtable.put(bin, list);
			}

			list.add(elements.size());
		}
		elements.add(element);
		coordinates.add(new double[] {sc.getLongitude(), sc.getLatitude()});
	}
	/**
	 * @return the number of element in this MatchTable
	 */
	public int size() {
		return elements.size();
	}
	private class IndexList {
		long key;
		IndexList(long key) {
			this.key = key;
		}		
	}
	private Object[] getBins(SphericalCoordinate sc) {
		Number[] raDecRadians = {Math.toRadians(sc.getLongitude()), Math.toRadians(sc.getLatitude())};
		Object[] bins = matchEngine.getBins(raDecRadians);
		return bins;
	}

	/**
	 * @param sc the coordinate
	 * @return the element closest to the coordinate, or null if nothing is close
	 */
	public Element bestMatch(SphericalCoordinate sc) {
		Object[] bins = getBins(sc);
		Element retval =  null;
		bestSeparation = Double.MAX_VALUE;
		for (Object bin : bins) {
			ArrayList<Integer> list = hashtable.get(bin);
			if (list != null) {
				for (int index : list) {
					double[] coord = coordinates.get(index);
					SphericalCoordinate temp = new SphericalCoordinate(coord[0], coord[1]);
					double separation = sc.separation(temp);
					if (separation < bestSeparation) {
						bestSeparation = separation;
						retval = elements.get(index);
					}
				}
			}
		}
		return retval;
	}

	/**
	 * gets all of the objects that are near the coordinate
	 * @param sc the coordinate
	 * @param separationArcsec maximum separation allowed
	 * @return elements near the coordinate
	 */
	public ArrayList<Element> getMatches(SphericalCoordinate sc, double separationArcsec) {
		double separationDegrees = separationArcsec/3600.0;
		Object[] bins = getBins(sc);
		ArrayList<Element> retval =  new ArrayList<Element>();
		bestSeparation = Double.MAX_VALUE;
		for (Object bin : bins) {
			ArrayList<Integer> list = hashtable.get(bin);
			for (int index : list) {
				Element element = elements.get(index);
				if (!retval.contains(element)) {
					double[] coord = coordinates.get(index);
					SphericalCoordinate temp = new SphericalCoordinate(coord[0], coord[1]);
					double separation = sc.separation(temp);
					if (separation < separationDegrees) {
						retval.add(element);
					}
				}
			}
		}
		return retval;
	}

	/**
	 * find the element of this MatchTable that is closest to the coordinate
	 * @param coordinate location of the object 
	 * @param separationArcsec only count matches that have a separation smaller than this, in arcseconds
	 * @return the Element of the best match, or null if there is no match
	 * 
	 * Warning:  if separationArcsec is greater than the separation you use in the MatchEngine that defines 
	 * this MatchTable, you may miss matches.
	 */
	public Element bestMatch(SphericalCoordinate coordinate, double separationArcsec) {
		Element retval = bestMatch(coordinate);
		if (bestSeparation > separationArcsec/3600.0) {
			retval = null;
		} 
		return retval;
	}

	/**
	 * demonstrates the use of MatchTable.  The HTMMathEngine uses less memory per object (752 instead
	 * of 2020 bytes) but it runs slower.  
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		Random random = new Random();
		random.setSeed(123456789L);
		int nCatalog = 1000;
		double[] ras = new double[nCatalog];
		double[] decs = new double[nCatalog];
		for (int i=0; i<nCatalog; i++) {
			ras[i] = 10 + 10*random.nextDouble();
			decs[i] = 30 + 10*random.nextDouble();			
		}		

		double separationArcsec = 3;
		double separationRadians = Math.toRadians(separationArcsec/3600.0);

		System.out.println("Begin Test with HEALPixMatchEngine ==================================");
		MatchEngine healMatchEngine = new HEALPixMatchEngine(separationRadians, false);
		doTest(healMatchEngine, ras, decs, separationArcsec);

		System.out.println("Begin Test with HTMMatchEngine     ==================================");
		MatchEngine htmMatchEngine = new HTMMatchEngine(separationRadians, false);
		doTest(htmMatchEngine, ras, decs, separationArcsec);
	}

	private static void doTest(MatchEngine matchEngine, double[] ras, double[] decs, double separationArcsec) {
		MatchTable<String> matchTable = new MatchTable<String>(matchEngine);
		System.out.println("before adding "+ras.length+" objects to table: "+Memory.report());
		long before = Memory.availableMemory();
		StopwatchCollection.start("add");
		for (int i=0; i<ras.length; i++) {
			if (i%10000 == 0) {
				System.out.print(".");
			}
			matchTable.add(new SphericalCoordinate(ras[i], decs[i]), Integer.toString(i));
		}
		System.out.println("");
		StopwatchCollection.stop("add");
		System.out.println(" after adding "+ras.length+" objects to table: "+Memory.report());
		long after = Memory.availableMemory();
		double size = (before-after)/ras.length;
		System.out.println("size="+size);
		int nFound = 0;
		StopwatchCollection.start("get");

		for (int i=0; i<ras.length; i++) {
			if (i%10000 == 0) {
				System.out.print(".");
			}
			String found = matchTable.bestMatch(new SphericalCoordinate(ras[i], decs[i]));
			if (found.equals(Integer.toString(i))) {
				nFound++;
			}
		}
		StopwatchCollection.stop("get");
		if (nFound == ras.length) {
			System.out.println("SUCCESS:  nFound="+nFound);
		}
		System.out.println("");
		StopwatchCollection.report(System.out);
	}
	/**
	 * @param fileName where to write a DS9 region file
	 * @param circleRadiusArcsec 
	 */
	public void makeRegionFile(String fileName, double circleRadiusArcsec) {
		double circleRadiusInDeg = circleRadiusArcsec/3600;
		PrintStream radecPs;
		try {
			radecPs = new PrintStream(new FileOutputStream(new File(fileName)));
		} catch (FileNotFoundException e) {
			throw new Error(e);
		}
		radecPs.println("FK5");
		for (double[] coordinate : coordinates) {
			radecPs.println("circle "+coordinate[0]+" "+coordinate[1]+" "+circleRadiusInDeg);
		}
		radecPs.close();
	}
}

