package gov.fnal.eag.util.misc;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.JarInputStream;
import java.util.regex.Pattern;
import java.util.zip.ZipException;



/**
 * list resources available from the classpath
 * @author stoughto
 * 
 * modified  by N. Kuropatkin to permit search for resources in
 * One-Jar jar file. 
 *
 */
public class ResourceList {

	private static String LIB_PREFIX = "lib/";
	private static String MAIN_PREFIX = "main/";
	/**
	 * for all elements of java.class.path get a Collection of resources
	 * Pattern pattern = Pattern.compile(".*"); gets all resources
	 * @param pattern the pattern to match
	 * @return the resources in the order they are found 
	 */
	public static Collection<String> getResources(Pattern pattern) {
		ArrayList<String> retval = new ArrayList<String>();
		String classPath = System.getProperty("java.class.path",".");
		String[] classPathElements = classPath.split(":");
		int i=0;
		for (String element : classPathElements) {			
			retval.addAll(getResources(element, pattern));
			i++;
		}
		return retval;
	}

	private static Collection<String> getResources(String element, Pattern pattern) {
		ArrayList<String> retval = new ArrayList<String>();
		File file = null;
		try {
			file = new File(element);
		} catch (Exception ex) {
			throw new Error("trouble creating a File for element="+element);
		}
		if (file.isDirectory()) {
			retval.addAll(getResourcesFromDirectory(file, pattern));
		} else {
			try {
				retval.addAll(getResourcesFromJarFile(file, pattern));
			} catch (IOException e) {
				System.err.println("Error opening a jar file");
				e.printStackTrace();
			}
		}
		return retval;
	}


	private static Collection<String> getResourcesFromJarFile(File file, Pattern pattern) throws IOException {
		ArrayList<String> retval = new ArrayList<String>();
		JarFile jarFile;
		try {
			jarFile = new JarFile(file);
			Enumeration<JarEntry> elements = jarFile.entries();
			while (elements.hasMoreElements()) {

				JarEntry entry = elements.nextElement();
				if (entry.isDirectory()) {
					continue;
				}
				String jar = entry.getName();
				if (jar.startsWith(LIB_PREFIX) || jar.startsWith(MAIN_PREFIX)) {
					if (jar.endsWith(".jar") && jar.contains("gov.fnal")) {

						JarInputStream is =  new JarInputStream(jarFile.getInputStream(entry));

						JarEntry nje;
						while ((nje = is.getNextJarEntry()) != null) {
							String fName = nje.getName();
							if (! fName.endsWith(".class")) {
								boolean accept = pattern.matcher(fName).matches();
								if (accept) {
									retval.add(fName);
								}
							}
						}
						is.close();
					}

				} else {
					boolean accept = pattern.matcher(jar).matches();
					if (accept) {
						retval.add(jar);
					}
				}
			}
			jarFile.close();
			return retval;
		} catch (ZipException ze) {
			return retval;
		} catch (Error e) {
			throw new Error("file="+file,e);
		}
	}

	private static Collection<String> getResourcesFromDirectory(File directory, Pattern pattern) {
		ArrayList<String> retval = new ArrayList<String>();
		String sep = java.io.File.separator;
		//
		// get current working directory-it will be a base for
		// calculation of the relative path
		//
		Properties pr = System.getProperties();
		String myDir = pr.getProperty("user.dir");
		//
		// Calculate the leading part of the absolute path
		//
		String basePath = myDir.substring(0,myDir.lastIndexOf(sep)+1);

		int cutIndex = basePath.length();
		File[] fileList = directory.listFiles();
		for (File file : fileList) {

			if (file.isDirectory()) {
				retval.addAll(getResourcesFromDirectory(file, pattern));
			} else {
				try {

					String fileName = file.getCanonicalPath();
					fileName = fileName.substring(cutIndex);
					StringTokenizer st = new StringTokenizer(fileName,sep);
					int nt = st.countTokens();
					String[] tokens = new String[nt];
					int startInd = 0;
					if (nt > 1) {
						for (int i=0; i<nt; i++){
							tokens[i] = st.nextToken();
							// is it a package?
							if ((i < nt-1) &&  (tokens[i].indexOf(".") > 0)) {
								startInd = i;
								tokens[i]= tokens[i].replaceAll("\\.", sep);
							}

						}
						fileName = "";
						for (int k=startInd; k<nt; k++) {
							fileName = fileName + tokens[k]+sep;
						}

						fileName = fileName.substring(0, fileName.length()-1);

					}
					boolean accept = pattern.matcher(fileName).matches();
					if (accept) {
						retval.add(fileName);
					}
				} catch (IOException e) {
					throw new Error(e);
				}
			}
		}
		return retval;
	}

	/**
	 * list the resources that match args[0]
	 * @param args args[0] is the pattern to match, or list all resources if there are no args
	 */
	public static void main(String[] args) {
		Pattern pattern;
		if (args.length < 1) {
			pattern = Pattern.compile(".*");
		} else {
			pattern = Pattern.compile(args[0]);	
		}
		Collection<String> list = ResourceList.getResources(pattern);
		System.out.println("ResourceList.main:  resources for pattern="+pattern);
		for (String name : list) {

			System.out.println(name);

		}

	}
}
