package gov.fnal.eag.util.plot;

import gov.fnal.eag.wcs.Coordinate;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Paint;
import java.awt.Shape;
import java.awt.Stroke;
import java.io.File;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;

import org.jfree.chart.ChartColor;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.annotations.XYAnnotation;
import org.jfree.chart.annotations.XYLineAnnotation;
import org.jfree.chart.axis.LogarithmicAxis;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.NumberTickUnit;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYDotRenderer;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

/**
 * a very short collection of helpful JFreechart utilities.
 * @author stoughto
 *
 */
public class JFreeChartUtil {

	/**
	 * the default paints indexed by series number
	 */
	public static Paint[] paints = ChartColor.createDefaultPaintArray();
	
	static final BasicStroke basicStroke = new BasicStroke();

	/**
	 * sets the renderer to put small dots at each point
	 * @param chart
	 */

	public static void setXYDotRenderer(JFreeChart chart) {
		final XYPlot plot = chart.getXYPlot();
		plot.setRenderer(new XYDotRenderer());
	}

	/**
	 * sets the range axis to be logarithmic
	 * @param chart
	 */
	public static void setLogarithmicRangeAxis(JFreeChart chart) {
		final XYPlot plot = chart.getXYPlot();
		final NumberAxis rangeAxis = new LogarithmicAxis(plot.getRangeAxis().getLabel());
		plot.setRangeAxis(rangeAxis);
	}
	/**
	 * sets the domain axis to be logarithmic
	 * @param chart
	 */
	public static void setLogarithmicDomainAxis(JFreeChart chart) {
		final XYPlot plot = chart.getXYPlot();
		final NumberAxis rangeAxis = new LogarithmicAxis(plot.getDomainAxis().getLabel());
		plot.setDomainAxis(rangeAxis);
	}

	/**
	 * sets the renderer to put dots at each point
	 * @param chart
	 * @param dotWidth 
	 * @param dotHeight 
	 */
	public static void setXYDotRenderer(JFreeChart chart, int dotWidth, int dotHeight) {
		final XYPlot plot = chart.getXYPlot();
		XYDotRenderer xydr = new XYDotRenderer();
		xydr.setDotWidth(dotWidth);
		xydr.setDotHeight(dotHeight);
		plot.setRenderer(xydr);
	}
	static final int defaultXSize = 800;
	static final int defaultYSize = 600;
	/**
	 * plot thin black error bars on the points
	 * @param chart where to put them
	 * @param values locations of the points
	 * @param errors radius of the error bar
	 */
	public static void plotErrorBars(JFreeChart chart, XYSeries values, XYSeries errors) {
		Stroke stroke = new BasicStroke();
		Paint paint = Color.black;
		XYPlot plot = chart.getXYPlot();
		if (values.getItemCount() != errors.getItemCount()) return;
		for (int i=0; i<values.getItemCount(); i++) {
			double x = values.getX(i).doubleValue();
			double y = values.getY(i).doubleValue();
			double dx = errors.getX(i).doubleValue();
			double dy = errors.getY(i).doubleValue();			
			XYLineAnnotation vertical = new XYLineAnnotation(x, y-dy, x, y+dy, stroke, paint);
			plot.addAnnotation(vertical);
			XYLineAnnotation horizontal = new XYLineAnnotation(x-dx, y, x+dx, y, stroke, paint);
			plot.addAnnotation(horizontal);
		}
	}

	/**
	 * plots the chart to the fileName using the default size of 800x600
	 * @param chart
	 * @param fileName
	 */
	public static void toJpeg(JFreeChart chart, String fileName) {
		toJpeg(chart, fileName, defaultXSize, defaultYSize);
	}
	/**
	 * plots the chart to the fileName using the default size of 800x600
	 * @param chart
	 * @param fileName
	 */
	public static void toPng(JFreeChart chart, String fileName) {
		toPng(chart, fileName, defaultXSize, defaultYSize);
	}
	/**
	 * plots the chart to the fileName using the specified sizes
	 * @param chart
	 * @param fileName
	 * @param xSize
	 * @param ySize
	 */
	public static void toPng(JFreeChart chart, String fileName, int xSize, int ySize) {
		try {
			ChartUtilities.saveChartAsPNG(new File(fileName), chart, xSize, ySize);
		} catch (IOException e) {
			throw new Error(e);
		}

	}
	/**
	 * plots the chart to the fileName using the specified sizes
	 * @param chart
	 * @param fileName
	 * @param xSize
	 * @param ySize
	 */
	public static void toJpeg(JFreeChart chart, String fileName, int xSize, int ySize) {
		try {
			ChartUtilities.saveChartAsJPEG(new File(fileName), chart, xSize, ySize);
		} catch (IOException e) {
			throw new Error(e);
		}

	}

	/**
	 * call NumberAxis.setTickUnit using E notation
	 * @param axis
	 */
	public static void setTickUnit(NumberAxis axis) {
		DecimalFormat decimalFormat = new DecimalFormat("0.###E0");
		axis.setNumberFormatOverride(decimalFormat);
		double delta = (axis.getUpperBound()-axis.getLowerBound())/10.0;
		double temp = Math.pow(10, Math.floor(Math.log10(delta)));
		//System.out.println("JFreeChartUtil.setTickUnit:  delta="+delta+" temp="+temp);

		NumberTickUnit numberTickUnit = new NumberTickUnit(temp);
		axis.setTickUnit(numberTickUnit);

	}

	/**
	 * defines ticks based on values in the domain range that will be displayed
	 * @param plot the plot to modify
	 * @param domainMinIn minimum on the x axis
	 * @param domainMaxIn maximum on the y axis
	 */
	public static void setRangeTicks(XYPlot plot, Double domainMinIn, Double domainMaxIn) {
		DecimalFormat decimalFormat = new DecimalFormat("0.###E0");
		Double domainMin, domainMax;
		if (domainMinIn == null) {
			domainMin = -Double.MAX_VALUE;
		} else {
			domainMin = domainMinIn;
		}
		if (domainMaxIn == null) {
			domainMax = Double.MAX_VALUE;
		} else {
			domainMax = domainMaxIn;
		}
		for (int iDataset=0; iDataset < plot.getDatasetCount(); iDataset++) {
			XYDataset dataset = plot.getDataset(iDataset);
			double rangeMin = Double.MAX_VALUE;
			double rangeMax = -Double.MAX_VALUE;
			for (int iSeries = 0; iSeries < dataset.getSeriesCount(); iSeries++) {
				NumberAxis rangeAxis = (NumberAxis) plot.getRangeAxis();
				for (int i=0; i<dataset.getItemCount(iSeries); i++) {
					Double x = (Double) dataset.getX(iSeries,i);
					if (x > domainMin && x < domainMax) {
						Double y = (Double) dataset.getY(iSeries,i);						
						rangeMin = Math.min(rangeMin, y);
						rangeMax = Math.max(rangeMax, y);
					}
				}				
				double delta = rangeMax-rangeMin;
				double temp = Math.pow(10, Math.floor(Math.log10(delta)));
				//System.out.println("JFreeChartUtil.setTickUnit:  delta="+delta+" temp="+temp);				
				NumberTickUnit numberTickUnit = new NumberTickUnit(temp);
				//System.out.println(" iSeries="+iSeries+" rangeAxis="+rangeAxis+" rangeMin="+rangeMin+" rangeMax="+rangeMax);
				rangeAxis.setTickUnit(numberTickUnit);
				rangeAxis.setLowerBound(rangeMin);
				rangeAxis.setUpperBound(rangeMax);
				rangeAxis.setNumberFormatOverride(decimalFormat);
			}
		}
	}

	/**
	 * convenience method to set the visibility of lines and shapes for everything in the chart
	 * @param chart
	 * @param linesVisible
	 * @param shapesVisible
	 */
	public static void linesShapesVisible(JFreeChart chart, boolean linesVisible, boolean shapesVisible) {
		{
			XYLineAndShapeRenderer xylsr = (XYLineAndShapeRenderer) chart.getXYPlot().getRenderer();
			xylsr.setBaseLinesVisible(linesVisible);
			xylsr.setBaseShapesVisible(shapesVisible);		
		}

	}
	/**
	 * convenience method to set the visibility of lines and shapes for the specified series in the chart
	 * @param chart
	 * @param iSeries 
	 * @param linesVisible
	 * @param shapesVisible
	 */
	public static void linesShapesVisible(JFreeChart chart, int iSeries, boolean linesVisible, boolean shapesVisible) {
		XYLineAndShapeRenderer xylsr = (XYLineAndShapeRenderer) chart.getXYPlot().getRenderer();
		xylsr.setSeriesLinesVisible(iSeries,linesVisible);
		xylsr.setSeriesShapesVisible(iSeries, shapesVisible);		
	}

	/**
	 * makes a chart out of the series
	 * @param xys
	 * @param title
	 * @param xLabel
	 * @param yLabel
	 * @param shape use this as the default shape if it is not null
	 * @return the chart
	 */
	public static JFreeChart makeChart(XYSeries xys, String title, String xLabel, String yLabel, Shape shape) {
		XYSeriesCollection xysc = new XYSeriesCollection(xys);
		return makeChart(xysc, title, xLabel, yLabel, shape);
	}


	/**
	 * makes a chart out of the series collection
	 * @param xysc
	 * @param title
	 * @param xLabel
	 * @param yLabel
	 * @param shape use this as the default shape if it is not null
	 * @return the chart
	 */
	public static JFreeChart makeChart(XYSeriesCollection xysc, String title, String xLabel, String yLabel, Shape shape) {
		return makeChart(XYPM.makeXYPMs(xysc), title, xLabel, yLabel, shape);
	}

	/**
	 * @param xypms
	 * @param title
	 * @param xLabel
	 * @param yLabel
	 * @param shape
	 * @return the chart
	 */
	public static JFreeChart makeChart(ArrayList<XYPM> xypms, String title, String xLabel, String yLabel, Shape shape) {
		return makeChart(xypms, title, xLabel, yLabel, shape, 0.01, 0.0, 0.0, new BasicStroke());
	}

	/**
	 * @param xypms
	 * @param title
	 * @param xLabel
	 * @param yLabel
	 * @param shape
	 * @param errorBarWidth size of horizontal ticks in fraction of domain axis size
	 * @param errorBarOffset horizontal offset of top (to the right) and bottom (to the left) of error bar
	 * @param dErrorBarOffsetDSeries change in the offset per each series with error bars
	 * @param stroke to draw the error bar
	 * @return the chart
	 */
	public static JFreeChart makeChart(ArrayList<XYPM> xypms, String title, String xLabel, String yLabel, Shape shape, double errorBarWidth, double errorBarOffset, double dErrorBarOffsetDSeries, Stroke stroke) {
		XYSeriesCollection xyscInvisible = new XYSeriesCollection();
		XYSeries xysInvisible = new XYSeries("invisible");
		xyscInvisible.addSeries(xysInvisible);
		XYSeriesCollection xysc = new XYSeriesCollection();
		for (XYPM xypm : xypms) {
			xysc.addSeries(xypm.data);
			for (int i=0; i<xypm.data.getItemCount(); i++) {
				double x = (Double) xypm.data.getX(i);
				double y = (Double) xypm.data.getY(i);
				xysInvisible.add(x,y);
				if (xypm.errors != null) {
					double p = y + (Double) xypm.errors.getX(i);
					double m = y + (Double) xypm.errors.getY(i);
					xysInvisible.add(x,p);
					xysInvisible.add(x,m);
				}
			}
		}
		JFreeChart tempChart = ChartFactory.createScatterPlot(title, xLabel, yLabel, xyscInvisible, PlotOrientation.VERTICAL, true, true, true);

		ValueAxis domainAxis = tempChart.getXYPlot().getDomainAxis();
		double domainSize = domainAxis.getUpperBound()-domainAxis.getLowerBound();
		ValueAxis rangeAxis  = tempChart.getXYPlot().getRangeAxis();
		//Range domainAxisRange = domainAxis.getRange();
		JFreeChart chart = ChartFactory.createScatterPlot(title, xLabel, yLabel, xysc, PlotOrientation.VERTICAL, true, true, true);
		if (shape != null) {
			XYPlot plot = chart.getXYPlot();
			XYLineAndShapeRenderer xyItemRenderer = (XYLineAndShapeRenderer) plot.getRendererForDataset(xysc);
			xyItemRenderer.setBaseShape(shape);
		}

		chart.getXYPlot().getDomainAxis().setRange(domainAxis.getRange());
		chart.getXYPlot().getRangeAxis().setRange(rangeAxis.getRange());
		int nErrorBars = 0;
		for (XYPM xypm : xypms) {
			if (xypm.errors != null) {
				nErrorBars++;
			}
		}
		double iErrorBar0 = (nErrorBars-1)/2.0;
		int iSeries=0;
		int iErrorBar = 0;
		drawChart(chart);
		XYPlot plot = chart.getXYPlot();
		//System.out.println("JFreeChartUtil:  plot="+plot);
		XYLineAndShapeRenderer xyItemRenderer = (XYLineAndShapeRenderer) plot.getRenderer(0);
		//System.out.println("JFreeChartUtil:  xyItemRenderer="+xyItemRenderer);

		for (XYPM xypm : xypms){
			if (xypm.errors != null) {
				double offset = errorBarOffset + dErrorBarOffsetDSeries*(iErrorBar-iErrorBar0);
				offset *= domainSize;
				//Paint paint = chart.getXYPlot().getRenderer(0).getSeriesPaint(iSeries);
				Paint paint = xyItemRenderer.getSeriesPaint(iSeries);
				//System.out.println("JFreeChartUtil.makeChart xypm="+xypm+" paint="+paint+" iSeries="+iSeries);
				if (paint == null) {
					paint = paints[iSeries];
					xyItemRenderer.setSeriesPaint(iSeries, paint);
				}
				iErrorBar++;
				for (int i=0; i < xypm.data.getItemCount(); i++) {
					double x = (Double) xypm.data.getX(i);
					double y = (Double) xypm.data.getY(i);
					double deltaP = (Double) xypm.errors.getX(i);
					double deltaM = (Double) xypm.errors.getY(i);
					double p = y + deltaP;
					double m = y + deltaM;
					//deltaP is positive and deltaM is negative
					{
						XYLineAnnotation line = new XYLineAnnotation(x-offset, m, x+offset, p, stroke, paint);
						chart.getXYPlot().addAnnotation(line);
					}
					double dx = domainSize * errorBarWidth;
					{
						XYLineAnnotation line = new XYLineAnnotation(x-offset-dx/2, m, x-offset+dx/2, m, basicStroke, paint);
						chart.getXYPlot().addAnnotation(line);
					}
					{
						XYLineAnnotation line = new XYLineAnnotation(x+offset-dx/2, p, x+offset+dx/2, p, basicStroke, paint);
						chart.getXYPlot().addAnnotation(line);
					}
				}
			}
			iSeries++;
		}
		return chart;

	}
	private static void drawChart(JFreeChart chart) {
		File file = new File("junk.png");
		try {
			ChartUtilities.saveChartAsPNG(file, chart, 100, 100);
		} catch (IOException e) {
			throw new Error(e);
		}
		file.delete();
	}

	/**
	 * @param coordinates
	 * @param name
	 * @return the series
	 */
	public static XYSeries getXYSeries(ArrayList<Coordinate> coordinates, String name) {
		XYSeries retval = new XYSeries(name);
		for (Coordinate coordinate : coordinates) {
			retval.add(coordinate.getX(), coordinate.getY());
		}
		return retval;
	}

	/**
	 * adds the annotations to the XYPlot in the chart
	 * @param chart
	 * @param xyAnnotations
	 */
	public static void addAnnotations(JFreeChart chart, ArrayList<XYAnnotation> xyAnnotations) {
		XYPlot plot = chart.getXYPlot();
		for (XYAnnotation xyAnnotation : xyAnnotations) {
			plot.addAnnotation(xyAnnotation);
		}
	}

}
