package gov.fnal.eag.util.recordJar;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.LinkedList;
import java.util.Properties;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

/**
 * Reads and writes {@link java.util.LinkedList LinkedList}s of {@link java.util.Properties Properties} in record-jar format.
 * See <cite>The Art of UNIX Programming</cite> by Eric Raymond,
 * <a href="http://www.faqs.org/docs/artu/ch05s02.html#id2906931">
 * Section 5.2.4</a> for details of the file format.
 * 
 * The reader can transparently read gzipped files. 
 * 
 * @author neilsen
 * @version $Name:  $ ($Id: RecordJar.java,v 1.4 2008/02/22 17:36:40 descvs Exp $)
 */
public class RecordJar extends LinkedList<Properties> {
	
	/**
	 * 
	 */
	private static final long serialVersionUID = -7052938645383445150L;

	/**
	 * Construct a record-jar using data from an input stream.
	 * The input stream may be gzipped.
	 * 
	 * @param inStream the stream from which to create the RecordJar
	 * @throws IOException 
	 * 
	 */
	public RecordJar(InputStream inStream) throws IOException {
		super();
		load(inStream);
	}

	/**
	 * Create a record-jar with no records.
	 */
	public RecordJar() {
		super();
	}
	
	/**
	 * Create a new record-jar, using data from a named URL of file
	 * 
	 * @param name the string representation of the URL or the file name
	 * @throws IOException thrown when reading from the URL fails
	 */
	public RecordJar(String name) throws IOException {
		super();
		load(name);
	}
	
	/**
	 * Reads a record-jar from an input stream
	 * 
	 * @param inStream
	 * @throws IOException if there is IO trouble
	 */
	public void load(InputStream inStream) throws IOException {
		// Use a buffered stream so we can test whether it is gzipped
		BufferedInputStream bufferedInStream =
			new BufferedInputStream(inStream);
		// If it is gzipped, read it through a GZIP steam
		InputStream uncompressedInStream = isGZIP(bufferedInStream) ? 
			new GZIPInputStream(bufferedInStream) : bufferedInStream;
		InputStreamReader converter =
			new InputStreamReader(uncompressedInStream);
		BufferedReader reader = new BufferedReader(converter);
		String line = "";
		String accumulatedRecord = "";
		
		while ((line = reader.readLine()) != null) {
			if (line.startsWith("%%")) {
				if (accumulatedRecord.length() > 0) {
					addRecord(accumulatedRecord);
					accumulatedRecord = "";
				}
			} else {
				accumulatedRecord += line + "\n";
			}
		}
		if (accumulatedRecord.length() > 0) {
			addRecord(accumulatedRecord);
			accumulatedRecord = "";
		}		
	}
	
	/**
	 * Load a record-jar from a URL
	 * 
	 * @param url the URL from which to get the record-jar
	 * @throws IOException thrown when reading from the URL fails
	 */
	public void load(URL url) throws IOException {
		load(url.openStream());
	}
	
	/**
	 * Load a record-jar from a file or URL represented by a string
	 * 
	 * @param name the string representation of the URL or the file name
	 * @throws IOException thrown when reading from the URL of file fails
	 */
	public void load(String name) throws IOException {
		try {
			load(new URL(name));
		} catch (MalformedURLException e) {
			File file = new File(name);
			URL url = file.toURI().toURL();
			load(url);
		} 
	}
	
	/**
	 * Determines whether an input stream is gzipped
	 * 
	 * @param in the input stream to check (must be a buffered stream)
	 * @return true if the stream is gzipped
	 * @throws IOException if there is a problem reading the stream
	 */
	private boolean isGZIP(InputStream in) throws IOException {
		if (!in.markSupported()) {
			throw new IllegalArgumentException(
					"Cannot check if this input is gzipped");
		}
		in.mark(3);
		int read1 = in.read();
		int read2 = in.read();
		in.reset();
		int magic2 = GZIPInputStream.GZIP_MAGIC >> 8;
		int magic1 = GZIPInputStream.GZIP_MAGIC & 0xff;
		return (read1 == magic1) && (read2 == magic2);
	}
	
	/**
	 * Add a record generated from an input stream.
	 * 
	 * @param recordString the stream from which to generate the Properties
	 * 			record
	 */
	private void addRecord(String recordString) throws IOException {
		
		// For the funny little dance to avoid 
		// using depricated StringBufferInputStream
		// See http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4094886
		byte[] recordInputBytes;
		ByteArrayInputStream recordInputStream;
		Properties record;
		record = new Properties();
		recordInputBytes = recordString.getBytes("ISO-8859-1");
		recordInputStream = new ByteArrayInputStream(recordInputBytes);
		record.load(recordInputStream);
		addLast(record);
	}
	
	/**
	 * Write the record-jar to an output stream
	 * 
	 * @param out the raw output stream
	 * @param compressed if true, the output will be gzipped before it
	 * 			it is written to the stream
	 * @throws IOException
	 */
	public void store(OutputStream out, boolean compressed) throws IOException {
		OutputStream outStream = compressed ? new GZIPOutputStream(out) : out;
		ByteArrayOutputStream propertyStream;
		String recordString = "";
		for(Properties record: this) {
			recordString = "";
			propertyStream = new ByteArrayOutputStream();
			record.store(propertyStream,"");
			for(String line: propertyStream.toString("ISO-8859-1").split("\n")) {
				line = line.replaceFirst("#","%%");
				line = line.replaceFirst("=",": ");
				recordString += line + "\n";
			}
			outStream.write(recordString.getBytes("ISO-8859-1"));
			outStream.flush();
		}
		outStream.close();
	}
	
}
