package gov.fnal.eag.util.stopwatch;


import java.io.File;
import java.io.IOException;
import java.io.PrintStream;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

/**
 * 
 * helps keep track of time in programs in a simple way
 * 
 */


public class Stopwatch {
	/**
	 * the maximum number of data points remembered for plotting
	 */
	public final int maxDataPointSetSize = 10000;
	/**
	 * the name of this stopwatch
	 */
	private String name;

	/**
	 * the time it was reset
	 */
	private long resetTime;

	/**
	 * accumulated time
	 */
	private long elapsedTime;

	/**
	 * time it was started
	 */
	private long startTime;
	
	
	/**
	 * true if this stopwatch is running right now
	 */
	private boolean isRunning;

	/**
	 * number of times it has been started
	 */
	private long nStart = 0;

	private XYSeriesCollection xySeriesCollection;
	private XYSeries xySeries;
	
	/**
	 * creates and resets a Stopwatch
	 * @param name name of this Stopwatch
	 */
	public Stopwatch(String name) {
		this.name = name;
		reset();
	}

	/**
	 * starts the stopwatch
	 */
	public void start() {
		if (isRunning)
			return;
		isRunning = true;
		startTime = System.currentTimeMillis();
		nStart++;
	}

	/**
	 * stops the stopwatch and adds time since start to the elapsed time
	 */
	public void stop() {
		if (!isRunning) {
			return;
		}
		isRunning = false;
		long endTime = System.currentTimeMillis();
		elapsedTime += endTime - startTime;
		if (xySeries.getItemCount() < maxDataPointSetSize) {
			xySeries.add(nStart, (endTime-resetTime)/1000.0);
		}
	}

	/**
	 * get the elapsed time
	 * @return the elapsed time
	 */
	public long getElapsedTime() {
		if (isRunning) {
			long endTime = System.currentTimeMillis();
			elapsedTime = elapsedTime + endTime - startTime;
			startTime = endTime;
		}
		return elapsedTime;
	}

	/**
	 * get the name
	 * @return the name
	 */
	public String getName() {
		return name;
	}

	/**
	 * stops the watch and resets the elapsed time to 0.
	 */
	public void reset() {
		resetTime = System.currentTimeMillis();
		elapsedTime = 0;
		nStart = 0;
		xySeries = new XYSeries(name);
		xySeriesCollection = new XYSeriesCollection();
		xySeriesCollection.addSeries(xySeries);
	}

	/**
	 * prints summary infomration
	 * @param ps PrintStream that receives the summary
	 */
	public void tops(PrintStream ps) {
		if (isRunning) {
			ps.println("Warning: " + name + " still running");
		}
		String line = String.format(" %-50s   seconds=%10.3f     nStart=%6d    average=%13.6f ", 
				name, elapsedTime/1000.0, nStart, elapsedTime/1000.0/nStart);
		ps.println(line);
	}

	/**
	 * make plots of startTime vs. start number
	 * @param plotFileBaseName base file name of .jpg and .ps plots
	 */
	public void toPng(String plotFileBaseName) {
		JFreeChart chart = getChart();
		try {
			ChartUtilities.saveChartAsPNG(new File(plotFileBaseName+".png"), chart, 500, 300);
		} catch (IOException e) {
			throw new Error(e);
		}		
	}

    private JFreeChart getChart() {
		JFreeChart chart = ChartFactory.createScatterPlot(name,"counter","start time (sec)", xySeriesCollection, PlotOrientation.VERTICAL, true, true, true);
		return chart;
	}

	/**
     * get the number of times this stopwatch has been started
     * @return the number
     */
    public long getNStart() {
        return nStart;
    }

    /**
     * get the size of the data point set
     * @return the size
     */
    public int getDataPointSetSize() {
           return xySeries.getItemCount();
    }
}