/**
 * write and read Yanny parameter files.
 */
package gov.fnal.eag.util.yanny;

import gov.fnal.eag.util.misc.StringUtil;
import gov.fnal.eag.util.plot.JFreeChartUtil;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

import uk.ac.starlink.table.ColumnInfo;
import uk.ac.starlink.table.DefaultValueInfo;
import uk.ac.starlink.table.DescribedValue;
import uk.ac.starlink.table.RowListStarTable;
import uk.ac.starlink.table.RowSequence;
import uk.ac.starlink.table.StarTable;

/**
 * read Yanny Parameter Files into Startable, and other utilities
 * @author stoughto
 *
 */
public class Table {
	private static HashMap<String,String> tableToYanny;
	private static HashMap<String,Class<?>> yannyToTable;
	/**
	 * set to true for some chatter while determining whether tables are "equal"
	 */

	public static boolean equalsVerbose = false;
	
	static {
		tableToYanny = new HashMap<String,String>();
		// "Double" is how java refers to the data type and "double" is how tables refers to the data type
		tableToYanny.put("Double","double");
		tableToYanny.put("Float", "float");
        tableToYanny.put("Integer","int");
        tableToYanny.put("Long","long");
		tableToYanny.put("String","char");
		yannyToTable = new HashMap<String,Class<?>>();
		Iterator<String> keysIt   = tableToYanny.keySet().iterator();
		while (keysIt.hasNext()) {
			String key = keysIt.next();
			String value = tableToYanny.get(key);
			Class<?> tableClass = getClassForName(key);
			// Note that the order is switched.
			yannyToTable.put(value, tableClass);
		}		
	}

	/**
	 * @param key
	 * @return the results of Class.forName("java.lang."+key)
	 * @throws Error if the class is not found
	 */
	public static Class<?> getClassForName(String key) {
		try {
			Class<?> retval = Class.forName("java.lang."+key);
			return retval;
		} catch (ClassNotFoundException e) {
			throw new Error(e);
		}
	}
	/**
	 * write the table to the file
	 * @param rlst table to write
	 * @param fileName file to write
	 * @throws IOException if there is trouble
	 */
	public static void write(RowListStarTable rlst, String fileName) throws IOException {
		OutputStream outputStream = new FileOutputStream(fileName);
		write(rlst, outputStream);
	}

	/**
	 * write the table to the output stream	 * @param rlst table to write
	 * @param rlst table to write
	 * @param outputStream output stream to write
	 * @throws IOException if there is trouble
	 */
	public static void write(RowListStarTable rlst, OutputStream outputStream) throws IOException {
		PrintStream ps = new PrintStream(outputStream);
        List<?> list = get(rlst);
        Iterator<?> it = list.iterator();
        while (it.hasNext()) {
            DescribedValue dv = (DescribedValue) it.next();
            ps.println(dv.getInfo().getName()+" "+dv.getValue());
        }
 
		ps.println("typedef struct {");
		for (int icol=0; icol<rlst.getColumnCount(); icol++) {
			ColumnInfo ci = rlst.getColumnInfo(icol);
			String tableName = ci.getName();
			String tableClassName = DefaultValueInfo.formatClass(ci.getContentClass());
			String yannyClass = tableToYanny.get(tableClassName);
 			ps.print("   "+yannyClass+" "+tableName);
			if (yannyClass.equals("char")) {
				ps.print("[100]");
			}
			ps.println(";");
		}
		String schemaName = rlst.getName();
		ps.println("} "+schemaName);
		ps.println();
		for (int irow=0; irow<rlst.getRowCount(); irow++) {
			ps.print(schemaName+" ");
			for (int icol=0; icol<rlst.getColumnCount(); icol++) {
				ps.print(rlst.getCell(irow, icol)+" ");
			}
			ps.print("\n");
		}
		outputStream.close();
	}
	
	/**
	 * create a RowListStarTable from the file
	 * @param fileName file to read
	 * @return the resulting RowListStarTable
	 * @throws IOException if there is trouble opening fileName as a FileReader
	 */
	public static RowListStarTable read(String fileName) throws IOException {
		BufferedReader bufferedReader = new BufferedReader(new FileReader(fileName));
		return read(bufferedReader);
	}

	/**
	 * create a RowListStarTable from the reader
	 * @param bufferedReader the buffered reader
	 * @return the resulting RowListStarTable
	 */
	public static RowListStarTable read(BufferedReader bufferedReader) {
		String line;
		boolean typedef = false;
        boolean header = true;
		String schemaName = "unknown";
		LinkedList<ColumnInfo> columnInfoList = new LinkedList<ColumnInfo>();
        LinkedList<DescribedValue> parameterList = new LinkedList<DescribedValue>();
		RowListStarTable rlst = null;
		// like for_file in tcl
			try {
				while ( (line = bufferedReader.readLine()) != null) {
				    //System.out.println(" Table.read:  line="+line);
				    line = line.trim();
					if (!isComment(line)) {
						//defines things from typedef { to close } to be typedef
						if (typedef) {
							if (isEndOfTypedefStruct(line)) {
								typedef = false;
								schemaName = getSchemaName(line);
								ColumnInfo[] columnInfo = getColumnInfo(columnInfoList);
								rlst = new RowListStarTable(columnInfo);
				                setAllParameters(parameterList, rlst);
				            } else {
				                columnInfoList.add(getColumnInfo(line));
				            }
				        } else {	
							if (isTypedefStruct(line)) {
								typedef = true;
				                header = false;
							} else {
				                putParameterOrAddLine(line, header, schemaName, parameterList, rlst);
							}
						}
					}
				}
			} catch (IOException e) {
				throw new Error(e);
			}
		if (rlst != null) {
			rlst.setName(schemaName);
		} else {
			ColumnInfo[] columnInfo = new ColumnInfo[0];
			rlst = new RowListStarTable(columnInfo);
		}
		return rlst;
	}

	private static void putParameterOrAddLine(String line, boolean header, String schemaName, LinkedList<DescribedValue> parameterList, RowListStarTable rlst) {
		if (header) {
		   putParameter(line, parameterList);
		} else {
		    add(line, schemaName, rlst);
		}
	}

	private static void setAllParameters(LinkedList<DescribedValue> parameterList, RowListStarTable rlst) {
		Iterator<DescribedValue> pit = parameterList.iterator();
		while (pit.hasNext()) {
		    rlst.setParameter(pit.next());
		}
	}

	private static ColumnInfo[] getColumnInfo(LinkedList<ColumnInfo> columnInfoList) {
		ColumnInfo[] columnInfo = new ColumnInfo[columnInfoList.size()];
		for (int i=0; i<columnInfoList.size(); i++) {
			columnInfo[i] = columnInfoList.get(i);
		}
		return columnInfo;
	}
	private static void putParameter(String line, LinkedList<DescribedValue> parameterList) {
        StringTokenizer st = new StringTokenizer(line, " ");
        if (st.countTokens() < 2) return;
        String name = st.nextToken();
        String value = st.nextToken();
        while (st.hasMoreTokens()) {
            value += " "+st.nextToken();
        }
        DefaultValueInfo dvi = new DefaultValueInfo(name, String.class);
        DescribedValue dv = new DescribedValue(dvi, value);
        parameterList.add(dv);
        return;
    }

    /**
	 * create a row out of the line, and append it to the table
	 * @param line the line containing the values for this object
     * @param schemaName a descriptive name for this schema
	 * @param rlst the table
	 * @throws IllegalAccessException 
	 * @throws InstantiationException 
	 * @throws NoSuchMethodException 
	 * @throws SecurityException 
	 * @throws InvocationTargetException 
	 * @throws IllegalArgumentException 
	 */
	@SuppressWarnings("unchecked")
	private static void add(String line, String schemaName, RowListStarTable rlst) {
        StringTokenizer st = new StringTokenizer(line, " \t");
        //System.out.println(" in add: st.countTokens()="+st.countTokens());
		if (st.countTokens() != rlst.getColumnCount()+1) {
			return;
		}
		Class[] classDeclaration = {String.class};
        String firstToken = st.nextToken();
 		if (firstToken.equalsIgnoreCase(schemaName)) {
			Object[] row = new Object[rlst.getColumnCount()];
			for (int icol=0; icol<rlst.getColumnCount(); icol++) {
				String[] value = {st.nextToken()};
				Constructor<?> constructor;
				try {
					constructor = rlst.getColumnInfo(icol).getContentClass().getDeclaredConstructor(classDeclaration);
					row[icol] = constructor.newInstance((Object[]) value);
				} catch (SecurityException e) {
					throw new Error(e);
				} catch (NoSuchMethodException e) {
					throw new Error(e);
				} catch (IllegalArgumentException e) {
					throw new Error(e);
				} catch (InstantiationException e) {
					throw new Error(e);
				} catch (IllegalAccessException e) {
					throw new Error(e);
				} catch (InvocationTargetException e) {
					throw new Error(e);
				}
			}
			rlst.addRow(row);
		}
	}

    /**
     * put the name,value pair into the header of the RowListStarTable
     * @param name parameter name
     * @param value parameter value, always a String
     * @param rlst the table
     */
    public static void put(String name, String value, RowListStarTable rlst) {
        DefaultValueInfo dvi = new DefaultValueInfo(name, String.class);
        DescribedValue dv = new DescribedValue(dvi, value);
        rlst.setParameter(dv);
    }
    
    /**
     * get the value of the names parameter
     * @param name the parameter to get
     * @param rlst the table
     * @return the value
     */
    public static String get(String name, RowListStarTable rlst) {
        DescribedValue dv = rlst.getParameterByName(name);
        if (dv == null) return null;
        return (String) dv.getValue();
    }
    
    /**
     * return all of the named parameters
     * @param rlst the table
     * @return List of DescribedValues
     */
    @SuppressWarnings("unchecked")
	public static List get(RowListStarTable rlst) {
        return rlst.getParameters();
    }
	/**
	 * return a ColumnInfo based on the line
	 * @param line input line
	 * @return ColumnInfo information
	 */
	private static ColumnInfo getColumnInfo(String line) {
		StringTokenizer st = new StringTokenizer(line, " ;[]");
		if (st.countTokens() < 2) {
			return null;
		}
		String yannyType = st.nextToken().trim();
		String name = st.nextToken().trim();
		Class<?> tableClass = yannyToTable.get(yannyType);
		return new ColumnInfo(name, tableClass, name);
	}

	/**
	 * extract the schema name from the line that ends the typedef struct stanze
	 * @param line input line
	 * @return the schema name
	 */
	private static String getSchemaName(String line) {
		StringTokenizer st = new StringTokenizer(line, " ;");
		if (st.countTokens() < 2 ) {
			return "unknown";
		} 
		st.nextToken();
		return st.nextToken();
	}

	/**
	 * true is this line ends the typedef struct stanza
	 * @param line input line
	 * @return true if the line contains a } followed by schema name
	 */
	private static boolean isEndOfTypedefStruct(String line) {
		if (line.startsWith("}")) return true;
		return false;
	}

	/**
	 * true if this line begins the typedef struct stanza
	 * @param line input line
	 * @return true if this line contains "typedef struct {"
	 */
	private static boolean isTypedefStruct(String line) {
		int  indexTypedef;
		if ( (indexTypedef = line.indexOf("typedef")) == -1) return false;
		int indexStruct;
		if ( (indexStruct = line.indexOf("struct", indexTypedef+1)) == -1) return false;
		if (line.indexOf("{", indexStruct) == -1) return false;
		return true;
	}

	/**
	 * true if this line is a comment
	 * @param line input line
	 * @return true if the line begins with #
	 */
	private static boolean isComment(String line) {
		if (line.startsWith("#")) return true;
		return false;
	}

	/**
	 * given a name find the first column that matches the name, ignoring case
	 * @param rlst the StarTable
	 * @param name the column name
	 * @return the index, or -1, if there is no such column
	 */
	public static int getIndexByName(StarTable rlst, String name) {
		String colNames = "";
		for (int col=0; col<rlst.getColumnCount(); col++) {
			ColumnInfo columnInfo = rlst.getColumnInfo(col);
			if (name.equalsIgnoreCase(columnInfo.getName())) {
				return col;
			}
			colNames += columnInfo.getName()+" ";
		}
		throw new Error("can not find name="+name+" in colNames="+colNames);
	}

	/**
	 * determine whether two StarTables are equal
	 * @param starTable1
	 * @param starTable2
	 * @return true or false
	 */
	public static boolean startableEquals(RowListStarTable starTable1, RowListStarTable starTable2) {
		long nr1 = starTable1.getRowCount();
		long nr2 = starTable2.getRowCount();
		if (nr1 != nr2) {
			if (equalsVerbose) {
				System.out.println(" in startableEquals:  nr1="+nr1+" nr2="+nr2);
			}
			return false;
		}
		return true;
	}
	/**
	 * @param table input 
	 * @return a copy of the input, as a RowListStarTable
	 */
	public static RowListStarTable makeRowListStarTable(StarTable table) {
		System.out.println(" in makeRowListStarTable:  size="+table.getRowCount());
		RowListStarTable rlst = new RowListStarTable(table);
		for (long i=0; i<table.getRowCount(); i++) {
			try {
				rlst.addRow(table.getRow(i));
			} catch (IOException e) {
				throw new Error("i="+i, e);
			}
		}
		return rlst;
	}
	/**
	 * @param tableDef defines the columns, for example "double ra; double dec; int nPhotons"
	 * the data types are:  double float int long char
	 * @return a blank table
	 */
	public static RowListStarTable makeBlankTable(String tableDef) {
		String[] typeNames = tableDef.split(";");
		ArrayList<ColumnInfo> columnInfos = new ArrayList<ColumnInfo>();
		for (String typeName : typeNames) {
			String[] tn = StringUtil.removeDuplicateWhitespace(typeName).toString().split(" ");
			Class<?> tableClass = yannyToTable.get(tn[0]);
			String name = tn[1];
			ColumnInfo columnInfo = new ColumnInfo(name, tableClass, name);
			columnInfos.add(columnInfo);
		}
		ColumnInfo[] a = new ColumnInfo[typeNames.length];
		RowListStarTable rlst = new RowListStarTable(columnInfos.toArray(a));
		return rlst;
	}

	/**
	 * make a png from the table
	 * @param fileName where to write the plot
	 * @param starTable the table to plot
	 * @param x variable on the x axis
	 * @param y variable on the y axis
	 */
	public static void toPng(String fileName, StarTable starTable, String x, String y) {
		JFreeChart chart = getChart(fileName, starTable, x, y);
		JFreeChartUtil.toPng(chart, fileName);
	}

	/**
	 * makes a scatter plot of two columns in a StarTable
	 * @param fileName
	 * @param starTable
	 * @param x
	 * @param y
	 * @return the chart
	 */
	public static JFreeChart getChart(String fileName, StarTable starTable, String x, String y) {		
		XYSeries xy = getXYSeries(fileName, starTable, x, y);	
		XYSeriesCollection xysc = new XYSeriesCollection(xy);
		JFreeChart chart = ChartFactory.createScatterPlot(fileName, x, y, xysc, PlotOrientation.VERTICAL, true, true, true);
		return chart;
	}

	/**
	 * makes an XYSeries of two columns in a StarTable
	 * @param name
	 * @param starTable
	 * @param xName
	 * @param yName
	 * @return the series
	 */
	public static XYSeries getXYSeries(String name, StarTable starTable, String xName, String yName) {
		XYSeries xy = new XYSeries(name);
		try {
			int xi = getIndexByName(starTable, xName);
			int yi = getIndexByName(starTable, yName);
			RowSequence rs = starTable.getRowSequence();
			while (rs.next()) {
				Object[] row = rs.getRow();
				double x = ((Number) row[xi]).doubleValue();
				double y = ((Number) row[yi]).doubleValue();
				xy.add(x,y);
			}
			return xy;
		} catch (IOException e) {
			throw new Error(e);
		}
	}
	/**
	 * @param inputStream
	 * @return the resulting RowListStarTable
	 */
	public static RowListStarTable read(InputStream inputStream) {
		return read(new BufferedInputStream(inputStream));
	}
}
