package gov.fnal.eag.wcs;

/**
 * is a matrix to rotate SphericalCoordinates
 * 
 * @author stoughto
 *
 * Transliterated from SLALIB modules 
 *   prebn.c for FK4 precession matrix
 *   proc.c for FK5 precession matrix
 *   deuler.c for Euler angles --> precession matrix conversion
 *   
 *  Reference:                                                                  
 *     Kinoshita, H. (1975) 'Formulas for precession', SAO Special              
 *     Report No. 364, Smithsonian Institution Astrophysical                    
 *     Observatory, Cambridge, Massachusetts.
 */
public class RotationMatrix {
	static final double DAS2R = Math.toRadians(1.0/3600.0);
	double[][] matrix = new double[3][3];
	/**
	 * defines a rotation matrix to apply precession
	 * @param epochBegin beginning epoch (years)
	 * @param epochEnd ending epoch (years)
	 * @param system FK4 or FK5
	 * 
	 * the epochs are Besselian for FK4 and TDB (loosely ET) Julian for FK5
	 */
	public RotationMatrix(double epochBegin, double epochEnd, ReferenceSystem system) {
		switch(system) {
		case FK4: {
			/* Interval between basic epoch B1850.0 and beginning epoch in TC */
			double bigt  = ( epochBegin - 1850.0 ) / 100.0;

			/* Interval over which precession required, in tropical centuries */
			double t = ( epochEnd - epochBegin ) / 100.0;

			/* Euler angles */
			double tas2r = t * DAS2R;
			double w = 2303.5548 + ( 1.39720 + 0.000059 * bigt ) * bigt;
			double zeta = (w + ( 0.30242 - 0.000269 * bigt + 0.017996 * t ) * t ) * tas2r;
			double z = (w + ( 1.09478 + 0.000387 * bigt + 0.018324 * t ) * t ) * tas2r;
			double theta = ( 2005.1125 + ( - 0.85294 - 0.000365* bigt ) * bigt +
					( - 0.42647 - 0.000365 * bigt - 0.041802 * t ) * t ) * tas2r;
			euler( "ZYZ", -zeta, theta, -z);
			break;
		}
		case FK5: {
			/*
			  Though the matrix method itself is rigorous, the precession             
			  angles are expressed through canonical polynomials which are            
			  valid only for a limited time span.  There are also known               
			  errors in the IAU precession rate.  The absolute accuracy               
			  of the present formulation is better than 0.1 arcsec from               
			  1960AD to 2040AD, better than 1 arcsec from 1640AD to 2360AD,           
			  and remains below 3 arcsec for the whole of the period                  
			  500BC to 3000AD.  The errors exceed 10 arcsec outside the               
			  range 1200BC to 3900AD, exceed 100 arcsec outside 4200BC to             
			  5600AD and exceed 1000 arcsec outside 6800BC to 8200AD.                 
			  The SLALIB routine slaPrecl implements a more elaborate                 
			  model which is suitable for problems spanning several                   
			  thousand years.                                                         
			 **                                                                              
			 **  References:                                                                 
			 **     Lieske,J.H., 1979. Astron. Astrophys.,73,282.                            
			 **          equations (6) & (7), p283.                                          
			 **     Kaplan,G.H., 1981. USNO circular no. 163, pa2.                         
			 */

			/* Interval between basic epoch J2000.0 and beginning epoch (JC) */
			double t0 = ( epochBegin - 2000.0 ) / 100.0;

			/* Interval over which precession required (JC) */
			double t =  ( epochEnd - epochBegin ) / 100.0;

			/* Euler angles */
			double tas2r = t * DAS2R;
			double w = 2306.2181 + ( ( 1.39656 - ( 0.000139 * t0 ) ) * t0 );
			double zeta = (w + ( ( 0.30188 - 0.000344 * t0 ) + 0.017998 * t ) * t ) * tas2r;
			double z = (w + ( ( 1.09468 + 0.000066 * t0 ) + 0.018203 * t ) * t ) * tas2r;
			double theta = ( ( 2004.3109 + ( - 0.85330 - 0.000217 * t0 ) * t0 )
					+ ( ( -0.42665 - 0.000217 * t0 ) - 0.041833 * t ) * t ) * tas2r;
			euler ( "ZYZ", -zeta, theta, -z);
			break;
		}
		}
	}

	private void euler(String order, double phi, double theta, double psi) {
		/*                                                                              
		 **  - - - - - - - - - -                                                         
		 **   s l a D e u l e r                                                          
		 **  - - - - - - - - - -                                                         
		 **                                                                              
		 **  Form a rotation matrix from the Euler angles - three successive             
		 **  rotations about specified Cartesian axes.                                   
		 **                                                                              
		 **  (double precision)                                                          
		 **                                                                              
		 **  Given:                                                                      
		 **    *order char     specifies about which axes the rotations occur            
		 **    phi    double   1st rotation (radians)                                    
		 **    theta  double   2nd rotation (   "   )                                    
		 **    psi    double   3rd rotation (   "   )                                    
		 **                                                                              
		 **  Returned:                                                                   
		 **    rmat   double[3][3]  rotation matrix                                      
		 **                                                                              
		 **  A rotation is positive when the reference frame rotates                     
		 **  anticlockwise as seen looking towards the origin from the                   
		 **  positive region of the specified axis.                                      
		 **                                                                              
		 **  The characters of order define which axes the three successive              
		 **  rotations are about.  A typical value is 'zxz', indicating that             
		 **  rmat is to become the direction cosine matrix corresponding to              
		 **  rotations of the reference frame through phi radians about the              
		 **  old z-axis, followed by theta radians about the resulting x-axis,           
		 **  then psi radians about the resulting z-axis.                                
		 **                                                                              
		 **  The axis names can be any of the following, in any order or                 
		 **  combination:  x, y, z, uppercase or lowercase, 1, 2, 3.  Normal             
		 **  axis labelling/numbering conventions apply;  the xyz (=123)                 
		 **  triad is right-handed.  Thus, the 'zxz' example given above                 
		 **  could be written 'zxz' or '313' (or even 'zxz' or '3xz').  Order            
		 **  is terminated by length or by the first unrecognized character.             
		 **                                                                              
		 **  Fewer than three rotations are acceptable, in which case the later          
		 **  angle arguments are ignored.  Zero rotations leaves rmat set to the         
		 **  identity matrix.                                                            
		 **                                                                              
		 **  Last revision:   9 December 1996                                            
		 **                                                                              
		 **  Copyright P.T.Wallace.  All rights reserved.                                
		 */
		int j, i, l, n, k;
		double angle, s, c , w;
		double[][] result = new double[3][3];
		double[][] rotn = new double[3][3];
		double[][] wm = new double[3][3];
		char axis;
		/* Initialize result matrix */
		for ( j = 0; j < 3; j++ ) {
			for ( i = 0; i < 3; i++ ) {
				result[i][j] = ( i == j ) ? 1.0 : 0.0;
			}
		}

		/* Establish length of axis string */
		l = order.length();

		/* Look at each character of axis string until finished */
		for ( n = 0; n < 3; n++ ) {
			if ( n <= l ) {

				/* Initialize rotation matrix for the current rotation */
				for ( j = 0; j < 3; j++ ) {
					for ( i = 0; i < 3; i++ ) {
						rotn[i][j] = ( i == j ) ? 1.0 : 0.0;
					}
				}

				/* Pick up the appropriate Euler angle and take sine & cosine */
				switch ( n ) {
				case 0 :
					angle = phi;
					break;
				case 1 :
					angle = theta;
					break;
				default:
					angle = psi;
				break;
				}
				s = Math.sin ( angle );
				c = Math.cos ( angle );

				/* Identify the axis */
				axis =  order.charAt(n);
				if ( ( axis == 'Y' ) || ( axis == 'y' ) || ( axis == '2' ) ) {

					/* Matrix for y-rotation */
					rotn[0][0] = c;
					rotn[0][2] = -s;
					rotn[2][0] = s;
					rotn[2][2] = c;
				} else {

					/* Matrix for z-rotation */
					rotn[0][0] = c;
					rotn[0][1] = s;
					rotn[1][0] = -s;
					rotn[1][1] = c;
				}
				/* Apply the current rotation (matrix rotn x matrix result) */
				for ( i = 0; i < 3; i++ ) {
					for ( j = 0; j < 3; j++ ) {
						w = 0.0;
						for ( k = 0; k < 3; k++ ) {
							w += rotn[i][k] * result[k][j];
						}
						wm[i][j] = w;
					}
				}
				for ( j = 0; j < 3; j++ ) {
					for ( i= 0; i < 3; i++ ) {
						result[i][j] = wm[i][j];
					}
				}
			}
		}

		/* Copy the result */
		for ( j = 0; j < 3; j++ ) {
			for ( i = 0; i < 3; i++ ) {
				matrix[i][j] = result[i][j];
			}
		}

	}
	/**
	 * apply this  matrix to the coordinate
	 * @param sc the coordinate to rotate
	 */
	public void rotate(SphericalCoordinate sc) {
		double[] in = {sc.getX(),sc.getY(), sc.getZ()};
		double[] out = new double[3];
		/* Matrix dm * vector va -> vector vw */
		for ( int j = 0; j < 3; j++ ) {
			for ( int i = 0; i < 3; i++ ) {
				out[j] += matrix[j][i] * in[i];
			}
		}
		sc.set(out);
	}
	/**
	 * specifies a reference system for coordinates
	 * @author stoughto
	 *
	 */
	public enum ReferenceSystem {
		/**
		 * the old system pre-IAU1976, Bessel-Newcomb model, using                     
		 * Kinoshita's formulation (double precision)
		 */
		FK4,
		/**
		 * the new system, IAU 1976, using an approximation good to 0.1 arcsec from 1960AD to 2040AD, and good to 1 arcsec from 1640AD to 2360AD
		 */
		FK5
	}
}
