package gov.fnal.eag.wcs;

import gov.fnal.eag.healpix.PixTools;

import java.awt.geom.Point2D;

import javax.vecmath.Vector3d;


/**
 * deals with locations on the unit sphere.  Longitude ranges from 0,360 degrees and latitude ranges from -90,90 degrees
 * @author stoughto
 *
 */
public class SphericalCoordinate {
	private double x,y,z;
	private double longitude, latitude;
	private static String[] longitudeTypes = {"RA", "LONG"}; 
	private static String[] latitudeTypes = {"DEC", "LAT"}; 
	private static double tolerence = 1e-15;
	private static final PixTools pixTools = new PixTools();
	/**
	 * 	Galactic north pole ra= 12h 51m 26.282s in degrees
	 */
	public static final double raGalacticNorthPole;
	/**
	 * 	Galactic north pole dec= 27d 7m 42.01s in degrees
	 */
	public static final double decGalacticNorthPole;
	static {
		raGalacticNorthPole = degreesFromHMS("12:51:26.282");
		decGalacticNorthPole = degreesFromDMS("27:07:42.01");
	}
	/**
	 * 
	 * @param longitude
	 * @param latitude
	 */
	public SphericalCoordinate(double longitude, double latitude) {
		init(longitude,latitude);
	}

	/**
	 * set values from the vector
	 * @param xyz
	 */
	public void set(double[] xyz) {
		x = xyz[0];
		y = xyz[1];
		z = xyz[2];
		double norm = Math.sqrt(x*x+y*y+z*z);
		x /= norm;
		y /= norm;
		z /= norm;
		calculateLongitudeLatitude();
	}
	/**
	 * set new values of longitude and lattitude
	 * @param longitude
	 * @param lattitude
	 */
	public void set(double longitude,double lattitude) {
		init(longitude,lattitude);
	}
	/**
	 * @param coord1 first coordinate in degrees
	 * @param ctype1 first coordinate type, e.g. RA
	 * @param coord2 second coordinate in degrees
	 * @param ctype2 second coordinate type, e.g. DEC
	 */
	public SphericalCoordinate(double coord1, String ctype1, double coord2, String ctype2) {
		if (longitudeLatitude(ctype1, ctype2)) { 
			init(coord1, coord2);
		} else {
			init(coord2, coord1);
		}
	}
	
	/**
	 * construct from xyz in the vector
	 * @param xyz
	 */
	public SphericalCoordinate(double[] xyz) {
		set(xyz);
	}

	private boolean longitudeLatitude(String ctype1, String ctype2) {
		// easy case first
		boolean ctype1IsLongitude = isLongitude(ctype1);
		boolean ctype2IsLatitude = isLatitude(ctype2);
		if (ctype1IsLongitude && ctype2IsLatitude) return true;
		
		// they are flipped
		boolean ctype2IsLongitude = isLongitude(ctype2);
		boolean ctype1IsLatitude = isLatitude(ctype1);
		if (ctype1IsLatitude && ctype2IsLongitude) return false;
		

		throw new Error("can not figure out which is longitude and latitude:  ctype1="+ctype1+" ctype2="+ctype2);
		
		
	}
	
	private boolean isLongitude(String ctype) {
		boolean retval = false;
		for (String type : longitudeTypes) {
			if (ctype.contains(type)) {
				retval = true;
				break;
			}
		}
		return retval;
	}
	
	private boolean isLatitude(String ctype) {
		boolean retval = false;
		for (String type : latitudeTypes) {
			if (ctype.contains(type)) {
				retval = true;
				break;
			}
		}
		return retval;
	}
		

	private void init(double longitudeIn, double latitudeIn) {
		double longitudeRad = longitudeIn*Math.PI/180;
		double latitudeRad  = latitudeIn *Math.PI/180;
		
		double rho = Math.abs(Math.cos(latitudeRad));
		x = rho*Math.cos(longitudeRad);
		y = rho*Math.sin(longitudeRad);
		z = Math.sin(latitudeRad);
		calculateLongitudeLatitude();
	}

	private void calculateLongitudeLatitude() {
		this.latitude = Math.asin(z)*180.0/Math.PI;
		this.longitude = Math.atan2(y, x)*180.0/Math.PI;
		if (longitude < 0) longitude += 360;
	}
	
	/**
	 * @return longitude in degrees 
	 */
	public double getLongitude() {
		return longitude;
	}
	
	/**
	 * @return latitude in degrees
	 */
	public double getLatitude() {
		return latitude;
	}
	
	/**
	 * @param that
	 * @return true if x, y, and z are equal for this and that
	 */
	public boolean equals(SphericalCoordinate that) {
		if (Math.abs(this.x - that.x) > tolerence || Math.abs(this.y - that.y) > tolerence || Math.abs(this.z - that.z) > tolerence ) {
			return false;
		}
		return true;
	}
	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		String retval = super.toString()+" long="+getLongitude()+" lat="+getLatitude();
		return retval;
	}
	
	/**
	 * @param that
	 * @return the separation between this and that, in degrees
	 */
	public double separation(SphericalCoordinate that) {
		double dotProduct = x*that.x + y*that.y + z*that.z;
		double theta;
		if (dotProduct >= 1) {
			theta = 0;
		} else if (dotProduct <= -1) {
			theta = 180.0;
		} else {
			theta = Math.toDegrees(Math.acos(dotProduct));
		}
		return theta;
	}
	
	/**
	 * @param that
	 * @return deltaRa,deltaDec in arcseconds
	 */
	public Point2D.Double delta(SphericalCoordinate that) {
		double cosDec = Math.cos(Math.toRadians(getLatitude()));
		double deltaRa = that.getLongitude()-getLongitude();
		if (deltaRa < -180) {
			deltaRa += 360;
		} else if (deltaRa > 180) {
			deltaRa -= 360;
		}
		deltaRa *= 3600.0*cosDec;
		double deltaDec = that.getLatitude()-getLatitude();
		deltaDec *= 3600.0;
		Point2D.Double retval = new Point2D.Double(deltaRa, deltaDec);
		return retval;
	}
	/**
	 * @param min
	 * @param max
	 * @return the area bounded by the two coordinates
	 */
	public static double steradians(SphericalCoordinate min, SphericalCoordinate max) {
    	double ra1 = min.getLongitude();
    	double ra2 = max.getLongitude();
    	double dec1 = min.getLatitude();
    	double dec2 = max.getLatitude();
    	double raFraction = Math.abs(ra2-ra1)/360.0;
    	if (ra2 < ra1) raFraction = 1 - raFraction;
    	double integral = Math.abs(Math.sin(dec2*Math.PI/180.0) -  Math.sin(dec1*Math.PI/180.0));
    	double steradians = 2 * Math.PI * raFraction * integral;
    	return steradians;
	}

	/**
	 * @return the x value
	 */
	public double getX() {
		return x;
	}
	/**
	 * @return the y value
	 */
	public double getY() {
		return y;
	}
	/**
	 * @return the z value
	 */
	public double getZ() {
		return z;
	}

	/**
	 * @return a deep copy of this coordinate
	 */
	public SphericalCoordinate copy() {
		SphericalCoordinate retval = new SphericalCoordinate(longitude, latitude);
		return retval;
	}
	
	/**
	 * @param degrees
	 * @param minutes
	 * @param seconds
	 * @return decimal degrees
	 */
	public static double degreesFromDMS(double degrees, double minutes, double seconds) {
		double retval = degrees + minutes/60.0 + seconds/3600.0;
		return retval;
	}
	
	/**
	 * @param hours
	 * @param minutes
	 * @param seconds
	 * @return decimal degrees
	 */
	public static double degreesFromHMS(double hours, double minutes, double seconds) {
		double retval = 15*(hours + minutes/60.0 + seconds/3600.0);
		return retval;
	}

	/**
	 * @param string colon-separated hours:minutes:seconds
	 * @return value in decimal degrees
	 */
	public static double degreesFromHMS(String string) {
		String[] tokens = string.split(":");
		double h = Double.parseDouble(tokens[0]);
		double m = Double.parseDouble(tokens[1]);
		double s = Double.parseDouble(tokens[2]);
		return degreesFromHMS(h,m,s);
	}

	/**
	 * @param string colon-separated degrees:minutes:seconds
	 * @return value in decimal degrees
	 */
	public static double degreesFromDMS(String string) {
		String[] tokens = string.split(":");
		double d = Double.parseDouble(tokens[0]);
		double m = Double.parseDouble(tokens[1]);
		double s = Double.parseDouble(tokens[2]);
		return degreesFromDMS(d,m,s);
	}

	/**
	 * @param base 
	 * @return value of the 31 lsb of the longitude and latitude, x0red, the x0red with the non-sign bits of base
	 */
	public int getSeed(int base) {
		long lonBitsLong = Double.doubleToLongBits(longitude);
		long latBitsLong = Double.doubleToLongBits(latitude);	
		//System.out.println(String.format("lonBitsLong=%16X latBitsLong=%16X", lonBitsLong, latBitsLong));
		int allMask = 0x7fffffff;
		int lonBitsInt = (int) (lonBitsLong & allMask);
		int latBitsInt = (int) (latBitsLong & allMask);
		//System.out.println(String.format(" lonBitsInt=%16X  latBitsInt=%16X", lonBitsInt, latBitsInt));
		int lonMask = 0x55555555;
		int latMask = 0x2aaaaaaa;
		int lon = lonBitsInt & lonMask;
		int lat = latBitsInt & latMask;
		//System.out.println(String.format("        lon=%16X         lat=%16X", lon, lat));
		int retval = (lon | lat) ^ (base & allMask);
		//System.out.println(String.format("     retval=%16X  ", retval));
		return retval;
	}

	/**
	 * rotates the coordinated around the Z axis
	 * @param d rotation angle in degrees
	 */
	public void rotateZ(double d) {
		longitude += d;
		init(longitude, latitude);
	}
	/**
	 * rotates the coordinate around the X axis
	 * @param theta rotation angle in degrees
	 */
	public void rotateX(double theta) {
		double ct = Math.cos(Math.toRadians(theta));
		double st = Math.sin(Math.toRadians(theta));
		
		double tempY = y*ct - z*st;
		z = y*st + z*ct;
		y = tempY;
		calculateLongitudeLatitude();
	}

	/**
	 * rotates the coordinate around the Y axis
	 * @param theta
	 */
	public void rotateY(double theta) {
		double ct = Math.cos(Math.toRadians(theta));
		double st = Math.sin(Math.toRadians(theta));
		double temp = z*ct - x*st;
		x = z*st + x*ct;
		z = temp;	
		calculateLongitudeLatitude();
	}
	/**
	 * converts this coordinate from equatorial (J2000) to galactic
	 */
	public void equatorialToGalactic() {
		rotateZ(-raGalacticNorthPole);
		rotateY(decGalacticNorthPole-90);
	}
	
	/**
	 * converts this coordinate from galactic to equatorial (J2000)
	 */
	public void galacticToEquatorial() {
		rotateY(90-decGalacticNorthPole);
		rotateZ(raGalacticNorthPole);		
	}

	/**
	 * @param nSide
	 * @return the HealPix ring index for this coordinate
	 */
	public long getIpix_ring(int nSide) {
		Vector3d vector = new Vector3d(x,y,z);
		long retval = pixTools.vect2pix_ring(nSide, vector);
		return retval;
	}
	
}
