package gov.fnal.eag.wcs;
import java.awt.geom.AffineTransform;

import nom.tam.fits.Header;

	/**
	 * linear transformation
	 * @author stoughto
	 *
	 */
	public class TransformLinear implements Transform, TransformPlanar {

		private String ctype1, ctype2;
		private double crpix1, crpix2, crval1, crval2;
		private double sm1_1, sm1_2, sm2_1, sm2_2, smDeterminant;
		private double invSm1_1, invSm1_2,invSm2_1, invSm2_2;
		
		/**
		 * default constructor for the identity transform
		 */
		public TransformLinear() {
			this(new Header());
		}
		/**
		 * constructor which calls fromHeader()
		 * @param header
		 */
		public TransformLinear(Header header) {
			fromHeader(header);
		}
		
		/**
		 * @param colRef
		 * @param rowRef
		 * @param xRef
		 * @param yRef
		 * @param metersPerPixel
		 * @param rotationDegrees
		 */
		public TransformLinear(
				double colRef,
				double rowRef,
				double xRef,
				double yRef,
				double metersPerPixel,
				double rotationDegrees
		)  {
			Header header = new Header();
			HUtil.replaceKeywordValue(header, "CTYPE1", "LINEAR", "column");
			HUtil.replaceKeywordValue(header, "CTYPE2", "LINEAR", "row");
			HUtil.replaceKeywordValue(header, "CRPIX1", colRef, "reference column");
			HUtil.replaceKeywordValue(header, "CRPIX2", rowRef, "reference row");
			HUtil.replaceKeywordValue(header, "CRVAL1", xRef, "reference x position (meters)");
			HUtil.replaceKeywordValue(header, "CRVAL2", yRef, "reference y position");
			HUtil.replaceKeywordValue(header, "CDELT1", metersPerPixel, "");
			HUtil.replaceKeywordValue(header, "CDELT2", metersPerPixel, "");
			HUtil.replaceKeywordValue(header, "CROTA2", rotationDegrees, "");
			fromHeader(header);			
		}

		/**
		 * sets values from the header
		 * @param header
		 */
		public void fromHeader(Header header) {
			if ((ctype1 = header.getStringValue("CTYPE1")) == null) ctype1="LINEAR";
			if (!ctype1.equals("LINEAR")) {
				throw new IllegalArgumentException("ctype1 must be LINEAR but ctype1="+ctype1);
			}
			if ((ctype2 = header.getStringValue("CTYPE2")) == null) ctype2="LINEAR";
			if (!ctype2.equals("LINEAR")) {
				throw new IllegalArgumentException("ctype2 must be LINEAR but ctype2="+ctype2);
			}
			
			
	        this.crpix1 = ValueCheck.getDoubleValue(header, "CRPIX1", 0.0);
	        this.crpix2 = ValueCheck.getDoubleValue(header, "CRPIX2", 0.0);
	        this.crval1 = ValueCheck.getDoubleValue(header, "CRVAL1", 0.0);
	        this.crval2 = ValueCheck.getDoubleValue(header, "CRVAL2", 0.0);
	        if (HUtil.definedWithCdValues(header)) {
	        	// If one CD_i_j is specified, all other values default to 0.0, and ignore CDELTi and CROTAi
	        	sm1_1 = ValueCheck.getDoubleValue(header, "CD1_1", 0.0);
	        	sm2_1 = ValueCheck.getDoubleValue(header, "CD2_1", 0.0);
	        	sm1_2 = ValueCheck.getDoubleValue(header, "CD1_2", 0.0);
	        	sm2_2 = ValueCheck.getDoubleValue(header, "CD2_2", 0.0);
	        } else if (HUtil.definedWithPcValues(header)) {
	        	// If one of CDELTi or PCi_j are defined, use defaults that are the identity transform
	        	double cdelt1 = ValueCheck.getDoubleValue(header, "CDELT1", 1.0);
	        	double cdelt2 = ValueCheck.getDoubleValue(header, "CDELT2", 1.0);
	        	sm1_1 = cdelt1*ValueCheck.getDoubleValue(header, "PC1_1", 1.0);
	        	sm2_1 = cdelt2*ValueCheck.getDoubleValue(header, "PC2_1", 0.0);
	        	sm1_2 = cdelt1*ValueCheck.getDoubleValue(header, "PC1_2", 0.0);
	        	sm2_2 = cdelt2*ValueCheck.getDoubleValue(header, "PC2_2", 1.0);
	        } else {
	        	// Otherwise, default to the identity transform, using scaling and rotation if they are geven
	        	double cdelt1 = ValueCheck.getDoubleValue(header, "CDELT1", 1.0);
	        	double cdelt2 = ValueCheck.getDoubleValue(header, "CDELT2", 1.0);
	        	double crota = ValueCheck.getDoubleValue(header, "CROTA1", 0.0);
	        	crota = ValueCheck.getDoubleValue(header, "CROTA2", crota);
	           	double cosTheta = Math.cos(Math.toRadians(crota));
	           	double sinTheta = Math.sin(Math.toRadians(crota));
	           	sm1_1 =  cdelt1 * cosTheta;
	           	sm2_1 =  cdelt2 * sinTheta;
	           	sm1_2 = -cdelt1 * sinTheta;
	           	sm2_2 =  cdelt2 * cosTheta;           	
	        }
	    	smDeterminant = sm1_1*sm2_2 - sm2_1*sm1_2;
	    	if (Math.abs(smDeterminant) < Double.MIN_VALUE) {
	    		throw new IllegalArgumentException("singular matrix:  smDeterminant="+smDeterminant);
	    	}
	    	invSm1_1 =  sm2_2/smDeterminant;
	    	invSm1_2 = -sm2_1/smDeterminant;
	    	invSm2_1 = -sm1_2/smDeterminant;
	    	invSm2_2 =  sm1_1/smDeterminant;	
		}

		// pixel --> focal plane
		/**
		 * @see gov.fnal.eag.wcs.Transform#forward(gov.fnal.eag.wcs.Coordinate)
		 */
		public SphericalCoordinate forward(Coordinate coordinate) {
			double dx = coordinate.getX() - crpix1;
			double dy = coordinate.getY() - crpix2;
			SphericalCoordinate retval = new SphericalCoordinate(crval1 + sm1_1*dx + sm1_2*dy, crval2 + sm2_1*dx + sm2_2*dy);
			return retval;
		}
		/**
		 * pixel -> focal plane
		 * apply the forward transformation to the coordinate and replace its values
		 * @param coordinate 
		 */
		public void forwardInPlace(Coordinate coordinate) {
			double dx = coordinate.getX() - crpix1;
			double dy = coordinate.getY() - crpix2;
			double x = crval1 + sm1_1*dx + sm1_2*dy;
			double y = crval2 + sm2_1*dx + sm2_2*dy;			
			coordinate.setX(x);
			coordinate.setY(y);
		}

		// sky --> pixel
		/**
		 * @see gov.fnal.eag.wcs.Transform#inverse(gov.fnal.eag.wcs.SphericalCoordinate)
		 */
		public Coordinate inverse(SphericalCoordinate coordinate) {
			double dx = coordinate.getLongitude() - crval1;
			double dy = coordinate.getLatitude() - crval2;
			Coordinate retval = new Coordinate(crpix1 + invSm1_1*dx + invSm1_2*dy,  crpix2 + invSm2_1*dx + invSm2_2*dy);		
			return retval;
		}

		/**
		 * focal plane -> pixel
		 * @param coordinate 
		 */
		public void inverseInPlace(Coordinate coordinate) {
			double dx = coordinate.getX() - crval1;
			double dy = coordinate.getY() - crval2;
			double x = crpix1 + invSm1_1*dx + invSm1_2*dy;
			double y = crpix2 + invSm2_1*dx + invSm2_2*dy;
			coordinate.setX(x);
			coordinate.setY(y);
		}
		/**
		 * @see gov.fnal.eag.wcs.Transform#toHeader(nom.tam.fits.Header)
		 */
		public void toHeader(Header header) {
			HUtil.eraseWcsInfo(header);
			HUtil.replaceKeywordValue(header, "CTYPE1", "LINEAR", "");
			HUtil.replaceKeywordValue(header, "CTYPE2", "LINEAR", "");
			HUtil.replaceKeywordValue(header, "CRPIX1", crpix1, "");
			HUtil.replaceKeywordValue(header, "CRPIX2", crpix2, "");
			HUtil.replaceKeywordValue(header, "CRVAL1", crval1, "");
			HUtil.replaceKeywordValue(header, "CRVAL2", crval2, "");
			HUtil.replaceKeywordValue(header, "CD1_1", sm1_1, "");
			HUtil.replaceKeywordValue(header, "CD1_2", sm1_2, "");
			HUtil.replaceKeywordValue(header, "CD2_1", sm2_1, "");
			HUtil.replaceKeywordValue(header, "CD2_2", sm2_2, "");
		}
		
		/**
		 * @see java.lang.Object#toString()
		 */
		@Override
		public String toString() {
			String retval = "TransformLinear with crpix1="+crpix1+" crpix2="+crpix2+" crval1="+crval1+" crval2="+crval2+" sm1_1="+sm1_1+" sm1_2="+sm1_2+" sm2_1="+sm2_1+" sm2_2="+sm2_2;
			return retval;
		}
		
		/**
		 * @param that
		 * @return true if the elements are equal
		 */
		public boolean equals(TransformLinear that) {
			// if (!ctype1.equals(that.ctype1)) return false;
			// if (!ctype2.equals(that.ctype2)) return false;
			if (crpix1 != that.crpix1) return false;
			if (crpix2 != that.crpix2) return false;
			if (crval1 != that.crval1) return false;
			if (crval2 != that.crval2) return false;
			if (sm1_1 != that.sm1_1) return false;
			if (sm1_2 != that.sm1_2) return false;
			if (sm2_1 != that.sm2_1) return false;
			if (sm2_2 != that.sm2_2) return false;
			return true;
		}
		
		/**
		 * @see gov.fnal.eag.wcs.Transform#arcsecPerPixel()
		 */
		public double arcsecPerPixel() {
			double det = Math.sqrt(smDeterminant)*3600.0;	
			return det;
		}

		/**
		 * @return the AffineTransform representation of the forward transform
		 */
		public AffineTransform getForward() {
			double m00 = sm1_1;
			double m01 = sm1_2;
			double m02 = crval1 - sm1_1*crpix1 - sm1_2*crpix2;
			double m10 = sm2_1;
			double m11 = sm2_2;
			double m12 = crval2 - sm2_1*crpix1 - sm2_2*crpix2;
			AffineTransform retval = new AffineTransform(m00, m10, m01, m11, m02, m12);
			return retval;
		}
		
		/**
		 * @return the AffineTransform representation of the inverse transform
		 */
		public AffineTransform getInverse() {
			double m00 = invSm1_1;
			double m01 = invSm1_2;
			double m02 = crpix1 - invSm1_1*crval1 - invSm1_2*crval2;
			double m10 = invSm2_1;
			double m11 = invSm2_2;
			double m12 = crpix2 - invSm2_1*crval1 - invSm2_2*crval2;
			AffineTransform retval = new AffineTransform(m00, m10, m01, m11, m02, m12);
			return retval;
		}
		/**
		 * @see gov.fnal.eag.wcs.TransformPlanar#getDescription()
		 */
		public String getDescription() {
			return "linear transform";
		}

	}


