package gov.fnal.eag.wcs;
import java.io.PrintStream;

import nom.tam.fits.Header;

	/**
	 * linear transformation
	 * @author stoughto
	 *
	 */
	public class TransformTangentPlaneA7 implements Transform {

		private String ctype1, ctype2;
		private double crpix1, crpix2, crval1, crval2;
		private double a7;
		private double sm1_1, sm1_2, sm2_1, sm2_2, smDeterminant;
		private double invSm1_1, invSm1_2,invSm2_1, invSm2_2;
		/**
		 * sanity check on the maximum value of a7 allowed.  Typical a7 values are a few percent.
		 */
		public static final double a7Max = 0.5;
		/**
		 * watch out too close to the poles
		 */
		public static final double dectTol = 1e-15;
		/**
		 * @param that
		 * @return true if the elements are equal
		 */
		public boolean equals(TransformTangentPlaneA7 that) {
			if (!ctype1.equals(that.ctype1)) return false;
			if (!ctype2.equals(that.ctype2)) return false;
			if (crpix1 != that.crpix1) return false;
			if (crpix2 != that.crpix2) return false;
			if (crval1 != that.crval1) return false;
			if (crval2 != that.crval2) return false;
			if (a7 != that.a7) return false;
			if (sm1_1 != that.sm1_1) return false;
			if (sm1_2 != that.sm1_2) return false;
			if (sm2_1 != that.sm2_1) return false;
			if (sm2_2 != that.sm2_2) return false;
			return true;
		}
		/**
		 * default constructor for the identity transform
		 */
		public TransformTangentPlaneA7() {
			this(new Header());
		}
		/**
		 * constructor which calls fromHeader()
		 * @param header
		 */
		public TransformTangentPlaneA7(Header header) {
			fromHeader(header);
		}

		/**
		 * @param colRef
		 * @param rowRef
		 * @param raRef
		 * @param decRef
		 * @param degreesPerPixel
		 * @param rotationDegrees
		 * @param a7In 
		 */
		public TransformTangentPlaneA7(
				double colRef,
				double rowRef,
				double raRef,
				double decRef,
				double degreesPerPixel,
				double rotationDegrees,
				double a7In
		)  {
			Header header = new Header();
			HUtil.replaceKeywordValue(header, "CTYPE1", "RA---TAN", "RA type of transformation");
			HUtil.replaceKeywordValue(header, "CTYPE2", "DEC--TAN", "Dec type of transformation");
			HUtil.replaceKeywordValue(header, "CRPIX1", colRef, "reference column");
			HUtil.replaceKeywordValue(header, "CRPIX2", rowRef, "reference row");
			HUtil.replaceKeywordValue(header, "CRVAL1", raRef, "reference ra");
			HUtil.replaceKeywordValue(header, "CRVAL2", decRef, "reference dec");
			HUtil.replaceKeywordValue(header, "CDELT1", degreesPerPixel, "");
			HUtil.replaceKeywordValue(header, "CDELT2", degreesPerPixel, "");
			HUtil.replaceKeywordValue(header, "CROTA2", rotationDegrees, "");
			HUtil.replaceKeywordValue(header, "PV1_7", a7In,"");
			HUtil.replaceKeywordValue(header, "PV1_9", a7In,"");
			HUtil.replaceKeywordValue(header, "PV2_7", a7In,"");
			HUtil.replaceKeywordValue(header, "PV2_9", a7In,"");

			fromHeader(header);
		}	

		
		private void fromHeader(Header header) {
			
			if ((ctype1 = header.getStringValue("CTYPE1")) != null) {
//				System.out.println(" hello with ctype1="+ctype1);
				if (!ctype1.endsWith("--TAN")) {
					throw new IllegalArgumentException("ctype1 must end in --TAN but ctype1="+ctype1);
				}
			} else {
				ctype1 = "RA---TAN";
			}
			if ((ctype2 = header.getStringValue("CTYPE2")) != null) {
				if (!ctype2.endsWith("--TAN")) {
					throw new IllegalArgumentException("ctype2 must end with --TAN but ctype2="+ctype2);
				}
			} else {
				ctype2 = "DEC--TAN";
			}
	        this.crpix1 = header.getDoubleValue("CRPIX1", 0.0);
	        this.crpix2 = header.getDoubleValue("CRPIX2", 0.0);
	        this.crval1 = header.getDoubleValue("CRVAL1", 0.0);
	        this.crval2 = header.getDoubleValue("CRVAL2", 0.0);
	        if (HUtil.definedWithCdValues(header)) {
	        	// If one CD_i_j is specified, all other values default to 0.0, and ignore CDELTi and CROTAi
	        	sm1_1 = header.getDoubleValue("CD1_1", 0.0);
	        	sm2_1 = header.getDoubleValue("CD2_1", 0.0);
	        	sm1_2 = header.getDoubleValue("CD1_2", 0.0);
	        	sm2_2 = header.getDoubleValue("CD2_2", 0.0);
	        } else if (HUtil.definedWithPcValues(header)) {
	        	// If one of CDELTi or PCi_j are defined, use defaults that are the identity transform
	        	double cdelt1 = header.getDoubleValue("CDELT1", 1.0);
	        	double cdelt2 = header.getDoubleValue("CDELT2", cdelt1);
	        	sm1_1 = cdelt1*header.getDoubleValue("PC1_1", 1.0);
	        	sm2_1 = cdelt2*header.getDoubleValue("PC2_1", 0.0);
	        	sm1_2 = cdelt1*header.getDoubleValue("PC1_2", 0.0);
	        	sm2_2 = cdelt2*header.getDoubleValue("PC2_2", 1.0);
	        } 
	        else {
	        	// Otherwise, default to the identity transform, using scaling and rotation if they are given
	        	double crota = header.getDoubleValue("CROTA1", 0.0);
	        	crota = header.getDoubleValue("CROTA2", crota);
	           	double cosTheta = Math.cos(Math.toRadians(crota));
	           	double sinTheta = Math.sin(Math.toRadians(crota));
	        	double cdelt1 = header.getDoubleValue("CDELT1", 1.0);
	        	double cdelt2 = header.getDoubleValue("CDELT2", cdelt1);
	        	double cdelt = 0.5*(cdelt1+cdelt2);
	           	sm1_1 =  cosTheta*cdelt;
	           	sm2_1 =  sinTheta*cdelt;
	           	sm1_2 = -sinTheta*cdelt;
	           	sm2_2 =  cosTheta*cdelt;           	
	        }
	    	smDeterminant = sm1_1*sm2_2 - sm2_1*sm1_2;
	    	if (Math.abs(smDeterminant) < Double.MIN_VALUE) {
	    		throw new IllegalArgumentException("singular matrix:  smDeterminant="+smDeterminant);
	    	}
	    	invSm1_1 =  sm2_2/smDeterminant;
	    	/*
	    	 * incorrect
	    	invSm1_2 = -sm2_1/smDeterminant;
	    	invSm2_1 = -sm1_2/smDeterminant;
	    	*/
	    	invSm1_2 = -sm1_2/smDeterminant; // corrected August 8, 2007
	    	invSm2_1 = -sm2_1/smDeterminant; // corrected August 8, 2007
	    	invSm2_2 =  sm1_1/smDeterminant;	
	    	
	    	a7 = 0.25*(header.getDoubleValue("PV1_7", 0.0)+header.getDoubleValue("PV1_9", 0.0)+header.getDoubleValue("PV2_7", 0.0)+header.getDoubleValue("PV2_9", 0.0));
	    	if (a7 > a7Max) {
	    		throw new IllegalArgumentException("a7 must be less than a7Max="+a7Max+" but a7="+a7);
	    	}
	    	a7 = header.getDoubleValue("A7", a7);
		}

		// pixel --> sky
		/**
		 * @see gov.fnal.eag.wcs.Transform#forward(gov.fnal.eag.wcs.Coordinate)
		 */
		public SphericalCoordinate forward(Coordinate coordinate)  {
			double dx = coordinate.getX() - crpix1;
			double dy = coordinate.getY() - crpix2;
			double tx = sm1_1*dx + sm1_2*dy;
			dy = sm2_1*dx + sm2_2*dy;
			dx = tx;
			double rad2 = dx*dx + dy*dy; // fixed August 8, 2007.
			dx *=1.0 + a7*rad2;
			dy *= 1.0 + a7*rad2;
			double ra0 = Math.toRadians(crval1);
			double dec0 = Math.toRadians(crval2);
			
			double l = Math.toRadians(dx);
			double m = Math.toRadians(dy);
			double sins = l*l + m*m;
			if (sins > 1.0) {
				throw new Error("sins must be <= 1 but sins="+sins);
			}
			
			double cos0 = Math.cos(dec0);
			double sin0 = Math.sin(dec0);
			
			double dect = cos0 - m*sin0;
			if (Math.abs(dect) < dectTol) {
				throw new Error("abs(dect) must be >= dectTol="+dectTol+" but dect="+dect);
			}
			double rat = ra0 + Math.atan2(l,dect);
			dect = Math.atan(Math.cos(rat-ra0) * (m * cos0 + sin0) / dect);
			
			double raOut = rat;
			double decOut = dect;
			if (raOut < 0) raOut += 2*Math.PI;
			raOut = Math.toDegrees(raOut);
			decOut = Math.toDegrees(decOut);
			SphericalCoordinate sphericalCoordinate = new SphericalCoordinate(raOut, decOut);
			return sphericalCoordinate;
		}

		// sky --> pixel
		/**
		 * @see gov.fnal.eag.wcs.Transform#inverse(gov.fnal.eag.wcs.SphericalCoordinate)
		 */
		public Coordinate inverse(SphericalCoordinate sphericalCoordinate) {
			
			double ra = sphericalCoordinate.getLongitude();
			double dec = sphericalCoordinate.getLatitude();

			double ra0 = Math.toRadians(crval1); 
			double dec0 = Math.toRadians(crval2); 
			double dt = ra - crval1;

			if (dt > 180.) ra  -= 360;

			ra = Math.toRadians(ra);
			dec = Math.toRadians(dec);

			double coss = Math.cos(dec);
			double sins = Math.sin(dec);
			double l = Math.sin(ra-ra0) * coss;
			
			double sint = sins * Math.sin(dec0) + coss * Math.cos(dec0) * Math.cos(ra-ra0);
			if (sint <= 0.0) {
				throw new Error(" sint must be > 0 but sint="+sint);
			}
        
			double m = sins * Math.sin(dec0) + coss * Math.cos(dec0) * Math.cos(ra-ra0);
			l /= m;
			m = (sins * Math.cos(dec0) - coss * Math.sin(dec0) * Math.cos(ra-ra0)) / m;

			double dx = Math.toDegrees(l);
			double dy = Math.toDegrees(m);

			double rad2 = dx*dx + dy*dy;
			dx = dx * (1.-a7*rad2);
			dy = dy * (1.-a7*rad2);

			double tx = dx * invSm1_1 + dy * invSm1_2;
			dy = dx * invSm2_1 + dy * invSm2_2;
			dx = tx;

			Coordinate coordinate = new Coordinate(dx + crpix1, dy + crpix2);
			return coordinate;
		}


		/**
		 * @see gov.fnal.eag.wcs.Transform#toHeader(nom.tam.fits.Header)
		 */
		public void toHeader(Header header) {
			HUtil.eraseWcsInfo(header);
			HUtil.replaceKeywordValue(header, "CTYPE1", "RA---TAN", "Coordinate Type");
			HUtil.replaceKeywordValue(header, "CTYPE2", "DEC--TAN", "Coordinate Type");
			HUtil.replaceKeywordValue(header, "CRPIX1", crpix1, "Coordinate Reference Pixel");
			HUtil.replaceKeywordValue(header, "CRPIX2", crpix2, "Coordinate Reference Pixel");
			HUtil.replaceKeywordValue(header, "CRVAL1", crval1, "Coordinate Reference Value");
			HUtil.replaceKeywordValue(header, "CRVAL2", crval2, "Coordinate Reference Value");
			
			HUtil.replaceKeywordValue(header, "CD1_1", sm1_1, " Coordinate Matrix");
			HUtil.replaceKeywordValue(header, "CD1_2", sm1_2, " Coordinate Matrix");
			HUtil.replaceKeywordValue(header, "CD2_1", sm2_1, " Coordinate Matrix");
			HUtil.replaceKeywordValue(header, "CD2_2", sm2_2, " Coordinate Matrix");
			HUtil.replaceKeywordValue(header, "PV1_0", 0.0, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV1_1", 1.0, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV1_2", 0.0, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV1_3", 0.0, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV1_4", 0.0, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV1_5", 0.0, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV1_6", 0.0, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV1_7", a7, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV1_8", 0.0, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV1_9", a7, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV1_10",0.0 , "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV2_0", 0.0, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV2_1", 1.0, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV2_2", 0.0, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV2_3", 0.0, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV2_4", 0.0, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV2_5", 0.0, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV2_6", 0.0, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV2_7", a7, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV2_8", 0.0, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV2_9", a7, "PV Distortion Coefficient");
			HUtil.replaceKeywordValue(header, "PV2_10",0.0, "PV Distortion Coefficient");
		}
		
		/**
		 * @see java.lang.Object#toString()
		 */
		@Override
		public String toString() {
			String retval = "TransformTangentPlaneA7 with crpix1="+crpix1+" crpix2="+crpix2+" crval1="+crval1+" crval2="+crval2+" sm1_1="+sm1_1;
			return retval;
		}
		/**
		 * @see gov.fnal.eag.wcs.Transform#arcsecPerPixel()
		 */
		public double arcsecPerPixel() {
			double det = Math.sqrt(smDeterminant)*3600.0;	
			return det;
		}
		/**
		 * prints the values of this transform
		 * @param out where to print
		 */
		public void dump(PrintStream out) {
			out.println(String.format("crpix1 = %.14e",crpix1));
			out.println(String.format("crpix2 = %.14e",crpix2));
			out.println(String.format("crval1 = %.14e",crval1));
			out.println(String.format("crval2 = %.14e",crval2));
			out.println(String.format("cd11 = %.14e",sm1_1));
			out.println(String.format("cd12 = %.14e",sm1_2));
			out.println(String.format("cd21 = %.14e",sm2_1));
			out.println(String.format("cd22 = %.14e",sm2_2));
			out.println(String.format("a7 = %.14e",a7));			
		}

	}


