package snap.snoop.framework.visualize;

import hep.aida.IAnalysisFactory;
import hep.aida.ITree;
import hep.aida.ITuple;
import hep.aida.ITupleFactory;

import java.io.IOException;
import java.util.Vector;

import snap.snoop.framework.api.FrameworkException;
import snap.snoop.framework.api.LogUtil;
import snap.snoop.framework.api.Noun;
import snap.snoop.framework.api.SNOID;
import snap.snoop.framework.mdr.MDR;
import snap.snoop.framework.mdr.VersionInfo;

/**
 * DataTuple class extends from Noun and hence can be saved and retrieved from
 * datasets.
 * 
 * Note: In the current version it can hold vectors of native types double and
 * int
 *  
 */

public class DataTuple extends Noun
{

    /***************************************************************************
     * Data Members *
     **************************************************************************/

    /** Get a JAS3 Analysis factory reference. */
    //---------------------------------------------
    // Note: It is transient as it can not be saved
    //---------------------------------------------
    transient private IAnalysisFactory analysisFactory;

    /** Similarily generate a refernce to a JAS3 tree. */
    transient private ITree tree;

    /** Get a refernce for the JAS3 tuple factory. */
    transient private ITupleFactory tupleFactory;

    /** Get a reference for the JAS3 ITuple data structure. */
    transient private ITuple tuple;

    //------------------------------------------------------
    // Note: Following data members are saved with the noun.
    //------------------------------------------------------

    /** Default path for the ITuple. */
    protected String path;

    /** Default title for the ITuple. */
    protected String title;

    /** String array of column names in the tuple. */
    protected String[] columnNames;

    /** String array of column types in the tuple, saved as Strings. */
    protected String[] columnTypesString;

    /** Collection array of the data from all the columns of type int. */
    protected int[] intData = null;

    /** Collection array of the data from all the columns of type double. */
    protected double[] doubleData = null;

    /** Offset positions of the column positions. */
    protected int[] intArrayOfOffsets = null;

    /** Position counter for clumn of type int. */
    protected int intPos = 0;

    /** Position counter for clumn of type double. */
    protected int dblPos = 0;

    /** */
    protected int columnLength = 0;

    //------------------------------------------------------
    // Note: Following data members are NOT saved with the noun.
    //------------------------------------------------------
    transient protected Vector offset = new Vector();

    transient protected Class[] columnTypes;

    /** Is true if the first column is fake. */
    transient private boolean defaultColumn = false;

    /***************************************************************************
     * Static Initializer *
     **************************************************************************/

    /**
     * Every Noun should have a Static Initializer.
     */
    static
    {
        VersionInfo.addVersion(DataTuple.class);
        try
        {
            MDR.addNoun(DataTuple.class);
        }
        catch (FrameworkException mde)
        {
            String stack = LogUtil.createExceptionStackString(mde);
            log.severe("Unable to add DataTuple to MDR!\n\n" + stack);
        }
    }

    /***************************************************************************
     * Constructors *
     **************************************************************************/

    public DataTuple (String iName, String path, String title,
            String[] colNames, Class[] colTypes)
    {
        super(iName);

        analysisFactory = IAnalysisFactory.create();
        tree = analysisFactory.createTreeFactory().create();
        tupleFactory = analysisFactory.createTupleFactory(tree);
        tuple = tupleFactory.create(path, title, colNames, colTypes);

        this.path = path;
        this.title = title;

        //initialize arrays and then fill them

        columnNames = new String[colNames.length];
        columnTypes = new Class[colTypes.length];

        System.arraycopy(colNames, 0, this.columnNames, 0, colNames.length);
        System.arraycopy(colTypes, 0, this.columnTypes, 0, colTypes.length);

        columnTypesString = new String[colTypes.length];
        columnTypesString = getColumnTypesString();
    }

    public DataTuple (String iName, String path, String title)
    {
        super(iName);

        analysisFactory = IAnalysisFactory.create();
        tree = analysisFactory.createTreeFactory().create();
        tupleFactory = analysisFactory.createTupleFactory(tree);

        //Default tuple columns. These get filled by the addColumn method
        String columns = "int Index=0";
        tuple = tupleFactory.create(path, title, columns);

        this.path = path;
        this.title = title;

        defaultColumn = true;
    }

    /**
     * Every Noun must at least have a default constructor that takes no
     * parameters and has no side effects.
     */
    public DataTuple ()
    {
        super();
    }

    /**
     * It is nice to provide a noun that includes an iname. Often the
     * constructor that takes an iname will be combined with other parameters.
     * 
     * @param iName
     *            The iName for the new Noun
     */
    public DataTuple (String iName)
    {
        super(iName);
    }

    /**
     * If a Noun needs to be 'cloned' or copied, do not use the cloneable
     * interface. Instead create a copy constructor or a copy method. See
     * "Efficient Java" for details.
     * 
     * Here is a template for a copy constructor.
     * 
     * @param nounToClone
     *            The noun instance to clone
     */
    //	public
    //	DataTuple(DataTuple nounToClone)
    //	{
    //		super();
    //
    //			/************************
    //			 * User Code Goes Below *
    //			 ************************/
    //	}
    //	
    /**
     * It is very often useful to have an 'assign' method as well as a copy
     * constructor. This is directly analogous to an assignment operator in c++.
     * 
     * If it makes sense for this Noun to have an assignment operator un-comment
     * out the following code and implement it.
     */
    /*
     * public void assign(Noun rhsNoun) { }
     */

    /***************************************************************************
     * Public Methods *
     **************************************************************************/

    /*
     * Setters
     */

    public void save (String fileName)
    {
        try
        {
            ITree saveTree = analysisFactory.createTreeFactory().create(
                    fileName + ".aida", "xml", false, true);
            ITupleFactory saveTupleFactory = analysisFactory
                    .createTupleFactory(saveTree);

            /** create a new tuple from old tuple */
            ITuple saveTuple = saveTupleFactory.create(tuple.title(),
                    getPath(), tuple.columnNames(), tuple.columnTypes());

            //	fill data into the new data tuple
            String[] names = tuple.columnNames();
            Class[] types = tuple.columnTypes();

            tuple.start();
            while (tuple.next())
            {
                for (int loop = 0; loop < tuple.columns(); loop++)
                {
                    fillRow(loop, saveTuple);
                }
                saveTuple.addRow();
            }

            //save
            saveTree.commit();
        }
        catch (IOException e)
        {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    /**
     * @param column
     * @param value
     */
    public void fill (int column, Object[] value)
    {
        tuple.fill(column, value);
    }

    /**
     * @param columnName
     * @param value
     */
    public void fill (String columnName, Object[] value)
    {
        tuple.fill(getColumnIndex(columnName), value);
    }

    /**
     * @param column
     * @param value
     */
    public void fill (int column, double[] value)
    {
        /** create a new tuple from old tuple */
        ITuple newTuple = tupleFactory.create(tuple.title(), getPath(), tuple
                .columnNames(), tuple.columnTypes());

        /*
         * copy the old and the new data in the new tuple
         */

        // if the tuple is brand new, i.e., all the columns are empty
        // @TO DO/CAVEAT: For now will only work with the constructor
        // that takes columns and types implicitly. Need to correct this.
        if (tuple.rows() == 0)
        {
            int col = 0;

            for (int i = 0; i < value.length; i++)
            {
                newTuple.fill(col, value[i]);
                newTuple.addRow();
            }
        }
        // but if it has previous data then copy data from
        // old tuple to the new one...
        else
        {
            int i = 0;
            int col = 0;
            if (newTuple.columnName(0).compareTo("Index") == 0)
            {
                col++;
            }
            tuple.start();
            while (tuple.next())
            {
                for (int loop = col; loop < newTuple.columns(); loop++)
                {
                    if (loop == column)
                    {
                        newTuple.fill(loop, value[i]); // gets filled with new
                        // data
                    }
                    else
                    {
                        fillRow(loop, newTuple);
                    }
                }
                newTuple.addRow();
                i++;
            }
        }

        /*
         * connect the newTuple to the oldTuple
         */
        tuple = newTuple;
    }

    /**
     * @param columnName
     * @param value
     */
    public void fill (String columnName, double[] value)
    {
        fill(getColumnIndex(columnName), value);
    }

    /**
     * @param column
     * @param value
     */
    public void fill (int column, int[] value)
    {
        /*
         * create a new tuple from old tuple
         */
        ITuple newTuple = tupleFactory.create(tuple.title(), getPath(), tuple
                .columnNames(), tuple.columnTypes());

        /*
         * copy the old and the new data in the new tuple
         */

        // if the tuple is brand new, i.e., all the columns are empty
        // @TO DO/CAVEAT: For now will only work with the constructor
        // that takes columns and types implicitly. Need to correct this.
        if (tuple.rows() == 0)
        {
            int col = 0;

            for (int i = 0; i < value.length; i++)
            {
                newTuple.fill(col, value[i]);
                newTuple.addRow();
            }
        }
        // but if it has previous data then copy data from
        // old tuple to the new one...
        else
        {
            int i = 0;
            int col = 0;

            tuple.start();
            while (tuple.next())
            {
                for (int loop = col; loop < newTuple.columns(); loop++)
                {
                    if (loop == column)
                    {
                        newTuple.fill(loop, value[i]); // gets filled with new
                        // data
                    }
                    else
                    {
                        fillRow(loop, newTuple);
                    }
                }
                newTuple.addRow();
                i++;
            }
        }

        /*
         * connect the newTuple to the oldTuple
         */
        tuple = newTuple;
    }

    /**
     * @param columnName
     * @param value
     */
    public void fill (String columnName, int[] value)
    {
    		int i = getColumnIndex(columnName);
        fill(i, value);
    }

    /**
     * @param column
     * @param value
     */
    public void fill (int column, float[] value)
    {
        tuple.fill(column, value);
    }

    /**
     * @param columnName
     * @param value
     */
    public void fill (String columnName, float[] value)
    {
        tuple.fill(getColumnIndex(columnName), value);
    }

    /**
     * @param column
     * @param value
     */
    public void fill (int column, boolean[] value)
    {
        tuple.fill(column, value);
    }

    /**
     * @param columnName
     * @param value
     */
    public void fill (String columnName, boolean[] value)
    {
        tuple.fill(getColumnIndex(columnName), value);
    }

    /**
     * @param column
     * @param value
     */
    public void fill (int column, String[] value)
    {
        tuple.fill(column, value);
    }

    /**
     * @param columnName
     * @param value
     */
    public void fill (String columnName, String[] value)
    {
        tuple.fill(getColumnIndex(columnName), value);
    }

    /**
     * Copy data member tuple to the new tuple with added column and assign the
     * reference back to the original tuple. NOTE: Find out how to assign the
     * reference back to the old tuple while it may be in use.
     * 
     * @param columnName
     * @param columnType
     */
    public void addColumn (String columnName, Class columnType)
    {
        if (tuple != null)
        {
            //int size = tuple.columns ();
            String[] tmpColumnNames = tuple.columnNames();
            Class[] tmpColumnTypes = tuple.columnTypes();
            String[] newColumnNames = null;
            Class[] newColumnTypes = null;

            if (defaultColumn)
            {
                newColumnNames = new String[tmpColumnNames.length];
                newColumnTypes = new Class[tmpColumnTypes.length];

                newColumnNames[0] = columnName;
                newColumnTypes[0] = columnType;

                defaultColumn = false;
            }
            else
            {
                newColumnNames = new String[tmpColumnNames.length + 1];
                newColumnTypes = new Class[tmpColumnTypes.length + 1];

                System.arraycopy(tmpColumnNames, 0, newColumnNames, 0,
                        tmpColumnNames.length);
                newColumnNames[newColumnNames.length - 1] = columnName;

                System.arraycopy(tmpColumnTypes, 0, newColumnTypes, 0,
                        tmpColumnTypes.length);
                newColumnTypes[newColumnTypes.length - 1] = columnType;
            }

            /** Initialize arrays. */
            if (columnName == null) columnNames = new String[newColumnNames.length];
            if (columnTypes == null) columnTypes = new Class[newColumnTypes.length];
            if (columnTypesString == null) columnTypesString = new String[columnTypes.length];

            columnNames = newColumnNames;
            columnTypes = newColumnTypes;
            columnTypesString = getColumnTypesString();

            analysisFactory = IAnalysisFactory.create();
            tree = analysisFactory.createTreeFactory().create();
            tupleFactory = analysisFactory.createTupleFactory(tree);
            ITuple newTuple = tupleFactory.create(getPath(), tuple.title(),
                    newColumnNames, newColumnTypes);

            // copy data from old tuple to the new one...

            // copy data over to the new tuple ONLY if there is any to copy
            if (tuple.rows() > 0)
            {
                tuple.start();
                int col = 0;

                while (tuple.next())
                {
                    for (int loop = col; loop < newTuple.columns(); loop++)
                    {
                        if (loop == newTuple.findColumn(columnName))
                        {
                            fillRowWithDefault(loop, columnName, newTuple);
                        }
                        else
                        {
                            fillRow(loop, newTuple);
                        }
                    }
                    newTuple.addRow();
                }
            }
            tuple = newTuple;
        }
    }

    /**
     * @param col
     * @param newTuple
     */
    private void fillRow (int col, ITuple newTuple)
    {
        Class type = newTuple.columnType(col);

        if (type.getName().compareTo("double") == 0)
        {
            newTuple.fill(col, tuple.getDouble(col));
        }
        else if (type.getName().compareTo("int") == 0)
        {
            newTuple.fill(col, tuple.getInt(col));
        }
    }

    /**
     * @param columnName
     * @param newTuple
     */
    private void fillRowWithDefault (int col, String columnName, ITuple newTuple)
    {
        Class type = newTuple.columnType(newTuple.findColumn(columnName));

        if (type.getName().compareTo("double") == 0)
        {
            newTuple.fill(col, 0.0);
        }
        else if (type.getName().compareTo("int") == 0)
        {
            newTuple.fill(col, 0);
        }
    }

    /*
     * Getters
     */
    /**
     * @return
     */
    public ITuple getTuple ()
    {
        return tuple;
    }

    /**
     * @param columnName
     * @return
     */
    public Object[] getColumnAsObject (String columnName)
    {
        int i = 0;
        Object column[] = new Object[tuple.rows()];

        tuple.start();
        int columnIndex = getColumnIndex(columnName);
        while (tuple.next())
        {
            column[i] = tuple.getObject(columnIndex);
            i++;
        }

        return column;
    }

    /**
     * @param columnName
     * @return
     */
    public double[] getColumnAsDouble (String columnName)
    {
        int i = 0;

        int columnIndex = tuple.findColumn(columnName);
        double column[] = new double[getTupleSize()];

        tuple.start();
        while (tuple.next())
        {
            column[i] = tuple.getDouble(columnIndex);
            i++;
        }

        return column;
    }

    /**
     * @return
     */
    int getTupleSize ()
    {
        // TODO Auto-generated method stub
        int i = 0;
        tuple.start();
        while (tuple.next())
        {
            i++;
        }
        return i;
    }

    /**
     * @param columnName
     * @return
     */
    public int[] getColumnAsInteger (String columnName)
    {
        int i = 0;
        int column[] = new int[tuple.rows()];
        int columnIndex = getColumnIndex(columnName);

        tuple.start();
        while (tuple.next())
        {
            column[i] = tuple.getInt(columnIndex);
            i++;
        }

        return column;
    }

    /**
     * @param columnName
     * @return
     */
    public float[] getColumnAsFloat (String columnName)
    {
        int i = 0;
        float column[] = new float[tuple.rows()];

        tuple.start();
        while (tuple.next())
        {
            column[i] = tuple.getFloat(getColumnIndex(columnName));
            i++;
        }

        return column;
    }

    /**
     * @param columnName
     * @return
     */
    public boolean[] getColumnAsBoolean (String columnName)
    {
        int i = 0;
        boolean column[] = new boolean[tuple.rows()];

        tuple.start();
        while (tuple.next())
        {
            column[i] = tuple.getBoolean(getColumnIndex(columnName));
            i++;
        }

        return column;
    }

    /**
     * @param columnName
     * @return
     */
    public String[] getColumnAsString (String columnName)
    {
        int i = 0;
        String column[] = new String[tuple.rows()];

        tuple.start();
        while (tuple.next())
        {
            column[i] = tuple.getString(getColumnIndex(columnName));
            i++;
        }

        return column;
    }

    /*
     * Utility functions
     */

    /**
     *  
     */
    public void start ()
    {
        tuple.start();
    }

    /**
     * @return
     */
    public boolean next ()
    {
        return tuple.next();
    }

    //    public void write(String fileName)
    //    {
    //        try
    //        {
    //            //ITree t =
    // analysisFactory.createTreeFactory().create(fileName,"xml",false,true);
    //            IAnalysisFactory wAnalysisFactory = IAnalysisFactory.create();
    //            ITree wTree =
    // wAnalysisFactory.createTreeFactory().create(fileName,"xml",false,true);
    //            ITupleFactory wTupleFactory = wAnalysisFactory.createTupleFactory(wTree);
    //            ITuple wTuple = wTupleFactory.create (getPath(), getTitle(),
    // getColumnNames(), getColumnTypes());
    //            wTuple = this.tuple;
    //            wTree.commit();
    //        }
    //        catch (IllegalArgumentException e)
    //        {
    //            // TODO Auto-generated catch block
    //            e.printStackTrace();
    //        } catch (IOException e)
    //        {
    //            // TODO Auto-generated catch block
    //            e.printStackTrace();
    //        }
    //    }

    //    public void writeASCII (String fileName, String delimiter)
    //    {
    //    }
    //
    //    public void print ()
    //    {
    //    		//Sent to std I/O
    //    }

    /***************************************************************************
     * Private Methods *
     **************************************************************************/

    private int getColumnIndex (String columnName)
    {
        //returns corresponding column
        return tuple.findColumn(columnName);
    }

    /**
     * @return Returns the path.
     * 
     * @uml.property name="path"
     */
    private String getPath ()
    {
        return path;
    }

    private ITuple copyTuple (ITuple newTuple, String columnName)
    {
        return null;
    }

    /**
     * @return
     */
    private double defaultValue (Class type)
    {
        // TODO Auto-generated method stub
        return 0.0;
    }

    /***************************************************************************
     * Noun Required Methods *
     **************************************************************************/

    /**
     * Get a short description of this Noun.
     * 
     * @return A short human readable description of the class.
     */
    public static String getClassDescription ()
    {
        return "Noun";
    }

    /**
     * Get the version of this Noun.
     * 
     * @return A string that uniquily defines the version of this Noun.
     */
    public static String getClassVersion ()
    {
        return "$Id: DataTuple.java,v 1.8 2005/10/05 18:47:03 hshukla Exp $";
    }

    /**
     * Get the ClassId of this Noun.
     * 
     * @return The ClassID for this Noun
     */
    public static SNOID getClassClassId ()
    {
        return new SNOID(0x13293be4df9000L);
    }

    /***************************************************************************
     * Methods Specific to this Noun *
     **************************************************************************/

    /**
     * Methods that are tightly coupled to this Noun should be defined below.
     * Other more complicated operations should be added as Verbs. It is always
     * permissible to use a Verb instead of a method as long as the Verb does
     * need access to unexposed private data. Always use a Verb : -> If the
     * algorithm is likely to be changed over time. -> If the algorithm is going
     * to be written by a different person than the Noun implementor. -> If the
     * algorithm is complex and uses many types of Nouns. -> If different
     * problems will need different variations of the algorithm
     * 
     * The key difference is that Nouns should be implemented AND maintained by
     * one person and be stable over time so that everyone in the collaberation
     * can rely on them. Verbs are decoupled from Nouns so they can be written
     * by many people and changed often without effecting other collaborators.
     */

    /**
     * It is often useful to provide a custom toString method. Every Noun
     * inherits a toString method that returns the Noun type and the iName. The
     * toString method is by java custom a simple string only identifying the
     * instance. It should usually be less than 80 characters and not include a
     * terminating linefeed.
     */
    /*
     * public String toString() { return "short string identity"; }
     */

    /**
     * Every Noun inherits a toDebugString that returns the Noun Type, iName and
     * value of most of the java built in types. Over time the default
     * toDebugString will be enhanced to output useful information for more
     * types. Using toDebugString is a very useful development technique.
     * 
     * It is often useful to have provide a custom toDebugString. A custom
     * toDebugString will understand the type of the Noun and output useful
     * information for that particular noun. For example, the ParameterVector
     * toDebugString might output all the parameters and their parameters in a
     * useful format.
     */
    /*
     * public String toDebugString() { return "debug string"; }
     */

    /**
     * If you want to be able to compare Nouns to each other, then you _must_
     * override the equals method. In java the == operator compares to see if
     * the instances are the exact same object, not if they contain logically
     * the same values.
     * 
     * If you override equals, you usually need to override hashCode as well.
     * See hashCode for more information.
     */
    /*
     * public boolean equals(Noun rhs) { return true if the values are equal; }
     */

    /**
     * If you override equals or want to optimize the storage of this particular
     * Noun in hash tables you should override the hashCode method.
     * 
     * The hashCode method is a 16 bit hash of the contents of the instance.
     * There is a contract in java that all objects that return true for
     * equals() should have the same hashCode. You need to insure that this is
     * the case.
     */
    /*
     * public int hashCode() { return hash; }
     */

    /**
     * @return Returns the columnNames.
     */
    public String[] getColumnNames ()
    {
        return columnNames;
    }

    /**
     * @param columnNames
     *            The columnNames to set.
     */
    public void setColumnNames (String[] columnNames)
    {
        this.columnNames = columnNames;
    }

    /**
     * @return Returns the columnTypes.
     */
    public Class[] getColumnTypes ()
    {
        return columnTypes;
    }

    /**
     * @param columnTypes
     *            The columnTypes to set.
     */
    public void setColumnTypes (Class[] columnTypes)
    {
        this.columnTypes = columnTypes;
    }

    /**
     * @return Returns the title.
     */
    public String getTitle ()
    {
        return title;
    }

    /**
     * @param title
     *            The title to set.
     */
    public void setTitle (String title)
    {
        this.title = title;
    }

    /**
     * @param path
     *            The path to set.
     */
    public void setPath (String path)
    {
        this.path = path;
    }

    /***************************************************************************
     * Methods called by Framework to save/retrieve data *
     **************************************************************************/
    /**
     * Saves data
     */
    public boolean beforeSave (Noun containingNoun) throws FrameworkException
    {
        super.beforeSave(containingNoun);

        //get all the tuple data to be wrapped all other information from the
        //tuple is saved as it is not transient.

        //Save column type as Strings since Framework doesn't suport
        //Class type.

        int[] offsetAsInt = null;
        if (offset.size() == 0)
        {
            offsetAsInt = new int[offset.size()];
        }

        //fill the offsetAsInt array
        for (int i = 0; i < offset.size(); i++)
        {
            Integer o = (Integer) offset.get(i);
            offsetAsInt[i] = o.intValue();
        }

        //Collect data in contiguous arrays of specific type
        intData = null;
        doubleData = null;

        //the types supported are all primitive types ???
        if (columnTypes != null)
        {
            for (int i = 0; i < columnTypes.length; i++)
            {
                if (columnTypes[i].getName().compareTo("int") == 0)
                {
                    int[] tmp = getColumnAsInteger(getTuple().columnName(i));
                    fillIntData(tmp);
                }
                else if (columnTypes[i].getName().compareTo("double") == 0)
                {
                    double[] tmp = getColumnAsDouble(getTuple().columnName(i));
                    fillDoubleData(tmp);
                }
                else
                {
                    return false;
                }
            }
        }

        fillOffsetArray();
        return true;
    }

    /**
     *  
     */
    private void fillOffsetArray ()
    {
        intArrayOfOffsets = new int[offset.size()];

        for (int i = 0; i < offset.size(); i++)
        {
            intArrayOfOffsets[i] = ((Integer) offset.get(i)).intValue();
        }
    }

    /**
     * Retrieves data
     */
    public void afterLoad (Noun containingNoun) throws FrameworkException
    {
        super.afterLoad(containingNoun);

//        if (columnNames == null || columnTypesString == null || intData == null
//                || doubleData == null) throw new FrameworkException(
//                "Attempting to unwrap DataTuple that has not been saved correctly!");
        
        if (columnNames == null || columnTypesString == null) 
        		throw new FrameworkException(
                "Attempting to unwrap DataTuple that has not been saved correctly!");
        //Create the tuple
        analysisFactory = IAnalysisFactory.create();
        tree = analysisFactory.createTreeFactory().create();
        tupleFactory = analysisFactory.createTupleFactory(tree);
        columnTypes = getColumnTypesClass(columnTypesString);
        //Default tuple columns. These get filled by the addColumn method
        String columns = "int Index=0";
        ITuple newTuple = tupleFactory.create(path, title, columnNames,
                columnTypes);
        tuple = newTuple;

        //Fill the respective Columns
        int nextOffset = 0;
        for (int i = 0; i < columnTypesString.length; i++)
        {
            if (columnTypesString[i].compareTo("int") == 0)
            {
                //                Integer tmp = (Integer) offset.get(i);
                //                int offsetValue = tmp.intValue();
                fill(i, getIntData(intArrayOfOffsets[i]));
            }
            else if (columnTypesString[i].compareTo("double") == 0)
            {
                //                Integer tmp = (Integer) offset.get(i);
                //                int offsetValue = tmp.intValue();
                fill(i, getDoubleData(intArrayOfOffsets[i]));
            }
        }

        //--unwrap--
    }

    /**
     * Frees data
     */
    public void afterSave (Noun containingNoun) throws FrameworkException
    {
        super.afterSave(containingNoun);

        columnNames = null;
        columnTypesString = null;
        intData = null;
        doubleData = null;
        offset = null;

        //--free--
    }

    /***************************************************************************
     * Private methods
     **************************************************************************/

    /**
     * @param tmp
     */
    private void fillIntData (int[] tmp)
    {
        if (intData == null)
        {
            offset.add(new Integer(0));
            columnLength = tmp.length;

            //overwrite
            int[] tmpArray = new int[tmp.length];
            System.arraycopy(tmp, 0, tmpArray, 0, tmp.length);
            intData = tmpArray;

            dblPos = dblPos + tmp.length;
        }
        else
        {
            offset.add(new Integer(dblPos));

            //append
            int[] tmpIntData = new int[intData.length + tmp.length];
            System.arraycopy(intData, 0, tmpIntData, 0, intData.length);
            System.arraycopy(tmp, 0, tmpIntData, intData.length, tmp.length);
            intData = tmpIntData;

            dblPos = dblPos + tmp.length;
        }
    }

    /**
     * @param tmp
     */
    private void fillDoubleData (double[] tmp)
    {
        if (doubleData == null)
        {
            offset.add(new Integer(0));
            columnLength = tmp.length;

            //overwrite
            double[] tmpArray = new double[tmp.length];
            System.arraycopy(tmp, 0, tmpArray, 0, tmp.length);
            doubleData = tmpArray;

            intPos = intPos + tmp.length;
        }
        else
        {
            offset.add(new Integer(intPos));

            //append
            double[] tmpDoubleData = new double[doubleData.length + tmp.length];
            System
                    .arraycopy(doubleData, 0, tmpDoubleData, 0,
                            doubleData.length);
            System.arraycopy(tmp, 0, tmpDoubleData, doubleData.length,
                    tmp.length);
            doubleData = tmpDoubleData;

            intPos = intPos + tmp.length;
        }
    }

    /**
     * @return
     */
    private double[] getDoubleData (int position)
    {
        double[] tmp = new double[columnLength];
        System.arraycopy(doubleData, position, tmp, 0, columnLength);

        return tmp;
    }

    /**
     * @return
     */
    private int[] getIntData (int position)
    {
        int[] tmp = new int[columnLength];
        System.arraycopy(intData, position, tmp, 0, columnLength);

        return tmp;
    }

    /**
     * @return Returns the columnTypesString.
     */
    private String[] getColumnTypesString ()
    {
        if (columnTypesString.length != columnTypes.length)
        {
            String[] newArray = new String[columnTypes.length];
            columnTypesString = newArray;
        }

        for (int i = 0; i < columnTypes.length; i++)
        {
            //            String tmp = columnTypes [i].getName();
            columnTypesString[i] = columnTypes[i].getName();
        }
        return columnTypesString;
    }

    /**
     * @return Returns the columnTypesString.
     */
    private Class[] getColumnTypesClass (String[] columnString)
    {
        Class[] columnTypesClass = new Class[columnString.length];

        for (int i = 0; i < columnString.length; i++)
        {
            if (columnString[i].compareTo("double") == 0)
            {
                columnTypesClass[i] = double.class;
            }
            else if (columnString[i].compareTo("int") == 0)
            {
                columnTypesClass[i] = int.class;
            }
        }
        return columnTypesClass;
    }
}