package snap.snoop.framework.visualize;

import java.awt.Component;
import java.awt.Container;
import java.util.Arrays;

import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.WindowConstants;

import snap.snoop.framework.api.SNAPRuntimeException;

import hep.aida.IAnalysisFactory;
import hep.aida.IBaseHistogram;
import hep.aida.ICloud1D;
import hep.aida.ICloud2D;
import hep.aida.IDataPointSet;
import hep.aida.IDataPointSetFactory;
import hep.aida.IHistogram1D;
import hep.aida.IHistogram2D;
import hep.aida.IHistogramFactory;
import hep.aida.IPlotter;
import hep.aida.IPlotterFactory;
import hep.aida.IPlotterStyle;
import hep.aida.ITree;
import hep.aida.ref.plotter.PlotterUtilities;

/**
 * Visualize is the base class extending <a
 * href="http://aida.freehep.org/doc/v3.2.1/api/index.html">AIDA </a> graphics
 * contexts which allow SNAP simulation framework to visually realize data
 * during development and debugging states. Data structures of both AIDA and
 * Java primitive types are plotted via various static plotting routines. A
 * Visualize object encapsulates the state information needed for the various
 * rendering operations. This state information includes:
 * <ul>
 * <li>The canvas to draw on
 * <li>The interactive interface. (Mouse Right Click event displays the
 * available properties)
 * <li>The plotter reference for advanced interaction
 * <li>Multiple plotting regions
 * <li>Plot overlays
 * <li>Printing, saving (standard image formats e.g. JPEG, GIF, TIFF etc.)
 * </ul>
 * <p>
 * The Visualize object is initialized via static initializer. The static
 * methods curtail the overhead of object creation while providing the
 * flexibilty of data visualization within the debug cycle. Through the Display
 * window of <a href="http://www.eclipse.org">Eclipse </a> (IDE) while in Debug
 * perspective one can plot data structures availble in the memory.
 * <p>
 * 
 * @author Hemant Shukla
 * @version 0.9 (Beta)
 * 
 */

public class Visualize
{
	private static IAnalysisFactory analysisFactory;

	private static ITree tree;

	/**
	 * By default the plots are turned ON. User can turn toggle using the
	 * display (boolean on) method.
	 */
	private static boolean display = true;

	/**
	 * In batchMode by default plots are OFF. For batchMode = true, member
	 * variable display is overidden and set to false.
	 */
	private static boolean batchMode = false;
	// private static int region = 0;

	/**
	 * Static Initializer.
	 */
	static
	{
		analysisFactory = IAnalysisFactory.create();
		tree = analysisFactory.createTreeFactory().create();

		// histPlotter = analysisFactory.createPlotterFactory().create("Plot");
		// Read the snoop.properties to check for batchmode boolean
	}

	/**
	 * 
	 * @param s
	 * @param buttonNames
	 * @return
	 */
	public static String plotDialog(SubPlot s, String[] buttonNames)
	{
		VisualizeFrame frame = new VisualizeFrame(s, buttonNames);
		frame.display();

		return frame.getButtonName();
	}

	public static void plot(SubPlot s)
	{
		JDialog jd = new JDialog();
		
		Container c = jd.getContentPane();
		
		c.add(PlotterUtilities.componentForPlotter(s.getPlotter()));
		
		jd.setTitle(s.getWindowTitle());
		
		jd.pack();
		
		jd.show();
	}

	public static void plot(IPlotter plotter, String title)
	{
		Component component = PlotterUtilities.componentForPlotter(plotter);

		JFrame frame = new JFrame(title);
		
		Container c = frame.getContentPane();
		
		c.add(component);
		
		frame.setTitle(title);
		
		frame.pack();
		
        frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

		frame.show();
	}
	
	public static void plot(IPlotter plotter)
	{
		plot(plotter, "");
	}

	/**
	 * @param y
	 *            The method plots an array of double. The x axis is determined
	 *            based upon the index.
	 */
	public static IPlotter plot(double[] y)
	{
		// Create data point set
		IDataPointSetFactory dps = analysisFactory
				.createDataPointSetFactory(tree);
		IDataPointSet dps1D = dps.create("dps1D", "<<Title>>", 1);

		// fill it
		for (int i = 0; i < y.length; i++)
		{
			dps1D.addPoint();
			dps1D.point(i).coordinate(0).setValue(y[i]);
		}

		// plot it
		return plotY(dps1D);
	}

	/**
	 * The method takes a DataTuple type and plots the single column based on
	 * the name provided by the string columnName. NOTE: This plots only double,
	 * int and float for now. Needs to exit elegantly.
	 * 
	 * @param tuple
	 *            of type DataTuple
	 * @param columnNameY
	 *            String name of the vector
	 * 
	 */
	public static IPlotter plot(DataTuple tuple, String columnNameY)
	{
		// Create data point set
		IDataPointSetFactory dps = analysisFactory
				.createDataPointSetFactory(tree);
		IDataPointSet dps1D = dps.create("dps1D", "<<Title>>", 1);

		Class type = tuple.getTuple().columnType(
				tuple.getTuple().findColumn(columnNameY));

		if (type.getName().compareTo("double") == 0)
		{
			double[] y = new double[tuple.getTupleSize()];
			y = tuple.getColumnAsDouble(columnNameY);

			// fill it
			for (int i = 0; i < y.length; i++)
			{
				dps1D.addPoint();
				dps1D.point(i).coordinate(0).setValue(y[i]);
			}
		}

		if (type.getName().compareTo("int") == 0)
		{
			int[] y = new int[tuple.getTupleSize()];
			y = tuple.getColumnAsInteger(columnNameY);

			// fill it
			for (int i = 0; i < y.length; i++)
			{
				dps1D.addPoint();
				dps1D.point(i).coordinate(0).setValue(y[i]);
			}
		}

		if (type.getName().compareTo("float") == 0)
		{
			float[] y = new float[tuple.getTupleSize()];
			y = tuple.getColumnAsFloat(columnNameY);

			// fill it
			for (int i = 0; i < y.length; i++)
			{
				dps1D.addPoint();
				dps1D.point(i).coordinate(0).setValue(y[i]);
			}
		}

		// plot it
		return plotY(dps1D);
	}

	/**
	 * @param x
	 * @param y
	 * @return
	 */
	public static IPlotter plot(double[] x, double[] y, double[] err)
	{
		// Create data point set
		IDataPointSetFactory dps = analysisFactory
				.createDataPointSetFactory(tree);
		IDataPointSet dps2D = dps.create("dps2D", "<<Title>>", 2);

		// check if x and y have same dimensions else
		// throw and exception...
		if (x.length != y.length)
			throw new SNAPRuntimeException("Dimensions not equal");

		// fill it
		for (int i = 0; i < y.length; i++)
		{
			dps2D.addPoint();
			dps2D.point(i).coordinate(0).setValue(x[i]);
			dps2D.point(i).coordinate(1).setValue(y[i]);
			dps2D.point(i).coordinate(1).setErrorMinus(err[i]);
			dps2D.point(i).coordinate(1).setErrorPlus(err[i]);
		}

		// plot it
		return plotXY(dps2D);
	}

	/**
	 * @param tuple
	 * @param columnNameX
	 * @param columnNameY
	 * @return
	 */
	public static IPlotter plot(DataTuple tuple, String columnNameX,
			String columnNameY)
	{
		// Create data point set
		IDataPointSetFactory dps = analysisFactory
				.createDataPointSetFactory(tree);
		IDataPointSet dps2D = dps.create("dps2D", "<<Title>>", 2);

		Class typeX = tuple.getTuple().columnType(
				tuple.getTuple().findColumn(columnNameX));
		Class typeY = tuple.getTuple().columnType(
				tuple.getTuple().findColumn(columnNameY));

		int size = tuple.getTupleSize();

		if (typeX.getName().compareTo("double") == 0)
		{
			double[] x = new double[size];
			x = tuple.getColumnAsDouble(columnNameX);

			// fill it
			for (int i = 0; i < x.length; i++)
			{
				dps2D.addPoint();
				dps2D.point(i).coordinate(0).setValue(x[i]);
			}
		}

		if (typeX.getName().compareTo("int") == 0)
		{
			int[] x = new int[size];
			x = tuple.getColumnAsInteger(columnNameX);

			// fill it
			for (int i = 0; i < x.length; i++)
			{
				dps2D.addPoint();
				dps2D.point(i).coordinate(0).setValue(x[i]);
			}
		}

		if (typeY.getName().compareTo("double") == 0)
		{
			double[] y = new double[size];
			y = tuple.getColumnAsDouble(columnNameY);

			// fill it
			for (int i = 0; i < y.length; i++)
			{
				dps2D.addPoint();
				dps2D.point(i).coordinate(1).setValue(y[i]);
			}
		}

		if (typeY.getName().compareTo("int") == 0)
		{
			int[] y = new int[size];
			y = tuple.getColumnAsInteger(columnNameY);

			// fill it
			for (int i = 0; i < y.length; i++)
			{
				dps2D.addPoint();
				dps2D.point(i).coordinate(1).setValue(y[i]);
			}
		}

		// plot it
		return plotXY(dps2D);
	}

	/***************************************************************************
	 * Plot Histogram
	 **************************************************************************/

	/**
	 * @param y
	 * @param bin
	 * @return
	 */
	public static IPlotter hist(double[] y, double binWidth, double minXAxis,
			double maxXAxis)
	{
		IHistogramFactory hf = analysisFactory.createHistogramFactory(tree);

		int bin = (int) ((maxXAxis - minXAxis) / binWidth);

		IHistogram1D h = hf.createHistogram1D(" ", "", bin, minXAxis, maxXAxis);

		for (int i = 0; i < y.length; i++)
		{
			h.fill(y[i]);
		}

		return plotHist(h);
	}

	/**
	 * @param y
	 * @param bin
	 * @return
	 */
	public static IPlotter hist(double[] y, int bin)
	{
		IHistogramFactory hf = analysisFactory.createHistogramFactory(tree);
		IHistogram1D h = hf.createHistogram1D(" ", "", bin, min(y)
				+ (min(y) * 0.2), max(y) + (max(y) * 0.2));

		for (int i = 0; i < y.length; i++)
		{
			h.fill(y[i]);
		}

		return plotHist(h);
	}

	/**
	 * @param tuple
	 * @param columnNameY
	 * @param bin
	 * @return
	 */
	public static IPlotter hist(DataTuple tuple, String columnNameY, int bin)
	{
		IHistogramFactory hf = analysisFactory.createHistogramFactory(tree);

		Class type = tuple.getTuple().columnType(
				tuple.getTuple().findColumn(columnNameY));
		IHistogram1D h = null;

		if (type.getName().compareTo("double") == 0)
		{
			double[] y = new double[tuple.getTupleSize()];
			y = tuple.getColumnAsDouble(columnNameY);
			h = hf.createHistogram1D(" ", "", bin, min(y) + (min(y) * 0.2),
					max(y) + (max(y) * 0.2));

			// fill it
			for (int i = 0; i < y.length; i++)
			{
				h.fill(y[i]);
			}
		}

		if (type.getName().compareTo("int") == 0)
		{
			int[] y = new int[tuple.getTupleSize()];
			y = tuple.getColumnAsInteger(columnNameY);

			h = hf.createHistogram1D(" ", "", bin, min(y) + (min(y) * 0.2),
					max(y) + (max(y) * 0.2));

			// fill it
			for (int i = 0; i < y.length; i++)
			{
				h.fill(y[i]);
			}
		}

		if (type.getName().compareTo("float") == 0)
		{
			float[] y = new float[tuple.getTupleSize()];
			y = tuple.getColumnAsFloat(columnNameY);

			h = hf.createHistogram1D(" ", "", bin, min(y) + (min(y) * 0.2),
					max(y) + (max(y) * 0.2));

			// fill it
			for (int i = 0; i < y.length; i++)
			{
				h.fill(y[i]);
			}
		}

		return plotHist(h);
	}

	/**
	 * 
	 * @param x
	 * @param y
	 * @return
	 */
	public static IPlotter plot(double[] x, double[] y)
	{
		// Create data point set
		IDataPointSetFactory dps = analysisFactory
				.createDataPointSetFactory(tree);
		IDataPointSet dps2D = dps.create("dps2D", "<<Title>>", 2);

		// check if x and y have same dimensions else
		// throw and exception...
		if (x.length != y.length)
			throw new SNAPRuntimeException("Dimensions not equal");

		// fill it
		for (int i = 0; i < y.length; i++)
		{
			dps2D.addPoint();
			dps2D.point(i).coordinate(0).setValue(x[i]);
			dps2D.point(i).coordinate(1).setValue(y[i]);
		}

		// plot it
		return plotXY(dps2D);
	}

	/**
	 * @param y
	 * @return
	 */
	public static IPlotter hist(double[] y)
	{
		double offset = autoScaleHistAxis(y);
		IHistogramFactory hf = analysisFactory.createHistogramFactory(tree);
		IHistogram1D h = hf.createHistogram1D(" ", "", 50, min(y) - offset,
				max(y) + offset);

		for (int i = 0; i < y.length; i++)
		{
			h.fill(y[i]);
		}

		return plotHist(h);
	}

	private static double autoScaleHistAxis(double[] y)
	{
		double value = 0.0;
		
		// check if the data array y is filled with zeroes then
		// return an offset that is small but non-zero
		if (isClear (y)) return 0.5;
		
		double len = Math.abs(max(y) - min(y));
		value = len * 0.15;
		return Math.abs(value);
	}

	private static boolean isClear (double [] array)
	{
		boolean equal = false;
		double [] test = new double [array.length];
		if (Arrays.equals (test, array))
			equal = true;
		return equal;
	}
	
	/**
	 * @param x
	 * @param y
	 * @return
	 */
	public static IPlotter hist(double[] x, double[] y)
	{
		double xOffset = autoScaleHistAxis(x);
		double yOffset = autoScaleHistAxis(y);

		IHistogramFactory hf = analysisFactory.createHistogramFactory(tree);
		IHistogram2D h = hf.createHistogram2D(" ", "", 50, min(x) - xOffset,
				max(x) + xOffset, 50, min(y) - yOffset, max(y) + yOffset);

		for (int i = 0; i < y.length; i++)
		{
			h.fill(x[i], y[i]);
		}

		return plotHist(h);
	}

	/**
	 * @param tuple
	 * @param columnNameY
	 * @return
	 */
	public static IPlotter hist(DataTuple tuple, String columnNameY)
	{
		return hist(tuple, columnNameY, 50);
	}

	/**
	 * @param tuple
	 * @param columnNameX
	 * @param columnNameY
	 * @return
	 */
	public static IPlotter hist(DataTuple tuple, String columnNameX,
			String columnNameY)
	{
		return hist(tuple, columnNameX, columnNameX, 50, 50);
	}

	/**
	 * @param tuple
	 * @param columnNameX
	 * @param columnNameY
	 * @param binX
	 * @param binY
	 * @return
	 */
	public static IPlotter hist(DataTuple tuple, String columnNameX,
			String columnNameY, int binX, int binY)
	{
		// Create data point set
		IHistogramFactory hf = analysisFactory.createHistogramFactory(tree);
		IHistogram2D h = null;

		Class typeX = tuple.getTuple().columnType(
				tuple.getTuple().findColumn(columnNameX));
		Class typeY = tuple.getTuple().columnType(
				tuple.getTuple().findColumn(columnNameY));

		int size = tuple.getTupleSize();

		// Get Data in x and y vectors
		if (typeX.getName().compareTo("double") == 0)
		{
			double[] x = new double[size];
			x = tuple.getColumnAsDouble(columnNameX);

			if (typeY.getName().compareTo("double") == 0)
			{
				double[] y = new double[size];
				y = tuple.getColumnAsDouble(columnNameY);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}

			if (typeY.getName().compareTo("int") == 0)
			{
				int[] y = new int[size];
				y = tuple.getColumnAsInteger(columnNameY);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}

			if (typeY.getName().compareTo("float") == 0)
			{
				float[] y = new float[size];
				y = tuple.getColumnAsFloat(columnNameY);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}
		}

		if (typeX.getName().compareTo("int") == 0)
		{
			int[] x = new int[size];
			x = tuple.getColumnAsInteger(columnNameX);

			if (typeY.getName().compareTo("double") == 0)
			{
				double[] y = new double[size];
				y = tuple.getColumnAsDouble(columnNameY);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}

			if (typeY.getName().compareTo("int") == 0)
			{
				int[] y = new int[size];
				y = tuple.getColumnAsInteger(columnNameY);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}

			if (typeY.getName().compareTo("float") == 0)
			{
				float[] y = new float[size];
				y = tuple.getColumnAsFloat(columnNameY);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}
		}

		if (typeX.getName().compareTo("float") == 0)
		{
			float[] x = new float[size];
			x = tuple.getColumnAsFloat(columnNameX);

			if (typeY.getName().compareTo("double") == 0)
			{
				double[] y = new double[size];
				y = tuple.getColumnAsDouble(columnNameY);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}

			if (typeY.getName().compareTo("int") == 0)
			{
				int[] y = new int[size];
				y = tuple.getColumnAsInteger(columnNameY);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}

			if (typeY.getName().compareTo("float") == 0)
			{
				float[] y = new float[size];
				y = tuple.getColumnAsFloat(columnNameY);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}
		}

		if (typeY.getName().compareTo("double") == 0)
		{
			double[] y = new double[size];
			y = tuple.getColumnAsDouble(columnNameY);

			if (typeX.getName().compareTo("double") == 0)
			{
				double[] x = new double[size];
				x = tuple.getColumnAsDouble(columnNameX);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}

			if (typeX.getName().compareTo("int") == 0)
			{
				int[] x = new int[size];
				x = tuple.getColumnAsInteger(columnNameX);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}

			if (typeX.getName().compareTo("float") == 0)
			{
				float[] x = new float[size];
				x = tuple.getColumnAsFloat(columnNameX);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}
		}

		if (typeY.getName().compareTo("int") == 0)
		{
			int[] y = new int[size];
			y = tuple.getColumnAsInteger(columnNameY);

			if (typeX.getName().compareTo("double") == 0)
			{
				double[] x = new double[size];
				x = tuple.getColumnAsDouble(columnNameX);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}

			if (typeX.getName().compareTo("int") == 0)
			{
				int[] x = new int[size];
				x = tuple.getColumnAsInteger(columnNameX);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}

			if (typeX.getName().compareTo("float") == 0)
			{
				float[] x = new float[size];
				x = tuple.getColumnAsFloat(columnNameX);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}
		}

		if (typeY.getName().compareTo("float") == 0)
		{
			float[] y = new float[size];
			y = tuple.getColumnAsFloat(columnNameY);

			if (typeX.getName().compareTo("double") == 0)
			{
				double[] x = new double[size];
				x = tuple.getColumnAsDouble(columnNameX);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}

			if (typeX.getName().compareTo("int") == 0)
			{
				int[] x = new int[size];
				x = tuple.getColumnAsInteger(columnNameX);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}

			if (typeX.getName().compareTo("float") == 0)
			{
				float[] x = new float[size];
				x = tuple.getColumnAsFloat(columnNameX);

				// Fill data in the 2-D Histogtram AIDA object
				h = hf.createHistogram2D(" ", "", binX,
						min(x) + (min(x) * 0.2), max(x) + (max(x) * 0.2), binY,
						min(y) + (min(y) * 0.2), max(y) + (max(y) * 0.2));

				for (int i = 0; i < y.length; i++)
				{
					h.fill(x[i], y[i]);
				}
			}
		}

		return plotHist(h);
	}

	/**
	 * @param histData
	 * 
	 * Use IHistogram for polymorphism. IHistogram subclasses IHistogram1D,
	 * IHistogram2D, IHistogram3D can all be passed as the argument and plotted.
	 */
	// public static void hist (DataHistogram histData)
	// {
	// }
	/***************************************************************************
	 * Plot Scatter
	 **************************************************************************/

	/**
	 * @param y
	 * @return
	 */
	public static IPlotter scatter(double[] y)
	{
		IHistogramFactory hf = analysisFactory.createHistogramFactory(tree);
		ICloud1D c = hf.createCloud1D(" ");

		for (int i = 0; i < y.length; i++)
		{
			c.fill(y[i]);
		}

		return plotScatter(c);
	}

	/**
	 * @param x
	 * @param y
	 * @return
	 */
	public static IPlotter scatter(double[] x, double[] y)
	{
		IHistogramFactory hf = analysisFactory.createHistogramFactory(tree);
		ICloud2D c = hf.createCloud2D(" ");

		for (int i = 0; i < y.length; i++)
		{
			c.fill(x[i], y[i]);
		}

		return plotScatter(c);
	}

	/**
	 * COMING SOON!!! (May 10, 2005) Following methods with enhanced features
	 */

	/***************************************************************************
	 * public static void scatter (DataTuple tuple, String columnNameX, String
	 * columnNameY) { } // Utitlities, styles etc. for data manipulation
	 * 
	 * public static void title (String title) { }
	 * 
	 * public static void xLabel (String label) { }
	 * 
	 * public static void yLabel (String label) { }
	 * 
	 * public static void show (boolean on) { display = on; }
	 * 
	 * public static void xFont (String size, String style, String color) {}
	 * public static void yFont (String size, String style, String color) {}
	 * public static void titleFont (String size, String style, String color) {}
	 * public static void xRange (double min, double max) {} public static void
	 * yRange (double min, double max) {} public static void marker (String
	 * size, String shape, String color) {}
	 * 
	 * 
	 * //Filter provides for cuts and ranges for data view
	 * 
	 * public static void filter (String filt) {}
	 * 
	 * //Evaluater provides expressions
	 * 
	 * public static void evaluater (String eval) {}
	 * 
	 * 
	 * //Create n x m reagions within a plot public static void setRegions (int
	 * n, int m) {}
	 * 
	 **************************************************************************/

	// public static void setPlotRegion (int plotRegion)
	// {
	// region = plotRegion;
	// }
	//	
	/**
	 * Private methods for plotting
	 */

	private static IPlotter plotY(IDataPointSet dps1D)
	{
		IPlotterFactory plotterFactory = analysisFactory.createPlotterFactory();
		IPlotter plotter = plotterFactory.create();

		IPlotterStyle regionStyle = plotter.region(0).style();

		setPlotStyle(regionStyle);

		plotter.region(0).plot(dps1D);

		if (display)
		{
			plotter.show();
		}

		return plotter;
	}

	private static IPlotter plotY(int plotRegion, IDataPointSet dps1D)
	{
		IPlotterFactory plotterFactory = analysisFactory.createPlotterFactory();
		IPlotter plotter = plotterFactory.create();
		plotter.createRegions(2, 2);

		IPlotterStyle regionStyle = plotter.region(plotRegion).style();

		setPlotStyle(regionStyle);

		plotter.region(plotRegion).plot(dps1D);

		if (display)
		{
			plotter.show();
		}

		return plotter;
	}

	private static IPlotter plotXY(IDataPointSet dps2D)
	{
		IPlotterFactory plotterFactory = analysisFactory.createPlotterFactory();
		IPlotter plotter = plotterFactory.create();

		IPlotterStyle regionStyle = plotter.region(0).style();

		setPlotStyle(regionStyle);

		plotter.region(0).plot(dps2D);
		if (display)
		{
			plotter.show();
		}

		return plotter;
	}

	/**
	 * @param regionStyle
	 */
	private static void setPlotStyle(IPlotterStyle regionStyle)
	{
//		regionStyle.dataStyle().setParameter("showDataPoints", "false");
//		regionStyle.dataStyle().setParameter("connectDataPoints", "true");
		regionStyle.dataStyle().markerStyle().setVisible(true);
		regionStyle.dataStyle().outlineStyle().setVisible(false);

		/**
		 * Set fonts
		 */
		String font = "Gill Sans";

		// Title
		regionStyle.titleStyle().textStyle().setFont(font);
		regionStyle.titleStyle().textStyle().setFontSize(14);

		// X Axis
		regionStyle.xAxisStyle().labelStyle().setFont(font);
		regionStyle.xAxisStyle().labelStyle().setFontSize(12);

		regionStyle.xAxisStyle().tickLabelStyle().setFont(font);
		regionStyle.xAxisStyle().tickLabelStyle().setFontSize(12);

		// Y Axis
		regionStyle.yAxisStyle().labelStyle().setFont(font);
		regionStyle.yAxisStyle().labelStyle().setFontSize(12);

		regionStyle.yAxisStyle().tickLabelStyle().setFont(font);
		regionStyle.yAxisStyle().tickLabelStyle().setFontSize(12);
	}

	private static IPlotter plotScatter(ICloud1D cl1D)
	{
		IPlotterFactory plotterFactory = analysisFactory.createPlotterFactory();
		IPlotter cloudPlotter = plotterFactory.create();

		setPlotStyle(cloudPlotter.region(0).style());

		cloudPlotter.region(0).plot(cl1D);
		if (display)
		{
			cloudPlotter.show();
		}

		return cloudPlotter;
	}

	private static IPlotter plotScatter(ICloud2D cl2D)
	{
		IPlotterFactory plotterFactory = analysisFactory.createPlotterFactory();
		IPlotter cloudPlotter = plotterFactory.create();

		setPlotStyle(cloudPlotter.region(0).style());

		cloudPlotter.region(0).plot(cl2D);
		if (display)
		{
			cloudPlotter.show();
		}

		return cloudPlotter;
	}

	/*
	 * public static void setLabel () { IPlotterStyle regionStyle =
	 * histPlotter.region(0).style();
	 * regionStyle.dataStyle().lineStyle().setParameter("color","black");
	 * regionStyle.xAxisStyle().setLabel("E/m");
	 * regionStyle.yAxisStyle().setLabel("# Evt");
	 * 
	 * if (display) { histPlotter.show(); } }
	 * 
	 * private static String title = "Default"; public static void setTitle
	 * (String t) { if (t != null) { title = t; } }
	 * 
	 * private static String getTitle () { return title; }
	 */
	// private static IPlotter histPlotter;
	/**
	 * NOTE: If we decide to have only one plotting window then perhaps this
	 * with become reduntant.
	 * 
	 * @return
	 * 
	 * @uml.property name="histPlotter"
	 */
	// public static IPlotter getHistPlotter() {
	// return histPlotter;
	// }
	private static IPlotter plotHist(IHistogram1D h)
	{
		IPlotterFactory plotterFactory = analysisFactory.createPlotterFactory();
		IPlotter histPlotter = plotterFactory.create();
		/*
		 * //This are the default styles for the region IPlotterStyle
		 * regionStyle = plotter.region(0).style();
		 * regionStyle.dataStyle().lineStyle().setParameter("color","black");
		 * regionStyle.xAxisStyle().setLabel("E/m");
		 * regionStyle.yAxisStyle().setLabel("# Evt");
		 * 
		 * regionStyle.titleStyle().textStyle().setFontSize(30);
		 * regionStyle.titleStyle().textStyle().setColor("orange");
		 * 
		 * regionStyle.xAxisStyle().labelStyle().setFontSize(24);
		 * regionStyle.xAxisStyle().labelStyle().setItalic(true);
		 * regionStyle.xAxisStyle().labelStyle().setColor("black");
		 * regionStyle.xAxisStyle().tickLabelStyle().setFontSize(14);
		 * regionStyle.xAxisStyle().tickLabelStyle().setBold(true);
		 * regionStyle.xAxisStyle().tickLabelStyle().setColor("blue");
		 * 
		 * regionStyle.yAxisStyle().labelStyle().setFontSize(24);
		 * regionStyle.yAxisStyle().labelStyle().setItalic(true);
		 * regionStyle.yAxisStyle().labelStyle().setColor("brown");
		 * regionStyle.yAxisStyle().tickLabelStyle().setFontSize(14);
		 * regionStyle.yAxisStyle().tickLabelStyle().setBold(true);
		 * regionStyle.yAxisStyle().tickLabelStyle().setColor("green");
		 * 
		 * //This styles overwrite some of the region styles IPlotterStyle style =
		 * plotterFactory.createPlotterStyle();
		 * style.dataStyle().markerStyle().setParameter("size","12");
		 * style.dataStyle().markerStyle().setParameter("shape","3");
		 * style.dataStyle().markerStyle().setParameter("color","blue");
		 */
		setPlotStyle(histPlotter.region(0).style());
		histPlotter.region(0).plot((IBaseHistogram) h);

		if (display)
		{
			histPlotter.show();
		}

		return histPlotter;
	}

	private static IPlotter plotHist(IHistogram2D h)
	{
		IPlotterFactory plotterFactory = analysisFactory.createPlotterFactory();
		IPlotter histPlotter = plotterFactory.create();

		setPlotStyle(histPlotter.region(0).style());
		histPlotter.region(0).plot((IBaseHistogram) h);

		if (display)
		{
			histPlotter.show();
		}

		return histPlotter;
	}

	/**
	 * @return
	 */
	/*
	 * public static IPlotterStyle getStyle () { IPlotterFactory
	 * newPlotterFactory = analysisFactory .createPlotterFactory();
	 * IPlotterStyle style = newPlotterFactory.createPlotterStyle(); return
	 * style; }
	 * 
	 * public static IPlotter plot (double[] x, double[] y, double[] err,
	 * IPlotterStyle style) { //Create data point set IDataPointSetFactory dps =
	 * analysisFactory .createDataPointSetFactory(tree); IDataPointSet dps2D =
	 * dps.create("dps2D", "<<Title>>", 2);
	 * 
	 * //TO DO: //check if x and y have same dimensions else //throw and
	 * exception...
	 * 
	 * //fill it for (int i = 0; i < y.length; i++) { dps2D.addPoint();
	 * dps2D.point(i).coordinate(0).setValue(x[i]);
	 * dps2D.point(i).coordinate(1).setValue(y[i]);
	 * dps2D.point(i).coordinate(1).setErrorMinus(err[i]);
	 * dps2D.point(i).coordinate(1).setErrorPlus(err[i]); }
	 * 
	 * IPlotterFactory plotterFactory = analysisFactory.createPlotterFactory();
	 * IPlotter plotter = plotterFactory.create();
	 * 
	 * //plot it plotter.region(0).plot(dps2D, style, "mode=overlay");
	 * 
	 * return plotter; }
	 */
	private static double min(double[] x)
	{
		double minX = x[0];

		for (int i = 1; i < x.length; i++)
		{
			if (minX > x[i])
			{
				minX = x[i];
			}
		}

		return minX;
	}

	public static void show(boolean on)
	{
		display = on;
	}

	private static double max(double[] x)
	{
		double maxX = x[0];

		for (int i = 1; i < x.length; i++)
		{
			if (maxX < x[i])
			{
				maxX = x[i];
			}
		}

		return maxX;
	}

	private static int min(int[] x)
	{
		int minX = x[0];

		for (int i = 1; i < x.length; i++)
		{
			if (minX > x[i])
			{
				minX = x[i];
			}
		}

		return minX;
	}

	private static int max(int[] x)
	{
		int maxX = x[0];

		for (int i = 1; i < x.length; i++)
		{
			if (maxX < x[i])
			{
				maxX = x[i];
			}
		}

		return maxX;
	}

	private static float min(float[] x)
	{
		float minX = x[0];

		for (int i = 1; i < x.length; i++)
		{
			if (minX > x[i])
			{
				minX = x[i];
			}
		}

		return minX;
	}

	private static float max(float[] x)
	{
		float maxX = x[0];

		for (int i = 1; i < x.length; i++)
		{
			if (maxX < x[i])
			{
				maxX = x[i];
			}
		}

		return maxX;
	}
}